"""
The YTFieldData object.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

class YTFieldData(dict):
    """
    A Container object for field data, instead of just having it be a dict.
    """
    pass
