/* Generated by Cython 0.29.20 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "yt/utilities/lib/tsearch.h"
        ],
        "include_dirs": [
            "yt/utilities/lib/"
        ],
        "libraries": [
            "m"
        ],
        "name": "yt.geometry.selection_routines",
        "sources": [
            "yt/geometry/selection_routines.pyx"
        ]
    },
    "module_name": "yt.geometry.selection_routines"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_20"
#define CYTHON_HEX_VERSION 0x001D14F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__yt__geometry__selection_routines
#define __PYX_HAVE_API__yt__geometry__selection_routines
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#include <stdlib.h>
#include <math.h>
#include "tsearch.h"
#include "math.h"
#include "pythread.h"
#include "pystate.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "yt/geometry/selection_routines.pyx",
  "yt/geometry/selection_routines.pxd",
  "stringsource",
  "__init__.pxd",
  "type.pxd",
  "yt/geometry/oct_visitors.pxd",
  "yt/utilities/lib/allocation_container.pxd",
  "yt/geometry/oct_container.pxd",
  "yt/utilities/lib/partitioned_grid.pxd",
  "yt/utilities/lib/image_samplers.pxd",
  "yt/utilities/lib/bitarray.pxd",
};
/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;

/* MemviewSliceStruct.proto */
struct __pyx_memoryview_obj;
typedef struct {
  struct __pyx_memoryview_obj *memview;
  char *data;
  Py_ssize_t shape[8];
  Py_ssize_t strides[8];
  Py_ssize_t suboffsets[8];
} __Pyx_memviewslice;
#define __Pyx_MemoryView_Len(m)  (m.shape[0])

/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __pyx_atomic_int_type int
#if CYTHON_ATOMICS && __GNUC__ >= 4 && (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL >= 2)) &&\
                    !defined(__i386__)
    #define __pyx_atomic_incr_aligned(value, lock) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_aligned(value, lock) __sync_fetch_and_sub(value, 1)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER) && 0
    #include <Windows.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type LONG
    #define __pyx_atomic_incr_aligned(value, lock) InterlockedIncrement(value)
    #define __pyx_atomic_decr_aligned(value, lock) InterlockedDecrement(value)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#elif CYTHON_ATOMICS && (defined(__ICC) || defined(__INTEL_COMPILER)) && 0
    #define __pyx_atomic_incr_aligned(value, lock) _InterlockedIncrement(value)
    #define __pyx_atomic_decr_aligned(value, lock) _InterlockedDecrement(value)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using Intel atomics"
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
typedef volatile __pyx_atomic_int_type __pyx_atomic_int;
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_aligned(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_aligned(__pyx_get_slice_count_pointer(memview), memview->lock)
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif


/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":775
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":776
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":777
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":778
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":782
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":783
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":784
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":785
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":789
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":790
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":799
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":800
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":801
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":803
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":804
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":805
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":807
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":808
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":810
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":811
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":812
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor;
struct __pyx_obj_2yt_8geometry_12oct_visitors_CountTotalOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_CountTotalCells;
struct __pyx_obj_2yt_8geometry_12oct_visitors_MarkOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_MaskOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_IndexOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_ICoordsOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_IResOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_FCoordsOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_FWidthOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_CopyArrayI64;
struct __pyx_obj_2yt_8geometry_12oct_visitors_CopyArrayF64;
struct __pyx_obj_2yt_8geometry_12oct_visitors_IdentifyOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_AssignDomainInd;
struct __pyx_obj_2yt_8geometry_12oct_visitors_FillFileIndicesO;
struct __pyx_obj_2yt_8geometry_12oct_visitors_FillFileIndicesR;
struct __pyx_obj_2yt_8geometry_12oct_visitors_CountByDomain;
struct __pyx_obj_2yt_8geometry_12oct_visitors_StoreOctree;
struct __pyx_obj_2yt_8geometry_12oct_visitors_LoadOctree;
struct __pyx_obj_2yt_9utilities_3lib_20allocation_container_ObjectPool;
struct __pyx_obj_2yt_8geometry_13oct_container_OctObjectPool;
struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer;
struct __pyx_obj_2yt_8geometry_13oct_container_SparseOctreeContainer;
struct __pyx_obj_2yt_8geometry_13oct_container_RAMSESOctreeContainer;
struct __pyx_obj_2yt_9utilities_3lib_16partitioned_grid_PartitionedGrid;
struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_ImageSampler;
struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_ProjectionSampler;
struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_InterpolatedProjectionSampler;
struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_VolumeRenderSampler;
struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_LightSourceRenderSampler;
struct __pyx_obj_2yt_9utilities_3lib_8bitarray_bitarray;
struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject;
struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__;
struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr;
struct __pyx_array_obj;
struct __pyx_MemviewEnum_obj;
struct __pyx_memoryview_obj;
struct __pyx_memoryviewslice_obj;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":814
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":815
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":816
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":818
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_t_2yt_8geometry_12oct_visitors_Oct;
struct __pyx_t_2yt_8geometry_12oct_visitors_OctPadded;

/* "oct_visitors.pxd":19
 * cimport numpy as np
 * 
 * cdef struct Oct             # <<<<<<<<<<<<<<
 * cdef struct Oct:
 *     np.int64_t file_ind     # index with respect to the order in which it was
 */
struct __pyx_t_2yt_8geometry_12oct_visitors_Oct {
  __pyx_t_5numpy_int64_t file_ind;
  __pyx_t_5numpy_int64_t domain_ind;
  __pyx_t_5numpy_int64_t domain;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct **children;
};

/* "oct_visitors.pxd":27
 *     Oct **children          # Up to 8 long
 * 
 * cdef struct OctPadded:             # <<<<<<<<<<<<<<
 *     np.int64_t file_ind
 *     np.int64_t domain_ind
 */
struct __pyx_t_2yt_8geometry_12oct_visitors_OctPadded {
  __pyx_t_5numpy_int64_t file_ind;
  __pyx_t_5numpy_int64_t domain_ind;
  __pyx_t_5numpy_int64_t domain;
  __pyx_t_5numpy_int64_t padding;
};
struct __pyx_t_2yt_8geometry_13grid_visitors_GridTreeNode;
struct __pyx_t_2yt_8geometry_13grid_visitors_GridTreeNodePadded;
struct __pyx_t_2yt_8geometry_13grid_visitors_GridVisitorData;

/* "grid_visitors.pxd":19
 * cimport numpy as np
 * 
 * cdef struct GridTreeNode:             # <<<<<<<<<<<<<<
 *     np.int32_t num_children
 *     np.int32_t level
 */
struct __pyx_t_2yt_8geometry_13grid_visitors_GridTreeNode {
  __pyx_t_5numpy_int32_t num_children;
  __pyx_t_5numpy_int32_t level;
  __pyx_t_5numpy_int64_t index;
  __pyx_t_5numpy_float64_t left_edge[3];
  __pyx_t_5numpy_float64_t right_edge[3];
  struct __pyx_t_2yt_8geometry_13grid_visitors_GridTreeNode **children;
  __pyx_t_5numpy_int64_t start_index[3];
  __pyx_t_5numpy_int32_t dims[3];
  __pyx_t_5numpy_float64_t dds[3];
};

/* "grid_visitors.pxd":30
 *     np.float64_t dds[3]
 * 
 * cdef struct GridTreeNodePadded:             # <<<<<<<<<<<<<<
 *     np.int32_t num_children
 *     np.int32_t level
 */
struct __pyx_t_2yt_8geometry_13grid_visitors_GridTreeNodePadded {
  __pyx_t_5numpy_int32_t num_children;
  __pyx_t_5numpy_int32_t level;
  __pyx_t_5numpy_int64_t index;
  __pyx_t_5numpy_float64_t left_edge_x;
  __pyx_t_5numpy_float64_t left_edge_y;
  __pyx_t_5numpy_float64_t left_edge_z;
  __pyx_t_5numpy_float64_t right_edge_x;
  __pyx_t_5numpy_float64_t right_edge_y;
  __pyx_t_5numpy_float64_t right_edge_z;
  __pyx_t_5numpy_int_t children_pointers;
  __pyx_t_5numpy_int64_t start_index_x;
  __pyx_t_5numpy_int64_t start_index_y;
  __pyx_t_5numpy_int64_t start_index_z;
  __pyx_t_5numpy_int32_t dims_x;
  __pyx_t_5numpy_int32_t dims_y;
  __pyx_t_5numpy_int32_t dims_z;
  __pyx_t_5numpy_float64_t dds_x;
  __pyx_t_5numpy_float64_t dds_y;
  __pyx_t_5numpy_float64_t dds_z;
};

/* "grid_visitors.pxd":51
 *     np.float64_t dds_z
 * 
 * cdef struct GridVisitorData:             # <<<<<<<<<<<<<<
 *     GridTreeNode *grid
 *     np.uint64_t index
 */
struct __pyx_t_2yt_8geometry_13grid_visitors_GridVisitorData {
  struct __pyx_t_2yt_8geometry_13grid_visitors_GridTreeNode *grid;
  __pyx_t_5numpy_uint64_t index;
  __pyx_t_5numpy_uint64_t global_index;
  __pyx_t_5numpy_int64_t pos[3];
  int n_tuples;
  int **child_tuples;
  void *array;
  int ref_factor;
};

/* "grid_visitors.pxd":67
 * cdef np.uint8_t check_child_masked(GridVisitorData *data) nogil
 * 
 * ctypedef void grid_visitor_function(GridVisitorData *data,             # <<<<<<<<<<<<<<
 *                                          np.uint8_t selected) nogil
 * # This is similar in spirit to the way oct visitor functions work.  However,
 */
typedef void __pyx_t_2yt_8geometry_13grid_visitors_grid_visitor_function(struct __pyx_t_2yt_8geometry_13grid_visitors_GridVisitorData *, __pyx_t_5numpy_uint8_t);
struct __pyx_t_2yt_9utilities_3lib_20allocation_container_AllocationContainer;
struct __pyx_opt_args_2yt_9utilities_3lib_20allocation_container_10ObjectPool_allocate_objs;

/* "yt/utilities/lib/allocation_container.pxd":19
 * from libc.stdlib cimport malloc, free, realloc
 * 
 * cdef struct AllocationContainer:             # <<<<<<<<<<<<<<
 *     np.uint64_t n
 *     np.uint64_t n_assigned
 */
struct __pyx_t_2yt_9utilities_3lib_20allocation_container_AllocationContainer {
  __pyx_t_5numpy_uint64_t n;
  __pyx_t_5numpy_uint64_t n_assigned;
  __pyx_t_5numpy_uint64_t offset;
  __pyx_t_5numpy_int64_t con_id;
  void *my_objs;
};

/* "yt/utilities/lib/allocation_container.pxd":30
 *     cdef np.uint64_t n_con
 *     cdef AllocationContainer* containers
 *     cdef void allocate_objs(self, int n_objs, np.int64_t con_id = ?) except *             # <<<<<<<<<<<<<<
 *     cdef void setup_objs(self, void *obj, np.uint64_t count,
 *                          np.uint64_t offset, np.int64_t con_id)
 */
struct __pyx_opt_args_2yt_9utilities_3lib_20allocation_container_10ObjectPool_allocate_objs {
  int __pyx_n;
  __pyx_t_5numpy_int64_t con_id;
};
struct __pyx_t_2yt_8geometry_13oct_container_OctKey;
struct __pyx_t_2yt_8geometry_13oct_container_OctInfo;
struct __pyx_t_2yt_8geometry_13oct_container_OctList;
struct __pyx_t_2yt_8geometry_13oct_container_OctAllocationContainer;
struct __pyx_opt_args_2yt_8geometry_13oct_container_15OctreeContainer_get;
struct __pyx_opt_args_2yt_8geometry_13oct_container_15OctreeContainer_visit_all_octs;

/* "oct_container.pxd":30
 * cdef int ORDER_MAX
 * 
 * cdef struct OctKey:             # <<<<<<<<<<<<<<
 *     np.int64_t key
 *     Oct *node
 */
struct __pyx_t_2yt_8geometry_13oct_container_OctKey {
  __pyx_t_5numpy_int64_t key;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *node;
};

/* "oct_container.pxd":34
 *     Oct *node
 * 
 * cdef struct OctInfo:             # <<<<<<<<<<<<<<
 *     np.float64_t left_edge[3]
 *     np.float64_t dds[3]
 */
struct __pyx_t_2yt_8geometry_13oct_container_OctInfo {
  __pyx_t_5numpy_float64_t left_edge[3];
  __pyx_t_5numpy_float64_t dds[3];
  __pyx_t_5numpy_int64_t ipos[3];
  __pyx_t_5numpy_int32_t level;
};

/* "oct_container.pxd":40
 *     np.int32_t level
 * 
 * cdef struct OctList             # <<<<<<<<<<<<<<
 * 
 * cdef struct OctList:
 */
struct __pyx_t_2yt_8geometry_13oct_container_OctList {
  struct __pyx_t_2yt_8geometry_13oct_container_OctList *next;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *o;
};

/* "oct_container.pxd":48
 * # NOTE: This object *has* to be the same size as the AllocationContainer
 * # object.  There's an assert in the __cinit__ function.
 * cdef struct OctAllocationContainer:             # <<<<<<<<<<<<<<
 *     np.uint64_t n
 *     np.uint64_t n_assigned
 */
struct __pyx_t_2yt_8geometry_13oct_container_OctAllocationContainer {
  __pyx_t_5numpy_uint64_t n;
  __pyx_t_5numpy_uint64_t n_assigned;
  __pyx_t_5numpy_uint64_t offset;
  __pyx_t_5numpy_int64_t con_id;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *my_objs;
};

/* "oct_container.pxd":74
 *     cdef public np.int64_t nocts
 *     cdef public int num_domains
 *     cdef Oct *get(self, np.float64_t ppos[3], OctInfo *oinfo = ?,             # <<<<<<<<<<<<<<
 *                   int max_level = ?)
 *     cdef int get_root(self, int ind[3], Oct **o)
 */
struct __pyx_opt_args_2yt_8geometry_13oct_container_15OctreeContainer_get {
  int __pyx_n;
  struct __pyx_t_2yt_8geometry_13oct_container_OctInfo *oinfo;
  int max_level;
};

/* "oct_container.pxd":83
 *     # AllocationContainer.offset if such a thing exists.
 *     cdef np.int64_t get_domain_offset(self, int domain_id)
 *     cdef void visit_all_octs(self,             # <<<<<<<<<<<<<<
 *                         selection_routines.SelectorObject selector,
 *                         OctVisitor visitor,
 */
struct __pyx_opt_args_2yt_8geometry_13oct_container_15OctreeContainer_visit_all_octs {
  int __pyx_n;
  int vc;
};
struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer;

/* "yt/utilities/lib/volume_container.pxd":19
 * cimport numpy as np
 * 
 * cdef struct VolumeContainer:             # <<<<<<<<<<<<<<
 *     int n_fields
 *     np.float64_t **data
 */
struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer {
  int n_fields;
  __pyx_t_5numpy_float64_t **data;
  __pyx_t_5numpy_uint8_t *mask;
  __pyx_t_5numpy_float64_t left_edge[3];
  __pyx_t_5numpy_float64_t right_edge[3];
  __pyx_t_5numpy_float64_t dds[3];
  __pyx_t_5numpy_float64_t idds[3];
  int dims[3];
};
struct __pyx_t_2yt_9utilities_3lib_14image_samplers_ImageAccumulator;

/* "image_samplers.pxd":30
 * cdef struct VolumeRenderAccumulator
 * 
 * ctypedef int calculate_extent_function(ImageSampler image,             # <<<<<<<<<<<<<<
 *             VolumeContainer *vc, np.int64_t rv[4]) nogil except -1
 * 
 */
typedef int __pyx_t_2yt_9utilities_3lib_14image_samplers_calculate_extent_function(struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_ImageSampler *, struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer *, __pyx_t_5numpy_int64_t *);

/* "image_samplers.pxd":33
 *             VolumeContainer *vc, np.int64_t rv[4]) nogil except -1
 * 
 * ctypedef void generate_vector_info_function(ImageSampler im,             # <<<<<<<<<<<<<<
 *             np.int64_t vi, np.int64_t vj,
 *             np.float64_t width[2],
 */
typedef void __pyx_t_2yt_9utilities_3lib_14image_samplers_generate_vector_info_function(struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_ImageSampler *, __pyx_t_5numpy_int64_t, __pyx_t_5numpy_int64_t, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *);

/* "image_samplers.pxd":38
 *             np.float64_t v_dir[3], np.float64_t v_pos[3]) nogil
 * 
 * cdef struct ImageAccumulator:             # <<<<<<<<<<<<<<
 *     np.float64_t rgba[Nch]
 *     void *supp_data
 */
struct __pyx_t_2yt_9utilities_3lib_14image_samplers_ImageAccumulator {
  __pyx_t_5numpy_float64_t rgba[4];
  void *supp_data;
};
struct __pyx_opt_args_2yt_9utilities_3lib_14grid_traversal_walk_volume;

/* "yt/utilities/lib/grid_traversal.pxd":23
 * from .volume_container cimport VolumeContainer, vc_index, vc_pos_index
 * 
 * ctypedef void sampler_function(             # <<<<<<<<<<<<<<
 *                 VolumeContainer *vc,
 *                 np.float64_t v_pos[3],
 */
typedef void __pyx_t_2yt_9utilities_3lib_14grid_traversal_sampler_function(struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, int *, void *);

/* "yt/utilities/lib/grid_traversal.pxd":32
 *                 void *data) nogil
 * 
 * cdef int walk_volume(VolumeContainer *vc,             # <<<<<<<<<<<<<<
 *                      np.float64_t v_pos[3],
 *                      np.float64_t v_dir[3],
 */
struct __pyx_opt_args_2yt_9utilities_3lib_14grid_traversal_walk_volume {
  int __pyx_n;
  __pyx_t_5numpy_float64_t *return_t;
  __pyx_t_5numpy_float64_t max_t;
};
struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_recursively_visit_octs;
struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid;
struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_visit_grid_cells;
struct __pyx_t_2yt_8geometry_18selection_routines_IntegrationAccumulator;
struct __pyx_opt_args_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_grid;
struct __pyx_opt_args_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_grid;
struct __pyx_opt_args_2yt_8geometry_18selection_routines_14AlwaysSelector_select_grid;
struct __pyx_opt_args_2yt_8geometry_18selection_routines_15ComposeSelector_select_grid;
struct __pyx_opt_args_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_grid;
struct __pyx_opt_args_2yt_8geometry_18selection_routines_17BooleanORSelector_select_grid;
struct __pyx_opt_args_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_grid;
struct __pyx_opt_args_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_grid;
struct __pyx_opt_args_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_grid;
struct __pyx_opt_args_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_grid;
struct __pyx_opt_args_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_grid;

/* "yt/geometry/selection_routines.pxd":38
 *     cdef np.int64_t _hash
 * 
 *     cdef void recursively_visit_octs(self, Oct *root,             # <<<<<<<<<<<<<<
 *                         np.float64_t pos[3], np.float64_t dds[3],
 *                         int level,
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_recursively_visit_octs {
  int __pyx_n;
  int visit_covered;
};

/* "yt/geometry/selection_routines.pxd":46
 *                               np.float64_t spos[3], np.float64_t sdds[3],
 *                               OctVisitor visitor, int i, int j, int k)
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3],
 *                                np.int32_t level, Oct *o = ?) nogil
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid {
  int __pyx_n;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *o;
};

/* "yt/geometry/selection_routines.pxd":61
 *                                 np.ndarray[np.uint8_t, ndim=3] mask,
 *                                 int level)
 *     cdef void visit_grid_cells(self, GridVisitorData *data,             # <<<<<<<<<<<<<<
 *                     grid_visitor_function *func, np.uint8_t *cached_mask = ?)
 * 
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_visit_grid_cells {
  int __pyx_n;
  __pyx_t_5numpy_uint8_t *cached_mask;
};

/* "yt/geometry/selection_routines.pyx":1381
 * ortho_ray_selector = OrthoRaySelector
 * 
 * cdef struct IntegrationAccumulator:             # <<<<<<<<<<<<<<
 *     np.float64_t *t
 *     np.float64_t *dt
 */
struct __pyx_t_2yt_8geometry_18selection_routines_IntegrationAccumulator {
  __pyx_t_5numpy_float64_t *t;
  __pyx_t_5numpy_float64_t *dt;
  __pyx_t_5numpy_uint8_t *child_mask;
  int hits;
};

/* "yt/geometry/selection_routines.pyx":1852
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_grid {
  int __pyx_n;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *o;
};

/* "yt/geometry/selection_routines.pyx":1932
 *         return self.base_selector.select_bbox(left_edge, right_edge)
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_grid {
  int __pyx_n;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *o;
};

/* "yt/geometry/selection_routines.pyx":1966
 *         return 1
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_14AlwaysSelector_select_grid {
  int __pyx_n;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *o;
};

/* "yt/geometry/selection_routines.pyx":2009
 *             return 0
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_15ComposeSelector_select_grid {
  int __pyx_n;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *o;
};

/* "yt/geometry/selection_routines.pyx":2123
 *         return 1
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_grid {
  int __pyx_n;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *o;
};

/* "yt/geometry/selection_routines.pyx":2167
 *         return 0
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_17BooleanORSelector_select_grid {
  int __pyx_n;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *o;
};

/* "yt/geometry/selection_routines.pyx":2209
 *         return 1
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_grid {
  int __pyx_n;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *o;
};

/* "yt/geometry/selection_routines.pyx":2241
 *         return 1
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_grid {
  int __pyx_n;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *o;
};

/* "yt/geometry/selection_routines.pyx":2277
 *         return self.sel1.select_bbox(left_edge, right_edge)
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_grid {
  int __pyx_n;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *o;
};

/* "yt/geometry/selection_routines.pyx":2334
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_grid {
  int __pyx_n;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *o;
};

/* "yt/geometry/selection_routines.pyx":2400
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_grid {
  int __pyx_n;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *o;
};

/* "oct_visitors.pxd":33
 *     np.int64_t padding
 * 
 * cdef class OctVisitor:             # <<<<<<<<<<<<<<
 *     cdef np.uint64_t index
 *     cdef np.uint64_t last
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor *__pyx_vtab;
  __pyx_t_5numpy_uint64_t index;
  __pyx_t_5numpy_uint64_t last;
  __pyx_t_5numpy_int64_t global_index;
  __pyx_t_5numpy_int64_t pos[3];
  __pyx_t_5numpy_uint8_t ind[3];
  int dims;
  __pyx_t_5numpy_int32_t domain;
  __pyx_t_5numpy_int8_t level;
  __pyx_t_5numpy_int8_t oref;
  __pyx_t_5numpy_int32_t nz;
};


/* "oct_visitors.pxd":59
 *         return (((self.ind[2]*d)+self.ind[1])*d+self.ind[0])
 * 
 * cdef class CountTotalOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_CountTotalOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};


/* "oct_visitors.pxd":62
 *     pass
 * 
 * cdef class CountTotalCells(OctVisitor):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_CountTotalCells {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};


/* "oct_visitors.pxd":65
 *     pass
 * 
 * cdef class MarkOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     # Unused
 *     cdef np.uint8_t[:,:,:,:] mark
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_MarkOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice mark;
};


/* "oct_visitors.pxd":69
 *     cdef np.uint8_t[:,:,:,:] mark
 * 
 * cdef class MaskOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:,:,:,:] mask
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_MaskOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice mask;
};


/* "oct_visitors.pxd":72
 *     cdef np.uint8_t[:,:,:,:] mask
 * 
 * cdef class IndexOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:] oct_index
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_IndexOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice oct_index;
};


/* "oct_visitors.pxd":75
 *     cdef np.int64_t[:] oct_index
 * 
 * cdef class ICoordsOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:,:] icoords
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_ICoordsOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice icoords;
};


/* "oct_visitors.pxd":78
 *     cdef np.int64_t[:,:] icoords
 * 
 * cdef class IResOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:] ires
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_IResOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice ires;
};


/* "oct_visitors.pxd":81
 *     cdef np.int64_t[:] ires
 * 
 * cdef class FCoordsOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t[:,:] fcoords
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_FCoordsOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice fcoords;
};


/* "oct_visitors.pxd":84
 *     cdef np.float64_t[:,:] fcoords
 * 
 * cdef class FWidthOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t[:,:] fwidth
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_FWidthOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice fwidth;
};


/* "oct_visitors.pxd":87
 *     cdef np.float64_t[:,:] fwidth
 * 
 * cdef class CopyArrayI64(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:,:,:,:,:,:] source
 *     cdef np.int64_t[:,:] dest
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_CopyArrayI64 {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice source;
  __Pyx_memviewslice dest;
};


/* "oct_visitors.pxd":91
 *     cdef np.int64_t[:,:] dest
 * 
 * cdef class CopyArrayF64(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t[:,:,:,:,:] source
 *     cdef np.float64_t[:,:] dest
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_CopyArrayF64 {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice source;
  __Pyx_memviewslice dest;
};


/* "oct_visitors.pxd":95
 *     cdef np.float64_t[:,:] dest
 * 
 * cdef class IdentifyOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] domain_mask
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_IdentifyOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice domain_mask;
};


/* "oct_visitors.pxd":98
 *     cdef np.uint8_t[:] domain_mask
 * 
 * cdef class AssignDomainInd(OctVisitor):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_AssignDomainInd {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};


/* "oct_visitors.pxd":101
 *     pass
 * 
 * cdef class FillFileIndicesO(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] levels
 *     cdef np.int64_t[:] file_inds
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_FillFileIndicesO {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice levels;
  __Pyx_memviewslice file_inds;
  __Pyx_memviewslice cell_inds;
};


/* "oct_visitors.pxd":106
 *     cdef np.uint8_t[:] cell_inds
 * 
 * cdef class FillFileIndicesR(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] levels
 *     cdef np.int64_t[:] file_inds
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_FillFileIndicesR {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice levels;
  __Pyx_memviewslice file_inds;
  __Pyx_memviewslice cell_inds;
};


/* "oct_visitors.pxd":111
 *     cdef np.uint8_t[:] cell_inds
 * 
 * cdef class CountByDomain(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:] domain_counts
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_CountByDomain {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice domain_counts;
};


/* "oct_visitors.pxd":114
 *     cdef np.int64_t[:] domain_counts
 * 
 * cdef class StoreOctree(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] ref_mask
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_StoreOctree {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice ref_mask;
};


/* "oct_visitors.pxd":117
 *     cdef np.uint8_t[:] ref_mask
 * 
 * cdef class LoadOctree(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] ref_mask
 *     cdef Oct* octs
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_LoadOctree {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice ref_mask;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *octs;
  __pyx_t_5numpy_uint64_t *nocts;
  __pyx_t_5numpy_uint64_t *nfinest;
};


/* "yt/utilities/lib/allocation_container.pxd":26
 *     void *my_objs
 * 
 * cdef class ObjectPool:             # <<<<<<<<<<<<<<
 *     cdef public np.uint64_t itemsize
 *     cdef np.uint64_t n_con
 */
struct __pyx_obj_2yt_9utilities_3lib_20allocation_container_ObjectPool {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_9utilities_3lib_20allocation_container_ObjectPool *__pyx_vtab;
  __pyx_t_5numpy_uint64_t itemsize;
  __pyx_t_5numpy_uint64_t n_con;
  struct __pyx_t_2yt_9utilities_3lib_20allocation_container_AllocationContainer *containers;
};


/* "oct_container.pxd":55
 *     Oct *my_objs
 * 
 * cdef class OctObjectPool(ObjectPool):             # <<<<<<<<<<<<<<
 *     cdef inline OctAllocationContainer *get_cont(self, int i):
 *         return <OctAllocationContainer*> (&self.containers[i])
 */
struct __pyx_obj_2yt_8geometry_13oct_container_OctObjectPool {
  struct __pyx_obj_2yt_9utilities_3lib_20allocation_container_ObjectPool __pyx_base;
};


/* "oct_container.pxd":63
 * cdef void OctList_delete(OctList *list)
 * 
 * cdef class OctreeContainer:             # <<<<<<<<<<<<<<
 *     cdef public OctObjectPool domains
 *     cdef Oct ****root_mesh
 */
struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctreeContainer *__pyx_vtab;
  struct __pyx_obj_2yt_8geometry_13oct_container_OctObjectPool *domains;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct ****root_mesh;
  int partial_coverage;
  int level_offset;
  int nn[3];
  __pyx_t_5numpy_uint8_t oref;
  __pyx_t_5numpy_float64_t DLE[3];
  __pyx_t_5numpy_float64_t DRE[3];
  __pyx_t_5numpy_int64_t nocts;
  int num_domains;
  PyObject *fill_style;
};


/* "oct_container.pxd":94
 *     cdef public object fill_style
 * 
 * cdef class SparseOctreeContainer(OctreeContainer):             # <<<<<<<<<<<<<<
 *     cdef OctKey *root_nodes
 *     cdef void *tree_root
 */
struct __pyx_obj_2yt_8geometry_13oct_container_SparseOctreeContainer {
  struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer __pyx_base;
  struct __pyx_t_2yt_8geometry_13oct_container_OctKey *root_nodes;
  void *tree_root;
  int num_root;
  int max_root;
};


/* "oct_container.pxd":102
 *     cdef np.int64_t ipos_to_key(self, int pos[3])
 * 
 * cdef class RAMSESOctreeContainer(SparseOctreeContainer):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_2yt_8geometry_13oct_container_RAMSESOctreeContainer {
  struct __pyx_obj_2yt_8geometry_13oct_container_SparseOctreeContainer __pyx_base;
};


/* "partitioned_grid.pxd":22
 * from .volume_container cimport VolumeContainer
 * 
 * cdef class PartitionedGrid:             # <<<<<<<<<<<<<<
 *     cdef public object my_data
 *     cdef public object source_mask
 */
struct __pyx_obj_2yt_9utilities_3lib_16partitioned_grid_PartitionedGrid {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_9utilities_3lib_16partitioned_grid_PartitionedGrid *__pyx_vtab;
  PyObject *my_data;
  PyObject *source_mask;
  PyObject *LeftEdge;
  PyObject *RightEdge;
  int parent_grid_id;
  struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer *container;
  __pyx_t_5numpy_float64_t star_er;
  __pyx_t_5numpy_float64_t star_sigma_num;
  __pyx_t_5numpy_float64_t star_coeff;
};


/* "image_samplers.pxd":42
 *     void *supp_data
 * 
 * cdef class ImageSampler:             # <<<<<<<<<<<<<<
 *     cdef np.float64_t[:,:,:] vp_pos
 *     cdef np.float64_t[:,:,:] vp_dir
 */
struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_ImageSampler {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_ImageSampler *__pyx_vtab;
  __Pyx_memviewslice vp_pos;
  __Pyx_memviewslice vp_dir;
  __pyx_t_5numpy_float64_t *center;
  __Pyx_memviewslice image;
  __Pyx_memviewslice zbuffer;
  __Pyx_memviewslice image_used;
  __Pyx_memviewslice mesh_lines;
  __pyx_t_5numpy_float64_t pdx;
  __pyx_t_5numpy_float64_t pdy;
  __pyx_t_5numpy_float64_t bounds[4];
  __Pyx_memviewslice camera_data;
  int nv[2];
  __pyx_t_5numpy_float64_t *x_vec;
  __pyx_t_5numpy_float64_t *y_vec;
  PyObject *acenter;
  PyObject *aimage;
  PyObject *ax_vec;
  PyObject *ay_vec;
  PyObject *azbuffer;
  PyObject *aimage_used;
  PyObject *amesh_lines;
  void *supp_data;
  __pyx_t_5numpy_float64_t width[3];
  PyObject *lens_type;
  __pyx_t_2yt_9utilities_3lib_14image_samplers_calculate_extent_function *extent_function;
  __pyx_t_2yt_9utilities_3lib_14image_samplers_generate_vector_info_function *vector_function;
};


/* "image_samplers.pxd":75
 *                 void *data) nogil
 * 
 * cdef class ProjectionSampler(ImageSampler):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_ProjectionSampler {
  struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_ImageSampler __pyx_base;
};


/* "image_samplers.pxd":78
 *     pass
 * 
 * cdef class InterpolatedProjectionSampler(ImageSampler):             # <<<<<<<<<<<<<<
 *     cdef VolumeRenderAccumulator *vra
 *     cdef public object tf_obj
 */
struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_InterpolatedProjectionSampler {
  struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_ImageSampler __pyx_base;
  struct __pyx_t_2yt_9utilities_3lib_14image_samplers_VolumeRenderAccumulator *vra;
  PyObject *tf_obj;
  PyObject *my_field_tables;
};


/* "image_samplers.pxd":83
 *     cdef public object my_field_tables
 * 
 * cdef class VolumeRenderSampler(ImageSampler):             # <<<<<<<<<<<<<<
 *     cdef VolumeRenderAccumulator *vra
 *     cdef public object tf_obj
 */
struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_VolumeRenderSampler {
  struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_ImageSampler __pyx_base;
  struct __pyx_t_2yt_9utilities_3lib_14image_samplers_VolumeRenderAccumulator *vra;
  PyObject *tf_obj;
  PyObject *my_field_tables;
  PyObject *tree_containers;
};


/* "image_samplers.pxd":89
 *     cdef object tree_containers
 * 
 * cdef class LightSourceRenderSampler(ImageSampler):             # <<<<<<<<<<<<<<
 *     cdef VolumeRenderAccumulator *vra
 *     cdef public object tf_obj
 */
struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_LightSourceRenderSampler {
  struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_ImageSampler __pyx_base;
  struct __pyx_t_2yt_9utilities_3lib_14image_samplers_VolumeRenderAccumulator *vra;
  PyObject *tf_obj;
  PyObject *my_field_tables;
};


/* "yt/utilities/lib/bitarray.pxd":30
 *     return 1
 * 
 * cdef class bitarray:             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t *buf
 *     cdef np.uint64_t size
 */
struct __pyx_obj_2yt_9utilities_3lib_8bitarray_bitarray {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_9utilities_3lib_8bitarray_bitarray *__pyx_vtab;
  __pyx_t_5numpy_uint8_t *buf;
  __pyx_t_5numpy_uint64_t size;
  __pyx_t_5numpy_uint64_t buf_size;
  PyObject *ibuf;
};


/* "yt/geometry/selection_routines.pxd":29
 *     return arr
 * 
 * cdef class SelectorObject:             # <<<<<<<<<<<<<<
 *     cdef public np.int32_t min_level
 *     cdef public np.int32_t max_level
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *__pyx_vtab;
  __pyx_t_5numpy_int32_t min_level;
  __pyx_t_5numpy_int32_t max_level;
  int overlap_cells;
  __pyx_t_5numpy_float64_t domain_width[3];
  int periodicity[3];
  int _hash_initialized;
  __pyx_t_5numpy_int64_t _hash;
};


/* "yt/geometry/selection_routines.pxd":67
 *     cdef np.float64_t difference(self, np.float64_t x1, np.float64_t x2, int d) nogil
 * 
 * cdef class AlwaysSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};


/* "yt/geometry/selection_routines.pxd":70
 *     pass
 * 
 * cdef class OctreeSubsetSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef SelectorObject base_selector
 *     cdef public np.int64_t domain_id
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *base_selector;
  __pyx_t_5numpy_int64_t domain_id;
};


/* "yt/geometry/selection_routines.pxd":74
 *     cdef public np.int64_t domain_id
 * 
 * cdef class BooleanSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef public SelectorObject sel1
 *     cdef public SelectorObject sel2
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *sel1;
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *sel2;
};


/* "yt/geometry/selection_routines.pyx":641
 * 
 * 
 * cdef class PointSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t p[3]
 * 
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  __pyx_t_5numpy_float64_t p[3];
};


/* "yt/geometry/selection_routines.pyx":698
 * 
 * 
 * cdef class SphereSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t radius
 *     cdef np.float64_t radius2
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  __pyx_t_5numpy_float64_t radius;
  __pyx_t_5numpy_float64_t radius2;
  __pyx_t_5numpy_float64_t center[3];
  __pyx_t_5numpy_float64_t bbox[3][2];
  int check_box[3];
};


/* "yt/geometry/selection_routines.pyx":800
 * sphere_selector = SphereSelector
 * 
 * cdef class RegionSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t left_edge[3]
 *     cdef np.float64_t right_edge[3]
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  __pyx_t_5numpy_float64_t left_edge[3];
  __pyx_t_5numpy_float64_t right_edge[3];
  __pyx_t_5numpy_float64_t right_edge_shift[3];
  int loose_selection;
  int check_period;
};


/* "yt/geometry/selection_routines.pyx":961
 * region_selector = RegionSelector
 * 
 * cdef class CutRegionSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef set _positions
 *     cdef tuple _conditionals
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  PyObject *_positions;
  PyObject *_conditionals;
};


/* "yt/geometry/selection_routines.pyx":995
 * cut_region_selector = CutRegionSelector
 * 
 * cdef class DiskSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t norm_vec[3]
 *     cdef np.float64_t center[3]
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  __pyx_t_5numpy_float64_t norm_vec[3];
  __pyx_t_5numpy_float64_t center[3];
  __pyx_t_5numpy_float64_t radius;
  __pyx_t_5numpy_float64_t radius2;
  __pyx_t_5numpy_float64_t height;
};


/* "yt/geometry/selection_routines.pyx":1110
 * disk_selector = DiskSelector
 * 
 * cdef class CuttingPlaneSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t norm_vec[3]
 *     cdef np.float64_t d
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  __pyx_t_5numpy_float64_t norm_vec[3];
  __pyx_t_5numpy_float64_t d;
};


/* "yt/geometry/selection_routines.pyx":1190
 * cutting_selector = CuttingPlaneSelector
 * 
 * cdef class SliceSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef int axis
 *     cdef np.float64_t coord
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  int axis;
  __pyx_t_5numpy_float64_t coord;
  int ax;
  int ay;
};


/* "yt/geometry/selection_routines.pyx":1279
 * slice_selector = SliceSelector
 * 
 * cdef class OrthoRaySelector(SelectorObject):             # <<<<<<<<<<<<<<
 * 
 *     cdef np.uint8_t px_ax
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  __pyx_t_5numpy_uint8_t px_ax;
  __pyx_t_5numpy_uint8_t py_ax;
  __pyx_t_5numpy_float64_t px;
  __pyx_t_5numpy_float64_t py;
  int axis;
};


/* "yt/geometry/selection_routines.pyx":1403
 *     am.dt[di] = (exit_t - enter_t)
 * 
 * cdef class RaySelector(SelectorObject):             # <<<<<<<<<<<<<<
 * 
 *     cdef np.float64_t p1[3]
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  __pyx_t_5numpy_float64_t p1[3];
  __pyx_t_5numpy_float64_t p2[3];
  __pyx_t_5numpy_float64_t vec[3];
};


/* "yt/geometry/selection_routines.pyx":1623
 * ray_selector = RaySelector
 * 
 * cdef class DataCollectionSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef object obj_ids
 *     cdef np.int64_t nids
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  PyObject *obj_ids;
  __pyx_t_5numpy_int64_t nids;
};


/* "yt/geometry/selection_routines.pyx":1660
 * data_collection_selector = DataCollectionSelector
 * 
 * cdef class EllipsoidSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t vec[3][3]
 *     cdef np.float64_t mag[3]
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  __pyx_t_5numpy_float64_t vec[3][3];
  __pyx_t_5numpy_float64_t mag[3];
  __pyx_t_5numpy_float64_t center[3];
};


/* "yt/geometry/selection_routines.pyx":1767
 * ellipsoid_selector = EllipsoidSelector
 * 
 * cdef class GridSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef object ind
 * 
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  PyObject *ind;
};


/* "yt/geometry/selection_routines.pyx":1870
 * octree_subset_selector = OctreeSubsetSelector
 * 
 * cdef class IndexedOctreeSubsetSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     # This is a numpy array, which will be a bool of ndim 1
 *     cdef np.uint64_t min_ind
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  __pyx_t_5numpy_uint64_t min_ind;
  __pyx_t_5numpy_uint64_t max_ind;
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *base_selector;
  int filter_bbox;
  __pyx_t_5numpy_float64_t DLE[3];
  __pyx_t_5numpy_float64_t DRE[3];
};


/* "yt/geometry/selection_routines.pyx":1986
 * always_selector = AlwaysSelector
 * 
 * cdef class ComposeSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef SelectorObject selector1
 *     cdef SelectorObject selector2
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *selector1;
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *selector2;
};


/* "yt/geometry/selection_routines.pyx":2045
 * compose_selector = ComposeSelector
 * 
 * cdef class HaloParticlesSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef public object base_source
 *     cdef SelectorObject base_selector
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  PyObject *base_source;
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *base_selector;
  PyObject *pind;
  __pyx_t_5numpy_int64_t halo_id;
};


/* "yt/geometry/selection_routines.pyx":2114
 *             self.sel2 = dobj.dobj2.selector
 * 
 * cdef class BooleanANDSelector(BooleanSelector):             # <<<<<<<<<<<<<<
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector __pyx_base;
};


/* "yt/geometry/selection_routines.pyx":2158
 *                 self.sel2._hash_vals())
 * 
 * cdef class BooleanORSelector(BooleanSelector):             # <<<<<<<<<<<<<<
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector __pyx_base;
};


/* "yt/geometry/selection_routines.pyx":2202
 *                 self.sel2._hash_vals())
 * 
 * cdef class BooleanNOTSelector(BooleanSelector):             # <<<<<<<<<<<<<<
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector __pyx_base;
};


/* "yt/geometry/selection_routines.pyx":2233
 *                 ("not",))
 * 
 * cdef class BooleanXORSelector(BooleanSelector):             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector __pyx_base;
};


/* "yt/geometry/selection_routines.pyx":2269
 *                 self.sel2._hash_vals())
 * 
 * cdef class BooleanNEGSelector(BooleanSelector):             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector __pyx_base;
};


/* "yt/geometry/selection_routines.pyx":2308
 *                 self.sel2._hash_vals())
 * 
 * cdef class ChainedBooleanSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef int n_obj
 *     cdef np.ndarray selectors
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  int n_obj;
  PyArrayObject *selectors;
};


/* "yt/geometry/selection_routines.pyx":2318
 *             self.selectors[i] = dobj.data_objects[i].selector
 * 
 * cdef class ChainedBooleanANDSelector(ChainedBooleanSelector):             # <<<<<<<<<<<<<<
 *     @cython.cdivision(True)
 *     @cython.boundscheck(False)
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector __pyx_base;
};


/* "yt/geometry/selection_routines.pyx":2384
 * intersection_selector = ChainedBooleanANDSelector
 * 
 * cdef class ChainedBooleanORSelector(ChainedBooleanSelector):             # <<<<<<<<<<<<<<
 *     @cython.cdivision(True)
 *     @cython.boundscheck(False)
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector __pyx_base;
};


/* "yt/geometry/selection_routines.pyx":965
 *     cdef tuple _conditionals
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         positions = np.array([dobj['x'], dobj['y'], dobj['z']]).T
 *         self._conditionals = tuple(dobj.conditionals)
 */
struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ {
  PyObject_HEAD
  PyObject *__pyx_v_positions;
};


/* "yt/geometry/selection_routines.pyx":968
 *         positions = np.array([dobj['x'], dobj['y'], dobj['z']]).T
 *         self._conditionals = tuple(dobj.conditionals)
 *         self._positions = set(tuple(position) for position in positions)             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_bbox(self,  np.float64_t left_edge[3],
 */
struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ *__pyx_outer_scope;
  PyObject *__pyx_v_position;
};


/* "View.MemoryView":105
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_array_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_array *__pyx_vtab;
  char *data;
  Py_ssize_t len;
  char *format;
  int ndim;
  Py_ssize_t *_shape;
  Py_ssize_t *_strides;
  Py_ssize_t itemsize;
  PyObject *mode;
  PyObject *_format;
  void (*callback_free_data)(void *);
  int free_data;
  int dtype_is_object;
};


/* "View.MemoryView":279
 * 
 * @cname('__pyx_MemviewEnum')
 * cdef class Enum(object):             # <<<<<<<<<<<<<<
 *     cdef object name
 *     def __init__(self, name):
 */
struct __pyx_MemviewEnum_obj {
  PyObject_HEAD
  PyObject *name;
};


/* "View.MemoryView":330
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */
struct __pyx_memoryview_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_memoryview *__pyx_vtab;
  PyObject *obj;
  PyObject *_size;
  PyObject *_array_interface;
  PyThread_type_lock lock;
  __pyx_atomic_int acquisition_count[2];
  __pyx_atomic_int *acquisition_count_aligned_p;
  Py_buffer view;
  int flags;
  int dtype_is_object;
  __Pyx_TypeInfo *typeinfo;
};


/* "View.MemoryView":965
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */
struct __pyx_memoryviewslice_obj {
  struct __pyx_memoryview_obj __pyx_base;
  __Pyx_memviewslice from_slice;
  PyObject *from_object;
  PyObject *(*to_object_func)(char *);
  int (*to_dtype_func)(char *, PyObject *);
};



/* "oct_visitors.pxd":33
 *     np.int64_t padding
 * 
 * cdef class OctVisitor:             # <<<<<<<<<<<<<<
 *     cdef np.uint64_t index
 *     cdef np.uint64_t last
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor {
  void (*visit)(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_uint8_t);
  int (*oind)(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *);
  int (*rind)(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *);
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor *__pyx_vtabptr_2yt_8geometry_12oct_visitors_OctVisitor;
static CYTHON_INLINE int __pyx_f_2yt_8geometry_12oct_visitors_10OctVisitor_oind(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *);
static CYTHON_INLINE int __pyx_f_2yt_8geometry_12oct_visitors_10OctVisitor_rind(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *);


/* "oct_visitors.pxd":59
 *         return (((self.ind[2]*d)+self.ind[1])*d+self.ind[0])
 * 
 * cdef class CountTotalOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountTotalOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountTotalOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_CountTotalOcts;


/* "oct_visitors.pxd":62
 *     pass
 * 
 * cdef class CountTotalCells(OctVisitor):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountTotalCells {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountTotalCells *__pyx_vtabptr_2yt_8geometry_12oct_visitors_CountTotalCells;


/* "oct_visitors.pxd":65
 *     pass
 * 
 * cdef class MarkOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     # Unused
 *     cdef np.uint8_t[:,:,:,:] mark
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MarkOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MarkOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_MarkOcts;


/* "oct_visitors.pxd":69
 *     cdef np.uint8_t[:,:,:,:] mark
 * 
 * cdef class MaskOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:,:,:,:] mask
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MaskOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MaskOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_MaskOcts;


/* "oct_visitors.pxd":72
 *     cdef np.uint8_t[:,:,:,:] mask
 * 
 * cdef class IndexOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:] oct_index
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IndexOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IndexOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_IndexOcts;


/* "oct_visitors.pxd":75
 *     cdef np.int64_t[:] oct_index
 * 
 * cdef class ICoordsOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:,:] icoords
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_ICoordsOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_ICoordsOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_ICoordsOcts;


/* "oct_visitors.pxd":78
 *     cdef np.int64_t[:,:] icoords
 * 
 * cdef class IResOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:] ires
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IResOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IResOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_IResOcts;


/* "oct_visitors.pxd":81
 *     cdef np.int64_t[:] ires
 * 
 * cdef class FCoordsOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t[:,:] fcoords
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FCoordsOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FCoordsOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_FCoordsOcts;


/* "oct_visitors.pxd":84
 *     cdef np.float64_t[:,:] fcoords
 * 
 * cdef class FWidthOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t[:,:] fwidth
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FWidthOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FWidthOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_FWidthOcts;


/* "oct_visitors.pxd":87
 *     cdef np.float64_t[:,:] fwidth
 * 
 * cdef class CopyArrayI64(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:,:,:,:,:,:] source
 *     cdef np.int64_t[:,:] dest
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CopyArrayI64 {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CopyArrayI64 *__pyx_vtabptr_2yt_8geometry_12oct_visitors_CopyArrayI64;


/* "oct_visitors.pxd":91
 *     cdef np.int64_t[:,:] dest
 * 
 * cdef class CopyArrayF64(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t[:,:,:,:,:] source
 *     cdef np.float64_t[:,:] dest
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CopyArrayF64 {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CopyArrayF64 *__pyx_vtabptr_2yt_8geometry_12oct_visitors_CopyArrayF64;


/* "oct_visitors.pxd":95
 *     cdef np.float64_t[:,:] dest
 * 
 * cdef class IdentifyOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] domain_mask
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IdentifyOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IdentifyOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_IdentifyOcts;


/* "oct_visitors.pxd":98
 *     cdef np.uint8_t[:] domain_mask
 * 
 * cdef class AssignDomainInd(OctVisitor):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_AssignDomainInd {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_AssignDomainInd *__pyx_vtabptr_2yt_8geometry_12oct_visitors_AssignDomainInd;


/* "oct_visitors.pxd":101
 *     pass
 * 
 * cdef class FillFileIndicesO(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] levels
 *     cdef np.int64_t[:] file_inds
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FillFileIndicesO {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FillFileIndicesO *__pyx_vtabptr_2yt_8geometry_12oct_visitors_FillFileIndicesO;


/* "oct_visitors.pxd":106
 *     cdef np.uint8_t[:] cell_inds
 * 
 * cdef class FillFileIndicesR(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] levels
 *     cdef np.int64_t[:] file_inds
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FillFileIndicesR {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FillFileIndicesR *__pyx_vtabptr_2yt_8geometry_12oct_visitors_FillFileIndicesR;


/* "oct_visitors.pxd":111
 *     cdef np.uint8_t[:] cell_inds
 * 
 * cdef class CountByDomain(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:] domain_counts
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountByDomain {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountByDomain *__pyx_vtabptr_2yt_8geometry_12oct_visitors_CountByDomain;


/* "oct_visitors.pxd":114
 *     cdef np.int64_t[:] domain_counts
 * 
 * cdef class StoreOctree(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] ref_mask
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_StoreOctree {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_StoreOctree *__pyx_vtabptr_2yt_8geometry_12oct_visitors_StoreOctree;


/* "oct_visitors.pxd":117
 *     cdef np.uint8_t[:] ref_mask
 * 
 * cdef class LoadOctree(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] ref_mask
 *     cdef Oct* octs
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_LoadOctree {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_LoadOctree *__pyx_vtabptr_2yt_8geometry_12oct_visitors_LoadOctree;


/* "yt/utilities/lib/allocation_container.pxd":26
 *     void *my_objs
 * 
 * cdef class ObjectPool:             # <<<<<<<<<<<<<<
 *     cdef public np.uint64_t itemsize
 *     cdef np.uint64_t n_con
 */

struct __pyx_vtabstruct_2yt_9utilities_3lib_20allocation_container_ObjectPool {
  void (*allocate_objs)(struct __pyx_obj_2yt_9utilities_3lib_20allocation_container_ObjectPool *, int, struct __pyx_opt_args_2yt_9utilities_3lib_20allocation_container_10ObjectPool_allocate_objs *__pyx_optional_args);
  void (*setup_objs)(struct __pyx_obj_2yt_9utilities_3lib_20allocation_container_ObjectPool *, void *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_int64_t);
  void (*teardown_objs)(struct __pyx_obj_2yt_9utilities_3lib_20allocation_container_ObjectPool *, void *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_int64_t);
};
static struct __pyx_vtabstruct_2yt_9utilities_3lib_20allocation_container_ObjectPool *__pyx_vtabptr_2yt_9utilities_3lib_20allocation_container_ObjectPool;


/* "oct_container.pxd":55
 *     Oct *my_objs
 * 
 * cdef class OctObjectPool(ObjectPool):             # <<<<<<<<<<<<<<
 *     cdef inline OctAllocationContainer *get_cont(self, int i):
 *         return <OctAllocationContainer*> (&self.containers[i])
 */

struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctObjectPool {
  struct __pyx_vtabstruct_2yt_9utilities_3lib_20allocation_container_ObjectPool __pyx_base;
  struct __pyx_t_2yt_8geometry_13oct_container_OctAllocationContainer *(*get_cont)(struct __pyx_obj_2yt_8geometry_13oct_container_OctObjectPool *, int);
};
static struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctObjectPool *__pyx_vtabptr_2yt_8geometry_13oct_container_OctObjectPool;
static CYTHON_INLINE struct __pyx_t_2yt_8geometry_13oct_container_OctAllocationContainer *__pyx_f_2yt_8geometry_13oct_container_13OctObjectPool_get_cont(struct __pyx_obj_2yt_8geometry_13oct_container_OctObjectPool *, int);


/* "oct_container.pxd":63
 * cdef void OctList_delete(OctList *list)
 * 
 * cdef class OctreeContainer:             # <<<<<<<<<<<<<<
 *     cdef public OctObjectPool domains
 *     cdef Oct ****root_mesh
 */

struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctreeContainer {
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *(*get)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, __pyx_t_5numpy_float64_t *, struct __pyx_opt_args_2yt_8geometry_13oct_container_15OctreeContainer_get *__pyx_optional_args);
  int (*get_root)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, int *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct **);
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct **(*neighbors)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, struct __pyx_t_2yt_8geometry_13oct_container_OctInfo *, __pyx_t_5numpy_int64_t *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, int *);
  void (*oct_bounds)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *);
  __pyx_t_5numpy_int64_t (*get_domain_offset)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, int);
  void (*visit_all_octs)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *, struct __pyx_opt_args_2yt_8geometry_13oct_container_15OctreeContainer_visit_all_octs *__pyx_optional_args);
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *(*next_root)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, int, int *);
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *(*next_child)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, int, int *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *);
  void (*append_domain)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, __pyx_t_5numpy_int64_t);
};
static struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctreeContainer *__pyx_vtabptr_2yt_8geometry_13oct_container_OctreeContainer;


/* "oct_container.pxd":94
 *     cdef public object fill_style
 * 
 * cdef class SparseOctreeContainer(OctreeContainer):             # <<<<<<<<<<<<<<
 *     cdef OctKey *root_nodes
 *     cdef void *tree_root
 */

struct __pyx_vtabstruct_2yt_8geometry_13oct_container_SparseOctreeContainer {
  struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctreeContainer __pyx_base;
  void (*key_to_ipos)(struct __pyx_obj_2yt_8geometry_13oct_container_SparseOctreeContainer *, __pyx_t_5numpy_int64_t, __pyx_t_5numpy_int64_t *);
  __pyx_t_5numpy_int64_t (*ipos_to_key)(struct __pyx_obj_2yt_8geometry_13oct_container_SparseOctreeContainer *, int *);
};
static struct __pyx_vtabstruct_2yt_8geometry_13oct_container_SparseOctreeContainer *__pyx_vtabptr_2yt_8geometry_13oct_container_SparseOctreeContainer;


/* "oct_container.pxd":102
 *     cdef np.int64_t ipos_to_key(self, int pos[3])
 * 
 * cdef class RAMSESOctreeContainer(SparseOctreeContainer):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_13oct_container_RAMSESOctreeContainer {
  struct __pyx_vtabstruct_2yt_8geometry_13oct_container_SparseOctreeContainer __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_13oct_container_RAMSESOctreeContainer *__pyx_vtabptr_2yt_8geometry_13oct_container_RAMSESOctreeContainer;


/* "partitioned_grid.pxd":22
 * from .volume_container cimport VolumeContainer
 * 
 * cdef class PartitionedGrid:             # <<<<<<<<<<<<<<
 *     cdef public object my_data
 *     cdef public object source_mask
 */

struct __pyx_vtabstruct_2yt_9utilities_3lib_16partitioned_grid_PartitionedGrid {
  void (*get_vector_field)(struct __pyx_obj_2yt_9utilities_3lib_16partitioned_grid_PartitionedGrid *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *);
};
static struct __pyx_vtabstruct_2yt_9utilities_3lib_16partitioned_grid_PartitionedGrid *__pyx_vtabptr_2yt_9utilities_3lib_16partitioned_grid_PartitionedGrid;


/* "image_samplers.pxd":42
 *     void *supp_data
 * 
 * cdef class ImageSampler:             # <<<<<<<<<<<<<<
 *     cdef np.float64_t[:,:,:] vp_pos
 *     cdef np.float64_t[:,:,:] vp_dir
 */

struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_ImageSampler {
  void (*setup)(struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_ImageSampler *, struct __pyx_obj_2yt_9utilities_3lib_16partitioned_grid_PartitionedGrid *);
  void (*sample)(struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, int *, void *);
};
static struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_ImageSampler *__pyx_vtabptr_2yt_9utilities_3lib_14image_samplers_ImageSampler;


/* "image_samplers.pxd":75
 *                 void *data) nogil
 * 
 * cdef class ProjectionSampler(ImageSampler):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_ProjectionSampler {
  struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_ImageSampler __pyx_base;
};
static struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_ProjectionSampler *__pyx_vtabptr_2yt_9utilities_3lib_14image_samplers_ProjectionSampler;


/* "image_samplers.pxd":78
 *     pass
 * 
 * cdef class InterpolatedProjectionSampler(ImageSampler):             # <<<<<<<<<<<<<<
 *     cdef VolumeRenderAccumulator *vra
 *     cdef public object tf_obj
 */

struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_InterpolatedProjectionSampler {
  struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_ImageSampler __pyx_base;
};
static struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_InterpolatedProjectionSampler *__pyx_vtabptr_2yt_9utilities_3lib_14image_samplers_InterpolatedProjectionSampler;


/* "image_samplers.pxd":83
 *     cdef public object my_field_tables
 * 
 * cdef class VolumeRenderSampler(ImageSampler):             # <<<<<<<<<<<<<<
 *     cdef VolumeRenderAccumulator *vra
 *     cdef public object tf_obj
 */

struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_VolumeRenderSampler {
  struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_ImageSampler __pyx_base;
};
static struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_VolumeRenderSampler *__pyx_vtabptr_2yt_9utilities_3lib_14image_samplers_VolumeRenderSampler;


/* "image_samplers.pxd":89
 *     cdef object tree_containers
 * 
 * cdef class LightSourceRenderSampler(ImageSampler):             # <<<<<<<<<<<<<<
 *     cdef VolumeRenderAccumulator *vra
 *     cdef public object tf_obj
 */

struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_LightSourceRenderSampler {
  struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_ImageSampler __pyx_base;
};
static struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_LightSourceRenderSampler *__pyx_vtabptr_2yt_9utilities_3lib_14image_samplers_LightSourceRenderSampler;


/* "yt/utilities/lib/bitarray.pxd":30
 *     return 1
 * 
 * cdef class bitarray:             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t *buf
 *     cdef np.uint64_t size
 */

struct __pyx_vtabstruct_2yt_9utilities_3lib_8bitarray_bitarray {
  void (*_set_value)(struct __pyx_obj_2yt_9utilities_3lib_8bitarray_bitarray *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint8_t);
  __pyx_t_5numpy_uint8_t (*_query_value)(struct __pyx_obj_2yt_9utilities_3lib_8bitarray_bitarray *, __pyx_t_5numpy_uint64_t);
};
static struct __pyx_vtabstruct_2yt_9utilities_3lib_8bitarray_bitarray *__pyx_vtabptr_2yt_9utilities_3lib_8bitarray_bitarray;


/* "yt/geometry/selection_routines.pyx":113
 *         raise RuntimeError
 * 
 * cdef class SelectorObject:             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, dobj, *args):
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject {
  void (*recursively_visit_octs)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, int, struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_recursively_visit_octs *__pyx_optional_args);
  void (*visit_oct_cells)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *, int, int, int);
  int (*select_grid)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int32_t, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid *__pyx_optional_args);
  int (*select_cell)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *);
  int (*select_point)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *);
  int (*select_sphere)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t);
  int (*select_bbox)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *);
  int (*fill_mask_selector)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, int *, PyArrayObject *, PyArrayObject *, int);
  void (*visit_grid_cells)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, struct __pyx_t_2yt_8geometry_13grid_visitors_GridVisitorData *, __pyx_t_2yt_8geometry_13grid_visitors_grid_visitor_function *, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_visit_grid_cells *__pyx_optional_args);
  __pyx_t_5numpy_float64_t (*difference)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, int);
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;


/* "yt/geometry/selection_routines.pyx":1944
 * indexed_octree_subset_selector = IndexedOctreeSubsetSelector
 * 
 * cdef class AlwaysSelector(SelectorObject):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, dobj):
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_AlwaysSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_AlwaysSelector;


/* "yt/geometry/selection_routines.pyx":1806
 * grid_selector = GridSelector
 * 
 * cdef class OctreeSubsetSelector(SelectorObject):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, dobj):
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_OctreeSubsetSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_OctreeSubsetSelector;


/* "yt/geometry/selection_routines.pyx":2100
 *     return mask
 * 
 * cdef class BooleanSelector(SelectorObject):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, dobj):
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanSelector;


/* "yt/geometry/selection_routines.pyx":641
 * 
 * 
 * cdef class PointSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t p[3]
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_PointSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_PointSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_PointSelector;


/* "yt/geometry/selection_routines.pyx":698
 * 
 * 
 * cdef class SphereSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t radius
 *     cdef np.float64_t radius2
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SphereSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SphereSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_SphereSelector;


/* "yt/geometry/selection_routines.pyx":800
 * sphere_selector = SphereSelector
 * 
 * cdef class RegionSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t left_edge[3]
 *     cdef np.float64_t right_edge[3]
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_RegionSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_RegionSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_RegionSelector;


/* "yt/geometry/selection_routines.pyx":961
 * region_selector = RegionSelector
 * 
 * cdef class CutRegionSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef set _positions
 *     cdef tuple _conditionals
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_CutRegionSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_CutRegionSelector;


/* "yt/geometry/selection_routines.pyx":995
 * cut_region_selector = CutRegionSelector
 * 
 * cdef class DiskSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t norm_vec[3]
 *     cdef np.float64_t center[3]
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_DiskSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_DiskSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_DiskSelector;


/* "yt/geometry/selection_routines.pyx":1110
 * disk_selector = DiskSelector
 * 
 * cdef class CuttingPlaneSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t norm_vec[3]
 *     cdef np.float64_t d
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_CuttingPlaneSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_CuttingPlaneSelector;


/* "yt/geometry/selection_routines.pyx":1190
 * cutting_selector = CuttingPlaneSelector
 * 
 * cdef class SliceSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef int axis
 *     cdef np.float64_t coord
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SliceSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SliceSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_SliceSelector;


/* "yt/geometry/selection_routines.pyx":1279
 * slice_selector = SliceSelector
 * 
 * cdef class OrthoRaySelector(SelectorObject):             # <<<<<<<<<<<<<<
 * 
 *     cdef np.uint8_t px_ax
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_OrthoRaySelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_OrthoRaySelector;


/* "yt/geometry/selection_routines.pyx":1403
 *     am.dt[di] = (exit_t - enter_t)
 * 
 * cdef class RaySelector(SelectorObject):             # <<<<<<<<<<<<<<
 * 
 *     cdef np.float64_t p1[3]
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_RaySelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_RaySelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_RaySelector;


/* "yt/geometry/selection_routines.pyx":1623
 * ray_selector = RaySelector
 * 
 * cdef class DataCollectionSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef object obj_ids
 *     cdef np.int64_t nids
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_DataCollectionSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_DataCollectionSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_DataCollectionSelector;


/* "yt/geometry/selection_routines.pyx":1660
 * data_collection_selector = DataCollectionSelector
 * 
 * cdef class EllipsoidSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t vec[3][3]
 *     cdef np.float64_t mag[3]
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_EllipsoidSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_EllipsoidSelector;


/* "yt/geometry/selection_routines.pyx":1767
 * ellipsoid_selector = EllipsoidSelector
 * 
 * cdef class GridSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef object ind
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_GridSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_GridSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_GridSelector;


/* "yt/geometry/selection_routines.pyx":1870
 * octree_subset_selector = OctreeSubsetSelector
 * 
 * cdef class IndexedOctreeSubsetSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     # This is a numpy array, which will be a bool of ndim 1
 *     cdef np.uint64_t min_ind
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector;


/* "yt/geometry/selection_routines.pyx":1986
 * always_selector = AlwaysSelector
 * 
 * cdef class ComposeSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef SelectorObject selector1
 *     cdef SelectorObject selector2
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_ComposeSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_ComposeSelector;


/* "yt/geometry/selection_routines.pyx":2045
 * compose_selector = ComposeSelector
 * 
 * cdef class HaloParticlesSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef public object base_source
 *     cdef SelectorObject base_selector
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_HaloParticlesSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_HaloParticlesSelector;


/* "yt/geometry/selection_routines.pyx":2114
 *             self.sel2 = dobj.dobj2.selector
 * 
 * cdef class BooleanANDSelector(BooleanSelector):             # <<<<<<<<<<<<<<
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanANDSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanSelector __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanANDSelector;


/* "yt/geometry/selection_routines.pyx":2158
 *                 self.sel2._hash_vals())
 * 
 * cdef class BooleanORSelector(BooleanSelector):             # <<<<<<<<<<<<<<
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanORSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanSelector __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanORSelector;


/* "yt/geometry/selection_routines.pyx":2202
 *                 self.sel2._hash_vals())
 * 
 * cdef class BooleanNOTSelector(BooleanSelector):             # <<<<<<<<<<<<<<
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanNOTSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanSelector __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanNOTSelector;


/* "yt/geometry/selection_routines.pyx":2233
 *                 ("not",))
 * 
 * cdef class BooleanXORSelector(BooleanSelector):             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanXORSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanSelector __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanXORSelector;


/* "yt/geometry/selection_routines.pyx":2269
 *                 self.sel2._hash_vals())
 * 
 * cdef class BooleanNEGSelector(BooleanSelector):             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanNEGSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanSelector __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanNEGSelector;


/* "yt/geometry/selection_routines.pyx":2308
 *                 self.sel2._hash_vals())
 * 
 * cdef class ChainedBooleanSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef int n_obj
 *     cdef np.ndarray selectors
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_ChainedBooleanSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_ChainedBooleanSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_ChainedBooleanSelector;


/* "yt/geometry/selection_routines.pyx":2318
 *             self.selectors[i] = dobj.data_objects[i].selector
 * 
 * cdef class ChainedBooleanANDSelector(ChainedBooleanSelector):             # <<<<<<<<<<<<<<
 *     @cython.cdivision(True)
 *     @cython.boundscheck(False)
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_ChainedBooleanSelector __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector;


/* "yt/geometry/selection_routines.pyx":2384
 * intersection_selector = ChainedBooleanANDSelector
 * 
 * cdef class ChainedBooleanORSelector(ChainedBooleanSelector):             # <<<<<<<<<<<<<<
 *     @cython.cdivision(True)
 *     @cython.boundscheck(False)
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_ChainedBooleanORSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_ChainedBooleanSelector __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_ChainedBooleanORSelector;


/* "View.MemoryView":105
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_array {
  PyObject *(*get_memview)(struct __pyx_array_obj *);
};
static struct __pyx_vtabstruct_array *__pyx_vtabptr_array;


/* "View.MemoryView":330
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

struct __pyx_vtabstruct_memoryview {
  char *(*get_item_pointer)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*is_slice)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_slice_assignment)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*setitem_slice_assign_scalar)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_indexed)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*convert_item_to_object)(struct __pyx_memoryview_obj *, char *);
  PyObject *(*assign_item_from_object)(struct __pyx_memoryview_obj *, char *, PyObject *);
};
static struct __pyx_vtabstruct_memoryview *__pyx_vtabptr_memoryview;


/* "View.MemoryView":965
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

struct __pyx_vtabstruct__memoryviewslice {
  struct __pyx_vtabstruct_memoryview __pyx_base;
};
static struct __pyx_vtabstruct__memoryviewslice *__pyx_vtabptr__memoryviewslice;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* BufferGetAndValidate.proto */
#define __Pyx_GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack)\
    ((obj == Py_None || obj == NULL) ?\
    (__Pyx_ZeroBuffer(buf), 0) :\
    __Pyx__GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack))
static int  __Pyx__GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static void __Pyx_ZeroBuffer(Py_buffer* buf);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static Py_ssize_t __Pyx_minusones[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static Py_ssize_t __Pyx_zeros[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* BufferFallbackError.proto */
static void __Pyx_RaiseBufferFallbackError(void);

#define __Pyx_BufPtrStrided3d(type, buf, i0, s0, i1, s1, i2, s2) (type)((char*)buf + i0 * s0 + i1 * s1 + i2 * s2)
#define __Pyx_BufPtrStrided2d(type, buf, i0, s0, i1, s1) (type)((char*)buf + i0 * s0 + i1 * s1)
/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

#define __Pyx_BufPtrStrided1d(type, buf, i0, s0) (type)((char*)buf + i0 * s0)
/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* BufferIndexError.proto */
static void __Pyx_RaiseBufferIndexError(int axis);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* MemviewSliceInit.proto */
#define __Pyx_BUF_MAX_NDIMS %(BUF_MAX_NDIMS)d
#define __Pyx_MEMVIEW_DIRECT   1
#define __Pyx_MEMVIEW_PTR      2
#define __Pyx_MEMVIEW_FULL     4
#define __Pyx_MEMVIEW_CONTIG   8
#define __Pyx_MEMVIEW_STRIDED  16
#define __Pyx_MEMVIEW_FOLLOW   32
#define __Pyx_IS_C_CONTIG 1
#define __Pyx_IS_F_CONTIG 2
static int __Pyx_init_memviewslice(
                struct __pyx_memoryview_obj *memview,
                int ndim,
                __Pyx_memviewslice *memviewslice,
                int memview_is_new_reference);
static CYTHON_INLINE int __pyx_add_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
static CYTHON_INLINE int __pyx_sub_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
#define __pyx_get_slice_count_pointer(memview) (memview->acquisition_count_aligned_p)
#define __pyx_get_slice_count(memview) (*__pyx_get_slice_count_pointer(memview))
#define __PYX_INC_MEMVIEW(slice, have_gil) __Pyx_INC_MEMVIEW(slice, have_gil, __LINE__)
#define __PYX_XDEC_MEMVIEW(slice, have_gil) __Pyx_XDEC_MEMVIEW(slice, have_gil, __LINE__)
static CYTHON_INLINE void __Pyx_INC_MEMVIEW(__Pyx_memviewslice *, int, int);
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *, int, int);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* UnicodeAsUCS4.proto */
static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject*);

/* object_ord.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Ord(c)\
    (likely(PyUnicode_Check(c)) ? (long)__Pyx_PyUnicode_AsPy_UCS4(c) : __Pyx__PyObject_Ord(c))
#else
#define __Pyx_PyObject_Ord(c) __Pyx__PyObject_Ord(c)
#endif
static long __Pyx__PyObject_Ord(PyObject* c);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PySetContains.proto */
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* None.proto */
static CYTHON_INLINE long __Pyx_mod_long(long, long);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_DivideCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_DivideCObj(op1, op2, floatval, inplace, zerodivision_check)\
    ((inplace ? __Pyx_PyNumber_InPlaceDivide(op1, op2) : __Pyx_PyNumber_Divide(op1, op2)))
    #endif

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* None.proto */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t, Py_ssize_t);

/* UnaryNegOverflows.proto */
#define UNARY_NEG_WOULD_OVERFLOW(x)\
        (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *); /*proto*/
/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* None.proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_dealloc);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* FusedFunction.proto */
typedef struct {
    __pyx_CyFunctionObject func;
    PyObject *__signatures__;
    PyObject *type;
    PyObject *self;
} __pyx_FusedFunctionObject;
static PyObject *__pyx_FusedFunction_New(PyMethodDef *ml, int flags,
                                         PyObject *qualname, PyObject *closure,
                                         PyObject *module, PyObject *globals,
                                         PyObject *code);
static int __pyx_FusedFunction_clear(__pyx_FusedFunctionObject *self);
static PyTypeObject *__pyx_FusedFunctionType = NULL;
static int __pyx_FusedFunction_init(void);
#define __Pyx_FusedFunction_USED

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

/* MemviewSliceIsContig.proto */
static int __pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim);

/* OverlappingSlices.proto */
static int __pyx_slices_overlap(__Pyx_memviewslice *slice1,
                                __Pyx_memviewslice *slice2,
                                int ndim, size_t itemsize);

/* Capsule.proto */
static CYTHON_INLINE PyObject *__pyx_capsule_create(void *p, const char *sig);

/* TypeInfoCompare.proto */
static int __pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b);

/* MemviewSliceValidateAndInit.proto */
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_float(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_double(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(PyObject *, int writable_flag);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int64(npy_int64 value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint64(npy_uint64 value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int32(npy_int32 value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint8(npy_uint8 value);

/* Print.proto */
static int __Pyx_Print(PyObject*, PyObject *, int);
#if CYTHON_COMPILING_IN_PYPY || PY_MAJOR_VERSION >= 3
static PyObject* __pyx_print = 0;
static PyObject* __pyx_print_kwargs = 0;
#endif

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_float64_t(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_float64_t(const char *itemp, PyObject *obj);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value);

/* MemviewSliceCopyTemplate.proto */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_int64 __Pyx_PyInt_As_npy_int64(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_int32 __Pyx_PyInt_As_npy_int32(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_uint8 __Pyx_PyInt_As_npy_uint8(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_As_Py_intptr_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_uint64 __Pyx_PyInt_As_npy_uint64(PyObject *);

/* BytesContains.proto */
static CYTHON_INLINE int __Pyx_BytesContains(PyObject* bytes, char character);

/* ImportNumPyArray.proto */
static PyObject *__pyx_numpy_ndarray = NULL;
static PyObject* __Pyx_ImportNumPyArrayTypeIfAvailable(void);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_unsigned_char(PyObject *, int writable_flag);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static void __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_recursively_visit_octs(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_root, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds, int __pyx_v_level, struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *__pyx_v_visitor, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_recursively_visit_octs *__pyx_optional_args); /* proto*/
static void __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_visit_oct_cells(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_root, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_ch, __pyx_t_5numpy_float64_t *__pyx_v_spos, __pyx_t_5numpy_float64_t *__pyx_v_sdds, struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *__pyx_v_visitor, int __pyx_v_i, int __pyx_v_j, int __pyx_v_k); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid *__pyx_optional_args); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_select_cell(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_select_sphere(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_select_bbox(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_difference(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, __pyx_t_5numpy_float64_t __pyx_v_x1, __pyx_t_5numpy_float64_t __pyx_v_x2, int __pyx_v_d); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_fill_mask_selector(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_float64_t *__pyx_v_dds, int *__pyx_v_dim, PyArrayObject *__pyx_v_child_mask, PyArrayObject *__pyx_v_mask, int __pyx_v_level); /* proto*/
static void __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_visit_grid_cells(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, struct __pyx_t_2yt_8geometry_13grid_visitors_GridVisitorData *__pyx_v_data, __pyx_t_2yt_8geometry_13grid_visitors_grid_visitor_function *__pyx_v_func, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_visit_grid_cells *__pyx_optional_args); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_13PointSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_13PointSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_13PointSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14SphereSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14SphereSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14SphereSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14SphereSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14RegionSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14RegionSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14RegionSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14RegionSelector_fill_mask_selector(struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_float64_t *__pyx_v_dds, int *__pyx_v_dim, PyArrayObject *__pyx_v_child_mask, PyArrayObject *__pyx_v_mask, int __pyx_v_level); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_17CutRegionSelector_select_bbox(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_17CutRegionSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_17CutRegionSelector_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_17CutRegionSelector_select_sphere(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_12DiskSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_12DiskSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_12DiskSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_12DiskSelector_select_bbox(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_13SliceSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_13SliceSelector_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_13SliceSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_13SliceSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_16OrthoRaySelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_16OrthoRaySelector_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_16OrthoRaySelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_16OrthoRaySelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_11RaySelector_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_11RaySelector_select_sphere(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_11RaySelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_11RaySelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_17EllipsoidSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_17EllipsoidSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_17EllipsoidSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_17EllipsoidSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_12GridSelector_select_cell(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_12GridSelector_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_sphere(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_cell(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_grid *__pyx_optional_args); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_sphere(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_cell(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_grid *__pyx_optional_args); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14AlwaysSelector_select_cell(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14AlwaysSelector_select_grid(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge, CYTHON_UNUSED __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14AlwaysSelector_select_grid *__pyx_optional_args); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14AlwaysSelector_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14AlwaysSelector_select_sphere(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_14AlwaysSelector_select_bbox(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_15ComposeSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_15ComposeSelector_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_15ComposeSelector_select_grid *__pyx_optional_args); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_15ComposeSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_15ComposeSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_15ComposeSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_grid *__pyx_optional_args); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_17BooleanORSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_17BooleanORSelector_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_17BooleanORSelector_select_grid *__pyx_optional_args); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_17BooleanORSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_17BooleanORSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_17BooleanORSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_bbox(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_grid(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge, CYTHON_UNUSED __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_grid *__pyx_optional_args); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_bbox(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_grid(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge, CYTHON_UNUSED __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_grid *__pyx_optional_args); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_grid *__pyx_optional_args); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_grid *__pyx_optional_args); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_grid *__pyx_optional_args); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos); /* proto*/
static int __pyx_f_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto*/
static CYTHON_INLINE int __pyx_f_2yt_8geometry_12oct_visitors_10OctVisitor_oind(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_2yt_8geometry_12oct_visitors_10OctVisitor_rind(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *__pyx_v_self); /* proto*/
static CYTHON_INLINE struct __pyx_t_2yt_8geometry_13oct_container_OctAllocationContainer *__pyx_f_2yt_8geometry_13oct_container_13OctObjectPool_get_cont(struct __pyx_obj_2yt_8geometry_13oct_container_OctObjectPool *__pyx_v_self, int __pyx_v_i); /* proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self); /* proto*/
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto*/
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/

/* Module declarations from 'yt.geometry.oct_visitors' */
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_OctVisitor = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalCells = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_MarkOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_MaskOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_IndexOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_ICoordsOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_IResOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_FCoordsOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_FWidthOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_CopyArrayI64 = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_CopyArrayF64 = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_IdentifyOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_AssignDomainInd = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_FillFileIndicesO = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_FillFileIndicesR = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_CountByDomain = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_StoreOctree = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_LoadOctree = 0;
static CYTHON_INLINE int __pyx_f_2yt_8geometry_12oct_visitors_cind(int, int, int); /*proto*/

/* Module declarations from 'yt.geometry.grid_visitors' */
static __pyx_t_5numpy_uint8_t (*__pyx_f_2yt_8geometry_13grid_visitors_check_child_masked)(struct __pyx_t_2yt_8geometry_13grid_visitors_GridVisitorData *); /*proto*/

/* Module declarations from 'cython.view' */

/* Module declarations from 'cython' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'yt.utilities.lib.fnv_hash' */
static __pyx_t_5numpy_int64_t (*__pyx_f_2yt_9utilities_3lib_8fnv_hash_c_fnv_hash)(__Pyx_memviewslice); /*proto*/

/* Module declarations from 'yt.utilities.lib.fp_utils' */
static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_imax(__pyx_t_5numpy_int64_t, __pyx_t_5numpy_int64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_fmax(__pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_imin(__pyx_t_5numpy_int64_t, __pyx_t_5numpy_int64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_fmin(__pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_iclip(__pyx_t_5numpy_int64_t, __pyx_t_5numpy_int64_t, __pyx_t_5numpy_int64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_fclip(__pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t); /*proto*/

/* Module declarations from 'libc.math' */

/* Module declarations from 'yt.utilities.lib.allocation_container' */
static PyTypeObject *__pyx_ptype_2yt_9utilities_3lib_20allocation_container_ObjectPool = 0;

/* Module declarations from 'yt.geometry.oct_container' */
static PyTypeObject *__pyx_ptype_2yt_8geometry_13oct_container_OctObjectPool = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_13oct_container_RAMSESOctreeContainer = 0;
static int *__pyx_vp_2yt_8geometry_13oct_container_ORDER_MAX = 0;
#define __pyx_v_2yt_8geometry_13oct_container_ORDER_MAX (*__pyx_vp_2yt_8geometry_13oct_container_ORDER_MAX)

/* Module declarations from 'yt.utilities.lib.volume_container' */
static CYTHON_INLINE int __pyx_f_2yt_9utilities_3lib_16volume_container_vc_index(struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer *, int, int, int); /*proto*/
static CYTHON_INLINE int __pyx_f_2yt_9utilities_3lib_16volume_container_vc_pos_index(struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer *, __pyx_t_5numpy_float64_t *); /*proto*/

/* Module declarations from 'yt.utilities.lib.partitioned_grid' */
static PyTypeObject *__pyx_ptype_2yt_9utilities_3lib_16partitioned_grid_PartitionedGrid = 0;

/* Module declarations from 'yt.utilities.lib.image_samplers' */
static PyTypeObject *__pyx_ptype_2yt_9utilities_3lib_14image_samplers_ImageSampler = 0;
static PyTypeObject *__pyx_ptype_2yt_9utilities_3lib_14image_samplers_ProjectionSampler = 0;
static PyTypeObject *__pyx_ptype_2yt_9utilities_3lib_14image_samplers_InterpolatedProjectionSampler = 0;
static PyTypeObject *__pyx_ptype_2yt_9utilities_3lib_14image_samplers_VolumeRenderSampler = 0;
static PyTypeObject *__pyx_ptype_2yt_9utilities_3lib_14image_samplers_LightSourceRenderSampler = 0;

/* Module declarations from 'yt.utilities.lib.grid_traversal' */
static int (*__pyx_f_2yt_9utilities_3lib_14grid_traversal_walk_volume)(struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_2yt_9utilities_3lib_14grid_traversal_sampler_function *, void *, struct __pyx_opt_args_2yt_9utilities_3lib_14grid_traversal_walk_volume *__pyx_optional_args); /*proto*/

/* Module declarations from 'yt.utilities.lib.bitarray' */
static PyTypeObject *__pyx_ptype_2yt_9utilities_3lib_8bitarray_bitarray = 0;
static CYTHON_INLINE void __pyx_f_2yt_9utilities_3lib_8bitarray_ba_set_value(__pyx_t_5numpy_uint8_t *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint8_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_uint8_t __pyx_f_2yt_9utilities_3lib_8bitarray_ba_get_value(__pyx_t_5numpy_uint8_t *, __pyx_t_5numpy_uint64_t); /*proto*/

/* Module declarations from 'yt.geometry.selection_routines' */
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_AlwaysSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_OctreeSubsetSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_BooleanSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_PointSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_SphereSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_RegionSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_CutRegionSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_DiskSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_CuttingPlaneSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_SliceSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_OrthoRaySelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_RaySelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_DataCollectionSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_EllipsoidSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_GridSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_ComposeSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_HaloParticlesSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_BooleanANDSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_BooleanORSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_BooleanNOTSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_BooleanXORSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_BooleanNEGSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_ChainedBooleanSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_ChainedBooleanORSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr = 0;
static PyTypeObject *__pyx_array_type = 0;
static PyTypeObject *__pyx_MemviewEnum_type = 0;
static PyTypeObject *__pyx_memoryview_type = 0;
static PyTypeObject *__pyx_memoryviewslice_type = 0;
static __pyx_t_5numpy_float64_t __pyx_v_2yt_8geometry_18selection_routines_grid_eps;
static PyObject *generic = 0;
static PyObject *strided = 0;
static PyObject *indirect = 0;
static PyObject *contiguous = 0;
static PyObject *indirect_contiguous = 0;
static int __pyx_memoryview_thread_locks_used;
static PyThread_type_lock __pyx_memoryview_thread_locks[8];
static CYTHON_INLINE PyObject *__pyx_f_2yt_8geometry_18selection_routines__ensure_code(PyObject *); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_18selection_routines__periodic_dist(__pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, int); /*proto*/
static void __pyx_f_2yt_8geometry_18selection_routines_dt_sampler(struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, int *, void *); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_2yt_8geometry_18selection_routines__mask_fill(PyArrayObject *, __pyx_t_5numpy_int64_t, PyArrayObject *, PyArrayObject *); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_2yt_8geometry_18selection_routines__mask_fill(PyArrayObject *, __pyx_t_5numpy_int64_t, PyArrayObject *, PyArrayObject *); /*proto*/
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/
static void *__pyx_align_pointer(void *, size_t); /*proto*/
static PyObject *__pyx_memoryview_new(PyObject *, int, int, __Pyx_TypeInfo *); /*proto*/
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *); /*proto*/
static PyObject *_unellipsify(PyObject *, int); /*proto*/
static PyObject *assert_direct_dimensions(Py_ssize_t *, int); /*proto*/
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *, PyObject *); /*proto*/
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int, int); /*proto*/
static char *__pyx_pybuffer_index(Py_buffer *, char *, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memslice_transpose(__Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice, int, PyObject *(*)(char *), int (*)(char *, PyObject *), int); /*proto*/
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *); /*proto*/
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static Py_ssize_t abs_py_ssize_t(Py_ssize_t); /*proto*/
static char __pyx_get_best_slice_order(__Pyx_memviewslice *, int); /*proto*/
static void _copy_strided_to_strided(char *, Py_ssize_t *, char *, Py_ssize_t *, Py_ssize_t *, Py_ssize_t *, int, size_t); /*proto*/
static void copy_strided_to_strided(__Pyx_memviewslice *, __Pyx_memviewslice *, int, size_t); /*proto*/
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *, int); /*proto*/
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *, Py_ssize_t *, Py_ssize_t, int, char); /*proto*/
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *, __Pyx_memviewslice *, char, int); /*proto*/
static int __pyx_memoryview_err_extents(int, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memoryview_err_dim(PyObject *, char *, int); /*proto*/
static int __pyx_memoryview_err(PyObject *, char *); /*proto*/
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice, __Pyx_memviewslice, int, int, int); /*proto*/
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *, int, int); /*proto*/
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *, int, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *, int, size_t, void *, int); /*proto*/
static void __pyx_memoryview__slice_assign_scalar(char *, Py_ssize_t *, Py_ssize_t *, int, size_t, void *); /*proto*/
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *, PyObject *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t = { "uint8_t", NULL, sizeof(__pyx_t_5numpy_uint8_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_uint8_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_uint8_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t = { "int64_t", NULL, sizeof(__pyx_t_5numpy_int64_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_int64_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_int64_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t = { "float64_t", NULL, sizeof(__pyx_t_5numpy_float64_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_float = { "float", NULL, sizeof(float), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_double = { "double", NULL, sizeof(double), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t = { "int32_t", NULL, sizeof(__pyx_t_5numpy_int32_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_int32_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_int32_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_unsigned_char = { "unsigned char", NULL, sizeof(unsigned char), { 0 }, 0, IS_UNSIGNED(unsigned char) ? 'U' : 'I', IS_UNSIGNED(unsigned char), 0 };
#define __Pyx_MODULE_NAME "yt.geometry.selection_routines"
extern int __pyx_module_is_main_yt__geometry__selection_routines;
int __pyx_module_is_main_yt__geometry__selection_routines = 0;

/* Implementation of 'yt.geometry.selection_routines' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_Ellipsis;
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_IndexError;
static const char __pyx_k_[] = "()";
static const char __pyx_k_A[] = "_A";
static const char __pyx_k_B[] = "_B";
static const char __pyx_k_C[] = "_C";
static const char __pyx_k_O[] = "O";
static const char __pyx_k_T[] = "T";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_d[] = "d";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_j[] = "j";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_p[] = "p";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k_z[] = "z";
static const char __pyx_k__2[] = "|";
static const char __pyx_k_cx[] = "cx";
static const char __pyx_k_cy[] = "cy";
static const char __pyx_k_cz[] = "cz";
static const char __pyx_k_ds[] = "ds";
static const char __pyx_k_dx[] = "dx";
static const char __pyx_k_dy[] = "dy";
static const char __pyx_k_dz[] = "dz";
static const char __pyx_k_e0[] = "_e0";
static const char __pyx_k_e1[] = "_e1";
static const char __pyx_k_e2[] = "_e2";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_nz[] = "nz";
static const char __pyx_k_or[] = "or";
static const char __pyx_k_px[] = "px";
static const char __pyx_k_py[] = "py";
static const char __pyx_k_pz[] = "pz";
static const char __pyx_k_and[] = "and";
static const char __pyx_k_d_2[] = "_d";
static const char __pyx_k_dds[] = "dds";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_eps[] = "eps";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_n_c[] = "n_c";
static const char __pyx_k_n_p[] = "n_p";
static const char __pyx_k_neg[] = "neg";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_not[] = "not";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_out[] = "out";
static const char __pyx_k_p_0[] = "p[0]";
static const char __pyx_k_p_1[] = "p[1]";
static const char __pyx_k_p_2[] = "p[2]";
static const char __pyx_k_pos[] = "pos";
static const char __pyx_k_vec[] = "vec";
static const char __pyx_k_xor[] = "xor";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_axis[] = "axis";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_bool[] = "bool";
static const char __pyx_k_cpos[] = "cpos";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_dobj[] = "dobj";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_fmod[] = "fmod";
static const char __pyx_k_kind[] = "kind";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mask[] = "mask";
static const char __pyx_k_mesh[] = "mesh";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_ones[] = "ones";
static const char __pyx_k_p1_0[] = "p1[0]";
static const char __pyx_k_p1_1[] = "p1[1]";
static const char __pyx_k_p1_2[] = "p1[2]";
static const char __pyx_k_p2_0[] = "p2[0]";
static const char __pyx_k_p2_1[] = "p2[1]";
static const char __pyx_k_p2_2[] = "p2[2]";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_vals[] = "vals";
static const char __pyx_k_view[] = "view";
static const char __pyx_k_ASCII[] = "ASCII";
static const char __pyx_k_Level[] = "Level";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_ascii[] = "ascii";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_coord[] = "coord";
static const char __pyx_k_count[] = "count";
static const char __pyx_k_dobj1[] = "dobj1";
static const char __pyx_k_dobj2[] = "dobj2";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_finfo[] = "finfo";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_float[] = "float";
static const char __pyx_k_int64[] = "int64";
static const char __pyx_k_mag_0[] = "mag[0]";
static const char __pyx_k_mag_1[] = "mag[1]";
static const char __pyx_k_mag_2[] = "mag[2]";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_px_ax[] = "px_ax";
static const char __pyx_k_py_ax[] = "py_ax";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_uint8[] = "uint8";
static const char __pyx_k_units[] = "units";
static const char __pyx_k_vec_0[] = "vec[0]";
static const char __pyx_k_vec_1[] = "vec[1]";
static const char __pyx_k_vec_2[] = "vec[2]";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_always[] = "always";
static const char __pyx_k_astype[] = "astype";
static const char __pyx_k_center[] = "center";
static const char __pyx_k_double[] = "double";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_extend[] = "extend";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_height[] = "height";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_levels[] = "levels";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_octree[] = "octree";
static const char __pyx_k_offset[] = "offset";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_radius[] = "radius";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_float32[] = "float32";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_fortran[] = "fortran";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_indices[] = "indices";
static const char __pyx_k_max_ind[] = "max_ind";
static const char __pyx_k_memview[] = "memview";
static const char __pyx_k_min_ind[] = "min_ind";
static const char __pyx_k_obj_ids[] = "_obj_ids";
static const char __pyx_k_radius2[] = "radius2";
static const char __pyx_k_vec_0_0[] = "vec[0][0]";
static const char __pyx_k_vec_0_1[] = "vec[0][1]";
static const char __pyx_k_vec_0_2[] = "vec[0][2]";
static const char __pyx_k_vec_1_0[] = "vec[1][0]";
static const char __pyx_k_vec_1_1[] = "vec[1][1]";
static const char __pyx_k_vec_1_2[] = "vec[1][2]";
static const char __pyx_k_vec_2_0[] = "vec[2][0]";
static const char __pyx_k_vec_2_1[] = "vec[2][1]";
static const char __pyx_k_vec_2_2[] = "vec[2][2]";
static const char __pyx_k_Ellipsis[] = "Ellipsis";
static const char __pyx_k_LeftEdge[] = "LeftEdge";
static const char __pyx_k_center_0[] = "center[0]";
static const char __pyx_k_center_1[] = "center[1]";
static const char __pyx_k_center_2[] = "center[2]";
static const char __pyx_k_defaults[] = "defaults";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_norm_vec[] = "_norm_vec";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_selector[] = "selector";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_tostring[] = "tostring";
static const char __pyx_k_RightEdge[] = "RightEdge";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_base_hash[] = "_base_hash";
static const char __pyx_k_domain_id[] = "domain_id";
static const char __pyx_k_end_point[] = "end_point";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_hash_vals[] = "_hash_vals";
static const char __pyx_k_id_offset[] = "_id_offset";
static const char __pyx_k_left_edge[] = "left_edge";
static const char __pyx_k_mask_fill[] = "mask_fill";
static const char __pyx_k_max_level[] = "max_level";
static const char __pyx_k_min_level[] = "min_level";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_selector1[] = "selector1";
static const char __pyx_k_selector2[] = "selector2";
static const char __pyx_k_transpose[] = "transpose";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_chained_or[] = "chained_or";
static const char __pyx_k_child_mask[] = "child_mask";
static const char __pyx_k_left_edges[] = "left_edges";
static const char __pyx_k_logical_or[] = "logical_or";
static const char __pyx_k_norm_vec_0[] = "norm_vec[0]";
static const char __pyx_k_norm_vec_1[] = "norm_vec[1]";
static const char __pyx_k_norm_vec_2[] = "norm_vec[2]";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_right_edge[] = "right_edge";
static const char __pyx_k_to_ndarray[] = "to_ndarray";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_RaySelector[] = "RaySelector";
static const char __pyx_k_base_source[] = "base_source";
static const char __pyx_k_chained_and[] = "chained_and";
static const char __pyx_k_code_length[] = "code_length";
static const char __pyx_k_filter_bbox[] = "filter_bbox";
static const char __pyx_k_left_edge_0[] = "left_edge[0]";
static const char __pyx_k_left_edge_1[] = "left_edge[1]";
static const char __pyx_k_left_edge_2[] = "left_edge[2]";
static const char __pyx_k_periodicity[] = "periodicity";
static const char __pyx_k_right_edges[] = "right_edges";
static const char __pyx_k_start_point[] = "start_point";
static const char __pyx_k_DiskSelector[] = "DiskSelector";
static const char __pyx_k_GridSelector[] = "GridSelector";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_conditionals[] = "conditionals";
static const char __pyx_k_count_points[] = "count_points";
static const char __pyx_k_data_objects[] = "data_objects";
static const char __pyx_k_domain_width[] = "domain_width";
static const char __pyx_k_index_offset[] = "_index_offset";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_ray_selector[] = "ray_selector";
static const char __pyx_k_right_edge_0[] = "right_edge[0]";
static const char __pyx_k_right_edge_1[] = "right_edge[1]";
static const char __pyx_k_right_edge_2[] = "right_edge[2]";
static const char __pyx_k_select_grids[] = "select_grids";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_PointSelector[] = "PointSelector";
static const char __pyx_k_SliceSelector[] = "SliceSelector";
static const char __pyx_k_base_selector[] = "base_selector";
static const char __pyx_k_conditional_s[] = "conditional[%s]";
static const char __pyx_k_disk_selector[] = "disk_selector";
static const char __pyx_k_grid_selector[] = "grid_selector";
static const char __pyx_k_overlap_cells[] = "overlap_cells";
static const char __pyx_k_periodicity_0[] = "periodicity[0]";
static const char __pyx_k_periodicity_1[] = "periodicity[1]";
static const char __pyx_k_periodicity_2[] = "periodicity[2]";
static const char __pyx_k_pyx_getbuffer[] = "__pyx_getbuffer";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_select_points[] = "select_points";
static const char __pyx_k_AlwaysSelector[] = "AlwaysSelector";
static const char __pyx_k_RegionSelector[] = "RegionSelector";
static const char __pyx_k_SelectorObject[] = "SelectorObject";
static const char __pyx_k_SphereSelector[] = "SphereSelector";
static const char __pyx_k_domain_width_0[] = "domain_width[0]";
static const char __pyx_k_domain_width_1[] = "domain_width[1]";
static const char __pyx_k_domain_width_2[] = "domain_width[2]";
static const char __pyx_k_halo_particles[] = "halo_particles";
static const char __pyx_k_point_selector[] = "point_selector";
static const char __pyx_k_slice_selector[] = "slice_selector";
static const char __pyx_k_union_selector[] = "union_selector";
static const char __pyx_k_BooleanSelector[] = "BooleanSelector";
static const char __pyx_k_ComposeSelector[] = "ComposeSelector";
static const char __pyx_k_View_MemoryView[] = "View.MemoryView";
static const char __pyx_k_allocate_buffer[] = "allocate_buffer";
static const char __pyx_k_always_selector[] = "always_selector";
static const char __pyx_k_dtype_is_object[] = "dtype_is_object";
static const char __pyx_k_loose_selection[] = "loose_selection";
static const char __pyx_k_points_in_cells[] = "points_in_cells";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_region_selector[] = "region_selector";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_sphere_selector[] = "sphere_selector";
static const char __pyx_k_ActiveDimensions[] = "ActiveDimensions";
static const char __pyx_k_OrthoRaySelector[] = "OrthoRaySelector";
static const char __pyx_k_compose_selector[] = "compose_selector";
static const char __pyx_k_convert_to_units[] = "convert_to_units";
static const char __pyx_k_cutting_selector[] = "cutting_selector";
static const char __pyx_k_domain_left_edge[] = "domain_left_edge";
static const char __pyx_k_particle_indices[] = "particle_indices";
static const char __pyx_k_BooleanORSelector[] = "BooleanORSelector";
static const char __pyx_k_CutRegionSelector[] = "CutRegionSelector";
static const char __pyx_k_EllipsoidSelector[] = "EllipsoidSelector";
static const char __pyx_k_domain_right_edge[] = "domain_right_edge";
static const char __pyx_k_pyx_unpickle_Enum[] = "__pyx_unpickle_Enum";
static const char __pyx_k_BooleanANDSelector[] = "BooleanANDSelector";
static const char __pyx_k_BooleanNEGSelector[] = "BooleanNEGSelector";
static const char __pyx_k_BooleanNOTSelector[] = "BooleanNOTSelector";
static const char __pyx_k_BooleanXORSelector[] = "BooleanXORSelector";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_ellipsoid_selector[] = "ellipsoid_selector";
static const char __pyx_k_ortho_ray_selector[] = "ortho_ray_selector";
static const char __pyx_k_strided_and_direct[] = "<strided and direct>";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_connectivity_coords[] = "connectivity_coords";
static const char __pyx_k_connectivity_length[] = "_connectivity_length";
static const char __pyx_k_cut_region_selector[] = "cut_region_selector";
static const char __pyx_k_CuttingPlaneSelector[] = "CuttingPlaneSelector";
static const char __pyx_k_OctreeSubsetSelector[] = "OctreeSubsetSelector";
static const char __pyx_k_connectivity_indices[] = "connectivity_indices";
static const char __pyx_k_strided_and_indirect[] = "<strided and indirect>";
static const char __pyx_k_HaloParticlesSelector[] = "HaloParticlesSelector";
static const char __pyx_k_contiguous_and_direct[] = "<contiguous and direct>";
static const char __pyx_k_init___locals_genexpr[] = "__init__.<locals>.genexpr";
static const char __pyx_k_intersection_selector[] = "intersection_selector";
static const char __pyx_k_ChainedBooleanSelector[] = "ChainedBooleanSelector";
static const char __pyx_k_DataCollectionSelector[] = "DataCollectionSelector";
static const char __pyx_k_MemoryView_of_r_object[] = "<MemoryView of %r object>";
static const char __pyx_k_octree_subset_selector[] = "octree_subset_selector";
static const char __pyx_k_MemoryView_of_r_at_0x_x[] = "<MemoryView of %r at 0x%x>";
static const char __pyx_k_contiguous_and_indirect[] = "<contiguous and indirect>";
static const char __pyx_k_convert_mask_to_indices[] = "convert_mask_to_indices";
static const char __pyx_k_halo_particles_selector[] = "halo_particles_selector";
static const char __pyx_k_Cannot_index_with_type_s[] = "Cannot index with type '%s'";
static const char __pyx_k_ChainedBooleanORSelector[] = "ChainedBooleanORSelector";
static const char __pyx_k_data_collection_selector[] = "data_collection_selector";
static const char __pyx_k_ChainedBooleanANDSelector[] = "ChainedBooleanANDSelector";
static const char __pyx_k_Invalid_shape_in_axis_d_d[] = "Invalid shape in axis %d: %d.";
static const char __pyx_k_Geometry_selection_routines[] = "\nGeometry selection routines.\n\n\n\n\n";
static const char __pyx_k_IndexedOctreeSubsetSelector[] = "IndexedOctreeSubsetSelector";
static const char __pyx_k_No_matching_signature_found[] = "No matching signature found";
static const char __pyx_k_SelectorObject_count_points[] = "SelectorObject.count_points";
static const char __pyx_k_itemsize_0_for_cython_array[] = "itemsize <= 0 for cython.array";
static const char __pyx_k_ndarray_is_not_C_contiguous[] = "ndarray is not C contiguous";
static const char __pyx_k_SelectorObject_select_points[] = "SelectorObject.select_points";
static const char __pyx_k_unable_to_allocate_array_data[] = "unable to allocate array data.";
static const char __pyx_k_indexed_octree_subset_selector[] = "indexed_octree_subset_selector";
static const char __pyx_k_strided_and_direct_or_indirect[] = "<strided and direct or indirect>";
static const char __pyx_k_yt_geometry_selection_routines[] = "yt.geometry.selection_routines";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_unknown_dtype_code_in_numpy_pxd[] = "unknown dtype code in numpy.pxd (%d)";
static const char __pyx_k_Buffer_view_does_not_expose_stri[] = "Buffer view does not expose strides";
static const char __pyx_k_Can_only_create_a_buffer_that_is[] = "Can only create a buffer that is contiguous in memory.";
static const char __pyx_k_Cannot_assign_to_read_only_memor[] = "Cannot assign to read-only memoryview";
static const char __pyx_k_Cannot_create_writable_memory_vi[] = "Cannot create writable memory view from read-only memoryview";
static const char __pyx_k_Empty_shape_tuple_for_cython_arr[] = "Empty shape tuple for cython.array";
static const char __pyx_k_Error_yt_attempted_to_read_outsi[] = "Error: yt attempted to read outside the boundaries of a non-periodic domain along dimension %s.\nRegion left edge = %s, Region right edge = %s\nDataset left edge = %s, Dataset right edge = %s\n\nThis commonly happens when trying to compute ghost cells up to the domain boundary. Two possible solutions are to load a smaller region that does not border the edge or override the periodicity for this dataset.";
static const char __pyx_k_Expected_at_least_d_argument_s_g[] = "Expected at least %d argument%s, got %d";
static const char __pyx_k_Format_string_allocated_too_shor[] = "Format string allocated too short, see comment in numpy.pxd";
static const char __pyx_k_Function_call_with_ambiguous_arg[] = "Function call with ambiguous argument types";
static const char __pyx_k_Incompatible_checksums_s_vs_0xb0[] = "Incompatible checksums (%s vs 0xb068931 = (name))";
static const char __pyx_k_Indirect_dimensions_not_supporte[] = "Indirect dimensions not supported";
static const char __pyx_k_Invalid_mode_expected_c_or_fortr[] = "Invalid mode, expected 'c' or 'fortran', got %s";
static const char __pyx_k_Non_native_byte_order_not_suppor[] = "Non-native byte order not supported";
static const char __pyx_k_Out_of_bounds_on_buffer_access_a[] = "Out of bounds on buffer access (axis %d)";
static const char __pyx_k_Region_right_edge_s_left_edge_wi[] = "Region right edge[%s] < left edge: width = %s";
static const char __pyx_k_Unable_to_convert_item_to_object[] = "Unable to convert item to object";
static const char __pyx_k_got_differing_extents_in_dimensi[] = "got differing extents in dimension %d (got %d and %d)";
static const char __pyx_k_ndarray_is_not_Fortran_contiguou[] = "ndarray is not Fortran contiguous";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_unable_to_allocate_shape_and_str[] = "unable to allocate shape and strides.";
static const char __pyx_k_yt_geometry_selection_routines_p[] = "yt/geometry/selection_routines.pyx";
static const char __pyx_k_Format_string_allocated_too_shor_2[] = "Format string allocated too short.";
static PyObject *__pyx_kp_s_;
static PyObject *__pyx_n_s_A;
static PyObject *__pyx_n_s_ASCII;
static PyObject *__pyx_n_s_ActiveDimensions;
static PyObject *__pyx_n_s_AlwaysSelector;
static PyObject *__pyx_n_s_B;
static PyObject *__pyx_n_s_BooleanANDSelector;
static PyObject *__pyx_n_s_BooleanNEGSelector;
static PyObject *__pyx_n_s_BooleanNOTSelector;
static PyObject *__pyx_n_s_BooleanORSelector;
static PyObject *__pyx_n_s_BooleanSelector;
static PyObject *__pyx_n_s_BooleanXORSelector;
static PyObject *__pyx_kp_s_Buffer_view_does_not_expose_stri;
static PyObject *__pyx_n_s_C;
static PyObject *__pyx_kp_s_Can_only_create_a_buffer_that_is;
static PyObject *__pyx_kp_s_Cannot_assign_to_read_only_memor;
static PyObject *__pyx_kp_s_Cannot_create_writable_memory_vi;
static PyObject *__pyx_kp_s_Cannot_index_with_type_s;
static PyObject *__pyx_n_s_ChainedBooleanANDSelector;
static PyObject *__pyx_n_s_ChainedBooleanORSelector;
static PyObject *__pyx_n_s_ChainedBooleanSelector;
static PyObject *__pyx_n_s_ComposeSelector;
static PyObject *__pyx_n_s_CutRegionSelector;
static PyObject *__pyx_n_s_CuttingPlaneSelector;
static PyObject *__pyx_n_s_DataCollectionSelector;
static PyObject *__pyx_n_s_DiskSelector;
static PyObject *__pyx_n_s_Ellipsis;
static PyObject *__pyx_n_s_EllipsoidSelector;
static PyObject *__pyx_kp_s_Empty_shape_tuple_for_cython_arr;
static PyObject *__pyx_kp_s_Error_yt_attempted_to_read_outsi;
static PyObject *__pyx_kp_s_Expected_at_least_d_argument_s_g;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor_2;
static PyObject *__pyx_kp_s_Function_call_with_ambiguous_arg;
static PyObject *__pyx_n_s_GridSelector;
static PyObject *__pyx_n_s_HaloParticlesSelector;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xb0;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_n_s_IndexedOctreeSubsetSelector;
static PyObject *__pyx_kp_s_Indirect_dimensions_not_supporte;
static PyObject *__pyx_kp_s_Invalid_mode_expected_c_or_fortr;
static PyObject *__pyx_kp_s_Invalid_shape_in_axis_d_d;
static PyObject *__pyx_n_s_LeftEdge;
static PyObject *__pyx_n_s_Level;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_kp_s_MemoryView_of_r_at_0x_x;
static PyObject *__pyx_kp_s_MemoryView_of_r_object;
static PyObject *__pyx_kp_s_No_matching_signature_found;
static PyObject *__pyx_kp_u_Non_native_byte_order_not_suppor;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_b_O;
static PyObject *__pyx_n_s_OctreeSubsetSelector;
static PyObject *__pyx_n_s_OrthoRaySelector;
static PyObject *__pyx_kp_s_Out_of_bounds_on_buffer_access_a;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_PointSelector;
static PyObject *__pyx_n_s_RaySelector;
static PyObject *__pyx_n_s_RegionSelector;
static PyObject *__pyx_kp_s_Region_right_edge_s_left_edge_wi;
static PyObject *__pyx_n_s_RightEdge;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_SelectorObject;
static PyObject *__pyx_n_s_SelectorObject_count_points;
static PyObject *__pyx_n_s_SelectorObject_select_points;
static PyObject *__pyx_n_s_SliceSelector;
static PyObject *__pyx_n_s_SphereSelector;
static PyObject *__pyx_n_s_T;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unable_to_convert_item_to_object;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_View_MemoryView;
static PyObject *__pyx_kp_s__2;
static PyObject *__pyx_n_s_allocate_buffer;
static PyObject *__pyx_n_s_always;
static PyObject *__pyx_n_s_always_selector;
static PyObject *__pyx_n_s_and;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_ascii;
static PyObject *__pyx_n_s_astype;
static PyObject *__pyx_n_s_axis;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_base_hash;
static PyObject *__pyx_n_s_base_selector;
static PyObject *__pyx_n_s_base_source;
static PyObject *__pyx_n_s_bool;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_u_c;
static PyObject *__pyx_n_s_center;
static PyObject *__pyx_kp_s_center_0;
static PyObject *__pyx_kp_s_center_1;
static PyObject *__pyx_kp_s_center_2;
static PyObject *__pyx_n_s_chained_and;
static PyObject *__pyx_n_s_chained_or;
static PyObject *__pyx_n_s_child_mask;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_code_length;
static PyObject *__pyx_n_s_compose_selector;
static PyObject *__pyx_kp_s_conditional_s;
static PyObject *__pyx_n_s_conditionals;
static PyObject *__pyx_n_s_connectivity_coords;
static PyObject *__pyx_n_s_connectivity_indices;
static PyObject *__pyx_n_s_connectivity_length;
static PyObject *__pyx_kp_s_contiguous_and_direct;
static PyObject *__pyx_kp_s_contiguous_and_indirect;
static PyObject *__pyx_n_s_convert_mask_to_indices;
static PyObject *__pyx_n_s_convert_to_units;
static PyObject *__pyx_n_s_coord;
static PyObject *__pyx_n_s_count;
static PyObject *__pyx_n_s_count_points;
static PyObject *__pyx_n_s_cpos;
static PyObject *__pyx_n_s_cut_region_selector;
static PyObject *__pyx_n_s_cutting_selector;
static PyObject *__pyx_n_s_cx;
static PyObject *__pyx_n_s_cy;
static PyObject *__pyx_n_s_cz;
static PyObject *__pyx_n_s_d;
static PyObject *__pyx_n_s_d_2;
static PyObject *__pyx_n_s_data_collection_selector;
static PyObject *__pyx_n_s_data_objects;
static PyObject *__pyx_n_s_dds;
static PyObject *__pyx_n_s_defaults;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_disk_selector;
static PyObject *__pyx_n_s_dobj;
static PyObject *__pyx_n_s_dobj1;
static PyObject *__pyx_n_s_dobj2;
static PyObject *__pyx_n_s_domain_id;
static PyObject *__pyx_n_s_domain_left_edge;
static PyObject *__pyx_n_s_domain_right_edge;
static PyObject *__pyx_n_s_domain_width;
static PyObject *__pyx_kp_s_domain_width_0;
static PyObject *__pyx_kp_s_domain_width_1;
static PyObject *__pyx_kp_s_domain_width_2;
static PyObject *__pyx_n_s_double;
static PyObject *__pyx_n_s_ds;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dtype_is_object;
static PyObject *__pyx_n_s_dx;
static PyObject *__pyx_n_s_dy;
static PyObject *__pyx_n_s_dz;
static PyObject *__pyx_n_s_e0;
static PyObject *__pyx_n_s_e1;
static PyObject *__pyx_n_s_e2;
static PyObject *__pyx_n_s_ellipsoid_selector;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_end_point;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_eps;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_extend;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_filter_bbox;
static PyObject *__pyx_n_s_finfo;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_float;
static PyObject *__pyx_n_s_float32;
static PyObject *__pyx_n_s_float64;
static PyObject *__pyx_n_s_fmod;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fortran;
static PyObject *__pyx_n_u_fortran;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_s_got_differing_extents_in_dimensi;
static PyObject *__pyx_n_s_grid_selector;
static PyObject *__pyx_n_s_halo_particles;
static PyObject *__pyx_n_s_halo_particles_selector;
static PyObject *__pyx_n_s_hash_vals;
static PyObject *__pyx_n_s_height;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_id_offset;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_index_offset;
static PyObject *__pyx_n_s_indexed_octree_subset_selector;
static PyObject *__pyx_n_s_indices;
static PyObject *__pyx_n_s_inf;
static PyObject *__pyx_n_s_init___locals_genexpr;
static PyObject *__pyx_n_s_int64;
static PyObject *__pyx_n_s_intersection_selector;
static PyObject *__pyx_n_s_itemsize;
static PyObject *__pyx_kp_s_itemsize_0_for_cython_array;
static PyObject *__pyx_n_s_j;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_kind;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_left_edge;
static PyObject *__pyx_kp_s_left_edge_0;
static PyObject *__pyx_kp_s_left_edge_1;
static PyObject *__pyx_kp_s_left_edge_2;
static PyObject *__pyx_n_s_left_edges;
static PyObject *__pyx_n_s_levels;
static PyObject *__pyx_n_s_logical_or;
static PyObject *__pyx_n_s_loose_selection;
static PyObject *__pyx_kp_s_mag_0;
static PyObject *__pyx_kp_s_mag_1;
static PyObject *__pyx_kp_s_mag_2;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mask;
static PyObject *__pyx_n_s_mask_fill;
static PyObject *__pyx_n_s_max_ind;
static PyObject *__pyx_n_s_max_level;
static PyObject *__pyx_n_s_memview;
static PyObject *__pyx_n_s_mesh;
static PyObject *__pyx_n_s_min_ind;
static PyObject *__pyx_n_s_min_level;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_n_c;
static PyObject *__pyx_n_s_n_p;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_kp_u_ndarray_is_not_C_contiguous;
static PyObject *__pyx_kp_u_ndarray_is_not_Fortran_contiguou;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_s_neg;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_norm_vec;
static PyObject *__pyx_kp_s_norm_vec_0;
static PyObject *__pyx_kp_s_norm_vec_1;
static PyObject *__pyx_kp_s_norm_vec_2;
static PyObject *__pyx_n_s_not;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_nz;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_obj_ids;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_s_octree;
static PyObject *__pyx_n_s_octree_subset_selector;
static PyObject *__pyx_n_s_offset;
static PyObject *__pyx_n_s_ones;
static PyObject *__pyx_n_s_or;
static PyObject *__pyx_n_s_ortho_ray_selector;
static PyObject *__pyx_n_s_out;
static PyObject *__pyx_n_s_overlap_cells;
static PyObject *__pyx_n_s_p;
static PyObject *__pyx_kp_s_p1_0;
static PyObject *__pyx_kp_s_p1_1;
static PyObject *__pyx_kp_s_p1_2;
static PyObject *__pyx_kp_s_p2_0;
static PyObject *__pyx_kp_s_p2_1;
static PyObject *__pyx_kp_s_p2_2;
static PyObject *__pyx_kp_s_p_0;
static PyObject *__pyx_kp_s_p_1;
static PyObject *__pyx_kp_s_p_2;
static PyObject *__pyx_n_s_pack;
static PyObject *__pyx_n_s_particle_indices;
static PyObject *__pyx_n_s_periodicity;
static PyObject *__pyx_kp_s_periodicity_0;
static PyObject *__pyx_kp_s_periodicity_1;
static PyObject *__pyx_kp_s_periodicity_2;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_point_selector;
static PyObject *__pyx_n_s_points_in_cells;
static PyObject *__pyx_n_s_pos;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_px;
static PyObject *__pyx_n_s_px_ax;
static PyObject *__pyx_n_s_py;
static PyObject *__pyx_n_s_py_ax;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_getbuffer;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_Enum;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_pz;
static PyObject *__pyx_n_s_radius;
static PyObject *__pyx_n_s_radius2;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_ray_selector;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_region_selector;
static PyObject *__pyx_n_s_right_edge;
static PyObject *__pyx_kp_s_right_edge_0;
static PyObject *__pyx_kp_s_right_edge_1;
static PyObject *__pyx_kp_s_right_edge_2;
static PyObject *__pyx_n_s_right_edges;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_n_s_select_grids;
static PyObject *__pyx_n_s_select_points;
static PyObject *__pyx_n_s_selector;
static PyObject *__pyx_n_s_selector1;
static PyObject *__pyx_n_s_selector2;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_slice_selector;
static PyObject *__pyx_n_s_sphere_selector;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_start_point;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_kp_s_strided_and_direct;
static PyObject *__pyx_kp_s_strided_and_direct_or_indirect;
static PyObject *__pyx_kp_s_strided_and_indirect;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_to_ndarray;
static PyObject *__pyx_n_s_tostring;
static PyObject *__pyx_n_s_transpose;
static PyObject *__pyx_n_s_uint8;
static PyObject *__pyx_kp_s_unable_to_allocate_array_data;
static PyObject *__pyx_kp_s_unable_to_allocate_shape_and_str;
static PyObject *__pyx_n_s_union_selector;
static PyObject *__pyx_n_s_units;
static PyObject *__pyx_kp_u_unknown_dtype_code_in_numpy_pxd;
static PyObject *__pyx_n_s_unpack;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_vals;
static PyObject *__pyx_n_s_vec;
static PyObject *__pyx_kp_s_vec_0;
static PyObject *__pyx_kp_s_vec_0_0;
static PyObject *__pyx_kp_s_vec_0_1;
static PyObject *__pyx_kp_s_vec_0_2;
static PyObject *__pyx_kp_s_vec_1;
static PyObject *__pyx_kp_s_vec_1_0;
static PyObject *__pyx_kp_s_vec_1_1;
static PyObject *__pyx_kp_s_vec_1_2;
static PyObject *__pyx_kp_s_vec_2;
static PyObject *__pyx_kp_s_vec_2_0;
static PyObject *__pyx_kp_s_vec_2_1;
static PyObject *__pyx_kp_s_vec_2_2;
static PyObject *__pyx_n_s_view;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_xor;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_n_s_yt_geometry_selection_routines;
static PyObject *__pyx_kp_s_yt_geometry_selection_routines_p;
static PyObject *__pyx_n_s_z;
static PyObject *__pyx_n_s_zeros;
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_convert_mask_to_indices(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_mask, int __pyx_v_count, int __pyx_v_transpose); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_2mask_fill(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_out, __pyx_t_5numpy_int64_t __pyx_v_offset, PyArrayObject *__pyx_v_mask, PyArrayObject *__pyx_v_vals); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject___cinit__(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyObject *__pyx_v_dobj, CYTHON_UNUSED PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_2select_grids(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyArrayObject *__pyx_v_left_edges, PyArrayObject *__pyx_v_right_edges, PyArrayObject *__pyx_v_levels); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_4count_octs(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *__pyx_v_octree, int __pyx_v_domain_id); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_6count_oct_cells(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *__pyx_v_octree, int __pyx_v_domain_id); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_8fill_mesh_mask(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyObject *__pyx_v_mesh); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_10fill_mesh_cell_mask(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyObject *__pyx_v_mesh); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_12fill_mask(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyObject *__pyx_v_gobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_14count_points(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_24count_points(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyArrayObject *__pyx_v_x, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_z, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_26count_points(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyArrayObject *__pyx_v_x, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_z, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_16select_points(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_30select_points(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyArrayObject *__pyx_v_x, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_z, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_32select_points(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyArrayObject *__pyx_v_x, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_z, __pyx_t_5numpy_float64_t __pyx_v_radius); /* proto */
static Py_hash_t __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_18__hash__(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_20_hash_vals(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_22_base_hash(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_9min_level___get__(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_9min_level_2__set__(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_9max_level___get__(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_9max_level_2__set__(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_36__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_38__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_13PointSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_13PointSelector_2_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_13PointSelector_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_13PointSelector_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_14SphereSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SphereSelector_2_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SphereSelector_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SphereSelector_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_14RegionSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14RegionSelector_2_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14RegionSelector_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14RegionSelector_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17CutRegionSelector_8__init___genexpr(PyObject *__pyx_self); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_17CutRegionSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17CutRegionSelector_2_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17CutRegionSelector_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17CutRegionSelector_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_12DiskSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_12DiskSelector_2_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_12DiskSelector_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_12DiskSelector_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_20CuttingPlaneSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_2_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_13SliceSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_13SliceSelector_2fill_mask(struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self, PyObject *__pyx_v_gobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_13SliceSelector_4_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_13SliceSelector_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_13SliceSelector_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_16OrthoRaySelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_16OrthoRaySelector_2fill_mask(struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self, PyObject *__pyx_v_gobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_16OrthoRaySelector_4_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_16OrthoRaySelector_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_16OrthoRaySelector_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_11RaySelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_2fill_mask(struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, PyObject *__pyx_v_gobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_4get_dt(struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, PyObject *__pyx_v_gobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_6get_dt_mesh(struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, PyObject *__pyx_v_mesh, PyObject *__pyx_v_nz, int __pyx_v_offset); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_8_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector_2select_grids(struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *__pyx_v_self, PyArrayObject *__pyx_v_left_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_right_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_levels); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector_4fill_mask(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *__pyx_v_self, PyObject *__pyx_v_gobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector_6_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_17EllipsoidSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17EllipsoidSelector_2_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17EllipsoidSelector_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17EllipsoidSelector_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_12GridSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_12GridSelector_2select_grids(struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *__pyx_v_self, PyArrayObject *__pyx_v_left_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_right_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_levels); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_12GridSelector_4fill_mask(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *__pyx_v_self, PyObject *__pyx_v_gobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_12GridSelector_6_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_12GridSelector_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_12GridSelector_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_2select_grids(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED PyArrayObject *__pyx_v_left_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_right_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_levels); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_4_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_9domain_id___get__(struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_9domain_id_2__set__(struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_2select_grids(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED PyArrayObject *__pyx_v_left_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_right_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_levels); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_4_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_14AlwaysSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14AlwaysSelector_2select_grids(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self, PyArrayObject *__pyx_v_left_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_right_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_levels); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14AlwaysSelector_4_hash_vals(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14AlwaysSelector_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14AlwaysSelector_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_15ComposeSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_dobj, PyObject *__pyx_v_selector1, PyObject *__pyx_v_selector2); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_15ComposeSelector_2select_grids(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self, PyArrayObject *__pyx_v_left_edges, PyArrayObject *__pyx_v_right_edges, PyArrayObject *__pyx_v_levels); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_15ComposeSelector_4_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_15ComposeSelector_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_15ComposeSelector_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_2_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source___get__(struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source_2__set__(struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source_4__del__(struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_7halo_id___get__(struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_7halo_id_2__set__(struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_4points_in_cells(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_cx, __Pyx_memviewslice __pyx_v_cy, __Pyx_memviewslice __pyx_v_cz, __Pyx_memviewslice __pyx_v_dx, __Pyx_memviewslice __pyx_v_dy, __Pyx_memviewslice __pyx_v_dz, __Pyx_memviewslice __pyx_v_px, __Pyx_memviewslice __pyx_v_py, __Pyx_memviewslice __pyx_v_pz); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1___get__(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1_2__set__(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1_4__del__(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2___get__(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2_2__set__(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2_4__del__(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanANDSelector__hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanANDSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanANDSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17BooleanORSelector__hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17BooleanORSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17BooleanORSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanNOTSelector__hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanNOTSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanNOTSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanXORSelector__hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanXORSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanXORSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanNEGSelector__hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanNEGSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanNEGSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_18selection_routines_22ChainedBooleanSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *__pyx_v_self, PyObject *__pyx_v_dobj); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector__hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector__hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object); /* proto */
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_AlwaysSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_OctreeSubsetSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_BooleanSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_PointSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_SphereSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_RegionSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_CutRegionSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_DiskSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_CuttingPlaneSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_SliceSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_OrthoRaySelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_RaySelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_DataCollectionSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_EllipsoidSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_GridSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_ComposeSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_HaloParticlesSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_BooleanANDSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_BooleanORSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_BooleanNOTSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_BooleanXORSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_BooleanNEGSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_ChainedBooleanSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_ChainedBooleanORSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines___pyx_scope_struct____init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyByteArray_Type_extend = {0, &__pyx_n_s_extend, 0, 0, 0};
static PyObject *__pyx_float_1_0;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_99;
static PyObject *__pyx_int_184977713;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__90;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__51;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__53;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__55;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__57;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__61;
static PyObject *__pyx_tuple__62;
static PyObject *__pyx_tuple__63;
static PyObject *__pyx_tuple__64;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__66;
static PyObject *__pyx_tuple__67;
static PyObject *__pyx_tuple__68;
static PyObject *__pyx_tuple__69;
static PyObject *__pyx_tuple__70;
static PyObject *__pyx_tuple__71;
static PyObject *__pyx_tuple__72;
static PyObject *__pyx_tuple__73;
static PyObject *__pyx_tuple__74;
static PyObject *__pyx_tuple__75;
static PyObject *__pyx_tuple__76;
static PyObject *__pyx_tuple__77;
static PyObject *__pyx_tuple__78;
static PyObject *__pyx_tuple__79;
static PyObject *__pyx_tuple__80;
static PyObject *__pyx_tuple__81;
static PyObject *__pyx_tuple__82;
static PyObject *__pyx_tuple__83;
static PyObject *__pyx_tuple__84;
static PyObject *__pyx_tuple__85;
static PyObject *__pyx_tuple__86;
static PyObject *__pyx_tuple__87;
static PyObject *__pyx_tuple__88;
static PyObject *__pyx_tuple__89;
static PyObject *__pyx_tuple__91;
static PyObject *__pyx_tuple__92;
static PyObject *__pyx_tuple__93;
static PyObject *__pyx_tuple__94;
static PyObject *__pyx_tuple__96;
static PyObject *__pyx_tuple__98;
static PyObject *__pyx_tuple__100;
static PyObject *__pyx_tuple__102;
static PyObject *__pyx_tuple__104;
static PyObject *__pyx_tuple__105;
static PyObject *__pyx_tuple__106;
static PyObject *__pyx_tuple__107;
static PyObject *__pyx_tuple__108;
static PyObject *__pyx_tuple__109;
static PyObject *__pyx_codeobj__95;
static PyObject *__pyx_codeobj__97;
static PyObject *__pyx_codeobj__99;
static PyObject *__pyx_codeobj__101;
static PyObject *__pyx_codeobj__103;
static PyObject *__pyx_codeobj__110;
/* Late includes */

/* "yt/geometry/selection_routines.pyx":63
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def convert_mask_to_indices(np.ndarray[np.uint8_t, ndim=3, cast=True] mask,             # <<<<<<<<<<<<<<
 *             int count, int transpose = 0):
 *     cdef int i, j, k, cpos
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_1convert_mask_to_indices(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_8geometry_18selection_routines_1convert_mask_to_indices = {"convert_mask_to_indices", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_18selection_routines_1convert_mask_to_indices, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_1convert_mask_to_indices(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_mask = 0;
  int __pyx_v_count;
  int __pyx_v_transpose;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("convert_mask_to_indices (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mask,&__pyx_n_s_count,&__pyx_n_s_transpose,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_count)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convert_mask_to_indices", 0, 2, 3, 1); __PYX_ERR(0, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_transpose);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "convert_mask_to_indices") < 0)) __PYX_ERR(0, 63, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_mask = ((PyArrayObject *)values[0]);
    __pyx_v_count = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_count == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 64, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_transpose = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_transpose == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 64, __pyx_L3_error)
    } else {
      __pyx_v_transpose = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("convert_mask_to_indices", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 63, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.convert_mask_to_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask), __pyx_ptype_5numpy_ndarray, 1, "mask", 0))) __PYX_ERR(0, 63, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_convert_mask_to_indices(__pyx_self, __pyx_v_mask, __pyx_v_count, __pyx_v_transpose);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_convert_mask_to_indices(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_mask, int __pyx_v_count, int __pyx_v_transpose) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_cpos;
  PyArrayObject *__pyx_v_indices = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyArrayObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  npy_intp __pyx_t_10;
  npy_intp __pyx_t_11;
  npy_intp __pyx_t_12;
  npy_intp __pyx_t_13;
  int __pyx_t_14;
  npy_intp __pyx_t_15;
  npy_intp __pyx_t_16;
  int __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  int __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_mask_to_indices", 0);
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack) == -1)) __PYX_ERR(0, 63, __pyx_L1_error)
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mask.diminfo[1].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mask.diminfo[1].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_mask.diminfo[2].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_mask.diminfo[2].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[2];

  /* "yt/geometry/selection_routines.pyx":67
 *     cdef int i, j, k, cpos
 *     cdef np.ndarray[np.int64_t, ndim=2] indices
 *     indices = np.zeros((count, 3), dtype='int64')             # <<<<<<<<<<<<<<
 *     cpos = 0
 *     for i in range(mask.shape[0]):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_3);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_n_s_int64) < 0) __PYX_ERR(0, 67, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 67, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
      __pyx_t_7 = __pyx_t_8 = __pyx_t_9 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_indices.diminfo[1].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_indices.diminfo[1].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 67, __pyx_L1_error)
  }
  __pyx_t_5 = 0;
  __pyx_v_indices = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":68
 *     cdef np.ndarray[np.int64_t, ndim=2] indices
 *     indices = np.zeros((count, 3), dtype='int64')
 *     cpos = 0             # <<<<<<<<<<<<<<
 *     for i in range(mask.shape[0]):
 *         for j in range(mask.shape[1]):
 */
  __pyx_v_cpos = 0;

  /* "yt/geometry/selection_routines.pyx":69
 *     indices = np.zeros((count, 3), dtype='int64')
 *     cpos = 0
 *     for i in range(mask.shape[0]):             # <<<<<<<<<<<<<<
 *         for j in range(mask.shape[1]):
 *             for k in range(mask.shape[2]):
 */
  __pyx_t_10 = (__pyx_v_mask->dimensions[0]);
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_11; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "yt/geometry/selection_routines.pyx":70
 *     cpos = 0
 *     for i in range(mask.shape[0]):
 *         for j in range(mask.shape[1]):             # <<<<<<<<<<<<<<
 *             for k in range(mask.shape[2]):
 *                 if mask[i, j, k] == 1:
 */
    __pyx_t_12 = (__pyx_v_mask->dimensions[1]);
    __pyx_t_13 = __pyx_t_12;
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
      __pyx_v_j = __pyx_t_14;

      /* "yt/geometry/selection_routines.pyx":71
 *     for i in range(mask.shape[0]):
 *         for j in range(mask.shape[1]):
 *             for k in range(mask.shape[2]):             # <<<<<<<<<<<<<<
 *                 if mask[i, j, k] == 1:
 *                     if transpose == 1:
 */
      __pyx_t_15 = (__pyx_v_mask->dimensions[2]);
      __pyx_t_16 = __pyx_t_15;
      for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
        __pyx_v_k = __pyx_t_17;

        /* "yt/geometry/selection_routines.pyx":72
 *         for j in range(mask.shape[1]):
 *             for k in range(mask.shape[2]):
 *                 if mask[i, j, k] == 1:             # <<<<<<<<<<<<<<
 *                     if transpose == 1:
 *                         indices[cpos, 0] = k
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_19 = __pyx_v_j;
        __pyx_t_20 = __pyx_v_k;
        __pyx_t_21 = (((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_mask.diminfo[0].strides, __pyx_t_19, __pyx_pybuffernd_mask.diminfo[1].strides, __pyx_t_20, __pyx_pybuffernd_mask.diminfo[2].strides)) == 1) != 0);
        if (__pyx_t_21) {

          /* "yt/geometry/selection_routines.pyx":73
 *             for k in range(mask.shape[2]):
 *                 if mask[i, j, k] == 1:
 *                     if transpose == 1:             # <<<<<<<<<<<<<<
 *                         indices[cpos, 0] = k
 *                         indices[cpos, 1] = j
 */
          __pyx_t_21 = ((__pyx_v_transpose == 1) != 0);
          if (__pyx_t_21) {

            /* "yt/geometry/selection_routines.pyx":74
 *                 if mask[i, j, k] == 1:
 *                     if transpose == 1:
 *                         indices[cpos, 0] = k             # <<<<<<<<<<<<<<
 *                         indices[cpos, 1] = j
 *                         indices[cpos, 2] = i
 */
            __pyx_t_20 = __pyx_v_cpos;
            __pyx_t_19 = 0;
            *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_indices.rcbuffer->pybuffer.buf, __pyx_t_20, __pyx_pybuffernd_indices.diminfo[0].strides, __pyx_t_19, __pyx_pybuffernd_indices.diminfo[1].strides) = __pyx_v_k;

            /* "yt/geometry/selection_routines.pyx":75
 *                     if transpose == 1:
 *                         indices[cpos, 0] = k
 *                         indices[cpos, 1] = j             # <<<<<<<<<<<<<<
 *                         indices[cpos, 2] = i
 *                     else:
 */
            __pyx_t_18 = __pyx_v_cpos;
            __pyx_t_22 = 1;
            *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_indices.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_indices.diminfo[0].strides, __pyx_t_22, __pyx_pybuffernd_indices.diminfo[1].strides) = __pyx_v_j;

            /* "yt/geometry/selection_routines.pyx":76
 *                         indices[cpos, 0] = k
 *                         indices[cpos, 1] = j
 *                         indices[cpos, 2] = i             # <<<<<<<<<<<<<<
 *                     else:
 *                         indices[cpos, 0] = i
 */
            __pyx_t_23 = __pyx_v_cpos;
            __pyx_t_24 = 2;
            *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_indices.rcbuffer->pybuffer.buf, __pyx_t_23, __pyx_pybuffernd_indices.diminfo[0].strides, __pyx_t_24, __pyx_pybuffernd_indices.diminfo[1].strides) = __pyx_v_i;

            /* "yt/geometry/selection_routines.pyx":73
 *             for k in range(mask.shape[2]):
 *                 if mask[i, j, k] == 1:
 *                     if transpose == 1:             # <<<<<<<<<<<<<<
 *                         indices[cpos, 0] = k
 *                         indices[cpos, 1] = j
 */
            goto __pyx_L10;
          }

          /* "yt/geometry/selection_routines.pyx":78
 *                         indices[cpos, 2] = i
 *                     else:
 *                         indices[cpos, 0] = i             # <<<<<<<<<<<<<<
 *                         indices[cpos, 1] = j
 *                         indices[cpos, 2] = k
 */
          /*else*/ {
            __pyx_t_25 = __pyx_v_cpos;
            __pyx_t_26 = 0;
            *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_indices.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_indices.diminfo[0].strides, __pyx_t_26, __pyx_pybuffernd_indices.diminfo[1].strides) = __pyx_v_i;

            /* "yt/geometry/selection_routines.pyx":79
 *                     else:
 *                         indices[cpos, 0] = i
 *                         indices[cpos, 1] = j             # <<<<<<<<<<<<<<
 *                         indices[cpos, 2] = k
 *                     cpos += 1
 */
            __pyx_t_27 = __pyx_v_cpos;
            __pyx_t_28 = 1;
            *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_indices.rcbuffer->pybuffer.buf, __pyx_t_27, __pyx_pybuffernd_indices.diminfo[0].strides, __pyx_t_28, __pyx_pybuffernd_indices.diminfo[1].strides) = __pyx_v_j;

            /* "yt/geometry/selection_routines.pyx":80
 *                         indices[cpos, 0] = i
 *                         indices[cpos, 1] = j
 *                         indices[cpos, 2] = k             # <<<<<<<<<<<<<<
 *                     cpos += 1
 *     return indices
 */
            __pyx_t_29 = __pyx_v_cpos;
            __pyx_t_30 = 2;
            *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_indices.rcbuffer->pybuffer.buf, __pyx_t_29, __pyx_pybuffernd_indices.diminfo[0].strides, __pyx_t_30, __pyx_pybuffernd_indices.diminfo[1].strides) = __pyx_v_k;
          }
          __pyx_L10:;

          /* "yt/geometry/selection_routines.pyx":81
 *                         indices[cpos, 1] = j
 *                         indices[cpos, 2] = k
 *                     cpos += 1             # <<<<<<<<<<<<<<
 *     return indices
 * 
 */
          __pyx_v_cpos = (__pyx_v_cpos + 1);

          /* "yt/geometry/selection_routines.pyx":72
 *         for j in range(mask.shape[1]):
 *             for k in range(mask.shape[2]):
 *                 if mask[i, j, k] == 1:             # <<<<<<<<<<<<<<
 *                     if transpose == 1:
 *                         indices[cpos, 0] = k
 */
        }
      }
    }
  }

  /* "yt/geometry/selection_routines.pyx":82
 *                         indices[cpos, 2] = k
 *                     cpos += 1
 *     return indices             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __pyx_r = ((PyObject *)__pyx_v_indices);
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":63
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def convert_mask_to_indices(np.ndarray[np.uint8_t, ndim=3, cast=True] mask,             # <<<<<<<<<<<<<<
 *             int count, int transpose = 0):
 *     cdef int i, j, k, cpos
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.convert_mask_to_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":88
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef _mask_fill(np.ndarray[np.float64_t, ndim=1] out,             # <<<<<<<<<<<<<<
 *                 np.int64_t offset,
 *                 np.ndarray[np.uint8_t, ndim=3, cast=True] mask,
 */

static PyObject *__pyx_fuse_0__pyx_f_2yt_8geometry_18selection_routines__mask_fill(PyArrayObject *__pyx_v_out, __pyx_t_5numpy_int64_t __pyx_v_offset, PyArrayObject *__pyx_v_mask, PyArrayObject *__pyx_v_vals) {
  __pyx_t_5numpy_int64_t __pyx_v_count;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_out;
  __Pyx_Buffer __pyx_pybuffer_out;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_vals;
  __Pyx_Buffer __pyx_pybuffer_vals;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  npy_intp __pyx_t_1;
  npy_intp __pyx_t_2;
  int __pyx_t_3;
  npy_intp __pyx_t_4;
  npy_intp __pyx_t_5;
  int __pyx_t_6;
  npy_intp __pyx_t_7;
  npy_intp __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  __pyx_t_5numpy_int64_t __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_0_mask_fill", 0);
  __pyx_pybuffer_out.pybuffer.buf = NULL;
  __pyx_pybuffer_out.refcount = 0;
  __pyx_pybuffernd_out.data = NULL;
  __pyx_pybuffernd_out.rcbuffer = &__pyx_pybuffer_out;
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  __pyx_pybuffer_vals.pybuffer.buf = NULL;
  __pyx_pybuffer_vals.refcount = 0;
  __pyx_pybuffernd_vals.data = NULL;
  __pyx_pybuffernd_vals.rcbuffer = &__pyx_pybuffer_vals;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_out.rcbuffer->pybuffer, (PyObject*)__pyx_v_out, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 88, __pyx_L1_error)
  }
  __pyx_pybuffernd_out.diminfo[0].strides = __pyx_pybuffernd_out.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_out.diminfo[0].shape = __pyx_pybuffernd_out.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack) == -1)) __PYX_ERR(0, 88, __pyx_L1_error)
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mask.diminfo[1].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mask.diminfo[1].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_mask.diminfo[2].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_mask.diminfo[2].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[2];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_vals.rcbuffer->pybuffer, (PyObject*)__pyx_v_vals, &__Pyx_TypeInfo_float, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) __PYX_ERR(0, 88, __pyx_L1_error)
  }
  __pyx_pybuffernd_vals.diminfo[0].strides = __pyx_pybuffernd_vals.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_vals.diminfo[0].shape = __pyx_pybuffernd_vals.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_vals.diminfo[1].strides = __pyx_pybuffernd_vals.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_vals.diminfo[1].shape = __pyx_pybuffernd_vals.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_vals.diminfo[2].strides = __pyx_pybuffernd_vals.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_vals.diminfo[2].shape = __pyx_pybuffernd_vals.rcbuffer->pybuffer.shape[2];

  /* "yt/geometry/selection_routines.pyx":92
 *                 np.ndarray[np.uint8_t, ndim=3, cast=True] mask,
 *                 np.ndarray[floating, ndim=3] vals):
 *     cdef np.int64_t count = 0             # <<<<<<<<<<<<<<
 *     cdef int i, j, k
 *     for i in range(mask.shape[0]):
 */
  __pyx_v_count = 0;

  /* "yt/geometry/selection_routines.pyx":94
 *     cdef np.int64_t count = 0
 *     cdef int i, j, k
 *     for i in range(mask.shape[0]):             # <<<<<<<<<<<<<<
 *         for j in range(mask.shape[1]):
 *             for k in range(mask.shape[2]):
 */
  __pyx_t_1 = (__pyx_v_mask->dimensions[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "yt/geometry/selection_routines.pyx":95
 *     cdef int i, j, k
 *     for i in range(mask.shape[0]):
 *         for j in range(mask.shape[1]):             # <<<<<<<<<<<<<<
 *             for k in range(mask.shape[2]):
 *                 if mask[i, j, k] == 1:
 */
    __pyx_t_4 = (__pyx_v_mask->dimensions[1]);
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "yt/geometry/selection_routines.pyx":96
 *     for i in range(mask.shape[0]):
 *         for j in range(mask.shape[1]):
 *             for k in range(mask.shape[2]):             # <<<<<<<<<<<<<<
 *                 if mask[i, j, k] == 1:
 *                     out[offset + count] = vals[i, j, k]
 */
      __pyx_t_7 = (__pyx_v_mask->dimensions[2]);
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_k = __pyx_t_9;

        /* "yt/geometry/selection_routines.pyx":97
 *         for j in range(mask.shape[1]):
 *             for k in range(mask.shape[2]):
 *                 if mask[i, j, k] == 1:             # <<<<<<<<<<<<<<
 *                     out[offset + count] = vals[i, j, k]
 *                     count += 1
 */
        __pyx_t_10 = __pyx_v_i;
        __pyx_t_11 = __pyx_v_j;
        __pyx_t_12 = __pyx_v_k;
        __pyx_t_13 = (((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_mask.diminfo[0].strides, __pyx_t_11, __pyx_pybuffernd_mask.diminfo[1].strides, __pyx_t_12, __pyx_pybuffernd_mask.diminfo[2].strides)) == 1) != 0);
        if (__pyx_t_13) {

          /* "yt/geometry/selection_routines.pyx":98
 *             for k in range(mask.shape[2]):
 *                 if mask[i, j, k] == 1:
 *                     out[offset + count] = vals[i, j, k]             # <<<<<<<<<<<<<<
 *                     count += 1
 *     return count
 */
          __pyx_t_12 = __pyx_v_i;
          __pyx_t_11 = __pyx_v_j;
          __pyx_t_10 = __pyx_v_k;
          __pyx_t_14 = (__pyx_v_offset + __pyx_v_count);
          *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_out.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_out.diminfo[0].strides) = (*__Pyx_BufPtrStrided3d(float *, __pyx_pybuffernd_vals.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_vals.diminfo[0].strides, __pyx_t_11, __pyx_pybuffernd_vals.diminfo[1].strides, __pyx_t_10, __pyx_pybuffernd_vals.diminfo[2].strides));

          /* "yt/geometry/selection_routines.pyx":99
 *                 if mask[i, j, k] == 1:
 *                     out[offset + count] = vals[i, j, k]
 *                     count += 1             # <<<<<<<<<<<<<<
 *     return count
 * 
 */
          __pyx_v_count = (__pyx_v_count + 1);

          /* "yt/geometry/selection_routines.pyx":97
 *         for j in range(mask.shape[1]):
 *             for k in range(mask.shape[2]):
 *                 if mask[i, j, k] == 1:             # <<<<<<<<<<<<<<
 *                     out[offset + count] = vals[i, j, k]
 *                     count += 1
 */
        }
      }
    }
  }

  /* "yt/geometry/selection_routines.pyx":100
 *                     out[offset + count] = vals[i, j, k]
 *                     count += 1
 *     return count             # <<<<<<<<<<<<<<
 * 
 * def mask_fill(np.ndarray[np.float64_t, ndim=1] out,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_15 = __Pyx_PyInt_From_npy_int64(__pyx_v_count); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_r = __pyx_t_15;
  __pyx_t_15 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":88
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef _mask_fill(np.ndarray[np.float64_t, ndim=1] out,             # <<<<<<<<<<<<<<
 *                 np.int64_t offset,
 *                 np.ndarray[np.uint8_t, ndim=3, cast=True] mask,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_15);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_vals.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines._mask_fill", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_vals.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_fuse_1__pyx_f_2yt_8geometry_18selection_routines__mask_fill(PyArrayObject *__pyx_v_out, __pyx_t_5numpy_int64_t __pyx_v_offset, PyArrayObject *__pyx_v_mask, PyArrayObject *__pyx_v_vals) {
  __pyx_t_5numpy_int64_t __pyx_v_count;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_out;
  __Pyx_Buffer __pyx_pybuffer_out;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_vals;
  __Pyx_Buffer __pyx_pybuffer_vals;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  npy_intp __pyx_t_1;
  npy_intp __pyx_t_2;
  int __pyx_t_3;
  npy_intp __pyx_t_4;
  npy_intp __pyx_t_5;
  int __pyx_t_6;
  npy_intp __pyx_t_7;
  npy_intp __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  __pyx_t_5numpy_int64_t __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_1_mask_fill", 0);
  __pyx_pybuffer_out.pybuffer.buf = NULL;
  __pyx_pybuffer_out.refcount = 0;
  __pyx_pybuffernd_out.data = NULL;
  __pyx_pybuffernd_out.rcbuffer = &__pyx_pybuffer_out;
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  __pyx_pybuffer_vals.pybuffer.buf = NULL;
  __pyx_pybuffer_vals.refcount = 0;
  __pyx_pybuffernd_vals.data = NULL;
  __pyx_pybuffernd_vals.rcbuffer = &__pyx_pybuffer_vals;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_out.rcbuffer->pybuffer, (PyObject*)__pyx_v_out, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 88, __pyx_L1_error)
  }
  __pyx_pybuffernd_out.diminfo[0].strides = __pyx_pybuffernd_out.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_out.diminfo[0].shape = __pyx_pybuffernd_out.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack) == -1)) __PYX_ERR(0, 88, __pyx_L1_error)
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mask.diminfo[1].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mask.diminfo[1].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_mask.diminfo[2].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_mask.diminfo[2].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[2];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_vals.rcbuffer->pybuffer, (PyObject*)__pyx_v_vals, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) __PYX_ERR(0, 88, __pyx_L1_error)
  }
  __pyx_pybuffernd_vals.diminfo[0].strides = __pyx_pybuffernd_vals.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_vals.diminfo[0].shape = __pyx_pybuffernd_vals.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_vals.diminfo[1].strides = __pyx_pybuffernd_vals.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_vals.diminfo[1].shape = __pyx_pybuffernd_vals.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_vals.diminfo[2].strides = __pyx_pybuffernd_vals.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_vals.diminfo[2].shape = __pyx_pybuffernd_vals.rcbuffer->pybuffer.shape[2];

  /* "yt/geometry/selection_routines.pyx":92
 *                 np.ndarray[np.uint8_t, ndim=3, cast=True] mask,
 *                 np.ndarray[floating, ndim=3] vals):
 *     cdef np.int64_t count = 0             # <<<<<<<<<<<<<<
 *     cdef int i, j, k
 *     for i in range(mask.shape[0]):
 */
  __pyx_v_count = 0;

  /* "yt/geometry/selection_routines.pyx":94
 *     cdef np.int64_t count = 0
 *     cdef int i, j, k
 *     for i in range(mask.shape[0]):             # <<<<<<<<<<<<<<
 *         for j in range(mask.shape[1]):
 *             for k in range(mask.shape[2]):
 */
  __pyx_t_1 = (__pyx_v_mask->dimensions[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "yt/geometry/selection_routines.pyx":95
 *     cdef int i, j, k
 *     for i in range(mask.shape[0]):
 *         for j in range(mask.shape[1]):             # <<<<<<<<<<<<<<
 *             for k in range(mask.shape[2]):
 *                 if mask[i, j, k] == 1:
 */
    __pyx_t_4 = (__pyx_v_mask->dimensions[1]);
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "yt/geometry/selection_routines.pyx":96
 *     for i in range(mask.shape[0]):
 *         for j in range(mask.shape[1]):
 *             for k in range(mask.shape[2]):             # <<<<<<<<<<<<<<
 *                 if mask[i, j, k] == 1:
 *                     out[offset + count] = vals[i, j, k]
 */
      __pyx_t_7 = (__pyx_v_mask->dimensions[2]);
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_k = __pyx_t_9;

        /* "yt/geometry/selection_routines.pyx":97
 *         for j in range(mask.shape[1]):
 *             for k in range(mask.shape[2]):
 *                 if mask[i, j, k] == 1:             # <<<<<<<<<<<<<<
 *                     out[offset + count] = vals[i, j, k]
 *                     count += 1
 */
        __pyx_t_10 = __pyx_v_i;
        __pyx_t_11 = __pyx_v_j;
        __pyx_t_12 = __pyx_v_k;
        __pyx_t_13 = (((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_mask.diminfo[0].strides, __pyx_t_11, __pyx_pybuffernd_mask.diminfo[1].strides, __pyx_t_12, __pyx_pybuffernd_mask.diminfo[2].strides)) == 1) != 0);
        if (__pyx_t_13) {

          /* "yt/geometry/selection_routines.pyx":98
 *             for k in range(mask.shape[2]):
 *                 if mask[i, j, k] == 1:
 *                     out[offset + count] = vals[i, j, k]             # <<<<<<<<<<<<<<
 *                     count += 1
 *     return count
 */
          __pyx_t_12 = __pyx_v_i;
          __pyx_t_11 = __pyx_v_j;
          __pyx_t_10 = __pyx_v_k;
          __pyx_t_14 = (__pyx_v_offset + __pyx_v_count);
          *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_out.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_out.diminfo[0].strides) = (*__Pyx_BufPtrStrided3d(double *, __pyx_pybuffernd_vals.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_vals.diminfo[0].strides, __pyx_t_11, __pyx_pybuffernd_vals.diminfo[1].strides, __pyx_t_10, __pyx_pybuffernd_vals.diminfo[2].strides));

          /* "yt/geometry/selection_routines.pyx":99
 *                 if mask[i, j, k] == 1:
 *                     out[offset + count] = vals[i, j, k]
 *                     count += 1             # <<<<<<<<<<<<<<
 *     return count
 * 
 */
          __pyx_v_count = (__pyx_v_count + 1);

          /* "yt/geometry/selection_routines.pyx":97
 *         for j in range(mask.shape[1]):
 *             for k in range(mask.shape[2]):
 *                 if mask[i, j, k] == 1:             # <<<<<<<<<<<<<<
 *                     out[offset + count] = vals[i, j, k]
 *                     count += 1
 */
        }
      }
    }
  }

  /* "yt/geometry/selection_routines.pyx":100
 *                     out[offset + count] = vals[i, j, k]
 *                     count += 1
 *     return count             # <<<<<<<<<<<<<<
 * 
 * def mask_fill(np.ndarray[np.float64_t, ndim=1] out,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_15 = __Pyx_PyInt_From_npy_int64(__pyx_v_count); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_r = __pyx_t_15;
  __pyx_t_15 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":88
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * cdef _mask_fill(np.ndarray[np.float64_t, ndim=1] out,             # <<<<<<<<<<<<<<
 *                 np.int64_t offset,
 *                 np.ndarray[np.uint8_t, ndim=3, cast=True] mask,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_15);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_vals.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines._mask_fill", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_vals.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":102
 *     return count
 * 
 * def mask_fill(np.ndarray[np.float64_t, ndim=1] out,             # <<<<<<<<<<<<<<
 *               np.int64_t offset,
 *               np.ndarray[np.uint8_t, ndim=3, cast=True] mask,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_3mask_fill(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_8geometry_18selection_routines_3mask_fill = {"mask_fill", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_18selection_routines_3mask_fill, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_3mask_fill(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_out = 0;
  __pyx_t_5numpy_int64_t __pyx_v_offset;
  PyArrayObject *__pyx_v_mask = 0;
  PyArrayObject *__pyx_v_vals = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mask_fill (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_out,&__pyx_n_s_offset,&__pyx_n_s_mask,&__pyx_n_s_vals,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_out)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mask_fill", 1, 4, 4, 1); __PYX_ERR(0, 102, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mask_fill", 1, 4, 4, 2); __PYX_ERR(0, 102, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_vals)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mask_fill", 1, 4, 4, 3); __PYX_ERR(0, 102, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "mask_fill") < 0)) __PYX_ERR(0, 102, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_out = ((PyArrayObject *)values[0]);
    __pyx_v_offset = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_offset == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 103, __pyx_L3_error)
    __pyx_v_mask = ((PyArrayObject *)values[2]);
    __pyx_v_vals = ((PyArrayObject *)values[3]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("mask_fill", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 102, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.mask_fill", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_out), __pyx_ptype_5numpy_ndarray, 1, "out", 0))) __PYX_ERR(0, 102, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask), __pyx_ptype_5numpy_ndarray, 1, "mask", 0))) __PYX_ERR(0, 104, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_vals), __pyx_ptype_5numpy_ndarray, 1, "vals", 0))) __PYX_ERR(0, 105, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_2mask_fill(__pyx_self, __pyx_v_out, __pyx_v_offset, __pyx_v_mask, __pyx_v_vals);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_2mask_fill(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_out, __pyx_t_5numpy_int64_t __pyx_v_offset, PyArrayObject *__pyx_v_mask, PyArrayObject *__pyx_v_vals) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_out;
  __Pyx_Buffer __pyx_pybuffer_out;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mask_fill", 0);
  __pyx_pybuffer_out.pybuffer.buf = NULL;
  __pyx_pybuffer_out.refcount = 0;
  __pyx_pybuffernd_out.data = NULL;
  __pyx_pybuffernd_out.rcbuffer = &__pyx_pybuffer_out;
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_out.rcbuffer->pybuffer, (PyObject*)__pyx_v_out, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 102, __pyx_L1_error)
  }
  __pyx_pybuffernd_out.diminfo[0].strides = __pyx_pybuffernd_out.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_out.diminfo[0].shape = __pyx_pybuffernd_out.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack) == -1)) __PYX_ERR(0, 102, __pyx_L1_error)
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mask.diminfo[1].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mask.diminfo[1].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_mask.diminfo[2].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_mask.diminfo[2].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[2];

  /* "yt/geometry/selection_routines.pyx":106
 *               np.ndarray[np.uint8_t, ndim=3, cast=True] mask,
 *               np.ndarray vals):
 *     if vals.dtype == np.float32:             # <<<<<<<<<<<<<<
 *         return _mask_fill[np.float32_t](out, offset, mask, vals)
 *     elif vals.dtype == np.float64:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_vals), __pyx_n_s_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_float32); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "yt/geometry/selection_routines.pyx":107
 *               np.ndarray vals):
 *     if vals.dtype == np.float32:
 *         return _mask_fill[np.float32_t](out, offset, mask, vals)             # <<<<<<<<<<<<<<
 *     elif vals.dtype == np.float64:
 *         return _mask_fill[np.float64_t](out, offset, mask, vals)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_fuse_0__pyx_f_2yt_8geometry_18selection_routines__mask_fill(((PyArrayObject *)__pyx_v_out), __pyx_v_offset, ((PyArrayObject *)__pyx_v_mask), ((PyArrayObject *)__pyx_v_vals)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":106
 *               np.ndarray[np.uint8_t, ndim=3, cast=True] mask,
 *               np.ndarray vals):
 *     if vals.dtype == np.float32:             # <<<<<<<<<<<<<<
 *         return _mask_fill[np.float32_t](out, offset, mask, vals)
 *     elif vals.dtype == np.float64:
 */
  }

  /* "yt/geometry/selection_routines.pyx":108
 *     if vals.dtype == np.float32:
 *         return _mask_fill[np.float32_t](out, offset, mask, vals)
 *     elif vals.dtype == np.float64:             # <<<<<<<<<<<<<<
 *         return _mask_fill[np.float64_t](out, offset, mask, vals)
 *     else:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_vals), __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(__pyx_t_4)) {

    /* "yt/geometry/selection_routines.pyx":109
 *         return _mask_fill[np.float32_t](out, offset, mask, vals)
 *     elif vals.dtype == np.float64:
 *         return _mask_fill[np.float64_t](out, offset, mask, vals)             # <<<<<<<<<<<<<<
 *     else:
 *         raise RuntimeError
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_fuse_1__pyx_f_2yt_8geometry_18selection_routines__mask_fill(((PyArrayObject *)__pyx_v_out), __pyx_v_offset, ((PyArrayObject *)__pyx_v_mask), ((PyArrayObject *)__pyx_v_vals)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":108
 *     if vals.dtype == np.float32:
 *         return _mask_fill[np.float32_t](out, offset, mask, vals)
 *     elif vals.dtype == np.float64:             # <<<<<<<<<<<<<<
 *         return _mask_fill[np.float64_t](out, offset, mask, vals)
 *     else:
 */
  }

  /* "yt/geometry/selection_routines.pyx":111
 *         return _mask_fill[np.float64_t](out, offset, mask, vals)
 *     else:
 *         raise RuntimeError             # <<<<<<<<<<<<<<
 * 
 * cdef class SelectorObject:
 */
  /*else*/ {
    __Pyx_Raise(__pyx_builtin_RuntimeError, 0, 0, 0);
    __PYX_ERR(0, 111, __pyx_L1_error)
  }

  /* "yt/geometry/selection_routines.pyx":102
 *     return count
 * 
 * def mask_fill(np.ndarray[np.float64_t, ndim=1] out,             # <<<<<<<<<<<<<<
 *               np.int64_t offset,
 *               np.ndarray[np.uint8_t, ndim=3, cast=True] mask,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.mask_fill", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":115
 * cdef class SelectorObject:
 * 
 *     def __cinit__(self, dobj, *args):             # <<<<<<<<<<<<<<
 *         self._hash_initialized = 0
 *         cdef np.float64_t [:] DLE
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (PyTuple_GET_SIZE(__pyx_args) > 1) {
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 1, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_args)) {
      __Pyx_RefNannyFinishContext();
      return -1;
    }
    __Pyx_GOTREF(__pyx_v_args);
  } else {
    __pyx_v_args = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 1) ? pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, used_pos_args, "__cinit__") < 0)) __PYX_ERR(0, 115, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) < 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 115, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject___cinit__(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_dobj, __pyx_v_args);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject___cinit__(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyObject *__pyx_v_dobj, CYTHON_UNUSED PyObject *__pyx_v_args) {
  __Pyx_memviewslice __pyx_v_DLE = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_DRE = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_ds = NULL;
  long __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int32_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  long __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "yt/geometry/selection_routines.pyx":116
 * 
 *     def __cinit__(self, dobj, *args):
 *         self._hash_initialized = 0             # <<<<<<<<<<<<<<
 *         cdef np.float64_t [:] DLE
 *         cdef np.float64_t [:] DRE
 */
  __pyx_v_self->_hash_initialized = 0;

  /* "yt/geometry/selection_routines.pyx":119
 *         cdef np.float64_t [:] DLE
 *         cdef np.float64_t [:] DRE
 *         self.min_level = getattr(dobj, "min_level", 0)             # <<<<<<<<<<<<<<
 *         self.max_level = getattr(dobj, "max_level", 99)
 *         self.overlap_cells = 0
 */
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_dobj, __pyx_n_s_min_level, __pyx_int_0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->min_level = __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":120
 *         cdef np.float64_t [:] DRE
 *         self.min_level = getattr(dobj, "min_level", 0)
 *         self.max_level = getattr(dobj, "max_level", 99)             # <<<<<<<<<<<<<<
 *         self.overlap_cells = 0
 * 
 */
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_dobj, __pyx_n_s_max_level, __pyx_int_99); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->max_level = __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":121
 *         self.min_level = getattr(dobj, "min_level", 0)
 *         self.max_level = getattr(dobj, "max_level", 99)
 *         self.overlap_cells = 0             # <<<<<<<<<<<<<<
 * 
 *         ds = getattr(dobj, 'ds', None)
 */
  __pyx_v_self->overlap_cells = 0;

  /* "yt/geometry/selection_routines.pyx":123
 *         self.overlap_cells = 0
 * 
 *         ds = getattr(dobj, 'ds', None)             # <<<<<<<<<<<<<<
 *         if ds is None:
 *             for i in range(3):
 */
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_dobj, __pyx_n_s_ds, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ds = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":124
 * 
 *         ds = getattr(dobj, 'ds', None)
 *         if ds is None:             # <<<<<<<<<<<<<<
 *             for i in range(3):
 *                 # NOTE that this is not universal.
 */
  __pyx_t_3 = (__pyx_v_ds == Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "yt/geometry/selection_routines.pyx":125
 *         ds = getattr(dobj, 'ds', None)
 *         if ds is None:
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 # NOTE that this is not universal.
 *                 self.domain_width[i] = 1.0
 */
    for (__pyx_t_5 = 0; __pyx_t_5 < 3; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "yt/geometry/selection_routines.pyx":127
 *             for i in range(3):
 *                 # NOTE that this is not universal.
 *                 self.domain_width[i] = 1.0             # <<<<<<<<<<<<<<
 *                 self.periodicity[i] = False
 *         else:
 */
      (__pyx_v_self->domain_width[__pyx_v_i]) = 1.0;

      /* "yt/geometry/selection_routines.pyx":128
 *                 # NOTE that this is not universal.
 *                 self.domain_width[i] = 1.0
 *                 self.periodicity[i] = False             # <<<<<<<<<<<<<<
 *         else:
 *             DLE = _ensure_code(ds.domain_left_edge)
 */
      (__pyx_v_self->periodicity[__pyx_v_i]) = 0;
    }

    /* "yt/geometry/selection_routines.pyx":124
 * 
 *         ds = getattr(dobj, 'ds', None)
 *         if ds is None:             # <<<<<<<<<<<<<<
 *             for i in range(3):
 *                 # NOTE that this is not universal.
 */
    goto __pyx_L3;
  }

  /* "yt/geometry/selection_routines.pyx":130
 *                 self.periodicity[i] = False
 *         else:
 *             DLE = _ensure_code(ds.domain_left_edge)             # <<<<<<<<<<<<<<
 *             DRE = _ensure_code(ds.domain_right_edge)
 *             for i in range(3):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_ds, __pyx_n_s_domain_left_edge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_DLE = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "yt/geometry/selection_routines.pyx":131
 *         else:
 *             DLE = _ensure_code(ds.domain_left_edge)
 *             DRE = _ensure_code(ds.domain_right_edge)             # <<<<<<<<<<<<<<
 *             for i in range(3):
 *                 self.domain_width[i] = DRE[i] - DLE[i]
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_ds, __pyx_n_s_domain_right_edge); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_DRE = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "yt/geometry/selection_routines.pyx":132
 *             DLE = _ensure_code(ds.domain_left_edge)
 *             DRE = _ensure_code(ds.domain_right_edge)
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 self.domain_width[i] = DRE[i] - DLE[i]
 *                 self.periodicity[i] = ds.periodicity[i]
 */
    for (__pyx_t_5 = 0; __pyx_t_5 < 3; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "yt/geometry/selection_routines.pyx":133
 *             DRE = _ensure_code(ds.domain_right_edge)
 *             for i in range(3):
 *                 self.domain_width[i] = DRE[i] - DLE[i]             # <<<<<<<<<<<<<<
 *                 self.periodicity[i] = ds.periodicity[i]
 * 
 */
      __pyx_t_8 = __pyx_v_i;
      __pyx_t_9 = -1;
      if (__pyx_t_8 < 0) {
        __pyx_t_8 += __pyx_v_DRE.shape[0];
        if (unlikely(__pyx_t_8 < 0)) __pyx_t_9 = 0;
      } else if (unlikely(__pyx_t_8 >= __pyx_v_DRE.shape[0])) __pyx_t_9 = 0;
      if (unlikely(__pyx_t_9 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_9);
        __PYX_ERR(0, 133, __pyx_L1_error)
      }
      __pyx_t_10 = __pyx_v_i;
      __pyx_t_9 = -1;
      if (__pyx_t_10 < 0) {
        __pyx_t_10 += __pyx_v_DLE.shape[0];
        if (unlikely(__pyx_t_10 < 0)) __pyx_t_9 = 0;
      } else if (unlikely(__pyx_t_10 >= __pyx_v_DLE.shape[0])) __pyx_t_9 = 0;
      if (unlikely(__pyx_t_9 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_9);
        __PYX_ERR(0, 133, __pyx_L1_error)
      }
      (__pyx_v_self->domain_width[__pyx_v_i]) = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_DRE.data + __pyx_t_8 * __pyx_v_DRE.strides[0]) ))) - (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_DLE.data + __pyx_t_10 * __pyx_v_DLE.strides[0]) ))));

      /* "yt/geometry/selection_routines.pyx":134
 *             for i in range(3):
 *                 self.domain_width[i] = DRE[i] - DLE[i]
 *                 self.periodicity[i] = ds.periodicity[i]             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_ds, __pyx_n_s_periodicity); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_i, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 134, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 134, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      (__pyx_v_self->periodicity[__pyx_v_i]) = __pyx_t_4;
    }
  }
  __pyx_L3:;

  /* "yt/geometry/selection_routines.pyx":115
 * cdef class SelectorObject:
 * 
 *     def __cinit__(self, dobj, *args):             # <<<<<<<<<<<<<<
 *         self._hash_initialized = 0
 *         cdef np.float64_t [:] DLE
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_DLE, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_DRE, 1);
  __Pyx_XDECREF(__pyx_v_ds);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":139
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_grids(self,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=2] left_edges,
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_3select_grids(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_3select_grids(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left_edges = 0;
  PyArrayObject *__pyx_v_right_edges = 0;
  PyArrayObject *__pyx_v_levels = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("select_grids (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left_edges,&__pyx_n_s_right_edges,&__pyx_n_s_levels,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left_edges)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right_edges)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, 1); __PYX_ERR(0, 139, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_levels)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, 2); __PYX_ERR(0, 139, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "select_grids") < 0)) __PYX_ERR(0, 139, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_left_edges = ((PyArrayObject *)values[0]);
    __pyx_v_right_edges = ((PyArrayObject *)values[1]);
    __pyx_v_levels = ((PyArrayObject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 139, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.select_grids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left_edges), __pyx_ptype_5numpy_ndarray, 1, "left_edges", 0))) __PYX_ERR(0, 140, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right_edges), __pyx_ptype_5numpy_ndarray, 1, "right_edges", 0))) __PYX_ERR(0, 141, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_levels), __pyx_ptype_5numpy_ndarray, 1, "levels", 0))) __PYX_ERR(0, 142, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_2select_grids(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_left_edges, __pyx_v_right_edges, __pyx_v_levels);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_2select_grids(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyArrayObject *__pyx_v_left_edges, PyArrayObject *__pyx_v_right_edges, PyArrayObject *__pyx_v_levels) {
  int __pyx_v_i;
  int __pyx_v_n;
  int __pyx_v_ng;
  PyArrayObject *__pyx_v_gridi = 0;
  __pyx_t_5numpy_float64_t __pyx_v_LE[3];
  __pyx_t_5numpy_float64_t __pyx_v_RE[3];
  __Pyx_LocalBuf_ND __pyx_pybuffernd_gridi;
  __Pyx_Buffer __pyx_pybuffer_gridi;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left_edges;
  __Pyx_Buffer __pyx_pybuffer_left_edges;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_levels;
  __Pyx_Buffer __pyx_pybuffer_levels;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right_edges;
  __Pyx_Buffer __pyx_pybuffer_right_edges;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyArrayObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("select_grids", 0);
  __pyx_pybuffer_gridi.pybuffer.buf = NULL;
  __pyx_pybuffer_gridi.refcount = 0;
  __pyx_pybuffernd_gridi.data = NULL;
  __pyx_pybuffernd_gridi.rcbuffer = &__pyx_pybuffer_gridi;
  __pyx_pybuffer_left_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_left_edges.refcount = 0;
  __pyx_pybuffernd_left_edges.data = NULL;
  __pyx_pybuffernd_left_edges.rcbuffer = &__pyx_pybuffer_left_edges;
  __pyx_pybuffer_right_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_right_edges.refcount = 0;
  __pyx_pybuffernd_right_edges.data = NULL;
  __pyx_pybuffernd_right_edges.rcbuffer = &__pyx_pybuffer_right_edges;
  __pyx_pybuffer_levels.pybuffer.buf = NULL;
  __pyx_pybuffer_levels.refcount = 0;
  __pyx_pybuffernd_levels.data = NULL;
  __pyx_pybuffernd_levels.rcbuffer = &__pyx_pybuffer_levels;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_left_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 139, __pyx_L1_error)
  }
  __pyx_pybuffernd_left_edges.diminfo[0].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left_edges.diminfo[0].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_left_edges.diminfo[1].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_left_edges.diminfo[1].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_right_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 139, __pyx_L1_error)
  }
  __pyx_pybuffernd_right_edges.diminfo[0].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right_edges.diminfo[0].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_right_edges.diminfo[1].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_right_edges.diminfo[1].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_levels.rcbuffer->pybuffer, (PyObject*)__pyx_v_levels, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 139, __pyx_L1_error)
  }
  __pyx_pybuffernd_levels.diminfo[0].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_levels.diminfo[0].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_levels.diminfo[1].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_levels.diminfo[1].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[1];

  /* "yt/geometry/selection_routines.pyx":144
 *                      np.ndarray[np.int32_t, ndim=2] levels):
 *         cdef int i, n
 *         cdef int ng = left_edges.shape[0]             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')
 *         cdef np.float64_t LE[3]
 */
  __pyx_v_ng = (__pyx_v_left_edges->dimensions[0]);

  /* "yt/geometry/selection_routines.pyx":145
 *         cdef int i, n
 *         cdef int ng = left_edges.shape[0]
 *         cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')             # <<<<<<<<<<<<<<
 *         cdef np.float64_t LE[3]
 *         cdef np.float64_t RE[3]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_ng); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 145, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_gridi.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_gridi = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_gridi.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 145, __pyx_L1_error)
    } else {__pyx_pybuffernd_gridi.diminfo[0].strides = __pyx_pybuffernd_gridi.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_gridi.diminfo[0].shape = __pyx_pybuffernd_gridi.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_5 = 0;
  __pyx_v_gridi = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":148
 *         cdef np.float64_t LE[3]
 *         cdef np.float64_t RE[3]
 *         _ensure_code(left_edges)             # <<<<<<<<<<<<<<
 *         _ensure_code(right_edges)
 *         with nogil:
 */
  __pyx_t_4 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(((PyObject *)__pyx_v_left_edges)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":149
 *         cdef np.float64_t RE[3]
 *         _ensure_code(left_edges)
 *         _ensure_code(right_edges)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             for n in range(ng):
 */
  __pyx_t_4 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(((PyObject *)__pyx_v_right_edges)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":150
 *         _ensure_code(left_edges)
 *         _ensure_code(right_edges)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for n in range(ng):
 *                 # Call our selector function
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "yt/geometry/selection_routines.pyx":151
 *         _ensure_code(right_edges)
 *         with nogil:
 *             for n in range(ng):             # <<<<<<<<<<<<<<
 *                 # Call our selector function
 *                 # Check if the sphere is inside the grid
 */
        __pyx_t_6 = __pyx_v_ng;
        __pyx_t_7 = __pyx_t_6;
        for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
          __pyx_v_n = __pyx_t_8;

          /* "yt/geometry/selection_routines.pyx":154
 *                 # Call our selector function
 *                 # Check if the sphere is inside the grid
 *                 for i in range(3):             # <<<<<<<<<<<<<<
 *                     LE[i] = left_edges[n, i]
 *                     RE[i] = right_edges[n, i]
 */
          for (__pyx_t_9 = 0; __pyx_t_9 < 3; __pyx_t_9+=1) {
            __pyx_v_i = __pyx_t_9;

            /* "yt/geometry/selection_routines.pyx":155
 *                 # Check if the sphere is inside the grid
 *                 for i in range(3):
 *                     LE[i] = left_edges[n, i]             # <<<<<<<<<<<<<<
 *                     RE[i] = right_edges[n, i]
 *                 gridi[n] = self.select_grid(LE, RE, levels[n, 0])
 */
            __pyx_t_10 = __pyx_v_n;
            __pyx_t_11 = __pyx_v_i;
            (__pyx_v_LE[__pyx_v_i]) = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_left_edges.diminfo[0].strides, __pyx_t_11, __pyx_pybuffernd_left_edges.diminfo[1].strides));

            /* "yt/geometry/selection_routines.pyx":156
 *                 for i in range(3):
 *                     LE[i] = left_edges[n, i]
 *                     RE[i] = right_edges[n, i]             # <<<<<<<<<<<<<<
 *                 gridi[n] = self.select_grid(LE, RE, levels[n, 0])
 *         return gridi.astype("bool")
 */
            __pyx_t_11 = __pyx_v_n;
            __pyx_t_10 = __pyx_v_i;
            (__pyx_v_RE[__pyx_v_i]) = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_right_edges.diminfo[0].strides, __pyx_t_10, __pyx_pybuffernd_right_edges.diminfo[1].strides));
          }

          /* "yt/geometry/selection_routines.pyx":157
 *                     LE[i] = left_edges[n, i]
 *                     RE[i] = right_edges[n, i]
 *                 gridi[n] = self.select_grid(LE, RE, levels[n, 0])             # <<<<<<<<<<<<<<
 *         return gridi.astype("bool")
 * 
 */
          __pyx_t_10 = __pyx_v_n;
          __pyx_t_11 = 0;
          __pyx_t_12 = __pyx_v_n;
          *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_gridi.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_gridi.diminfo[0].strides) = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_grid(__pyx_v_self, __pyx_v_LE, __pyx_v_RE, (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int32_t *, __pyx_pybuffernd_levels.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_levels.diminfo[0].strides, __pyx_t_11, __pyx_pybuffernd_levels.diminfo[1].strides)), NULL);
        }
      }

      /* "yt/geometry/selection_routines.pyx":150
 *         _ensure_code(left_edges)
 *         _ensure_code(right_edges)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for n in range(ng):
 *                 # Call our selector function
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "yt/geometry/selection_routines.pyx":158
 *                     RE[i] = right_edges[n, i]
 *                 gridi[n] = self.select_grid(LE, RE, levels[n, 0])
 *         return gridi.astype("bool")             # <<<<<<<<<<<<<<
 * 
 *     def count_octs(self, OctreeContainer octree, int domain_id = -1):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_gridi), __pyx_n_s_astype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_n_s_bool) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_n_s_bool);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":139
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_grids(self,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=2] left_edges,
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_gridi.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.select_grids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_gridi.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_gridi);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":160
 *         return gridi.astype("bool")
 * 
 *     def count_octs(self, OctreeContainer octree, int domain_id = -1):             # <<<<<<<<<<<<<<
 *         cdef oct_visitors.CountTotalOcts visitor
 *         visitor = oct_visitors.CountTotalOcts(octree, domain_id)
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_5count_octs(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_5count_octs(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *__pyx_v_octree = 0;
  int __pyx_v_domain_id;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("count_octs (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_octree,&__pyx_n_s_domain_id,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_octree)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_domain_id);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "count_octs") < 0)) __PYX_ERR(0, 160, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_octree = ((struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *)values[0]);
    if (values[1]) {
      __pyx_v_domain_id = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_domain_id == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 160, __pyx_L3_error)
    } else {
      __pyx_v_domain_id = ((int)-1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("count_octs", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 160, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.count_octs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_octree), __pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer, 1, "octree", 0))) __PYX_ERR(0, 160, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_4count_octs(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_octree, __pyx_v_domain_id);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_4count_octs(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *__pyx_v_octree, int __pyx_v_domain_id) {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_CountTotalOcts *__pyx_v_visitor = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_octs", 0);

  /* "yt/geometry/selection_routines.pyx":162
 *     def count_octs(self, OctreeContainer octree, int domain_id = -1):
 *         cdef oct_visitors.CountTotalOcts visitor
 *         visitor = oct_visitors.CountTotalOcts(octree, domain_id)             # <<<<<<<<<<<<<<
 *         octree.visit_all_octs(self, visitor)
 *         return visitor.index
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_domain_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_octree));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_octree));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_octree));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalOcts), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_visitor = ((struct __pyx_obj_2yt_8geometry_12oct_visitors_CountTotalOcts *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":163
 *         cdef oct_visitors.CountTotalOcts visitor
 *         visitor = oct_visitors.CountTotalOcts(octree, domain_id)
 *         octree.visit_all_octs(self, visitor)             # <<<<<<<<<<<<<<
 *         return visitor.index
 * 
 */
  ((struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctreeContainer *)__pyx_v_octree->__pyx_vtab)->visit_all_octs(__pyx_v_octree, __pyx_v_self, ((struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *)__pyx_v_visitor), NULL);

  /* "yt/geometry/selection_routines.pyx":164
 *         visitor = oct_visitors.CountTotalOcts(octree, domain_id)
 *         octree.visit_all_octs(self, visitor)
 *         return visitor.index             # <<<<<<<<<<<<<<
 * 
 *     def count_oct_cells(self, OctreeContainer octree, int domain_id = -1):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_visitor->__pyx_base.index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":160
 *         return gridi.astype("bool")
 * 
 *     def count_octs(self, OctreeContainer octree, int domain_id = -1):             # <<<<<<<<<<<<<<
 *         cdef oct_visitors.CountTotalOcts visitor
 *         visitor = oct_visitors.CountTotalOcts(octree, domain_id)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.count_octs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_visitor);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":166
 *         return visitor.index
 * 
 *     def count_oct_cells(self, OctreeContainer octree, int domain_id = -1):             # <<<<<<<<<<<<<<
 *         cdef oct_visitors.CountTotalCells visitor
 *         visitor = oct_visitors.CountTotalCells(octree, domain_id)
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_7count_oct_cells(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_7count_oct_cells(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *__pyx_v_octree = 0;
  int __pyx_v_domain_id;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("count_oct_cells (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_octree,&__pyx_n_s_domain_id,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_octree)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_domain_id);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "count_oct_cells") < 0)) __PYX_ERR(0, 166, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_octree = ((struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *)values[0]);
    if (values[1]) {
      __pyx_v_domain_id = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_domain_id == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 166, __pyx_L3_error)
    } else {
      __pyx_v_domain_id = ((int)-1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("count_oct_cells", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 166, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.count_oct_cells", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_octree), __pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer, 1, "octree", 0))) __PYX_ERR(0, 166, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_6count_oct_cells(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_octree, __pyx_v_domain_id);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_6count_oct_cells(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *__pyx_v_octree, int __pyx_v_domain_id) {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_CountTotalCells *__pyx_v_visitor = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_oct_cells", 0);

  /* "yt/geometry/selection_routines.pyx":168
 *     def count_oct_cells(self, OctreeContainer octree, int domain_id = -1):
 *         cdef oct_visitors.CountTotalCells visitor
 *         visitor = oct_visitors.CountTotalCells(octree, domain_id)             # <<<<<<<<<<<<<<
 *         octree.visit_all_octs(self, visitor)
 *         return visitor.index
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_domain_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_octree));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_octree));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_octree));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalCells), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_visitor = ((struct __pyx_obj_2yt_8geometry_12oct_visitors_CountTotalCells *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":169
 *         cdef oct_visitors.CountTotalCells visitor
 *         visitor = oct_visitors.CountTotalCells(octree, domain_id)
 *         octree.visit_all_octs(self, visitor)             # <<<<<<<<<<<<<<
 *         return visitor.index
 * 
 */
  ((struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctreeContainer *)__pyx_v_octree->__pyx_vtab)->visit_all_octs(__pyx_v_octree, __pyx_v_self, ((struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *)__pyx_v_visitor), NULL);

  /* "yt/geometry/selection_routines.pyx":170
 *         visitor = oct_visitors.CountTotalCells(octree, domain_id)
 *         octree.visit_all_octs(self, visitor)
 *         return visitor.index             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_visitor->__pyx_base.index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":166
 *         return visitor.index
 * 
 *     def count_oct_cells(self, OctreeContainer octree, int domain_id = -1):             # <<<<<<<<<<<<<<
 *         cdef oct_visitors.CountTotalCells visitor
 *         visitor = oct_visitors.CountTotalCells(octree, domain_id)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.count_oct_cells", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_visitor);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":175
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void recursively_visit_octs(self, Oct *root,             # <<<<<<<<<<<<<<
 *                         np.float64_t pos[3], np.float64_t dds[3],
 *                         int level,
 */

static void __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_recursively_visit_octs(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_root, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds, int __pyx_v_level, struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *__pyx_v_visitor, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_recursively_visit_octs *__pyx_optional_args) {
  int __pyx_v_visit_covered = ((int)0);
  __pyx_t_5numpy_float64_t __pyx_v_LE[3];
  __pyx_t_5numpy_float64_t __pyx_v_RE[3];
  __pyx_t_5numpy_float64_t __pyx_v_sdds[3];
  __pyx_t_5numpy_float64_t __pyx_v_spos[3];
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_res;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_ch;
  int __pyx_v_increment;
  int __pyx_v_next_level;
  int __pyx_v_this_level;
  int __pyx_v_iter;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_recursively_visit_octs __pyx_t_7;
  long __pyx_t_8;
  __Pyx_RefNannySetupContext("recursively_visit_octs", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_visit_covered = __pyx_optional_args->visit_covered;
    }
  }

  /* "yt/geometry/selection_routines.pyx":192
 *         # Remember that pos is the *center* of the oct, and dds is the oct
 *         # width.  So to get to the edges, we add/subtract half of dds.
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             # sdds is the cell width
 *             sdds[i] = dds[i]/2.0
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":194
 *         for i in range(3):
 *             # sdds is the cell width
 *             sdds[i] = dds[i]/2.0             # <<<<<<<<<<<<<<
 *             LE[i] = pos[i] - dds[i]/2.0
 *             RE[i] = pos[i] + dds[i]/2.0
 */
    (__pyx_v_sdds[__pyx_v_i]) = ((__pyx_v_dds[__pyx_v_i]) / 2.0);

    /* "yt/geometry/selection_routines.pyx":195
 *             # sdds is the cell width
 *             sdds[i] = dds[i]/2.0
 *             LE[i] = pos[i] - dds[i]/2.0             # <<<<<<<<<<<<<<
 *             RE[i] = pos[i] + dds[i]/2.0
 *         #print LE[0], RE[0], LE[1], RE[1], LE[2], RE[2]
 */
    (__pyx_v_LE[__pyx_v_i]) = ((__pyx_v_pos[__pyx_v_i]) - ((__pyx_v_dds[__pyx_v_i]) / 2.0));

    /* "yt/geometry/selection_routines.pyx":196
 *             sdds[i] = dds[i]/2.0
 *             LE[i] = pos[i] - dds[i]/2.0
 *             RE[i] = pos[i] + dds[i]/2.0             # <<<<<<<<<<<<<<
 *         #print LE[0], RE[0], LE[1], RE[1], LE[2], RE[2]
 *         res = self.select_grid(LE, RE, level, root)
 */
    (__pyx_v_RE[__pyx_v_i]) = ((__pyx_v_pos[__pyx_v_i]) + ((__pyx_v_dds[__pyx_v_i]) / 2.0));
  }

  /* "yt/geometry/selection_routines.pyx":198
 *             RE[i] = pos[i] + dds[i]/2.0
 *         #print LE[0], RE[0], LE[1], RE[1], LE[2], RE[2]
 *         res = self.select_grid(LE, RE, level, root)             # <<<<<<<<<<<<<<
 *         if res == 1 and visitor.domain > 0 and root.domain != visitor.domain:
 *             res = -1
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.o = __pyx_v_root;
  __pyx_t_1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_grid(__pyx_v_self, __pyx_v_LE, __pyx_v_RE, __pyx_v_level, &__pyx_t_2); 
  __pyx_v_res = __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":199
 *         #print LE[0], RE[0], LE[1], RE[1], LE[2], RE[2]
 *         res = self.select_grid(LE, RE, level, root)
 *         if res == 1 and visitor.domain > 0 and root.domain != visitor.domain:             # <<<<<<<<<<<<<<
 *             res = -1
 *         cdef int increment = 1
 */
  __pyx_t_4 = ((__pyx_v_res == 1) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_v_visitor->domain > 0) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_v_root->domain != __pyx_v_visitor->domain) != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_3) {

    /* "yt/geometry/selection_routines.pyx":200
 *         res = self.select_grid(LE, RE, level, root)
 *         if res == 1 and visitor.domain > 0 and root.domain != visitor.domain:
 *             res = -1             # <<<<<<<<<<<<<<
 *         cdef int increment = 1
 *         cdef int next_level, this_level
 */
    __pyx_v_res = -1;

    /* "yt/geometry/selection_routines.pyx":199
 *         #print LE[0], RE[0], LE[1], RE[1], LE[2], RE[2]
 *         res = self.select_grid(LE, RE, level, root)
 *         if res == 1 and visitor.domain > 0 and root.domain != visitor.domain:             # <<<<<<<<<<<<<<
 *             res = -1
 *         cdef int increment = 1
 */
  }

  /* "yt/geometry/selection_routines.pyx":201
 *         if res == 1 and visitor.domain > 0 and root.domain != visitor.domain:
 *             res = -1
 *         cdef int increment = 1             # <<<<<<<<<<<<<<
 *         cdef int next_level, this_level
 *         # next_level: an int that says whether or not we can progress to children
 */
  __pyx_v_increment = 1;

  /* "yt/geometry/selection_routines.pyx":206
 *         # this_level: an int that says whether or not we can select from this
 *         # level
 *         next_level = this_level = 1             # <<<<<<<<<<<<<<
 *         if res == -1:
 *             # This happens when we do domain selection but the oct has
 */
  __pyx_v_next_level = 1;
  __pyx_v_this_level = 1;

  /* "yt/geometry/selection_routines.pyx":207
 *         # level
 *         next_level = this_level = 1
 *         if res == -1:             # <<<<<<<<<<<<<<
 *             # This happens when we do domain selection but the oct has
 *             # children.  This would allow an oct to pass to its children but
 */
  __pyx_t_3 = ((__pyx_v_res == -1L) != 0);
  if (__pyx_t_3) {

    /* "yt/geometry/selection_routines.pyx":211
 *             # children.  This would allow an oct to pass to its children but
 *             # not get accessed itself.
 *             next_level = 1             # <<<<<<<<<<<<<<
 *             this_level = 0
 *         elif level == self.max_level:
 */
    __pyx_v_next_level = 1;

    /* "yt/geometry/selection_routines.pyx":212
 *             # not get accessed itself.
 *             next_level = 1
 *             this_level = 0             # <<<<<<<<<<<<<<
 *         elif level == self.max_level:
 *             next_level = 0
 */
    __pyx_v_this_level = 0;

    /* "yt/geometry/selection_routines.pyx":207
 *         # level
 *         next_level = this_level = 1
 *         if res == -1:             # <<<<<<<<<<<<<<
 *             # This happens when we do domain selection but the oct has
 *             # children.  This would allow an oct to pass to its children but
 */
    goto __pyx_L9;
  }

  /* "yt/geometry/selection_routines.pyx":213
 *             next_level = 1
 *             this_level = 0
 *         elif level == self.max_level:             # <<<<<<<<<<<<<<
 *             next_level = 0
 *         elif level < self.min_level or level > self.max_level:
 */
  __pyx_t_3 = ((__pyx_v_level == __pyx_v_self->max_level) != 0);
  if (__pyx_t_3) {

    /* "yt/geometry/selection_routines.pyx":214
 *             this_level = 0
 *         elif level == self.max_level:
 *             next_level = 0             # <<<<<<<<<<<<<<
 *         elif level < self.min_level or level > self.max_level:
 *             this_level = 0
 */
    __pyx_v_next_level = 0;

    /* "yt/geometry/selection_routines.pyx":213
 *             next_level = 1
 *             this_level = 0
 *         elif level == self.max_level:             # <<<<<<<<<<<<<<
 *             next_level = 0
 *         elif level < self.min_level or level > self.max_level:
 */
    goto __pyx_L9;
  }

  /* "yt/geometry/selection_routines.pyx":215
 *         elif level == self.max_level:
 *             next_level = 0
 *         elif level < self.min_level or level > self.max_level:             # <<<<<<<<<<<<<<
 *             this_level = 0
 *         if res == 0 and this_level == 1:
 */
  __pyx_t_4 = ((__pyx_v_level < __pyx_v_self->min_level) != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_v_level > __pyx_v_self->max_level) != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L10_bool_binop_done:;
  if (__pyx_t_3) {

    /* "yt/geometry/selection_routines.pyx":216
 *             next_level = 0
 *         elif level < self.min_level or level > self.max_level:
 *             this_level = 0             # <<<<<<<<<<<<<<
 *         if res == 0 and this_level == 1:
 *             return
 */
    __pyx_v_this_level = 0;

    /* "yt/geometry/selection_routines.pyx":215
 *         elif level == self.max_level:
 *             next_level = 0
 *         elif level < self.min_level or level > self.max_level:             # <<<<<<<<<<<<<<
 *             this_level = 0
 *         if res == 0 and this_level == 1:
 */
  }
  __pyx_L9:;

  /* "yt/geometry/selection_routines.pyx":217
 *         elif level < self.min_level or level > self.max_level:
 *             this_level = 0
 *         if res == 0 and this_level == 1:             # <<<<<<<<<<<<<<
 *             return
 *         # Now we visit all our children.  We subtract off sdds for the first
 */
  __pyx_t_4 = ((__pyx_v_res == 0) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L13_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_v_this_level == 1) != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L13_bool_binop_done:;
  if (__pyx_t_3) {

    /* "yt/geometry/selection_routines.pyx":218
 *             this_level = 0
 *         if res == 0 and this_level == 1:
 *             return             # <<<<<<<<<<<<<<
 *         # Now we visit all our children.  We subtract off sdds for the first
 *         # pass because we center it on the first cell.
 */
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":217
 *         elif level < self.min_level or level > self.max_level:
 *             this_level = 0
 *         if res == 0 and this_level == 1:             # <<<<<<<<<<<<<<
 *             return
 *         # Now we visit all our children.  We subtract off sdds for the first
 */
  }

  /* "yt/geometry/selection_routines.pyx":221
 *         # Now we visit all our children.  We subtract off sdds for the first
 *         # pass because we center it on the first cell.
 *         cdef int iter = 1 - visit_covered # 2 if 1, 1 if 0.             # <<<<<<<<<<<<<<
 *         # So the order here goes like so.  If visit_covered is 1, which usually
 *         # comes from "partial_coverage", we visit the components of a zone even
 */
  __pyx_v_iter = (1 - __pyx_v_visit_covered);

  /* "yt/geometry/selection_routines.pyx":228
 *         # visit potentially covered cells.  The next time through, we visit
 *         # child cells.
 *         while iter < 2:             # <<<<<<<<<<<<<<
 *             spos[0] = pos[0] - sdds[0]/2.0
 *             for i in range(2):
 */
  while (1) {
    __pyx_t_3 = ((__pyx_v_iter < 2) != 0);
    if (!__pyx_t_3) break;

    /* "yt/geometry/selection_routines.pyx":229
 *         # child cells.
 *         while iter < 2:
 *             spos[0] = pos[0] - sdds[0]/2.0             # <<<<<<<<<<<<<<
 *             for i in range(2):
 *                 spos[1] = pos[1] - sdds[1]/2.0
 */
    (__pyx_v_spos[0]) = ((__pyx_v_pos[0]) - ((__pyx_v_sdds[0]) / 2.0));

    /* "yt/geometry/selection_routines.pyx":230
 *         while iter < 2:
 *             spos[0] = pos[0] - sdds[0]/2.0
 *             for i in range(2):             # <<<<<<<<<<<<<<
 *                 spos[1] = pos[1] - sdds[1]/2.0
 *                 for j in range(2):
 */
    for (__pyx_t_1 = 0; __pyx_t_1 < 2; __pyx_t_1+=1) {
      __pyx_v_i = __pyx_t_1;

      /* "yt/geometry/selection_routines.pyx":231
 *             spos[0] = pos[0] - sdds[0]/2.0
 *             for i in range(2):
 *                 spos[1] = pos[1] - sdds[1]/2.0             # <<<<<<<<<<<<<<
 *                 for j in range(2):
 *                     spos[2] = pos[2] - sdds[2]/2.0
 */
      (__pyx_v_spos[1]) = ((__pyx_v_pos[1]) - ((__pyx_v_sdds[1]) / 2.0));

      /* "yt/geometry/selection_routines.pyx":232
 *             for i in range(2):
 *                 spos[1] = pos[1] - sdds[1]/2.0
 *                 for j in range(2):             # <<<<<<<<<<<<<<
 *                     spos[2] = pos[2] - sdds[2]/2.0
 *                     for k in range(2):
 */
      for (__pyx_t_5 = 0; __pyx_t_5 < 2; __pyx_t_5+=1) {
        __pyx_v_j = __pyx_t_5;

        /* "yt/geometry/selection_routines.pyx":233
 *                 spos[1] = pos[1] - sdds[1]/2.0
 *                 for j in range(2):
 *                     spos[2] = pos[2] - sdds[2]/2.0             # <<<<<<<<<<<<<<
 *                     for k in range(2):
 *                         ch = NULL
 */
        (__pyx_v_spos[2]) = ((__pyx_v_pos[2]) - ((__pyx_v_sdds[2]) / 2.0));

        /* "yt/geometry/selection_routines.pyx":234
 *                 for j in range(2):
 *                     spos[2] = pos[2] - sdds[2]/2.0
 *                     for k in range(2):             # <<<<<<<<<<<<<<
 *                         ch = NULL
 *                         # We only supply a child if we are actually going to
 */
        for (__pyx_t_6 = 0; __pyx_t_6 < 2; __pyx_t_6+=1) {
          __pyx_v_k = __pyx_t_6;

          /* "yt/geometry/selection_routines.pyx":235
 *                     spos[2] = pos[2] - sdds[2]/2.0
 *                     for k in range(2):
 *                         ch = NULL             # <<<<<<<<<<<<<<
 *                         # We only supply a child if we are actually going to
 *                         # look at the next level.
 */
          __pyx_v_ch = NULL;

          /* "yt/geometry/selection_routines.pyx":238
 *                         # We only supply a child if we are actually going to
 *                         # look at the next level.
 *                         if root.children != NULL and next_level == 1:             # <<<<<<<<<<<<<<
 *                             ch = root.children[cind(i, j, k)]
 *                         if iter == 1 and next_level == 1 and ch != NULL:
 */
          __pyx_t_4 = ((__pyx_v_root->children != NULL) != 0);
          if (__pyx_t_4) {
          } else {
            __pyx_t_3 = __pyx_t_4;
            goto __pyx_L24_bool_binop_done;
          }
          __pyx_t_4 = ((__pyx_v_next_level == 1) != 0);
          __pyx_t_3 = __pyx_t_4;
          __pyx_L24_bool_binop_done:;
          if (__pyx_t_3) {

            /* "yt/geometry/selection_routines.pyx":239
 *                         # look at the next level.
 *                         if root.children != NULL and next_level == 1:
 *                             ch = root.children[cind(i, j, k)]             # <<<<<<<<<<<<<<
 *                         if iter == 1 and next_level == 1 and ch != NULL:
 *                             # Note that visitor.pos is always going to be the
 */
            __pyx_v_ch = (__pyx_v_root->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]);

            /* "yt/geometry/selection_routines.pyx":238
 *                         # We only supply a child if we are actually going to
 *                         # look at the next level.
 *                         if root.children != NULL and next_level == 1:             # <<<<<<<<<<<<<<
 *                             ch = root.children[cind(i, j, k)]
 *                         if iter == 1 and next_level == 1 and ch != NULL:
 */
          }

          /* "yt/geometry/selection_routines.pyx":240
 *                         if root.children != NULL and next_level == 1:
 *                             ch = root.children[cind(i, j, k)]
 *                         if iter == 1 and next_level == 1 and ch != NULL:             # <<<<<<<<<<<<<<
 *                             # Note that visitor.pos is always going to be the
 *                             # position of the Oct -- it is *not* always going
 */
          __pyx_t_4 = ((__pyx_v_iter == 1) != 0);
          if (__pyx_t_4) {
          } else {
            __pyx_t_3 = __pyx_t_4;
            goto __pyx_L27_bool_binop_done;
          }
          __pyx_t_4 = ((__pyx_v_next_level == 1) != 0);
          if (__pyx_t_4) {
          } else {
            __pyx_t_3 = __pyx_t_4;
            goto __pyx_L27_bool_binop_done;
          }
          __pyx_t_4 = ((__pyx_v_ch != NULL) != 0);
          __pyx_t_3 = __pyx_t_4;
          __pyx_L27_bool_binop_done:;
          if (__pyx_t_3) {

            /* "yt/geometry/selection_routines.pyx":245
 *                             # to be the same as the position of the cell under
 *                             # investigation.
 *                             visitor.pos[0] = (visitor.pos[0] << 1) + i             # <<<<<<<<<<<<<<
 *                             visitor.pos[1] = (visitor.pos[1] << 1) + j
 *                             visitor.pos[2] = (visitor.pos[2] << 1) + k
 */
            (__pyx_v_visitor->pos[0]) = (((__pyx_v_visitor->pos[0]) << 1) + __pyx_v_i);

            /* "yt/geometry/selection_routines.pyx":246
 *                             # investigation.
 *                             visitor.pos[0] = (visitor.pos[0] << 1) + i
 *                             visitor.pos[1] = (visitor.pos[1] << 1) + j             # <<<<<<<<<<<<<<
 *                             visitor.pos[2] = (visitor.pos[2] << 1) + k
 *                             visitor.level += 1
 */
            (__pyx_v_visitor->pos[1]) = (((__pyx_v_visitor->pos[1]) << 1) + __pyx_v_j);

            /* "yt/geometry/selection_routines.pyx":247
 *                             visitor.pos[0] = (visitor.pos[0] << 1) + i
 *                             visitor.pos[1] = (visitor.pos[1] << 1) + j
 *                             visitor.pos[2] = (visitor.pos[2] << 1) + k             # <<<<<<<<<<<<<<
 *                             visitor.level += 1
 *                             self.recursively_visit_octs(
 */
            (__pyx_v_visitor->pos[2]) = (((__pyx_v_visitor->pos[2]) << 1) + __pyx_v_k);

            /* "yt/geometry/selection_routines.pyx":248
 *                             visitor.pos[1] = (visitor.pos[1] << 1) + j
 *                             visitor.pos[2] = (visitor.pos[2] << 1) + k
 *                             visitor.level += 1             # <<<<<<<<<<<<<<
 *                             self.recursively_visit_octs(
 *                                 ch, spos, sdds, level + 1, visitor,
 */
            __pyx_v_visitor->level = (__pyx_v_visitor->level + 1);

            /* "yt/geometry/selection_routines.pyx":249
 *                             visitor.pos[2] = (visitor.pos[2] << 1) + k
 *                             visitor.level += 1
 *                             self.recursively_visit_octs(             # <<<<<<<<<<<<<<
 *                                 ch, spos, sdds, level + 1, visitor,
 *                                 visit_covered)
 */
            __pyx_t_7.__pyx_n = 1;
            __pyx_t_7.visit_covered = __pyx_v_visit_covered;
            ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->recursively_visit_octs(__pyx_v_self, __pyx_v_ch, __pyx_v_spos, __pyx_v_sdds, (__pyx_v_level + 1), __pyx_v_visitor, &__pyx_t_7); 

            /* "yt/geometry/selection_routines.pyx":252
 *                                 ch, spos, sdds, level + 1, visitor,
 *                                 visit_covered)
 *                             visitor.pos[0] = (visitor.pos[0] >> 1)             # <<<<<<<<<<<<<<
 *                             visitor.pos[1] = (visitor.pos[1] >> 1)
 *                             visitor.pos[2] = (visitor.pos[2] >> 1)
 */
            (__pyx_v_visitor->pos[0]) = ((__pyx_v_visitor->pos[0]) >> 1);

            /* "yt/geometry/selection_routines.pyx":253
 *                                 visit_covered)
 *                             visitor.pos[0] = (visitor.pos[0] >> 1)
 *                             visitor.pos[1] = (visitor.pos[1] >> 1)             # <<<<<<<<<<<<<<
 *                             visitor.pos[2] = (visitor.pos[2] >> 1)
 *                             visitor.level -= 1
 */
            (__pyx_v_visitor->pos[1]) = ((__pyx_v_visitor->pos[1]) >> 1);

            /* "yt/geometry/selection_routines.pyx":254
 *                             visitor.pos[0] = (visitor.pos[0] >> 1)
 *                             visitor.pos[1] = (visitor.pos[1] >> 1)
 *                             visitor.pos[2] = (visitor.pos[2] >> 1)             # <<<<<<<<<<<<<<
 *                             visitor.level -= 1
 *                         elif this_level == 1 and visitor.oref > 0:
 */
            (__pyx_v_visitor->pos[2]) = ((__pyx_v_visitor->pos[2]) >> 1);

            /* "yt/geometry/selection_routines.pyx":255
 *                             visitor.pos[1] = (visitor.pos[1] >> 1)
 *                             visitor.pos[2] = (visitor.pos[2] >> 1)
 *                             visitor.level -= 1             # <<<<<<<<<<<<<<
 *                         elif this_level == 1 and visitor.oref > 0:
 *                             visitor.global_index += increment
 */
            __pyx_v_visitor->level = (__pyx_v_visitor->level - 1);

            /* "yt/geometry/selection_routines.pyx":240
 *                         if root.children != NULL and next_level == 1:
 *                             ch = root.children[cind(i, j, k)]
 *                         if iter == 1 and next_level == 1 and ch != NULL:             # <<<<<<<<<<<<<<
 *                             # Note that visitor.pos is always going to be the
 *                             # position of the Oct -- it is *not* always going
 */
            goto __pyx_L26;
          }

          /* "yt/geometry/selection_routines.pyx":256
 *                             visitor.pos[2] = (visitor.pos[2] >> 1)
 *                             visitor.level -= 1
 *                         elif this_level == 1 and visitor.oref > 0:             # <<<<<<<<<<<<<<
 *                             visitor.global_index += increment
 *                             increment = 0
 */
          __pyx_t_4 = ((__pyx_v_this_level == 1) != 0);
          if (__pyx_t_4) {
          } else {
            __pyx_t_3 = __pyx_t_4;
            goto __pyx_L30_bool_binop_done;
          }
          __pyx_t_4 = ((__pyx_v_visitor->oref > 0) != 0);
          __pyx_t_3 = __pyx_t_4;
          __pyx_L30_bool_binop_done:;
          if (__pyx_t_3) {

            /* "yt/geometry/selection_routines.pyx":257
 *                             visitor.level -= 1
 *                         elif this_level == 1 and visitor.oref > 0:
 *                             visitor.global_index += increment             # <<<<<<<<<<<<<<
 *                             increment = 0
 *                             self.visit_oct_cells(root, ch, spos, sdds,
 */
            __pyx_v_visitor->global_index = (__pyx_v_visitor->global_index + __pyx_v_increment);

            /* "yt/geometry/selection_routines.pyx":258
 *                         elif this_level == 1 and visitor.oref > 0:
 *                             visitor.global_index += increment
 *                             increment = 0             # <<<<<<<<<<<<<<
 *                             self.visit_oct_cells(root, ch, spos, sdds,
 *                                                  visitor, i, j, k)
 */
            __pyx_v_increment = 0;

            /* "yt/geometry/selection_routines.pyx":259
 *                             visitor.global_index += increment
 *                             increment = 0
 *                             self.visit_oct_cells(root, ch, spos, sdds,             # <<<<<<<<<<<<<<
 *                                                  visitor, i, j, k)
 *                         elif this_level == 1 and increment == 1:
 */
            ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->visit_oct_cells(__pyx_v_self, __pyx_v_root, __pyx_v_ch, __pyx_v_spos, __pyx_v_sdds, __pyx_v_visitor, __pyx_v_i, __pyx_v_j, __pyx_v_k);

            /* "yt/geometry/selection_routines.pyx":256
 *                             visitor.pos[2] = (visitor.pos[2] >> 1)
 *                             visitor.level -= 1
 *                         elif this_level == 1 and visitor.oref > 0:             # <<<<<<<<<<<<<<
 *                             visitor.global_index += increment
 *                             increment = 0
 */
            goto __pyx_L26;
          }

          /* "yt/geometry/selection_routines.pyx":261
 *                             self.visit_oct_cells(root, ch, spos, sdds,
 *                                                  visitor, i, j, k)
 *                         elif this_level == 1 and increment == 1:             # <<<<<<<<<<<<<<
 *                             visitor.global_index += increment
 *                             increment = 0
 */
          __pyx_t_4 = ((__pyx_v_this_level == 1) != 0);
          if (__pyx_t_4) {
          } else {
            __pyx_t_3 = __pyx_t_4;
            goto __pyx_L32_bool_binop_done;
          }
          __pyx_t_4 = ((__pyx_v_increment == 1) != 0);
          __pyx_t_3 = __pyx_t_4;
          __pyx_L32_bool_binop_done:;
          if (__pyx_t_3) {

            /* "yt/geometry/selection_routines.pyx":262
 *                                                  visitor, i, j, k)
 *                         elif this_level == 1 and increment == 1:
 *                             visitor.global_index += increment             # <<<<<<<<<<<<<<
 *                             increment = 0
 *                             visitor.ind[0] = visitor.ind[1] = visitor.ind[2] = 0
 */
            __pyx_v_visitor->global_index = (__pyx_v_visitor->global_index + __pyx_v_increment);

            /* "yt/geometry/selection_routines.pyx":263
 *                         elif this_level == 1 and increment == 1:
 *                             visitor.global_index += increment
 *                             increment = 0             # <<<<<<<<<<<<<<
 *                             visitor.ind[0] = visitor.ind[1] = visitor.ind[2] = 0
 *                             visitor.visit(root, 1)
 */
            __pyx_v_increment = 0;

            /* "yt/geometry/selection_routines.pyx":264
 *                             visitor.global_index += increment
 *                             increment = 0
 *                             visitor.ind[0] = visitor.ind[1] = visitor.ind[2] = 0             # <<<<<<<<<<<<<<
 *                             visitor.visit(root, 1)
 *                         spos[2] += sdds[2]
 */
            (__pyx_v_visitor->ind[0]) = 0;
            (__pyx_v_visitor->ind[1]) = 0;
            (__pyx_v_visitor->ind[2]) = 0;

            /* "yt/geometry/selection_routines.pyx":265
 *                             increment = 0
 *                             visitor.ind[0] = visitor.ind[1] = visitor.ind[2] = 0
 *                             visitor.visit(root, 1)             # <<<<<<<<<<<<<<
 *                         spos[2] += sdds[2]
 *                     spos[1] += sdds[1]
 */
            ((struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor *)__pyx_v_visitor->__pyx_vtab)->visit(__pyx_v_visitor, __pyx_v_root, 1);

            /* "yt/geometry/selection_routines.pyx":261
 *                             self.visit_oct_cells(root, ch, spos, sdds,
 *                                                  visitor, i, j, k)
 *                         elif this_level == 1 and increment == 1:             # <<<<<<<<<<<<<<
 *                             visitor.global_index += increment
 *                             increment = 0
 */
          }
          __pyx_L26:;

          /* "yt/geometry/selection_routines.pyx":266
 *                             visitor.ind[0] = visitor.ind[1] = visitor.ind[2] = 0
 *                             visitor.visit(root, 1)
 *                         spos[2] += sdds[2]             # <<<<<<<<<<<<<<
 *                     spos[1] += sdds[1]
 *                 spos[0] += sdds[0]
 */
          __pyx_t_8 = 2;
          (__pyx_v_spos[__pyx_t_8]) = ((__pyx_v_spos[__pyx_t_8]) + (__pyx_v_sdds[2]));
        }

        /* "yt/geometry/selection_routines.pyx":267
 *                             visitor.visit(root, 1)
 *                         spos[2] += sdds[2]
 *                     spos[1] += sdds[1]             # <<<<<<<<<<<<<<
 *                 spos[0] += sdds[0]
 *             this_level = 0 # We turn this off for the second pass.
 */
        __pyx_t_8 = 1;
        (__pyx_v_spos[__pyx_t_8]) = ((__pyx_v_spos[__pyx_t_8]) + (__pyx_v_sdds[1]));
      }

      /* "yt/geometry/selection_routines.pyx":268
 *                         spos[2] += sdds[2]
 *                     spos[1] += sdds[1]
 *                 spos[0] += sdds[0]             # <<<<<<<<<<<<<<
 *             this_level = 0 # We turn this off for the second pass.
 *             iter += 1
 */
      __pyx_t_8 = 0;
      (__pyx_v_spos[__pyx_t_8]) = ((__pyx_v_spos[__pyx_t_8]) + (__pyx_v_sdds[0]));
    }

    /* "yt/geometry/selection_routines.pyx":269
 *                     spos[1] += sdds[1]
 *                 spos[0] += sdds[0]
 *             this_level = 0 # We turn this off for the second pass.             # <<<<<<<<<<<<<<
 *             iter += 1
 * 
 */
    __pyx_v_this_level = 0;

    /* "yt/geometry/selection_routines.pyx":270
 *                 spos[0] += sdds[0]
 *             this_level = 0 # We turn this off for the second pass.
 *             iter += 1             # <<<<<<<<<<<<<<
 * 
 *     cdef void visit_oct_cells(self, Oct *root, Oct *ch,
 */
    __pyx_v_iter = (__pyx_v_iter + 1);
  }

  /* "yt/geometry/selection_routines.pyx":175
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void recursively_visit_octs(self, Oct *root,             # <<<<<<<<<<<<<<
 *                         np.float64_t pos[3], np.float64_t dds[3],
 *                         int level,
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/selection_routines.pyx":272
 *             iter += 1
 * 
 *     cdef void visit_oct_cells(self, Oct *root, Oct *ch,             # <<<<<<<<<<<<<<
 *                               np.float64_t spos[3], np.float64_t sdds[3],
 *                               OctVisitor visitor, int i, int j, int k):
 */

static void __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_visit_oct_cells(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_root, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_ch, __pyx_t_5numpy_float64_t *__pyx_v_spos, __pyx_t_5numpy_float64_t *__pyx_v_sdds, struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *__pyx_v_visitor, int __pyx_v_i, int __pyx_v_j, int __pyx_v_k) {
  int __pyx_v_selected;
  __pyx_t_5numpy_float64_t __pyx_v_dds[3];
  __pyx_t_5numpy_float64_t __pyx_v_pos[3];
  int __pyx_v_ci;
  int __pyx_v_cj;
  int __pyx_v_ck;
  int __pyx_v_nr;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  long __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("visit_oct_cells", 0);

  /* "yt/geometry/selection_routines.pyx":278
 *         # This saves us some funny-business.
 *         cdef int selected
 *         if visitor.oref == 1:             # <<<<<<<<<<<<<<
 *             selected = self.select_cell(spos, sdds)
 *             if ch != NULL:
 */
  __pyx_t_1 = ((__pyx_v_visitor->oref == 1) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":279
 *         cdef int selected
 *         if visitor.oref == 1:
 *             selected = self.select_cell(spos, sdds)             # <<<<<<<<<<<<<<
 *             if ch != NULL:
 *                 selected *= self.overlap_cells
 */
    __pyx_v_selected = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_cell(__pyx_v_self, __pyx_v_spos, __pyx_v_sdds);

    /* "yt/geometry/selection_routines.pyx":280
 *         if visitor.oref == 1:
 *             selected = self.select_cell(spos, sdds)
 *             if ch != NULL:             # <<<<<<<<<<<<<<
 *                 selected *= self.overlap_cells
 *             # visitor.ind refers to the cell, not to the oct.
 */
    __pyx_t_1 = ((__pyx_v_ch != NULL) != 0);
    if (__pyx_t_1) {

      /* "yt/geometry/selection_routines.pyx":281
 *             selected = self.select_cell(spos, sdds)
 *             if ch != NULL:
 *                 selected *= self.overlap_cells             # <<<<<<<<<<<<<<
 *             # visitor.ind refers to the cell, not to the oct.
 *             visitor.ind[0] = i
 */
      __pyx_v_selected = (__pyx_v_selected * __pyx_v_self->overlap_cells);

      /* "yt/geometry/selection_routines.pyx":280
 *         if visitor.oref == 1:
 *             selected = self.select_cell(spos, sdds)
 *             if ch != NULL:             # <<<<<<<<<<<<<<
 *                 selected *= self.overlap_cells
 *             # visitor.ind refers to the cell, not to the oct.
 */
    }

    /* "yt/geometry/selection_routines.pyx":283
 *                 selected *= self.overlap_cells
 *             # visitor.ind refers to the cell, not to the oct.
 *             visitor.ind[0] = i             # <<<<<<<<<<<<<<
 *             visitor.ind[1] = j
 *             visitor.ind[2] = k
 */
    (__pyx_v_visitor->ind[0]) = __pyx_v_i;

    /* "yt/geometry/selection_routines.pyx":284
 *             # visitor.ind refers to the cell, not to the oct.
 *             visitor.ind[0] = i
 *             visitor.ind[1] = j             # <<<<<<<<<<<<<<
 *             visitor.ind[2] = k
 *             visitor.visit(root, selected)
 */
    (__pyx_v_visitor->ind[1]) = __pyx_v_j;

    /* "yt/geometry/selection_routines.pyx":285
 *             visitor.ind[0] = i
 *             visitor.ind[1] = j
 *             visitor.ind[2] = k             # <<<<<<<<<<<<<<
 *             visitor.visit(root, selected)
 *             return
 */
    (__pyx_v_visitor->ind[2]) = __pyx_v_k;

    /* "yt/geometry/selection_routines.pyx":286
 *             visitor.ind[1] = j
 *             visitor.ind[2] = k
 *             visitor.visit(root, selected)             # <<<<<<<<<<<<<<
 *             return
 *         # Okay, now that we've got that out of the way, we have to do some
 */
    ((struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor *)__pyx_v_visitor->__pyx_vtab)->visit(__pyx_v_visitor, __pyx_v_root, __pyx_v_selected);

    /* "yt/geometry/selection_routines.pyx":287
 *             visitor.ind[2] = k
 *             visitor.visit(root, selected)
 *             return             # <<<<<<<<<<<<<<
 *         # Okay, now that we've got that out of the way, we have to do some
 *         # other checks here.  In this case, spos[] is the position of the
 */
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":278
 *         # This saves us some funny-business.
 *         cdef int selected
 *         if visitor.oref == 1:             # <<<<<<<<<<<<<<
 *             selected = self.select_cell(spos, sdds)
 *             if ch != NULL:
 */
  }

  /* "yt/geometry/selection_routines.pyx":296
 *         cdef np.float64_t pos[3]
 *         cdef int ci, cj, ck
 *         cdef int nr = (1 << (visitor.oref - 1))             # <<<<<<<<<<<<<<
 *         for ci in range(3):
 *             dds[ci] = sdds[ci] / nr
 */
  __pyx_v_nr = (1 << (__pyx_v_visitor->oref - 1));

  /* "yt/geometry/selection_routines.pyx":297
 *         cdef int ci, cj, ck
 *         cdef int nr = (1 << (visitor.oref - 1))
 *         for ci in range(3):             # <<<<<<<<<<<<<<
 *             dds[ci] = sdds[ci] / nr
 *         # Boot strap at the first index.
 */
  for (__pyx_t_2 = 0; __pyx_t_2 < 3; __pyx_t_2+=1) {
    __pyx_v_ci = __pyx_t_2;

    /* "yt/geometry/selection_routines.pyx":298
 *         cdef int nr = (1 << (visitor.oref - 1))
 *         for ci in range(3):
 *             dds[ci] = sdds[ci] / nr             # <<<<<<<<<<<<<<
 *         # Boot strap at the first index.
 *         pos[0] = (spos[0] - sdds[0]/2.0) + dds[0] * 0.5
 */
    if (unlikely(__pyx_v_nr == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 298, __pyx_L1_error)
    }
    (__pyx_v_dds[__pyx_v_ci]) = ((__pyx_v_sdds[__pyx_v_ci]) / __pyx_v_nr);
  }

  /* "yt/geometry/selection_routines.pyx":300
 *             dds[ci] = sdds[ci] / nr
 *         # Boot strap at the first index.
 *         pos[0] = (spos[0] - sdds[0]/2.0) + dds[0] * 0.5             # <<<<<<<<<<<<<<
 *         for ci in range(nr):
 *             pos[1] = (spos[1] - sdds[1]/2.0) + dds[1] * 0.5
 */
  (__pyx_v_pos[0]) = (((__pyx_v_spos[0]) - ((__pyx_v_sdds[0]) / 2.0)) + ((__pyx_v_dds[0]) * 0.5));

  /* "yt/geometry/selection_routines.pyx":301
 *         # Boot strap at the first index.
 *         pos[0] = (spos[0] - sdds[0]/2.0) + dds[0] * 0.5
 *         for ci in range(nr):             # <<<<<<<<<<<<<<
 *             pos[1] = (spos[1] - sdds[1]/2.0) + dds[1] * 0.5
 *             for cj in range(nr):
 */
  __pyx_t_2 = __pyx_v_nr;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_ci = __pyx_t_4;

    /* "yt/geometry/selection_routines.pyx":302
 *         pos[0] = (spos[0] - sdds[0]/2.0) + dds[0] * 0.5
 *         for ci in range(nr):
 *             pos[1] = (spos[1] - sdds[1]/2.0) + dds[1] * 0.5             # <<<<<<<<<<<<<<
 *             for cj in range(nr):
 *                 pos[2] = (spos[2] - sdds[2]/2.0) + dds[2] * 0.5
 */
    (__pyx_v_pos[1]) = (((__pyx_v_spos[1]) - ((__pyx_v_sdds[1]) / 2.0)) + ((__pyx_v_dds[1]) * 0.5));

    /* "yt/geometry/selection_routines.pyx":303
 *         for ci in range(nr):
 *             pos[1] = (spos[1] - sdds[1]/2.0) + dds[1] * 0.5
 *             for cj in range(nr):             # <<<<<<<<<<<<<<
 *                 pos[2] = (spos[2] - sdds[2]/2.0) + dds[2] * 0.5
 *                 for ck in range(nr):
 */
    __pyx_t_5 = __pyx_v_nr;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_cj = __pyx_t_7;

      /* "yt/geometry/selection_routines.pyx":304
 *             pos[1] = (spos[1] - sdds[1]/2.0) + dds[1] * 0.5
 *             for cj in range(nr):
 *                 pos[2] = (spos[2] - sdds[2]/2.0) + dds[2] * 0.5             # <<<<<<<<<<<<<<
 *                 for ck in range(nr):
 *                     selected = self.select_cell(pos, dds)
 */
      (__pyx_v_pos[2]) = (((__pyx_v_spos[2]) - ((__pyx_v_sdds[2]) / 2.0)) + ((__pyx_v_dds[2]) * 0.5));

      /* "yt/geometry/selection_routines.pyx":305
 *             for cj in range(nr):
 *                 pos[2] = (spos[2] - sdds[2]/2.0) + dds[2] * 0.5
 *                 for ck in range(nr):             # <<<<<<<<<<<<<<
 *                     selected = self.select_cell(pos, dds)
 *                     if ch != NULL:
 */
      __pyx_t_8 = __pyx_v_nr;
      __pyx_t_9 = __pyx_t_8;
      for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
        __pyx_v_ck = __pyx_t_10;

        /* "yt/geometry/selection_routines.pyx":306
 *                 pos[2] = (spos[2] - sdds[2]/2.0) + dds[2] * 0.5
 *                 for ck in range(nr):
 *                     selected = self.select_cell(pos, dds)             # <<<<<<<<<<<<<<
 *                     if ch != NULL:
 *                         selected *= self.overlap_cells
 */
        __pyx_v_selected = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_cell(__pyx_v_self, __pyx_v_pos, __pyx_v_dds);

        /* "yt/geometry/selection_routines.pyx":307
 *                 for ck in range(nr):
 *                     selected = self.select_cell(pos, dds)
 *                     if ch != NULL:             # <<<<<<<<<<<<<<
 *                         selected *= self.overlap_cells
 *                     visitor.ind[0] = ci + i * nr
 */
        __pyx_t_1 = ((__pyx_v_ch != NULL) != 0);
        if (__pyx_t_1) {

          /* "yt/geometry/selection_routines.pyx":308
 *                     selected = self.select_cell(pos, dds)
 *                     if ch != NULL:
 *                         selected *= self.overlap_cells             # <<<<<<<<<<<<<<
 *                     visitor.ind[0] = ci + i * nr
 *                     visitor.ind[1] = cj + j * nr
 */
          __pyx_v_selected = (__pyx_v_selected * __pyx_v_self->overlap_cells);

          /* "yt/geometry/selection_routines.pyx":307
 *                 for ck in range(nr):
 *                     selected = self.select_cell(pos, dds)
 *                     if ch != NULL:             # <<<<<<<<<<<<<<
 *                         selected *= self.overlap_cells
 *                     visitor.ind[0] = ci + i * nr
 */
        }

        /* "yt/geometry/selection_routines.pyx":309
 *                     if ch != NULL:
 *                         selected *= self.overlap_cells
 *                     visitor.ind[0] = ci + i * nr             # <<<<<<<<<<<<<<
 *                     visitor.ind[1] = cj + j * nr
 *                     visitor.ind[2] = ck + k * nr
 */
        (__pyx_v_visitor->ind[0]) = (__pyx_v_ci + (__pyx_v_i * __pyx_v_nr));

        /* "yt/geometry/selection_routines.pyx":310
 *                         selected *= self.overlap_cells
 *                     visitor.ind[0] = ci + i * nr
 *                     visitor.ind[1] = cj + j * nr             # <<<<<<<<<<<<<<
 *                     visitor.ind[2] = ck + k * nr
 *                     visitor.visit(root, selected)
 */
        (__pyx_v_visitor->ind[1]) = (__pyx_v_cj + (__pyx_v_j * __pyx_v_nr));

        /* "yt/geometry/selection_routines.pyx":311
 *                     visitor.ind[0] = ci + i * nr
 *                     visitor.ind[1] = cj + j * nr
 *                     visitor.ind[2] = ck + k * nr             # <<<<<<<<<<<<<<
 *                     visitor.visit(root, selected)
 *                     pos[2] += dds[2]
 */
        (__pyx_v_visitor->ind[2]) = (__pyx_v_ck + (__pyx_v_k * __pyx_v_nr));

        /* "yt/geometry/selection_routines.pyx":312
 *                     visitor.ind[1] = cj + j * nr
 *                     visitor.ind[2] = ck + k * nr
 *                     visitor.visit(root, selected)             # <<<<<<<<<<<<<<
 *                     pos[2] += dds[2]
 *                 pos[1] += dds[1]
 */
        ((struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor *)__pyx_v_visitor->__pyx_vtab)->visit(__pyx_v_visitor, __pyx_v_root, __pyx_v_selected);

        /* "yt/geometry/selection_routines.pyx":313
 *                     visitor.ind[2] = ck + k * nr
 *                     visitor.visit(root, selected)
 *                     pos[2] += dds[2]             # <<<<<<<<<<<<<<
 *                 pos[1] += dds[1]
 *             pos[0] += dds[0]
 */
        __pyx_t_11 = 2;
        (__pyx_v_pos[__pyx_t_11]) = ((__pyx_v_pos[__pyx_t_11]) + (__pyx_v_dds[2]));
      }

      /* "yt/geometry/selection_routines.pyx":314
 *                     visitor.visit(root, selected)
 *                     pos[2] += dds[2]
 *                 pos[1] += dds[1]             # <<<<<<<<<<<<<<
 *             pos[0] += dds[0]
 * 
 */
      __pyx_t_11 = 1;
      (__pyx_v_pos[__pyx_t_11]) = ((__pyx_v_pos[__pyx_t_11]) + (__pyx_v_dds[1]));
    }

    /* "yt/geometry/selection_routines.pyx":315
 *                     pos[2] += dds[2]
 *                 pos[1] += dds[1]
 *             pos[0] += dds[0]             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    __pyx_t_11 = 0;
    (__pyx_v_pos[__pyx_t_11]) = ((__pyx_v_pos[__pyx_t_11]) + (__pyx_v_dds[0]));
  }

  /* "yt/geometry/selection_routines.pyx":272
 *             iter += 1
 * 
 *     cdef void visit_oct_cells(self, Oct *root, Oct *ch,             # <<<<<<<<<<<<<<
 *                               np.float64_t spos[3], np.float64_t sdds[3],
 *                               OctVisitor visitor, int i, int j, int k):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("yt.geometry.selection_routines.SelectorObject.visit_oct_cells", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/selection_routines.pyx":320
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3],
 *                                np.int32_t level, Oct *o = NULL) nogil:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid *__pyx_optional_args) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  if (__pyx_optional_args) {
  }

  /* "yt/geometry/selection_routines.pyx":323
 *                                np.float64_t right_edge[3],
 *                                np.int32_t level, Oct *o = NULL) nogil:
 *         if level < self.min_level or level > self.max_level: return 0             # <<<<<<<<<<<<<<
 *         return self.select_bbox(left_edge, right_edge)
 * 
 */
  __pyx_t_2 = ((__pyx_v_level < __pyx_v_self->min_level) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_level > __pyx_v_self->max_level) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":324
 *                                np.int32_t level, Oct *o = NULL) nogil:
 *         if level < self.min_level or level > self.max_level: return 0
 *         return self.select_bbox(left_edge, right_edge)             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 */
  __pyx_r = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_bbox(__pyx_v_self, __pyx_v_left_edge, __pyx_v_right_edge);
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":320
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3],
 *                                np.int32_t level, Oct *o = NULL) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":326
 *         return self.select_bbox(left_edge, right_edge)
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_select_cell(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":327
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":326
 *         return self.select_bbox(left_edge, right_edge)
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":329
 *         return 0
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":330
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":329
 *         return 0
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":332
 *         return 0
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_select_sphere(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":333
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":332
 *         return 0
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":335
 *         return 0
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         return 0
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_select_bbox(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":337
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":335
 *         return 0
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         return 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":342
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef np.float64_t difference(self, np.float64_t x1, np.float64_t x2, int d) nogil:             # <<<<<<<<<<<<<<
 *         # domain_width is already in code units, and we assume what is fed in
 *         # is too.
 */

static __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_difference(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, __pyx_t_5numpy_float64_t __pyx_v_x1, __pyx_t_5numpy_float64_t __pyx_v_x2, int __pyx_v_d) {
  __pyx_t_5numpy_float64_t __pyx_v_rel;
  __pyx_t_5numpy_float64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":345
 *         # domain_width is already in code units, and we assume what is fed in
 *         # is too.
 *         cdef np.float64_t rel = x1 - x2             # <<<<<<<<<<<<<<
 *         if self.periodicity[d]:
 *             if rel > self.domain_width[d] * 0.5:
 */
  __pyx_v_rel = (__pyx_v_x1 - __pyx_v_x2);

  /* "yt/geometry/selection_routines.pyx":346
 *         # is too.
 *         cdef np.float64_t rel = x1 - x2
 *         if self.periodicity[d]:             # <<<<<<<<<<<<<<
 *             if rel > self.domain_width[d] * 0.5:
 *                 rel -= self.domain_width[d]
 */
  __pyx_t_1 = ((__pyx_v_self->periodicity[__pyx_v_d]) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":347
 *         cdef np.float64_t rel = x1 - x2
 *         if self.periodicity[d]:
 *             if rel > self.domain_width[d] * 0.5:             # <<<<<<<<<<<<<<
 *                 rel -= self.domain_width[d]
 *             elif rel < -self.domain_width[d] * 0.5:
 */
    __pyx_t_1 = ((__pyx_v_rel > ((__pyx_v_self->domain_width[__pyx_v_d]) * 0.5)) != 0);
    if (__pyx_t_1) {

      /* "yt/geometry/selection_routines.pyx":348
 *         if self.periodicity[d]:
 *             if rel > self.domain_width[d] * 0.5:
 *                 rel -= self.domain_width[d]             # <<<<<<<<<<<<<<
 *             elif rel < -self.domain_width[d] * 0.5:
 *                 rel += self.domain_width[d]
 */
      __pyx_v_rel = (__pyx_v_rel - (__pyx_v_self->domain_width[__pyx_v_d]));

      /* "yt/geometry/selection_routines.pyx":347
 *         cdef np.float64_t rel = x1 - x2
 *         if self.periodicity[d]:
 *             if rel > self.domain_width[d] * 0.5:             # <<<<<<<<<<<<<<
 *                 rel -= self.domain_width[d]
 *             elif rel < -self.domain_width[d] * 0.5:
 */
      goto __pyx_L4;
    }

    /* "yt/geometry/selection_routines.pyx":349
 *             if rel > self.domain_width[d] * 0.5:
 *                 rel -= self.domain_width[d]
 *             elif rel < -self.domain_width[d] * 0.5:             # <<<<<<<<<<<<<<
 *                 rel += self.domain_width[d]
 *         return rel
 */
    __pyx_t_1 = ((__pyx_v_rel < ((-(__pyx_v_self->domain_width[__pyx_v_d])) * 0.5)) != 0);
    if (__pyx_t_1) {

      /* "yt/geometry/selection_routines.pyx":350
 *                 rel -= self.domain_width[d]
 *             elif rel < -self.domain_width[d] * 0.5:
 *                 rel += self.domain_width[d]             # <<<<<<<<<<<<<<
 *         return rel
 * 
 */
      __pyx_v_rel = (__pyx_v_rel + (__pyx_v_self->domain_width[__pyx_v_d]));

      /* "yt/geometry/selection_routines.pyx":349
 *             if rel > self.domain_width[d] * 0.5:
 *                 rel -= self.domain_width[d]
 *             elif rel < -self.domain_width[d] * 0.5:             # <<<<<<<<<<<<<<
 *                 rel += self.domain_width[d]
 *         return rel
 */
    }
    __pyx_L4:;

    /* "yt/geometry/selection_routines.pyx":346
 *         # is too.
 *         cdef np.float64_t rel = x1 - x2
 *         if self.periodicity[d]:             # <<<<<<<<<<<<<<
 *             if rel > self.domain_width[d] * 0.5:
 *                 rel -= self.domain_width[d]
 */
  }

  /* "yt/geometry/selection_routines.pyx":351
 *             elif rel < -self.domain_width[d] * 0.5:
 *                 rel += self.domain_width[d]
 *         return rel             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = __pyx_v_rel;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":342
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef np.float64_t difference(self, np.float64_t x1, np.float64_t x2, int d) nogil:             # <<<<<<<<<<<<<<
 *         # domain_width is already in code units, and we assume what is fed in
 *         # is too.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":356
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def fill_mesh_mask(self, mesh):             # <<<<<<<<<<<<<<
 *         cdef np.float64_t pos[3]
 *         cdef np.ndarray[np.int64_t, ndim=2] indices
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_9fill_mesh_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_mesh); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_9fill_mesh_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_mesh) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fill_mesh_mask (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_8fill_mesh_mask(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), ((PyObject *)__pyx_v_mesh));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_8fill_mesh_mask(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyObject *__pyx_v_mesh) {
  __pyx_t_5numpy_float64_t __pyx_v_pos[3];
  PyArrayObject *__pyx_v_indices = 0;
  PyArrayObject *__pyx_v_coords = 0;
  PyArrayObject *__pyx_v_mask = 0;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_selected;
  int __pyx_v_npoints;
  int __pyx_v_nv;
  int __pyx_v_total;
  int __pyx_v_offset;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_coords;
  __Pyx_Buffer __pyx_pybuffer_coords;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyArrayObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyArrayObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  __pyx_t_5numpy_int64_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  int __pyx_t_22;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fill_mesh_mask", 0);
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_coords.pybuffer.buf = NULL;
  __pyx_pybuffer_coords.refcount = 0;
  __pyx_pybuffernd_coords.data = NULL;
  __pyx_pybuffernd_coords.rcbuffer = &__pyx_pybuffer_coords;
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;

  /* "yt/geometry/selection_routines.pyx":362
 *         cdef np.ndarray[np.uint8_t, ndim=1] mask
 *         cdef int i, j, k, selected
 *         cdef int npoints, nv = mesh._connectivity_length             # <<<<<<<<<<<<<<
 *         cdef int total = 0
 *         cdef int offset = mesh._index_offset
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_mesh, __pyx_n_s_connectivity_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_nv = __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":363
 *         cdef int i, j, k, selected
 *         cdef int npoints, nv = mesh._connectivity_length
 *         cdef int total = 0             # <<<<<<<<<<<<<<
 *         cdef int offset = mesh._index_offset
 *         coords = _ensure_code(mesh.connectivity_coords)
 */
  __pyx_v_total = 0;

  /* "yt/geometry/selection_routines.pyx":364
 *         cdef int npoints, nv = mesh._connectivity_length
 *         cdef int total = 0
 *         cdef int offset = mesh._index_offset             # <<<<<<<<<<<<<<
 *         coords = _ensure_code(mesh.connectivity_coords)
 *         indices = mesh.connectivity_indices
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_mesh, __pyx_n_s_index_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_offset = __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":365
 *         cdef int total = 0
 *         cdef int offset = mesh._index_offset
 *         coords = _ensure_code(mesh.connectivity_coords)             # <<<<<<<<<<<<<<
 *         indices = mesh.connectivity_indices
 *         npoints = indices.shape[0]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_mesh, __pyx_n_s_connectivity_coords); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 365, __pyx_L1_error)
  __pyx_t_4 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords.rcbuffer->pybuffer);
    __pyx_t_2 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_coords.rcbuffer->pybuffer, (PyObject*)__pyx_t_4, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_2 < 0)) {
      PyErr_Fetch(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_coords.rcbuffer->pybuffer, (PyObject*)__pyx_v_coords, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_7);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      }
      __pyx_t_5 = __pyx_t_6 = __pyx_t_7 = 0;
    }
    __pyx_pybuffernd_coords.diminfo[0].strides = __pyx_pybuffernd_coords.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_coords.diminfo[0].shape = __pyx_pybuffernd_coords.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_coords.diminfo[1].strides = __pyx_pybuffernd_coords.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_coords.diminfo[1].shape = __pyx_pybuffernd_coords.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 365, __pyx_L1_error)
  }
  __pyx_t_4 = 0;
  __pyx_v_coords = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":366
 *         cdef int offset = mesh._index_offset
 *         coords = _ensure_code(mesh.connectivity_coords)
 *         indices = mesh.connectivity_indices             # <<<<<<<<<<<<<<
 *         npoints = indices.shape[0]
 *         mask = np.zeros(npoints, dtype='uint8')
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_mesh, __pyx_n_s_connectivity_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 366, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_2 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_2 < 0)) {
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_6, &__pyx_t_5);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_7, __pyx_t_6, __pyx_t_5);
      }
      __pyx_t_7 = __pyx_t_6 = __pyx_t_5 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_indices.diminfo[1].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_indices.diminfo[1].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 366, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __pyx_v_indices = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":367
 *         coords = _ensure_code(mesh.connectivity_coords)
 *         indices = mesh.connectivity_indices
 *         npoints = indices.shape[0]             # <<<<<<<<<<<<<<
 *         mask = np.zeros(npoints, dtype='uint8')
 *         for i in range(npoints):
 */
  __pyx_v_npoints = (__pyx_v_indices->dimensions[0]);

  /* "yt/geometry/selection_routines.pyx":368
 *         indices = mesh.connectivity_indices
 *         npoints = indices.shape[0]
 *         mask = np.zeros(npoints, dtype='uint8')             # <<<<<<<<<<<<<<
 *         for i in range(npoints):
 *             selected = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_npoints); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 368, __pyx_L1_error)
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_10) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_10, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 368, __pyx_L1_error)
  __pyx_t_11 = ((PyArrayObject *)__pyx_t_10);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __pyx_t_2 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_11, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_2 < 0)) {
      PyErr_Fetch(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_7);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      }
      __pyx_t_5 = __pyx_t_6 = __pyx_t_7 = 0;
    }
    __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 368, __pyx_L1_error)
  }
  __pyx_t_11 = 0;
  __pyx_v_mask = ((PyArrayObject *)__pyx_t_10);
  __pyx_t_10 = 0;

  /* "yt/geometry/selection_routines.pyx":369
 *         npoints = indices.shape[0]
 *         mask = np.zeros(npoints, dtype='uint8')
 *         for i in range(npoints):             # <<<<<<<<<<<<<<
 *             selected = 0
 *             for j in range(nv):
 */
  __pyx_t_2 = __pyx_v_npoints;
  __pyx_t_12 = __pyx_t_2;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_i = __pyx_t_13;

    /* "yt/geometry/selection_routines.pyx":370
 *         mask = np.zeros(npoints, dtype='uint8')
 *         for i in range(npoints):
 *             selected = 0             # <<<<<<<<<<<<<<
 *             for j in range(nv):
 *                 for k in range(3):
 */
    __pyx_v_selected = 0;

    /* "yt/geometry/selection_routines.pyx":371
 *         for i in range(npoints):
 *             selected = 0
 *             for j in range(nv):             # <<<<<<<<<<<<<<
 *                 for k in range(3):
 *                     pos[k] = coords[indices[i, j] - offset, k]
 */
    __pyx_t_14 = __pyx_v_nv;
    __pyx_t_15 = __pyx_t_14;
    for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
      __pyx_v_j = __pyx_t_16;

      /* "yt/geometry/selection_routines.pyx":372
 *             selected = 0
 *             for j in range(nv):
 *                 for k in range(3):             # <<<<<<<<<<<<<<
 *                     pos[k] = coords[indices[i, j] - offset, k]
 *                 selected = self.select_point(pos)
 */
      for (__pyx_t_17 = 0; __pyx_t_17 < 3; __pyx_t_17+=1) {
        __pyx_v_k = __pyx_t_17;

        /* "yt/geometry/selection_routines.pyx":373
 *             for j in range(nv):
 *                 for k in range(3):
 *                     pos[k] = coords[indices[i, j] - offset, k]             # <<<<<<<<<<<<<<
 *                 selected = self.select_point(pos)
 *                 if selected == 1: break
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_19 = __pyx_v_j;
        __pyx_t_20 = ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_indices.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_indices.diminfo[0].strides, __pyx_t_19, __pyx_pybuffernd_indices.diminfo[1].strides)) - __pyx_v_offset);
        __pyx_t_21 = __pyx_v_k;
        (__pyx_v_pos[__pyx_v_k]) = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_coords.rcbuffer->pybuffer.buf, __pyx_t_20, __pyx_pybuffernd_coords.diminfo[0].strides, __pyx_t_21, __pyx_pybuffernd_coords.diminfo[1].strides));
      }

      /* "yt/geometry/selection_routines.pyx":374
 *                 for k in range(3):
 *                     pos[k] = coords[indices[i, j] - offset, k]
 *                 selected = self.select_point(pos)             # <<<<<<<<<<<<<<
 *                 if selected == 1: break
 *             total += selected
 */
      __pyx_v_selected = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_point(__pyx_v_self, __pyx_v_pos);

      /* "yt/geometry/selection_routines.pyx":375
 *                     pos[k] = coords[indices[i, j] - offset, k]
 *                 selected = self.select_point(pos)
 *                 if selected == 1: break             # <<<<<<<<<<<<<<
 *             total += selected
 *             mask[i] = selected
 */
      __pyx_t_22 = ((__pyx_v_selected == 1) != 0);
      if (__pyx_t_22) {
        goto __pyx_L6_break;
      }
    }
    __pyx_L6_break:;

    /* "yt/geometry/selection_routines.pyx":376
 *                 selected = self.select_point(pos)
 *                 if selected == 1: break
 *             total += selected             # <<<<<<<<<<<<<<
 *             mask[i] = selected
 *         if total == 0: return None
 */
    __pyx_v_total = (__pyx_v_total + __pyx_v_selected);

    /* "yt/geometry/selection_routines.pyx":377
 *                 if selected == 1: break
 *             total += selected
 *             mask[i] = selected             # <<<<<<<<<<<<<<
 *         if total == 0: return None
 *         return mask.astype("bool")
 */
    __pyx_t_19 = __pyx_v_i;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_19, __pyx_pybuffernd_mask.diminfo[0].strides) = __pyx_v_selected;
  }

  /* "yt/geometry/selection_routines.pyx":378
 *             total += selected
 *             mask[i] = selected
 *         if total == 0: return None             # <<<<<<<<<<<<<<
 *         return mask.astype("bool")
 * 
 */
  __pyx_t_22 = ((__pyx_v_total == 0) != 0);
  if (__pyx_t_22) {
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":379
 *             mask[i] = selected
 *         if total == 0: return None
 *         return mask.astype("bool")             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_mask), __pyx_n_s_astype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_10 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_9, __pyx_n_s_bool) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_s_bool);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":356
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def fill_mesh_mask(self, mesh):             # <<<<<<<<<<<<<<
 *         cdef np.float64_t pos[3]
 *         cdef np.ndarray[np.int64_t, ndim=2] indices
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.fill_mesh_mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XDECREF((PyObject *)__pyx_v_coords);
  __Pyx_XDECREF((PyObject *)__pyx_v_mask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":384
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def fill_mesh_cell_mask(self, mesh):             # <<<<<<<<<<<<<<
 *         cdef np.float64_t pos
 *         cdef np.float64_t le[3]
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_11fill_mesh_cell_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_mesh); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_11fill_mesh_cell_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_mesh) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fill_mesh_cell_mask (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_10fill_mesh_cell_mask(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), ((PyObject *)__pyx_v_mesh));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_10fill_mesh_cell_mask(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyObject *__pyx_v_mesh) {
  __pyx_t_5numpy_float64_t __pyx_v_pos;
  __pyx_t_5numpy_float64_t __pyx_v_le[3];
  __pyx_t_5numpy_float64_t __pyx_v_re[3];
  PyArrayObject *__pyx_v_indices = 0;
  PyArrayObject *__pyx_v_coords = 0;
  PyArrayObject *__pyx_v_mask = 0;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_selected;
  int __pyx_v_npoints;
  int __pyx_v_nv;
  int __pyx_v_ndim;
  int __pyx_v_total;
  int __pyx_v_offset;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_coords;
  __Pyx_Buffer __pyx_pybuffer_coords;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyArrayObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyArrayObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __pyx_t_5numpy_int64_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  int __pyx_t_24;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fill_mesh_cell_mask", 0);
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_coords.pybuffer.buf = NULL;
  __pyx_pybuffer_coords.refcount = 0;
  __pyx_pybuffernd_coords.data = NULL;
  __pyx_pybuffernd_coords.rcbuffer = &__pyx_pybuffer_coords;
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;

  /* "yt/geometry/selection_routines.pyx":392
 *         cdef np.ndarray[np.uint8_t, ndim=1] mask
 *         cdef int i, j, k, selected
 *         cdef int npoints, nv = mesh._connectivity_length             # <<<<<<<<<<<<<<
 *         cdef int ndim = mesh.connectivity_coords.shape[1]
 *         cdef int total = 0
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_mesh, __pyx_n_s_connectivity_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_nv = __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":393
 *         cdef int i, j, k, selected
 *         cdef int npoints, nv = mesh._connectivity_length
 *         cdef int ndim = mesh.connectivity_coords.shape[1]             # <<<<<<<<<<<<<<
 *         cdef int total = 0
 *         cdef int offset = mesh._index_offset
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_mesh, __pyx_n_s_connectivity_coords); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ndim = __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":394
 *         cdef int npoints, nv = mesh._connectivity_length
 *         cdef int ndim = mesh.connectivity_coords.shape[1]
 *         cdef int total = 0             # <<<<<<<<<<<<<<
 *         cdef int offset = mesh._index_offset
 *         coords = _ensure_code(mesh.connectivity_coords)
 */
  __pyx_v_total = 0;

  /* "yt/geometry/selection_routines.pyx":395
 *         cdef int ndim = mesh.connectivity_coords.shape[1]
 *         cdef int total = 0
 *         cdef int offset = mesh._index_offset             # <<<<<<<<<<<<<<
 *         coords = _ensure_code(mesh.connectivity_coords)
 *         indices = mesh.connectivity_indices
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_mesh, __pyx_n_s_index_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_offset = __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":396
 *         cdef int total = 0
 *         cdef int offset = mesh._index_offset
 *         coords = _ensure_code(mesh.connectivity_coords)             # <<<<<<<<<<<<<<
 *         indices = mesh.connectivity_indices
 *         npoints = indices.shape[0]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_mesh, __pyx_n_s_connectivity_coords); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 396, __pyx_L1_error)
  __pyx_t_4 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords.rcbuffer->pybuffer);
    __pyx_t_2 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_coords.rcbuffer->pybuffer, (PyObject*)__pyx_t_4, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_2 < 0)) {
      PyErr_Fetch(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_coords.rcbuffer->pybuffer, (PyObject*)__pyx_v_coords, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_7);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      }
      __pyx_t_5 = __pyx_t_6 = __pyx_t_7 = 0;
    }
    __pyx_pybuffernd_coords.diminfo[0].strides = __pyx_pybuffernd_coords.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_coords.diminfo[0].shape = __pyx_pybuffernd_coords.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_coords.diminfo[1].strides = __pyx_pybuffernd_coords.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_coords.diminfo[1].shape = __pyx_pybuffernd_coords.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 396, __pyx_L1_error)
  }
  __pyx_t_4 = 0;
  __pyx_v_coords = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":397
 *         cdef int offset = mesh._index_offset
 *         coords = _ensure_code(mesh.connectivity_coords)
 *         indices = mesh.connectivity_indices             # <<<<<<<<<<<<<<
 *         npoints = indices.shape[0]
 *         mask = np.zeros(npoints, dtype='uint8')
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_mesh, __pyx_n_s_connectivity_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 397, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_2 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_2 < 0)) {
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_6, &__pyx_t_5);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_7, __pyx_t_6, __pyx_t_5);
      }
      __pyx_t_7 = __pyx_t_6 = __pyx_t_5 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_indices.diminfo[1].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_indices.diminfo[1].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 397, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __pyx_v_indices = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":398
 *         coords = _ensure_code(mesh.connectivity_coords)
 *         indices = mesh.connectivity_indices
 *         npoints = indices.shape[0]             # <<<<<<<<<<<<<<
 *         mask = np.zeros(npoints, dtype='uint8')
 *         for i in range(npoints):
 */
  __pyx_v_npoints = (__pyx_v_indices->dimensions[0]);

  /* "yt/geometry/selection_routines.pyx":399
 *         indices = mesh.connectivity_indices
 *         npoints = indices.shape[0]
 *         mask = np.zeros(npoints, dtype='uint8')             # <<<<<<<<<<<<<<
 *         for i in range(npoints):
 *             selected = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_npoints); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_10) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_10, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 399, __pyx_L1_error)
  __pyx_t_11 = ((PyArrayObject *)__pyx_t_10);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __pyx_t_2 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_11, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_2 < 0)) {
      PyErr_Fetch(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_7);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      }
      __pyx_t_5 = __pyx_t_6 = __pyx_t_7 = 0;
    }
    __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 399, __pyx_L1_error)
  }
  __pyx_t_11 = 0;
  __pyx_v_mask = ((PyArrayObject *)__pyx_t_10);
  __pyx_t_10 = 0;

  /* "yt/geometry/selection_routines.pyx":400
 *         npoints = indices.shape[0]
 *         mask = np.zeros(npoints, dtype='uint8')
 *         for i in range(npoints):             # <<<<<<<<<<<<<<
 *             selected = 0
 *             for k in range(ndim):
 */
  __pyx_t_2 = __pyx_v_npoints;
  __pyx_t_12 = __pyx_t_2;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_i = __pyx_t_13;

    /* "yt/geometry/selection_routines.pyx":401
 *         mask = np.zeros(npoints, dtype='uint8')
 *         for i in range(npoints):
 *             selected = 0             # <<<<<<<<<<<<<<
 *             for k in range(ndim):
 *                 le[k] = 1e60
 */
    __pyx_v_selected = 0;

    /* "yt/geometry/selection_routines.pyx":402
 *         for i in range(npoints):
 *             selected = 0
 *             for k in range(ndim):             # <<<<<<<<<<<<<<
 *                 le[k] = 1e60
 *                 re[k] = -1e60
 */
    __pyx_t_14 = __pyx_v_ndim;
    __pyx_t_15 = __pyx_t_14;
    for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
      __pyx_v_k = __pyx_t_16;

      /* "yt/geometry/selection_routines.pyx":403
 *             selected = 0
 *             for k in range(ndim):
 *                 le[k] = 1e60             # <<<<<<<<<<<<<<
 *                 re[k] = -1e60
 *             for j in range(nv):
 */
      (__pyx_v_le[__pyx_v_k]) = 1e60;

      /* "yt/geometry/selection_routines.pyx":404
 *             for k in range(ndim):
 *                 le[k] = 1e60
 *                 re[k] = -1e60             # <<<<<<<<<<<<<<
 *             for j in range(nv):
 *                 for k in range(ndim):
 */
      (__pyx_v_re[__pyx_v_k]) = -1e60;
    }

    /* "yt/geometry/selection_routines.pyx":405
 *                 le[k] = 1e60
 *                 re[k] = -1e60
 *             for j in range(nv):             # <<<<<<<<<<<<<<
 *                 for k in range(ndim):
 *                     pos = coords[indices[i, j] - offset, k]
 */
    __pyx_t_14 = __pyx_v_nv;
    __pyx_t_15 = __pyx_t_14;
    for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
      __pyx_v_j = __pyx_t_16;

      /* "yt/geometry/selection_routines.pyx":406
 *                 re[k] = -1e60
 *             for j in range(nv):
 *                 for k in range(ndim):             # <<<<<<<<<<<<<<
 *                     pos = coords[indices[i, j] - offset, k]
 *                     le[k] = fmin(pos, le[k])
 */
      __pyx_t_17 = __pyx_v_ndim;
      __pyx_t_18 = __pyx_t_17;
      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
        __pyx_v_k = __pyx_t_19;

        /* "yt/geometry/selection_routines.pyx":407
 *             for j in range(nv):
 *                 for k in range(ndim):
 *                     pos = coords[indices[i, j] - offset, k]             # <<<<<<<<<<<<<<
 *                     le[k] = fmin(pos, le[k])
 *                     re[k] = fmax(pos, re[k])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_21 = __pyx_v_j;
        __pyx_t_22 = ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_indices.rcbuffer->pybuffer.buf, __pyx_t_20, __pyx_pybuffernd_indices.diminfo[0].strides, __pyx_t_21, __pyx_pybuffernd_indices.diminfo[1].strides)) - __pyx_v_offset);
        __pyx_t_23 = __pyx_v_k;
        __pyx_v_pos = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_coords.rcbuffer->pybuffer.buf, __pyx_t_22, __pyx_pybuffernd_coords.diminfo[0].strides, __pyx_t_23, __pyx_pybuffernd_coords.diminfo[1].strides));

        /* "yt/geometry/selection_routines.pyx":408
 *                 for k in range(ndim):
 *                     pos = coords[indices[i, j] - offset, k]
 *                     le[k] = fmin(pos, le[k])             # <<<<<<<<<<<<<<
 *                     re[k] = fmax(pos, re[k])
 *             selected = self.select_bbox(le, re)
 */
        (__pyx_v_le[__pyx_v_k]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_fmin(__pyx_v_pos, (__pyx_v_le[__pyx_v_k]));

        /* "yt/geometry/selection_routines.pyx":409
 *                     pos = coords[indices[i, j] - offset, k]
 *                     le[k] = fmin(pos, le[k])
 *                     re[k] = fmax(pos, re[k])             # <<<<<<<<<<<<<<
 *             selected = self.select_bbox(le, re)
 *             total += selected
 */
        (__pyx_v_re[__pyx_v_k]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_fmax(__pyx_v_pos, (__pyx_v_re[__pyx_v_k]));
      }
    }

    /* "yt/geometry/selection_routines.pyx":410
 *                     le[k] = fmin(pos, le[k])
 *                     re[k] = fmax(pos, re[k])
 *             selected = self.select_bbox(le, re)             # <<<<<<<<<<<<<<
 *             total += selected
 *             mask[i] = selected
 */
    __pyx_v_selected = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_bbox(__pyx_v_self, __pyx_v_le, __pyx_v_re);

    /* "yt/geometry/selection_routines.pyx":411
 *                     re[k] = fmax(pos, re[k])
 *             selected = self.select_bbox(le, re)
 *             total += selected             # <<<<<<<<<<<<<<
 *             mask[i] = selected
 *         if total == 0: return None
 */
    __pyx_v_total = (__pyx_v_total + __pyx_v_selected);

    /* "yt/geometry/selection_routines.pyx":412
 *             selected = self.select_bbox(le, re)
 *             total += selected
 *             mask[i] = selected             # <<<<<<<<<<<<<<
 *         if total == 0: return None
 *         return mask.astype("bool")
 */
    __pyx_t_21 = __pyx_v_i;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_21, __pyx_pybuffernd_mask.diminfo[0].strides) = __pyx_v_selected;
  }

  /* "yt/geometry/selection_routines.pyx":413
 *             total += selected
 *             mask[i] = selected
 *         if total == 0: return None             # <<<<<<<<<<<<<<
 *         return mask.astype("bool")
 * 
 */
  __pyx_t_24 = ((__pyx_v_total == 0) != 0);
  if (__pyx_t_24) {
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":414
 *             mask[i] = selected
 *         if total == 0: return None
 *         return mask.astype("bool")             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_mask), __pyx_n_s_astype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_10 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_9, __pyx_n_s_bool) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_s_bool);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":384
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def fill_mesh_cell_mask(self, mesh):             # <<<<<<<<<<<<<<
 *         cdef np.float64_t pos
 *         cdef np.float64_t le[3]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.fill_mesh_cell_mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XDECREF((PyObject *)__pyx_v_coords);
  __Pyx_XDECREF((PyObject *)__pyx_v_mask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":419
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def fill_mask(self, gobj):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
 *         child_mask = gobj.child_mask
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_13fill_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_gobj); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_13fill_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_gobj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fill_mask (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_12fill_mask(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), ((PyObject *)__pyx_v_gobj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_12fill_mask(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyObject *__pyx_v_gobj) {
  PyArrayObject *__pyx_v_child_mask = 0;
  PyArrayObject *__pyx_v_mask = 0;
  int __pyx_v_dim[3];
  PyArrayObject *__pyx_v_odds = 0;
  PyArrayObject *__pyx_v_oleft_edge = 0;
  PyArrayObject *__pyx_v_oright_edge = 0;
  int __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_dds[3];
  __pyx_t_5numpy_float64_t __pyx_v_left_edge[3];
  __pyx_t_5numpy_float64_t __pyx_v_right_edge[3];
  __pyx_t_5numpy_int32_t __pyx_v_level;
  int __pyx_v_total;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_child_mask;
  __Pyx_Buffer __pyx_pybuffer_child_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_odds;
  __Pyx_Buffer __pyx_pybuffer_odds;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_oleft_edge;
  __Pyx_Buffer __pyx_pybuffer_oleft_edge;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_oright_edge;
  __Pyx_Buffer __pyx_pybuffer_oright_edge;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyArrayObject *__pyx_t_9 = NULL;
  PyArrayObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyArrayObject *__pyx_t_15 = NULL;
  __pyx_t_5numpy_int32_t __pyx_t_16;
  int __pyx_t_17;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fill_mask", 0);
  __pyx_pybuffer_child_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_child_mask.refcount = 0;
  __pyx_pybuffernd_child_mask.data = NULL;
  __pyx_pybuffernd_child_mask.rcbuffer = &__pyx_pybuffer_child_mask;
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  __pyx_pybuffer_odds.pybuffer.buf = NULL;
  __pyx_pybuffer_odds.refcount = 0;
  __pyx_pybuffernd_odds.data = NULL;
  __pyx_pybuffernd_odds.rcbuffer = &__pyx_pybuffer_odds;
  __pyx_pybuffer_oleft_edge.pybuffer.buf = NULL;
  __pyx_pybuffer_oleft_edge.refcount = 0;
  __pyx_pybuffernd_oleft_edge.data = NULL;
  __pyx_pybuffernd_oleft_edge.rcbuffer = &__pyx_pybuffer_oleft_edge;
  __pyx_pybuffer_oright_edge.pybuffer.buf = NULL;
  __pyx_pybuffer_oright_edge.refcount = 0;
  __pyx_pybuffernd_oright_edge.data = NULL;
  __pyx_pybuffernd_oright_edge.rcbuffer = &__pyx_pybuffer_oright_edge;

  /* "yt/geometry/selection_routines.pyx":421
 *     def fill_mask(self, gobj):
 *         cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
 *         child_mask = gobj.child_mask             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=3] mask
 *         cdef int dim[3]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_child_mask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 421, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_child_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_child_mask.diminfo[0].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_child_mask.diminfo[0].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_child_mask.diminfo[1].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_child_mask.diminfo[1].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_child_mask.diminfo[2].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_child_mask.diminfo[2].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 421, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __pyx_v_child_mask = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":424
 *         cdef np.ndarray[np.uint8_t, ndim=3] mask
 *         cdef int dim[3]
 *         _ensure_code(gobj.dds)             # <<<<<<<<<<<<<<
 *         _ensure_code(gobj.LeftEdge)
 *         _ensure_code(gobj.RightEdge)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_dds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "yt/geometry/selection_routines.pyx":425
 *         cdef int dim[3]
 *         _ensure_code(gobj.dds)
 *         _ensure_code(gobj.LeftEdge)             # <<<<<<<<<<<<<<
 *         _ensure_code(gobj.RightEdge)
 *         cdef np.ndarray[np.float64_t, ndim=1] odds = gobj.dds.d
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_LeftEdge); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":426
 *         _ensure_code(gobj.dds)
 *         _ensure_code(gobj.LeftEdge)
 *         _ensure_code(gobj.RightEdge)             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.float64_t, ndim=1] odds = gobj.dds.d
 *         cdef np.ndarray[np.float64_t, ndim=1] oleft_edge = gobj.LeftEdge.d
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_RightEdge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "yt/geometry/selection_routines.pyx":427
 *         _ensure_code(gobj.LeftEdge)
 *         _ensure_code(gobj.RightEdge)
 *         cdef np.ndarray[np.float64_t, ndim=1] odds = gobj.dds.d             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.float64_t, ndim=1] oleft_edge = gobj.LeftEdge.d
 *         cdef np.ndarray[np.float64_t, ndim=1] oright_edge = gobj.RightEdge.d
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_dds); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_d); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 427, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_odds.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_odds = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_odds.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 427, __pyx_L1_error)
    } else {__pyx_pybuffernd_odds.diminfo[0].strides = __pyx_pybuffernd_odds.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_odds.diminfo[0].shape = __pyx_pybuffernd_odds.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_8 = 0;
  __pyx_v_odds = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":428
 *         _ensure_code(gobj.RightEdge)
 *         cdef np.ndarray[np.float64_t, ndim=1] odds = gobj.dds.d
 *         cdef np.ndarray[np.float64_t, ndim=1] oleft_edge = gobj.LeftEdge.d             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.float64_t, ndim=1] oright_edge = gobj.RightEdge.d
 *         cdef int i
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_LeftEdge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_d); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 428, __pyx_L1_error)
  __pyx_t_9 = ((PyArrayObject *)__pyx_t_7);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_oleft_edge.rcbuffer->pybuffer, (PyObject*)__pyx_t_9, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_oleft_edge = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_oleft_edge.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 428, __pyx_L1_error)
    } else {__pyx_pybuffernd_oleft_edge.diminfo[0].strides = __pyx_pybuffernd_oleft_edge.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_oleft_edge.diminfo[0].shape = __pyx_pybuffernd_oleft_edge.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_9 = 0;
  __pyx_v_oleft_edge = ((PyArrayObject *)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "yt/geometry/selection_routines.pyx":429
 *         cdef np.ndarray[np.float64_t, ndim=1] odds = gobj.dds.d
 *         cdef np.ndarray[np.float64_t, ndim=1] oleft_edge = gobj.LeftEdge.d
 *         cdef np.ndarray[np.float64_t, ndim=1] oright_edge = gobj.RightEdge.d             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.float64_t dds[3]
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_RightEdge); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_d); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 429, __pyx_L1_error)
  __pyx_t_10 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_oright_edge.rcbuffer->pybuffer, (PyObject*)__pyx_t_10, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_oright_edge = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_oright_edge.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 429, __pyx_L1_error)
    } else {__pyx_pybuffernd_oright_edge.diminfo[0].strides = __pyx_pybuffernd_oright_edge.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_oright_edge.diminfo[0].shape = __pyx_pybuffernd_oright_edge.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_10 = 0;
  __pyx_v_oright_edge = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":434
 *         cdef np.float64_t left_edge[3]
 *         cdef np.float64_t right_edge[3]
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             dds[i] = odds[i]
 *             dim[i] = gobj.ActiveDimensions[i]
 */
  for (__pyx_t_3 = 0; __pyx_t_3 < 3; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "yt/geometry/selection_routines.pyx":435
 *         cdef np.float64_t right_edge[3]
 *         for i in range(3):
 *             dds[i] = odds[i]             # <<<<<<<<<<<<<<
 *             dim[i] = gobj.ActiveDimensions[i]
 *             left_edge[i] = oleft_edge[i]
 */
    __pyx_t_11 = __pyx_v_i;
    (__pyx_v_dds[__pyx_v_i]) = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_odds.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_odds.diminfo[0].strides));

    /* "yt/geometry/selection_routines.pyx":436
 *         for i in range(3):
 *             dds[i] = odds[i]
 *             dim[i] = gobj.ActiveDimensions[i]             # <<<<<<<<<<<<<<
 *             left_edge[i] = oleft_edge[i]
 *             right_edge[i] = oright_edge[i]
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_ActiveDimensions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_7); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 436, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    (__pyx_v_dim[__pyx_v_i]) = __pyx_t_12;

    /* "yt/geometry/selection_routines.pyx":437
 *             dds[i] = odds[i]
 *             dim[i] = gobj.ActiveDimensions[i]
 *             left_edge[i] = oleft_edge[i]             # <<<<<<<<<<<<<<
 *             right_edge[i] = oright_edge[i]
 *         mask = np.zeros(gobj.ActiveDimensions, dtype='uint8')
 */
    __pyx_t_11 = __pyx_v_i;
    (__pyx_v_left_edge[__pyx_v_i]) = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_oleft_edge.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_oleft_edge.diminfo[0].strides));

    /* "yt/geometry/selection_routines.pyx":438
 *             dim[i] = gobj.ActiveDimensions[i]
 *             left_edge[i] = oleft_edge[i]
 *             right_edge[i] = oright_edge[i]             # <<<<<<<<<<<<<<
 *         mask = np.zeros(gobj.ActiveDimensions, dtype='uint8')
 *         # Check for the level bounds
 */
    __pyx_t_11 = __pyx_v_i;
    (__pyx_v_right_edge[__pyx_v_i]) = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_oright_edge.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_oright_edge.diminfo[0].strides));
  }

  /* "yt/geometry/selection_routines.pyx":439
 *             left_edge[i] = oleft_edge[i]
 *             right_edge[i] = oright_edge[i]
 *         mask = np.zeros(gobj.ActiveDimensions, dtype='uint8')             # <<<<<<<<<<<<<<
 *         # Check for the level bounds
 *         cdef np.int32_t level = gobj.Level
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_ActiveDimensions); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_13 = PyTuple_New(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 439, __pyx_L1_error)
  __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_13, __pyx_t_7); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (!(likely(((__pyx_t_14) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_14, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 439, __pyx_L1_error)
  __pyx_t_15 = ((PyArrayObject *)__pyx_t_14);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_15, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mask.diminfo[1].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mask.diminfo[1].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_mask.diminfo[2].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_mask.diminfo[2].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 439, __pyx_L1_error)
  }
  __pyx_t_15 = 0;
  __pyx_v_mask = ((PyArrayObject *)__pyx_t_14);
  __pyx_t_14 = 0;

  /* "yt/geometry/selection_routines.pyx":441
 *         mask = np.zeros(gobj.ActiveDimensions, dtype='uint8')
 *         # Check for the level bounds
 *         cdef np.int32_t level = gobj.Level             # <<<<<<<<<<<<<<
 *         # We set this to 1 if we ignore child_mask
 *         cdef int total
 */
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_Level); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_16 = __Pyx_PyInt_As_npy_int32(__pyx_t_14); if (unlikely((__pyx_t_16 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_v_level = __pyx_t_16;

  /* "yt/geometry/selection_routines.pyx":444
 *         # We set this to 1 if we ignore child_mask
 *         cdef int total
 *         total = self.fill_mask_selector(left_edge, right_edge, dds, dim,             # <<<<<<<<<<<<<<
 *                                         child_mask, mask, level)
 *         if total == 0: return None
 */
  __pyx_v_total = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->fill_mask_selector(__pyx_v_self, __pyx_v_left_edge, __pyx_v_right_edge, __pyx_v_dds, __pyx_v_dim, ((PyArrayObject *)__pyx_v_child_mask), ((PyArrayObject *)__pyx_v_mask), __pyx_v_level);

  /* "yt/geometry/selection_routines.pyx":446
 *         total = self.fill_mask_selector(left_edge, right_edge, dds, dim,
 *                                         child_mask, mask, level)
 *         if total == 0: return None             # <<<<<<<<<<<<<<
 *         return mask.astype("bool")
 * 
 */
  __pyx_t_17 = ((__pyx_v_total == 0) != 0);
  if (__pyx_t_17) {
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":447
 *                                         child_mask, mask, level)
 *         if total == 0: return None
 *         return mask.astype("bool")             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_mask), __pyx_n_s_astype); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_13 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_13)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_13);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_14 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_13, __pyx_n_s_bool) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_n_s_bool);
  __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
  if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_r = __pyx_t_14;
  __pyx_t_14 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":419
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def fill_mask(self, gobj):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
 *         child_mask = gobj.child_mask
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_odds.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_oleft_edge.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_oright_edge.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.fill_mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_odds.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_oleft_edge.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_oright_edge.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child_mask);
  __Pyx_XDECREF((PyObject *)__pyx_v_mask);
  __Pyx_XDECREF((PyObject *)__pyx_v_odds);
  __Pyx_XDECREF((PyObject *)__pyx_v_oleft_edge);
  __Pyx_XDECREF((PyObject *)__pyx_v_oright_edge);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":452
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int fill_mask_selector(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                 np.float64_t right_edge[3],
 *                                 np.float64_t dds[3], int dim[3],
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_fill_mask_selector(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_float64_t *__pyx_v_dds, int *__pyx_v_dim, PyArrayObject *__pyx_v_child_mask, PyArrayObject *__pyx_v_mask, int __pyx_v_level) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_total;
  int __pyx_v_this_level;
  __pyx_t_5numpy_float64_t __pyx_v_pos[3];
  __Pyx_LocalBuf_ND __pyx_pybuffernd_child_mask;
  __Pyx_Buffer __pyx_pybuffer_child_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  long __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fill_mask_selector", 0);
  __pyx_pybuffer_child_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_child_mask.refcount = 0;
  __pyx_pybuffernd_child_mask.data = NULL;
  __pyx_pybuffernd_child_mask.rcbuffer = &__pyx_pybuffer_child_mask;
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_child_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack) == -1)) __PYX_ERR(0, 452, __pyx_L1_error)
  }
  __pyx_pybuffernd_child_mask.diminfo[0].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_child_mask.diminfo[0].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_child_mask.diminfo[1].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_child_mask.diminfo[1].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_child_mask.diminfo[2].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_child_mask.diminfo[2].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[2];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) __PYX_ERR(0, 452, __pyx_L1_error)
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mask.diminfo[1].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mask.diminfo[1].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_mask.diminfo[2].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_mask.diminfo[2].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[2];

  /* "yt/geometry/selection_routines.pyx":459
 *                                 int level):
 *         cdef int i, j, k
 *         cdef int total = 0, this_level = 0             # <<<<<<<<<<<<<<
 *         cdef np.float64_t pos[3]
 *         if level < self.min_level or level > self.max_level:
 */
  __pyx_v_total = 0;
  __pyx_v_this_level = 0;

  /* "yt/geometry/selection_routines.pyx":461
 *         cdef int total = 0, this_level = 0
 *         cdef np.float64_t pos[3]
 *         if level < self.min_level or level > self.max_level:             # <<<<<<<<<<<<<<
 *             return 0
 *         if level == self.max_level:
 */
  __pyx_t_2 = ((__pyx_v_level < __pyx_v_self->min_level) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_level > __pyx_v_self->max_level) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":462
 *         cdef np.float64_t pos[3]
 *         if level < self.min_level or level > self.max_level:
 *             return 0             # <<<<<<<<<<<<<<
 *         if level == self.max_level:
 *             this_level = 1
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":461
 *         cdef int total = 0, this_level = 0
 *         cdef np.float64_t pos[3]
 *         if level < self.min_level or level > self.max_level:             # <<<<<<<<<<<<<<
 *             return 0
 *         if level == self.max_level:
 */
  }

  /* "yt/geometry/selection_routines.pyx":463
 *         if level < self.min_level or level > self.max_level:
 *             return 0
 *         if level == self.max_level:             # <<<<<<<<<<<<<<
 *             this_level = 1
 *         with nogil:
 */
  __pyx_t_1 = ((__pyx_v_level == __pyx_v_self->max_level) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":464
 *             return 0
 *         if level == self.max_level:
 *             this_level = 1             # <<<<<<<<<<<<<<
 *         with nogil:
 *             pos[0] = left_edge[0] + dds[0] * 0.5
 */
    __pyx_v_this_level = 1;

    /* "yt/geometry/selection_routines.pyx":463
 *         if level < self.min_level or level > self.max_level:
 *             return 0
 *         if level == self.max_level:             # <<<<<<<<<<<<<<
 *             this_level = 1
 *         with nogil:
 */
  }

  /* "yt/geometry/selection_routines.pyx":465
 *         if level == self.max_level:
 *             this_level = 1
 *         with nogil:             # <<<<<<<<<<<<<<
 *             pos[0] = left_edge[0] + dds[0] * 0.5
 *             for i in range(dim[0]):
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "yt/geometry/selection_routines.pyx":466
 *             this_level = 1
 *         with nogil:
 *             pos[0] = left_edge[0] + dds[0] * 0.5             # <<<<<<<<<<<<<<
 *             for i in range(dim[0]):
 *                 pos[1] = left_edge[1] + dds[1] * 0.5
 */
        (__pyx_v_pos[0]) = ((__pyx_v_left_edge[0]) + ((__pyx_v_dds[0]) * 0.5));

        /* "yt/geometry/selection_routines.pyx":467
 *         with nogil:
 *             pos[0] = left_edge[0] + dds[0] * 0.5
 *             for i in range(dim[0]):             # <<<<<<<<<<<<<<
 *                 pos[1] = left_edge[1] + dds[1] * 0.5
 *                 for j in range(dim[1]):
 */
        __pyx_t_3 = (__pyx_v_dim[0]);
        __pyx_t_4 = __pyx_t_3;
        for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
          __pyx_v_i = __pyx_t_5;

          /* "yt/geometry/selection_routines.pyx":468
 *             pos[0] = left_edge[0] + dds[0] * 0.5
 *             for i in range(dim[0]):
 *                 pos[1] = left_edge[1] + dds[1] * 0.5             # <<<<<<<<<<<<<<
 *                 for j in range(dim[1]):
 *                     pos[2] = left_edge[2] + dds[2] * 0.5
 */
          (__pyx_v_pos[1]) = ((__pyx_v_left_edge[1]) + ((__pyx_v_dds[1]) * 0.5));

          /* "yt/geometry/selection_routines.pyx":469
 *             for i in range(dim[0]):
 *                 pos[1] = left_edge[1] + dds[1] * 0.5
 *                 for j in range(dim[1]):             # <<<<<<<<<<<<<<
 *                     pos[2] = left_edge[2] + dds[2] * 0.5
 *                     for k in range(dim[2]):
 */
          __pyx_t_6 = (__pyx_v_dim[1]);
          __pyx_t_7 = __pyx_t_6;
          for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
            __pyx_v_j = __pyx_t_8;

            /* "yt/geometry/selection_routines.pyx":470
 *                 pos[1] = left_edge[1] + dds[1] * 0.5
 *                 for j in range(dim[1]):
 *                     pos[2] = left_edge[2] + dds[2] * 0.5             # <<<<<<<<<<<<<<
 *                     for k in range(dim[2]):
 *                         if child_mask[i, j, k] == 1 or this_level == 1:
 */
            (__pyx_v_pos[2]) = ((__pyx_v_left_edge[2]) + ((__pyx_v_dds[2]) * 0.5));

            /* "yt/geometry/selection_routines.pyx":471
 *                 for j in range(dim[1]):
 *                     pos[2] = left_edge[2] + dds[2] * 0.5
 *                     for k in range(dim[2]):             # <<<<<<<<<<<<<<
 *                         if child_mask[i, j, k] == 1 or this_level == 1:
 *                             mask[i, j, k] = self.select_cell(pos, dds)
 */
            __pyx_t_9 = (__pyx_v_dim[2]);
            __pyx_t_10 = __pyx_t_9;
            for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
              __pyx_v_k = __pyx_t_11;

              /* "yt/geometry/selection_routines.pyx":472
 *                     pos[2] = left_edge[2] + dds[2] * 0.5
 *                     for k in range(dim[2]):
 *                         if child_mask[i, j, k] == 1 or this_level == 1:             # <<<<<<<<<<<<<<
 *                             mask[i, j, k] = self.select_cell(pos, dds)
 *                             total += mask[i, j, k]
 */
              __pyx_t_12 = __pyx_v_i;
              __pyx_t_13 = __pyx_v_j;
              __pyx_t_14 = __pyx_v_k;
              __pyx_t_2 = (((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_child_mask.diminfo[0].strides, __pyx_t_13, __pyx_pybuffernd_child_mask.diminfo[1].strides, __pyx_t_14, __pyx_pybuffernd_child_mask.diminfo[2].strides)) == 1) != 0);
              if (!__pyx_t_2) {
              } else {
                __pyx_t_1 = __pyx_t_2;
                goto __pyx_L17_bool_binop_done;
              }
              __pyx_t_2 = ((__pyx_v_this_level == 1) != 0);
              __pyx_t_1 = __pyx_t_2;
              __pyx_L17_bool_binop_done:;
              if (__pyx_t_1) {

                /* "yt/geometry/selection_routines.pyx":473
 *                     for k in range(dim[2]):
 *                         if child_mask[i, j, k] == 1 or this_level == 1:
 *                             mask[i, j, k] = self.select_cell(pos, dds)             # <<<<<<<<<<<<<<
 *                             total += mask[i, j, k]
 *                         pos[2] += dds[2]
 */
                __pyx_t_14 = __pyx_v_i;
                __pyx_t_13 = __pyx_v_j;
                __pyx_t_12 = __pyx_v_k;
                *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_mask.diminfo[0].strides, __pyx_t_13, __pyx_pybuffernd_mask.diminfo[1].strides, __pyx_t_12, __pyx_pybuffernd_mask.diminfo[2].strides) = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_cell(__pyx_v_self, __pyx_v_pos, __pyx_v_dds);

                /* "yt/geometry/selection_routines.pyx":474
 *                         if child_mask[i, j, k] == 1 or this_level == 1:
 *                             mask[i, j, k] = self.select_cell(pos, dds)
 *                             total += mask[i, j, k]             # <<<<<<<<<<<<<<
 *                         pos[2] += dds[2]
 *                     pos[1] += dds[1]
 */
                __pyx_t_15 = __pyx_v_i;
                __pyx_t_16 = __pyx_v_j;
                __pyx_t_17 = __pyx_v_k;
                __pyx_v_total = (__pyx_v_total + (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_mask.diminfo[0].strides, __pyx_t_16, __pyx_pybuffernd_mask.diminfo[1].strides, __pyx_t_17, __pyx_pybuffernd_mask.diminfo[2].strides)));

                /* "yt/geometry/selection_routines.pyx":472
 *                     pos[2] = left_edge[2] + dds[2] * 0.5
 *                     for k in range(dim[2]):
 *                         if child_mask[i, j, k] == 1 or this_level == 1:             # <<<<<<<<<<<<<<
 *                             mask[i, j, k] = self.select_cell(pos, dds)
 *                             total += mask[i, j, k]
 */
              }

              /* "yt/geometry/selection_routines.pyx":475
 *                             mask[i, j, k] = self.select_cell(pos, dds)
 *                             total += mask[i, j, k]
 *                         pos[2] += dds[2]             # <<<<<<<<<<<<<<
 *                     pos[1] += dds[1]
 *                 pos[0] += dds[0]
 */
              __pyx_t_18 = 2;
              (__pyx_v_pos[__pyx_t_18]) = ((__pyx_v_pos[__pyx_t_18]) + (__pyx_v_dds[2]));
            }

            /* "yt/geometry/selection_routines.pyx":476
 *                             total += mask[i, j, k]
 *                         pos[2] += dds[2]
 *                     pos[1] += dds[1]             # <<<<<<<<<<<<<<
 *                 pos[0] += dds[0]
 *         return total
 */
            __pyx_t_18 = 1;
            (__pyx_v_pos[__pyx_t_18]) = ((__pyx_v_pos[__pyx_t_18]) + (__pyx_v_dds[1]));
          }

          /* "yt/geometry/selection_routines.pyx":477
 *                         pos[2] += dds[2]
 *                     pos[1] += dds[1]
 *                 pos[0] += dds[0]             # <<<<<<<<<<<<<<
 *         return total
 * 
 */
          __pyx_t_18 = 0;
          (__pyx_v_pos[__pyx_t_18]) = ((__pyx_v_pos[__pyx_t_18]) + (__pyx_v_dds[0]));
        }
      }

      /* "yt/geometry/selection_routines.pyx":465
 *         if level == self.max_level:
 *             this_level = 1
 *         with nogil:             # <<<<<<<<<<<<<<
 *             pos[0] = left_edge[0] + dds[0] * 0.5
 *             for i in range(dim[0]):
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L9;
        }
        __pyx_L9:;
      }
  }

  /* "yt/geometry/selection_routines.pyx":478
 *                     pos[1] += dds[1]
 *                 pos[0] += dds[0]
 *         return total             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = __pyx_v_total;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":452
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int fill_mask_selector(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                 np.float64_t right_edge[3],
 *                                 np.float64_t dds[3], int dim[3],
 */

  /* function exit code */
  __pyx_L1_error:;
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_WriteUnraisable("yt.geometry.selection_routines.SelectorObject.fill_mask_selector", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":483
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void visit_grid_cells(self, GridVisitorData *data,             # <<<<<<<<<<<<<<
 *                               grid_visitor_function *func,
 *                               np.uint8_t *cached_mask = NULL):
 */

static void __pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_visit_grid_cells(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, struct __pyx_t_2yt_8geometry_13grid_visitors_GridVisitorData *__pyx_v_data, __pyx_t_2yt_8geometry_13grid_visitors_grid_visitor_function *__pyx_v_func, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_visit_grid_cells *__pyx_optional_args) {

  /* "yt/geometry/selection_routines.pyx":485
 *     cdef void visit_grid_cells(self, GridVisitorData *data,
 *                               grid_visitor_function *func,
 *                               np.uint8_t *cached_mask = NULL):             # <<<<<<<<<<<<<<
 *         # This function accepts a grid visitor function, the data that
 *         # corresponds to the current grid being examined (the most important
 */
  __pyx_t_5numpy_uint8_t *__pyx_v_cached_mask = ((__pyx_t_5numpy_uint8_t *)NULL);
  __pyx_t_5numpy_float64_t __pyx_v_left_edge[3];
  __pyx_t_5numpy_float64_t __pyx_v_right_edge[3];
  __pyx_t_5numpy_float64_t __pyx_v_dds[3];
  int __pyx_v_dim[3];
  int __pyx_v_this_level;
  int __pyx_v_level;
  int __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_pos[3];
  __pyx_t_5numpy_uint8_t __pyx_v_child_masked;
  __pyx_t_5numpy_uint8_t __pyx_v_selected;
  CYTHON_UNUSED int __pyx_v_j;
  CYTHON_UNUSED int __pyx_v_k;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_int32_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  long __pyx_t_13;
  __Pyx_RefNannySetupContext("visit_grid_cells", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_cached_mask = __pyx_optional_args->cached_mask;
    }
  }

  /* "yt/geometry/selection_routines.pyx":495
 *         cdef np.float64_t dds[3]
 *         cdef int dim[3]
 *         cdef int this_level = 0, level, i             # <<<<<<<<<<<<<<
 *         cdef np.float64_t pos[3]
 *         level = data.grid.level
 */
  __pyx_v_this_level = 0;

  /* "yt/geometry/selection_routines.pyx":497
 *         cdef int this_level = 0, level, i
 *         cdef np.float64_t pos[3]
 *         level = data.grid.level             # <<<<<<<<<<<<<<
 *         if level < self.min_level or level > self.max_level:
 *             return
 */
  __pyx_t_1 = __pyx_v_data->grid->level;
  __pyx_v_level = __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":498
 *         cdef np.float64_t pos[3]
 *         level = data.grid.level
 *         if level < self.min_level or level > self.max_level:             # <<<<<<<<<<<<<<
 *             return
 *         if level == self.max_level:
 */
  __pyx_t_3 = ((__pyx_v_level < __pyx_v_self->min_level) != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = ((__pyx_v_level > __pyx_v_self->max_level) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "yt/geometry/selection_routines.pyx":499
 *         level = data.grid.level
 *         if level < self.min_level or level > self.max_level:
 *             return             # <<<<<<<<<<<<<<
 *         if level == self.max_level:
 *             this_level = 1
 */
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":498
 *         cdef np.float64_t pos[3]
 *         level = data.grid.level
 *         if level < self.min_level or level > self.max_level:             # <<<<<<<<<<<<<<
 *             return
 *         if level == self.max_level:
 */
  }

  /* "yt/geometry/selection_routines.pyx":500
 *         if level < self.min_level or level > self.max_level:
 *             return
 *         if level == self.max_level:             # <<<<<<<<<<<<<<
 *             this_level = 1
 *         cdef np.uint8_t child_masked, selected
 */
  __pyx_t_2 = ((__pyx_v_level == __pyx_v_self->max_level) != 0);
  if (__pyx_t_2) {

    /* "yt/geometry/selection_routines.pyx":501
 *             return
 *         if level == self.max_level:
 *             this_level = 1             # <<<<<<<<<<<<<<
 *         cdef np.uint8_t child_masked, selected
 *         for i in range(3):
 */
    __pyx_v_this_level = 1;

    /* "yt/geometry/selection_routines.pyx":500
 *         if level < self.min_level or level > self.max_level:
 *             return
 *         if level == self.max_level:             # <<<<<<<<<<<<<<
 *             this_level = 1
 *         cdef np.uint8_t child_masked, selected
 */
  }

  /* "yt/geometry/selection_routines.pyx":503
 *             this_level = 1
 *         cdef np.uint8_t child_masked, selected
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             left_edge[i] = data.grid.left_edge[i]
 *             right_edge[i] = data.grid.right_edge[i]
 */
  for (__pyx_t_4 = 0; __pyx_t_4 < 3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "yt/geometry/selection_routines.pyx":504
 *         cdef np.uint8_t child_masked, selected
 *         for i in range(3):
 *             left_edge[i] = data.grid.left_edge[i]             # <<<<<<<<<<<<<<
 *             right_edge[i] = data.grid.right_edge[i]
 *             dds[i] = (right_edge[i] - left_edge[i])/data.grid.dims[i]
 */
    (__pyx_v_left_edge[__pyx_v_i]) = (__pyx_v_data->grid->left_edge[__pyx_v_i]);

    /* "yt/geometry/selection_routines.pyx":505
 *         for i in range(3):
 *             left_edge[i] = data.grid.left_edge[i]
 *             right_edge[i] = data.grid.right_edge[i]             # <<<<<<<<<<<<<<
 *             dds[i] = (right_edge[i] - left_edge[i])/data.grid.dims[i]
 *             dim[i] = data.grid.dims[i]
 */
    (__pyx_v_right_edge[__pyx_v_i]) = (__pyx_v_data->grid->right_edge[__pyx_v_i]);

    /* "yt/geometry/selection_routines.pyx":506
 *             left_edge[i] = data.grid.left_edge[i]
 *             right_edge[i] = data.grid.right_edge[i]
 *             dds[i] = (right_edge[i] - left_edge[i])/data.grid.dims[i]             # <<<<<<<<<<<<<<
 *             dim[i] = data.grid.dims[i]
 *         with nogil:
 */
    (__pyx_v_dds[__pyx_v_i]) = (((__pyx_v_right_edge[__pyx_v_i]) - (__pyx_v_left_edge[__pyx_v_i])) / (__pyx_v_data->grid->dims[__pyx_v_i]));

    /* "yt/geometry/selection_routines.pyx":507
 *             right_edge[i] = data.grid.right_edge[i]
 *             dds[i] = (right_edge[i] - left_edge[i])/data.grid.dims[i]
 *             dim[i] = data.grid.dims[i]             # <<<<<<<<<<<<<<
 *         with nogil:
 *             pos[0] = left_edge[0] + dds[0] * 0.5
 */
    (__pyx_v_dim[__pyx_v_i]) = (__pyx_v_data->grid->dims[__pyx_v_i]);
  }

  /* "yt/geometry/selection_routines.pyx":508
 *             dds[i] = (right_edge[i] - left_edge[i])/data.grid.dims[i]
 *             dim[i] = data.grid.dims[i]
 *         with nogil:             # <<<<<<<<<<<<<<
 *             pos[0] = left_edge[0] + dds[0] * 0.5
 *             data.pos[0] = 0
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "yt/geometry/selection_routines.pyx":509
 *             dim[i] = data.grid.dims[i]
 *         with nogil:
 *             pos[0] = left_edge[0] + dds[0] * 0.5             # <<<<<<<<<<<<<<
 *             data.pos[0] = 0
 *             for i in range(dim[0]):
 */
        (__pyx_v_pos[0]) = ((__pyx_v_left_edge[0]) + ((__pyx_v_dds[0]) * 0.5));

        /* "yt/geometry/selection_routines.pyx":510
 *         with nogil:
 *             pos[0] = left_edge[0] + dds[0] * 0.5
 *             data.pos[0] = 0             # <<<<<<<<<<<<<<
 *             for i in range(dim[0]):
 *                 pos[1] = left_edge[1] + dds[1] * 0.5
 */
        (__pyx_v_data->pos[0]) = 0;

        /* "yt/geometry/selection_routines.pyx":511
 *             pos[0] = left_edge[0] + dds[0] * 0.5
 *             data.pos[0] = 0
 *             for i in range(dim[0]):             # <<<<<<<<<<<<<<
 *                 pos[1] = left_edge[1] + dds[1] * 0.5
 *                 data.pos[1] = 0
 */
        __pyx_t_4 = (__pyx_v_dim[0]);
        __pyx_t_5 = __pyx_t_4;
        for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
          __pyx_v_i = __pyx_t_6;

          /* "yt/geometry/selection_routines.pyx":512
 *             data.pos[0] = 0
 *             for i in range(dim[0]):
 *                 pos[1] = left_edge[1] + dds[1] * 0.5             # <<<<<<<<<<<<<<
 *                 data.pos[1] = 0
 *                 for j in range(dim[1]):
 */
          (__pyx_v_pos[1]) = ((__pyx_v_left_edge[1]) + ((__pyx_v_dds[1]) * 0.5));

          /* "yt/geometry/selection_routines.pyx":513
 *             for i in range(dim[0]):
 *                 pos[1] = left_edge[1] + dds[1] * 0.5
 *                 data.pos[1] = 0             # <<<<<<<<<<<<<<
 *                 for j in range(dim[1]):
 *                     pos[2] = left_edge[2] + dds[2] * 0.5
 */
          (__pyx_v_data->pos[1]) = 0;

          /* "yt/geometry/selection_routines.pyx":514
 *                 pos[1] = left_edge[1] + dds[1] * 0.5
 *                 data.pos[1] = 0
 *                 for j in range(dim[1]):             # <<<<<<<<<<<<<<
 *                     pos[2] = left_edge[2] + dds[2] * 0.5
 *                     data.pos[2] = 0
 */
          __pyx_t_7 = (__pyx_v_dim[1]);
          __pyx_t_8 = __pyx_t_7;
          for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
            __pyx_v_j = __pyx_t_9;

            /* "yt/geometry/selection_routines.pyx":515
 *                 data.pos[1] = 0
 *                 for j in range(dim[1]):
 *                     pos[2] = left_edge[2] + dds[2] * 0.5             # <<<<<<<<<<<<<<
 *                     data.pos[2] = 0
 *                     for k in range(dim[2]):
 */
            (__pyx_v_pos[2]) = ((__pyx_v_left_edge[2]) + ((__pyx_v_dds[2]) * 0.5));

            /* "yt/geometry/selection_routines.pyx":516
 *                 for j in range(dim[1]):
 *                     pos[2] = left_edge[2] + dds[2] * 0.5
 *                     data.pos[2] = 0             # <<<<<<<<<<<<<<
 *                     for k in range(dim[2]):
 *                         # We short-circuit if we have a cache; if we don't, we
 */
            (__pyx_v_data->pos[2]) = 0;

            /* "yt/geometry/selection_routines.pyx":517
 *                     pos[2] = left_edge[2] + dds[2] * 0.5
 *                     data.pos[2] = 0
 *                     for k in range(dim[2]):             # <<<<<<<<<<<<<<
 *                         # We short-circuit if we have a cache; if we don't, we
 *                         # only set selected to true if it's *not* masked by a
 */
            __pyx_t_10 = (__pyx_v_dim[2]);
            __pyx_t_11 = __pyx_t_10;
            for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
              __pyx_v_k = __pyx_t_12;

              /* "yt/geometry/selection_routines.pyx":521
 *                         # only set selected to true if it's *not* masked by a
 *                         # child and it *is* selected.
 *                         if cached_mask != NULL:             # <<<<<<<<<<<<<<
 *                             selected = ba_get_value(cached_mask,
 *                                                     data.global_index)
 */
              __pyx_t_2 = ((__pyx_v_cached_mask != NULL) != 0);
              if (__pyx_t_2) {

                /* "yt/geometry/selection_routines.pyx":522
 *                         # child and it *is* selected.
 *                         if cached_mask != NULL:
 *                             selected = ba_get_value(cached_mask,             # <<<<<<<<<<<<<<
 *                                                     data.global_index)
 *                         else:
 */
                __pyx_v_selected = __pyx_f_2yt_9utilities_3lib_8bitarray_ba_get_value(__pyx_v_cached_mask, __pyx_v_data->global_index);

                /* "yt/geometry/selection_routines.pyx":521
 *                         # only set selected to true if it's *not* masked by a
 *                         # child and it *is* selected.
 *                         if cached_mask != NULL:             # <<<<<<<<<<<<<<
 *                             selected = ba_get_value(cached_mask,
 *                                                     data.global_index)
 */
                goto __pyx_L18;
              }

              /* "yt/geometry/selection_routines.pyx":525
 *                                                     data.global_index)
 *                         else:
 *                             if this_level == 1:             # <<<<<<<<<<<<<<
 *                                 child_masked = 0
 *                             else:
 */
              /*else*/ {
                __pyx_t_2 = ((__pyx_v_this_level == 1) != 0);
                if (__pyx_t_2) {

                  /* "yt/geometry/selection_routines.pyx":526
 *                         else:
 *                             if this_level == 1:
 *                                 child_masked = 0             # <<<<<<<<<<<<<<
 *                             else:
 *                                 child_masked = check_child_masked(data)
 */
                  __pyx_v_child_masked = 0;

                  /* "yt/geometry/selection_routines.pyx":525
 *                                                     data.global_index)
 *                         else:
 *                             if this_level == 1:             # <<<<<<<<<<<<<<
 *                                 child_masked = 0
 *                             else:
 */
                  goto __pyx_L19;
                }

                /* "yt/geometry/selection_routines.pyx":528
 *                                 child_masked = 0
 *                             else:
 *                                 child_masked = check_child_masked(data)             # <<<<<<<<<<<<<<
 *                             if child_masked == 0:
 *                                 selected = self.select_cell(pos, dds)
 */
                /*else*/ {
                  __pyx_v_child_masked = __pyx_f_2yt_8geometry_13grid_visitors_check_child_masked(__pyx_v_data);
                }
                __pyx_L19:;

                /* "yt/geometry/selection_routines.pyx":529
 *                             else:
 *                                 child_masked = check_child_masked(data)
 *                             if child_masked == 0:             # <<<<<<<<<<<<<<
 *                                 selected = self.select_cell(pos, dds)
 *                             else:
 */
                __pyx_t_2 = ((__pyx_v_child_masked == 0) != 0);
                if (__pyx_t_2) {

                  /* "yt/geometry/selection_routines.pyx":530
 *                                 child_masked = check_child_masked(data)
 *                             if child_masked == 0:
 *                                 selected = self.select_cell(pos, dds)             # <<<<<<<<<<<<<<
 *                             else:
 *                                 selected = 0
 */
                  __pyx_v_selected = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_cell(__pyx_v_self, __pyx_v_pos, __pyx_v_dds);

                  /* "yt/geometry/selection_routines.pyx":529
 *                             else:
 *                                 child_masked = check_child_masked(data)
 *                             if child_masked == 0:             # <<<<<<<<<<<<<<
 *                                 selected = self.select_cell(pos, dds)
 *                             else:
 */
                  goto __pyx_L20;
                }

                /* "yt/geometry/selection_routines.pyx":532
 *                                 selected = self.select_cell(pos, dds)
 *                             else:
 *                                 selected = 0             # <<<<<<<<<<<<<<
 *                         func(data, selected)
 *                         data.global_index += 1
 */
                /*else*/ {
                  __pyx_v_selected = 0;
                }
                __pyx_L20:;
              }
              __pyx_L18:;

              /* "yt/geometry/selection_routines.pyx":533
 *                             else:
 *                                 selected = 0
 *                         func(data, selected)             # <<<<<<<<<<<<<<
 *                         data.global_index += 1
 *                         pos[2] += dds[2]
 */
              __pyx_v_func(__pyx_v_data, __pyx_v_selected);

              /* "yt/geometry/selection_routines.pyx":534
 *                                 selected = 0
 *                         func(data, selected)
 *                         data.global_index += 1             # <<<<<<<<<<<<<<
 *                         pos[2] += dds[2]
 *                         data.pos[2] += 1
 */
              __pyx_v_data->global_index = (__pyx_v_data->global_index + 1);

              /* "yt/geometry/selection_routines.pyx":535
 *                         func(data, selected)
 *                         data.global_index += 1
 *                         pos[2] += dds[2]             # <<<<<<<<<<<<<<
 *                         data.pos[2] += 1
 *                     pos[1] += dds[1]
 */
              __pyx_t_13 = 2;
              (__pyx_v_pos[__pyx_t_13]) = ((__pyx_v_pos[__pyx_t_13]) + (__pyx_v_dds[2]));

              /* "yt/geometry/selection_routines.pyx":536
 *                         data.global_index += 1
 *                         pos[2] += dds[2]
 *                         data.pos[2] += 1             # <<<<<<<<<<<<<<
 *                     pos[1] += dds[1]
 *                     data.pos[1] += 1
 */
              __pyx_t_13 = 2;
              (__pyx_v_data->pos[__pyx_t_13]) = ((__pyx_v_data->pos[__pyx_t_13]) + 1);
            }

            /* "yt/geometry/selection_routines.pyx":537
 *                         pos[2] += dds[2]
 *                         data.pos[2] += 1
 *                     pos[1] += dds[1]             # <<<<<<<<<<<<<<
 *                     data.pos[1] += 1
 *                 pos[0] += dds[0]
 */
            __pyx_t_13 = 1;
            (__pyx_v_pos[__pyx_t_13]) = ((__pyx_v_pos[__pyx_t_13]) + (__pyx_v_dds[1]));

            /* "yt/geometry/selection_routines.pyx":538
 *                         data.pos[2] += 1
 *                     pos[1] += dds[1]
 *                     data.pos[1] += 1             # <<<<<<<<<<<<<<
 *                 pos[0] += dds[0]
 *                 data.pos[0] += 1
 */
            __pyx_t_13 = 1;
            (__pyx_v_data->pos[__pyx_t_13]) = ((__pyx_v_data->pos[__pyx_t_13]) + 1);
          }

          /* "yt/geometry/selection_routines.pyx":539
 *                     pos[1] += dds[1]
 *                     data.pos[1] += 1
 *                 pos[0] += dds[0]             # <<<<<<<<<<<<<<
 *                 data.pos[0] += 1
 * 
 */
          __pyx_t_13 = 0;
          (__pyx_v_pos[__pyx_t_13]) = ((__pyx_v_pos[__pyx_t_13]) + (__pyx_v_dds[0]));

          /* "yt/geometry/selection_routines.pyx":540
 *                     data.pos[1] += 1
 *                 pos[0] += dds[0]
 *                 data.pos[0] += 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
          __pyx_t_13 = 0;
          (__pyx_v_data->pos[__pyx_t_13]) = ((__pyx_v_data->pos[__pyx_t_13]) + 1);
        }
      }

      /* "yt/geometry/selection_routines.pyx":508
 *             dds[i] = (right_edge[i] - left_edge[i])/data.grid.dims[i]
 *             dim[i] = data.grid.dims[i]
 *         with nogil:             # <<<<<<<<<<<<<<
 *             pos[0] = left_edge[0] + dds[0] * 0.5
 *             data.pos[0] = 0
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L11;
        }
        __pyx_L11:;
      }
  }

  /* "yt/geometry/selection_routines.pyx":483
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void visit_grid_cells(self, GridVisitorData *data,             # <<<<<<<<<<<<<<
 *                               grid_visitor_function *func,
 *                               np.uint8_t *cached_mask = NULL):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/selection_routines.pyx":545
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def count_points(self, np.ndarray[floating, ndim=1] x,             # <<<<<<<<<<<<<<
 *                            np.ndarray[floating, ndim=1] y,
 *                            np.ndarray[floating, ndim=1] z,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_15count_points(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_8geometry_18selection_routines_14SelectorObject_15count_points = {"count_points", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_15count_points, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_15count_points(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(0, 545, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(0, 545, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(0, 545, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 545, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_14count_points(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_14count_points(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  PyTypeObject *__pyx_v_ndarray = 0;
  __Pyx_memviewslice __pyx_v_memslice;
  Py_ssize_t __pyx_v_itemsize;
  CYTHON_UNUSED int __pyx_v_dtype_signed;
  char __pyx_v_kind;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_dtype = NULL;
  PyObject *__pyx_v_arg_base = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  __Pyx_memviewslice __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_points", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 545, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 545, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 545, __pyx_L1_error)
  __pyx_t_2 = ((1 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 545, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 545, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_x, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 545, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 545, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 545, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 545, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_5);
    __Pyx_GIVEREF(__pyx_int_5);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_5);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 545, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = (__pyx_v_ndarray != ((PyTypeObject*)Py_None));
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg, __pyx_v_ndarray); 
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_dtype = __pyx_t_6;
        __pyx_t_6 = 0;
        goto __pyx_L12;
      }
      __pyx_t_2 = __pyx_memoryview_check(__pyx_v_arg); 
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_base); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_arg_base = __pyx_t_6;
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg_base, __pyx_v_ndarray); 
        __pyx_t_2 = (__pyx_t_3 != 0);
        if (__pyx_t_2) {
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg_base, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_v_dtype = __pyx_t_6;
          __pyx_t_6 = 0;
          goto __pyx_L13;
        }
        /*else*/ {
          __Pyx_INCREF(Py_None);
          __pyx_v_dtype = Py_None;
        }
        __pyx_L13:;
        goto __pyx_L12;
      }
      /*else*/ {
        __Pyx_INCREF(Py_None);
        __pyx_v_dtype = Py_None;
      }
      __pyx_L12:;
      __pyx_v_itemsize = -1L;
      __pyx_t_2 = (__pyx_v_dtype != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 545, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_itemsize = __pyx_t_5;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyObject_Ord(__pyx_t_6); if (unlikely(__pyx_t_7 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(0, 545, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_kind = __pyx_t_7;
        __pyx_v_dtype_signed = (__pyx_v_kind == 'i');
        switch (__pyx_v_kind) {
          case 'i':
          case 'u':
          break;
          case 'f':
          __pyx_t_2 = (((sizeof(float)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 545, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 1) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L16_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 545, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          __pyx_t_2 = (((sizeof(double)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L19_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 545, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 1) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L19_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_double, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 545, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          break;
          case 'c':
          break;
          case 'O':
          break;
          default: break;
        }
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L22_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(float))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L22_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_float(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 545, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L26_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(double))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L26_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_double, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 545, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 545, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 545, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_9), (&__pyx_t_10)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_9, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_10);
    if (unlikely(__pyx_t_11 == 0)) break;
    if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_12 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_14, __pyx_kp_s_) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s_);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_split); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_12, __pyx_kp_s__2) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s__2);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_15 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 545, __pyx_L1_error)
    __pyx_t_16 = __pyx_t_15;
    for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
      __pyx_v_i = __pyx_t_17;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_3 = (__pyx_v_dst_type != Py_None);
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_13 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 545, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 545, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (__pyx_t_2) {
          __pyx_v_match_found = 1;
          goto __pyx_L34;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L32_break;
        }
        __pyx_L34:;
      }
    }
    __pyx_L32_break:;
    __pyx_t_2 = (__pyx_v_match_found != 0);
    if (__pyx_t_2) {
      __pyx_t_18 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 545, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 545, __pyx_L1_error)
  }
  __pyx_t_9 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 545, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_9 > 1) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 545, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 545, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_dtype);
  __Pyx_XDECREF(__pyx_v_arg_base);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_0__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_25count_points(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_2yt_8geometry_18selection_routines_14SelectorObject_25count_points = {"__pyx_fuse_0count_points", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_0__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_25count_points, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_fuse_0__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_25count_points(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_x = 0;
  PyArrayObject *__pyx_v_y = 0;
  PyArrayObject *__pyx_v_z = 0;
  __pyx_t_5numpy_float64_t __pyx_v_radius;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("count_points (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,&__pyx_n_s_z,&__pyx_n_s_radius,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("count_points", 1, 4, 4, 1); __PYX_ERR(0, 545, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_z)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("count_points", 1, 4, 4, 2); __PYX_ERR(0, 545, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_radius)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("count_points", 1, 4, 4, 3); __PYX_ERR(0, 545, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "count_points") < 0)) __PYX_ERR(0, 545, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_x = ((PyArrayObject *)values[0]);
    __pyx_v_y = ((PyArrayObject *)values[1]);
    __pyx_v_z = ((PyArrayObject *)values[2]);
    __pyx_v_radius = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_radius == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 548, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("count_points", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 545, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.count_points", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_x), __pyx_ptype_5numpy_ndarray, 1, "x", 0))) __PYX_ERR(0, 545, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_5numpy_ndarray, 1, "y", 0))) __PYX_ERR(0, 546, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_z), __pyx_ptype_5numpy_ndarray, 1, "z", 0))) __PYX_ERR(0, 547, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_24count_points(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_x, __pyx_v_y, __pyx_v_z, __pyx_v_radius);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_24count_points(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyArrayObject *__pyx_v_x, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_z, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_v_count;
  int __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_pos[3];
  __Pyx_LocalBuf_ND __pyx_pybuffernd_x;
  __Pyx_Buffer __pyx_pybuffer_x;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_y;
  __Pyx_Buffer __pyx_pybuffer_y;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_z;
  __Pyx_Buffer __pyx_pybuffer_z;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_0count_points", 0);
  __pyx_pybuffer_x.pybuffer.buf = NULL;
  __pyx_pybuffer_x.refcount = 0;
  __pyx_pybuffernd_x.data = NULL;
  __pyx_pybuffernd_x.rcbuffer = &__pyx_pybuffer_x;
  __pyx_pybuffer_y.pybuffer.buf = NULL;
  __pyx_pybuffer_y.refcount = 0;
  __pyx_pybuffernd_y.data = NULL;
  __pyx_pybuffernd_y.rcbuffer = &__pyx_pybuffer_y;
  __pyx_pybuffer_z.pybuffer.buf = NULL;
  __pyx_pybuffer_z.refcount = 0;
  __pyx_pybuffernd_z.data = NULL;
  __pyx_pybuffernd_z.rcbuffer = &__pyx_pybuffer_z;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_x.rcbuffer->pybuffer, (PyObject*)__pyx_v_x, &__Pyx_TypeInfo_float, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 545, __pyx_L1_error)
  }
  __pyx_pybuffernd_x.diminfo[0].strides = __pyx_pybuffernd_x.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_x.diminfo[0].shape = __pyx_pybuffernd_x.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_y.rcbuffer->pybuffer, (PyObject*)__pyx_v_y, &__Pyx_TypeInfo_float, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 545, __pyx_L1_error)
  }
  __pyx_pybuffernd_y.diminfo[0].strides = __pyx_pybuffernd_y.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_y.diminfo[0].shape = __pyx_pybuffernd_y.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_z.rcbuffer->pybuffer, (PyObject*)__pyx_v_z, &__Pyx_TypeInfo_float, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 545, __pyx_L1_error)
  }
  __pyx_pybuffernd_z.diminfo[0].strides = __pyx_pybuffernd_z.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_z.diminfo[0].shape = __pyx_pybuffernd_z.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/selection_routines.pyx":549
 *                            np.ndarray[floating, ndim=1] z,
 *                            np.float64_t radius):
 *         cdef int count = 0             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.float64_t pos[3]
 */
  __pyx_v_count = 0;

  /* "yt/geometry/selection_routines.pyx":552
 *         cdef int i
 *         cdef np.float64_t pos[3]
 *         _ensure_code(x)             # <<<<<<<<<<<<<<
 *         _ensure_code(y)
 *         _ensure_code(z)
 */
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(((PyObject *)__pyx_v_x)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":553
 *         cdef np.float64_t pos[3]
 *         _ensure_code(x)
 *         _ensure_code(y)             # <<<<<<<<<<<<<<
 *         _ensure_code(z)
 *         with nogil:
 */
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(((PyObject *)__pyx_v_y)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":554
 *         _ensure_code(x)
 *         _ensure_code(y)
 *         _ensure_code(z)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             if radius == 0.0 :
 */
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(((PyObject *)__pyx_v_z)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":555
 *         _ensure_code(y)
 *         _ensure_code(z)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             if radius == 0.0 :
 *                 for i in range(x.shape[0]):
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "yt/geometry/selection_routines.pyx":556
 *         _ensure_code(z)
 *         with nogil:
 *             if radius == 0.0 :             # <<<<<<<<<<<<<<
 *                 for i in range(x.shape[0]):
 *                     pos[0] = x[i]
 */
        __pyx_t_2 = ((__pyx_v_radius == 0.0) != 0);
        if (__pyx_t_2) {

          /* "yt/geometry/selection_routines.pyx":557
 *         with nogil:
 *             if radius == 0.0 :
 *                 for i in range(x.shape[0]):             # <<<<<<<<<<<<<<
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]
 */
          __pyx_t_3 = (__pyx_v_x->dimensions[0]);
          __pyx_t_4 = __pyx_t_3;
          for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
            __pyx_v_i = __pyx_t_5;

            /* "yt/geometry/selection_routines.pyx":558
 *             if radius == 0.0 :
 *                 for i in range(x.shape[0]):
 *                     pos[0] = x[i]             # <<<<<<<<<<<<<<
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]
 */
            __pyx_t_6 = __pyx_v_i;
            (__pyx_v_pos[0]) = (*__Pyx_BufPtrStrided1d(float *, __pyx_pybuffernd_x.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_x.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":559
 *                 for i in range(x.shape[0]):
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]             # <<<<<<<<<<<<<<
 *                     pos[2] = z[i]
 *                     count += self.select_point(pos)
 */
            __pyx_t_6 = __pyx_v_i;
            (__pyx_v_pos[1]) = (*__Pyx_BufPtrStrided1d(float *, __pyx_pybuffernd_y.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_y.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":560
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]             # <<<<<<<<<<<<<<
 *                     count += self.select_point(pos)
 *             else :
 */
            __pyx_t_6 = __pyx_v_i;
            (__pyx_v_pos[2]) = (*__Pyx_BufPtrStrided1d(float *, __pyx_pybuffernd_z.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_z.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":561
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]
 *                     count += self.select_point(pos)             # <<<<<<<<<<<<<<
 *             else :
 *                 for i in range(x.shape[0]):
 */
            __pyx_v_count = (__pyx_v_count + ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_point(__pyx_v_self, __pyx_v_pos));
          }

          /* "yt/geometry/selection_routines.pyx":556
 *         _ensure_code(z)
 *         with nogil:
 *             if radius == 0.0 :             # <<<<<<<<<<<<<<
 *                 for i in range(x.shape[0]):
 *                     pos[0] = x[i]
 */
          goto __pyx_L6;
        }

        /* "yt/geometry/selection_routines.pyx":563
 *                     count += self.select_point(pos)
 *             else :
 *                 for i in range(x.shape[0]):             # <<<<<<<<<<<<<<
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]
 */
        /*else*/ {
          __pyx_t_3 = (__pyx_v_x->dimensions[0]);
          __pyx_t_4 = __pyx_t_3;
          for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
            __pyx_v_i = __pyx_t_5;

            /* "yt/geometry/selection_routines.pyx":564
 *             else :
 *                 for i in range(x.shape[0]):
 *                     pos[0] = x[i]             # <<<<<<<<<<<<<<
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]
 */
            __pyx_t_6 = __pyx_v_i;
            (__pyx_v_pos[0]) = (*__Pyx_BufPtrStrided1d(float *, __pyx_pybuffernd_x.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_x.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":565
 *                 for i in range(x.shape[0]):
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]             # <<<<<<<<<<<<<<
 *                     pos[2] = z[i]
 *                     count += self.select_sphere(pos, radius)
 */
            __pyx_t_6 = __pyx_v_i;
            (__pyx_v_pos[1]) = (*__Pyx_BufPtrStrided1d(float *, __pyx_pybuffernd_y.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_y.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":566
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]             # <<<<<<<<<<<<<<
 *                     count += self.select_sphere(pos, radius)
 *         return count
 */
            __pyx_t_6 = __pyx_v_i;
            (__pyx_v_pos[2]) = (*__Pyx_BufPtrStrided1d(float *, __pyx_pybuffernd_z.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_z.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":567
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]
 *                     count += self.select_sphere(pos, radius)             # <<<<<<<<<<<<<<
 *         return count
 * 
 */
            __pyx_v_count = (__pyx_v_count + ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_sphere(__pyx_v_self, __pyx_v_pos, __pyx_v_radius));
          }
        }
        __pyx_L6:;
      }

      /* "yt/geometry/selection_routines.pyx":555
 *         _ensure_code(y)
 *         _ensure_code(z)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             if radius == 0.0 :
 *                 for i in range(x.shape[0]):
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "yt/geometry/selection_routines.pyx":568
 *                     pos[2] = z[i]
 *                     count += self.select_sphere(pos, radius)
 *         return count             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":545
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def count_points(self, np.ndarray[floating, ndim=1] x,             # <<<<<<<<<<<<<<
 *                            np.ndarray[floating, ndim=1] y,
 *                            np.ndarray[floating, ndim=1] z,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_x.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_y.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_z.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.count_points", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_x.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_y.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_z.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_1__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_27count_points(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_2yt_8geometry_18selection_routines_14SelectorObject_27count_points = {"__pyx_fuse_1count_points", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_1__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_27count_points, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_fuse_1__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_27count_points(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_x = 0;
  PyArrayObject *__pyx_v_y = 0;
  PyArrayObject *__pyx_v_z = 0;
  __pyx_t_5numpy_float64_t __pyx_v_radius;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("count_points (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,&__pyx_n_s_z,&__pyx_n_s_radius,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("count_points", 1, 4, 4, 1); __PYX_ERR(0, 545, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_z)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("count_points", 1, 4, 4, 2); __PYX_ERR(0, 545, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_radius)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("count_points", 1, 4, 4, 3); __PYX_ERR(0, 545, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "count_points") < 0)) __PYX_ERR(0, 545, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_x = ((PyArrayObject *)values[0]);
    __pyx_v_y = ((PyArrayObject *)values[1]);
    __pyx_v_z = ((PyArrayObject *)values[2]);
    __pyx_v_radius = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_radius == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 548, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("count_points", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 545, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.count_points", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_x), __pyx_ptype_5numpy_ndarray, 1, "x", 0))) __PYX_ERR(0, 545, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_5numpy_ndarray, 1, "y", 0))) __PYX_ERR(0, 546, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_z), __pyx_ptype_5numpy_ndarray, 1, "z", 0))) __PYX_ERR(0, 547, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_26count_points(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_x, __pyx_v_y, __pyx_v_z, __pyx_v_radius);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_26count_points(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyArrayObject *__pyx_v_x, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_z, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_v_count;
  int __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_pos[3];
  __Pyx_LocalBuf_ND __pyx_pybuffernd_x;
  __Pyx_Buffer __pyx_pybuffer_x;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_y;
  __Pyx_Buffer __pyx_pybuffer_y;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_z;
  __Pyx_Buffer __pyx_pybuffer_z;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_1count_points", 0);
  __pyx_pybuffer_x.pybuffer.buf = NULL;
  __pyx_pybuffer_x.refcount = 0;
  __pyx_pybuffernd_x.data = NULL;
  __pyx_pybuffernd_x.rcbuffer = &__pyx_pybuffer_x;
  __pyx_pybuffer_y.pybuffer.buf = NULL;
  __pyx_pybuffer_y.refcount = 0;
  __pyx_pybuffernd_y.data = NULL;
  __pyx_pybuffernd_y.rcbuffer = &__pyx_pybuffer_y;
  __pyx_pybuffer_z.pybuffer.buf = NULL;
  __pyx_pybuffer_z.refcount = 0;
  __pyx_pybuffernd_z.data = NULL;
  __pyx_pybuffernd_z.rcbuffer = &__pyx_pybuffer_z;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_x.rcbuffer->pybuffer, (PyObject*)__pyx_v_x, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 545, __pyx_L1_error)
  }
  __pyx_pybuffernd_x.diminfo[0].strides = __pyx_pybuffernd_x.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_x.diminfo[0].shape = __pyx_pybuffernd_x.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_y.rcbuffer->pybuffer, (PyObject*)__pyx_v_y, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 545, __pyx_L1_error)
  }
  __pyx_pybuffernd_y.diminfo[0].strides = __pyx_pybuffernd_y.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_y.diminfo[0].shape = __pyx_pybuffernd_y.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_z.rcbuffer->pybuffer, (PyObject*)__pyx_v_z, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 545, __pyx_L1_error)
  }
  __pyx_pybuffernd_z.diminfo[0].strides = __pyx_pybuffernd_z.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_z.diminfo[0].shape = __pyx_pybuffernd_z.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/selection_routines.pyx":549
 *                            np.ndarray[floating, ndim=1] z,
 *                            np.float64_t radius):
 *         cdef int count = 0             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.float64_t pos[3]
 */
  __pyx_v_count = 0;

  /* "yt/geometry/selection_routines.pyx":552
 *         cdef int i
 *         cdef np.float64_t pos[3]
 *         _ensure_code(x)             # <<<<<<<<<<<<<<
 *         _ensure_code(y)
 *         _ensure_code(z)
 */
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(((PyObject *)__pyx_v_x)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":553
 *         cdef np.float64_t pos[3]
 *         _ensure_code(x)
 *         _ensure_code(y)             # <<<<<<<<<<<<<<
 *         _ensure_code(z)
 *         with nogil:
 */
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(((PyObject *)__pyx_v_y)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":554
 *         _ensure_code(x)
 *         _ensure_code(y)
 *         _ensure_code(z)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             if radius == 0.0 :
 */
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(((PyObject *)__pyx_v_z)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":555
 *         _ensure_code(y)
 *         _ensure_code(z)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             if radius == 0.0 :
 *                 for i in range(x.shape[0]):
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "yt/geometry/selection_routines.pyx":556
 *         _ensure_code(z)
 *         with nogil:
 *             if radius == 0.0 :             # <<<<<<<<<<<<<<
 *                 for i in range(x.shape[0]):
 *                     pos[0] = x[i]
 */
        __pyx_t_2 = ((__pyx_v_radius == 0.0) != 0);
        if (__pyx_t_2) {

          /* "yt/geometry/selection_routines.pyx":557
 *         with nogil:
 *             if radius == 0.0 :
 *                 for i in range(x.shape[0]):             # <<<<<<<<<<<<<<
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]
 */
          __pyx_t_3 = (__pyx_v_x->dimensions[0]);
          __pyx_t_4 = __pyx_t_3;
          for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
            __pyx_v_i = __pyx_t_5;

            /* "yt/geometry/selection_routines.pyx":558
 *             if radius == 0.0 :
 *                 for i in range(x.shape[0]):
 *                     pos[0] = x[i]             # <<<<<<<<<<<<<<
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]
 */
            __pyx_t_6 = __pyx_v_i;
            (__pyx_v_pos[0]) = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_x.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_x.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":559
 *                 for i in range(x.shape[0]):
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]             # <<<<<<<<<<<<<<
 *                     pos[2] = z[i]
 *                     count += self.select_point(pos)
 */
            __pyx_t_6 = __pyx_v_i;
            (__pyx_v_pos[1]) = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_y.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_y.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":560
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]             # <<<<<<<<<<<<<<
 *                     count += self.select_point(pos)
 *             else :
 */
            __pyx_t_6 = __pyx_v_i;
            (__pyx_v_pos[2]) = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_z.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_z.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":561
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]
 *                     count += self.select_point(pos)             # <<<<<<<<<<<<<<
 *             else :
 *                 for i in range(x.shape[0]):
 */
            __pyx_v_count = (__pyx_v_count + ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_point(__pyx_v_self, __pyx_v_pos));
          }

          /* "yt/geometry/selection_routines.pyx":556
 *         _ensure_code(z)
 *         with nogil:
 *             if radius == 0.0 :             # <<<<<<<<<<<<<<
 *                 for i in range(x.shape[0]):
 *                     pos[0] = x[i]
 */
          goto __pyx_L6;
        }

        /* "yt/geometry/selection_routines.pyx":563
 *                     count += self.select_point(pos)
 *             else :
 *                 for i in range(x.shape[0]):             # <<<<<<<<<<<<<<
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]
 */
        /*else*/ {
          __pyx_t_3 = (__pyx_v_x->dimensions[0]);
          __pyx_t_4 = __pyx_t_3;
          for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
            __pyx_v_i = __pyx_t_5;

            /* "yt/geometry/selection_routines.pyx":564
 *             else :
 *                 for i in range(x.shape[0]):
 *                     pos[0] = x[i]             # <<<<<<<<<<<<<<
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]
 */
            __pyx_t_6 = __pyx_v_i;
            (__pyx_v_pos[0]) = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_x.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_x.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":565
 *                 for i in range(x.shape[0]):
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]             # <<<<<<<<<<<<<<
 *                     pos[2] = z[i]
 *                     count += self.select_sphere(pos, radius)
 */
            __pyx_t_6 = __pyx_v_i;
            (__pyx_v_pos[1]) = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_y.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_y.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":566
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]             # <<<<<<<<<<<<<<
 *                     count += self.select_sphere(pos, radius)
 *         return count
 */
            __pyx_t_6 = __pyx_v_i;
            (__pyx_v_pos[2]) = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_z.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_z.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":567
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]
 *                     count += self.select_sphere(pos, radius)             # <<<<<<<<<<<<<<
 *         return count
 * 
 */
            __pyx_v_count = (__pyx_v_count + ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_sphere(__pyx_v_self, __pyx_v_pos, __pyx_v_radius));
          }
        }
        __pyx_L6:;
      }

      /* "yt/geometry/selection_routines.pyx":555
 *         _ensure_code(y)
 *         _ensure_code(z)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             if radius == 0.0 :
 *                 for i in range(x.shape[0]):
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "yt/geometry/selection_routines.pyx":568
 *                     pos[2] = z[i]
 *                     count += self.select_sphere(pos, radius)
 *         return count             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":545
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def count_points(self, np.ndarray[floating, ndim=1] x,             # <<<<<<<<<<<<<<
 *                            np.ndarray[floating, ndim=1] y,
 *                            np.ndarray[floating, ndim=1] z,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_x.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_y.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_z.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.count_points", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_x.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_y.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_z.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":573
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_points(self, np.ndarray[floating, ndim=1] x,             # <<<<<<<<<<<<<<
 *                             np.ndarray[floating, ndim=1] y,
 *                             np.ndarray[floating, ndim=1] z,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_17select_points(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_8geometry_18selection_routines_14SelectorObject_17select_points = {"select_points", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_17select_points, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_17select_points(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(0, 573, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(0, 573, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(0, 573, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 573, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_16select_points(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_16select_points(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  PyTypeObject *__pyx_v_ndarray = 0;
  __Pyx_memviewslice __pyx_v_memslice;
  Py_ssize_t __pyx_v_itemsize;
  CYTHON_UNUSED int __pyx_v_dtype_signed;
  char __pyx_v_kind;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_dtype = NULL;
  PyObject *__pyx_v_arg_base = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  __Pyx_memviewslice __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("select_points", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 573, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 573, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 573, __pyx_L1_error)
  __pyx_t_2 = ((1 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 573, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 573, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_x, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 573, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 573, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 573, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 573, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_5);
    __Pyx_GIVEREF(__pyx_int_5);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_5);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 573, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = (__pyx_v_ndarray != ((PyTypeObject*)Py_None));
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg, __pyx_v_ndarray); 
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 573, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_dtype = __pyx_t_6;
        __pyx_t_6 = 0;
        goto __pyx_L12;
      }
      __pyx_t_2 = __pyx_memoryview_check(__pyx_v_arg); 
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_base); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 573, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_arg_base = __pyx_t_6;
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg_base, __pyx_v_ndarray); 
        __pyx_t_2 = (__pyx_t_3 != 0);
        if (__pyx_t_2) {
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg_base, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 573, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_v_dtype = __pyx_t_6;
          __pyx_t_6 = 0;
          goto __pyx_L13;
        }
        /*else*/ {
          __Pyx_INCREF(Py_None);
          __pyx_v_dtype = Py_None;
        }
        __pyx_L13:;
        goto __pyx_L12;
      }
      /*else*/ {
        __Pyx_INCREF(Py_None);
        __pyx_v_dtype = Py_None;
      }
      __pyx_L12:;
      __pyx_v_itemsize = -1L;
      __pyx_t_2 = (__pyx_v_dtype != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 573, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 573, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_itemsize = __pyx_t_5;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 573, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyObject_Ord(__pyx_t_6); if (unlikely(__pyx_t_7 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(0, 573, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_kind = __pyx_t_7;
        __pyx_v_dtype_signed = (__pyx_v_kind == 'i');
        switch (__pyx_v_kind) {
          case 'i':
          case 'u':
          break;
          case 'f':
          __pyx_t_2 = (((sizeof(float)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 573, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 573, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 1) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L16_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 573, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          __pyx_t_2 = (((sizeof(double)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L19_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 573, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 573, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 1) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L19_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_double, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 573, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          break;
          case 'c':
          break;
          case 'O':
          break;
          default: break;
        }
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L22_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(float))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L22_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_float(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 573, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L26_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(double))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L26_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_double, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 573, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 573, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 573, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_9), (&__pyx_t_10)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_9, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_10);
    if (unlikely(__pyx_t_11 == 0)) break;
    if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_12 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_14, __pyx_kp_s_) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s_);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_split); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_12, __pyx_kp_s__2) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s__2);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_15 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 573, __pyx_L1_error)
    __pyx_t_16 = __pyx_t_15;
    for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
      __pyx_v_i = __pyx_t_17;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_3 = (__pyx_v_dst_type != Py_None);
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_13 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 573, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 573, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (__pyx_t_2) {
          __pyx_v_match_found = 1;
          goto __pyx_L34;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L32_break;
        }
        __pyx_L34:;
      }
    }
    __pyx_L32_break:;
    __pyx_t_2 = (__pyx_v_match_found != 0);
    if (__pyx_t_2) {
      __pyx_t_18 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 573, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 573, __pyx_L1_error)
  }
  __pyx_t_9 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 573, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_9 > 1) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 573, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 573, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_dtype);
  __Pyx_XDECREF(__pyx_v_arg_base);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_0__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_31select_points(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_2yt_8geometry_18selection_routines_14SelectorObject_31select_points = {"__pyx_fuse_0select_points", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_0__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_31select_points, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_fuse_0__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_31select_points(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_x = 0;
  PyArrayObject *__pyx_v_y = 0;
  PyArrayObject *__pyx_v_z = 0;
  __pyx_t_5numpy_float64_t __pyx_v_radius;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("select_points (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,&__pyx_n_s_z,&__pyx_n_s_radius,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_points", 1, 4, 4, 1); __PYX_ERR(0, 573, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_z)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_points", 1, 4, 4, 2); __PYX_ERR(0, 573, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_radius)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_points", 1, 4, 4, 3); __PYX_ERR(0, 573, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "select_points") < 0)) __PYX_ERR(0, 573, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_x = ((PyArrayObject *)values[0]);
    __pyx_v_y = ((PyArrayObject *)values[1]);
    __pyx_v_z = ((PyArrayObject *)values[2]);
    __pyx_v_radius = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_radius == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 576, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("select_points", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 573, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.select_points", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_x), __pyx_ptype_5numpy_ndarray, 1, "x", 0))) __PYX_ERR(0, 573, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_5numpy_ndarray, 1, "y", 0))) __PYX_ERR(0, 574, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_z), __pyx_ptype_5numpy_ndarray, 1, "z", 0))) __PYX_ERR(0, 575, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_30select_points(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_x, __pyx_v_y, __pyx_v_z, __pyx_v_radius);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_30select_points(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyArrayObject *__pyx_v_x, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_z, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_v_count;
  int __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_pos[3];
  PyArrayObject *__pyx_v_mask = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_x;
  __Pyx_Buffer __pyx_pybuffer_x;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_y;
  __Pyx_Buffer __pyx_pybuffer_y;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_z;
  __Pyx_Buffer __pyx_pybuffer_z;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyArrayObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  npy_intp __pyx_t_11;
  npy_intp __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_0select_points", 0);
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  __pyx_pybuffer_x.pybuffer.buf = NULL;
  __pyx_pybuffer_x.refcount = 0;
  __pyx_pybuffernd_x.data = NULL;
  __pyx_pybuffernd_x.rcbuffer = &__pyx_pybuffer_x;
  __pyx_pybuffer_y.pybuffer.buf = NULL;
  __pyx_pybuffer_y.refcount = 0;
  __pyx_pybuffernd_y.data = NULL;
  __pyx_pybuffernd_y.rcbuffer = &__pyx_pybuffer_y;
  __pyx_pybuffer_z.pybuffer.buf = NULL;
  __pyx_pybuffer_z.refcount = 0;
  __pyx_pybuffernd_z.data = NULL;
  __pyx_pybuffernd_z.rcbuffer = &__pyx_pybuffer_z;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_x.rcbuffer->pybuffer, (PyObject*)__pyx_v_x, &__Pyx_TypeInfo_float, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 573, __pyx_L1_error)
  }
  __pyx_pybuffernd_x.diminfo[0].strides = __pyx_pybuffernd_x.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_x.diminfo[0].shape = __pyx_pybuffernd_x.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_y.rcbuffer->pybuffer, (PyObject*)__pyx_v_y, &__Pyx_TypeInfo_float, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 573, __pyx_L1_error)
  }
  __pyx_pybuffernd_y.diminfo[0].strides = __pyx_pybuffernd_y.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_y.diminfo[0].shape = __pyx_pybuffernd_y.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_z.rcbuffer->pybuffer, (PyObject*)__pyx_v_z, &__Pyx_TypeInfo_float, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 573, __pyx_L1_error)
  }
  __pyx_pybuffernd_z.diminfo[0].strides = __pyx_pybuffernd_z.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_z.diminfo[0].shape = __pyx_pybuffernd_z.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/selection_routines.pyx":577
 *                             np.ndarray[floating, ndim=1] z,
 *                             np.float64_t radius):
 *         cdef int count = 0             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.float64_t pos[3]
 */
  __pyx_v_count = 0;

  /* "yt/geometry/selection_routines.pyx":581
 *         cdef np.float64_t pos[3]
 *         cdef np.ndarray[np.uint8_t, ndim=1] mask
 *         mask = np.empty(x.shape[0], dtype='uint8')             # <<<<<<<<<<<<<<
 *         _ensure_code(x)
 *         _ensure_code(y)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t((__pyx_v_x->dimensions[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 581, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 581, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
      __pyx_t_7 = __pyx_t_8 = __pyx_t_9 = 0;
    }
    __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 581, __pyx_L1_error)
  }
  __pyx_t_5 = 0;
  __pyx_v_mask = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":582
 *         cdef np.ndarray[np.uint8_t, ndim=1] mask
 *         mask = np.empty(x.shape[0], dtype='uint8')
 *         _ensure_code(x)             # <<<<<<<<<<<<<<
 *         _ensure_code(y)
 *         _ensure_code(z)
 */
  __pyx_t_4 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(((PyObject *)__pyx_v_x)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":583
 *         mask = np.empty(x.shape[0], dtype='uint8')
 *         _ensure_code(x)
 *         _ensure_code(y)             # <<<<<<<<<<<<<<
 *         _ensure_code(z)
 * 
 */
  __pyx_t_4 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(((PyObject *)__pyx_v_y)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":584
 *         _ensure_code(x)
 *         _ensure_code(y)
 *         _ensure_code(z)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(((PyObject *)__pyx_v_z)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":592
 *         # between a ray and a point is null, while ray and a
 *         # sphere is allowed)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             if radius == 0.0 :
 *                 for i in range(x.shape[0]) :
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "yt/geometry/selection_routines.pyx":593
 *         # sphere is allowed)
 *         with nogil:
 *             if radius == 0.0 :             # <<<<<<<<<<<<<<
 *                 for i in range(x.shape[0]) :
 *                     pos[0] = x[i]
 */
        __pyx_t_10 = ((__pyx_v_radius == 0.0) != 0);
        if (__pyx_t_10) {

          /* "yt/geometry/selection_routines.pyx":594
 *         with nogil:
 *             if radius == 0.0 :
 *                 for i in range(x.shape[0]) :             # <<<<<<<<<<<<<<
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]
 */
          __pyx_t_11 = (__pyx_v_x->dimensions[0]);
          __pyx_t_12 = __pyx_t_11;
          for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_12; __pyx_t_6+=1) {
            __pyx_v_i = __pyx_t_6;

            /* "yt/geometry/selection_routines.pyx":595
 *             if radius == 0.0 :
 *                 for i in range(x.shape[0]) :
 *                     pos[0] = x[i]             # <<<<<<<<<<<<<<
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]
 */
            __pyx_t_13 = __pyx_v_i;
            (__pyx_v_pos[0]) = (*__Pyx_BufPtrStrided1d(float *, __pyx_pybuffernd_x.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_x.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":596
 *                 for i in range(x.shape[0]) :
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]             # <<<<<<<<<<<<<<
 *                     pos[2] = z[i]
 *                     mask[i] = self.select_point(pos)
 */
            __pyx_t_13 = __pyx_v_i;
            (__pyx_v_pos[1]) = (*__Pyx_BufPtrStrided1d(float *, __pyx_pybuffernd_y.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_y.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":597
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]             # <<<<<<<<<<<<<<
 *                     mask[i] = self.select_point(pos)
 *                     count += mask[i]
 */
            __pyx_t_13 = __pyx_v_i;
            (__pyx_v_pos[2]) = (*__Pyx_BufPtrStrided1d(float *, __pyx_pybuffernd_z.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_z.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":598
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]
 *                     mask[i] = self.select_point(pos)             # <<<<<<<<<<<<<<
 *                     count += mask[i]
 *             else :
 */
            __pyx_t_13 = __pyx_v_i;
            *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_mask.diminfo[0].strides) = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_point(__pyx_v_self, __pyx_v_pos);

            /* "yt/geometry/selection_routines.pyx":599
 *                     pos[2] = z[i]
 *                     mask[i] = self.select_point(pos)
 *                     count += mask[i]             # <<<<<<<<<<<<<<
 *             else :
 *                 for i in range(x.shape[0]):
 */
            __pyx_t_14 = __pyx_v_i;
            __pyx_v_count = (__pyx_v_count + (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_mask.diminfo[0].strides)));
          }

          /* "yt/geometry/selection_routines.pyx":593
 *         # sphere is allowed)
 *         with nogil:
 *             if radius == 0.0 :             # <<<<<<<<<<<<<<
 *                 for i in range(x.shape[0]) :
 *                     pos[0] = x[i]
 */
          goto __pyx_L6;
        }

        /* "yt/geometry/selection_routines.pyx":601
 *                     count += mask[i]
 *             else :
 *                 for i in range(x.shape[0]):             # <<<<<<<<<<<<<<
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]
 */
        /*else*/ {
          __pyx_t_11 = (__pyx_v_x->dimensions[0]);
          __pyx_t_12 = __pyx_t_11;
          for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_12; __pyx_t_6+=1) {
            __pyx_v_i = __pyx_t_6;

            /* "yt/geometry/selection_routines.pyx":602
 *             else :
 *                 for i in range(x.shape[0]):
 *                     pos[0] = x[i]             # <<<<<<<<<<<<<<
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]
 */
            __pyx_t_14 = __pyx_v_i;
            (__pyx_v_pos[0]) = (*__Pyx_BufPtrStrided1d(float *, __pyx_pybuffernd_x.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_x.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":603
 *                 for i in range(x.shape[0]):
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]             # <<<<<<<<<<<<<<
 *                     pos[2] = z[i]
 *                     mask[i] = self.select_sphere(pos, radius)
 */
            __pyx_t_14 = __pyx_v_i;
            (__pyx_v_pos[1]) = (*__Pyx_BufPtrStrided1d(float *, __pyx_pybuffernd_y.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_y.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":604
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]             # <<<<<<<<<<<<<<
 *                     mask[i] = self.select_sphere(pos, radius)
 *                     count += mask[i]
 */
            __pyx_t_14 = __pyx_v_i;
            (__pyx_v_pos[2]) = (*__Pyx_BufPtrStrided1d(float *, __pyx_pybuffernd_z.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_z.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":605
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]
 *                     mask[i] = self.select_sphere(pos, radius)             # <<<<<<<<<<<<<<
 *                     count += mask[i]
 *         if count == 0: return None
 */
            __pyx_t_14 = __pyx_v_i;
            *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_mask.diminfo[0].strides) = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_sphere(__pyx_v_self, __pyx_v_pos, __pyx_v_radius);

            /* "yt/geometry/selection_routines.pyx":606
 *                     pos[2] = z[i]
 *                     mask[i] = self.select_sphere(pos, radius)
 *                     count += mask[i]             # <<<<<<<<<<<<<<
 *         if count == 0: return None
 *         return mask.view("bool")
 */
            __pyx_t_15 = __pyx_v_i;
            __pyx_v_count = (__pyx_v_count + (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_mask.diminfo[0].strides)));
          }
        }
        __pyx_L6:;
      }

      /* "yt/geometry/selection_routines.pyx":592
 *         # between a ray and a point is null, while ray and a
 *         # sphere is allowed)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             if radius == 0.0 :
 *                 for i in range(x.shape[0]) :
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "yt/geometry/selection_routines.pyx":607
 *                     mask[i] = self.select_sphere(pos, radius)
 *                     count += mask[i]
 *         if count == 0: return None             # <<<<<<<<<<<<<<
 *         return mask.view("bool")
 * 
 */
  __pyx_t_10 = ((__pyx_v_count == 0) != 0);
  if (__pyx_t_10) {
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":608
 *                     count += mask[i]
 *         if count == 0: return None
 *         return mask.view("bool")             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_mask), __pyx_n_s_view); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_n_s_bool) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_n_s_bool);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":573
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_points(self, np.ndarray[floating, ndim=1] x,             # <<<<<<<<<<<<<<
 *                             np.ndarray[floating, ndim=1] y,
 *                             np.ndarray[floating, ndim=1] z,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_x.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_y.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_z.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.select_points", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_x.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_y.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_z.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_1__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_33select_points(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_2yt_8geometry_18selection_routines_14SelectorObject_33select_points = {"__pyx_fuse_1select_points", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_1__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_33select_points, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_fuse_1__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_33select_points(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_x = 0;
  PyArrayObject *__pyx_v_y = 0;
  PyArrayObject *__pyx_v_z = 0;
  __pyx_t_5numpy_float64_t __pyx_v_radius;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("select_points (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,&__pyx_n_s_z,&__pyx_n_s_radius,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_points", 1, 4, 4, 1); __PYX_ERR(0, 573, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_z)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_points", 1, 4, 4, 2); __PYX_ERR(0, 573, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_radius)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_points", 1, 4, 4, 3); __PYX_ERR(0, 573, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "select_points") < 0)) __PYX_ERR(0, 573, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_x = ((PyArrayObject *)values[0]);
    __pyx_v_y = ((PyArrayObject *)values[1]);
    __pyx_v_z = ((PyArrayObject *)values[2]);
    __pyx_v_radius = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_radius == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 576, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("select_points", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 573, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.select_points", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_x), __pyx_ptype_5numpy_ndarray, 1, "x", 0))) __PYX_ERR(0, 573, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_y), __pyx_ptype_5numpy_ndarray, 1, "y", 0))) __PYX_ERR(0, 574, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_z), __pyx_ptype_5numpy_ndarray, 1, "z", 0))) __PYX_ERR(0, 575, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_32select_points(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_x, __pyx_v_y, __pyx_v_z, __pyx_v_radius);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_32select_points(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyArrayObject *__pyx_v_x, PyArrayObject *__pyx_v_y, PyArrayObject *__pyx_v_z, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_v_count;
  int __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_pos[3];
  PyArrayObject *__pyx_v_mask = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_x;
  __Pyx_Buffer __pyx_pybuffer_x;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_y;
  __Pyx_Buffer __pyx_pybuffer_y;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_z;
  __Pyx_Buffer __pyx_pybuffer_z;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyArrayObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  npy_intp __pyx_t_11;
  npy_intp __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_1select_points", 0);
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  __pyx_pybuffer_x.pybuffer.buf = NULL;
  __pyx_pybuffer_x.refcount = 0;
  __pyx_pybuffernd_x.data = NULL;
  __pyx_pybuffernd_x.rcbuffer = &__pyx_pybuffer_x;
  __pyx_pybuffer_y.pybuffer.buf = NULL;
  __pyx_pybuffer_y.refcount = 0;
  __pyx_pybuffernd_y.data = NULL;
  __pyx_pybuffernd_y.rcbuffer = &__pyx_pybuffer_y;
  __pyx_pybuffer_z.pybuffer.buf = NULL;
  __pyx_pybuffer_z.refcount = 0;
  __pyx_pybuffernd_z.data = NULL;
  __pyx_pybuffernd_z.rcbuffer = &__pyx_pybuffer_z;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_x.rcbuffer->pybuffer, (PyObject*)__pyx_v_x, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 573, __pyx_L1_error)
  }
  __pyx_pybuffernd_x.diminfo[0].strides = __pyx_pybuffernd_x.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_x.diminfo[0].shape = __pyx_pybuffernd_x.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_y.rcbuffer->pybuffer, (PyObject*)__pyx_v_y, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 573, __pyx_L1_error)
  }
  __pyx_pybuffernd_y.diminfo[0].strides = __pyx_pybuffernd_y.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_y.diminfo[0].shape = __pyx_pybuffernd_y.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_z.rcbuffer->pybuffer, (PyObject*)__pyx_v_z, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 573, __pyx_L1_error)
  }
  __pyx_pybuffernd_z.diminfo[0].strides = __pyx_pybuffernd_z.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_z.diminfo[0].shape = __pyx_pybuffernd_z.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/selection_routines.pyx":577
 *                             np.ndarray[floating, ndim=1] z,
 *                             np.float64_t radius):
 *         cdef int count = 0             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.float64_t pos[3]
 */
  __pyx_v_count = 0;

  /* "yt/geometry/selection_routines.pyx":581
 *         cdef np.float64_t pos[3]
 *         cdef np.ndarray[np.uint8_t, ndim=1] mask
 *         mask = np.empty(x.shape[0], dtype='uint8')             # <<<<<<<<<<<<<<
 *         _ensure_code(x)
 *         _ensure_code(y)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t((__pyx_v_x->dimensions[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 581, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 581, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
      __pyx_t_7 = __pyx_t_8 = __pyx_t_9 = 0;
    }
    __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 581, __pyx_L1_error)
  }
  __pyx_t_5 = 0;
  __pyx_v_mask = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":582
 *         cdef np.ndarray[np.uint8_t, ndim=1] mask
 *         mask = np.empty(x.shape[0], dtype='uint8')
 *         _ensure_code(x)             # <<<<<<<<<<<<<<
 *         _ensure_code(y)
 *         _ensure_code(z)
 */
  __pyx_t_4 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(((PyObject *)__pyx_v_x)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":583
 *         mask = np.empty(x.shape[0], dtype='uint8')
 *         _ensure_code(x)
 *         _ensure_code(y)             # <<<<<<<<<<<<<<
 *         _ensure_code(z)
 * 
 */
  __pyx_t_4 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(((PyObject *)__pyx_v_y)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":584
 *         _ensure_code(x)
 *         _ensure_code(y)
 *         _ensure_code(z)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(((PyObject *)__pyx_v_z)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":592
 *         # between a ray and a point is null, while ray and a
 *         # sphere is allowed)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             if radius == 0.0 :
 *                 for i in range(x.shape[0]) :
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "yt/geometry/selection_routines.pyx":593
 *         # sphere is allowed)
 *         with nogil:
 *             if radius == 0.0 :             # <<<<<<<<<<<<<<
 *                 for i in range(x.shape[0]) :
 *                     pos[0] = x[i]
 */
        __pyx_t_10 = ((__pyx_v_radius == 0.0) != 0);
        if (__pyx_t_10) {

          /* "yt/geometry/selection_routines.pyx":594
 *         with nogil:
 *             if radius == 0.0 :
 *                 for i in range(x.shape[0]) :             # <<<<<<<<<<<<<<
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]
 */
          __pyx_t_11 = (__pyx_v_x->dimensions[0]);
          __pyx_t_12 = __pyx_t_11;
          for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_12; __pyx_t_6+=1) {
            __pyx_v_i = __pyx_t_6;

            /* "yt/geometry/selection_routines.pyx":595
 *             if radius == 0.0 :
 *                 for i in range(x.shape[0]) :
 *                     pos[0] = x[i]             # <<<<<<<<<<<<<<
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]
 */
            __pyx_t_13 = __pyx_v_i;
            (__pyx_v_pos[0]) = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_x.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_x.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":596
 *                 for i in range(x.shape[0]) :
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]             # <<<<<<<<<<<<<<
 *                     pos[2] = z[i]
 *                     mask[i] = self.select_point(pos)
 */
            __pyx_t_13 = __pyx_v_i;
            (__pyx_v_pos[1]) = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_y.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_y.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":597
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]             # <<<<<<<<<<<<<<
 *                     mask[i] = self.select_point(pos)
 *                     count += mask[i]
 */
            __pyx_t_13 = __pyx_v_i;
            (__pyx_v_pos[2]) = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_z.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_z.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":598
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]
 *                     mask[i] = self.select_point(pos)             # <<<<<<<<<<<<<<
 *                     count += mask[i]
 *             else :
 */
            __pyx_t_13 = __pyx_v_i;
            *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_mask.diminfo[0].strides) = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_point(__pyx_v_self, __pyx_v_pos);

            /* "yt/geometry/selection_routines.pyx":599
 *                     pos[2] = z[i]
 *                     mask[i] = self.select_point(pos)
 *                     count += mask[i]             # <<<<<<<<<<<<<<
 *             else :
 *                 for i in range(x.shape[0]):
 */
            __pyx_t_14 = __pyx_v_i;
            __pyx_v_count = (__pyx_v_count + (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_mask.diminfo[0].strides)));
          }

          /* "yt/geometry/selection_routines.pyx":593
 *         # sphere is allowed)
 *         with nogil:
 *             if radius == 0.0 :             # <<<<<<<<<<<<<<
 *                 for i in range(x.shape[0]) :
 *                     pos[0] = x[i]
 */
          goto __pyx_L6;
        }

        /* "yt/geometry/selection_routines.pyx":601
 *                     count += mask[i]
 *             else :
 *                 for i in range(x.shape[0]):             # <<<<<<<<<<<<<<
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]
 */
        /*else*/ {
          __pyx_t_11 = (__pyx_v_x->dimensions[0]);
          __pyx_t_12 = __pyx_t_11;
          for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_12; __pyx_t_6+=1) {
            __pyx_v_i = __pyx_t_6;

            /* "yt/geometry/selection_routines.pyx":602
 *             else :
 *                 for i in range(x.shape[0]):
 *                     pos[0] = x[i]             # <<<<<<<<<<<<<<
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]
 */
            __pyx_t_14 = __pyx_v_i;
            (__pyx_v_pos[0]) = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_x.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_x.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":603
 *                 for i in range(x.shape[0]):
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]             # <<<<<<<<<<<<<<
 *                     pos[2] = z[i]
 *                     mask[i] = self.select_sphere(pos, radius)
 */
            __pyx_t_14 = __pyx_v_i;
            (__pyx_v_pos[1]) = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_y.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_y.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":604
 *                     pos[0] = x[i]
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]             # <<<<<<<<<<<<<<
 *                     mask[i] = self.select_sphere(pos, radius)
 *                     count += mask[i]
 */
            __pyx_t_14 = __pyx_v_i;
            (__pyx_v_pos[2]) = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_z.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_z.diminfo[0].strides));

            /* "yt/geometry/selection_routines.pyx":605
 *                     pos[1] = y[i]
 *                     pos[2] = z[i]
 *                     mask[i] = self.select_sphere(pos, radius)             # <<<<<<<<<<<<<<
 *                     count += mask[i]
 *         if count == 0: return None
 */
            __pyx_t_14 = __pyx_v_i;
            *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_mask.diminfo[0].strides) = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_vtab)->select_sphere(__pyx_v_self, __pyx_v_pos, __pyx_v_radius);

            /* "yt/geometry/selection_routines.pyx":606
 *                     pos[2] = z[i]
 *                     mask[i] = self.select_sphere(pos, radius)
 *                     count += mask[i]             # <<<<<<<<<<<<<<
 *         if count == 0: return None
 *         return mask.view("bool")
 */
            __pyx_t_15 = __pyx_v_i;
            __pyx_v_count = (__pyx_v_count + (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_mask.diminfo[0].strides)));
          }
        }
        __pyx_L6:;
      }

      /* "yt/geometry/selection_routines.pyx":592
 *         # between a ray and a point is null, while ray and a
 *         # sphere is allowed)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             if radius == 0.0 :
 *                 for i in range(x.shape[0]) :
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "yt/geometry/selection_routines.pyx":607
 *                     mask[i] = self.select_sphere(pos, radius)
 *                     count += mask[i]
 *         if count == 0: return None             # <<<<<<<<<<<<<<
 *         return mask.view("bool")
 * 
 */
  __pyx_t_10 = ((__pyx_v_count == 0) != 0);
  if (__pyx_t_10) {
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":608
 *                     count += mask[i]
 *         if count == 0: return None
 *         return mask.view("bool")             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_mask), __pyx_n_s_view); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_n_s_bool) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_n_s_bool);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":573
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_points(self, np.ndarray[floating, ndim=1] x,             # <<<<<<<<<<<<<<
 *                             np.ndarray[floating, ndim=1] y,
 *                             np.ndarray[floating, ndim=1] z,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_x.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_y.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_z.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.select_points", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_x.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_y.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_z.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":610
 *         return mask.view("bool")
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         # convert data to be hashed to a byte array, which FNV algorithm expects
 *         if self._hash_initialized == 1:
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_19__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_19__hash__(PyObject *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_18__hash__(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_18__hash__(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self) {
  PyObject *__pyx_v_hash_data = NULL;
  PyObject *__pyx_v_v = NULL;
  __pyx_t_5numpy_int64_t __pyx_v_hash_value;
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "yt/geometry/selection_routines.pyx":612
 *     def __hash__(self):
 *         # convert data to be hashed to a byte array, which FNV algorithm expects
 *         if self._hash_initialized == 1:             # <<<<<<<<<<<<<<
 *             return self._hash
 *         hash_data = bytearray()
 */
  __pyx_t_1 = ((__pyx_v_self->_hash_initialized == 1) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":613
 *         # convert data to be hashed to a byte array, which FNV algorithm expects
 *         if self._hash_initialized == 1:
 *             return self._hash             # <<<<<<<<<<<<<<
 *         hash_data = bytearray()
 *         for v in self._hash_vals() + self._base_hash():
 */
    __pyx_r = __pyx_v_self->_hash;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":612
 *     def __hash__(self):
 *         # convert data to be hashed to a byte array, which FNV algorithm expects
 *         if self._hash_initialized == 1:             # <<<<<<<<<<<<<<
 *             return self._hash
 *         hash_data = bytearray()
 */
  }

  /* "yt/geometry/selection_routines.pyx":614
 *         if self._hash_initialized == 1:
 *             return self._hash
 *         hash_data = bytearray()             # <<<<<<<<<<<<<<
 *         for v in self._hash_vals() + self._base_hash():
 *             if isinstance(v, tuple):
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyByteArray_Type))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_hash_data = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":615
 *             return self._hash
 *         hash_data = bytearray()
 *         for v in self._hash_vals() + self._base_hash():             # <<<<<<<<<<<<<<
 *             if isinstance(v, tuple):
 *                 hash_data.extend(v[0].encode('ascii'))
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_hash_vals); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base_hash); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyNumber_Add(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
    __pyx_t_3 = __pyx_t_4; __Pyx_INCREF(__pyx_t_3); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 615, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 615, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 615, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 615, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 615, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_7(__pyx_t_3);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 615, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "yt/geometry/selection_routines.pyx":616
 *         hash_data = bytearray()
 *         for v in self._hash_vals() + self._base_hash():
 *             if isinstance(v, tuple):             # <<<<<<<<<<<<<<
 *                 hash_data.extend(v[0].encode('ascii'))
 *                 hash_data.extend(repr(v[1]).encode('ascii'))
 */
    __pyx_t_1 = PyTuple_Check(__pyx_v_v); 
    __pyx_t_8 = (__pyx_t_1 != 0);
    if (__pyx_t_8) {

      /* "yt/geometry/selection_routines.pyx":617
 *         for v in self._hash_vals() + self._base_hash():
 *             if isinstance(v, tuple):
 *                 hash_data.extend(v[0].encode('ascii'))             # <<<<<<<<<<<<<<
 *                 hash_data.extend(repr(v[1]).encode('ascii'))
 *             else:
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_v, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 617, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 617, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_n_s_ascii) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_s_ascii);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 617, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyByteArray_Type_extend, __pyx_v_hash_data, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 617, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "yt/geometry/selection_routines.pyx":618
 *             if isinstance(v, tuple):
 *                 hash_data.extend(v[0].encode('ascii'))
 *                 hash_data.extend(repr(v[1]).encode('ascii'))             # <<<<<<<<<<<<<<
 *             else:
 *                 hash_data.extend(repr(v).encode('ascii'))
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_v, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 618, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = PyObject_Repr(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 618, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 618, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_n_s_ascii) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_s_ascii);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 618, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyByteArray_Type_extend, __pyx_v_hash_data, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 618, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "yt/geometry/selection_routines.pyx":616
 *         hash_data = bytearray()
 *         for v in self._hash_vals() + self._base_hash():
 *             if isinstance(v, tuple):             # <<<<<<<<<<<<<<
 *                 hash_data.extend(v[0].encode('ascii'))
 *                 hash_data.extend(repr(v[1]).encode('ascii'))
 */
      goto __pyx_L6;
    }

    /* "yt/geometry/selection_routines.pyx":620
 *                 hash_data.extend(repr(v[1]).encode('ascii'))
 *             else:
 *                 hash_data.extend(repr(v).encode('ascii'))             # <<<<<<<<<<<<<<
 *         cdef np.int64_t hash_value = fnv_hash(hash_data)
 *         self._hash = hash_value
 */
    /*else*/ {
      __pyx_t_5 = PyObject_Repr(__pyx_v_v); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 620, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 620, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_n_s_ascii) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_n_s_ascii);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 620, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyByteArray_Type_extend, __pyx_v_hash_data, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 620, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L6:;

    /* "yt/geometry/selection_routines.pyx":615
 *             return self._hash
 *         hash_data = bytearray()
 *         for v in self._hash_vals() + self._base_hash():             # <<<<<<<<<<<<<<
 *             if isinstance(v, tuple):
 *                 hash_data.extend(v[0].encode('ascii'))
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":621
 *             else:
 *                 hash_data.extend(repr(v).encode('ascii'))
 *         cdef np.int64_t hash_value = fnv_hash(hash_data)             # <<<<<<<<<<<<<<
 *         self._hash = hash_value
 *         self._hash_initialized = 1
 */
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_unsigned_char(__pyx_v_hash_data, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 621, __pyx_L1_error)
  __pyx_v_hash_value = __pyx_f_2yt_9utilities_3lib_8fnv_hash_c_fnv_hash(__pyx_t_9);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "yt/geometry/selection_routines.pyx":622
 *                 hash_data.extend(repr(v).encode('ascii'))
 *         cdef np.int64_t hash_value = fnv_hash(hash_data)
 *         self._hash = hash_value             # <<<<<<<<<<<<<<
 *         self._hash_initialized = 1
 *         return hash_value
 */
  __pyx_v_self->_hash = __pyx_v_hash_value;

  /* "yt/geometry/selection_routines.pyx":623
 *         cdef np.int64_t hash_value = fnv_hash(hash_data)
 *         self._hash = hash_value
 *         self._hash_initialized = 1             # <<<<<<<<<<<<<<
 *         return hash_value
 * 
 */
  __pyx_v_self->_hash_initialized = 1;

  /* "yt/geometry/selection_routines.pyx":624
 *         self._hash = hash_value
 *         self._hash_initialized = 1
 *         return hash_value             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_r = __pyx_v_hash_value;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":610
 *         return mask.view("bool")
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         # convert data to be hashed to a byte array, which FNV algorithm expects
 *         if self._hash_initialized == 1:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_hash_data);
  __Pyx_XDECREF(__pyx_v_v);
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":626
 *         return hash_value
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_21_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_21_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_20_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_20_hash_vals(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":627
 * 
 *     def _hash_vals(self):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     def _base_hash(self):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(0, 627, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":626
 *         return hash_value
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":629
 *         raise NotImplementedError
 * 
 *     def _base_hash(self):             # <<<<<<<<<<<<<<
 *         return (("min_level", self.min_level),
 *                 ("max_level", self.max_level),
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_23_base_hash(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_23_base_hash(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_base_hash (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_22_base_hash(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_22_base_hash(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_base_hash", 0);

  /* "yt/geometry/selection_routines.pyx":630
 * 
 *     def _base_hash(self):
 *         return (("min_level", self.min_level),             # <<<<<<<<<<<<<<
 *                 ("max_level", self.max_level),
 *                 ("overlap_cells", self.overlap_cells),
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->min_level); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_min_level);
  __Pyx_GIVEREF(__pyx_n_s_min_level);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_min_level);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":631
 *     def _base_hash(self):
 *         return (("min_level", self.min_level),
 *                 ("max_level", self.max_level),             # <<<<<<<<<<<<<<
 *                 ("overlap_cells", self.overlap_cells),
 *                 ("periodicity[0]", self.periodicity[0]),
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->max_level); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_max_level);
  __Pyx_GIVEREF(__pyx_n_s_max_level);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_max_level);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":632
 *         return (("min_level", self.min_level),
 *                 ("max_level", self.max_level),
 *                 ("overlap_cells", self.overlap_cells),             # <<<<<<<<<<<<<<
 *                 ("periodicity[0]", self.periodicity[0]),
 *                 ("periodicity[1]", self.periodicity[1]),
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->overlap_cells); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_overlap_cells);
  __Pyx_GIVEREF(__pyx_n_s_overlap_cells);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_overlap_cells);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":633
 *                 ("max_level", self.max_level),
 *                 ("overlap_cells", self.overlap_cells),
 *                 ("periodicity[0]", self.periodicity[0]),             # <<<<<<<<<<<<<<
 *                 ("periodicity[1]", self.periodicity[1]),
 *                 ("periodicity[2]", self.periodicity[2]),
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->periodicity[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_kp_s_periodicity_0);
  __Pyx_GIVEREF(__pyx_kp_s_periodicity_0);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_s_periodicity_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":634
 *                 ("overlap_cells", self.overlap_cells),
 *                 ("periodicity[0]", self.periodicity[0]),
 *                 ("periodicity[1]", self.periodicity[1]),             # <<<<<<<<<<<<<<
 *                 ("periodicity[2]", self.periodicity[2]),
 *                 ("domain_width[0]", self.domain_width[0]),
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->periodicity[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_kp_s_periodicity_1);
  __Pyx_GIVEREF(__pyx_kp_s_periodicity_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_periodicity_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":635
 *                 ("periodicity[0]", self.periodicity[0]),
 *                 ("periodicity[1]", self.periodicity[1]),
 *                 ("periodicity[2]", self.periodicity[2]),             # <<<<<<<<<<<<<<
 *                 ("domain_width[0]", self.domain_width[0]),
 *                 ("domain_width[1]", self.domain_width[1]),
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->periodicity[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_kp_s_periodicity_2);
  __Pyx_GIVEREF(__pyx_kp_s_periodicity_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_kp_s_periodicity_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":636
 *                 ("periodicity[1]", self.periodicity[1]),
 *                 ("periodicity[2]", self.periodicity[2]),
 *                 ("domain_width[0]", self.domain_width[0]),             # <<<<<<<<<<<<<<
 *                 ("domain_width[1]", self.domain_width[1]),
 *                 ("domain_width[2]", self.domain_width[2]))
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->domain_width[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_kp_s_domain_width_0);
  __Pyx_GIVEREF(__pyx_kp_s_domain_width_0);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_kp_s_domain_width_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":637
 *                 ("periodicity[2]", self.periodicity[2]),
 *                 ("domain_width[0]", self.domain_width[0]),
 *                 ("domain_width[1]", self.domain_width[1]),             # <<<<<<<<<<<<<<
 *                 ("domain_width[2]", self.domain_width[2]))
 * 
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->domain_width[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_kp_s_domain_width_1);
  __Pyx_GIVEREF(__pyx_kp_s_domain_width_1);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_kp_s_domain_width_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":638
 *                 ("domain_width[0]", self.domain_width[0]),
 *                 ("domain_width[1]", self.domain_width[1]),
 *                 ("domain_width[2]", self.domain_width[2]))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->domain_width[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_kp_s_domain_width_2);
  __Pyx_GIVEREF(__pyx_kp_s_domain_width_2);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_kp_s_domain_width_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":630
 * 
 *     def _base_hash(self):
 *         return (("min_level", self.min_level),             # <<<<<<<<<<<<<<
 *                 ("max_level", self.max_level),
 *                 ("overlap_cells", self.overlap_cells),
 */
  __pyx_t_1 = PyTuple_New(9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_t_10);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":629
 *         raise NotImplementedError
 * 
 *     def _base_hash(self):             # <<<<<<<<<<<<<<
 *         return (("min_level", self.min_level),
 *                 ("max_level", self.max_level),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject._base_hash", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pxd":30
 * 
 * cdef class SelectorObject:
 *     cdef public np.int32_t min_level             # <<<<<<<<<<<<<<
 *     cdef public np.int32_t max_level
 *     cdef int overlap_cells
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_9min_level_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_9min_level_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_9min_level___get__(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_9min_level___get__(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->min_level); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.min_level.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_9min_level_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_9min_level_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_9min_level_2__set__(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_9min_level_2__set__(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_int32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 30, __pyx_L1_error)
  __pyx_v_self->min_level = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.min_level.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pxd":31
 * cdef class SelectorObject:
 *     cdef public np.int32_t min_level
 *     cdef public np.int32_t max_level             # <<<<<<<<<<<<<<
 *     cdef int overlap_cells
 *     cdef np.float64_t domain_width[3]
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_9max_level_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_9max_level_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_9max_level___get__(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_9max_level___get__(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->max_level); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.max_level.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_9max_level_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_9max_level_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_9max_level_2__set__(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_9max_level_2__set__(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_int32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 31, __pyx_L1_error)
  __pyx_v_self->max_level = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.max_level.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_37__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_37__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_36__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_36__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_39__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_39__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_38__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SelectorObject_38__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SelectorObject.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":644
 *     cdef np.float64_t p[3]
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             self.p[i] = _ensure_code(dobj.p[i])
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_13PointSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_13PointSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 644, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 644, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.PointSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_13PointSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_13PointSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  long __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_5numpy_float64_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  long __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":645
 * 
 *     def __init__(self, dobj):
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             self.p[i] = _ensure_code(dobj.p[i])
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":646
 *     def __init__(self, dobj):
 *         for i in range(3):
 *             self.p[i] = _ensure_code(dobj.p[i])             # <<<<<<<<<<<<<<
 * 
 *             # ensure the point lies in the domain
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_p); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_i, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_self->p[__pyx_v_i]) = __pyx_t_4;

    /* "yt/geometry/selection_routines.pyx":649
 * 
 *             # ensure the point lies in the domain
 *             if self.periodicity[i]:             # <<<<<<<<<<<<<<
 *                 self.p[i] = np.fmod(self.p[i], self.domain_width[i])
 *                 if self.p[i] < 0.0:
 */
    __pyx_t_5 = ((__pyx_v_self->__pyx_base.periodicity[__pyx_v_i]) != 0);
    if (__pyx_t_5) {

      /* "yt/geometry/selection_routines.pyx":650
 *             # ensure the point lies in the domain
 *             if self.periodicity[i]:
 *                 self.p[i] = np.fmod(self.p[i], self.domain_width[i])             # <<<<<<<<<<<<<<
 *                 if self.p[i] < 0.0:
 *                     self.p[i] += self.domain_width[i]
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 650, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_fmod); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 650, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyFloat_FromDouble((__pyx_v_self->p[__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 650, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = PyFloat_FromDouble((__pyx_v_self->__pyx_base.domain_width[__pyx_v_i])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 650, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_3, __pyx_t_7};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 650, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_3, __pyx_t_7};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 650, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 650, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_8) {
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
        __pyx_t_3 = 0;
        __pyx_t_7 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 650, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 650, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      (__pyx_v_self->p[__pyx_v_i]) = __pyx_t_4;

      /* "yt/geometry/selection_routines.pyx":651
 *             if self.periodicity[i]:
 *                 self.p[i] = np.fmod(self.p[i], self.domain_width[i])
 *                 if self.p[i] < 0.0:             # <<<<<<<<<<<<<<
 *                     self.p[i] += self.domain_width[i]
 * 
 */
      __pyx_t_5 = (((__pyx_v_self->p[__pyx_v_i]) < 0.0) != 0);
      if (__pyx_t_5) {

        /* "yt/geometry/selection_routines.pyx":652
 *                 self.p[i] = np.fmod(self.p[i], self.domain_width[i])
 *                 if self.p[i] < 0.0:
 *                     self.p[i] += self.domain_width[i]             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
        __pyx_t_11 = __pyx_v_i;
        (__pyx_v_self->p[__pyx_t_11]) = ((__pyx_v_self->p[__pyx_t_11]) + (__pyx_v_self->__pyx_base.domain_width[__pyx_v_i]));

        /* "yt/geometry/selection_routines.pyx":651
 *             if self.periodicity[i]:
 *                 self.p[i] = np.fmod(self.p[i], self.domain_width[i])
 *                 if self.p[i] < 0.0:             # <<<<<<<<<<<<<<
 *                     self.p[i] += self.domain_width[i]
 * 
 */
      }

      /* "yt/geometry/selection_routines.pyx":649
 * 
 *             # ensure the point lies in the domain
 *             if self.periodicity[i]:             # <<<<<<<<<<<<<<
 *                 self.p[i] = np.fmod(self.p[i], self.domain_width[i])
 *                 if self.p[i] < 0.0:
 */
    }
  }

  /* "yt/geometry/selection_routines.pyx":644
 *     cdef np.float64_t p[3]
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             self.p[i] = _ensure_code(dobj.p[i])
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("yt.geometry.selection_routines.PointSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":657
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         if (pos[0] - 0.5*dds[0] <= self.p[0] < pos[0]+0.5*dds[0] and
 *             pos[1] - 0.5*dds[1] <= self.p[1] < pos[1]+0.5*dds[1] and
 */

static int __pyx_f_2yt_8geometry_18selection_routines_13PointSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":658
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         if (pos[0] - 0.5*dds[0] <= self.p[0] < pos[0]+0.5*dds[0] and             # <<<<<<<<<<<<<<
 *             pos[1] - 0.5*dds[1] <= self.p[1] < pos[1]+0.5*dds[1] and
 *             pos[2] - 0.5*dds[2] <= self.p[2] < pos[2]+0.5*dds[2]):
 */
  __pyx_t_2 = (((__pyx_v_pos[0]) - (0.5 * (__pyx_v_dds[0]))) <= (__pyx_v_self->p[0]));
  if (__pyx_t_2) {
    __pyx_t_2 = ((__pyx_v_self->p[0]) < ((__pyx_v_pos[0]) + (0.5 * (__pyx_v_dds[0]))));
  }
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":659
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         if (pos[0] - 0.5*dds[0] <= self.p[0] < pos[0]+0.5*dds[0] and
 *             pos[1] - 0.5*dds[1] <= self.p[1] < pos[1]+0.5*dds[1] and             # <<<<<<<<<<<<<<
 *             pos[2] - 0.5*dds[2] <= self.p[2] < pos[2]+0.5*dds[2]):
 *             return 1
 */
  __pyx_t_3 = (((__pyx_v_pos[1]) - (0.5 * (__pyx_v_dds[1]))) <= (__pyx_v_self->p[1]));
  if (__pyx_t_3) {
    __pyx_t_3 = ((__pyx_v_self->p[1]) < ((__pyx_v_pos[1]) + (0.5 * (__pyx_v_dds[1]))));
  }
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":660
 *         if (pos[0] - 0.5*dds[0] <= self.p[0] < pos[0]+0.5*dds[0] and
 *             pos[1] - 0.5*dds[1] <= self.p[1] < pos[1]+0.5*dds[1] and
 *             pos[2] - 0.5*dds[2] <= self.p[2] < pos[2]+0.5*dds[2]):             # <<<<<<<<<<<<<<
 *             return 1
 *         else:
 */
  __pyx_t_2 = (((__pyx_v_pos[2]) - (0.5 * (__pyx_v_dds[2]))) <= (__pyx_v_self->p[2]));
  if (__pyx_t_2) {
    __pyx_t_2 = ((__pyx_v_self->p[2]) < ((__pyx_v_pos[2]) + (0.5 * (__pyx_v_dds[2]))));
  }
  __pyx_t_3 = (__pyx_t_2 != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;

  /* "yt/geometry/selection_routines.pyx":658
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         if (pos[0] - 0.5*dds[0] <= self.p[0] < pos[0]+0.5*dds[0] and             # <<<<<<<<<<<<<<
 *             pos[1] - 0.5*dds[1] <= self.p[1] < pos[1]+0.5*dds[1] and
 *             pos[2] - 0.5*dds[2] <= self.p[2] < pos[2]+0.5*dds[2]):
 */
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":661
 *             pos[1] - 0.5*dds[1] <= self.p[1] < pos[1]+0.5*dds[1] and
 *             pos[2] - 0.5*dds[2] <= self.p[2] < pos[2]+0.5*dds[2]):
 *             return 1             # <<<<<<<<<<<<<<
 *         else:
 *             return 0
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":658
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         if (pos[0] - 0.5*dds[0] <= self.p[0] < pos[0]+0.5*dds[0] and             # <<<<<<<<<<<<<<
 *             pos[1] - 0.5*dds[1] <= self.p[1] < pos[1]+0.5*dds[1] and
 *             pos[2] - 0.5*dds[2] <= self.p[2] < pos[2]+0.5*dds[2]):
 */
  }

  /* "yt/geometry/selection_routines.pyx":663
 *             return 1
 *         else:
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":657
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         if (pos[0] - 0.5*dds[0] <= self.p[0] < pos[0]+0.5*dds[0] and
 *             pos[1] - 0.5*dds[1] <= self.p[1] < pos[1]+0.5*dds[1] and
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":668
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.float64_t dist, dist2 = 0
 */

static int __pyx_f_2yt_8geometry_18selection_routines_13PointSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_dist;
  __pyx_t_5numpy_float64_t __pyx_v_dist2;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":670
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef int i
 *         cdef np.float64_t dist, dist2 = 0             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             dist = self.difference(pos[i], self.p[i], i)
 */
  __pyx_v_dist2 = 0.0;

  /* "yt/geometry/selection_routines.pyx":671
 *         cdef int i
 *         cdef np.float64_t dist, dist2 = 0
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             dist = self.difference(pos[i], self.p[i], i)
 *             dist2 += dist*dist
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":672
 *         cdef np.float64_t dist, dist2 = 0
 *         for i in range(3):
 *             dist = self.difference(pos[i], self.p[i], i)             # <<<<<<<<<<<<<<
 *             dist2 += dist*dist
 *         if dist2 <= radius*radius: return 1
 */
    __pyx_v_dist = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_PointSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.difference(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), (__pyx_v_pos[__pyx_v_i]), (__pyx_v_self->p[__pyx_v_i]), __pyx_v_i);

    /* "yt/geometry/selection_routines.pyx":673
 *         for i in range(3):
 *             dist = self.difference(pos[i], self.p[i], i)
 *             dist2 += dist*dist             # <<<<<<<<<<<<<<
 *         if dist2 <= radius*radius: return 1
 *         return 0
 */
    __pyx_v_dist2 = (__pyx_v_dist2 + (__pyx_v_dist * __pyx_v_dist));
  }

  /* "yt/geometry/selection_routines.pyx":674
 *             dist = self.difference(pos[i], self.p[i], i)
 *             dist2 += dist*dist
 *         if dist2 <= radius*radius: return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_2 = ((__pyx_v_dist2 <= (__pyx_v_radius * __pyx_v_radius)) != 0);
  if (__pyx_t_2) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":675
 *             dist2 += dist*dist
 *         if dist2 <= radius*radius: return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":668
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.float64_t dist, dist2 = 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":680
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         # point definitely can only be in one cell
 */

static int __pyx_f_2yt_8geometry_18selection_routines_13PointSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":683
 *                                np.float64_t right_edge[3]) nogil:
 *         # point definitely can only be in one cell
 *         if (left_edge[0] <= self.p[0] < right_edge[0] and             # <<<<<<<<<<<<<<
 *             left_edge[1] <= self.p[1] < right_edge[1] and
 *             left_edge[2] <= self.p[2] < right_edge[2]):
 */
  __pyx_t_2 = ((__pyx_v_left_edge[0]) <= (__pyx_v_self->p[0]));
  if (__pyx_t_2) {
    __pyx_t_2 = ((__pyx_v_self->p[0]) < (__pyx_v_right_edge[0]));
  }
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":684
 *         # point definitely can only be in one cell
 *         if (left_edge[0] <= self.p[0] < right_edge[0] and
 *             left_edge[1] <= self.p[1] < right_edge[1] and             # <<<<<<<<<<<<<<
 *             left_edge[2] <= self.p[2] < right_edge[2]):
 *             return 1
 */
  __pyx_t_3 = ((__pyx_v_left_edge[1]) <= (__pyx_v_self->p[1]));
  if (__pyx_t_3) {
    __pyx_t_3 = ((__pyx_v_self->p[1]) < (__pyx_v_right_edge[1]));
  }
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":685
 *         if (left_edge[0] <= self.p[0] < right_edge[0] and
 *             left_edge[1] <= self.p[1] < right_edge[1] and
 *             left_edge[2] <= self.p[2] < right_edge[2]):             # <<<<<<<<<<<<<<
 *             return 1
 *         else:
 */
  __pyx_t_2 = ((__pyx_v_left_edge[2]) <= (__pyx_v_self->p[2]));
  if (__pyx_t_2) {
    __pyx_t_2 = ((__pyx_v_self->p[2]) < (__pyx_v_right_edge[2]));
  }
  __pyx_t_3 = (__pyx_t_2 != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;

  /* "yt/geometry/selection_routines.pyx":683
 *                                np.float64_t right_edge[3]) nogil:
 *         # point definitely can only be in one cell
 *         if (left_edge[0] <= self.p[0] < right_edge[0] and             # <<<<<<<<<<<<<<
 *             left_edge[1] <= self.p[1] < right_edge[1] and
 *             left_edge[2] <= self.p[2] < right_edge[2]):
 */
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":686
 *             left_edge[1] <= self.p[1] < right_edge[1] and
 *             left_edge[2] <= self.p[2] < right_edge[2]):
 *             return 1             # <<<<<<<<<<<<<<
 *         else:
 *             return 0
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":683
 *                                np.float64_t right_edge[3]) nogil:
 *         # point definitely can only be in one cell
 *         if (left_edge[0] <= self.p[0] < right_edge[0] and             # <<<<<<<<<<<<<<
 *             left_edge[1] <= self.p[1] < right_edge[1] and
 *             left_edge[2] <= self.p[2] < right_edge[2]):
 */
  }

  /* "yt/geometry/selection_routines.pyx":688
 *             return 1
 *         else:
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":680
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         # point definitely can only be in one cell
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":690
 *             return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("p[0]", self.p[0]),
 *                 ("p[1]", self.p[1]),
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_13PointSelector_3_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_13PointSelector_3_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_13PointSelector_2_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_13PointSelector_2_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":691
 * 
 *     def _hash_vals(self):
 *         return (("p[0]", self.p[0]),             # <<<<<<<<<<<<<<
 *                 ("p[1]", self.p[1]),
 *                 ("p[2]", self.p[2]))
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->p[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_p_0);
  __Pyx_GIVEREF(__pyx_kp_s_p_0);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_s_p_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":692
 *     def _hash_vals(self):
 *         return (("p[0]", self.p[0]),
 *                 ("p[1]", self.p[1]),             # <<<<<<<<<<<<<<
 *                 ("p[2]", self.p[2]))
 * 
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->p[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_kp_s_p_1);
  __Pyx_GIVEREF(__pyx_kp_s_p_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_s_p_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":693
 *         return (("p[0]", self.p[0]),
 *                 ("p[1]", self.p[1]),
 *                 ("p[2]", self.p[2]))             # <<<<<<<<<<<<<<
 * 
 * point_selector = PointSelector
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->p[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_kp_s_p_2);
  __Pyx_GIVEREF(__pyx_kp_s_p_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_s_p_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":691
 * 
 *     def _hash_vals(self):
 *         return (("p[0]", self.p[0]),             # <<<<<<<<<<<<<<
 *                 ("p[1]", self.p[1]),
 *                 ("p[2]", self.p[2]))
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":690
 *             return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("p[0]", self.p[0]),
 *                 ("p[1]", self.p[1]),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("yt.geometry.selection_routines.PointSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_13PointSelector_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_13PointSelector_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_13PointSelector_4__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_13PointSelector_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.PointSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_13PointSelector_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_13PointSelector_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_13PointSelector_6__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_13PointSelector_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.PointSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":705
 *     cdef bint check_box[3]
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             self.center[i] = _ensure_code(dobj.center[i])
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_14SphereSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_14SphereSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 705, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 705, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.SphereSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SphereSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_14SphereSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  long __pyx_v_i;
  PyObject *__pyx_v_center = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_5numpy_float64_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":706
 * 
 *     def __init__(self, dobj):
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             self.center[i] = _ensure_code(dobj.center[i])
 *         self.radius = _ensure_code(dobj.radius)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":707
 *     def __init__(self, dobj):
 *         for i in range(3):
 *             self.center[i] = _ensure_code(dobj.center[i])             # <<<<<<<<<<<<<<
 *         self.radius = _ensure_code(dobj.radius)
 *         self.radius2 = self.radius * self.radius
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_center); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_i, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_self->center[__pyx_v_i]) = __pyx_t_4;
  }

  /* "yt/geometry/selection_routines.pyx":708
 *         for i in range(3):
 *             self.center[i] = _ensure_code(dobj.center[i])
 *         self.radius = _ensure_code(dobj.radius)             # <<<<<<<<<<<<<<
 *         self.radius2 = self.radius * self.radius
 *         center = _ensure_code(dobj.center)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_radius); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->radius = __pyx_t_4;

  /* "yt/geometry/selection_routines.pyx":709
 *             self.center[i] = _ensure_code(dobj.center[i])
 *         self.radius = _ensure_code(dobj.radius)
 *         self.radius2 = self.radius * self.radius             # <<<<<<<<<<<<<<
 *         center = _ensure_code(dobj.center)
 *         for i in range(3):
 */
  __pyx_v_self->radius2 = (__pyx_v_self->radius * __pyx_v_self->radius);

  /* "yt/geometry/selection_routines.pyx":710
 *         self.radius = _ensure_code(dobj.radius)
 *         self.radius2 = self.radius * self.radius
 *         center = _ensure_code(dobj.center)             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             self.center[i] = center[i]
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_center); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_center = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":711
 *         self.radius2 = self.radius * self.radius
 *         center = _ensure_code(dobj.center)
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             self.center[i] = center[i]
 *             self.bbox[i][0] = self.center[i] - self.radius
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":712
 *         center = _ensure_code(dobj.center)
 *         for i in range(3):
 *             self.center[i] = center[i]             # <<<<<<<<<<<<<<
 *             self.bbox[i][0] = self.center[i] - self.radius
 *             self.bbox[i][1] = self.center[i] + self.radius
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_center, __pyx_v_i, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_self->center[__pyx_v_i]) = __pyx_t_4;

    /* "yt/geometry/selection_routines.pyx":713
 *         for i in range(3):
 *             self.center[i] = center[i]
 *             self.bbox[i][0] = self.center[i] - self.radius             # <<<<<<<<<<<<<<
 *             self.bbox[i][1] = self.center[i] + self.radius
 *             if self.bbox[i][0] < dobj.ds.domain_left_edge[i]:
 */
    ((__pyx_v_self->bbox[__pyx_v_i])[0]) = ((__pyx_v_self->center[__pyx_v_i]) - __pyx_v_self->radius);

    /* "yt/geometry/selection_routines.pyx":714
 *             self.center[i] = center[i]
 *             self.bbox[i][0] = self.center[i] - self.radius
 *             self.bbox[i][1] = self.center[i] + self.radius             # <<<<<<<<<<<<<<
 *             if self.bbox[i][0] < dobj.ds.domain_left_edge[i]:
 *                 self.check_box[i] = False
 */
    ((__pyx_v_self->bbox[__pyx_v_i])[1]) = ((__pyx_v_self->center[__pyx_v_i]) + __pyx_v_self->radius);

    /* "yt/geometry/selection_routines.pyx":715
 *             self.bbox[i][0] = self.center[i] - self.radius
 *             self.bbox[i][1] = self.center[i] + self.radius
 *             if self.bbox[i][0] < dobj.ds.domain_left_edge[i]:             # <<<<<<<<<<<<<<
 *                 self.check_box[i] = False
 *             elif self.bbox[i][1] > dobj.ds.domain_right_edge[i]:
 */
    __pyx_t_2 = PyFloat_FromDouble(((__pyx_v_self->bbox[__pyx_v_i])[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 715, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_ds); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 715, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_domain_left_edge); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 715, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, __pyx_v_i, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 715, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 715, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 715, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_6) {

      /* "yt/geometry/selection_routines.pyx":716
 *             self.bbox[i][1] = self.center[i] + self.radius
 *             if self.bbox[i][0] < dobj.ds.domain_left_edge[i]:
 *                 self.check_box[i] = False             # <<<<<<<<<<<<<<
 *             elif self.bbox[i][1] > dobj.ds.domain_right_edge[i]:
 *                 self.check_box[i] = False
 */
      (__pyx_v_self->check_box[__pyx_v_i]) = 0;

      /* "yt/geometry/selection_routines.pyx":715
 *             self.bbox[i][0] = self.center[i] - self.radius
 *             self.bbox[i][1] = self.center[i] + self.radius
 *             if self.bbox[i][0] < dobj.ds.domain_left_edge[i]:             # <<<<<<<<<<<<<<
 *                 self.check_box[i] = False
 *             elif self.bbox[i][1] > dobj.ds.domain_right_edge[i]:
 */
      goto __pyx_L7;
    }

    /* "yt/geometry/selection_routines.pyx":717
 *             if self.bbox[i][0] < dobj.ds.domain_left_edge[i]:
 *                 self.check_box[i] = False
 *             elif self.bbox[i][1] > dobj.ds.domain_right_edge[i]:             # <<<<<<<<<<<<<<
 *                 self.check_box[i] = False
 *             else:
 */
    __pyx_t_5 = PyFloat_FromDouble(((__pyx_v_self->bbox[__pyx_v_i])[1])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 717, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_ds); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 717, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_domain_right_edge); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 717, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_i, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 717, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_5, __pyx_t_3, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 717, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 717, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_6) {

      /* "yt/geometry/selection_routines.pyx":718
 *                 self.check_box[i] = False
 *             elif self.bbox[i][1] > dobj.ds.domain_right_edge[i]:
 *                 self.check_box[i] = False             # <<<<<<<<<<<<<<
 *             else:
 *                 self.check_box[i] = True
 */
      (__pyx_v_self->check_box[__pyx_v_i]) = 0;

      /* "yt/geometry/selection_routines.pyx":717
 *             if self.bbox[i][0] < dobj.ds.domain_left_edge[i]:
 *                 self.check_box[i] = False
 *             elif self.bbox[i][1] > dobj.ds.domain_right_edge[i]:             # <<<<<<<<<<<<<<
 *                 self.check_box[i] = False
 *             else:
 */
      goto __pyx_L7;
    }

    /* "yt/geometry/selection_routines.pyx":720
 *                 self.check_box[i] = False
 *             else:
 *                 self.check_box[i] = True             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    /*else*/ {
      (__pyx_v_self->check_box[__pyx_v_i]) = 1;
    }
    __pyx_L7:;
  }

  /* "yt/geometry/selection_routines.pyx":705
 *     cdef bint check_box[3]
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             self.center[i] = _ensure_code(dobj.center[i])
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SphereSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_center);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":725
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         # sphere center either inside cell or center of cell lies inside sphere
 *         if (pos[0] - 0.5*dds[0] <= self.center[0] <= pos[0]+0.5*dds[0] and
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14SphereSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":727
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         # sphere center either inside cell or center of cell lies inside sphere
 *         if (pos[0] - 0.5*dds[0] <= self.center[0] <= pos[0]+0.5*dds[0] and             # <<<<<<<<<<<<<<
 *             pos[1] - 0.5*dds[1] <= self.center[1] <= pos[1]+0.5*dds[1] and
 *             pos[2] - 0.5*dds[2] <= self.center[2] <= pos[2]+0.5*dds[2]):
 */
  __pyx_t_2 = (((__pyx_v_pos[0]) - (0.5 * (__pyx_v_dds[0]))) <= (__pyx_v_self->center[0]));
  if (__pyx_t_2) {
    __pyx_t_2 = ((__pyx_v_self->center[0]) <= ((__pyx_v_pos[0]) + (0.5 * (__pyx_v_dds[0]))));
  }
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":728
 *         # sphere center either inside cell or center of cell lies inside sphere
 *         if (pos[0] - 0.5*dds[0] <= self.center[0] <= pos[0]+0.5*dds[0] and
 *             pos[1] - 0.5*dds[1] <= self.center[1] <= pos[1]+0.5*dds[1] and             # <<<<<<<<<<<<<<
 *             pos[2] - 0.5*dds[2] <= self.center[2] <= pos[2]+0.5*dds[2]):
 *             return 1
 */
  __pyx_t_3 = (((__pyx_v_pos[1]) - (0.5 * (__pyx_v_dds[1]))) <= (__pyx_v_self->center[1]));
  if (__pyx_t_3) {
    __pyx_t_3 = ((__pyx_v_self->center[1]) <= ((__pyx_v_pos[1]) + (0.5 * (__pyx_v_dds[1]))));
  }
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":729
 *         if (pos[0] - 0.5*dds[0] <= self.center[0] <= pos[0]+0.5*dds[0] and
 *             pos[1] - 0.5*dds[1] <= self.center[1] <= pos[1]+0.5*dds[1] and
 *             pos[2] - 0.5*dds[2] <= self.center[2] <= pos[2]+0.5*dds[2]):             # <<<<<<<<<<<<<<
 *             return 1
 *         return self.select_point(pos)
 */
  __pyx_t_2 = (((__pyx_v_pos[2]) - (0.5 * (__pyx_v_dds[2]))) <= (__pyx_v_self->center[2]));
  if (__pyx_t_2) {
    __pyx_t_2 = ((__pyx_v_self->center[2]) <= ((__pyx_v_pos[2]) + (0.5 * (__pyx_v_dds[2]))));
  }
  __pyx_t_3 = (__pyx_t_2 != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;

  /* "yt/geometry/selection_routines.pyx":727
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         # sphere center either inside cell or center of cell lies inside sphere
 *         if (pos[0] - 0.5*dds[0] <= self.center[0] <= pos[0]+0.5*dds[0] and             # <<<<<<<<<<<<<<
 *             pos[1] - 0.5*dds[1] <= self.center[1] <= pos[1]+0.5*dds[1] and
 *             pos[2] - 0.5*dds[2] <= self.center[2] <= pos[2]+0.5*dds[2]):
 */
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":730
 *             pos[1] - 0.5*dds[1] <= self.center[1] <= pos[1]+0.5*dds[1] and
 *             pos[2] - 0.5*dds[2] <= self.center[2] <= pos[2]+0.5*dds[2]):
 *             return 1             # <<<<<<<<<<<<<<
 *         return self.select_point(pos)
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":727
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         # sphere center either inside cell or center of cell lies inside sphere
 *         if (pos[0] - 0.5*dds[0] <= self.center[0] <= pos[0]+0.5*dds[0] and             # <<<<<<<<<<<<<<
 *             pos[1] - 0.5*dds[1] <= self.center[1] <= pos[1]+0.5*dds[1] and
 *             pos[2] - 0.5*dds[2] <= self.center[2] <= pos[2]+0.5*dds[2]):
 */
  }

  /* "yt/geometry/selection_routines.pyx":731
 *             pos[2] - 0.5*dds[2] <= self.center[2] <= pos[2]+0.5*dds[2]):
 *             return 1
 *         return self.select_point(pos)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SphereSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.select_point(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_pos);
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":725
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         # sphere center either inside cell or center of cell lies inside sphere
 *         if (pos[0] - 0.5*dds[0] <= self.center[0] <= pos[0]+0.5*dds[0] and
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":736
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.float64_t dist, dist2 = 0
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14SphereSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_dist;
  __pyx_t_5numpy_float64_t __pyx_v_dist2;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":738
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         cdef int i
 *         cdef np.float64_t dist, dist2 = 0             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             if self.check_box[i] and \
 */
  __pyx_v_dist2 = 0.0;

  /* "yt/geometry/selection_routines.pyx":739
 *         cdef int i
 *         cdef np.float64_t dist, dist2 = 0
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             if self.check_box[i] and \
 *               (pos[i] < self.bbox[i][0] or
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":740
 *         cdef np.float64_t dist, dist2 = 0
 *         for i in range(3):
 *             if self.check_box[i] and \             # <<<<<<<<<<<<<<
 *               (pos[i] < self.bbox[i][0] or
 *                pos[i] > self.bbox[i][1]):
 */
    __pyx_t_3 = ((__pyx_v_self->check_box[__pyx_v_i]) != 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L6_bool_binop_done;
    }

    /* "yt/geometry/selection_routines.pyx":741
 *         for i in range(3):
 *             if self.check_box[i] and \
 *               (pos[i] < self.bbox[i][0] or             # <<<<<<<<<<<<<<
 *                pos[i] > self.bbox[i][1]):
 *                 return 0
 */
    __pyx_t_3 = (((__pyx_v_pos[__pyx_v_i]) < ((__pyx_v_self->bbox[__pyx_v_i])[0])) != 0);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L6_bool_binop_done;
    }

    /* "yt/geometry/selection_routines.pyx":742
 *             if self.check_box[i] and \
 *               (pos[i] < self.bbox[i][0] or
 *                pos[i] > self.bbox[i][1]):             # <<<<<<<<<<<<<<
 *                 return 0
 *             dist = _periodic_dist(pos[i], self.center[i], self.domain_width[i],
 */
    __pyx_t_3 = (((__pyx_v_pos[__pyx_v_i]) > ((__pyx_v_self->bbox[__pyx_v_i])[1])) != 0);
    __pyx_t_2 = __pyx_t_3;
    __pyx_L6_bool_binop_done:;

    /* "yt/geometry/selection_routines.pyx":740
 *         cdef np.float64_t dist, dist2 = 0
 *         for i in range(3):
 *             if self.check_box[i] and \             # <<<<<<<<<<<<<<
 *               (pos[i] < self.bbox[i][0] or
 *                pos[i] > self.bbox[i][1]):
 */
    if (__pyx_t_2) {

      /* "yt/geometry/selection_routines.pyx":743
 *               (pos[i] < self.bbox[i][0] or
 *                pos[i] > self.bbox[i][1]):
 *                 return 0             # <<<<<<<<<<<<<<
 *             dist = _periodic_dist(pos[i], self.center[i], self.domain_width[i],
 *                                   self.periodicity[i])
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "yt/geometry/selection_routines.pyx":740
 *         cdef np.float64_t dist, dist2 = 0
 *         for i in range(3):
 *             if self.check_box[i] and \             # <<<<<<<<<<<<<<
 *               (pos[i] < self.bbox[i][0] or
 *                pos[i] > self.bbox[i][1]):
 */
    }

    /* "yt/geometry/selection_routines.pyx":744
 *                pos[i] > self.bbox[i][1]):
 *                 return 0
 *             dist = _periodic_dist(pos[i], self.center[i], self.domain_width[i],             # <<<<<<<<<<<<<<
 *                                   self.periodicity[i])
 *             dist2 += dist*dist
 */
    __pyx_v_dist = __pyx_f_2yt_8geometry_18selection_routines__periodic_dist((__pyx_v_pos[__pyx_v_i]), (__pyx_v_self->center[__pyx_v_i]), (__pyx_v_self->__pyx_base.domain_width[__pyx_v_i]), (__pyx_v_self->__pyx_base.periodicity[__pyx_v_i]));

    /* "yt/geometry/selection_routines.pyx":746
 *             dist = _periodic_dist(pos[i], self.center[i], self.domain_width[i],
 *                                   self.periodicity[i])
 *             dist2 += dist*dist             # <<<<<<<<<<<<<<
 *             if dist2 > self.radius2: return 0
 *         return 1
 */
    __pyx_v_dist2 = (__pyx_v_dist2 + (__pyx_v_dist * __pyx_v_dist));

    /* "yt/geometry/selection_routines.pyx":747
 *                                   self.periodicity[i])
 *             dist2 += dist*dist
 *             if dist2 > self.radius2: return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
    __pyx_t_2 = ((__pyx_v_dist2 > __pyx_v_self->radius2) != 0);
    if (__pyx_t_2) {
      __pyx_r = 0;
      goto __pyx_L0;
    }
  }

  /* "yt/geometry/selection_routines.pyx":748
 *             dist2 += dist*dist
 *             if dist2 > self.radius2: return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":736
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.float64_t dist, dist2 = 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":753
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.float64_t dist, dist2 = 0
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14SphereSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_dist;
  __pyx_t_5numpy_float64_t __pyx_v_dist2;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":755
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef int i
 *         cdef np.float64_t dist, dist2 = 0             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             dist = self.difference(pos[i], self.center[i], i)
 */
  __pyx_v_dist2 = 0.0;

  /* "yt/geometry/selection_routines.pyx":756
 *         cdef int i
 *         cdef np.float64_t dist, dist2 = 0
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             dist = self.difference(pos[i], self.center[i], i)
 *             dist2 += dist*dist
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":757
 *         cdef np.float64_t dist, dist2 = 0
 *         for i in range(3):
 *             dist = self.difference(pos[i], self.center[i], i)             # <<<<<<<<<<<<<<
 *             dist2 += dist*dist
 *         dist = self.radius+radius
 */
    __pyx_v_dist = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SphereSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.difference(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), (__pyx_v_pos[__pyx_v_i]), (__pyx_v_self->center[__pyx_v_i]), __pyx_v_i);

    /* "yt/geometry/selection_routines.pyx":758
 *         for i in range(3):
 *             dist = self.difference(pos[i], self.center[i], i)
 *             dist2 += dist*dist             # <<<<<<<<<<<<<<
 *         dist = self.radius+radius
 *         if dist2 <= dist*dist: return 1
 */
    __pyx_v_dist2 = (__pyx_v_dist2 + (__pyx_v_dist * __pyx_v_dist));
  }

  /* "yt/geometry/selection_routines.pyx":759
 *             dist = self.difference(pos[i], self.center[i], i)
 *             dist2 += dist*dist
 *         dist = self.radius+radius             # <<<<<<<<<<<<<<
 *         if dist2 <= dist*dist: return 1
 *         return 0
 */
  __pyx_v_dist = (__pyx_v_self->radius + __pyx_v_radius);

  /* "yt/geometry/selection_routines.pyx":760
 *             dist2 += dist*dist
 *         dist = self.radius+radius
 *         if dist2 <= dist*dist: return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_2 = ((__pyx_v_dist2 <= (__pyx_v_dist * __pyx_v_dist)) != 0);
  if (__pyx_t_2) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":761
 *         dist = self.radius+radius
 *         if dist2 <= dist*dist: return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":753
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.float64_t dist, dist2 = 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":766
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef np.float64_t box_center, relcenter, closest, dist, edge
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14SphereSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  __pyx_t_5numpy_float64_t __pyx_v_box_center;
  __pyx_t_5numpy_float64_t __pyx_v_relcenter;
  __pyx_t_5numpy_float64_t __pyx_v_closest;
  __pyx_t_5numpy_float64_t __pyx_v_dist;
  __pyx_t_5numpy_float64_t __pyx_v_edge;
  int __pyx_v_i;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "yt/geometry/selection_routines.pyx":770
 *         cdef np.float64_t box_center, relcenter, closest, dist, edge
 *         cdef int i
 *         if (left_edge[0] <= self.center[0] <= right_edge[0] and             # <<<<<<<<<<<<<<
 *             left_edge[1] <= self.center[1] <= right_edge[1] and
 *             left_edge[2] <= self.center[2] <= right_edge[2]):
 */
  __pyx_t_2 = ((__pyx_v_left_edge[0]) <= (__pyx_v_self->center[0]));
  if (__pyx_t_2) {
    __pyx_t_2 = ((__pyx_v_self->center[0]) <= (__pyx_v_right_edge[0]));
  }
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":771
 *         cdef int i
 *         if (left_edge[0] <= self.center[0] <= right_edge[0] and
 *             left_edge[1] <= self.center[1] <= right_edge[1] and             # <<<<<<<<<<<<<<
 *             left_edge[2] <= self.center[2] <= right_edge[2]):
 *             return 1
 */
  __pyx_t_3 = ((__pyx_v_left_edge[1]) <= (__pyx_v_self->center[1]));
  if (__pyx_t_3) {
    __pyx_t_3 = ((__pyx_v_self->center[1]) <= (__pyx_v_right_edge[1]));
  }
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":772
 *         if (left_edge[0] <= self.center[0] <= right_edge[0] and
 *             left_edge[1] <= self.center[1] <= right_edge[1] and
 *             left_edge[2] <= self.center[2] <= right_edge[2]):             # <<<<<<<<<<<<<<
 *             return 1
 *         for i in range(3):
 */
  __pyx_t_2 = ((__pyx_v_left_edge[2]) <= (__pyx_v_self->center[2]));
  if (__pyx_t_2) {
    __pyx_t_2 = ((__pyx_v_self->center[2]) <= (__pyx_v_right_edge[2]));
  }
  __pyx_t_3 = (__pyx_t_2 != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;

  /* "yt/geometry/selection_routines.pyx":770
 *         cdef np.float64_t box_center, relcenter, closest, dist, edge
 *         cdef int i
 *         if (left_edge[0] <= self.center[0] <= right_edge[0] and             # <<<<<<<<<<<<<<
 *             left_edge[1] <= self.center[1] <= right_edge[1] and
 *             left_edge[2] <= self.center[2] <= right_edge[2]):
 */
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":773
 *             left_edge[1] <= self.center[1] <= right_edge[1] and
 *             left_edge[2] <= self.center[2] <= right_edge[2]):
 *             return 1             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             if not self.check_box[i]: continue
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":770
 *         cdef np.float64_t box_center, relcenter, closest, dist, edge
 *         cdef int i
 *         if (left_edge[0] <= self.center[0] <= right_edge[0] and             # <<<<<<<<<<<<<<
 *             left_edge[1] <= self.center[1] <= right_edge[1] and
 *             left_edge[2] <= self.center[2] <= right_edge[2]):
 */
  }

  /* "yt/geometry/selection_routines.pyx":774
 *             left_edge[2] <= self.center[2] <= right_edge[2]):
 *             return 1
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             if not self.check_box[i]: continue
 *             if right_edge[i] < self.bbox[i][0] or \
 */
  for (__pyx_t_4 = 0; __pyx_t_4 < 3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "yt/geometry/selection_routines.pyx":775
 *             return 1
 *         for i in range(3):
 *             if not self.check_box[i]: continue             # <<<<<<<<<<<<<<
 *             if right_edge[i] < self.bbox[i][0] or \
 *                left_edge[i] > self.bbox[i][1]:
 */
    __pyx_t_1 = ((!((__pyx_v_self->check_box[__pyx_v_i]) != 0)) != 0);
    if (__pyx_t_1) {
      goto __pyx_L7_continue;
    }

    /* "yt/geometry/selection_routines.pyx":776
 *         for i in range(3):
 *             if not self.check_box[i]: continue
 *             if right_edge[i] < self.bbox[i][0] or \             # <<<<<<<<<<<<<<
 *                left_edge[i] > self.bbox[i][1]:
 *                 return 0
 */
    __pyx_t_3 = (((__pyx_v_right_edge[__pyx_v_i]) < ((__pyx_v_self->bbox[__pyx_v_i])[0])) != 0);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L11_bool_binop_done;
    }

    /* "yt/geometry/selection_routines.pyx":777
 *             if not self.check_box[i]: continue
 *             if right_edge[i] < self.bbox[i][0] or \
 *                left_edge[i] > self.bbox[i][1]:             # <<<<<<<<<<<<<<
 *                 return 0
 *         # http://www.gamedev.net/topic/335465-is-this-the-simplest-sphere-aabb-collision-test/
 */
    __pyx_t_3 = (((__pyx_v_left_edge[__pyx_v_i]) > ((__pyx_v_self->bbox[__pyx_v_i])[1])) != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L11_bool_binop_done:;

    /* "yt/geometry/selection_routines.pyx":776
 *         for i in range(3):
 *             if not self.check_box[i]: continue
 *             if right_edge[i] < self.bbox[i][0] or \             # <<<<<<<<<<<<<<
 *                left_edge[i] > self.bbox[i][1]:
 *                 return 0
 */
    if (__pyx_t_1) {

      /* "yt/geometry/selection_routines.pyx":778
 *             if right_edge[i] < self.bbox[i][0] or \
 *                left_edge[i] > self.bbox[i][1]:
 *                 return 0             # <<<<<<<<<<<<<<
 *         # http://www.gamedev.net/topic/335465-is-this-the-simplest-sphere-aabb-collision-test/
 *         dist = 0
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "yt/geometry/selection_routines.pyx":776
 *         for i in range(3):
 *             if not self.check_box[i]: continue
 *             if right_edge[i] < self.bbox[i][0] or \             # <<<<<<<<<<<<<<
 *                left_edge[i] > self.bbox[i][1]:
 *                 return 0
 */
    }
    __pyx_L7_continue:;
  }

  /* "yt/geometry/selection_routines.pyx":780
 *                 return 0
 *         # http://www.gamedev.net/topic/335465-is-this-the-simplest-sphere-aabb-collision-test/
 *         dist = 0             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             # Early terminate
 */
  __pyx_v_dist = 0.0;

  /* "yt/geometry/selection_routines.pyx":781
 *         # http://www.gamedev.net/topic/335465-is-this-the-simplest-sphere-aabb-collision-test/
 *         dist = 0
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             # Early terminate
 *             box_center = (right_edge[i] + left_edge[i])/2.0
 */
  for (__pyx_t_4 = 0; __pyx_t_4 < 3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "yt/geometry/selection_routines.pyx":783
 *         for i in range(3):
 *             # Early terminate
 *             box_center = (right_edge[i] + left_edge[i])/2.0             # <<<<<<<<<<<<<<
 *             relcenter = self.difference(box_center, self.center[i], i)
 *             edge = right_edge[i] - left_edge[i]
 */
    __pyx_v_box_center = (((__pyx_v_right_edge[__pyx_v_i]) + (__pyx_v_left_edge[__pyx_v_i])) / 2.0);

    /* "yt/geometry/selection_routines.pyx":784
 *             # Early terminate
 *             box_center = (right_edge[i] + left_edge[i])/2.0
 *             relcenter = self.difference(box_center, self.center[i], i)             # <<<<<<<<<<<<<<
 *             edge = right_edge[i] - left_edge[i]
 *             closest = relcenter - fclip(relcenter, -edge/2.0, edge/2.0)
 */
    __pyx_v_relcenter = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SphereSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.difference(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_box_center, (__pyx_v_self->center[__pyx_v_i]), __pyx_v_i);

    /* "yt/geometry/selection_routines.pyx":785
 *             box_center = (right_edge[i] + left_edge[i])/2.0
 *             relcenter = self.difference(box_center, self.center[i], i)
 *             edge = right_edge[i] - left_edge[i]             # <<<<<<<<<<<<<<
 *             closest = relcenter - fclip(relcenter, -edge/2.0, edge/2.0)
 *             dist += closest*closest
 */
    __pyx_v_edge = ((__pyx_v_right_edge[__pyx_v_i]) - (__pyx_v_left_edge[__pyx_v_i]));

    /* "yt/geometry/selection_routines.pyx":786
 *             relcenter = self.difference(box_center, self.center[i], i)
 *             edge = right_edge[i] - left_edge[i]
 *             closest = relcenter - fclip(relcenter, -edge/2.0, edge/2.0)             # <<<<<<<<<<<<<<
 *             dist += closest*closest
 *             if dist > self.radius2: return 0
 */
    __pyx_v_closest = (__pyx_v_relcenter - __pyx_f_2yt_9utilities_3lib_8fp_utils_fclip(__pyx_v_relcenter, ((-__pyx_v_edge) / 2.0), (__pyx_v_edge / 2.0)));

    /* "yt/geometry/selection_routines.pyx":787
 *             edge = right_edge[i] - left_edge[i]
 *             closest = relcenter - fclip(relcenter, -edge/2.0, edge/2.0)
 *             dist += closest*closest             # <<<<<<<<<<<<<<
 *             if dist > self.radius2: return 0
 *         return 1
 */
    __pyx_v_dist = (__pyx_v_dist + (__pyx_v_closest * __pyx_v_closest));

    /* "yt/geometry/selection_routines.pyx":788
 *             closest = relcenter - fclip(relcenter, -edge/2.0, edge/2.0)
 *             dist += closest*closest
 *             if dist > self.radius2: return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
    __pyx_t_1 = ((__pyx_v_dist > __pyx_v_self->radius2) != 0);
    if (__pyx_t_1) {
      __pyx_r = 0;
      goto __pyx_L0;
    }
  }

  /* "yt/geometry/selection_routines.pyx":789
 *             dist += closest*closest
 *             if dist > self.radius2: return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":766
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef np.float64_t box_center, relcenter, closest, dist, edge
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":791
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("radius", self.radius),
 *                 ("radius2", self.radius2),
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SphereSelector_3_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SphereSelector_3_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SphereSelector_2_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SphereSelector_2_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":792
 * 
 *     def _hash_vals(self):
 *         return (("radius", self.radius),             # <<<<<<<<<<<<<<
 *                 ("radius2", self.radius2),
 *                 ("center[0]", self.center[0]),
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->radius); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_radius);
  __Pyx_GIVEREF(__pyx_n_s_radius);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_radius);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":793
 *     def _hash_vals(self):
 *         return (("radius", self.radius),
 *                 ("radius2", self.radius2),             # <<<<<<<<<<<<<<
 *                 ("center[0]", self.center[0]),
 *                 ("center[1]", self.center[1]),
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->radius2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_radius2);
  __Pyx_GIVEREF(__pyx_n_s_radius2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_radius2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":794
 *         return (("radius", self.radius),
 *                 ("radius2", self.radius2),
 *                 ("center[0]", self.center[0]),             # <<<<<<<<<<<<<<
 *                 ("center[1]", self.center[1]),
 *                 ("center[2]", self.center[2]))
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->center[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_kp_s_center_0);
  __Pyx_GIVEREF(__pyx_kp_s_center_0);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_s_center_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":795
 *                 ("radius2", self.radius2),
 *                 ("center[0]", self.center[0]),
 *                 ("center[1]", self.center[1]),             # <<<<<<<<<<<<<<
 *                 ("center[2]", self.center[2]))
 * 
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->center[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_kp_s_center_1);
  __Pyx_GIVEREF(__pyx_kp_s_center_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_s_center_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":796
 *                 ("center[0]", self.center[0]),
 *                 ("center[1]", self.center[1]),
 *                 ("center[2]", self.center[2]))             # <<<<<<<<<<<<<<
 * 
 * sphere_selector = SphereSelector
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->center[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_kp_s_center_2);
  __Pyx_GIVEREF(__pyx_kp_s_center_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_center_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":792
 * 
 *     def _hash_vals(self):
 *         return (("radius", self.radius),             # <<<<<<<<<<<<<<
 *                 ("radius2", self.radius2),
 *                 ("center[0]", self.center[0]),
 */
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":791
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("radius", self.radius),
 *                 ("radius2", self.radius2),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SphereSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SphereSelector_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SphereSelector_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SphereSelector_4__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SphereSelector_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SphereSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SphereSelector_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14SphereSelector_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14SphereSelector_6__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14SphereSelector_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SphereSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":809
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         # We are modifying dobj.left_edge and dobj.right_edge , so here we will
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_14RegionSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_14RegionSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 809, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 809, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.RegionSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14RegionSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_14RegionSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  int __pyx_v_i;
  __Pyx_memviewslice __pyx_v_RE = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_LE = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_DW = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_DLE = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_DRE = { 0, 0, { 0 }, { 0 }, { 0 } };
  __pyx_t_5numpy_float64_t __pyx_v_region_width[3];
  int __pyx_v_p[3];
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  __pyx_t_5numpy_float64_t __pyx_t_17;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":813
 *         # We are modifying dobj.left_edge and dobj.right_edge , so here we will
 *         # do an in-place conversion of those arrays.
 *         cdef np.float64_t[:] RE = _ensure_code(dobj.right_edge)             # <<<<<<<<<<<<<<
 *         cdef np.float64_t[:] LE = _ensure_code(dobj.left_edge)
 *         cdef np.float64_t[:] DW = _ensure_code(dobj.ds.domain_width)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_right_edge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_RE = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;

  /* "yt/geometry/selection_routines.pyx":814
 *         # do an in-place conversion of those arrays.
 *         cdef np.float64_t[:] RE = _ensure_code(dobj.right_edge)
 *         cdef np.float64_t[:] LE = _ensure_code(dobj.left_edge)             # <<<<<<<<<<<<<<
 *         cdef np.float64_t[:] DW = _ensure_code(dobj.ds.domain_width)
 *         cdef np.float64_t[:] DLE = _ensure_code(dobj.ds.domain_left_edge)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_left_edge); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_LE = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;

  /* "yt/geometry/selection_routines.pyx":815
 *         cdef np.float64_t[:] RE = _ensure_code(dobj.right_edge)
 *         cdef np.float64_t[:] LE = _ensure_code(dobj.left_edge)
 *         cdef np.float64_t[:] DW = _ensure_code(dobj.ds.domain_width)             # <<<<<<<<<<<<<<
 *         cdef np.float64_t[:] DLE = _ensure_code(dobj.ds.domain_left_edge)
 *         cdef np.float64_t[:] DRE = _ensure_code(dobj.ds.domain_right_edge)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_ds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_domain_width); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_DW = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;

  /* "yt/geometry/selection_routines.pyx":816
 *         cdef np.float64_t[:] LE = _ensure_code(dobj.left_edge)
 *         cdef np.float64_t[:] DW = _ensure_code(dobj.ds.domain_width)
 *         cdef np.float64_t[:] DLE = _ensure_code(dobj.ds.domain_left_edge)             # <<<<<<<<<<<<<<
 *         cdef np.float64_t[:] DRE = _ensure_code(dobj.ds.domain_right_edge)
 *         cdef np.float64_t region_width[3]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_ds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_domain_left_edge); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_DLE = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;

  /* "yt/geometry/selection_routines.pyx":817
 *         cdef np.float64_t[:] DW = _ensure_code(dobj.ds.domain_width)
 *         cdef np.float64_t[:] DLE = _ensure_code(dobj.ds.domain_left_edge)
 *         cdef np.float64_t[:] DRE = _ensure_code(dobj.ds.domain_right_edge)             # <<<<<<<<<<<<<<
 *         cdef np.float64_t region_width[3]
 *         cdef bint p[3]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_ds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_domain_right_edge); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_DRE = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;

  /* "yt/geometry/selection_routines.pyx":822
 *         # This is for if we want to include zones that overlap and whose
 *         # centers are not strictly included.
 *         self.loose_selection = getattr(dobj, "loose_selection", False)             # <<<<<<<<<<<<<<
 *         self.check_period = False
 * 
 */
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_dobj, __pyx_n_s_loose_selection, Py_False); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->loose_selection = __pyx_t_4;

  /* "yt/geometry/selection_routines.pyx":823
 *         # centers are not strictly included.
 *         self.loose_selection = getattr(dobj, "loose_selection", False)
 *         self.check_period = False             # <<<<<<<<<<<<<<
 * 
 *         for i in range(3):
 */
  __pyx_v_self->check_period = 0;

  /* "yt/geometry/selection_routines.pyx":825
 *         self.check_period = False
 * 
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             region_width[i] = RE[i] - LE[i]
 *             p[i] = dobj.ds.periodicity[i]
 */
  for (__pyx_t_5 = 0; __pyx_t_5 < 3; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "yt/geometry/selection_routines.pyx":826
 * 
 *         for i in range(3):
 *             region_width[i] = RE[i] - LE[i]             # <<<<<<<<<<<<<<
 *             p[i] = dobj.ds.periodicity[i]
 *             if region_width[i] <= 0:
 */
    __pyx_t_6 = __pyx_v_i;
    __pyx_t_7 = __pyx_v_i;
    (__pyx_v_region_width[__pyx_v_i]) = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_RE.data + __pyx_t_6 * __pyx_v_RE.strides[0]) ))) - (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_LE.data + __pyx_t_7 * __pyx_v_LE.strides[0]) ))));

    /* "yt/geometry/selection_routines.pyx":827
 *         for i in range(3):
 *             region_width[i] = RE[i] - LE[i]
 *             p[i] = dobj.ds.periodicity[i]             # <<<<<<<<<<<<<<
 *             if region_width[i] <= 0:
 *                 raise RuntimeError(
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_ds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_periodicity); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_p[__pyx_v_i]) = __pyx_t_4;

    /* "yt/geometry/selection_routines.pyx":828
 *             region_width[i] = RE[i] - LE[i]
 *             p[i] = dobj.ds.periodicity[i]
 *             if region_width[i] <= 0:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     "Region right edge[%s] < left edge: width = %s" % (
 */
    __pyx_t_4 = (((__pyx_v_region_width[__pyx_v_i]) <= 0.0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "yt/geometry/selection_routines.pyx":831
 *                 raise RuntimeError(
 *                     "Region right edge[%s] < left edge: width = %s" % (
 *                         i, region_width[i]))             # <<<<<<<<<<<<<<
 * 
 *         for i in range(3):
 */
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = PyFloat_FromDouble((__pyx_v_region_width[__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 831, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 831, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2);
      __pyx_t_1 = 0;
      __pyx_t_2 = 0;

      /* "yt/geometry/selection_routines.pyx":830
 *             if region_width[i] <= 0:
 *                 raise RuntimeError(
 *                     "Region right edge[%s] < left edge: width = %s" % (             # <<<<<<<<<<<<<<
 *                         i, region_width[i]))
 * 
 */
      __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_Region_right_edge_s_left_edge_wi, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 830, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "yt/geometry/selection_routines.pyx":829
 *             p[i] = dobj.ds.periodicity[i]
 *             if region_width[i] <= 0:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     "Region right edge[%s] < left edge: width = %s" % (
 *                         i, region_width[i]))
 */
      __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 829, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 829, __pyx_L1_error)

      /* "yt/geometry/selection_routines.pyx":828
 *             region_width[i] = RE[i] - LE[i]
 *             p[i] = dobj.ds.periodicity[i]
 *             if region_width[i] <= 0:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     "Region right edge[%s] < left edge: width = %s" % (
 */
    }
  }

  /* "yt/geometry/selection_routines.pyx":833
 *                         i, region_width[i]))
 * 
 *         for i in range(3):             # <<<<<<<<<<<<<<
 * 
 *             if p[i]:
 */
  for (__pyx_t_5 = 0; __pyx_t_5 < 3; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "yt/geometry/selection_routines.pyx":835
 *         for i in range(3):
 * 
 *             if p[i]:             # <<<<<<<<<<<<<<
 *                 # First, we check if any criteria requires a period check,
 *                 # without any adjustments.  This is for short-circuiting the
 */
    __pyx_t_4 = ((__pyx_v_p[__pyx_v_i]) != 0);
    if (__pyx_t_4) {

      /* "yt/geometry/selection_routines.pyx":839
 *                 # without any adjustments.  This is for short-circuiting the
 *                 # short-circuit of the loop down below in mask filling.
 *                 if LE[i] < DLE[i] or LE[i] > DRE[i] or RE[i] > DRE[i]:             # <<<<<<<<<<<<<<
 *                     self.check_period = True
 *                 # shift so left_edge guaranteed in domain
 */
      __pyx_t_7 = __pyx_v_i;
      __pyx_t_6 = __pyx_v_i;
      __pyx_t_9 = (((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_LE.data + __pyx_t_7 * __pyx_v_LE.strides[0]) ))) < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_DLE.data + __pyx_t_6 * __pyx_v_DLE.strides[0]) )))) != 0);
      if (!__pyx_t_9) {
      } else {
        __pyx_t_4 = __pyx_t_9;
        goto __pyx_L10_bool_binop_done;
      }
      __pyx_t_6 = __pyx_v_i;
      __pyx_t_7 = __pyx_v_i;
      __pyx_t_9 = (((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_LE.data + __pyx_t_6 * __pyx_v_LE.strides[0]) ))) > (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_DRE.data + __pyx_t_7 * __pyx_v_DRE.strides[0]) )))) != 0);
      if (!__pyx_t_9) {
      } else {
        __pyx_t_4 = __pyx_t_9;
        goto __pyx_L10_bool_binop_done;
      }
      __pyx_t_7 = __pyx_v_i;
      __pyx_t_6 = __pyx_v_i;
      __pyx_t_9 = (((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_RE.data + __pyx_t_7 * __pyx_v_RE.strides[0]) ))) > (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_DRE.data + __pyx_t_6 * __pyx_v_DRE.strides[0]) )))) != 0);
      __pyx_t_4 = __pyx_t_9;
      __pyx_L10_bool_binop_done:;
      if (__pyx_t_4) {

        /* "yt/geometry/selection_routines.pyx":840
 *                 # short-circuit of the loop down below in mask filling.
 *                 if LE[i] < DLE[i] or LE[i] > DRE[i] or RE[i] > DRE[i]:
 *                     self.check_period = True             # <<<<<<<<<<<<<<
 *                 # shift so left_edge guaranteed in domain
 *                 if LE[i] < DLE[i]:
 */
        __pyx_v_self->check_period = 1;

        /* "yt/geometry/selection_routines.pyx":839
 *                 # without any adjustments.  This is for short-circuiting the
 *                 # short-circuit of the loop down below in mask filling.
 *                 if LE[i] < DLE[i] or LE[i] > DRE[i] or RE[i] > DRE[i]:             # <<<<<<<<<<<<<<
 *                     self.check_period = True
 *                 # shift so left_edge guaranteed in domain
 */
      }

      /* "yt/geometry/selection_routines.pyx":842
 *                     self.check_period = True
 *                 # shift so left_edge guaranteed in domain
 *                 if LE[i] < DLE[i]:             # <<<<<<<<<<<<<<
 *                     LE[i] += DW[i]
 *                     RE[i] += DW[i]
 */
      __pyx_t_6 = __pyx_v_i;
      __pyx_t_7 = __pyx_v_i;
      __pyx_t_4 = (((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_LE.data + __pyx_t_6 * __pyx_v_LE.strides[0]) ))) < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_DLE.data + __pyx_t_7 * __pyx_v_DLE.strides[0]) )))) != 0);
      if (__pyx_t_4) {

        /* "yt/geometry/selection_routines.pyx":843
 *                 # shift so left_edge guaranteed in domain
 *                 if LE[i] < DLE[i]:
 *                     LE[i] += DW[i]             # <<<<<<<<<<<<<<
 *                     RE[i] += DW[i]
 *                 elif LE[i] > DRE[i]:
 */
        __pyx_t_7 = __pyx_v_i;
        __pyx_t_6 = __pyx_v_i;
        *((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_LE.data + __pyx_t_6 * __pyx_v_LE.strides[0]) )) += (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_DW.data + __pyx_t_7 * __pyx_v_DW.strides[0]) )));

        /* "yt/geometry/selection_routines.pyx":844
 *                 if LE[i] < DLE[i]:
 *                     LE[i] += DW[i]
 *                     RE[i] += DW[i]             # <<<<<<<<<<<<<<
 *                 elif LE[i] > DRE[i]:
 *                     LE[i] -= DW[i]
 */
        __pyx_t_7 = __pyx_v_i;
        __pyx_t_10 = __pyx_v_i;
        *((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_RE.data + __pyx_t_10 * __pyx_v_RE.strides[0]) )) += (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_DW.data + __pyx_t_7 * __pyx_v_DW.strides[0]) )));

        /* "yt/geometry/selection_routines.pyx":842
 *                     self.check_period = True
 *                 # shift so left_edge guaranteed in domain
 *                 if LE[i] < DLE[i]:             # <<<<<<<<<<<<<<
 *                     LE[i] += DW[i]
 *                     RE[i] += DW[i]
 */
        goto __pyx_L13;
      }

      /* "yt/geometry/selection_routines.pyx":845
 *                     LE[i] += DW[i]
 *                     RE[i] += DW[i]
 *                 elif LE[i] > DRE[i]:             # <<<<<<<<<<<<<<
 *                     LE[i] -= DW[i]
 *                     RE[i] -= DW[i]
 */
      __pyx_t_7 = __pyx_v_i;
      __pyx_t_11 = __pyx_v_i;
      __pyx_t_4 = (((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_LE.data + __pyx_t_7 * __pyx_v_LE.strides[0]) ))) > (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_DRE.data + __pyx_t_11 * __pyx_v_DRE.strides[0]) )))) != 0);
      if (__pyx_t_4) {

        /* "yt/geometry/selection_routines.pyx":846
 *                     RE[i] += DW[i]
 *                 elif LE[i] > DRE[i]:
 *                     LE[i] -= DW[i]             # <<<<<<<<<<<<<<
 *                     RE[i] -= DW[i]
 *             else:
 */
        __pyx_t_11 = __pyx_v_i;
        __pyx_t_7 = __pyx_v_i;
        *((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_LE.data + __pyx_t_7 * __pyx_v_LE.strides[0]) )) -= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_DW.data + __pyx_t_11 * __pyx_v_DW.strides[0]) )));

        /* "yt/geometry/selection_routines.pyx":847
 *                 elif LE[i] > DRE[i]:
 *                     LE[i] -= DW[i]
 *                     RE[i] -= DW[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 if LE[i] < DLE[i] or RE[i] > DRE[i]:
 */
        __pyx_t_11 = __pyx_v_i;
        __pyx_t_12 = __pyx_v_i;
        *((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_RE.data + __pyx_t_12 * __pyx_v_RE.strides[0]) )) -= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_DW.data + __pyx_t_11 * __pyx_v_DW.strides[0]) )));

        /* "yt/geometry/selection_routines.pyx":845
 *                     LE[i] += DW[i]
 *                     RE[i] += DW[i]
 *                 elif LE[i] > DRE[i]:             # <<<<<<<<<<<<<<
 *                     LE[i] -= DW[i]
 *                     RE[i] -= DW[i]
 */
      }
      __pyx_L13:;

      /* "yt/geometry/selection_routines.pyx":835
 *         for i in range(3):
 * 
 *             if p[i]:             # <<<<<<<<<<<<<<
 *                 # First, we check if any criteria requires a period check,
 *                 # without any adjustments.  This is for short-circuiting the
 */
      goto __pyx_L8;
    }

    /* "yt/geometry/selection_routines.pyx":849
 *                     RE[i] -= DW[i]
 *             else:
 *                 if LE[i] < DLE[i] or RE[i] > DRE[i]:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         "Error: yt attempted to read outside the boundaries of "
 */
    /*else*/ {
      __pyx_t_11 = __pyx_v_i;
      __pyx_t_13 = __pyx_v_i;
      __pyx_t_9 = (((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_LE.data + __pyx_t_11 * __pyx_v_LE.strides[0]) ))) < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_DLE.data + __pyx_t_13 * __pyx_v_DLE.strides[0]) )))) != 0);
      if (!__pyx_t_9) {
      } else {
        __pyx_t_4 = __pyx_t_9;
        goto __pyx_L15_bool_binop_done;
      }
      __pyx_t_13 = __pyx_v_i;
      __pyx_t_11 = __pyx_v_i;
      __pyx_t_9 = (((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_RE.data + __pyx_t_13 * __pyx_v_RE.strides[0]) ))) > (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_DRE.data + __pyx_t_11 * __pyx_v_DRE.strides[0]) )))) != 0);
      __pyx_t_4 = __pyx_t_9;
      __pyx_L15_bool_binop_done:;
      if (unlikely(__pyx_t_4)) {

        /* "yt/geometry/selection_routines.pyx":859
 *                         "load a smaller region that does not border the edge or "
 *                         "override the periodicity for this dataset." % \
 *                         (i, dobj.left_edge[i], dobj.right_edge[i],             # <<<<<<<<<<<<<<
 *                          dobj.ds.domain_left_edge[i], dobj.ds.domain_right_edge[i])
 *                     )
 */
        __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 859, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_left_edge); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 859, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 859, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_right_edge); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 859, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_14 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 859, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "yt/geometry/selection_routines.pyx":860
 *                         "override the periodicity for this dataset." % \
 *                         (i, dobj.left_edge[i], dobj.right_edge[i],
 *                          dobj.ds.domain_left_edge[i], dobj.ds.domain_right_edge[i])             # <<<<<<<<<<<<<<
 *                     )
 *             # Already ensured in code
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_ds); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 860, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_domain_left_edge); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 860, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_15, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 860, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_ds); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 860, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_domain_right_edge); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 860, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_15 = __Pyx_GetItemInt(__pyx_t_16, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 860, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

        /* "yt/geometry/selection_routines.pyx":859
 *                         "load a smaller region that does not border the edge or "
 *                         "override the periodicity for this dataset." % \
 *                         (i, dobj.left_edge[i], dobj.right_edge[i],             # <<<<<<<<<<<<<<
 *                          dobj.ds.domain_left_edge[i], dobj.ds.domain_right_edge[i])
 *                     )
 */
        __pyx_t_16 = PyTuple_New(5); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 859, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_14);
        PyTuple_SET_ITEM(__pyx_t_16, 2, __pyx_t_14);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_16, 3, __pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_15);
        PyTuple_SET_ITEM(__pyx_t_16, 4, __pyx_t_15);
        __pyx_t_8 = 0;
        __pyx_t_1 = 0;
        __pyx_t_14 = 0;
        __pyx_t_2 = 0;
        __pyx_t_15 = 0;

        /* "yt/geometry/selection_routines.pyx":858
 *                         "up to the domain boundary. Two possible solutions are to "
 *                         "load a smaller region that does not border the edge or "
 *                         "override the periodicity for this dataset." % \             # <<<<<<<<<<<<<<
 *                         (i, dobj.left_edge[i], dobj.right_edge[i],
 *                          dobj.ds.domain_left_edge[i], dobj.ds.domain_right_edge[i])
 */
        __pyx_t_15 = __Pyx_PyString_Format(__pyx_kp_s_Error_yt_attempted_to_read_outsi, __pyx_t_16); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 858, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

        /* "yt/geometry/selection_routines.pyx":850
 *             else:
 *                 if LE[i] < DLE[i] or RE[i] > DRE[i]:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         "Error: yt attempted to read outside the boundaries of "
 *                         "a non-periodic domain along dimension %s.\n"
 */
        __pyx_t_16 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 850, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_Raise(__pyx_t_16, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __PYX_ERR(0, 850, __pyx_L1_error)

        /* "yt/geometry/selection_routines.pyx":849
 *                     RE[i] -= DW[i]
 *             else:
 *                 if LE[i] < DLE[i] or RE[i] > DRE[i]:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         "Error: yt attempted to read outside the boundaries of "
 */
      }
    }
    __pyx_L8:;

    /* "yt/geometry/selection_routines.pyx":863
 *                     )
 *             # Already ensured in code
 *             self.left_edge[i] = LE[i]             # <<<<<<<<<<<<<<
 *             self.right_edge[i] = RE[i]
 *             self.right_edge_shift[i] = RE[i] - DW[i]
 */
    __pyx_t_11 = __pyx_v_i;
    (__pyx_v_self->left_edge[__pyx_v_i]) = (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_LE.data + __pyx_t_11 * __pyx_v_LE.strides[0]) )));

    /* "yt/geometry/selection_routines.pyx":864
 *             # Already ensured in code
 *             self.left_edge[i] = LE[i]
 *             self.right_edge[i] = RE[i]             # <<<<<<<<<<<<<<
 *             self.right_edge_shift[i] = RE[i] - DW[i]
 *             if not self.periodicity[i]:
 */
    __pyx_t_11 = __pyx_v_i;
    (__pyx_v_self->right_edge[__pyx_v_i]) = (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_RE.data + __pyx_t_11 * __pyx_v_RE.strides[0]) )));

    /* "yt/geometry/selection_routines.pyx":865
 *             self.left_edge[i] = LE[i]
 *             self.right_edge[i] = RE[i]
 *             self.right_edge_shift[i] = RE[i] - DW[i]             # <<<<<<<<<<<<<<
 *             if not self.periodicity[i]:
 *                 self.right_edge_shift[i] = -np.inf
 */
    __pyx_t_11 = __pyx_v_i;
    __pyx_t_13 = __pyx_v_i;
    (__pyx_v_self->right_edge_shift[__pyx_v_i]) = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_RE.data + __pyx_t_11 * __pyx_v_RE.strides[0]) ))) - (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_DW.data + __pyx_t_13 * __pyx_v_DW.strides[0]) ))));

    /* "yt/geometry/selection_routines.pyx":866
 *             self.right_edge[i] = RE[i]
 *             self.right_edge_shift[i] = RE[i] - DW[i]
 *             if not self.periodicity[i]:             # <<<<<<<<<<<<<<
 *                 self.right_edge_shift[i] = -np.inf
 * 
 */
    __pyx_t_4 = ((!((__pyx_v_self->__pyx_base.periodicity[__pyx_v_i]) != 0)) != 0);
    if (__pyx_t_4) {

      /* "yt/geometry/selection_routines.pyx":867
 *             self.right_edge_shift[i] = RE[i] - DW[i]
 *             if not self.periodicity[i]:
 *                 self.right_edge_shift[i] = -np.inf             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_np); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 867, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_inf); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 867, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_16 = PyNumber_Negative(__pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 867, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_16); if (unlikely((__pyx_t_17 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 867, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      (__pyx_v_self->right_edge_shift[__pyx_v_i]) = __pyx_t_17;

      /* "yt/geometry/selection_routines.pyx":866
 *             self.right_edge[i] = RE[i]
 *             self.right_edge_shift[i] = RE[i] - DW[i]
 *             if not self.periodicity[i]:             # <<<<<<<<<<<<<<
 *                 self.right_edge_shift[i] = -np.inf
 * 
 */
    }
  }

  /* "yt/geometry/selection_routines.pyx":809
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         # We are modifying dobj.left_edge and dobj.right_edge , so here we will
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("yt.geometry.selection_routines.RegionSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_RE, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_LE, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_DW, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_DLE, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_DRE, 1);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":872
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef int i
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14RegionSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_v_i;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":875
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef int i
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             if (right_edge[i] < self.left_edge[i] and \
 *                 left_edge[i] >= self.right_edge_shift[i]) or \
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":876
 *         cdef int i
 *         for i in range(3):
 *             if (right_edge[i] < self.left_edge[i] and \             # <<<<<<<<<<<<<<
 *                 left_edge[i] >= self.right_edge_shift[i]) or \
 *                 left_edge[i] >= self.right_edge[i]:
 */
    __pyx_t_3 = (((__pyx_v_right_edge[__pyx_v_i]) < (__pyx_v_self->left_edge[__pyx_v_i])) != 0);
    if (!__pyx_t_3) {
      goto __pyx_L7_next_or;
    } else {
    }

    /* "yt/geometry/selection_routines.pyx":877
 *         for i in range(3):
 *             if (right_edge[i] < self.left_edge[i] and \
 *                 left_edge[i] >= self.right_edge_shift[i]) or \             # <<<<<<<<<<<<<<
 *                 left_edge[i] >= self.right_edge[i]:
 *                 return 0
 */
    __pyx_t_3 = (((__pyx_v_left_edge[__pyx_v_i]) >= (__pyx_v_self->right_edge_shift[__pyx_v_i])) != 0);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_L7_next_or:;

    /* "yt/geometry/selection_routines.pyx":878
 *             if (right_edge[i] < self.left_edge[i] and \
 *                 left_edge[i] >= self.right_edge_shift[i]) or \
 *                 left_edge[i] >= self.right_edge[i]:             # <<<<<<<<<<<<<<
 *                 return 0
 *         return 1
 */
    __pyx_t_3 = (((__pyx_v_left_edge[__pyx_v_i]) >= (__pyx_v_self->right_edge[__pyx_v_i])) != 0);
    __pyx_t_2 = __pyx_t_3;
    __pyx_L6_bool_binop_done:;

    /* "yt/geometry/selection_routines.pyx":876
 *         cdef int i
 *         for i in range(3):
 *             if (right_edge[i] < self.left_edge[i] and \             # <<<<<<<<<<<<<<
 *                 left_edge[i] >= self.right_edge_shift[i]) or \
 *                 left_edge[i] >= self.right_edge[i]:
 */
    if (__pyx_t_2) {

      /* "yt/geometry/selection_routines.pyx":879
 *                 left_edge[i] >= self.right_edge_shift[i]) or \
 *                 left_edge[i] >= self.right_edge[i]:
 *                 return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "yt/geometry/selection_routines.pyx":876
 *         cdef int i
 *         for i in range(3):
 *             if (right_edge[i] < self.left_edge[i] and \             # <<<<<<<<<<<<<<
 *                 left_edge[i] >= self.right_edge_shift[i]) or \
 *                 left_edge[i] >= self.right_edge[i]:
 */
    }
  }

  /* "yt/geometry/selection_routines.pyx":880
 *                 left_edge[i] >= self.right_edge[i]:
 *                 return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":872
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef int i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":885
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef np.float64_t left_edge[3]
 *         cdef np.float64_t right_edge[3]
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14RegionSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  __pyx_t_5numpy_float64_t __pyx_v_left_edge[3];
  __pyx_t_5numpy_float64_t __pyx_v_right_edge[3];
  int __pyx_v_i;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":889
 *         cdef np.float64_t right_edge[3]
 *         cdef int i
 *         if self.loose_selection:             # <<<<<<<<<<<<<<
 *             for i in range(3):
 *                 left_edge[i] = pos[i] - dds[i]*0.5
 */
  __pyx_t_1 = (__pyx_v_self->loose_selection != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":890
 *         cdef int i
 *         if self.loose_selection:
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 left_edge[i] = pos[i] - dds[i]*0.5
 *                 right_edge[i] = pos[i] + dds[i]*0.5
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 3; __pyx_t_2+=1) {
      __pyx_v_i = __pyx_t_2;

      /* "yt/geometry/selection_routines.pyx":891
 *         if self.loose_selection:
 *             for i in range(3):
 *                 left_edge[i] = pos[i] - dds[i]*0.5             # <<<<<<<<<<<<<<
 *                 right_edge[i] = pos[i] + dds[i]*0.5
 *             return self.select_bbox(left_edge, right_edge)
 */
      (__pyx_v_left_edge[__pyx_v_i]) = ((__pyx_v_pos[__pyx_v_i]) - ((__pyx_v_dds[__pyx_v_i]) * 0.5));

      /* "yt/geometry/selection_routines.pyx":892
 *             for i in range(3):
 *                 left_edge[i] = pos[i] - dds[i]*0.5
 *                 right_edge[i] = pos[i] + dds[i]*0.5             # <<<<<<<<<<<<<<
 *             return self.select_bbox(left_edge, right_edge)
 *         return self.select_point(pos)
 */
      (__pyx_v_right_edge[__pyx_v_i]) = ((__pyx_v_pos[__pyx_v_i]) + ((__pyx_v_dds[__pyx_v_i]) * 0.5));
    }

    /* "yt/geometry/selection_routines.pyx":893
 *                 left_edge[i] = pos[i] - dds[i]*0.5
 *                 right_edge[i] = pos[i] + dds[i]*0.5
 *             return self.select_bbox(left_edge, right_edge)             # <<<<<<<<<<<<<<
 *         return self.select_point(pos)
 * 
 */
    __pyx_r = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_RegionSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.select_bbox(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_left_edge, __pyx_v_right_edge);
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":889
 *         cdef np.float64_t right_edge[3]
 *         cdef int i
 *         if self.loose_selection:             # <<<<<<<<<<<<<<
 *             for i in range(3):
 *                 left_edge[i] = pos[i] - dds[i]*0.5
 */
  }

  /* "yt/geometry/selection_routines.pyx":894
 *                 right_edge[i] = pos[i] + dds[i]*0.5
 *             return self.select_bbox(left_edge, right_edge)
 *         return self.select_point(pos)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_RegionSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.select_point(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_pos);
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":885
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef np.float64_t left_edge[3]
 *         cdef np.float64_t right_edge[3]
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":899
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(3):
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14RegionSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_v_i;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "yt/geometry/selection_routines.pyx":901
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         cdef int i
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             if (self.right_edge_shift[i] <= pos[i] < self.left_edge[i]) or \
 *                pos[i] >= self.right_edge[i]:
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":902
 *         cdef int i
 *         for i in range(3):
 *             if (self.right_edge_shift[i] <= pos[i] < self.left_edge[i]) or \             # <<<<<<<<<<<<<<
 *                pos[i] >= self.right_edge[i]:
 *                 return 0
 */
    __pyx_t_3 = ((__pyx_v_self->right_edge_shift[__pyx_v_i]) <= (__pyx_v_pos[__pyx_v_i]));
    if (__pyx_t_3) {
      __pyx_t_3 = ((__pyx_v_pos[__pyx_v_i]) < (__pyx_v_self->left_edge[__pyx_v_i]));
    }
    __pyx_t_4 = (__pyx_t_3 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }

    /* "yt/geometry/selection_routines.pyx":903
 *         for i in range(3):
 *             if (self.right_edge_shift[i] <= pos[i] < self.left_edge[i]) or \
 *                pos[i] >= self.right_edge[i]:             # <<<<<<<<<<<<<<
 *                 return 0
 *         return 1
 */
    __pyx_t_4 = (((__pyx_v_pos[__pyx_v_i]) >= (__pyx_v_self->right_edge[__pyx_v_i])) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L6_bool_binop_done:;

    /* "yt/geometry/selection_routines.pyx":902
 *         cdef int i
 *         for i in range(3):
 *             if (self.right_edge_shift[i] <= pos[i] < self.left_edge[i]) or \             # <<<<<<<<<<<<<<
 *                pos[i] >= self.right_edge[i]:
 *                 return 0
 */
    if (__pyx_t_2) {

      /* "yt/geometry/selection_routines.pyx":904
 *             if (self.right_edge_shift[i] <= pos[i] < self.left_edge[i]) or \
 *                pos[i] >= self.right_edge[i]:
 *                 return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "yt/geometry/selection_routines.pyx":902
 *         cdef int i
 *         for i in range(3):
 *             if (self.right_edge_shift[i] <= pos[i] < self.left_edge[i]) or \             # <<<<<<<<<<<<<<
 *                pos[i] >= self.right_edge[i]:
 *                 return 0
 */
    }
  }

  /* "yt/geometry/selection_routines.pyx":905
 *                pos[i] >= self.right_edge[i]:
 *                 return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":899
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(3):
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":910
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int fill_mask_selector(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                 np.float64_t right_edge[3],
 *                                 np.float64_t dds[3], int dim[3],
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14RegionSelector_fill_mask_selector(struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_float64_t *__pyx_v_dds, int *__pyx_v_dim, PyArrayObject *__pyx_v_child_mask, PyArrayObject *__pyx_v_mask, int __pyx_v_level) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_total;
  int __pyx_v_this_level;
  __pyx_t_5numpy_float64_t __pyx_v_pos[3];
  __pyx_t_5numpy_int64_t __pyx_v_si[3];
  __pyx_t_5numpy_int64_t __pyx_v_ei[3];
  __Pyx_LocalBuf_ND __pyx_pybuffernd_child_mask;
  __Pyx_Buffer __pyx_pybuffer_child_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __pyx_t_5numpy_int64_t __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_int64_t __pyx_t_6;
  __pyx_t_5numpy_int64_t __pyx_t_7;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  long __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fill_mask_selector", 0);
  __pyx_pybuffer_child_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_child_mask.refcount = 0;
  __pyx_pybuffernd_child_mask.data = NULL;
  __pyx_pybuffernd_child_mask.rcbuffer = &__pyx_pybuffer_child_mask;
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_child_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack) == -1)) __PYX_ERR(0, 910, __pyx_L1_error)
  }
  __pyx_pybuffernd_child_mask.diminfo[0].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_child_mask.diminfo[0].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_child_mask.diminfo[1].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_child_mask.diminfo[1].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_child_mask.diminfo[2].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_child_mask.diminfo[2].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[2];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) __PYX_ERR(0, 910, __pyx_L1_error)
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mask.diminfo[1].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mask.diminfo[1].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_mask.diminfo[2].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_mask.diminfo[2].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[2];

  /* "yt/geometry/selection_routines.pyx":917
 *                                 int level):
 *         cdef int i, j, k
 *         cdef int total = 0, this_level = 0             # <<<<<<<<<<<<<<
 *         cdef np.float64_t pos[3]
 *         if level < self.min_level or level > self.max_level:
 */
  __pyx_v_total = 0;
  __pyx_v_this_level = 0;

  /* "yt/geometry/selection_routines.pyx":919
 *         cdef int total = 0, this_level = 0
 *         cdef np.float64_t pos[3]
 *         if level < self.min_level or level > self.max_level:             # <<<<<<<<<<<<<<
 *             return 0
 *         if level == self.max_level:
 */
  __pyx_t_2 = ((__pyx_v_level < __pyx_v_self->__pyx_base.min_level) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_level > __pyx_v_self->__pyx_base.max_level) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":920
 *         cdef np.float64_t pos[3]
 *         if level < self.min_level or level > self.max_level:
 *             return 0             # <<<<<<<<<<<<<<
 *         if level == self.max_level:
 *             this_level = 1
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":919
 *         cdef int total = 0, this_level = 0
 *         cdef np.float64_t pos[3]
 *         if level < self.min_level or level > self.max_level:             # <<<<<<<<<<<<<<
 *             return 0
 *         if level == self.max_level:
 */
  }

  /* "yt/geometry/selection_routines.pyx":921
 *         if level < self.min_level or level > self.max_level:
 *             return 0
 *         if level == self.max_level:             # <<<<<<<<<<<<<<
 *             this_level = 1
 *         cdef np.int64_t si[3]
 */
  __pyx_t_1 = ((__pyx_v_level == __pyx_v_self->__pyx_base.max_level) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":922
 *             return 0
 *         if level == self.max_level:
 *             this_level = 1             # <<<<<<<<<<<<<<
 *         cdef np.int64_t si[3]
 *         cdef np.int64_t ei[3]
 */
    __pyx_v_this_level = 1;

    /* "yt/geometry/selection_routines.pyx":921
 *         if level < self.min_level or level > self.max_level:
 *             return 0
 *         if level == self.max_level:             # <<<<<<<<<<<<<<
 *             this_level = 1
 *         cdef np.int64_t si[3]
 */
  }

  /* "yt/geometry/selection_routines.pyx":925
 *         cdef np.int64_t si[3]
 *         cdef np.int64_t ei[3]
 *         if not self.check_period:             # <<<<<<<<<<<<<<
 *             for i in range(3):
 *                 si[i] = <np.int64_t> ((self.left_edge[i] - left_edge[i])/dds[i])
 */
  __pyx_t_1 = ((!(__pyx_v_self->check_period != 0)) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":926
 *         cdef np.int64_t ei[3]
 *         if not self.check_period:
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 si[i] = <np.int64_t> ((self.left_edge[i] - left_edge[i])/dds[i])
 *                 ei[i] = <np.int64_t> ((self.right_edge[i] - left_edge[i])/dds[i])
 */
    for (__pyx_t_3 = 0; __pyx_t_3 < 3; __pyx_t_3+=1) {
      __pyx_v_i = __pyx_t_3;

      /* "yt/geometry/selection_routines.pyx":927
 *         if not self.check_period:
 *             for i in range(3):
 *                 si[i] = <np.int64_t> ((self.left_edge[i] - left_edge[i])/dds[i])             # <<<<<<<<<<<<<<
 *                 ei[i] = <np.int64_t> ((self.right_edge[i] - left_edge[i])/dds[i])
 *                 si[i] = iclip(si[i] - 1, 0, dim[i])
 */
      (__pyx_v_si[__pyx_v_i]) = ((__pyx_t_5numpy_int64_t)(((__pyx_v_self->left_edge[__pyx_v_i]) - (__pyx_v_left_edge[__pyx_v_i])) / (__pyx_v_dds[__pyx_v_i])));

      /* "yt/geometry/selection_routines.pyx":928
 *             for i in range(3):
 *                 si[i] = <np.int64_t> ((self.left_edge[i] - left_edge[i])/dds[i])
 *                 ei[i] = <np.int64_t> ((self.right_edge[i] - left_edge[i])/dds[i])             # <<<<<<<<<<<<<<
 *                 si[i] = iclip(si[i] - 1, 0, dim[i])
 *                 ei[i] = iclip(ei[i] + 1, 0, dim[i])
 */
      (__pyx_v_ei[__pyx_v_i]) = ((__pyx_t_5numpy_int64_t)(((__pyx_v_self->right_edge[__pyx_v_i]) - (__pyx_v_left_edge[__pyx_v_i])) / (__pyx_v_dds[__pyx_v_i])));

      /* "yt/geometry/selection_routines.pyx":929
 *                 si[i] = <np.int64_t> ((self.left_edge[i] - left_edge[i])/dds[i])
 *                 ei[i] = <np.int64_t> ((self.right_edge[i] - left_edge[i])/dds[i])
 *                 si[i] = iclip(si[i] - 1, 0, dim[i])             # <<<<<<<<<<<<<<
 *                 ei[i] = iclip(ei[i] + 1, 0, dim[i])
 *         else:
 */
      (__pyx_v_si[__pyx_v_i]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_iclip(((__pyx_v_si[__pyx_v_i]) - 1), 0, (__pyx_v_dim[__pyx_v_i]));

      /* "yt/geometry/selection_routines.pyx":930
 *                 ei[i] = <np.int64_t> ((self.right_edge[i] - left_edge[i])/dds[i])
 *                 si[i] = iclip(si[i] - 1, 0, dim[i])
 *                 ei[i] = iclip(ei[i] + 1, 0, dim[i])             # <<<<<<<<<<<<<<
 *         else:
 *             for i in range(3):
 */
      (__pyx_v_ei[__pyx_v_i]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_iclip(((__pyx_v_ei[__pyx_v_i]) + 1), 0, (__pyx_v_dim[__pyx_v_i]));
    }

    /* "yt/geometry/selection_routines.pyx":925
 *         cdef np.int64_t si[3]
 *         cdef np.int64_t ei[3]
 *         if not self.check_period:             # <<<<<<<<<<<<<<
 *             for i in range(3):
 *                 si[i] = <np.int64_t> ((self.left_edge[i] - left_edge[i])/dds[i])
 */
    goto __pyx_L7;
  }

  /* "yt/geometry/selection_routines.pyx":932
 *                 ei[i] = iclip(ei[i] + 1, 0, dim[i])
 *         else:
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 si[i] = 0
 *                 ei[i] = dim[i]
 */
  /*else*/ {
    for (__pyx_t_3 = 0; __pyx_t_3 < 3; __pyx_t_3+=1) {
      __pyx_v_i = __pyx_t_3;

      /* "yt/geometry/selection_routines.pyx":933
 *         else:
 *             for i in range(3):
 *                 si[i] = 0             # <<<<<<<<<<<<<<
 *                 ei[i] = dim[i]
 *         with nogil:
 */
      (__pyx_v_si[__pyx_v_i]) = 0;

      /* "yt/geometry/selection_routines.pyx":934
 *             for i in range(3):
 *                 si[i] = 0
 *                 ei[i] = dim[i]             # <<<<<<<<<<<<<<
 *         with nogil:
 *             pos[0] = left_edge[0] + (si[0] + 0.5) * dds[0]
 */
      (__pyx_v_ei[__pyx_v_i]) = (__pyx_v_dim[__pyx_v_i]);
    }
  }
  __pyx_L7:;

  /* "yt/geometry/selection_routines.pyx":935
 *                 si[i] = 0
 *                 ei[i] = dim[i]
 *         with nogil:             # <<<<<<<<<<<<<<
 *             pos[0] = left_edge[0] + (si[0] + 0.5) * dds[0]
 *             for i in range(si[0], ei[0]):
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "yt/geometry/selection_routines.pyx":936
 *                 ei[i] = dim[i]
 *         with nogil:
 *             pos[0] = left_edge[0] + (si[0] + 0.5) * dds[0]             # <<<<<<<<<<<<<<
 *             for i in range(si[0], ei[0]):
 *                 pos[1] = left_edge[1] + (si[1] + 0.5) * dds[1]
 */
        (__pyx_v_pos[0]) = ((__pyx_v_left_edge[0]) + (((__pyx_v_si[0]) + 0.5) * (__pyx_v_dds[0])));

        /* "yt/geometry/selection_routines.pyx":937
 *         with nogil:
 *             pos[0] = left_edge[0] + (si[0] + 0.5) * dds[0]
 *             for i in range(si[0], ei[0]):             # <<<<<<<<<<<<<<
 *                 pos[1] = left_edge[1] + (si[1] + 0.5) * dds[1]
 *                 for j in range(si[1], ei[1]):
 */
        __pyx_t_4 = (__pyx_v_ei[0]);
        __pyx_t_5 = __pyx_t_4;
        for (__pyx_t_3 = (__pyx_v_si[0]); __pyx_t_3 < __pyx_t_5; __pyx_t_3+=1) {
          __pyx_v_i = __pyx_t_3;

          /* "yt/geometry/selection_routines.pyx":938
 *             pos[0] = left_edge[0] + (si[0] + 0.5) * dds[0]
 *             for i in range(si[0], ei[0]):
 *                 pos[1] = left_edge[1] + (si[1] + 0.5) * dds[1]             # <<<<<<<<<<<<<<
 *                 for j in range(si[1], ei[1]):
 *                     pos[2] = left_edge[2] + (si[2] + 0.5) * dds[2]
 */
          (__pyx_v_pos[1]) = ((__pyx_v_left_edge[1]) + (((__pyx_v_si[1]) + 0.5) * (__pyx_v_dds[1])));

          /* "yt/geometry/selection_routines.pyx":939
 *             for i in range(si[0], ei[0]):
 *                 pos[1] = left_edge[1] + (si[1] + 0.5) * dds[1]
 *                 for j in range(si[1], ei[1]):             # <<<<<<<<<<<<<<
 *                     pos[2] = left_edge[2] + (si[2] + 0.5) * dds[2]
 *                     for k in range(si[2], ei[2]):
 */
          __pyx_t_6 = (__pyx_v_ei[1]);
          __pyx_t_7 = __pyx_t_6;
          for (__pyx_t_8 = (__pyx_v_si[1]); __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
            __pyx_v_j = __pyx_t_8;

            /* "yt/geometry/selection_routines.pyx":940
 *                 pos[1] = left_edge[1] + (si[1] + 0.5) * dds[1]
 *                 for j in range(si[1], ei[1]):
 *                     pos[2] = left_edge[2] + (si[2] + 0.5) * dds[2]             # <<<<<<<<<<<<<<
 *                     for k in range(si[2], ei[2]):
 *                         if child_mask[i, j, k] == 1 or this_level == 1:
 */
            (__pyx_v_pos[2]) = ((__pyx_v_left_edge[2]) + (((__pyx_v_si[2]) + 0.5) * (__pyx_v_dds[2])));

            /* "yt/geometry/selection_routines.pyx":941
 *                 for j in range(si[1], ei[1]):
 *                     pos[2] = left_edge[2] + (si[2] + 0.5) * dds[2]
 *                     for k in range(si[2], ei[2]):             # <<<<<<<<<<<<<<
 *                         if child_mask[i, j, k] == 1 or this_level == 1:
 *                             mask[i, j, k] = self.select_cell(pos, dds)
 */
            __pyx_t_9 = (__pyx_v_ei[2]);
            __pyx_t_10 = __pyx_t_9;
            for (__pyx_t_11 = (__pyx_v_si[2]); __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
              __pyx_v_k = __pyx_t_11;

              /* "yt/geometry/selection_routines.pyx":942
 *                     pos[2] = left_edge[2] + (si[2] + 0.5) * dds[2]
 *                     for k in range(si[2], ei[2]):
 *                         if child_mask[i, j, k] == 1 or this_level == 1:             # <<<<<<<<<<<<<<
 *                             mask[i, j, k] = self.select_cell(pos, dds)
 *                             total += mask[i, j, k]
 */
              __pyx_t_12 = __pyx_v_i;
              __pyx_t_13 = __pyx_v_j;
              __pyx_t_14 = __pyx_v_k;
              __pyx_t_2 = (((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_child_mask.diminfo[0].strides, __pyx_t_13, __pyx_pybuffernd_child_mask.diminfo[1].strides, __pyx_t_14, __pyx_pybuffernd_child_mask.diminfo[2].strides)) == 1) != 0);
              if (!__pyx_t_2) {
              } else {
                __pyx_t_1 = __pyx_t_2;
                goto __pyx_L22_bool_binop_done;
              }
              __pyx_t_2 = ((__pyx_v_this_level == 1) != 0);
              __pyx_t_1 = __pyx_t_2;
              __pyx_L22_bool_binop_done:;
              if (__pyx_t_1) {

                /* "yt/geometry/selection_routines.pyx":943
 *                     for k in range(si[2], ei[2]):
 *                         if child_mask[i, j, k] == 1 or this_level == 1:
 *                             mask[i, j, k] = self.select_cell(pos, dds)             # <<<<<<<<<<<<<<
 *                             total += mask[i, j, k]
 *                         pos[2] += dds[2]
 */
                __pyx_t_14 = __pyx_v_i;
                __pyx_t_13 = __pyx_v_j;
                __pyx_t_12 = __pyx_v_k;
                *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_mask.diminfo[0].strides, __pyx_t_13, __pyx_pybuffernd_mask.diminfo[1].strides, __pyx_t_12, __pyx_pybuffernd_mask.diminfo[2].strides) = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_RegionSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.select_cell(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_pos, __pyx_v_dds);

                /* "yt/geometry/selection_routines.pyx":944
 *                         if child_mask[i, j, k] == 1 or this_level == 1:
 *                             mask[i, j, k] = self.select_cell(pos, dds)
 *                             total += mask[i, j, k]             # <<<<<<<<<<<<<<
 *                         pos[2] += dds[2]
 *                     pos[1] += dds[1]
 */
                __pyx_t_15 = __pyx_v_i;
                __pyx_t_16 = __pyx_v_j;
                __pyx_t_17 = __pyx_v_k;
                __pyx_v_total = (__pyx_v_total + (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_15, __pyx_pybuffernd_mask.diminfo[0].strides, __pyx_t_16, __pyx_pybuffernd_mask.diminfo[1].strides, __pyx_t_17, __pyx_pybuffernd_mask.diminfo[2].strides)));

                /* "yt/geometry/selection_routines.pyx":942
 *                     pos[2] = left_edge[2] + (si[2] + 0.5) * dds[2]
 *                     for k in range(si[2], ei[2]):
 *                         if child_mask[i, j, k] == 1 or this_level == 1:             # <<<<<<<<<<<<<<
 *                             mask[i, j, k] = self.select_cell(pos, dds)
 *                             total += mask[i, j, k]
 */
              }

              /* "yt/geometry/selection_routines.pyx":945
 *                             mask[i, j, k] = self.select_cell(pos, dds)
 *                             total += mask[i, j, k]
 *                         pos[2] += dds[2]             # <<<<<<<<<<<<<<
 *                     pos[1] += dds[1]
 *                 pos[0] += dds[0]
 */
              __pyx_t_18 = 2;
              (__pyx_v_pos[__pyx_t_18]) = ((__pyx_v_pos[__pyx_t_18]) + (__pyx_v_dds[2]));
            }

            /* "yt/geometry/selection_routines.pyx":946
 *                             total += mask[i, j, k]
 *                         pos[2] += dds[2]
 *                     pos[1] += dds[1]             # <<<<<<<<<<<<<<
 *                 pos[0] += dds[0]
 *         return total
 */
            __pyx_t_18 = 1;
            (__pyx_v_pos[__pyx_t_18]) = ((__pyx_v_pos[__pyx_t_18]) + (__pyx_v_dds[1]));
          }

          /* "yt/geometry/selection_routines.pyx":947
 *                         pos[2] += dds[2]
 *                     pos[1] += dds[1]
 *                 pos[0] += dds[0]             # <<<<<<<<<<<<<<
 *         return total
 * 
 */
          __pyx_t_18 = 0;
          (__pyx_v_pos[__pyx_t_18]) = ((__pyx_v_pos[__pyx_t_18]) + (__pyx_v_dds[0]));
        }
      }

      /* "yt/geometry/selection_routines.pyx":935
 *                 si[i] = 0
 *                 ei[i] = dim[i]
 *         with nogil:             # <<<<<<<<<<<<<<
 *             pos[0] = left_edge[0] + (si[0] + 0.5) * dds[0]
 *             for i in range(si[0], ei[0]):
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L14;
        }
        __pyx_L14:;
      }
  }

  /* "yt/geometry/selection_routines.pyx":948
 *                     pos[1] += dds[1]
 *                 pos[0] += dds[0]
 *         return total             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_total;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":910
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int fill_mask_selector(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                 np.float64_t right_edge[3],
 *                                 np.float64_t dds[3], int dim[3],
 */

  /* function exit code */
  __pyx_L1_error:;
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_WriteUnraisable("yt.geometry.selection_routines.RegionSelector.fill_mask_selector", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":951
 * 
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("left_edge[0]", self.left_edge[0]),
 *                 ("left_edge[1]", self.left_edge[1]),
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14RegionSelector_3_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14RegionSelector_3_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14RegionSelector_2_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14RegionSelector_2_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":952
 * 
 *     def _hash_vals(self):
 *         return (("left_edge[0]", self.left_edge[0]),             # <<<<<<<<<<<<<<
 *                 ("left_edge[1]", self.left_edge[1]),
 *                 ("left_edge[2]", self.left_edge[2]),
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->left_edge[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_left_edge_0);
  __Pyx_GIVEREF(__pyx_kp_s_left_edge_0);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_s_left_edge_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":953
 *     def _hash_vals(self):
 *         return (("left_edge[0]", self.left_edge[0]),
 *                 ("left_edge[1]", self.left_edge[1]),             # <<<<<<<<<<<<<<
 *                 ("left_edge[2]", self.left_edge[2]),
 *                 ("right_edge[0]", self.right_edge[0]),
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->left_edge[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_kp_s_left_edge_1);
  __Pyx_GIVEREF(__pyx_kp_s_left_edge_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_s_left_edge_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":954
 *         return (("left_edge[0]", self.left_edge[0]),
 *                 ("left_edge[1]", self.left_edge[1]),
 *                 ("left_edge[2]", self.left_edge[2]),             # <<<<<<<<<<<<<<
 *                 ("right_edge[0]", self.right_edge[0]),
 *                 ("right_edge[1]", self.right_edge[1]),
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->left_edge[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_kp_s_left_edge_2);
  __Pyx_GIVEREF(__pyx_kp_s_left_edge_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_s_left_edge_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":955
 *                 ("left_edge[1]", self.left_edge[1]),
 *                 ("left_edge[2]", self.left_edge[2]),
 *                 ("right_edge[0]", self.right_edge[0]),             # <<<<<<<<<<<<<<
 *                 ("right_edge[1]", self.right_edge[1]),
 *                 ("right_edge[2]", self.right_edge[2]))
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->right_edge[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_kp_s_right_edge_0);
  __Pyx_GIVEREF(__pyx_kp_s_right_edge_0);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_s_right_edge_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":956
 *                 ("left_edge[2]", self.left_edge[2]),
 *                 ("right_edge[0]", self.right_edge[0]),
 *                 ("right_edge[1]", self.right_edge[1]),             # <<<<<<<<<<<<<<
 *                 ("right_edge[2]", self.right_edge[2]))
 * 
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->right_edge[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 956, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 956, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_kp_s_right_edge_1);
  __Pyx_GIVEREF(__pyx_kp_s_right_edge_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_right_edge_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":957
 *                 ("right_edge[0]", self.right_edge[0]),
 *                 ("right_edge[1]", self.right_edge[1]),
 *                 ("right_edge[2]", self.right_edge[2]))             # <<<<<<<<<<<<<<
 * 
 * region_selector = RegionSelector
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->right_edge[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_kp_s_right_edge_2);
  __Pyx_GIVEREF(__pyx_kp_s_right_edge_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_kp_s_right_edge_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":952
 * 
 *     def _hash_vals(self):
 *         return (("left_edge[0]", self.left_edge[0]),             # <<<<<<<<<<<<<<
 *                 ("left_edge[1]", self.left_edge[1]),
 *                 ("left_edge[2]", self.left_edge[2]),
 */
  __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_7);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":951
 * 
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("left_edge[0]", self.left_edge[0]),
 *                 ("left_edge[1]", self.left_edge[1]),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("yt.geometry.selection_routines.RegionSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14RegionSelector_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14RegionSelector_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14RegionSelector_4__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14RegionSelector_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.RegionSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14RegionSelector_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14RegionSelector_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14RegionSelector_6__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14RegionSelector_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.RegionSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":965
 *     cdef tuple _conditionals
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         positions = np.array([dobj['x'], dobj['y'], dobj['z']]).T
 *         self._conditionals = tuple(dobj.conditionals)
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_17CutRegionSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_17CutRegionSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 965, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 965, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.CutRegionSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_17CutRegionSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_2yt_8geometry_18selection_routines_17CutRegionSelector_8__init___2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "yt/geometry/selection_routines.pyx":968
 *         positions = np.array([dobj['x'], dobj['y'], dobj['z']]).T
 *         self._conditionals = tuple(dobj.conditionals)
 *         self._positions = set(tuple(position) for position in positions)             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_bbox(self,  np.float64_t left_edge[3],
 */

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17CutRegionSelector_8__init___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr *)__pyx_tp_new_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr(__pyx_ptype_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 968, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_2yt_8geometry_18selection_routines_17CutRegionSelector_8__init___2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_init___locals_genexpr, __pyx_n_s_yt_geometry_selection_routines); if (unlikely(!gen)) __PYX_ERR(0, 968, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.CutRegionSelector.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_2yt_8geometry_18selection_routines_17CutRegionSelector_8__init___2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 968, __pyx_L1_error)
  __pyx_r = PySet_New(NULL); if (unlikely(!__pyx_r)) __PYX_ERR(0, 968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_positions)) { __Pyx_RaiseClosureNameError("positions"); __PYX_ERR(0, 968, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_positions)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_positions)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_positions; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_positions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 968, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 968, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 968, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 968, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 968, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 968, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_position);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_position, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_cur_scope->__pyx_v_position); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(PySet_Add(__pyx_r, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 968, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":965
 *     cdef tuple _conditionals
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         positions = np.array([dobj['x'], dobj['y'], dobj['z']]).T
 *         self._conditionals = tuple(dobj.conditionals)
 */

static int __pyx_pf_2yt_8geometry_18selection_routines_17CutRegionSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ *__pyx_cur_scope;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ *)__pyx_tp_new_2yt_8geometry_18selection_routines___pyx_scope_struct____init__(__pyx_ptype_2yt_8geometry_18selection_routines___pyx_scope_struct____init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 965, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "yt/geometry/selection_routines.pyx":966
 * 
 *     def __init__(self, dobj):
 *         positions = np.array([dobj['x'], dobj['y'], dobj['z']]).T             # <<<<<<<<<<<<<<
 *         self._conditionals = tuple(dobj.conditionals)
 *         self._positions = set(tuple(position) for position in positions)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_dobj, __pyx_n_s_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_dobj, __pyx_n_s_y); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_v_dobj, __pyx_n_s_z); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyList_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyList_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyList_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_T); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_positions = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":967
 *     def __init__(self, dobj):
 *         positions = np.array([dobj['x'], dobj['y'], dobj['z']]).T
 *         self._conditionals = tuple(dobj.conditionals)             # <<<<<<<<<<<<<<
 *         self._positions = set(tuple(position) for position in positions)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_conditionals); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_conditionals);
  __Pyx_DECREF(__pyx_v_self->_conditionals);
  __pyx_v_self->_conditionals = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":968
 *         positions = np.array([dobj['x'], dobj['y'], dobj['z']]).T
 *         self._conditionals = tuple(dobj.conditionals)
 *         self._positions = set(tuple(position) for position in positions)             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_bbox(self,  np.float64_t left_edge[3],
 */
  __pyx_t_1 = __pyx_pf_2yt_8geometry_18selection_routines_17CutRegionSelector_8__init___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_positions);
  __Pyx_DECREF(__pyx_v_self->_positions);
  __pyx_v_self->_positions = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":965
 *     cdef tuple _conditionals
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         positions = np.array([dobj['x'], dobj['y'], dobj['z']]).T
 *         self._conditionals = tuple(dobj.conditionals)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("yt.geometry.selection_routines.CutRegionSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":970
 *         self._positions = set(tuple(position) for position in positions)
 * 
 *     cdef int select_bbox(self,  np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                      np.float64_t right_edge[3]) nogil:
 *         return 1
 */

static int __pyx_f_2yt_8geometry_18selection_routines_17CutRegionSelector_select_bbox(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":972
 *     cdef int select_bbox(self,  np.float64_t left_edge[3],
 *                      np.float64_t right_edge[3]) nogil:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":970
 *         self._positions = set(tuple(position) for position in positions)
 * 
 *     cdef int select_bbox(self,  np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                      np.float64_t right_edge[3]) nogil:
 *         return 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":974
 *         return 1
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         with gil:
 *             if (pos[0], pos[1], pos[2]) in self._positions:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_17CutRegionSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("select_cell", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "yt/geometry/selection_routines.pyx":975
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             if (pos[0], pos[1], pos[2]) in self._positions:
 *                 return 1
 */
  /*try:*/ {
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "yt/geometry/selection_routines.pyx":976
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         with gil:
 *             if (pos[0], pos[1], pos[2]) in self._positions:             # <<<<<<<<<<<<<<
 *                 return 1
 *             else:
 */
          __pyx_t_1 = PyFloat_FromDouble((__pyx_v_pos[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 976, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = PyFloat_FromDouble((__pyx_v_pos[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 976, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = PyFloat_FromDouble((__pyx_v_pos[2])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 976, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 976, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
          __pyx_t_1 = 0;
          __pyx_t_2 = 0;
          __pyx_t_3 = 0;
          if (unlikely(__pyx_v_self->_positions == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
            __PYX_ERR(0, 976, __pyx_L7_error)
          }
          __pyx_t_5 = (__Pyx_PySet_ContainsTF(__pyx_t_4, __pyx_v_self->_positions, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 976, __pyx_L7_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_6 = (__pyx_t_5 != 0);
          if (__pyx_t_6) {

            /* "yt/geometry/selection_routines.pyx":977
 *         with gil:
 *             if (pos[0], pos[1], pos[2]) in self._positions:
 *                 return 1             # <<<<<<<<<<<<<<
 *             else:
 *                 return 0
 */
            __pyx_r = 1;
            goto __pyx_L6_return;

            /* "yt/geometry/selection_routines.pyx":976
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         with gil:
 *             if (pos[0], pos[1], pos[2]) in self._positions:             # <<<<<<<<<<<<<<
 *                 return 1
 *             else:
 */
          }

          /* "yt/geometry/selection_routines.pyx":979
 *                 return 1
 *             else:
 *                 return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 */
          /*else*/ {
            __pyx_r = 0;
            goto __pyx_L6_return;
          }
        }

        /* "yt/geometry/selection_routines.pyx":975
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             if (pos[0], pos[1], pos[2]) in self._positions:
 *                 return 1
 */
        /*finally:*/ {
          __pyx_L6_return: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L3_return;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
        }
    }
  }
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "yt/geometry/selection_routines.pyx":974
 *         return 1
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         with gil:
 *             if (pos[0], pos[1], pos[2]) in self._positions:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("yt.geometry.selection_routines.CutRegionSelector.select_cell", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":981
 *                 return 0
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

static int __pyx_f_2yt_8geometry_18selection_routines_17CutRegionSelector_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":982
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":981
 *                 return 0
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":984
 *         return 1
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

static int __pyx_f_2yt_8geometry_18selection_routines_17CutRegionSelector_select_sphere(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":985
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":984
 *         return 1
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":987
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         t = ()
 *         for i, c in enumerate(self._conditionals):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17CutRegionSelector_3_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17CutRegionSelector_3_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_17CutRegionSelector_2_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17CutRegionSelector_2_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_v_self) {
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_c = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":988
 * 
 *     def _hash_vals(self):
 *         t = ()             # <<<<<<<<<<<<<<
 *         for i, c in enumerate(self._conditionals):
 *             t += ("conditional[%s]" % i, c)
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_t = __pyx_empty_tuple;

  /* "yt/geometry/selection_routines.pyx":989
 *     def _hash_vals(self):
 *         t = ()
 *         for i, c in enumerate(self._conditionals):             # <<<<<<<<<<<<<<
 *             t += ("conditional[%s]" % i, c)
 *         return ("conditionals", t)
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_1 = __pyx_int_0;
  __pyx_t_2 = __pyx_v_self->_conditionals; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
  for (;;) {
    if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 989, __pyx_L1_error)
    #else
    __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 989, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_c, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 989, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "yt/geometry/selection_routines.pyx":990
 *         t = ()
 *         for i, c in enumerate(self._conditionals):
 *             t += ("conditional[%s]" % i, c)             # <<<<<<<<<<<<<<
 *         return ("conditionals", t)
 * 
 */
    __pyx_t_4 = __Pyx_PyString_FormatSafe(__pyx_kp_s_conditional_s, __pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 990, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 990, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_c);
    __Pyx_GIVEREF(__pyx_v_c);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_c);
    __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_t, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 990, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_t, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "yt/geometry/selection_routines.pyx":989
 *     def _hash_vals(self):
 *         t = ()
 *         for i, c in enumerate(self._conditionals):             # <<<<<<<<<<<<<<
 *             t += ("conditional[%s]" % i, c)
 *         return ("conditionals", t)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":991
 *         for i, c in enumerate(self._conditionals):
 *             t += ("conditional[%s]" % i, c)
 *         return ("conditionals", t)             # <<<<<<<<<<<<<<
 * 
 * cut_region_selector = CutRegionSelector
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_conditionals);
  __Pyx_GIVEREF(__pyx_n_s_conditionals);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_conditionals);
  __Pyx_INCREF(__pyx_v_t);
  __Pyx_GIVEREF(__pyx_v_t);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_t);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":987
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         t = ()
 *         for i, c in enumerate(self._conditionals):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("yt.geometry.selection_routines.CutRegionSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17CutRegionSelector_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17CutRegionSelector_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_17CutRegionSelector_4__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17CutRegionSelector_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.CutRegionSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17CutRegionSelector_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17CutRegionSelector_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_17CutRegionSelector_6__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17CutRegionSelector_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.CutRegionSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1001
 *     cdef np.float64_t height
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(3):
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_12DiskSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_12DiskSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1001, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1001, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.DiskSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_12DiskSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_12DiskSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_5numpy_float64_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":1003
 *     def __init__(self, dobj):
 *         cdef int i
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             self.norm_vec[i] = dobj._norm_vec[i]
 *             self.center[i] = _ensure_code(dobj.center[i])
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":1004
 *         cdef int i
 *         for i in range(3):
 *             self.norm_vec[i] = dobj._norm_vec[i]             # <<<<<<<<<<<<<<
 *             self.center[i] = _ensure_code(dobj.center[i])
 *         self.radius = _ensure_code(dobj.radius)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_norm_vec); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1004, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1004, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1004, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_self->norm_vec[__pyx_v_i]) = __pyx_t_4;

    /* "yt/geometry/selection_routines.pyx":1005
 *         for i in range(3):
 *             self.norm_vec[i] = dobj._norm_vec[i]
 *             self.center[i] = _ensure_code(dobj.center[i])             # <<<<<<<<<<<<<<
 *         self.radius = _ensure_code(dobj.radius)
 *         self.radius2 = self.radius * self.radius
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_center); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1005, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1005, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1005, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1005, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_self->center[__pyx_v_i]) = __pyx_t_4;
  }

  /* "yt/geometry/selection_routines.pyx":1006
 *             self.norm_vec[i] = dobj._norm_vec[i]
 *             self.center[i] = _ensure_code(dobj.center[i])
 *         self.radius = _ensure_code(dobj.radius)             # <<<<<<<<<<<<<<
 *         self.radius2 = self.radius * self.radius
 *         self.height = _ensure_code(dobj.height)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_radius); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self->radius = __pyx_t_4;

  /* "yt/geometry/selection_routines.pyx":1007
 *             self.center[i] = _ensure_code(dobj.center[i])
 *         self.radius = _ensure_code(dobj.radius)
 *         self.radius2 = self.radius * self.radius             # <<<<<<<<<<<<<<
 *         self.height = _ensure_code(dobj.height)
 * 
 */
  __pyx_v_self->radius2 = (__pyx_v_self->radius * __pyx_v_self->radius);

  /* "yt/geometry/selection_routines.pyx":1008
 *         self.radius = _ensure_code(dobj.radius)
 *         self.radius2 = self.radius * self.radius
 *         self.height = _ensure_code(dobj.height)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_height); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->height = __pyx_t_4;

  /* "yt/geometry/selection_routines.pyx":1001
 *     cdef np.float64_t height
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(3):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("yt.geometry.selection_routines.DiskSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1013
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         return self.select_point(pos)
 * 
 */

static int __pyx_f_2yt_8geometry_18selection_routines_12DiskSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1014
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         return self.select_point(pos)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_DiskSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.select_point(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_pos);
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1013
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         return self.select_point(pos)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1019
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef np.float64_t h, d, r2, temp
 *         cdef int i
 */

static int __pyx_f_2yt_8geometry_18selection_routines_12DiskSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  __pyx_t_5numpy_float64_t __pyx_v_h;
  __pyx_t_5numpy_float64_t __pyx_v_d;
  __pyx_t_5numpy_float64_t __pyx_v_r2;
  __pyx_t_5numpy_float64_t __pyx_v_temp;
  int __pyx_v_i;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":1022
 *         cdef np.float64_t h, d, r2, temp
 *         cdef int i
 *         h = d = 0             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             temp = self.difference(pos[i], self.center[i], i)
 */
  __pyx_v_h = 0.0;
  __pyx_v_d = 0.0;

  /* "yt/geometry/selection_routines.pyx":1023
 *         cdef int i
 *         h = d = 0
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             temp = self.difference(pos[i], self.center[i], i)
 *             h += temp * self.norm_vec[i]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":1024
 *         h = d = 0
 *         for i in range(3):
 *             temp = self.difference(pos[i], self.center[i], i)             # <<<<<<<<<<<<<<
 *             h += temp * self.norm_vec[i]
 *             d += temp*temp
 */
    __pyx_v_temp = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_DiskSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.difference(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), (__pyx_v_pos[__pyx_v_i]), (__pyx_v_self->center[__pyx_v_i]), __pyx_v_i);

    /* "yt/geometry/selection_routines.pyx":1025
 *         for i in range(3):
 *             temp = self.difference(pos[i], self.center[i], i)
 *             h += temp * self.norm_vec[i]             # <<<<<<<<<<<<<<
 *             d += temp*temp
 *         r2 = (d - h*h)
 */
    __pyx_v_h = (__pyx_v_h + (__pyx_v_temp * (__pyx_v_self->norm_vec[__pyx_v_i])));

    /* "yt/geometry/selection_routines.pyx":1026
 *             temp = self.difference(pos[i], self.center[i], i)
 *             h += temp * self.norm_vec[i]
 *             d += temp*temp             # <<<<<<<<<<<<<<
 *         r2 = (d - h*h)
 *         if fabs(h) <= self.height and r2 <= self.radius2: return 1
 */
    __pyx_v_d = (__pyx_v_d + (__pyx_v_temp * __pyx_v_temp));
  }

  /* "yt/geometry/selection_routines.pyx":1027
 *             h += temp * self.norm_vec[i]
 *             d += temp*temp
 *         r2 = (d - h*h)             # <<<<<<<<<<<<<<
 *         if fabs(h) <= self.height and r2 <= self.radius2: return 1
 *         return 0
 */
  __pyx_v_r2 = (__pyx_v_d - (__pyx_v_h * __pyx_v_h));

  /* "yt/geometry/selection_routines.pyx":1028
 *             d += temp*temp
 *         r2 = (d - h*h)
 *         if fabs(h) <= self.height and r2 <= self.radius2: return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_3 = ((fabs(__pyx_v_h) <= __pyx_v_self->height) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_3 = ((__pyx_v_r2 <= __pyx_v_self->radius2) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_2) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":1029
 *         r2 = (d - h*h)
 *         if fabs(h) <= self.height and r2 <= self.radius2: return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1019
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef np.float64_t h, d, r2, temp
 *         cdef int i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1034
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef np.float64_t h, d, r2, temp
 *         cdef int i
 */

static int __pyx_f_2yt_8geometry_18selection_routines_12DiskSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  __pyx_t_5numpy_float64_t __pyx_v_h;
  __pyx_t_5numpy_float64_t __pyx_v_d;
  __pyx_t_5numpy_float64_t __pyx_v_r2;
  __pyx_t_5numpy_float64_t __pyx_v_temp;
  int __pyx_v_i;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":1037
 *         cdef np.float64_t h, d, r2, temp
 *         cdef int i
 *         h = d = 0             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             temp = self.difference(pos[i], self.center[i], i)
 */
  __pyx_v_h = 0.0;
  __pyx_v_d = 0.0;

  /* "yt/geometry/selection_routines.pyx":1038
 *         cdef int i
 *         h = d = 0
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             temp = self.difference(pos[i], self.center[i], i)
 *             h += pos[i] * self.norm_vec[i]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":1039
 *         h = d = 0
 *         for i in range(3):
 *             temp = self.difference(pos[i], self.center[i], i)             # <<<<<<<<<<<<<<
 *             h += pos[i] * self.norm_vec[i]
 *             d += temp*temp
 */
    __pyx_v_temp = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_DiskSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.difference(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), (__pyx_v_pos[__pyx_v_i]), (__pyx_v_self->center[__pyx_v_i]), __pyx_v_i);

    /* "yt/geometry/selection_routines.pyx":1040
 *         for i in range(3):
 *             temp = self.difference(pos[i], self.center[i], i)
 *             h += pos[i] * self.norm_vec[i]             # <<<<<<<<<<<<<<
 *             d += temp*temp
 *         r2 = (d - h*h)
 */
    __pyx_v_h = (__pyx_v_h + ((__pyx_v_pos[__pyx_v_i]) * (__pyx_v_self->norm_vec[__pyx_v_i])));

    /* "yt/geometry/selection_routines.pyx":1041
 *             temp = self.difference(pos[i], self.center[i], i)
 *             h += pos[i] * self.norm_vec[i]
 *             d += temp*temp             # <<<<<<<<<<<<<<
 *         r2 = (d - h*h)
 *         d = self.radius+radius
 */
    __pyx_v_d = (__pyx_v_d + (__pyx_v_temp * __pyx_v_temp));
  }

  /* "yt/geometry/selection_routines.pyx":1042
 *             h += pos[i] * self.norm_vec[i]
 *             d += temp*temp
 *         r2 = (d - h*h)             # <<<<<<<<<<<<<<
 *         d = self.radius+radius
 *         if fabs(h) <= self.height+radius and r2 <= d*d: return 1
 */
  __pyx_v_r2 = (__pyx_v_d - (__pyx_v_h * __pyx_v_h));

  /* "yt/geometry/selection_routines.pyx":1043
 *             d += temp*temp
 *         r2 = (d - h*h)
 *         d = self.radius+radius             # <<<<<<<<<<<<<<
 *         if fabs(h) <= self.height+radius and r2 <= d*d: return 1
 *         return 0
 */
  __pyx_v_d = (__pyx_v_self->radius + __pyx_v_radius);

  /* "yt/geometry/selection_routines.pyx":1044
 *         r2 = (d - h*h)
 *         d = self.radius+radius
 *         if fabs(h) <= self.height+radius and r2 <= d*d: return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_3 = ((fabs(__pyx_v_h) <= (__pyx_v_self->height + __pyx_v_radius)) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_3 = ((__pyx_v_r2 <= (__pyx_v_d * __pyx_v_d)) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_2) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":1045
 *         d = self.radius+radius
 *         if fabs(h) <= self.height+radius and r2 <= d*d: return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1034
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef np.float64_t h, d, r2, temp
 *         cdef int i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1050
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         # Until we can get our OBB/OBB intersection correct, disable this.
 */

static int __pyx_f_2yt_8geometry_18selection_routines_12DiskSelector_select_bbox(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1053
 *                                np.float64_t right_edge[3]) nogil:
 *         # Until we can get our OBB/OBB intersection correct, disable this.
 *         return 1             # <<<<<<<<<<<<<<
 *         # cdef np.float64_t *arr[2]
 *         # cdef np.float64_t pos[3]
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1050
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         # Until we can get our OBB/OBB intersection correct, disable this.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1097
 *         # return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("norm_vec[0]", self.norm_vec[0]),
 *                 ("norm_vec[1]", self.norm_vec[1]),
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_12DiskSelector_3_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_12DiskSelector_3_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_12DiskSelector_2_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_12DiskSelector_2_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":1098
 * 
 *     def _hash_vals(self):
 *         return (("norm_vec[0]", self.norm_vec[0]),             # <<<<<<<<<<<<<<
 *                 ("norm_vec[1]", self.norm_vec[1]),
 *                 ("norm_vec[2]", self.norm_vec[2]),
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->norm_vec[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_norm_vec_0);
  __Pyx_GIVEREF(__pyx_kp_s_norm_vec_0);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_s_norm_vec_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1099
 *     def _hash_vals(self):
 *         return (("norm_vec[0]", self.norm_vec[0]),
 *                 ("norm_vec[1]", self.norm_vec[1]),             # <<<<<<<<<<<<<<
 *                 ("norm_vec[2]", self.norm_vec[2]),
 *                 ("center[0]", self.center[0]),
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->norm_vec[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_kp_s_norm_vec_1);
  __Pyx_GIVEREF(__pyx_kp_s_norm_vec_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_s_norm_vec_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1100
 *         return (("norm_vec[0]", self.norm_vec[0]),
 *                 ("norm_vec[1]", self.norm_vec[1]),
 *                 ("norm_vec[2]", self.norm_vec[2]),             # <<<<<<<<<<<<<<
 *                 ("center[0]", self.center[0]),
 *                 ("center[1]", self.center[1]),
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->norm_vec[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_kp_s_norm_vec_2);
  __Pyx_GIVEREF(__pyx_kp_s_norm_vec_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_s_norm_vec_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1101
 *                 ("norm_vec[1]", self.norm_vec[1]),
 *                 ("norm_vec[2]", self.norm_vec[2]),
 *                 ("center[0]", self.center[0]),             # <<<<<<<<<<<<<<
 *                 ("center[1]", self.center[1]),
 *                 ("center[2]", self.center[2]),
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->center[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_kp_s_center_0);
  __Pyx_GIVEREF(__pyx_kp_s_center_0);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_s_center_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1102
 *                 ("norm_vec[2]", self.norm_vec[2]),
 *                 ("center[0]", self.center[0]),
 *                 ("center[1]", self.center[1]),             # <<<<<<<<<<<<<<
 *                 ("center[2]", self.center[2]),
 *                 ("radius", self.radius),
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->center[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_kp_s_center_1);
  __Pyx_GIVEREF(__pyx_kp_s_center_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_center_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1103
 *                 ("center[0]", self.center[0]),
 *                 ("center[1]", self.center[1]),
 *                 ("center[2]", self.center[2]),             # <<<<<<<<<<<<<<
 *                 ("radius", self.radius),
 *                 ("radius2", self.radius2),
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->center[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_kp_s_center_2);
  __Pyx_GIVEREF(__pyx_kp_s_center_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_kp_s_center_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1104
 *                 ("center[1]", self.center[1]),
 *                 ("center[2]", self.center[2]),
 *                 ("radius", self.radius),             # <<<<<<<<<<<<<<
 *                 ("radius2", self.radius2),
 *                 ("height", self.height))
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->radius); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_n_s_radius);
  __Pyx_GIVEREF(__pyx_n_s_radius);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_n_s_radius);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1105
 *                 ("center[2]", self.center[2]),
 *                 ("radius", self.radius),
 *                 ("radius2", self.radius2),             # <<<<<<<<<<<<<<
 *                 ("height", self.height))
 * 
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->radius2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_n_s_radius2);
  __Pyx_GIVEREF(__pyx_n_s_radius2);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_radius2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1106
 *                 ("radius", self.radius),
 *                 ("radius2", self.radius2),
 *                 ("height", self.height))             # <<<<<<<<<<<<<<
 * 
 * disk_selector = DiskSelector
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->height); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_n_s_height);
  __Pyx_GIVEREF(__pyx_n_s_height);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_n_s_height);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1098
 * 
 *     def _hash_vals(self):
 *         return (("norm_vec[0]", self.norm_vec[0]),             # <<<<<<<<<<<<<<
 *                 ("norm_vec[1]", self.norm_vec[1]),
 *                 ("norm_vec[2]", self.norm_vec[2]),
 */
  __pyx_t_1 = PyTuple_New(9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_t_10);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1097
 *         # return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("norm_vec[0]", self.norm_vec[0]),
 *                 ("norm_vec[1]", self.norm_vec[1]),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("yt.geometry.selection_routines.DiskSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_12DiskSelector_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_12DiskSelector_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_12DiskSelector_4__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_12DiskSelector_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.DiskSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_12DiskSelector_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_12DiskSelector_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_12DiskSelector_6__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_12DiskSelector_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.DiskSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1114
 *     cdef np.float64_t d
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(3):
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1114, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1114, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.CuttingPlaneSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_20CuttingPlaneSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_20CuttingPlaneSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_5numpy_float64_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":1116
 *     def __init__(self, dobj):
 *         cdef int i
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             self.norm_vec[i] = dobj._norm_vec[i]
 *         self.d = _ensure_code(dobj._d)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":1117
 *         cdef int i
 *         for i in range(3):
 *             self.norm_vec[i] = dobj._norm_vec[i]             # <<<<<<<<<<<<<<
 *         self.d = _ensure_code(dobj._d)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_norm_vec); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1117, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_self->norm_vec[__pyx_v_i]) = __pyx_t_4;
  }

  /* "yt/geometry/selection_routines.pyx":1118
 *         for i in range(3):
 *             self.norm_vec[i] = dobj._norm_vec[i]
 *         self.d = _ensure_code(dobj._d)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_d_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self->d = __pyx_t_4;

  /* "yt/geometry/selection_routines.pyx":1114
 *     cdef np.float64_t d
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(3):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("yt.geometry.selection_routines.CuttingPlaneSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1123
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef np.float64_t left_edge[3]
 *         cdef np.float64_t right_edge[3]
 */

static int __pyx_f_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  __pyx_t_5numpy_float64_t __pyx_v_left_edge[3];
  __pyx_t_5numpy_float64_t __pyx_v_right_edge[3];
  int __pyx_v_i;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":1127
 *         cdef np.float64_t right_edge[3]
 *         cdef int i
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             left_edge[i] = pos[i] - 0.5*dds[i]
 *             right_edge[i] = pos[i] + 0.5*dds[i]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":1128
 *         cdef int i
 *         for i in range(3):
 *             left_edge[i] = pos[i] - 0.5*dds[i]             # <<<<<<<<<<<<<<
 *             right_edge[i] = pos[i] + 0.5*dds[i]
 *         return self.select_bbox(left_edge, right_edge)
 */
    (__pyx_v_left_edge[__pyx_v_i]) = ((__pyx_v_pos[__pyx_v_i]) - (0.5 * (__pyx_v_dds[__pyx_v_i])));

    /* "yt/geometry/selection_routines.pyx":1129
 *         for i in range(3):
 *             left_edge[i] = pos[i] - 0.5*dds[i]
 *             right_edge[i] = pos[i] + 0.5*dds[i]             # <<<<<<<<<<<<<<
 *         return self.select_bbox(left_edge, right_edge)
 * 
 */
    (__pyx_v_right_edge[__pyx_v_i]) = ((__pyx_v_pos[__pyx_v_i]) + (0.5 * (__pyx_v_dds[__pyx_v_i])));
  }

  /* "yt/geometry/selection_routines.pyx":1130
 *             left_edge[i] = pos[i] - 0.5*dds[i]
 *             right_edge[i] = pos[i] + 0.5*dds[i]
 *         return self.select_bbox(left_edge, right_edge)             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 */
  __pyx_r = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_CuttingPlaneSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.select_bbox(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_left_edge, __pyx_v_right_edge);
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1123
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef np.float64_t left_edge[3]
 *         cdef np.float64_t right_edge[3]
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1132
 *         return self.select_bbox(left_edge, right_edge)
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         # two 0-volume constructs don't intersect
 *         return 0
 */

static int __pyx_f_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1134
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         # two 0-volume constructs don't intersect
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1132
 *         return self.select_bbox(left_edge, right_edge)
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         # two 0-volume constructs don't intersect
 *         return 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1139
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.float64_t height = self.d
 */

static int __pyx_f_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_height;
  int __pyx_r;
  __pyx_t_5numpy_float64_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":1141
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef int i
 *         cdef np.float64_t height = self.d             # <<<<<<<<<<<<<<
 *         for i in range(3) :
 *             height += pos[i] * self.norm_vec[i]
 */
  __pyx_t_1 = __pyx_v_self->d;
  __pyx_v_height = __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":1142
 *         cdef int i
 *         cdef np.float64_t height = self.d
 *         for i in range(3) :             # <<<<<<<<<<<<<<
 *             height += pos[i] * self.norm_vec[i]
 *         if height*height <= radius*radius : return 1
 */
  for (__pyx_t_2 = 0; __pyx_t_2 < 3; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "yt/geometry/selection_routines.pyx":1143
 *         cdef np.float64_t height = self.d
 *         for i in range(3) :
 *             height += pos[i] * self.norm_vec[i]             # <<<<<<<<<<<<<<
 *         if height*height <= radius*radius : return 1
 *         return 0
 */
    __pyx_v_height = (__pyx_v_height + ((__pyx_v_pos[__pyx_v_i]) * (__pyx_v_self->norm_vec[__pyx_v_i])));
  }

  /* "yt/geometry/selection_routines.pyx":1144
 *         for i in range(3) :
 *             height += pos[i] * self.norm_vec[i]
 *         if height*height <= radius*radius : return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_3 = (((__pyx_v_height * __pyx_v_height) <= (__pyx_v_radius * __pyx_v_radius)) != 0);
  if (__pyx_t_3) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":1145
 *             height += pos[i] * self.norm_vec[i]
 *         if height*height <= radius*radius : return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1139
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.float64_t height = self.d
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1150
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef int i, j, k, n
 */

static int __pyx_f_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_n;
  __pyx_t_5numpy_float64_t *__pyx_v_arr[2];
  __pyx_t_5numpy_float64_t __pyx_v_pos[3];
  __pyx_t_5numpy_float64_t __pyx_v_gd;
  long __pyx_v_all_under;
  long __pyx_v_all_over;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __pyx_t_5numpy_float64_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;

  /* "yt/geometry/selection_routines.pyx":1156
 *         cdef np.float64_t pos[3]
 *         cdef np.float64_t gd
 *         arr[0] = left_edge             # <<<<<<<<<<<<<<
 *         arr[1] = right_edge
 *         all_under = 1
 */
  (__pyx_v_arr[0]) = __pyx_v_left_edge;

  /* "yt/geometry/selection_routines.pyx":1157
 *         cdef np.float64_t gd
 *         arr[0] = left_edge
 *         arr[1] = right_edge             # <<<<<<<<<<<<<<
 *         all_under = 1
 *         all_over = 1
 */
  (__pyx_v_arr[1]) = __pyx_v_right_edge;

  /* "yt/geometry/selection_routines.pyx":1158
 *         arr[0] = left_edge
 *         arr[1] = right_edge
 *         all_under = 1             # <<<<<<<<<<<<<<
 *         all_over = 1
 *         # Check each corner
 */
  __pyx_v_all_under = 1;

  /* "yt/geometry/selection_routines.pyx":1159
 *         arr[1] = right_edge
 *         all_under = 1
 *         all_over = 1             # <<<<<<<<<<<<<<
 *         # Check each corner
 *         for i in range(2):
 */
  __pyx_v_all_over = 1;

  /* "yt/geometry/selection_routines.pyx":1161
 *         all_over = 1
 *         # Check each corner
 *         for i in range(2):             # <<<<<<<<<<<<<<
 *             pos[0] = arr[i][0]
 *             for j in range(2):
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 2; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":1162
 *         # Check each corner
 *         for i in range(2):
 *             pos[0] = arr[i][0]             # <<<<<<<<<<<<<<
 *             for j in range(2):
 *                 pos[1] = arr[j][1]
 */
    (__pyx_v_pos[0]) = ((__pyx_v_arr[__pyx_v_i])[0]);

    /* "yt/geometry/selection_routines.pyx":1163
 *         for i in range(2):
 *             pos[0] = arr[i][0]
 *             for j in range(2):             # <<<<<<<<<<<<<<
 *                 pos[1] = arr[j][1]
 *                 for k in range(2):
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 2; __pyx_t_2+=1) {
      __pyx_v_j = __pyx_t_2;

      /* "yt/geometry/selection_routines.pyx":1164
 *             pos[0] = arr[i][0]
 *             for j in range(2):
 *                 pos[1] = arr[j][1]             # <<<<<<<<<<<<<<
 *                 for k in range(2):
 *                     pos[2] = arr[k][2]
 */
      (__pyx_v_pos[1]) = ((__pyx_v_arr[__pyx_v_j])[1]);

      /* "yt/geometry/selection_routines.pyx":1165
 *             for j in range(2):
 *                 pos[1] = arr[j][1]
 *                 for k in range(2):             # <<<<<<<<<<<<<<
 *                     pos[2] = arr[k][2]
 *                     gd = self.d
 */
      for (__pyx_t_3 = 0; __pyx_t_3 < 2; __pyx_t_3+=1) {
        __pyx_v_k = __pyx_t_3;

        /* "yt/geometry/selection_routines.pyx":1166
 *                 pos[1] = arr[j][1]
 *                 for k in range(2):
 *                     pos[2] = arr[k][2]             # <<<<<<<<<<<<<<
 *                     gd = self.d
 *                     for n in range(3):
 */
        (__pyx_v_pos[2]) = ((__pyx_v_arr[__pyx_v_k])[2]);

        /* "yt/geometry/selection_routines.pyx":1167
 *                 for k in range(2):
 *                     pos[2] = arr[k][2]
 *                     gd = self.d             # <<<<<<<<<<<<<<
 *                     for n in range(3):
 *                         gd += pos[n] * self.norm_vec[n]
 */
        __pyx_t_4 = __pyx_v_self->d;
        __pyx_v_gd = __pyx_t_4;

        /* "yt/geometry/selection_routines.pyx":1168
 *                     pos[2] = arr[k][2]
 *                     gd = self.d
 *                     for n in range(3):             # <<<<<<<<<<<<<<
 *                         gd += pos[n] * self.norm_vec[n]
 *                     # this allows corners and faces on the low-end to
 */
        for (__pyx_t_5 = 0; __pyx_t_5 < 3; __pyx_t_5+=1) {
          __pyx_v_n = __pyx_t_5;

          /* "yt/geometry/selection_routines.pyx":1169
 *                     gd = self.d
 *                     for n in range(3):
 *                         gd += pos[n] * self.norm_vec[n]             # <<<<<<<<<<<<<<
 *                     # this allows corners and faces on the low-end to
 *                     # collide, while not selecting cells on the high-side
 */
          __pyx_v_gd = (__pyx_v_gd + ((__pyx_v_pos[__pyx_v_n]) * (__pyx_v_self->norm_vec[__pyx_v_n])));
        }

        /* "yt/geometry/selection_routines.pyx":1172
 *                     # this allows corners and faces on the low-end to
 *                     # collide, while not selecting cells on the high-side
 *                     if i == 0 and j == 0 and k == 0 :             # <<<<<<<<<<<<<<
 *                         if gd <= 0: all_over = 0
 *                         if gd >= 0: all_under = 0
 */
        __pyx_t_7 = ((__pyx_v_i == 0) != 0);
        if (__pyx_t_7) {
        } else {
          __pyx_t_6 = __pyx_t_7;
          goto __pyx_L12_bool_binop_done;
        }
        __pyx_t_7 = ((__pyx_v_j == 0) != 0);
        if (__pyx_t_7) {
        } else {
          __pyx_t_6 = __pyx_t_7;
          goto __pyx_L12_bool_binop_done;
        }
        __pyx_t_7 = ((__pyx_v_k == 0) != 0);
        __pyx_t_6 = __pyx_t_7;
        __pyx_L12_bool_binop_done:;
        if (__pyx_t_6) {

          /* "yt/geometry/selection_routines.pyx":1173
 *                     # collide, while not selecting cells on the high-side
 *                     if i == 0 and j == 0 and k == 0 :
 *                         if gd <= 0: all_over = 0             # <<<<<<<<<<<<<<
 *                         if gd >= 0: all_under = 0
 *                     else :
 */
          __pyx_t_6 = ((__pyx_v_gd <= 0.0) != 0);
          if (__pyx_t_6) {
            __pyx_v_all_over = 0;
          }

          /* "yt/geometry/selection_routines.pyx":1174
 *                     if i == 0 and j == 0 and k == 0 :
 *                         if gd <= 0: all_over = 0
 *                         if gd >= 0: all_under = 0             # <<<<<<<<<<<<<<
 *                     else :
 *                         if gd < 0: all_over = 0
 */
          __pyx_t_6 = ((__pyx_v_gd >= 0.0) != 0);
          if (__pyx_t_6) {
            __pyx_v_all_under = 0;
          }

          /* "yt/geometry/selection_routines.pyx":1172
 *                     # this allows corners and faces on the low-end to
 *                     # collide, while not selecting cells on the high-side
 *                     if i == 0 and j == 0 and k == 0 :             # <<<<<<<<<<<<<<
 *                         if gd <= 0: all_over = 0
 *                         if gd >= 0: all_under = 0
 */
          goto __pyx_L11;
        }

        /* "yt/geometry/selection_routines.pyx":1176
 *                         if gd >= 0: all_under = 0
 *                     else :
 *                         if gd < 0: all_over = 0             # <<<<<<<<<<<<<<
 *                         if gd > 0: all_under = 0
 *         if all_over == 1 or all_under == 1:
 */
        /*else*/ {
          __pyx_t_6 = ((__pyx_v_gd < 0.0) != 0);
          if (__pyx_t_6) {
            __pyx_v_all_over = 0;
          }

          /* "yt/geometry/selection_routines.pyx":1177
 *                     else :
 *                         if gd < 0: all_over = 0
 *                         if gd > 0: all_under = 0             # <<<<<<<<<<<<<<
 *         if all_over == 1 or all_under == 1:
 *             return 0
 */
          __pyx_t_6 = ((__pyx_v_gd > 0.0) != 0);
          if (__pyx_t_6) {
            __pyx_v_all_under = 0;
          }
        }
        __pyx_L11:;
      }
    }
  }

  /* "yt/geometry/selection_routines.pyx":1178
 *                         if gd < 0: all_over = 0
 *                         if gd > 0: all_under = 0
 *         if all_over == 1 or all_under == 1:             # <<<<<<<<<<<<<<
 *             return 0
 *         return 1
 */
  __pyx_t_7 = ((__pyx_v_all_over == 1) != 0);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L20_bool_binop_done;
  }
  __pyx_t_7 = ((__pyx_v_all_under == 1) != 0);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L20_bool_binop_done:;
  if (__pyx_t_6) {

    /* "yt/geometry/selection_routines.pyx":1179
 *                         if gd > 0: all_under = 0
 *         if all_over == 1 or all_under == 1:
 *             return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":1178
 *                         if gd < 0: all_over = 0
 *                         if gd > 0: all_under = 0
 *         if all_over == 1 or all_under == 1:             # <<<<<<<<<<<<<<
 *             return 0
 *         return 1
 */
  }

  /* "yt/geometry/selection_routines.pyx":1180
 *         if all_over == 1 or all_under == 1:
 *             return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1150
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef int i, j, k, n
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1182
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("norm_vec[0]", self.norm_vec[0]),
 *                 ("norm_vec[1]", self.norm_vec[1]),
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_3_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_3_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_2_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_2_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":1183
 * 
 *     def _hash_vals(self):
 *         return (("norm_vec[0]", self.norm_vec[0]),             # <<<<<<<<<<<<<<
 *                 ("norm_vec[1]", self.norm_vec[1]),
 *                 ("norm_vec[2]", self.norm_vec[2]),
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->norm_vec[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_norm_vec_0);
  __Pyx_GIVEREF(__pyx_kp_s_norm_vec_0);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_s_norm_vec_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1184
 *     def _hash_vals(self):
 *         return (("norm_vec[0]", self.norm_vec[0]),
 *                 ("norm_vec[1]", self.norm_vec[1]),             # <<<<<<<<<<<<<<
 *                 ("norm_vec[2]", self.norm_vec[2]),
 *                 ("d", self.d))
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->norm_vec[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_kp_s_norm_vec_1);
  __Pyx_GIVEREF(__pyx_kp_s_norm_vec_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_s_norm_vec_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1185
 *         return (("norm_vec[0]", self.norm_vec[0]),
 *                 ("norm_vec[1]", self.norm_vec[1]),
 *                 ("norm_vec[2]", self.norm_vec[2]),             # <<<<<<<<<<<<<<
 *                 ("d", self.d))
 * 
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->norm_vec[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_kp_s_norm_vec_2);
  __Pyx_GIVEREF(__pyx_kp_s_norm_vec_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_s_norm_vec_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1186
 *                 ("norm_vec[1]", self.norm_vec[1]),
 *                 ("norm_vec[2]", self.norm_vec[2]),
 *                 ("d", self.d))             # <<<<<<<<<<<<<<
 * 
 * cutting_selector = CuttingPlaneSelector
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->d); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_s_d);
  __Pyx_GIVEREF(__pyx_n_s_d);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_d);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1183
 * 
 *     def _hash_vals(self):
 *         return (("norm_vec[0]", self.norm_vec[0]),             # <<<<<<<<<<<<<<
 *                 ("norm_vec[1]", self.norm_vec[1]),
 *                 ("norm_vec[2]", self.norm_vec[2]),
 */
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_5);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1182
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("norm_vec[0]", self.norm_vec[0]),
 *                 ("norm_vec[1]", self.norm_vec[1]),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("yt.geometry.selection_routines.CuttingPlaneSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_4__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.CuttingPlaneSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_6__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.CuttingPlaneSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1195
 *     cdef int ax, ay
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         self.axis = dobj.axis
 *         self.coord = _ensure_code(dobj.coord)
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_13SliceSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_13SliceSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1195, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1195, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.SliceSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_13SliceSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_13SliceSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_5numpy_float64_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":1196
 * 
 *     def __init__(self, dobj):
 *         self.axis = dobj.axis             # <<<<<<<<<<<<<<
 *         self.coord = _ensure_code(dobj.coord)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_axis); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->axis = __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":1197
 *     def __init__(self, dobj):
 *         self.axis = dobj.axis
 *         self.coord = _ensure_code(dobj.coord)             # <<<<<<<<<<<<<<
 * 
 *         self.ax = (self.axis+1) % 3
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_coord); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->coord = __pyx_t_4;

  /* "yt/geometry/selection_routines.pyx":1199
 *         self.coord = _ensure_code(dobj.coord)
 * 
 *         self.ax = (self.axis+1) % 3             # <<<<<<<<<<<<<<
 *         self.ay = (self.axis+2) % 3
 * 
 */
  __pyx_v_self->ax = __Pyx_mod_long((__pyx_v_self->axis + 1), 3);

  /* "yt/geometry/selection_routines.pyx":1200
 * 
 *         self.ax = (self.axis+1) % 3
 *         self.ay = (self.axis+2) % 3             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_v_self->ay = __Pyx_mod_long((__pyx_v_self->axis + 2), 3);

  /* "yt/geometry/selection_routines.pyx":1195
 *     cdef int ax, ay
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         self.axis = dobj.axis
 *         self.coord = _ensure_code(dobj.coord)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SliceSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1205
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def fill_mask(self, gobj):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=3] mask
 *         cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_13SliceSelector_3fill_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_gobj); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_13SliceSelector_3fill_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_gobj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fill_mask (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_13SliceSelector_2fill_mask(((struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *)__pyx_v_self), ((PyObject *)__pyx_v_gobj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_13SliceSelector_2fill_mask(struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self, PyObject *__pyx_v_gobj) {
  PyArrayObject *__pyx_v_mask = 0;
  PyArrayObject *__pyx_v_child_mask = 0;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_total;
  int __pyx_v_this_level;
  int __pyx_v_ind[3][2];
  __pyx_t_5numpy_int32_t __pyx_v_level;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_child_mask;
  __Pyx_Buffer __pyx_pybuffer_child_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int32_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyArrayObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyArrayObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_t_21;
  int __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fill_mask", 0);
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  __pyx_pybuffer_child_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_child_mask.refcount = 0;
  __pyx_pybuffernd_child_mask.data = NULL;
  __pyx_pybuffernd_child_mask.rcbuffer = &__pyx_pybuffer_child_mask;

  /* "yt/geometry/selection_routines.pyx":1209
 *         cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
 *         cdef int i, j, k
 *         cdef int total = 0             # <<<<<<<<<<<<<<
 *         cdef int this_level = 0
 *         cdef int ind[3][2]
 */
  __pyx_v_total = 0;

  /* "yt/geometry/selection_routines.pyx":1210
 *         cdef int i, j, k
 *         cdef int total = 0
 *         cdef int this_level = 0             # <<<<<<<<<<<<<<
 *         cdef int ind[3][2]
 *         cdef np.int32_t level = gobj.Level
 */
  __pyx_v_this_level = 0;

  /* "yt/geometry/selection_routines.pyx":1212
 *         cdef int this_level = 0
 *         cdef int ind[3][2]
 *         cdef np.int32_t level = gobj.Level             # <<<<<<<<<<<<<<
 *         _ensure_code(gobj.LeftEdge)
 *         _ensure_code(gobj.dds)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_Level); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_level = __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":1213
 *         cdef int ind[3][2]
 *         cdef np.int32_t level = gobj.Level
 *         _ensure_code(gobj.LeftEdge)             # <<<<<<<<<<<<<<
 *         _ensure_code(gobj.dds)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_LeftEdge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":1214
 *         cdef np.int32_t level = gobj.Level
 *         _ensure_code(gobj.LeftEdge)
 *         _ensure_code(gobj.dds)             # <<<<<<<<<<<<<<
 * 
 *         if level < self.min_level or level > self.max_level:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_dds); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1216
 *         _ensure_code(gobj.dds)
 * 
 *         if level < self.min_level or level > self.max_level:             # <<<<<<<<<<<<<<
 *             return None
 *         else:
 */
  __pyx_t_5 = ((__pyx_v_level < __pyx_v_self->__pyx_base.min_level) != 0);
  if (!__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = ((__pyx_v_level > __pyx_v_self->__pyx_base.max_level) != 0);
  __pyx_t_4 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "yt/geometry/selection_routines.pyx":1217
 * 
 *         if level < self.min_level or level > self.max_level:
 *             return None             # <<<<<<<<<<<<<<
 *         else:
 *             child_mask = gobj.child_mask
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":1216
 *         _ensure_code(gobj.dds)
 * 
 *         if level < self.min_level or level > self.max_level:             # <<<<<<<<<<<<<<
 *             return None
 *         else:
 */
  }

  /* "yt/geometry/selection_routines.pyx":1219
 *             return None
 *         else:
 *             child_mask = gobj.child_mask             # <<<<<<<<<<<<<<
 *             mask = np.zeros(gobj.ActiveDimensions, dtype=np.uint8)
 *             if level == self.max_level:
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_child_mask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1219, __pyx_L1_error)
    __pyx_t_6 = ((PyArrayObject *)__pyx_t_1);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
      __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack);
      if (unlikely(__pyx_t_7 < 0)) {
        PyErr_Fetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_child_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
        }
        __pyx_t_8 = __pyx_t_9 = __pyx_t_10 = 0;
      }
      __pyx_pybuffernd_child_mask.diminfo[0].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_child_mask.diminfo[0].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_child_mask.diminfo[1].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_child_mask.diminfo[1].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_child_mask.diminfo[2].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_child_mask.diminfo[2].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[2];
      if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1219, __pyx_L1_error)
    }
    __pyx_t_6 = 0;
    __pyx_v_child_mask = ((PyArrayObject *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "yt/geometry/selection_routines.pyx":1220
 *         else:
 *             child_mask = gobj.child_mask
 *             mask = np.zeros(gobj.ActiveDimensions, dtype=np.uint8)             # <<<<<<<<<<<<<<
 *             if level == self.max_level:
 *                 this_level = 1
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_ActiveDimensions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_np); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_uint8); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_13) < 0) __PYX_ERR(0, 1220, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_11, __pyx_t_1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(((__pyx_t_13) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_13, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1220, __pyx_L1_error)
    __pyx_t_14 = ((PyArrayObject *)__pyx_t_13);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
      __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_14, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack);
      if (unlikely(__pyx_t_7 < 0)) {
        PyErr_Fetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
        }
        __pyx_t_10 = __pyx_t_9 = __pyx_t_8 = 0;
      }
      __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mask.diminfo[1].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mask.diminfo[1].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_mask.diminfo[2].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_mask.diminfo[2].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[2];
      if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1220, __pyx_L1_error)
    }
    __pyx_t_14 = 0;
    __pyx_v_mask = ((PyArrayObject *)__pyx_t_13);
    __pyx_t_13 = 0;

    /* "yt/geometry/selection_routines.pyx":1221
 *             child_mask = gobj.child_mask
 *             mask = np.zeros(gobj.ActiveDimensions, dtype=np.uint8)
 *             if level == self.max_level:             # <<<<<<<<<<<<<<
 *                 this_level = 1
 *             for i in range(3):
 */
    __pyx_t_4 = ((__pyx_v_level == __pyx_v_self->__pyx_base.max_level) != 0);
    if (__pyx_t_4) {

      /* "yt/geometry/selection_routines.pyx":1222
 *             mask = np.zeros(gobj.ActiveDimensions, dtype=np.uint8)
 *             if level == self.max_level:
 *                 this_level = 1             # <<<<<<<<<<<<<<
 *             for i in range(3):
 *                 if i == self.axis:
 */
      __pyx_v_this_level = 1;

      /* "yt/geometry/selection_routines.pyx":1221
 *             child_mask = gobj.child_mask
 *             mask = np.zeros(gobj.ActiveDimensions, dtype=np.uint8)
 *             if level == self.max_level:             # <<<<<<<<<<<<<<
 *                 this_level = 1
 *             for i in range(3):
 */
    }

    /* "yt/geometry/selection_routines.pyx":1223
 *             if level == self.max_level:
 *                 this_level = 1
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 if i == self.axis:
 *                     ind[i][0] = \
 */
    for (__pyx_t_7 = 0; __pyx_t_7 < 3; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "yt/geometry/selection_routines.pyx":1224
 *                 this_level = 1
 *             for i in range(3):
 *                 if i == self.axis:             # <<<<<<<<<<<<<<
 *                     ind[i][0] = \
 *                         <int> ((self.coord - (gobj.LeftEdge[i]).to_ndarray()) /
 */
      __pyx_t_4 = ((__pyx_v_i == __pyx_v_self->axis) != 0);
      if (__pyx_t_4) {

        /* "yt/geometry/selection_routines.pyx":1226
 *                 if i == self.axis:
 *                     ind[i][0] = \
 *                         <int> ((self.coord - (gobj.LeftEdge[i]).to_ndarray()) /             # <<<<<<<<<<<<<<
 *                                gobj.dds[i])
 *                     ind[i][1] = ind[i][0] + 1
 */
        __pyx_t_13 = PyFloat_FromDouble(__pyx_v_self->coord); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1226, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_LeftEdge); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1226, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_11, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1226, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_to_ndarray); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1226, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
          }
        }
        __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_11);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1226, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = PyNumber_Subtract(__pyx_t_13, __pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1226, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "yt/geometry/selection_routines.pyx":1227
 *                     ind[i][0] = \
 *                         <int> ((self.coord - (gobj.LeftEdge[i]).to_ndarray()) /
 *                                gobj.dds[i])             # <<<<<<<<<<<<<<
 *                     ind[i][1] = ind[i][0] + 1
 *                 else:
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_dds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1227, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_13 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1227, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "yt/geometry/selection_routines.pyx":1226
 *                 if i == self.axis:
 *                     ind[i][0] = \
 *                         <int> ((self.coord - (gobj.LeftEdge[i]).to_ndarray()) /             # <<<<<<<<<<<<<<
 *                                gobj.dds[i])
 *                     ind[i][1] = ind[i][0] + 1
 */
        __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_11, __pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1226, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_15 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_15 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1226, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "yt/geometry/selection_routines.pyx":1225
 *             for i in range(3):
 *                 if i == self.axis:
 *                     ind[i][0] = \             # <<<<<<<<<<<<<<
 *                         <int> ((self.coord - (gobj.LeftEdge[i]).to_ndarray()) /
 *                                gobj.dds[i])
 */
        ((__pyx_v_ind[__pyx_v_i])[0]) = ((int)__pyx_t_15);

        /* "yt/geometry/selection_routines.pyx":1228
 *                         <int> ((self.coord - (gobj.LeftEdge[i]).to_ndarray()) /
 *                                gobj.dds[i])
 *                     ind[i][1] = ind[i][0] + 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     ind[i][0] = 0
 */
        ((__pyx_v_ind[__pyx_v_i])[1]) = (((__pyx_v_ind[__pyx_v_i])[0]) + 1);

        /* "yt/geometry/selection_routines.pyx":1224
 *                 this_level = 1
 *             for i in range(3):
 *                 if i == self.axis:             # <<<<<<<<<<<<<<
 *                     ind[i][0] = \
 *                         <int> ((self.coord - (gobj.LeftEdge[i]).to_ndarray()) /
 */
        goto __pyx_L9;
      }

      /* "yt/geometry/selection_routines.pyx":1230
 *                     ind[i][1] = ind[i][0] + 1
 *                 else:
 *                     ind[i][0] = 0             # <<<<<<<<<<<<<<
 *                     ind[i][1] = gobj.ActiveDimensions[i]
 *             with nogil:
 */
      /*else*/ {
        ((__pyx_v_ind[__pyx_v_i])[0]) = 0;

        /* "yt/geometry/selection_routines.pyx":1231
 *                 else:
 *                     ind[i][0] = 0
 *                     ind[i][1] = gobj.ActiveDimensions[i]             # <<<<<<<<<<<<<<
 *             with nogil:
 *                 for i in range(ind[0][0], ind[0][1]):
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_ActiveDimensions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1231, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_13 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1231, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_15 = __Pyx_PyInt_As_int(__pyx_t_13); if (unlikely((__pyx_t_15 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1231, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        ((__pyx_v_ind[__pyx_v_i])[1]) = __pyx_t_15;
      }
      __pyx_L9:;
    }

    /* "yt/geometry/selection_routines.pyx":1232
 *                     ind[i][0] = 0
 *                     ind[i][1] = gobj.ActiveDimensions[i]
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 for i in range(ind[0][0], ind[0][1]):
 *                     for j in range(ind[1][0], ind[1][1]):
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {

          /* "yt/geometry/selection_routines.pyx":1233
 *                     ind[i][1] = gobj.ActiveDimensions[i]
 *             with nogil:
 *                 for i in range(ind[0][0], ind[0][1]):             # <<<<<<<<<<<<<<
 *                     for j in range(ind[1][0], ind[1][1]):
 *                         for k in range(ind[2][0], ind[2][1]):
 */
          __pyx_t_7 = ((__pyx_v_ind[0])[1]);
          __pyx_t_15 = __pyx_t_7;
          for (__pyx_t_16 = ((__pyx_v_ind[0])[0]); __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
            __pyx_v_i = __pyx_t_16;

            /* "yt/geometry/selection_routines.pyx":1234
 *             with nogil:
 *                 for i in range(ind[0][0], ind[0][1]):
 *                     for j in range(ind[1][0], ind[1][1]):             # <<<<<<<<<<<<<<
 *                         for k in range(ind[2][0], ind[2][1]):
 *                             if this_level == 1 or child_mask[i, j, k]:
 */
            __pyx_t_17 = ((__pyx_v_ind[1])[1]);
            __pyx_t_18 = __pyx_t_17;
            for (__pyx_t_19 = ((__pyx_v_ind[1])[0]); __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
              __pyx_v_j = __pyx_t_19;

              /* "yt/geometry/selection_routines.pyx":1235
 *                 for i in range(ind[0][0], ind[0][1]):
 *                     for j in range(ind[1][0], ind[1][1]):
 *                         for k in range(ind[2][0], ind[2][1]):             # <<<<<<<<<<<<<<
 *                             if this_level == 1 or child_mask[i, j, k]:
 *                                 mask[i, j, k] = 1
 */
              __pyx_t_20 = ((__pyx_v_ind[2])[1]);
              __pyx_t_21 = __pyx_t_20;
              for (__pyx_t_22 = ((__pyx_v_ind[2])[0]); __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
                __pyx_v_k = __pyx_t_22;

                /* "yt/geometry/selection_routines.pyx":1236
 *                     for j in range(ind[1][0], ind[1][1]):
 *                         for k in range(ind[2][0], ind[2][1]):
 *                             if this_level == 1 or child_mask[i, j, k]:             # <<<<<<<<<<<<<<
 *                                 mask[i, j, k] = 1
 *                                 total += 1
 */
                __pyx_t_5 = ((__pyx_v_this_level == 1) != 0);
                if (!__pyx_t_5) {
                } else {
                  __pyx_t_4 = __pyx_t_5;
                  goto __pyx_L20_bool_binop_done;
                }
                __pyx_t_23 = __pyx_v_i;
                __pyx_t_24 = __pyx_v_j;
                __pyx_t_25 = __pyx_v_k;
                __pyx_t_5 = ((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.buf, __pyx_t_23, __pyx_pybuffernd_child_mask.diminfo[0].strides, __pyx_t_24, __pyx_pybuffernd_child_mask.diminfo[1].strides, __pyx_t_25, __pyx_pybuffernd_child_mask.diminfo[2].strides)) != 0);
                __pyx_t_4 = __pyx_t_5;
                __pyx_L20_bool_binop_done:;
                if (__pyx_t_4) {

                  /* "yt/geometry/selection_routines.pyx":1237
 *                         for k in range(ind[2][0], ind[2][1]):
 *                             if this_level == 1 or child_mask[i, j, k]:
 *                                 mask[i, j, k] = 1             # <<<<<<<<<<<<<<
 *                                 total += 1
 *             if total == 0: return None
 */
                  __pyx_t_25 = __pyx_v_i;
                  __pyx_t_24 = __pyx_v_j;
                  __pyx_t_23 = __pyx_v_k;
                  *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_mask.diminfo[0].strides, __pyx_t_24, __pyx_pybuffernd_mask.diminfo[1].strides, __pyx_t_23, __pyx_pybuffernd_mask.diminfo[2].strides) = 1;

                  /* "yt/geometry/selection_routines.pyx":1238
 *                             if this_level == 1 or child_mask[i, j, k]:
 *                                 mask[i, j, k] = 1
 *                                 total += 1             # <<<<<<<<<<<<<<
 *             if total == 0: return None
 *             return mask.astype("bool")
 */
                  __pyx_v_total = (__pyx_v_total + 1);

                  /* "yt/geometry/selection_routines.pyx":1236
 *                     for j in range(ind[1][0], ind[1][1]):
 *                         for k in range(ind[2][0], ind[2][1]):
 *                             if this_level == 1 or child_mask[i, j, k]:             # <<<<<<<<<<<<<<
 *                                 mask[i, j, k] = 1
 *                                 total += 1
 */
                }
              }
            }
          }
        }

        /* "yt/geometry/selection_routines.pyx":1232
 *                     ind[i][0] = 0
 *                     ind[i][1] = gobj.ActiveDimensions[i]
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 for i in range(ind[0][0], ind[0][1]):
 *                     for j in range(ind[1][0], ind[1][1]):
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L12;
          }
          __pyx_L12:;
        }
    }

    /* "yt/geometry/selection_routines.pyx":1239
 *                                 mask[i, j, k] = 1
 *                                 total += 1
 *             if total == 0: return None             # <<<<<<<<<<<<<<
 *             return mask.astype("bool")
 * 
 */
    __pyx_t_4 = ((__pyx_v_total == 0) != 0);
    if (__pyx_t_4) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;
    }

    /* "yt/geometry/selection_routines.pyx":1240
 *                                 total += 1
 *             if total == 0: return None
 *             return mask.astype("bool")             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_mask), __pyx_n_s_astype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1240, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_13 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_11, __pyx_n_s_bool) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_n_s_bool);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1240, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":1205
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def fill_mask(self, gobj):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=3] mask
 *         cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.SliceSelector.fill_mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mask);
  __Pyx_XDECREF((PyObject *)__pyx_v_child_mask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1245
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         if pos[self.axis] + 0.5*dds[self.axis] > self.coord \
 *            and pos[self.axis] - 0.5*dds[self.axis] - grid_eps <= self.coord:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_13SliceSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":1246
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         if pos[self.axis] + 0.5*dds[self.axis] > self.coord \             # <<<<<<<<<<<<<<
 *            and pos[self.axis] - 0.5*dds[self.axis] - grid_eps <= self.coord:
 *             return 1
 */
  __pyx_t_2 = ((((__pyx_v_pos[__pyx_v_self->axis]) + (0.5 * (__pyx_v_dds[__pyx_v_self->axis]))) > __pyx_v_self->coord) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":1247
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         if pos[self.axis] + 0.5*dds[self.axis] > self.coord \
 *            and pos[self.axis] - 0.5*dds[self.axis] - grid_eps <= self.coord:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  __pyx_t_2 = (((((__pyx_v_pos[__pyx_v_self->axis]) - (0.5 * (__pyx_v_dds[__pyx_v_self->axis]))) - __pyx_v_2yt_8geometry_18selection_routines_grid_eps) <= __pyx_v_self->coord) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "yt/geometry/selection_routines.pyx":1246
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         if pos[self.axis] + 0.5*dds[self.axis] > self.coord \             # <<<<<<<<<<<<<<
 *            and pos[self.axis] - 0.5*dds[self.axis] - grid_eps <= self.coord:
 *             return 1
 */
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":1248
 *         if pos[self.axis] + 0.5*dds[self.axis] > self.coord \
 *            and pos[self.axis] - 0.5*dds[self.axis] - grid_eps <= self.coord:
 *             return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":1246
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         if pos[self.axis] + 0.5*dds[self.axis] > self.coord \             # <<<<<<<<<<<<<<
 *            and pos[self.axis] - 0.5*dds[self.axis] - grid_eps <= self.coord:
 *             return 1
 */
  }

  /* "yt/geometry/selection_routines.pyx":1249
 *            and pos[self.axis] - 0.5*dds[self.axis] - grid_eps <= self.coord:
 *             return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1245
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         if pos[self.axis] + 0.5*dds[self.axis] > self.coord \
 *            and pos[self.axis] - 0.5*dds[self.axis] - grid_eps <= self.coord:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1251
 *         return 0
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         # two 0-volume constructs don't intersect
 *         return 0
 */

static int __pyx_f_2yt_8geometry_18selection_routines_13SliceSelector_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1253
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         # two 0-volume constructs don't intersect
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1251
 *         return 0
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         # two 0-volume constructs don't intersect
 *         return 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1258
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef np.float64_t dist = self.difference(pos[self.axis], self.coord, self.axis)
 *         if dist*dist < radius*radius:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_13SliceSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  __pyx_t_5numpy_float64_t __pyx_v_dist;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":1259
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef np.float64_t dist = self.difference(pos[self.axis], self.coord, self.axis)             # <<<<<<<<<<<<<<
 *         if dist*dist < radius*radius:
 *             return 1
 */
  __pyx_v_dist = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SliceSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.difference(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), (__pyx_v_pos[__pyx_v_self->axis]), __pyx_v_self->coord, __pyx_v_self->axis);

  /* "yt/geometry/selection_routines.pyx":1260
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef np.float64_t dist = self.difference(pos[self.axis], self.coord, self.axis)
 *         if dist*dist < radius*radius:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  __pyx_t_1 = (((__pyx_v_dist * __pyx_v_dist) < (__pyx_v_radius * __pyx_v_radius)) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":1261
 *         cdef np.float64_t dist = self.difference(pos[self.axis], self.coord, self.axis)
 *         if dist*dist < radius*radius:
 *             return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":1260
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef np.float64_t dist = self.difference(pos[self.axis], self.coord, self.axis)
 *         if dist*dist < radius*radius:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  }

  /* "yt/geometry/selection_routines.pyx":1262
 *         if dist*dist < radius*radius:
 *             return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1258
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef np.float64_t dist = self.difference(pos[self.axis], self.coord, self.axis)
 *         if dist*dist < radius*radius:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1267
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         if left_edge[self.axis] - grid_eps <= self.coord < right_edge[self.axis]:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_13SliceSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":1269
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 *         if left_edge[self.axis] - grid_eps <= self.coord < right_edge[self.axis]:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  __pyx_t_1 = (((__pyx_v_left_edge[__pyx_v_self->axis]) - __pyx_v_2yt_8geometry_18selection_routines_grid_eps) <= __pyx_v_self->coord);
  if (__pyx_t_1) {
    __pyx_t_1 = (__pyx_v_self->coord < (__pyx_v_right_edge[__pyx_v_self->axis]));
  }
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "yt/geometry/selection_routines.pyx":1270
 *                                np.float64_t right_edge[3]) nogil:
 *         if left_edge[self.axis] - grid_eps <= self.coord < right_edge[self.axis]:
 *             return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":1269
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 *         if left_edge[self.axis] - grid_eps <= self.coord < right_edge[self.axis]:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  }

  /* "yt/geometry/selection_routines.pyx":1271
 *         if left_edge[self.axis] - grid_eps <= self.coord < right_edge[self.axis]:
 *             return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1267
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         if left_edge[self.axis] - grid_eps <= self.coord < right_edge[self.axis]:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1273
 *         return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("axis", self.axis),
 *                 ("coord", self.coord))
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_13SliceSelector_5_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_13SliceSelector_5_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_13SliceSelector_4_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_13SliceSelector_4_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":1274
 * 
 *     def _hash_vals(self):
 *         return (("axis", self.axis),             # <<<<<<<<<<<<<<
 *                 ("coord", self.coord))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->axis); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_axis);
  __Pyx_GIVEREF(__pyx_n_s_axis);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_axis);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1275
 *     def _hash_vals(self):
 *         return (("axis", self.axis),
 *                 ("coord", self.coord))             # <<<<<<<<<<<<<<
 * 
 * slice_selector = SliceSelector
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->coord); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_coord);
  __Pyx_GIVEREF(__pyx_n_s_coord);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_coord);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1274
 * 
 *     def _hash_vals(self):
 *         return (("axis", self.axis),             # <<<<<<<<<<<<<<
 *                 ("coord", self.coord))
 * 
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1273
 *         return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("axis", self.axis),
 *                 ("coord", self.coord))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SliceSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_13SliceSelector_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_13SliceSelector_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_13SliceSelector_6__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_13SliceSelector_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SliceSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_13SliceSelector_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_13SliceSelector_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_13SliceSelector_8__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_13SliceSelector_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.SliceSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1287
 *     cdef int axis
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         self.axis = dobj.axis
 *         self.px_ax = dobj.px_ax
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_16OrthoRaySelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_16OrthoRaySelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1287, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1287, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.OrthoRaySelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_16OrthoRaySelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_16OrthoRaySelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __pyx_t_5numpy_uint8_t __pyx_t_3;
  __pyx_t_5numpy_float64_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":1288
 * 
 *     def __init__(self, dobj):
 *         self.axis = dobj.axis             # <<<<<<<<<<<<<<
 *         self.px_ax = dobj.px_ax
 *         self.py_ax = dobj.py_ax
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_axis); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1288, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->axis = __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":1289
 *     def __init__(self, dobj):
 *         self.axis = dobj.axis
 *         self.px_ax = dobj.px_ax             # <<<<<<<<<<<<<<
 *         self.py_ax = dobj.py_ax
 *         self.px = dobj.px
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_px_ax); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_npy_uint8(__pyx_t_1); if (unlikely((__pyx_t_3 == ((npy_uint8)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->px_ax = __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":1290
 *         self.axis = dobj.axis
 *         self.px_ax = dobj.px_ax
 *         self.py_ax = dobj.py_ax             # <<<<<<<<<<<<<<
 *         self.px = dobj.px
 *         self.py = dobj.py
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_py_ax); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_npy_uint8(__pyx_t_1); if (unlikely((__pyx_t_3 == ((npy_uint8)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->py_ax = __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":1291
 *         self.px_ax = dobj.px_ax
 *         self.py_ax = dobj.py_ax
 *         self.px = dobj.px             # <<<<<<<<<<<<<<
 *         self.py = dobj.py
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_px); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1291, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->px = __pyx_t_4;

  /* "yt/geometry/selection_routines.pyx":1292
 *         self.py_ax = dobj.py_ax
 *         self.px = dobj.px
 *         self.py = dobj.py             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_py); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1292, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->py = __pyx_t_4;

  /* "yt/geometry/selection_routines.pyx":1287
 *     cdef int axis
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         self.axis = dobj.axis
 *         self.px_ax = dobj.px_ax
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.OrthoRaySelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1297
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def fill_mask(self, gobj):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=3] mask
 *         cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_16OrthoRaySelector_3fill_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_gobj); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_16OrthoRaySelector_3fill_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_gobj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fill_mask (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_16OrthoRaySelector_2fill_mask(((struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *)__pyx_v_self), ((PyObject *)__pyx_v_gobj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_16OrthoRaySelector_2fill_mask(struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self, PyObject *__pyx_v_gobj) {
  PyArrayObject *__pyx_v_mask = 0;
  PyArrayObject *__pyx_v_child_mask = 0;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_total;
  int __pyx_v_this_level;
  int __pyx_v_ind[3][2];
  __pyx_t_5numpy_int32_t __pyx_v_level;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_child_mask;
  __Pyx_Buffer __pyx_pybuffer_child_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int32_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyArrayObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyArrayObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_t_21;
  int __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fill_mask", 0);
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  __pyx_pybuffer_child_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_child_mask.refcount = 0;
  __pyx_pybuffernd_child_mask.data = NULL;
  __pyx_pybuffernd_child_mask.rcbuffer = &__pyx_pybuffer_child_mask;

  /* "yt/geometry/selection_routines.pyx":1301
 *         cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
 *         cdef int i, j, k
 *         cdef int total = 0             # <<<<<<<<<<<<<<
 *         cdef int this_level = 0
 *         cdef int ind[3][2]
 */
  __pyx_v_total = 0;

  /* "yt/geometry/selection_routines.pyx":1302
 *         cdef int i, j, k
 *         cdef int total = 0
 *         cdef int this_level = 0             # <<<<<<<<<<<<<<
 *         cdef int ind[3][2]
 *         cdef np.int32_t level = gobj.Level
 */
  __pyx_v_this_level = 0;

  /* "yt/geometry/selection_routines.pyx":1304
 *         cdef int this_level = 0
 *         cdef int ind[3][2]
 *         cdef np.int32_t level = gobj.Level             # <<<<<<<<<<<<<<
 *         _ensure_code(gobj.LeftEdge)
 *         _ensure_code(gobj.RightEdge)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_Level); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1304, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_level = __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":1305
 *         cdef int ind[3][2]
 *         cdef np.int32_t level = gobj.Level
 *         _ensure_code(gobj.LeftEdge)             # <<<<<<<<<<<<<<
 *         _ensure_code(gobj.RightEdge)
 *         _ensure_code(gobj.dds)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_LeftEdge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":1306
 *         cdef np.int32_t level = gobj.Level
 *         _ensure_code(gobj.LeftEdge)
 *         _ensure_code(gobj.RightEdge)             # <<<<<<<<<<<<<<
 *         _ensure_code(gobj.dds)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_RightEdge); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1307
 *         _ensure_code(gobj.LeftEdge)
 *         _ensure_code(gobj.RightEdge)
 *         _ensure_code(gobj.dds)             # <<<<<<<<<<<<<<
 * 
 *         if level < self.min_level or level > self.max_level:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_dds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":1309
 *         _ensure_code(gobj.dds)
 * 
 *         if level < self.min_level or level > self.max_level:             # <<<<<<<<<<<<<<
 *             return None
 *         else:
 */
  __pyx_t_5 = ((__pyx_v_level < __pyx_v_self->__pyx_base.min_level) != 0);
  if (!__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = ((__pyx_v_level > __pyx_v_self->__pyx_base.max_level) != 0);
  __pyx_t_4 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "yt/geometry/selection_routines.pyx":1310
 * 
 *         if level < self.min_level or level > self.max_level:
 *             return None             # <<<<<<<<<<<<<<
 *         else:
 *             child_mask = gobj.child_mask
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":1309
 *         _ensure_code(gobj.dds)
 * 
 *         if level < self.min_level or level > self.max_level:             # <<<<<<<<<<<<<<
 *             return None
 *         else:
 */
  }

  /* "yt/geometry/selection_routines.pyx":1312
 *             return None
 *         else:
 *             child_mask = gobj.child_mask             # <<<<<<<<<<<<<<
 *             mask = np.zeros(gobj.ActiveDimensions, dtype=np.uint8)
 *             if level == self.max_level:
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_child_mask); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1312, __pyx_L1_error)
    __pyx_t_6 = ((PyArrayObject *)__pyx_t_3);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
      __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack);
      if (unlikely(__pyx_t_7 < 0)) {
        PyErr_Fetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_child_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
        }
        __pyx_t_8 = __pyx_t_9 = __pyx_t_10 = 0;
      }
      __pyx_pybuffernd_child_mask.diminfo[0].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_child_mask.diminfo[0].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_child_mask.diminfo[1].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_child_mask.diminfo[1].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_child_mask.diminfo[2].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_child_mask.diminfo[2].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[2];
      if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1312, __pyx_L1_error)
    }
    __pyx_t_6 = 0;
    __pyx_v_child_mask = ((PyArrayObject *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "yt/geometry/selection_routines.pyx":1313
 *         else:
 *             child_mask = gobj.child_mask
 *             mask = np.zeros(gobj.ActiveDimensions, dtype=np.uint8)             # <<<<<<<<<<<<<<
 *             if level == self.max_level:
 *                 this_level = 1
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_ActiveDimensions); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_np); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_uint8); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_13) < 0) __PYX_ERR(0, 1313, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_11, __pyx_t_3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(((__pyx_t_13) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_13, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1313, __pyx_L1_error)
    __pyx_t_14 = ((PyArrayObject *)__pyx_t_13);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
      __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_14, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack);
      if (unlikely(__pyx_t_7 < 0)) {
        PyErr_Fetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
        }
        __pyx_t_10 = __pyx_t_9 = __pyx_t_8 = 0;
      }
      __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mask.diminfo[1].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mask.diminfo[1].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_mask.diminfo[2].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_mask.diminfo[2].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[2];
      if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1313, __pyx_L1_error)
    }
    __pyx_t_14 = 0;
    __pyx_v_mask = ((PyArrayObject *)__pyx_t_13);
    __pyx_t_13 = 0;

    /* "yt/geometry/selection_routines.pyx":1314
 *             child_mask = gobj.child_mask
 *             mask = np.zeros(gobj.ActiveDimensions, dtype=np.uint8)
 *             if level == self.max_level:             # <<<<<<<<<<<<<<
 *                 this_level = 1
 *             ind[self.axis][0] = 0
 */
    __pyx_t_4 = ((__pyx_v_level == __pyx_v_self->__pyx_base.max_level) != 0);
    if (__pyx_t_4) {

      /* "yt/geometry/selection_routines.pyx":1315
 *             mask = np.zeros(gobj.ActiveDimensions, dtype=np.uint8)
 *             if level == self.max_level:
 *                 this_level = 1             # <<<<<<<<<<<<<<
 *             ind[self.axis][0] = 0
 *             ind[self.axis][1] = gobj.ActiveDimensions[self.axis]
 */
      __pyx_v_this_level = 1;

      /* "yt/geometry/selection_routines.pyx":1314
 *             child_mask = gobj.child_mask
 *             mask = np.zeros(gobj.ActiveDimensions, dtype=np.uint8)
 *             if level == self.max_level:             # <<<<<<<<<<<<<<
 *                 this_level = 1
 *             ind[self.axis][0] = 0
 */
    }

    /* "yt/geometry/selection_routines.pyx":1316
 *             if level == self.max_level:
 *                 this_level = 1
 *             ind[self.axis][0] = 0             # <<<<<<<<<<<<<<
 *             ind[self.axis][1] = gobj.ActiveDimensions[self.axis]
 *             ind[self.px_ax][0] = \
 */
    ((__pyx_v_ind[__pyx_v_self->axis])[0]) = 0;

    /* "yt/geometry/selection_routines.pyx":1317
 *                 this_level = 1
 *             ind[self.axis][0] = 0
 *             ind[self.axis][1] = gobj.ActiveDimensions[self.axis]             # <<<<<<<<<<<<<<
 *             ind[self.px_ax][0] = \
 *                 <int> ((self.px - (gobj.LeftEdge).to_ndarray()[self.px_ax]) /
 */
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_ActiveDimensions); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_13, __pyx_v_self->axis, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1317, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    ((__pyx_v_ind[__pyx_v_self->axis])[1]) = __pyx_t_7;

    /* "yt/geometry/selection_routines.pyx":1319
 *             ind[self.axis][1] = gobj.ActiveDimensions[self.axis]
 *             ind[self.px_ax][0] = \
 *                 <int> ((self.px - (gobj.LeftEdge).to_ndarray()[self.px_ax]) /             # <<<<<<<<<<<<<<
 *                        gobj.dds[self.px_ax])
 *             ind[self.px_ax][1] = ind[self.px_ax][0] + 1
 */
    __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->px); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_LeftEdge); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_to_ndarray); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_13 = (__pyx_t_11) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_11) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_13, __pyx_v_self->px_ax, __pyx_t_5numpy_uint8_t, 0, __Pyx_PyInt_From_npy_uint8, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = PyNumber_Subtract(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "yt/geometry/selection_routines.pyx":1320
 *             ind[self.px_ax][0] = \
 *                 <int> ((self.px - (gobj.LeftEdge).to_ndarray()[self.px_ax]) /
 *                        gobj.dds[self.px_ax])             # <<<<<<<<<<<<<<
 *             ind[self.px_ax][1] = ind[self.px_ax][0] + 1
 *             ind[self.py_ax][0] = \
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_dds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_self->px_ax, __pyx_t_5numpy_uint8_t, 0, __Pyx_PyInt_From_npy_uint8, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "yt/geometry/selection_routines.pyx":1319
 *             ind[self.axis][1] = gobj.ActiveDimensions[self.axis]
 *             ind[self.px_ax][0] = \
 *                 <int> ((self.px - (gobj.LeftEdge).to_ndarray()[self.px_ax]) /             # <<<<<<<<<<<<<<
 *                        gobj.dds[self.px_ax])
 *             ind[self.px_ax][1] = ind[self.px_ax][0] + 1
 */
    __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_13, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1319, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "yt/geometry/selection_routines.pyx":1318
 *             ind[self.axis][0] = 0
 *             ind[self.axis][1] = gobj.ActiveDimensions[self.axis]
 *             ind[self.px_ax][0] = \             # <<<<<<<<<<<<<<
 *                 <int> ((self.px - (gobj.LeftEdge).to_ndarray()[self.px_ax]) /
 *                        gobj.dds[self.px_ax])
 */
    ((__pyx_v_ind[__pyx_v_self->px_ax])[0]) = ((int)__pyx_t_7);

    /* "yt/geometry/selection_routines.pyx":1321
 *                 <int> ((self.px - (gobj.LeftEdge).to_ndarray()[self.px_ax]) /
 *                        gobj.dds[self.px_ax])
 *             ind[self.px_ax][1] = ind[self.px_ax][0] + 1             # <<<<<<<<<<<<<<
 *             ind[self.py_ax][0] = \
 *                 <int> ((self.py - (gobj.LeftEdge).to_ndarray()[self.py_ax]) /
 */
    ((__pyx_v_ind[__pyx_v_self->px_ax])[1]) = (((__pyx_v_ind[__pyx_v_self->px_ax])[0]) + 1);

    /* "yt/geometry/selection_routines.pyx":1323
 *             ind[self.px_ax][1] = ind[self.px_ax][0] + 1
 *             ind[self.py_ax][0] = \
 *                 <int> ((self.py - (gobj.LeftEdge).to_ndarray()[self.py_ax]) /             # <<<<<<<<<<<<<<
 *                        gobj.dds[self.py_ax])
 *             ind[self.py_ax][1] = ind[self.py_ax][0] + 1
 */
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->py); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_LeftEdge); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_to_ndarray); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_3 = (__pyx_t_13) ? __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_13) : __Pyx_PyObject_CallNoArg(__pyx_t_11);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_self->py_ax, __pyx_t_5numpy_uint8_t, 0, __Pyx_PyInt_From_npy_uint8, 0, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_1, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "yt/geometry/selection_routines.pyx":1324
 *             ind[self.py_ax][0] = \
 *                 <int> ((self.py - (gobj.LeftEdge).to_ndarray()[self.py_ax]) /
 *                        gobj.dds[self.py_ax])             # <<<<<<<<<<<<<<
 *             ind[self.py_ax][1] = ind[self.py_ax][0] + 1
 * 
 */
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_dds); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_11, __pyx_v_self->py_ax, __pyx_t_5numpy_uint8_t, 0, __Pyx_PyInt_From_npy_uint8, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "yt/geometry/selection_routines.pyx":1323
 *             ind[self.px_ax][1] = ind[self.px_ax][0] + 1
 *             ind[self.py_ax][0] = \
 *                 <int> ((self.py - (gobj.LeftEdge).to_ndarray()[self.py_ax]) /             # <<<<<<<<<<<<<<
 *                        gobj.dds[self.py_ax])
 *             ind[self.py_ax][1] = ind[self.py_ax][0] + 1
 */
    __pyx_t_11 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_11); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1323, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "yt/geometry/selection_routines.pyx":1322
 *                        gobj.dds[self.px_ax])
 *             ind[self.px_ax][1] = ind[self.px_ax][0] + 1
 *             ind[self.py_ax][0] = \             # <<<<<<<<<<<<<<
 *                 <int> ((self.py - (gobj.LeftEdge).to_ndarray()[self.py_ax]) /
 *                        gobj.dds[self.py_ax])
 */
    ((__pyx_v_ind[__pyx_v_self->py_ax])[0]) = ((int)__pyx_t_7);

    /* "yt/geometry/selection_routines.pyx":1325
 *                 <int> ((self.py - (gobj.LeftEdge).to_ndarray()[self.py_ax]) /
 *                        gobj.dds[self.py_ax])
 *             ind[self.py_ax][1] = ind[self.py_ax][0] + 1             # <<<<<<<<<<<<<<
 * 
 *             with nogil:
 */
    ((__pyx_v_ind[__pyx_v_self->py_ax])[1]) = (((__pyx_v_ind[__pyx_v_self->py_ax])[0]) + 1);

    /* "yt/geometry/selection_routines.pyx":1327
 *             ind[self.py_ax][1] = ind[self.py_ax][0] + 1
 * 
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 for i in range(ind[0][0], ind[0][1]):
 *                     for j in range(ind[1][0], ind[1][1]):
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {

          /* "yt/geometry/selection_routines.pyx":1328
 * 
 *             with nogil:
 *                 for i in range(ind[0][0], ind[0][1]):             # <<<<<<<<<<<<<<
 *                     for j in range(ind[1][0], ind[1][1]):
 *                         for k in range(ind[2][0], ind[2][1]):
 */
          __pyx_t_7 = ((__pyx_v_ind[0])[1]);
          __pyx_t_15 = __pyx_t_7;
          for (__pyx_t_16 = ((__pyx_v_ind[0])[0]); __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
            __pyx_v_i = __pyx_t_16;

            /* "yt/geometry/selection_routines.pyx":1329
 *             with nogil:
 *                 for i in range(ind[0][0], ind[0][1]):
 *                     for j in range(ind[1][0], ind[1][1]):             # <<<<<<<<<<<<<<
 *                         for k in range(ind[2][0], ind[2][1]):
 *                             if this_level == 1 or child_mask[i, j, k]:
 */
            __pyx_t_17 = ((__pyx_v_ind[1])[1]);
            __pyx_t_18 = __pyx_t_17;
            for (__pyx_t_19 = ((__pyx_v_ind[1])[0]); __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
              __pyx_v_j = __pyx_t_19;

              /* "yt/geometry/selection_routines.pyx":1330
 *                 for i in range(ind[0][0], ind[0][1]):
 *                     for j in range(ind[1][0], ind[1][1]):
 *                         for k in range(ind[2][0], ind[2][1]):             # <<<<<<<<<<<<<<
 *                             if this_level == 1 or child_mask[i, j, k]:
 *                                 mask[i, j, k] = 1
 */
              __pyx_t_20 = ((__pyx_v_ind[2])[1]);
              __pyx_t_21 = __pyx_t_20;
              for (__pyx_t_22 = ((__pyx_v_ind[2])[0]); __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
                __pyx_v_k = __pyx_t_22;

                /* "yt/geometry/selection_routines.pyx":1331
 *                     for j in range(ind[1][0], ind[1][1]):
 *                         for k in range(ind[2][0], ind[2][1]):
 *                             if this_level == 1 or child_mask[i, j, k]:             # <<<<<<<<<<<<<<
 *                                 mask[i, j, k] = 1
 *                                 total += 1
 */
                __pyx_t_5 = ((__pyx_v_this_level == 1) != 0);
                if (!__pyx_t_5) {
                } else {
                  __pyx_t_4 = __pyx_t_5;
                  goto __pyx_L17_bool_binop_done;
                }
                __pyx_t_23 = __pyx_v_i;
                __pyx_t_24 = __pyx_v_j;
                __pyx_t_25 = __pyx_v_k;
                __pyx_t_5 = ((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.buf, __pyx_t_23, __pyx_pybuffernd_child_mask.diminfo[0].strides, __pyx_t_24, __pyx_pybuffernd_child_mask.diminfo[1].strides, __pyx_t_25, __pyx_pybuffernd_child_mask.diminfo[2].strides)) != 0);
                __pyx_t_4 = __pyx_t_5;
                __pyx_L17_bool_binop_done:;
                if (__pyx_t_4) {

                  /* "yt/geometry/selection_routines.pyx":1332
 *                         for k in range(ind[2][0], ind[2][1]):
 *                             if this_level == 1 or child_mask[i, j, k]:
 *                                 mask[i, j, k] = 1             # <<<<<<<<<<<<<<
 *                                 total += 1
 *             if total == 0: return None
 */
                  __pyx_t_25 = __pyx_v_i;
                  __pyx_t_24 = __pyx_v_j;
                  __pyx_t_23 = __pyx_v_k;
                  *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_mask.diminfo[0].strides, __pyx_t_24, __pyx_pybuffernd_mask.diminfo[1].strides, __pyx_t_23, __pyx_pybuffernd_mask.diminfo[2].strides) = 1;

                  /* "yt/geometry/selection_routines.pyx":1333
 *                             if this_level == 1 or child_mask[i, j, k]:
 *                                 mask[i, j, k] = 1
 *                                 total += 1             # <<<<<<<<<<<<<<
 *             if total == 0: return None
 *             return mask.astype("bool")
 */
                  __pyx_v_total = (__pyx_v_total + 1);

                  /* "yt/geometry/selection_routines.pyx":1331
 *                     for j in range(ind[1][0], ind[1][1]):
 *                         for k in range(ind[2][0], ind[2][1]):
 *                             if this_level == 1 or child_mask[i, j, k]:             # <<<<<<<<<<<<<<
 *                                 mask[i, j, k] = 1
 *                                 total += 1
 */
                }
              }
            }
          }
        }

        /* "yt/geometry/selection_routines.pyx":1327
 *             ind[self.py_ax][1] = ind[self.py_ax][0] + 1
 * 
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 for i in range(ind[0][0], ind[0][1]):
 *                     for j in range(ind[1][0], ind[1][1]):
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L9;
          }
          __pyx_L9:;
        }
    }

    /* "yt/geometry/selection_routines.pyx":1334
 *                                 mask[i, j, k] = 1
 *                                 total += 1
 *             if total == 0: return None             # <<<<<<<<<<<<<<
 *             return mask.astype("bool")
 * 
 */
    __pyx_t_4 = ((__pyx_v_total == 0) != 0);
    if (__pyx_t_4) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;
    }

    /* "yt/geometry/selection_routines.pyx":1335
 *                                 total += 1
 *             if total == 0: return None
 *             return mask.astype("bool")             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_mask), __pyx_n_s_astype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_11 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_n_s_bool) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_n_s_bool);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_11;
    __pyx_t_11 = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":1297
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def fill_mask(self, gobj):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=3] mask
 *         cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.OrthoRaySelector.fill_mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mask);
  __Pyx_XDECREF((PyObject *)__pyx_v_child_mask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1340
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         if self.px >= pos[self.px_ax] - 0.5*dds[self.px_ax] and \
 *            self.px <  pos[self.px_ax] + 0.5*dds[self.px_ax] and \
 */

static int __pyx_f_2yt_8geometry_18selection_routines_16OrthoRaySelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":1341
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         if self.px >= pos[self.px_ax] - 0.5*dds[self.px_ax] and \             # <<<<<<<<<<<<<<
 *            self.px <  pos[self.px_ax] + 0.5*dds[self.px_ax] and \
 *            self.py >= pos[self.py_ax] - 0.5*dds[self.py_ax] and \
 */
  __pyx_t_2 = ((__pyx_v_self->px >= ((__pyx_v_pos[__pyx_v_self->px_ax]) - (0.5 * (__pyx_v_dds[__pyx_v_self->px_ax])))) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":1342
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         if self.px >= pos[self.px_ax] - 0.5*dds[self.px_ax] and \
 *            self.px <  pos[self.px_ax] + 0.5*dds[self.px_ax] and \             # <<<<<<<<<<<<<<
 *            self.py >= pos[self.py_ax] - 0.5*dds[self.py_ax] and \
 *            self.py <  pos[self.py_ax] + 0.5*dds[self.py_ax]:
 */
  __pyx_t_2 = ((__pyx_v_self->px < ((__pyx_v_pos[__pyx_v_self->px_ax]) + (0.5 * (__pyx_v_dds[__pyx_v_self->px_ax])))) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":1343
 *         if self.px >= pos[self.px_ax] - 0.5*dds[self.px_ax] and \
 *            self.px <  pos[self.px_ax] + 0.5*dds[self.px_ax] and \
 *            self.py >= pos[self.py_ax] - 0.5*dds[self.py_ax] and \             # <<<<<<<<<<<<<<
 *            self.py <  pos[self.py_ax] + 0.5*dds[self.py_ax]:
 *             return 1
 */
  __pyx_t_2 = ((__pyx_v_self->py >= ((__pyx_v_pos[__pyx_v_self->py_ax]) - (0.5 * (__pyx_v_dds[__pyx_v_self->py_ax])))) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":1344
 *            self.px <  pos[self.px_ax] + 0.5*dds[self.px_ax] and \
 *            self.py >= pos[self.py_ax] - 0.5*dds[self.py_ax] and \
 *            self.py <  pos[self.py_ax] + 0.5*dds[self.py_ax]:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  __pyx_t_2 = ((__pyx_v_self->py < ((__pyx_v_pos[__pyx_v_self->py_ax]) + (0.5 * (__pyx_v_dds[__pyx_v_self->py_ax])))) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "yt/geometry/selection_routines.pyx":1341
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         if self.px >= pos[self.px_ax] - 0.5*dds[self.px_ax] and \             # <<<<<<<<<<<<<<
 *            self.px <  pos[self.px_ax] + 0.5*dds[self.px_ax] and \
 *            self.py >= pos[self.py_ax] - 0.5*dds[self.py_ax] and \
 */
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":1345
 *            self.py >= pos[self.py_ax] - 0.5*dds[self.py_ax] and \
 *            self.py <  pos[self.py_ax] + 0.5*dds[self.py_ax]:
 *             return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":1341
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         if self.px >= pos[self.px_ax] - 0.5*dds[self.px_ax] and \             # <<<<<<<<<<<<<<
 *            self.px <  pos[self.px_ax] + 0.5*dds[self.px_ax] and \
 *            self.py >= pos[self.py_ax] - 0.5*dds[self.py_ax] and \
 */
  }

  /* "yt/geometry/selection_routines.pyx":1346
 *            self.py <  pos[self.py_ax] + 0.5*dds[self.py_ax]:
 *             return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1340
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         if self.px >= pos[self.px_ax] - 0.5*dds[self.px_ax] and \
 *            self.px <  pos[self.px_ax] + 0.5*dds[self.px_ax] and \
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1348
 *         return 0
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         # two 0-volume constructs don't intersect
 *         return 0
 */

static int __pyx_f_2yt_8geometry_18selection_routines_16OrthoRaySelector_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1350
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         # two 0-volume constructs don't intersect
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1348
 *         return 0
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         # two 0-volume constructs don't intersect
 *         return 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1355
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef np.float64_t dx = self.difference(pos[self.px_ax], self.px, self.px_ax)
 *         cdef np.float64_t dy = self.difference(pos[self.py_ax], self.py, self.py_ax)
 */

static int __pyx_f_2yt_8geometry_18selection_routines_16OrthoRaySelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  __pyx_t_5numpy_float64_t __pyx_v_dx;
  __pyx_t_5numpy_float64_t __pyx_v_dy;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":1356
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef np.float64_t dx = self.difference(pos[self.px_ax], self.px, self.px_ax)             # <<<<<<<<<<<<<<
 *         cdef np.float64_t dy = self.difference(pos[self.py_ax], self.py, self.py_ax)
 *         if dx*dx + dy*dy < radius*radius:
 */
  __pyx_v_dx = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_OrthoRaySelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.difference(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), (__pyx_v_pos[__pyx_v_self->px_ax]), __pyx_v_self->px, __pyx_v_self->px_ax);

  /* "yt/geometry/selection_routines.pyx":1357
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef np.float64_t dx = self.difference(pos[self.px_ax], self.px, self.px_ax)
 *         cdef np.float64_t dy = self.difference(pos[self.py_ax], self.py, self.py_ax)             # <<<<<<<<<<<<<<
 *         if dx*dx + dy*dy < radius*radius:
 *             return 1
 */
  __pyx_v_dy = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_OrthoRaySelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.difference(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), (__pyx_v_pos[__pyx_v_self->py_ax]), __pyx_v_self->py, __pyx_v_self->py_ax);

  /* "yt/geometry/selection_routines.pyx":1358
 *         cdef np.float64_t dx = self.difference(pos[self.px_ax], self.px, self.px_ax)
 *         cdef np.float64_t dy = self.difference(pos[self.py_ax], self.py, self.py_ax)
 *         if dx*dx + dy*dy < radius*radius:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  __pyx_t_1 = ((((__pyx_v_dx * __pyx_v_dx) + (__pyx_v_dy * __pyx_v_dy)) < (__pyx_v_radius * __pyx_v_radius)) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":1359
 *         cdef np.float64_t dy = self.difference(pos[self.py_ax], self.py, self.py_ax)
 *         if dx*dx + dy*dy < radius*radius:
 *             return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":1358
 *         cdef np.float64_t dx = self.difference(pos[self.px_ax], self.px, self.px_ax)
 *         cdef np.float64_t dy = self.difference(pos[self.py_ax], self.py, self.py_ax)
 *         if dx*dx + dy*dy < radius*radius:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  }

  /* "yt/geometry/selection_routines.pyx":1360
 *         if dx*dx + dy*dy < radius*radius:
 *             return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1355
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef np.float64_t dx = self.difference(pos[self.px_ax], self.px, self.px_ax)
 *         cdef np.float64_t dy = self.difference(pos[self.py_ax], self.py, self.py_ax)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1365
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         if left_edge[self.px_ax] <= self.px < right_edge[self.px_ax] and \
 */

static int __pyx_f_2yt_8geometry_18selection_routines_16OrthoRaySelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":1367
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 *         if left_edge[self.px_ax] <= self.px < right_edge[self.px_ax] and \             # <<<<<<<<<<<<<<
 *            left_edge[self.py_ax] <= self.py < right_edge[self.py_ax] :
 *             return 1
 */
  __pyx_t_2 = ((__pyx_v_left_edge[__pyx_v_self->px_ax]) <= __pyx_v_self->px);
  if (__pyx_t_2) {
    __pyx_t_2 = (__pyx_v_self->px < (__pyx_v_right_edge[__pyx_v_self->px_ax]));
  }
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":1368
 *                                np.float64_t right_edge[3]) nogil:
 *         if left_edge[self.px_ax] <= self.px < right_edge[self.px_ax] and \
 *            left_edge[self.py_ax] <= self.py < right_edge[self.py_ax] :             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  __pyx_t_3 = ((__pyx_v_left_edge[__pyx_v_self->py_ax]) <= __pyx_v_self->py);
  if (__pyx_t_3) {
    __pyx_t_3 = (__pyx_v_self->py < (__pyx_v_right_edge[__pyx_v_self->py_ax]));
  }
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "yt/geometry/selection_routines.pyx":1367
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 *         if left_edge[self.px_ax] <= self.px < right_edge[self.px_ax] and \             # <<<<<<<<<<<<<<
 *            left_edge[self.py_ax] <= self.py < right_edge[self.py_ax] :
 *             return 1
 */
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":1369
 *         if left_edge[self.px_ax] <= self.px < right_edge[self.px_ax] and \
 *            left_edge[self.py_ax] <= self.py < right_edge[self.py_ax] :
 *             return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":1367
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 *         if left_edge[self.px_ax] <= self.px < right_edge[self.px_ax] and \             # <<<<<<<<<<<<<<
 *            left_edge[self.py_ax] <= self.py < right_edge[self.py_ax] :
 *             return 1
 */
  }

  /* "yt/geometry/selection_routines.pyx":1370
 *            left_edge[self.py_ax] <= self.py < right_edge[self.py_ax] :
 *             return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1365
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         if left_edge[self.px_ax] <= self.px < right_edge[self.px_ax] and \
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1372
 *         return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("px_ax", self.px_ax),
 *                 ("py_ax", self.py_ax),
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_16OrthoRaySelector_5_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_16OrthoRaySelector_5_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_16OrthoRaySelector_4_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_16OrthoRaySelector_4_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":1373
 * 
 *     def _hash_vals(self):
 *         return (("px_ax", self.px_ax),             # <<<<<<<<<<<<<<
 *                 ("py_ax", self.py_ax),
 *                 ("px", self.px),
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint8(__pyx_v_self->px_ax); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_px_ax);
  __Pyx_GIVEREF(__pyx_n_s_px_ax);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_px_ax);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1374
 *     def _hash_vals(self):
 *         return (("px_ax", self.px_ax),
 *                 ("py_ax", self.py_ax),             # <<<<<<<<<<<<<<
 *                 ("px", self.px),
 *                 ("py", self.py),
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint8(__pyx_v_self->py_ax); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_py_ax);
  __Pyx_GIVEREF(__pyx_n_s_py_ax);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_py_ax);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1375
 *         return (("px_ax", self.px_ax),
 *                 ("py_ax", self.py_ax),
 *                 ("px", self.px),             # <<<<<<<<<<<<<<
 *                 ("py", self.py),
 *                 ("axis", self.axis))
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->px); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_px);
  __Pyx_GIVEREF(__pyx_n_s_px);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_px);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1376
 *                 ("py_ax", self.py_ax),
 *                 ("px", self.px),
 *                 ("py", self.py),             # <<<<<<<<<<<<<<
 *                 ("axis", self.axis))
 * 
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->py); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_s_py);
  __Pyx_GIVEREF(__pyx_n_s_py);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_py);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1377
 *                 ("px", self.px),
 *                 ("py", self.py),
 *                 ("axis", self.axis))             # <<<<<<<<<<<<<<
 * 
 * ortho_ray_selector = OrthoRaySelector
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->axis); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_s_axis);
  __Pyx_GIVEREF(__pyx_n_s_axis);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_s_axis);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1373
 * 
 *     def _hash_vals(self):
 *         return (("px_ax", self.px_ax),             # <<<<<<<<<<<<<<
 *                 ("py_ax", self.py_ax),
 *                 ("px", self.px),
 */
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1372
 *         return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("px_ax", self.px_ax),
 *                 ("py_ax", self.py_ax),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("yt.geometry.selection_routines.OrthoRaySelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_16OrthoRaySelector_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_16OrthoRaySelector_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_16OrthoRaySelector_6__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_16OrthoRaySelector_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.OrthoRaySelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_16OrthoRaySelector_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_16OrthoRaySelector_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_16OrthoRaySelector_8__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_16OrthoRaySelector_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.OrthoRaySelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1387
 *     int hits
 * 
 * cdef void dt_sampler(             # <<<<<<<<<<<<<<
 *              VolumeContainer *vc,
 *              np.float64_t v_pos[3],
 */

static void __pyx_f_2yt_8geometry_18selection_routines_dt_sampler(struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer *__pyx_v_vc, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_v_dir, __pyx_t_5numpy_float64_t __pyx_v_enter_t, __pyx_t_5numpy_float64_t __pyx_v_exit_t, int *__pyx_v_index, void *__pyx_v_data) {
  struct __pyx_t_2yt_8geometry_18selection_routines_IntegrationAccumulator *__pyx_v_am;
  int __pyx_v_di;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":1395
 *              int index[3],
 *              void *data) nogil:
 *     cdef IntegrationAccumulator *am = <IntegrationAccumulator *> data             # <<<<<<<<<<<<<<
 *     cdef int di = (index[0]*vc.dims[1]+index[1])*vc.dims[2]+index[2]
 *     if am.child_mask[di] == 0 or enter_t == exit_t:
 */
  __pyx_v_am = ((struct __pyx_t_2yt_8geometry_18selection_routines_IntegrationAccumulator *)__pyx_v_data);

  /* "yt/geometry/selection_routines.pyx":1396
 *              void *data) nogil:
 *     cdef IntegrationAccumulator *am = <IntegrationAccumulator *> data
 *     cdef int di = (index[0]*vc.dims[1]+index[1])*vc.dims[2]+index[2]             # <<<<<<<<<<<<<<
 *     if am.child_mask[di] == 0 or enter_t == exit_t:
 *         return
 */
  __pyx_v_di = (((((__pyx_v_index[0]) * (__pyx_v_vc->dims[1])) + (__pyx_v_index[1])) * (__pyx_v_vc->dims[2])) + (__pyx_v_index[2]));

  /* "yt/geometry/selection_routines.pyx":1397
 *     cdef IntegrationAccumulator *am = <IntegrationAccumulator *> data
 *     cdef int di = (index[0]*vc.dims[1]+index[1])*vc.dims[2]+index[2]
 *     if am.child_mask[di] == 0 or enter_t == exit_t:             # <<<<<<<<<<<<<<
 *         return
 *     am.hits += 1
 */
  __pyx_t_2 = (((__pyx_v_am->child_mask[__pyx_v_di]) == 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_enter_t == __pyx_v_exit_t) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":1398
 *     cdef int di = (index[0]*vc.dims[1]+index[1])*vc.dims[2]+index[2]
 *     if am.child_mask[di] == 0 or enter_t == exit_t:
 *         return             # <<<<<<<<<<<<<<
 *     am.hits += 1
 *     am.t[di] = enter_t
 */
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":1397
 *     cdef IntegrationAccumulator *am = <IntegrationAccumulator *> data
 *     cdef int di = (index[0]*vc.dims[1]+index[1])*vc.dims[2]+index[2]
 *     if am.child_mask[di] == 0 or enter_t == exit_t:             # <<<<<<<<<<<<<<
 *         return
 *     am.hits += 1
 */
  }

  /* "yt/geometry/selection_routines.pyx":1399
 *     if am.child_mask[di] == 0 or enter_t == exit_t:
 *         return
 *     am.hits += 1             # <<<<<<<<<<<<<<
 *     am.t[di] = enter_t
 *     am.dt[di] = (exit_t - enter_t)
 */
  __pyx_v_am->hits = (__pyx_v_am->hits + 1);

  /* "yt/geometry/selection_routines.pyx":1400
 *         return
 *     am.hits += 1
 *     am.t[di] = enter_t             # <<<<<<<<<<<<<<
 *     am.dt[di] = (exit_t - enter_t)
 * 
 */
  (__pyx_v_am->t[__pyx_v_di]) = __pyx_v_enter_t;

  /* "yt/geometry/selection_routines.pyx":1401
 *     am.hits += 1
 *     am.t[di] = enter_t
 *     am.dt[di] = (exit_t - enter_t)             # <<<<<<<<<<<<<<
 * 
 * cdef class RaySelector(SelectorObject):
 */
  (__pyx_v_am->dt[__pyx_v_di]) = (__pyx_v_exit_t - __pyx_v_enter_t);

  /* "yt/geometry/selection_routines.pyx":1387
 *     int hits
 * 
 * cdef void dt_sampler(             # <<<<<<<<<<<<<<
 *              VolumeContainer *vc,
 *              np.float64_t v_pos[3],
 */

  /* function exit code */
  __pyx_L0:;
}

/* "yt/geometry/selection_routines.pyx":1409
 *     cdef np.float64_t vec[3]
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         _ensure_code(dobj.start_point)
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1409, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1409, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.RaySelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_11RaySelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_11RaySelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __pyx_t_5numpy_float64_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":1411
 *     def __init__(self, dobj):
 *         cdef int i
 *         _ensure_code(dobj.start_point)             # <<<<<<<<<<<<<<
 *         _ensure_code(dobj.end_point)
 *         for i in range(3):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_start_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":1412
 *         cdef int i
 *         _ensure_code(dobj.start_point)
 *         _ensure_code(dobj.end_point)             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             self.vec[i] = dobj.vec[i]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_end_point); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1413
 *         _ensure_code(dobj.start_point)
 *         _ensure_code(dobj.end_point)
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             self.vec[i] = dobj.vec[i]
 *             self.p1[i] = dobj.start_point[i]
 */
  for (__pyx_t_3 = 0; __pyx_t_3 < 3; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "yt/geometry/selection_routines.pyx":1414
 *         _ensure_code(dobj.end_point)
 *         for i in range(3):
 *             self.vec[i] = dobj.vec[i]             # <<<<<<<<<<<<<<
 *             self.p1[i] = dobj.start_point[i]
 *             self.p2[i] = dobj.end_point[i]
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_vec); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1414, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_self->vec[__pyx_v_i]) = __pyx_t_4;

    /* "yt/geometry/selection_routines.pyx":1415
 *         for i in range(3):
 *             self.vec[i] = dobj.vec[i]
 *             self.p1[i] = dobj.start_point[i]             # <<<<<<<<<<<<<<
 *             self.p2[i] = dobj.end_point[i]
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_start_point); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1415, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1415, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1415, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_self->p1[__pyx_v_i]) = __pyx_t_4;

    /* "yt/geometry/selection_routines.pyx":1416
 *             self.vec[i] = dobj.vec[i]
 *             self.p1[i] = dobj.start_point[i]
 *             self.p2[i] = dobj.end_point[i]             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_end_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1416, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_self->p2[__pyx_v_i]) = __pyx_t_4;
  }

  /* "yt/geometry/selection_routines.pyx":1409
 *     cdef np.float64_t vec[3]
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         _ensure_code(dobj.start_point)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("yt.geometry.selection_routines.RaySelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1421
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def fill_mask(self, gobj):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.float64_t, ndim=3] t, dt
 *         cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_3fill_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_gobj); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_3fill_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_gobj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fill_mask (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_2fill_mask(((struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *)__pyx_v_self), ((PyObject *)__pyx_v_gobj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_2fill_mask(struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, PyObject *__pyx_v_gobj) {
  PyArrayObject *__pyx_v_t = 0;
  PyArrayObject *__pyx_v_dt = 0;
  PyArrayObject *__pyx_v_child_mask = 0;
  int __pyx_v_i;
  int __pyx_v_total;
  struct __pyx_t_2yt_8geometry_18selection_routines_IntegrationAccumulator *__pyx_v_ia;
  struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer __pyx_v_vc;
  PyObject *__pyx_v_mask = NULL;
  npy_intp __pyx_v_j;
  npy_intp __pyx_v_k;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_child_mask;
  __Pyx_Buffer __pyx_pybuffer_child_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_dt;
  __Pyx_Buffer __pyx_pybuffer_dt;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_t;
  __Pyx_Buffer __pyx_pybuffer_t;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyArrayObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyArrayObject *__pyx_t_10 = NULL;
  __pyx_t_5numpy_float64_t __pyx_t_11;
  npy_intp __pyx_t_12;
  npy_intp __pyx_t_13;
  npy_intp __pyx_t_14;
  npy_intp __pyx_t_15;
  npy_intp __pyx_t_16;
  npy_intp __pyx_t_17;
  npy_intp __pyx_t_18;
  npy_intp __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  int __pyx_t_23;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fill_mask", 0);
  __pyx_pybuffer_t.pybuffer.buf = NULL;
  __pyx_pybuffer_t.refcount = 0;
  __pyx_pybuffernd_t.data = NULL;
  __pyx_pybuffernd_t.rcbuffer = &__pyx_pybuffer_t;
  __pyx_pybuffer_dt.pybuffer.buf = NULL;
  __pyx_pybuffer_dt.refcount = 0;
  __pyx_pybuffernd_dt.data = NULL;
  __pyx_pybuffernd_dt.rcbuffer = &__pyx_pybuffer_dt;
  __pyx_pybuffer_child_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_child_mask.refcount = 0;
  __pyx_pybuffernd_child_mask.data = NULL;
  __pyx_pybuffernd_child_mask.rcbuffer = &__pyx_pybuffer_child_mask;

  /* "yt/geometry/selection_routines.pyx":1425
 *         cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
 *         cdef int i
 *         cdef int total = 0             # <<<<<<<<<<<<<<
 *         cdef IntegrationAccumulator *ia
 *         ia = <IntegrationAccumulator *> malloc(sizeof(IntegrationAccumulator))
 */
  __pyx_v_total = 0;

  /* "yt/geometry/selection_routines.pyx":1427
 *         cdef int total = 0
 *         cdef IntegrationAccumulator *ia
 *         ia = <IntegrationAccumulator *> malloc(sizeof(IntegrationAccumulator))             # <<<<<<<<<<<<<<
 *         cdef VolumeContainer vc
 *         mask = np.zeros(gobj.ActiveDimensions, dtype='uint8')
 */
  __pyx_v_ia = ((struct __pyx_t_2yt_8geometry_18selection_routines_IntegrationAccumulator *)malloc((sizeof(struct __pyx_t_2yt_8geometry_18selection_routines_IntegrationAccumulator))));

  /* "yt/geometry/selection_routines.pyx":1429
 *         ia = <IntegrationAccumulator *> malloc(sizeof(IntegrationAccumulator))
 *         cdef VolumeContainer vc
 *         mask = np.zeros(gobj.ActiveDimensions, dtype='uint8')             # <<<<<<<<<<<<<<
 *         t = np.zeros(gobj.ActiveDimensions, dtype="float64")
 *         dt = np.zeros(gobj.ActiveDimensions, dtype="float64") - 1
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_ActiveDimensions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 1429, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mask = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":1430
 *         cdef VolumeContainer vc
 *         mask = np.zeros(gobj.ActiveDimensions, dtype='uint8')
 *         t = np.zeros(gobj.ActiveDimensions, dtype="float64")             # <<<<<<<<<<<<<<
 *         dt = np.zeros(gobj.ActiveDimensions, dtype="float64") - 1
 *         child_mask = gobj.child_mask
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_ActiveDimensions); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_n_s_float64) < 0) __PYX_ERR(0, 1430, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1430, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_t.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_t.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_t.rcbuffer->pybuffer, (PyObject*)__pyx_v_t, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
      __pyx_t_7 = __pyx_t_8 = __pyx_t_9 = 0;
    }
    __pyx_pybuffernd_t.diminfo[0].strides = __pyx_pybuffernd_t.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_t.diminfo[0].shape = __pyx_pybuffernd_t.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_t.diminfo[1].strides = __pyx_pybuffernd_t.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_t.diminfo[1].shape = __pyx_pybuffernd_t.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_t.diminfo[2].strides = __pyx_pybuffernd_t.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_t.diminfo[2].shape = __pyx_pybuffernd_t.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1430, __pyx_L1_error)
  }
  __pyx_t_5 = 0;
  __pyx_v_t = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":1431
 *         mask = np.zeros(gobj.ActiveDimensions, dtype='uint8')
 *         t = np.zeros(gobj.ActiveDimensions, dtype="float64")
 *         dt = np.zeros(gobj.ActiveDimensions, dtype="float64") - 1             # <<<<<<<<<<<<<<
 *         child_mask = gobj.child_mask
 *         ia.t = <np.float64_t *> t.data
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_ActiveDimensions); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_n_s_float64) < 0) __PYX_ERR(0, 1431, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1431, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dt.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dt.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dt.rcbuffer->pybuffer, (PyObject*)__pyx_v_dt, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_7);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      }
      __pyx_t_9 = __pyx_t_8 = __pyx_t_7 = 0;
    }
    __pyx_pybuffernd_dt.diminfo[0].strides = __pyx_pybuffernd_dt.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_dt.diminfo[0].shape = __pyx_pybuffernd_dt.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_dt.diminfo[1].strides = __pyx_pybuffernd_dt.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_dt.diminfo[1].shape = __pyx_pybuffernd_dt.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_dt.diminfo[2].strides = __pyx_pybuffernd_dt.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_dt.diminfo[2].shape = __pyx_pybuffernd_dt.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1431, __pyx_L1_error)
  }
  __pyx_t_5 = 0;
  __pyx_v_dt = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":1432
 *         t = np.zeros(gobj.ActiveDimensions, dtype="float64")
 *         dt = np.zeros(gobj.ActiveDimensions, dtype="float64") - 1
 *         child_mask = gobj.child_mask             # <<<<<<<<<<<<<<
 *         ia.t = <np.float64_t *> t.data
 *         ia.dt = <np.float64_t *> dt.data
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_child_mask); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1432, __pyx_L1_error)
  __pyx_t_10 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_10, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_child_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
      __pyx_t_7 = __pyx_t_8 = __pyx_t_9 = 0;
    }
    __pyx_pybuffernd_child_mask.diminfo[0].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_child_mask.diminfo[0].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_child_mask.diminfo[1].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_child_mask.diminfo[1].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_child_mask.diminfo[2].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_child_mask.diminfo[2].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1432, __pyx_L1_error)
  }
  __pyx_t_10 = 0;
  __pyx_v_child_mask = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":1433
 *         dt = np.zeros(gobj.ActiveDimensions, dtype="float64") - 1
 *         child_mask = gobj.child_mask
 *         ia.t = <np.float64_t *> t.data             # <<<<<<<<<<<<<<
 *         ia.dt = <np.float64_t *> dt.data
 *         ia.child_mask = <np.uint8_t *> child_mask.data
 */
  __pyx_v_ia->t = ((__pyx_t_5numpy_float64_t *)__pyx_v_t->data);

  /* "yt/geometry/selection_routines.pyx":1434
 *         child_mask = gobj.child_mask
 *         ia.t = <np.float64_t *> t.data
 *         ia.dt = <np.float64_t *> dt.data             # <<<<<<<<<<<<<<
 *         ia.child_mask = <np.uint8_t *> child_mask.data
 *         ia.hits = 0
 */
  __pyx_v_ia->dt = ((__pyx_t_5numpy_float64_t *)__pyx_v_dt->data);

  /* "yt/geometry/selection_routines.pyx":1435
 *         ia.t = <np.float64_t *> t.data
 *         ia.dt = <np.float64_t *> dt.data
 *         ia.child_mask = <np.uint8_t *> child_mask.data             # <<<<<<<<<<<<<<
 *         ia.hits = 0
 *         _ensure_code(gobj.LeftEdge)
 */
  __pyx_v_ia->child_mask = ((__pyx_t_5numpy_uint8_t *)__pyx_v_child_mask->data);

  /* "yt/geometry/selection_routines.pyx":1436
 *         ia.dt = <np.float64_t *> dt.data
 *         ia.child_mask = <np.uint8_t *> child_mask.data
 *         ia.hits = 0             # <<<<<<<<<<<<<<
 *         _ensure_code(gobj.LeftEdge)
 *         _ensure_code(gobj.RightEdge)
 */
  __pyx_v_ia->hits = 0;

  /* "yt/geometry/selection_routines.pyx":1437
 *         ia.child_mask = <np.uint8_t *> child_mask.data
 *         ia.hits = 0
 *         _ensure_code(gobj.LeftEdge)             # <<<<<<<<<<<<<<
 *         _ensure_code(gobj.RightEdge)
 *         _ensure_code(gobj.dds)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_LeftEdge); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1438
 *         ia.hits = 0
 *         _ensure_code(gobj.LeftEdge)
 *         _ensure_code(gobj.RightEdge)             # <<<<<<<<<<<<<<
 *         _ensure_code(gobj.dds)
 *         for i in range(3):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_RightEdge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":1439
 *         _ensure_code(gobj.LeftEdge)
 *         _ensure_code(gobj.RightEdge)
 *         _ensure_code(gobj.dds)             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             vc.left_edge[i] = gobj.LeftEdge[i]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_dds); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1440
 *         _ensure_code(gobj.RightEdge)
 *         _ensure_code(gobj.dds)
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             vc.left_edge[i] = gobj.LeftEdge[i]
 *             vc.right_edge[i] = gobj.RightEdge[i]
 */
  for (__pyx_t_6 = 0; __pyx_t_6 < 3; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "yt/geometry/selection_routines.pyx":1441
 *         _ensure_code(gobj.dds)
 *         for i in range(3):
 *             vc.left_edge[i] = gobj.LeftEdge[i]             # <<<<<<<<<<<<<<
 *             vc.right_edge[i] = gobj.RightEdge[i]
 *             vc.dds[i] = gobj.dds[i]
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_LeftEdge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_11 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1441, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_vc.left_edge[__pyx_v_i]) = __pyx_t_11;

    /* "yt/geometry/selection_routines.pyx":1442
 *         for i in range(3):
 *             vc.left_edge[i] = gobj.LeftEdge[i]
 *             vc.right_edge[i] = gobj.RightEdge[i]             # <<<<<<<<<<<<<<
 *             vc.dds[i] = gobj.dds[i]
 *             vc.idds[i] = 1.0/gobj.dds[i]
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_RightEdge); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_11 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1442, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_vc.right_edge[__pyx_v_i]) = __pyx_t_11;

    /* "yt/geometry/selection_routines.pyx":1443
 *             vc.left_edge[i] = gobj.LeftEdge[i]
 *             vc.right_edge[i] = gobj.RightEdge[i]
 *             vc.dds[i] = gobj.dds[i]             # <<<<<<<<<<<<<<
 *             vc.idds[i] = 1.0/gobj.dds[i]
 *             vc.dims[i] = dt.shape[i]
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_dds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1443, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1443, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_11 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1443, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_vc.dds[__pyx_v_i]) = __pyx_t_11;

    /* "yt/geometry/selection_routines.pyx":1444
 *             vc.right_edge[i] = gobj.RightEdge[i]
 *             vc.dds[i] = gobj.dds[i]
 *             vc.idds[i] = 1.0/gobj.dds[i]             # <<<<<<<<<<<<<<
 *             vc.dims[i] = dt.shape[i]
 *         walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_dds); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyFloat_DivideCObj(__pyx_float_1_0, __pyx_t_1, 1.0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_11 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1444, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_vc.idds[__pyx_v_i]) = __pyx_t_11;

    /* "yt/geometry/selection_routines.pyx":1445
 *             vc.dds[i] = gobj.dds[i]
 *             vc.idds[i] = 1.0/gobj.dds[i]
 *             vc.dims[i] = dt.shape[i]             # <<<<<<<<<<<<<<
 *         walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 *         for i in range(dt.shape[0]):
 */
    (__pyx_v_vc.dims[__pyx_v_i]) = (__pyx_v_dt->dimensions[__pyx_v_i]);
  }

  /* "yt/geometry/selection_routines.pyx":1446
 *             vc.idds[i] = 1.0/gobj.dds[i]
 *             vc.dims[i] = dt.shape[i]
 *         walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)             # <<<<<<<<<<<<<<
 *         for i in range(dt.shape[0]):
 *             for j in range(dt.shape[1]):
 */
  (void)(__pyx_f_2yt_9utilities_3lib_14grid_traversal_walk_volume((&__pyx_v_vc), __pyx_v_self->p1, __pyx_v_self->vec, __pyx_f_2yt_8geometry_18selection_routines_dt_sampler, ((void *)__pyx_v_ia), NULL));

  /* "yt/geometry/selection_routines.pyx":1447
 *             vc.dims[i] = dt.shape[i]
 *         walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 *         for i in range(dt.shape[0]):             # <<<<<<<<<<<<<<
 *             for j in range(dt.shape[1]):
 *                 for k in range(dt.shape[2]):
 */
  __pyx_t_12 = (__pyx_v_dt->dimensions[0]);
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_13; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "yt/geometry/selection_routines.pyx":1448
 *         walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 *         for i in range(dt.shape[0]):
 *             for j in range(dt.shape[1]):             # <<<<<<<<<<<<<<
 *                 for k in range(dt.shape[2]):
 *                     if dt[i, j, k] >= 0:
 */
    __pyx_t_14 = (__pyx_v_dt->dimensions[1]);
    __pyx_t_15 = __pyx_t_14;
    for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
      __pyx_v_j = __pyx_t_16;

      /* "yt/geometry/selection_routines.pyx":1449
 *         for i in range(dt.shape[0]):
 *             for j in range(dt.shape[1]):
 *                 for k in range(dt.shape[2]):             # <<<<<<<<<<<<<<
 *                     if dt[i, j, k] >= 0:
 *                         mask[i, j, k] = 1
 */
      __pyx_t_17 = (__pyx_v_dt->dimensions[2]);
      __pyx_t_18 = __pyx_t_17;
      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
        __pyx_v_k = __pyx_t_19;

        /* "yt/geometry/selection_routines.pyx":1450
 *             for j in range(dt.shape[1]):
 *                 for k in range(dt.shape[2]):
 *                     if dt[i, j, k] >= 0:             # <<<<<<<<<<<<<<
 *                         mask[i, j, k] = 1
 *                         total += 1
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_21 = __pyx_v_j;
        __pyx_t_22 = __pyx_v_k;
        __pyx_t_23 = (((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_dt.rcbuffer->pybuffer.buf, __pyx_t_20, __pyx_pybuffernd_dt.diminfo[0].strides, __pyx_t_21, __pyx_pybuffernd_dt.diminfo[1].strides, __pyx_t_22, __pyx_pybuffernd_dt.diminfo[2].strides)) >= 0.0) != 0);
        if (__pyx_t_23) {

          /* "yt/geometry/selection_routines.pyx":1451
 *                 for k in range(dt.shape[2]):
 *                     if dt[i, j, k] >= 0:
 *                         mask[i, j, k] = 1             # <<<<<<<<<<<<<<
 *                         total += 1
 *         free(ia)
 */
          __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1451, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_j); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1451, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyInt_From_Py_intptr_t(__pyx_v_k); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1451, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1451, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
          __pyx_t_2 = 0;
          __pyx_t_1 = 0;
          __pyx_t_3 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_mask, __pyx_t_4, __pyx_int_1) < 0)) __PYX_ERR(0, 1451, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "yt/geometry/selection_routines.pyx":1452
 *                     if dt[i, j, k] >= 0:
 *                         mask[i, j, k] = 1
 *                         total += 1             # <<<<<<<<<<<<<<
 *         free(ia)
 *         if total == 0: return None
 */
          __pyx_v_total = (__pyx_v_total + 1);

          /* "yt/geometry/selection_routines.pyx":1450
 *             for j in range(dt.shape[1]):
 *                 for k in range(dt.shape[2]):
 *                     if dt[i, j, k] >= 0:             # <<<<<<<<<<<<<<
 *                         mask[i, j, k] = 1
 *                         total += 1
 */
        }
      }
    }
  }

  /* "yt/geometry/selection_routines.pyx":1453
 *                         mask[i, j, k] = 1
 *                         total += 1
 *         free(ia)             # <<<<<<<<<<<<<<
 *         if total == 0: return None
 *         return mask.astype("bool")
 */
  free(__pyx_v_ia);

  /* "yt/geometry/selection_routines.pyx":1454
 *                         total += 1
 *         free(ia)
 *         if total == 0: return None             # <<<<<<<<<<<<<<
 *         return mask.astype("bool")
 * 
 */
  __pyx_t_23 = ((__pyx_v_total == 0) != 0);
  if (__pyx_t_23) {
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":1455
 *         free(ia)
 *         if total == 0: return None
 *         return mask.astype("bool")             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_mask, __pyx_n_s_astype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_n_s_bool) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_s_bool);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1421
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def fill_mask(self, gobj):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.float64_t, ndim=3] t, dt
 *         cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dt.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_t.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.RaySelector.fill_mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dt.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_t.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_t);
  __Pyx_XDECREF((PyObject *)__pyx_v_dt);
  __Pyx_XDECREF((PyObject *)__pyx_v_child_mask);
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1460
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def get_dt(self, gobj):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.float64_t, ndim=3] t, dt
 *         cdef np.ndarray[np.float64_t, ndim=1] tr, dtr
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_5get_dt(PyObject *__pyx_v_self, PyObject *__pyx_v_gobj); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_5get_dt(PyObject *__pyx_v_self, PyObject *__pyx_v_gobj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_dt (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_4get_dt(((struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *)__pyx_v_self), ((PyObject *)__pyx_v_gobj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_4get_dt(struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, PyObject *__pyx_v_gobj) {
  PyArrayObject *__pyx_v_t = 0;
  PyArrayObject *__pyx_v_dt = 0;
  PyArrayObject *__pyx_v_tr = 0;
  PyArrayObject *__pyx_v_dtr = 0;
  PyArrayObject *__pyx_v_child_mask = 0;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_ni;
  struct __pyx_t_2yt_8geometry_18selection_routines_IntegrationAccumulator *__pyx_v_ia;
  struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer __pyx_v_vc;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_child_mask;
  __Pyx_Buffer __pyx_pybuffer_child_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_dt;
  __Pyx_Buffer __pyx_pybuffer_dt;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_dtr;
  __Pyx_Buffer __pyx_pybuffer_dtr;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_t;
  __Pyx_Buffer __pyx_pybuffer_t;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_tr;
  __Pyx_Buffer __pyx_pybuffer_tr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyArrayObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyArrayObject *__pyx_t_10 = NULL;
  __pyx_t_5numpy_float64_t __pyx_t_11;
  PyArrayObject *__pyx_t_12 = NULL;
  npy_intp __pyx_t_13;
  npy_intp __pyx_t_14;
  npy_intp __pyx_t_15;
  npy_intp __pyx_t_16;
  int __pyx_t_17;
  npy_intp __pyx_t_18;
  npy_intp __pyx_t_19;
  int __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  int __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_dt", 0);
  __pyx_pybuffer_t.pybuffer.buf = NULL;
  __pyx_pybuffer_t.refcount = 0;
  __pyx_pybuffernd_t.data = NULL;
  __pyx_pybuffernd_t.rcbuffer = &__pyx_pybuffer_t;
  __pyx_pybuffer_dt.pybuffer.buf = NULL;
  __pyx_pybuffer_dt.refcount = 0;
  __pyx_pybuffernd_dt.data = NULL;
  __pyx_pybuffernd_dt.rcbuffer = &__pyx_pybuffer_dt;
  __pyx_pybuffer_tr.pybuffer.buf = NULL;
  __pyx_pybuffer_tr.refcount = 0;
  __pyx_pybuffernd_tr.data = NULL;
  __pyx_pybuffernd_tr.rcbuffer = &__pyx_pybuffer_tr;
  __pyx_pybuffer_dtr.pybuffer.buf = NULL;
  __pyx_pybuffer_dtr.refcount = 0;
  __pyx_pybuffernd_dtr.data = NULL;
  __pyx_pybuffernd_dtr.rcbuffer = &__pyx_pybuffer_dtr;
  __pyx_pybuffer_child_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_child_mask.refcount = 0;
  __pyx_pybuffernd_child_mask.data = NULL;
  __pyx_pybuffernd_child_mask.rcbuffer = &__pyx_pybuffer_child_mask;

  /* "yt/geometry/selection_routines.pyx":1466
 *         cdef int i, j, k, ni
 *         cdef IntegrationAccumulator *ia
 *         ia = <IntegrationAccumulator *> malloc(sizeof(IntegrationAccumulator))             # <<<<<<<<<<<<<<
 *         cdef VolumeContainer vc
 *         t = np.zeros(gobj.ActiveDimensions, dtype="float64")
 */
  __pyx_v_ia = ((struct __pyx_t_2yt_8geometry_18selection_routines_IntegrationAccumulator *)malloc((sizeof(struct __pyx_t_2yt_8geometry_18selection_routines_IntegrationAccumulator))));

  /* "yt/geometry/selection_routines.pyx":1468
 *         ia = <IntegrationAccumulator *> malloc(sizeof(IntegrationAccumulator))
 *         cdef VolumeContainer vc
 *         t = np.zeros(gobj.ActiveDimensions, dtype="float64")             # <<<<<<<<<<<<<<
 *         dt = np.zeros(gobj.ActiveDimensions, dtype="float64") - 1
 *         child_mask = gobj.child_mask
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_ActiveDimensions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_float64) < 0) __PYX_ERR(0, 1468, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1468, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_t.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_t.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_t.rcbuffer->pybuffer, (PyObject*)__pyx_v_t, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
      __pyx_t_7 = __pyx_t_8 = __pyx_t_9 = 0;
    }
    __pyx_pybuffernd_t.diminfo[0].strides = __pyx_pybuffernd_t.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_t.diminfo[0].shape = __pyx_pybuffernd_t.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_t.diminfo[1].strides = __pyx_pybuffernd_t.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_t.diminfo[1].shape = __pyx_pybuffernd_t.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_t.diminfo[2].strides = __pyx_pybuffernd_t.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_t.diminfo[2].shape = __pyx_pybuffernd_t.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1468, __pyx_L1_error)
  }
  __pyx_t_5 = 0;
  __pyx_v_t = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":1469
 *         cdef VolumeContainer vc
 *         t = np.zeros(gobj.ActiveDimensions, dtype="float64")
 *         dt = np.zeros(gobj.ActiveDimensions, dtype="float64") - 1             # <<<<<<<<<<<<<<
 *         child_mask = gobj.child_mask
 *         ia.t = <np.float64_t *> t.data
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_ActiveDimensions); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_n_s_float64) < 0) __PYX_ERR(0, 1469, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1469, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dt.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dt.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dt.rcbuffer->pybuffer, (PyObject*)__pyx_v_dt, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_7);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      }
      __pyx_t_9 = __pyx_t_8 = __pyx_t_7 = 0;
    }
    __pyx_pybuffernd_dt.diminfo[0].strides = __pyx_pybuffernd_dt.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_dt.diminfo[0].shape = __pyx_pybuffernd_dt.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_dt.diminfo[1].strides = __pyx_pybuffernd_dt.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_dt.diminfo[1].shape = __pyx_pybuffernd_dt.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_dt.diminfo[2].strides = __pyx_pybuffernd_dt.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_dt.diminfo[2].shape = __pyx_pybuffernd_dt.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1469, __pyx_L1_error)
  }
  __pyx_t_5 = 0;
  __pyx_v_dt = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":1470
 *         t = np.zeros(gobj.ActiveDimensions, dtype="float64")
 *         dt = np.zeros(gobj.ActiveDimensions, dtype="float64") - 1
 *         child_mask = gobj.child_mask             # <<<<<<<<<<<<<<
 *         ia.t = <np.float64_t *> t.data
 *         ia.dt = <np.float64_t *> dt.data
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_child_mask); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1470, __pyx_L1_error)
  __pyx_t_10 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_10, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_child_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
      __pyx_t_7 = __pyx_t_8 = __pyx_t_9 = 0;
    }
    __pyx_pybuffernd_child_mask.diminfo[0].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_child_mask.diminfo[0].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_child_mask.diminfo[1].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_child_mask.diminfo[1].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_child_mask.diminfo[2].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_child_mask.diminfo[2].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1470, __pyx_L1_error)
  }
  __pyx_t_10 = 0;
  __pyx_v_child_mask = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":1471
 *         dt = np.zeros(gobj.ActiveDimensions, dtype="float64") - 1
 *         child_mask = gobj.child_mask
 *         ia.t = <np.float64_t *> t.data             # <<<<<<<<<<<<<<
 *         ia.dt = <np.float64_t *> dt.data
 *         ia.child_mask = <np.uint8_t *> child_mask.data
 */
  __pyx_v_ia->t = ((__pyx_t_5numpy_float64_t *)__pyx_v_t->data);

  /* "yt/geometry/selection_routines.pyx":1472
 *         child_mask = gobj.child_mask
 *         ia.t = <np.float64_t *> t.data
 *         ia.dt = <np.float64_t *> dt.data             # <<<<<<<<<<<<<<
 *         ia.child_mask = <np.uint8_t *> child_mask.data
 *         ia.hits = 0
 */
  __pyx_v_ia->dt = ((__pyx_t_5numpy_float64_t *)__pyx_v_dt->data);

  /* "yt/geometry/selection_routines.pyx":1473
 *         ia.t = <np.float64_t *> t.data
 *         ia.dt = <np.float64_t *> dt.data
 *         ia.child_mask = <np.uint8_t *> child_mask.data             # <<<<<<<<<<<<<<
 *         ia.hits = 0
 *         _ensure_code(gobj.LeftEdge)
 */
  __pyx_v_ia->child_mask = ((__pyx_t_5numpy_uint8_t *)__pyx_v_child_mask->data);

  /* "yt/geometry/selection_routines.pyx":1474
 *         ia.dt = <np.float64_t *> dt.data
 *         ia.child_mask = <np.uint8_t *> child_mask.data
 *         ia.hits = 0             # <<<<<<<<<<<<<<
 *         _ensure_code(gobj.LeftEdge)
 *         _ensure_code(gobj.RightEdge)
 */
  __pyx_v_ia->hits = 0;

  /* "yt/geometry/selection_routines.pyx":1475
 *         ia.child_mask = <np.uint8_t *> child_mask.data
 *         ia.hits = 0
 *         _ensure_code(gobj.LeftEdge)             # <<<<<<<<<<<<<<
 *         _ensure_code(gobj.RightEdge)
 *         _ensure_code(gobj.dds)
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_LeftEdge); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":1476
 *         ia.hits = 0
 *         _ensure_code(gobj.LeftEdge)
 *         _ensure_code(gobj.RightEdge)             # <<<<<<<<<<<<<<
 *         _ensure_code(gobj.dds)
 *         for i in range(3):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_RightEdge); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":1477
 *         _ensure_code(gobj.LeftEdge)
 *         _ensure_code(gobj.RightEdge)
 *         _ensure_code(gobj.dds)             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             vc.left_edge[i] = gobj.LeftEdge[i]
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_dds); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":1478
 *         _ensure_code(gobj.RightEdge)
 *         _ensure_code(gobj.dds)
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             vc.left_edge[i] = gobj.LeftEdge[i]
 *             vc.right_edge[i] = gobj.RightEdge[i]
 */
  for (__pyx_t_6 = 0; __pyx_t_6 < 3; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "yt/geometry/selection_routines.pyx":1479
 *         _ensure_code(gobj.dds)
 *         for i in range(3):
 *             vc.left_edge[i] = gobj.LeftEdge[i]             # <<<<<<<<<<<<<<
 *             vc.right_edge[i] = gobj.RightEdge[i]
 *             vc.dds[i] = gobj.dds[i]
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_LeftEdge); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1479, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1479, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_11 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1479, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    (__pyx_v_vc.left_edge[__pyx_v_i]) = __pyx_t_11;

    /* "yt/geometry/selection_routines.pyx":1480
 *         for i in range(3):
 *             vc.left_edge[i] = gobj.LeftEdge[i]
 *             vc.right_edge[i] = gobj.RightEdge[i]             # <<<<<<<<<<<<<<
 *             vc.dds[i] = gobj.dds[i]
 *             vc.idds[i] = 1.0/gobj.dds[i]
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_RightEdge); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_4, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_11 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1480, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_vc.right_edge[__pyx_v_i]) = __pyx_t_11;

    /* "yt/geometry/selection_routines.pyx":1481
 *             vc.left_edge[i] = gobj.LeftEdge[i]
 *             vc.right_edge[i] = gobj.RightEdge[i]
 *             vc.dds[i] = gobj.dds[i]             # <<<<<<<<<<<<<<
 *             vc.idds[i] = 1.0/gobj.dds[i]
 *             vc.dims[i] = dt.shape[i]
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_dds); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_11 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1481, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    (__pyx_v_vc.dds[__pyx_v_i]) = __pyx_t_11;

    /* "yt/geometry/selection_routines.pyx":1482
 *             vc.right_edge[i] = gobj.RightEdge[i]
 *             vc.dds[i] = gobj.dds[i]
 *             vc.idds[i] = 1.0/gobj.dds[i]             # <<<<<<<<<<<<<<
 *             vc.dims[i] = dt.shape[i]
 *         walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_dds); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1482, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_4, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1482, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyFloat_DivideCObj(__pyx_float_1_0, __pyx_t_2, 1.0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1482, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_11 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1482, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    (__pyx_v_vc.idds[__pyx_v_i]) = __pyx_t_11;

    /* "yt/geometry/selection_routines.pyx":1483
 *             vc.dds[i] = gobj.dds[i]
 *             vc.idds[i] = 1.0/gobj.dds[i]
 *             vc.dims[i] = dt.shape[i]             # <<<<<<<<<<<<<<
 *         walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 *         tr = np.zeros(ia.hits, dtype="float64")
 */
    (__pyx_v_vc.dims[__pyx_v_i]) = (__pyx_v_dt->dimensions[__pyx_v_i]);
  }

  /* "yt/geometry/selection_routines.pyx":1484
 *             vc.idds[i] = 1.0/gobj.dds[i]
 *             vc.dims[i] = dt.shape[i]
 *         walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)             # <<<<<<<<<<<<<<
 *         tr = np.zeros(ia.hits, dtype="float64")
 *         dtr = np.zeros(ia.hits, dtype="float64")
 */
  (void)(__pyx_f_2yt_9utilities_3lib_14grid_traversal_walk_volume((&__pyx_v_vc), __pyx_v_self->p1, __pyx_v_self->vec, __pyx_f_2yt_8geometry_18selection_routines_dt_sampler, ((void *)__pyx_v_ia), NULL));

  /* "yt/geometry/selection_routines.pyx":1485
 *             vc.dims[i] = dt.shape[i]
 *         walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 *         tr = np.zeros(ia.hits, dtype="float64")             # <<<<<<<<<<<<<<
 *         dtr = np.zeros(ia.hits, dtype="float64")
 *         ni = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_ia->hits); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_n_s_float64) < 0) __PYX_ERR(0, 1485, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1485, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tr.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tr.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tr.rcbuffer->pybuffer, (PyObject*)__pyx_v_tr, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_7);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      }
      __pyx_t_9 = __pyx_t_8 = __pyx_t_7 = 0;
    }
    __pyx_pybuffernd_tr.diminfo[0].strides = __pyx_pybuffernd_tr.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_tr.diminfo[0].shape = __pyx_pybuffernd_tr.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1485, __pyx_L1_error)
  }
  __pyx_t_12 = 0;
  __pyx_v_tr = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1486
 *         walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 *         tr = np.zeros(ia.hits, dtype="float64")
 *         dtr = np.zeros(ia.hits, dtype="float64")             # <<<<<<<<<<<<<<
 *         ni = 0
 *         for i in range(dt.shape[0]):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_ia->hits); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_float64) < 0) __PYX_ERR(0, 1486, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1486, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dtr.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dtr.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dtr.rcbuffer->pybuffer, (PyObject*)__pyx_v_dtr, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
      __pyx_t_7 = __pyx_t_8 = __pyx_t_9 = 0;
    }
    __pyx_pybuffernd_dtr.diminfo[0].strides = __pyx_pybuffernd_dtr.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_dtr.diminfo[0].shape = __pyx_pybuffernd_dtr.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1486, __pyx_L1_error)
  }
  __pyx_t_12 = 0;
  __pyx_v_dtr = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":1487
 *         tr = np.zeros(ia.hits, dtype="float64")
 *         dtr = np.zeros(ia.hits, dtype="float64")
 *         ni = 0             # <<<<<<<<<<<<<<
 *         for i in range(dt.shape[0]):
 *             for j in range(dt.shape[1]):
 */
  __pyx_v_ni = 0;

  /* "yt/geometry/selection_routines.pyx":1488
 *         dtr = np.zeros(ia.hits, dtype="float64")
 *         ni = 0
 *         for i in range(dt.shape[0]):             # <<<<<<<<<<<<<<
 *             for j in range(dt.shape[1]):
 *                 for k in range(dt.shape[2]):
 */
  __pyx_t_13 = (__pyx_v_dt->dimensions[0]);
  __pyx_t_14 = __pyx_t_13;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_14; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "yt/geometry/selection_routines.pyx":1489
 *         ni = 0
 *         for i in range(dt.shape[0]):
 *             for j in range(dt.shape[1]):             # <<<<<<<<<<<<<<
 *                 for k in range(dt.shape[2]):
 *                     if dt[i, j, k] >= 0:
 */
    __pyx_t_15 = (__pyx_v_dt->dimensions[1]);
    __pyx_t_16 = __pyx_t_15;
    for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
      __pyx_v_j = __pyx_t_17;

      /* "yt/geometry/selection_routines.pyx":1490
 *         for i in range(dt.shape[0]):
 *             for j in range(dt.shape[1]):
 *                 for k in range(dt.shape[2]):             # <<<<<<<<<<<<<<
 *                     if dt[i, j, k] >= 0:
 *                         tr[ni] = t[i, j, k]
 */
      __pyx_t_18 = (__pyx_v_dt->dimensions[2]);
      __pyx_t_19 = __pyx_t_18;
      for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
        __pyx_v_k = __pyx_t_20;

        /* "yt/geometry/selection_routines.pyx":1491
 *             for j in range(dt.shape[1]):
 *                 for k in range(dt.shape[2]):
 *                     if dt[i, j, k] >= 0:             # <<<<<<<<<<<<<<
 *                         tr[ni] = t[i, j, k]
 *                         dtr[ni] = dt[i, j, k]
 */
        __pyx_t_21 = __pyx_v_i;
        __pyx_t_22 = __pyx_v_j;
        __pyx_t_23 = __pyx_v_k;
        __pyx_t_24 = (((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_dt.rcbuffer->pybuffer.buf, __pyx_t_21, __pyx_pybuffernd_dt.diminfo[0].strides, __pyx_t_22, __pyx_pybuffernd_dt.diminfo[1].strides, __pyx_t_23, __pyx_pybuffernd_dt.diminfo[2].strides)) >= 0.0) != 0);
        if (__pyx_t_24) {

          /* "yt/geometry/selection_routines.pyx":1492
 *                 for k in range(dt.shape[2]):
 *                     if dt[i, j, k] >= 0:
 *                         tr[ni] = t[i, j, k]             # <<<<<<<<<<<<<<
 *                         dtr[ni] = dt[i, j, k]
 *                         ni += 1
 */
          __pyx_t_23 = __pyx_v_i;
          __pyx_t_22 = __pyx_v_j;
          __pyx_t_21 = __pyx_v_k;
          __pyx_t_25 = __pyx_v_ni;
          *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tr.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_tr.diminfo[0].strides) = (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_t.rcbuffer->pybuffer.buf, __pyx_t_23, __pyx_pybuffernd_t.diminfo[0].strides, __pyx_t_22, __pyx_pybuffernd_t.diminfo[1].strides, __pyx_t_21, __pyx_pybuffernd_t.diminfo[2].strides));

          /* "yt/geometry/selection_routines.pyx":1493
 *                     if dt[i, j, k] >= 0:
 *                         tr[ni] = t[i, j, k]
 *                         dtr[ni] = dt[i, j, k]             # <<<<<<<<<<<<<<
 *                         ni += 1
 *         if not (ni == ia.hits):
 */
          __pyx_t_21 = __pyx_v_i;
          __pyx_t_22 = __pyx_v_j;
          __pyx_t_23 = __pyx_v_k;
          __pyx_t_26 = __pyx_v_ni;
          *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_dtr.rcbuffer->pybuffer.buf, __pyx_t_26, __pyx_pybuffernd_dtr.diminfo[0].strides) = (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_dt.rcbuffer->pybuffer.buf, __pyx_t_21, __pyx_pybuffernd_dt.diminfo[0].strides, __pyx_t_22, __pyx_pybuffernd_dt.diminfo[1].strides, __pyx_t_23, __pyx_pybuffernd_dt.diminfo[2].strides));

          /* "yt/geometry/selection_routines.pyx":1494
 *                         tr[ni] = t[i, j, k]
 *                         dtr[ni] = dt[i, j, k]
 *                         ni += 1             # <<<<<<<<<<<<<<
 *         if not (ni == ia.hits):
 *             print ni, ia.hits
 */
          __pyx_v_ni = (__pyx_v_ni + 1);

          /* "yt/geometry/selection_routines.pyx":1491
 *             for j in range(dt.shape[1]):
 *                 for k in range(dt.shape[2]):
 *                     if dt[i, j, k] >= 0:             # <<<<<<<<<<<<<<
 *                         tr[ni] = t[i, j, k]
 *                         dtr[ni] = dt[i, j, k]
 */
        }
      }
    }
  }

  /* "yt/geometry/selection_routines.pyx":1495
 *                         dtr[ni] = dt[i, j, k]
 *                         ni += 1
 *         if not (ni == ia.hits):             # <<<<<<<<<<<<<<
 *             print ni, ia.hits
 *         free(ia)
 */
  __pyx_t_24 = ((!((__pyx_v_ni == __pyx_v_ia->hits) != 0)) != 0);
  if (__pyx_t_24) {

    /* "yt/geometry/selection_routines.pyx":1496
 *                         ni += 1
 *         if not (ni == ia.hits):
 *             print ni, ia.hits             # <<<<<<<<<<<<<<
 *         free(ia)
 *         return dtr, tr
 */
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_ni); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1496, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_ia->hits); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1496, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1496, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_1 = 0;
    if (__Pyx_Print(0, __pyx_t_3, 1) < 0) __PYX_ERR(0, 1496, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "yt/geometry/selection_routines.pyx":1495
 *                         dtr[ni] = dt[i, j, k]
 *                         ni += 1
 *         if not (ni == ia.hits):             # <<<<<<<<<<<<<<
 *             print ni, ia.hits
 *         free(ia)
 */
  }

  /* "yt/geometry/selection_routines.pyx":1497
 *         if not (ni == ia.hits):
 *             print ni, ia.hits
 *         free(ia)             # <<<<<<<<<<<<<<
 *         return dtr, tr
 * 
 */
  free(__pyx_v_ia);

  /* "yt/geometry/selection_routines.pyx":1498
 *             print ni, ia.hits
 *         free(ia)
 *         return dtr, tr             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_dtr));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_dtr));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_dtr));
  __Pyx_INCREF(((PyObject *)__pyx_v_tr));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_tr));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_tr));
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1460
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def get_dt(self, gobj):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.float64_t, ndim=3] t, dt
 *         cdef np.ndarray[np.float64_t, ndim=1] tr, dtr
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dt.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dtr.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_t.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tr.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.RaySelector.get_dt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dt.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dtr.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_t.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tr.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_t);
  __Pyx_XDECREF((PyObject *)__pyx_v_dt);
  __Pyx_XDECREF((PyObject *)__pyx_v_tr);
  __Pyx_XDECREF((PyObject *)__pyx_v_dtr);
  __Pyx_XDECREF((PyObject *)__pyx_v_child_mask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1503
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def get_dt_mesh(self, mesh, nz, int offset):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.float64_t, ndim=3] t, dt
 *         cdef np.ndarray[np.float64_t, ndim=1] tr, dtr
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_7get_dt_mesh(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_7get_dt_mesh(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_mesh = 0;
  PyObject *__pyx_v_nz = 0;
  int __pyx_v_offset;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_dt_mesh (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mesh,&__pyx_n_s_nz,&__pyx_n_s_offset,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mesh)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_dt_mesh", 1, 3, 3, 1); __PYX_ERR(0, 1503, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_dt_mesh", 1, 3, 3, 2); __PYX_ERR(0, 1503, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_dt_mesh") < 0)) __PYX_ERR(0, 1503, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_mesh = values[0];
    __pyx_v_nz = values[1];
    __pyx_v_offset = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_offset == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1503, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_dt_mesh", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1503, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.RaySelector.get_dt_mesh", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_6get_dt_mesh(((struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *)__pyx_v_self), __pyx_v_mesh, __pyx_v_nz, __pyx_v_offset);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_6get_dt_mesh(struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, PyObject *__pyx_v_mesh, PyObject *__pyx_v_nz, int __pyx_v_offset) {
  PyArrayObject *__pyx_v_t = 0;
  PyArrayObject *__pyx_v_dt = 0;
  PyArrayObject *__pyx_v_tr = 0;
  PyArrayObject *__pyx_v_dtr = 0;
  PyArrayObject *__pyx_v_child_mask = 0;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_ni;
  __pyx_t_5numpy_float64_t __pyx_v_LE[3];
  __pyx_t_5numpy_float64_t __pyx_v_RE[3];
  __pyx_t_5numpy_float64_t __pyx_v_pos;
  struct __pyx_t_2yt_8geometry_18selection_routines_IntegrationAccumulator *__pyx_v_ia;
  PyArrayObject *__pyx_v_coords = 0;
  PyArrayObject *__pyx_v_indices = 0;
  int __pyx_v_nc;
  int __pyx_v_nv;
  struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer __pyx_v_vc;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_child_mask;
  __Pyx_Buffer __pyx_pybuffer_child_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_coords;
  __Pyx_Buffer __pyx_pybuffer_coords;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_dt;
  __Pyx_Buffer __pyx_pybuffer_dt;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_dtr;
  __Pyx_Buffer __pyx_pybuffer_dtr;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_t;
  __Pyx_Buffer __pyx_pybuffer_t;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_tr;
  __Pyx_Buffer __pyx_pybuffer_tr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyArrayObject *__pyx_t_11 = NULL;
  PyArrayObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyArrayObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  __pyx_t_5numpy_int64_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  Py_ssize_t __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_dt_mesh", 0);
  __pyx_pybuffer_t.pybuffer.buf = NULL;
  __pyx_pybuffer_t.refcount = 0;
  __pyx_pybuffernd_t.data = NULL;
  __pyx_pybuffernd_t.rcbuffer = &__pyx_pybuffer_t;
  __pyx_pybuffer_dt.pybuffer.buf = NULL;
  __pyx_pybuffer_dt.refcount = 0;
  __pyx_pybuffernd_dt.data = NULL;
  __pyx_pybuffernd_dt.rcbuffer = &__pyx_pybuffer_dt;
  __pyx_pybuffer_tr.pybuffer.buf = NULL;
  __pyx_pybuffer_tr.refcount = 0;
  __pyx_pybuffernd_tr.data = NULL;
  __pyx_pybuffernd_tr.rcbuffer = &__pyx_pybuffer_tr;
  __pyx_pybuffer_dtr.pybuffer.buf = NULL;
  __pyx_pybuffer_dtr.refcount = 0;
  __pyx_pybuffernd_dtr.data = NULL;
  __pyx_pybuffernd_dtr.rcbuffer = &__pyx_pybuffer_dtr;
  __pyx_pybuffer_child_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_child_mask.refcount = 0;
  __pyx_pybuffernd_child_mask.data = NULL;
  __pyx_pybuffernd_child_mask.rcbuffer = &__pyx_pybuffer_child_mask;
  __pyx_pybuffer_coords.pybuffer.buf = NULL;
  __pyx_pybuffer_coords.refcount = 0;
  __pyx_pybuffernd_coords.data = NULL;
  __pyx_pybuffernd_coords.rcbuffer = &__pyx_pybuffer_coords;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;

  /* "yt/geometry/selection_routines.pyx":1512
 *         cdef np.float64_t pos
 *         cdef IntegrationAccumulator *ia
 *         ia = <IntegrationAccumulator *> malloc(sizeof(IntegrationAccumulator))             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.float64_t, ndim=2] coords
 *         cdef np.ndarray[np.int64_t, ndim=2] indices
 */
  __pyx_v_ia = ((struct __pyx_t_2yt_8geometry_18selection_routines_IntegrationAccumulator *)malloc((sizeof(struct __pyx_t_2yt_8geometry_18selection_routines_IntegrationAccumulator))));

  /* "yt/geometry/selection_routines.pyx":1515
 *         cdef np.ndarray[np.float64_t, ndim=2] coords
 *         cdef np.ndarray[np.int64_t, ndim=2] indices
 *         indices = mesh.connectivity_indices             # <<<<<<<<<<<<<<
 *         coords = _ensure_code(mesh.connectivity_coords)
 *         cdef int nc = indices.shape[0]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_mesh, __pyx_n_s_connectivity_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1515, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_indices.diminfo[1].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_indices.diminfo[1].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1515, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __pyx_v_indices = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1516
 *         cdef np.ndarray[np.int64_t, ndim=2] indices
 *         indices = mesh.connectivity_indices
 *         coords = _ensure_code(mesh.connectivity_coords)             # <<<<<<<<<<<<<<
 *         cdef int nc = indices.shape[0]
 *         cdef int nv = indices.shape[1]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_mesh, __pyx_n_s_connectivity_coords); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1516, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_7);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_coords.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_coords.rcbuffer->pybuffer, (PyObject*)__pyx_v_coords, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_coords.diminfo[0].strides = __pyx_pybuffernd_coords.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_coords.diminfo[0].shape = __pyx_pybuffernd_coords.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_coords.diminfo[1].strides = __pyx_pybuffernd_coords.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_coords.diminfo[1].shape = __pyx_pybuffernd_coords.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1516, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __pyx_v_coords = ((PyArrayObject *)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "yt/geometry/selection_routines.pyx":1517
 *         indices = mesh.connectivity_indices
 *         coords = _ensure_code(mesh.connectivity_coords)
 *         cdef int nc = indices.shape[0]             # <<<<<<<<<<<<<<
 *         cdef int nv = indices.shape[1]
 *         if nv != 8:
 */
  __pyx_v_nc = (__pyx_v_indices->dimensions[0]);

  /* "yt/geometry/selection_routines.pyx":1518
 *         coords = _ensure_code(mesh.connectivity_coords)
 *         cdef int nc = indices.shape[0]
 *         cdef int nv = indices.shape[1]             # <<<<<<<<<<<<<<
 *         if nv != 8:
 *             raise NotImplementedError
 */
  __pyx_v_nv = (__pyx_v_indices->dimensions[1]);

  /* "yt/geometry/selection_routines.pyx":1519
 *         cdef int nc = indices.shape[0]
 *         cdef int nv = indices.shape[1]
 *         if nv != 8:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError
 *         cdef VolumeContainer vc
 */
  __pyx_t_9 = ((__pyx_v_nv != 8) != 0);
  if (unlikely(__pyx_t_9)) {

    /* "yt/geometry/selection_routines.pyx":1520
 *         cdef int nv = indices.shape[1]
 *         if nv != 8:
 *             raise NotImplementedError             # <<<<<<<<<<<<<<
 *         cdef VolumeContainer vc
 *         child_mask = np.ones((1,1,1), dtype="uint8")
 */
    __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
    __PYX_ERR(0, 1520, __pyx_L1_error)

    /* "yt/geometry/selection_routines.pyx":1519
 *         cdef int nc = indices.shape[0]
 *         cdef int nv = indices.shape[1]
 *         if nv != 8:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError
 *         cdef VolumeContainer vc
 */
  }

  /* "yt/geometry/selection_routines.pyx":1522
 *             raise NotImplementedError
 *         cdef VolumeContainer vc
 *         child_mask = np.ones((1,1,1), dtype="uint8")             # <<<<<<<<<<<<<<
 *         t = np.zeros((1,1,1), dtype="float64")
 *         dt = np.zeros((1,1,1), dtype="float64") - 1
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ones); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 1522, __pyx_L1_error)
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__24, __pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (!(likely(((__pyx_t_10) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_10, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1522, __pyx_L1_error)
  __pyx_t_11 = ((PyArrayObject *)__pyx_t_10);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_11, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_child_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 1, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_child_mask.diminfo[0].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_child_mask.diminfo[0].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_child_mask.diminfo[1].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_child_mask.diminfo[1].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_child_mask.diminfo[2].strides = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_child_mask.diminfo[2].shape = __pyx_pybuffernd_child_mask.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1522, __pyx_L1_error)
  }
  __pyx_t_11 = 0;
  __pyx_v_child_mask = ((PyArrayObject *)__pyx_t_10);
  __pyx_t_10 = 0;

  /* "yt/geometry/selection_routines.pyx":1523
 *         cdef VolumeContainer vc
 *         child_mask = np.ones((1,1,1), dtype="uint8")
 *         t = np.zeros((1,1,1), dtype="float64")             # <<<<<<<<<<<<<<
 *         dt = np.zeros((1,1,1), dtype="float64") - 1
 *         tr = np.zeros(nz, dtype="float64")
 */
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_np); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_dtype, __pyx_n_s_float64) < 0) __PYX_ERR(0, 1523, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__24, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1523, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_t.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_t.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_t.rcbuffer->pybuffer, (PyObject*)__pyx_v_t, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_t.diminfo[0].strides = __pyx_pybuffernd_t.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_t.diminfo[0].shape = __pyx_pybuffernd_t.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_t.diminfo[1].strides = __pyx_pybuffernd_t.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_t.diminfo[1].shape = __pyx_pybuffernd_t.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_t.diminfo[2].strides = __pyx_pybuffernd_t.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_t.diminfo[2].shape = __pyx_pybuffernd_t.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1523, __pyx_L1_error)
  }
  __pyx_t_12 = 0;
  __pyx_v_t = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1524
 *         child_mask = np.ones((1,1,1), dtype="uint8")
 *         t = np.zeros((1,1,1), dtype="float64")
 *         dt = np.zeros((1,1,1), dtype="float64") - 1             # <<<<<<<<<<<<<<
 *         tr = np.zeros(nz, dtype="float64")
 *         dtr = np.zeros(nz, dtype="float64")
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_float64) < 0) __PYX_ERR(0, 1524, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_tuple__24, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_t_7, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1524, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dt.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dt.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dt.rcbuffer->pybuffer, (PyObject*)__pyx_v_dt, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_dt.diminfo[0].strides = __pyx_pybuffernd_dt.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_dt.diminfo[0].shape = __pyx_pybuffernd_dt.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_dt.diminfo[1].strides = __pyx_pybuffernd_dt.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_dt.diminfo[1].shape = __pyx_pybuffernd_dt.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_dt.diminfo[2].strides = __pyx_pybuffernd_dt.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_dt.diminfo[2].shape = __pyx_pybuffernd_dt.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1524, __pyx_L1_error)
  }
  __pyx_t_12 = 0;
  __pyx_v_dt = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1525
 *         t = np.zeros((1,1,1), dtype="float64")
 *         dt = np.zeros((1,1,1), dtype="float64") - 1
 *         tr = np.zeros(nz, dtype="float64")             # <<<<<<<<<<<<<<
 *         dtr = np.zeros(nz, dtype="float64")
 *         ia.t = <np.float64_t *> t.data
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_nz);
  __Pyx_GIVEREF(__pyx_v_nz);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_nz);
  __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_dtype, __pyx_n_s_float64) < 0) __PYX_ERR(0, 1525, __pyx_L1_error)
  __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_1, __pyx_t_10); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (!(likely(((__pyx_t_13) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_13, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1525, __pyx_L1_error)
  __pyx_t_14 = ((PyArrayObject *)__pyx_t_13);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tr.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tr.rcbuffer->pybuffer, (PyObject*)__pyx_t_14, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tr.rcbuffer->pybuffer, (PyObject*)__pyx_v_tr, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_tr.diminfo[0].strides = __pyx_pybuffernd_tr.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_tr.diminfo[0].shape = __pyx_pybuffernd_tr.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1525, __pyx_L1_error)
  }
  __pyx_t_14 = 0;
  __pyx_v_tr = ((PyArrayObject *)__pyx_t_13);
  __pyx_t_13 = 0;

  /* "yt/geometry/selection_routines.pyx":1526
 *         dt = np.zeros((1,1,1), dtype="float64") - 1
 *         tr = np.zeros(nz, dtype="float64")
 *         dtr = np.zeros(nz, dtype="float64")             # <<<<<<<<<<<<<<
 *         ia.t = <np.float64_t *> t.data
 *         ia.dt = <np.float64_t *> dt.data
 */
  __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_np); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_zeros); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = PyTuple_New(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_INCREF(__pyx_v_nz);
  __Pyx_GIVEREF(__pyx_v_nz);
  PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_v_nz);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_float64) < 0) __PYX_ERR(0, 1526, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_13, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1526, __pyx_L1_error)
  __pyx_t_14 = ((PyArrayObject *)__pyx_t_7);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dtr.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dtr.rcbuffer->pybuffer, (PyObject*)__pyx_t_14, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dtr.rcbuffer->pybuffer, (PyObject*)__pyx_v_dtr, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_dtr.diminfo[0].strides = __pyx_pybuffernd_dtr.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_dtr.diminfo[0].shape = __pyx_pybuffernd_dtr.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1526, __pyx_L1_error)
  }
  __pyx_t_14 = 0;
  __pyx_v_dtr = ((PyArrayObject *)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "yt/geometry/selection_routines.pyx":1527
 *         tr = np.zeros(nz, dtype="float64")
 *         dtr = np.zeros(nz, dtype="float64")
 *         ia.t = <np.float64_t *> t.data             # <<<<<<<<<<<<<<
 *         ia.dt = <np.float64_t *> dt.data
 *         ia.child_mask = <np.uint8_t *> child_mask.data
 */
  __pyx_v_ia->t = ((__pyx_t_5numpy_float64_t *)__pyx_v_t->data);

  /* "yt/geometry/selection_routines.pyx":1528
 *         dtr = np.zeros(nz, dtype="float64")
 *         ia.t = <np.float64_t *> t.data
 *         ia.dt = <np.float64_t *> dt.data             # <<<<<<<<<<<<<<
 *         ia.child_mask = <np.uint8_t *> child_mask.data
 *         ia.hits = 0
 */
  __pyx_v_ia->dt = ((__pyx_t_5numpy_float64_t *)__pyx_v_dt->data);

  /* "yt/geometry/selection_routines.pyx":1529
 *         ia.t = <np.float64_t *> t.data
 *         ia.dt = <np.float64_t *> dt.data
 *         ia.child_mask = <np.uint8_t *> child_mask.data             # <<<<<<<<<<<<<<
 *         ia.hits = 0
 *         ni = 0
 */
  __pyx_v_ia->child_mask = ((__pyx_t_5numpy_uint8_t *)__pyx_v_child_mask->data);

  /* "yt/geometry/selection_routines.pyx":1530
 *         ia.dt = <np.float64_t *> dt.data
 *         ia.child_mask = <np.uint8_t *> child_mask.data
 *         ia.hits = 0             # <<<<<<<<<<<<<<
 *         ni = 0
 *         for i in range(nc):
 */
  __pyx_v_ia->hits = 0;

  /* "yt/geometry/selection_routines.pyx":1531
 *         ia.child_mask = <np.uint8_t *> child_mask.data
 *         ia.hits = 0
 *         ni = 0             # <<<<<<<<<<<<<<
 *         for i in range(nc):
 *             for j in range(3):
 */
  __pyx_v_ni = 0;

  /* "yt/geometry/selection_routines.pyx":1532
 *         ia.hits = 0
 *         ni = 0
 *         for i in range(nc):             # <<<<<<<<<<<<<<
 *             for j in range(3):
 *                 LE[j] = 1e60
 */
  __pyx_t_3 = __pyx_v_nc;
  __pyx_t_15 = __pyx_t_3;
  for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
    __pyx_v_i = __pyx_t_16;

    /* "yt/geometry/selection_routines.pyx":1533
 *         ni = 0
 *         for i in range(nc):
 *             for j in range(3):             # <<<<<<<<<<<<<<
 *                 LE[j] = 1e60
 *                 RE[j] = -1e60
 */
    for (__pyx_t_17 = 0; __pyx_t_17 < 3; __pyx_t_17+=1) {
      __pyx_v_j = __pyx_t_17;

      /* "yt/geometry/selection_routines.pyx":1534
 *         for i in range(nc):
 *             for j in range(3):
 *                 LE[j] = 1e60             # <<<<<<<<<<<<<<
 *                 RE[j] = -1e60
 *             for j in range(nv):
 */
      (__pyx_v_LE[__pyx_v_j]) = 1e60;

      /* "yt/geometry/selection_routines.pyx":1535
 *             for j in range(3):
 *                 LE[j] = 1e60
 *                 RE[j] = -1e60             # <<<<<<<<<<<<<<
 *             for j in range(nv):
 *                 for k in range(3):
 */
      (__pyx_v_RE[__pyx_v_j]) = -1e60;
    }

    /* "yt/geometry/selection_routines.pyx":1536
 *                 LE[j] = 1e60
 *                 RE[j] = -1e60
 *             for j in range(nv):             # <<<<<<<<<<<<<<
 *                 for k in range(3):
 *                     pos = coords[indices[i, j] - offset, k]
 */
    __pyx_t_17 = __pyx_v_nv;
    __pyx_t_18 = __pyx_t_17;
    for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
      __pyx_v_j = __pyx_t_19;

      /* "yt/geometry/selection_routines.pyx":1537
 *                 RE[j] = -1e60
 *             for j in range(nv):
 *                 for k in range(3):             # <<<<<<<<<<<<<<
 *                     pos = coords[indices[i, j] - offset, k]
 *                     LE[k] = fmin(pos, LE[k])
 */
      for (__pyx_t_20 = 0; __pyx_t_20 < 3; __pyx_t_20+=1) {
        __pyx_v_k = __pyx_t_20;

        /* "yt/geometry/selection_routines.pyx":1538
 *             for j in range(nv):
 *                 for k in range(3):
 *                     pos = coords[indices[i, j] - offset, k]             # <<<<<<<<<<<<<<
 *                     LE[k] = fmin(pos, LE[k])
 *                     RE[k] = fmax(pos, RE[k])
 */
        __pyx_t_21 = __pyx_v_i;
        __pyx_t_22 = __pyx_v_j;
        __pyx_t_23 = ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_indices.rcbuffer->pybuffer.buf, __pyx_t_21, __pyx_pybuffernd_indices.diminfo[0].strides, __pyx_t_22, __pyx_pybuffernd_indices.diminfo[1].strides)) - __pyx_v_offset);
        __pyx_t_24 = __pyx_v_k;
        __pyx_v_pos = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_coords.rcbuffer->pybuffer.buf, __pyx_t_23, __pyx_pybuffernd_coords.diminfo[0].strides, __pyx_t_24, __pyx_pybuffernd_coords.diminfo[1].strides));

        /* "yt/geometry/selection_routines.pyx":1539
 *                 for k in range(3):
 *                     pos = coords[indices[i, j] - offset, k]
 *                     LE[k] = fmin(pos, LE[k])             # <<<<<<<<<<<<<<
 *                     RE[k] = fmax(pos, RE[k])
 *             for j in range(3):
 */
        (__pyx_v_LE[__pyx_v_k]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_fmin(__pyx_v_pos, (__pyx_v_LE[__pyx_v_k]));

        /* "yt/geometry/selection_routines.pyx":1540
 *                     pos = coords[indices[i, j] - offset, k]
 *                     LE[k] = fmin(pos, LE[k])
 *                     RE[k] = fmax(pos, RE[k])             # <<<<<<<<<<<<<<
 *             for j in range(3):
 *                 vc.left_edge[j] = LE[j]
 */
        (__pyx_v_RE[__pyx_v_k]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_fmax(__pyx_v_pos, (__pyx_v_RE[__pyx_v_k]));
      }
    }

    /* "yt/geometry/selection_routines.pyx":1541
 *                     LE[k] = fmin(pos, LE[k])
 *                     RE[k] = fmax(pos, RE[k])
 *             for j in range(3):             # <<<<<<<<<<<<<<
 *                 vc.left_edge[j] = LE[j]
 *                 vc.right_edge[j] = RE[j]
 */
    for (__pyx_t_17 = 0; __pyx_t_17 < 3; __pyx_t_17+=1) {
      __pyx_v_j = __pyx_t_17;

      /* "yt/geometry/selection_routines.pyx":1542
 *                     RE[k] = fmax(pos, RE[k])
 *             for j in range(3):
 *                 vc.left_edge[j] = LE[j]             # <<<<<<<<<<<<<<
 *                 vc.right_edge[j] = RE[j]
 *                 vc.dds[j] = RE[j] - LE[j]
 */
      (__pyx_v_vc.left_edge[__pyx_v_j]) = (__pyx_v_LE[__pyx_v_j]);

      /* "yt/geometry/selection_routines.pyx":1543
 *             for j in range(3):
 *                 vc.left_edge[j] = LE[j]
 *                 vc.right_edge[j] = RE[j]             # <<<<<<<<<<<<<<
 *                 vc.dds[j] = RE[j] - LE[j]
 *                 vc.idds[j] = 1.0/vc.dds[j]
 */
      (__pyx_v_vc.right_edge[__pyx_v_j]) = (__pyx_v_RE[__pyx_v_j]);

      /* "yt/geometry/selection_routines.pyx":1544
 *                 vc.left_edge[j] = LE[j]
 *                 vc.right_edge[j] = RE[j]
 *                 vc.dds[j] = RE[j] - LE[j]             # <<<<<<<<<<<<<<
 *                 vc.idds[j] = 1.0/vc.dds[j]
 *                 vc.dims[j] = 1
 */
      (__pyx_v_vc.dds[__pyx_v_j]) = ((__pyx_v_RE[__pyx_v_j]) - (__pyx_v_LE[__pyx_v_j]));

      /* "yt/geometry/selection_routines.pyx":1545
 *                 vc.right_edge[j] = RE[j]
 *                 vc.dds[j] = RE[j] - LE[j]
 *                 vc.idds[j] = 1.0/vc.dds[j]             # <<<<<<<<<<<<<<
 *                 vc.dims[j] = 1
 *             t[0,0,0] = dt[0,0,0] = -1
 */
      (__pyx_v_vc.idds[__pyx_v_j]) = (1.0 / (__pyx_v_vc.dds[__pyx_v_j]));

      /* "yt/geometry/selection_routines.pyx":1546
 *                 vc.dds[j] = RE[j] - LE[j]
 *                 vc.idds[j] = 1.0/vc.dds[j]
 *                 vc.dims[j] = 1             # <<<<<<<<<<<<<<
 *             t[0,0,0] = dt[0,0,0] = -1
 *             walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 */
      (__pyx_v_vc.dims[__pyx_v_j]) = 1;
    }

    /* "yt/geometry/selection_routines.pyx":1547
 *                 vc.idds[j] = 1.0/vc.dds[j]
 *                 vc.dims[j] = 1
 *             t[0,0,0] = dt[0,0,0] = -1             # <<<<<<<<<<<<<<
 *             walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 *             if dt[0,0,0] >= 0:
 */
    __pyx_t_22 = 0;
    __pyx_t_21 = 0;
    __pyx_t_24 = 0;
    *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_t.rcbuffer->pybuffer.buf, __pyx_t_22, __pyx_pybuffernd_t.diminfo[0].strides, __pyx_t_21, __pyx_pybuffernd_t.diminfo[1].strides, __pyx_t_24, __pyx_pybuffernd_t.diminfo[2].strides) = -1.0;
    __pyx_t_25 = 0;
    __pyx_t_26 = 0;
    __pyx_t_27 = 0;
    *__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_dt.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_dt.diminfo[0].strides, __pyx_t_26, __pyx_pybuffernd_dt.diminfo[1].strides, __pyx_t_27, __pyx_pybuffernd_dt.diminfo[2].strides) = -1.0;

    /* "yt/geometry/selection_routines.pyx":1548
 *                 vc.dims[j] = 1
 *             t[0,0,0] = dt[0,0,0] = -1
 *             walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)             # <<<<<<<<<<<<<<
 *             if dt[0,0,0] >= 0:
 *                 tr[ni] = t[0,0,0]
 */
    (void)(__pyx_f_2yt_9utilities_3lib_14grid_traversal_walk_volume((&__pyx_v_vc), __pyx_v_self->p1, __pyx_v_self->vec, __pyx_f_2yt_8geometry_18selection_routines_dt_sampler, ((void *)__pyx_v_ia), NULL));

    /* "yt/geometry/selection_routines.pyx":1549
 *             t[0,0,0] = dt[0,0,0] = -1
 *             walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 *             if dt[0,0,0] >= 0:             # <<<<<<<<<<<<<<
 *                 tr[ni] = t[0,0,0]
 *                 dtr[ni] = dt[0,0,0]
 */
    __pyx_t_28 = 0;
    __pyx_t_29 = 0;
    __pyx_t_30 = 0;
    __pyx_t_9 = (((*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_dt.rcbuffer->pybuffer.buf, __pyx_t_28, __pyx_pybuffernd_dt.diminfo[0].strides, __pyx_t_29, __pyx_pybuffernd_dt.diminfo[1].strides, __pyx_t_30, __pyx_pybuffernd_dt.diminfo[2].strides)) >= 0.0) != 0);
    if (__pyx_t_9) {

      /* "yt/geometry/selection_routines.pyx":1550
 *             walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 *             if dt[0,0,0] >= 0:
 *                 tr[ni] = t[0,0,0]             # <<<<<<<<<<<<<<
 *                 dtr[ni] = dt[0,0,0]
 *                 ni += 1
 */
      __pyx_t_30 = 0;
      __pyx_t_29 = 0;
      __pyx_t_28 = 0;
      __pyx_t_31 = __pyx_v_ni;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_tr.rcbuffer->pybuffer.buf, __pyx_t_31, __pyx_pybuffernd_tr.diminfo[0].strides) = (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_t.rcbuffer->pybuffer.buf, __pyx_t_30, __pyx_pybuffernd_t.diminfo[0].strides, __pyx_t_29, __pyx_pybuffernd_t.diminfo[1].strides, __pyx_t_28, __pyx_pybuffernd_t.diminfo[2].strides));

      /* "yt/geometry/selection_routines.pyx":1551
 *             if dt[0,0,0] >= 0:
 *                 tr[ni] = t[0,0,0]
 *                 dtr[ni] = dt[0,0,0]             # <<<<<<<<<<<<<<
 *                 ni += 1
 *         free(ia)
 */
      __pyx_t_28 = 0;
      __pyx_t_29 = 0;
      __pyx_t_30 = 0;
      __pyx_t_32 = __pyx_v_ni;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_dtr.rcbuffer->pybuffer.buf, __pyx_t_32, __pyx_pybuffernd_dtr.diminfo[0].strides) = (*__Pyx_BufPtrStrided3d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_dt.rcbuffer->pybuffer.buf, __pyx_t_28, __pyx_pybuffernd_dt.diminfo[0].strides, __pyx_t_29, __pyx_pybuffernd_dt.diminfo[1].strides, __pyx_t_30, __pyx_pybuffernd_dt.diminfo[2].strides));

      /* "yt/geometry/selection_routines.pyx":1552
 *                 tr[ni] = t[0,0,0]
 *                 dtr[ni] = dt[0,0,0]
 *                 ni += 1             # <<<<<<<<<<<<<<
 *         free(ia)
 *         return dtr, tr
 */
      __pyx_v_ni = (__pyx_v_ni + 1);

      /* "yt/geometry/selection_routines.pyx":1549
 *             t[0,0,0] = dt[0,0,0] = -1
 *             walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 *             if dt[0,0,0] >= 0:             # <<<<<<<<<<<<<<
 *                 tr[ni] = t[0,0,0]
 *                 dtr[ni] = dt[0,0,0]
 */
    }
  }

  /* "yt/geometry/selection_routines.pyx":1553
 *                 dtr[ni] = dt[0,0,0]
 *                 ni += 1
 *         free(ia)             # <<<<<<<<<<<<<<
 *         return dtr, tr
 * 
 */
  free(__pyx_v_ia);

  /* "yt/geometry/selection_routines.pyx":1554
 *                 ni += 1
 *         free(ia)
 *         return dtr, tr             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(((PyObject *)__pyx_v_dtr));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_dtr));
  PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)__pyx_v_dtr));
  __Pyx_INCREF(((PyObject *)__pyx_v_tr));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_tr));
  PyTuple_SET_ITEM(__pyx_t_7, 1, ((PyObject *)__pyx_v_tr));
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1503
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def get_dt_mesh(self, mesh, nz, int offset):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.float64_t, ndim=3] t, dt
 *         cdef np.ndarray[np.float64_t, ndim=1] tr, dtr
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_13);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dt.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dtr.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_t.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tr.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.RaySelector.get_dt_mesh", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_child_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dt.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dtr.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_t.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tr.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_t);
  __Pyx_XDECREF((PyObject *)__pyx_v_dt);
  __Pyx_XDECREF((PyObject *)__pyx_v_tr);
  __Pyx_XDECREF((PyObject *)__pyx_v_dtr);
  __Pyx_XDECREF((PyObject *)__pyx_v_child_mask);
  __Pyx_XDECREF((PyObject *)__pyx_v_coords);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1556
 *         return dtr, tr
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         # two 0-volume constructs don't intersect
 *         return 0
 */

static int __pyx_f_2yt_8geometry_18selection_routines_11RaySelector_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1558
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         # two 0-volume constructs don't intersect
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1556
 *         return dtr, tr
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         # two 0-volume constructs don't intersect
 *         return 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1560
 *         return 0
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         # not implemented
 *         return 0
 */

static int __pyx_f_2yt_8geometry_18selection_routines_11RaySelector_select_sphere(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1562
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         # not implemented
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1560
 *         return 0
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         # not implemented
 *         return 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1567
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef int i, rv
 */

static int __pyx_f_2yt_8geometry_18selection_routines_11RaySelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_v_i;
  int __pyx_v_rv;
  struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer __pyx_v_vc;
  struct __pyx_t_2yt_8geometry_18selection_routines_IntegrationAccumulator *__pyx_v_ia;
  __pyx_t_5numpy_float64_t __pyx_v_dt[1];
  __pyx_t_5numpy_float64_t __pyx_v_t[1];
  __pyx_t_5numpy_uint8_t __pyx_v_cm[1];
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":1572
 *         cdef VolumeContainer vc
 *         cdef IntegrationAccumulator *ia
 *         ia = <IntegrationAccumulator *> malloc(sizeof(IntegrationAccumulator))             # <<<<<<<<<<<<<<
 *         cdef np.float64_t dt[1]
 *         cdef np.float64_t t[1]
 */
  __pyx_v_ia = ((struct __pyx_t_2yt_8geometry_18selection_routines_IntegrationAccumulator *)malloc((sizeof(struct __pyx_t_2yt_8geometry_18selection_routines_IntegrationAccumulator))));

  /* "yt/geometry/selection_routines.pyx":1576
 *         cdef np.float64_t t[1]
 *         cdef np.uint8_t cm[1]
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             vc.left_edge[i] = left_edge[i]
 *             vc.right_edge[i] = right_edge[i]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":1577
 *         cdef np.uint8_t cm[1]
 *         for i in range(3):
 *             vc.left_edge[i] = left_edge[i]             # <<<<<<<<<<<<<<
 *             vc.right_edge[i] = right_edge[i]
 *             vc.dds[i] = right_edge[i] - left_edge[i]
 */
    (__pyx_v_vc.left_edge[__pyx_v_i]) = (__pyx_v_left_edge[__pyx_v_i]);

    /* "yt/geometry/selection_routines.pyx":1578
 *         for i in range(3):
 *             vc.left_edge[i] = left_edge[i]
 *             vc.right_edge[i] = right_edge[i]             # <<<<<<<<<<<<<<
 *             vc.dds[i] = right_edge[i] - left_edge[i]
 *             vc.idds[i] = 1.0/vc.dds[i]
 */
    (__pyx_v_vc.right_edge[__pyx_v_i]) = (__pyx_v_right_edge[__pyx_v_i]);

    /* "yt/geometry/selection_routines.pyx":1579
 *             vc.left_edge[i] = left_edge[i]
 *             vc.right_edge[i] = right_edge[i]
 *             vc.dds[i] = right_edge[i] - left_edge[i]             # <<<<<<<<<<<<<<
 *             vc.idds[i] = 1.0/vc.dds[i]
 *             vc.dims[i] = 1
 */
    (__pyx_v_vc.dds[__pyx_v_i]) = ((__pyx_v_right_edge[__pyx_v_i]) - (__pyx_v_left_edge[__pyx_v_i]));

    /* "yt/geometry/selection_routines.pyx":1580
 *             vc.right_edge[i] = right_edge[i]
 *             vc.dds[i] = right_edge[i] - left_edge[i]
 *             vc.idds[i] = 1.0/vc.dds[i]             # <<<<<<<<<<<<<<
 *             vc.dims[i] = 1
 *         t[0] = dt[0] = 0.0
 */
    (__pyx_v_vc.idds[__pyx_v_i]) = (1.0 / (__pyx_v_vc.dds[__pyx_v_i]));

    /* "yt/geometry/selection_routines.pyx":1581
 *             vc.dds[i] = right_edge[i] - left_edge[i]
 *             vc.idds[i] = 1.0/vc.dds[i]
 *             vc.dims[i] = 1             # <<<<<<<<<<<<<<
 *         t[0] = dt[0] = 0.0
 *         cm[0] = 1
 */
    (__pyx_v_vc.dims[__pyx_v_i]) = 1;
  }

  /* "yt/geometry/selection_routines.pyx":1582
 *             vc.idds[i] = 1.0/vc.dds[i]
 *             vc.dims[i] = 1
 *         t[0] = dt[0] = 0.0             # <<<<<<<<<<<<<<
 *         cm[0] = 1
 *         ia.t = t
 */
  (__pyx_v_t[0]) = 0.0;
  (__pyx_v_dt[0]) = 0.0;

  /* "yt/geometry/selection_routines.pyx":1583
 *             vc.dims[i] = 1
 *         t[0] = dt[0] = 0.0
 *         cm[0] = 1             # <<<<<<<<<<<<<<
 *         ia.t = t
 *         ia.dt = dt
 */
  (__pyx_v_cm[0]) = 1;

  /* "yt/geometry/selection_routines.pyx":1584
 *         t[0] = dt[0] = 0.0
 *         cm[0] = 1
 *         ia.t = t             # <<<<<<<<<<<<<<
 *         ia.dt = dt
 *         ia.child_mask = cm
 */
  __pyx_v_ia->t = __pyx_v_t;

  /* "yt/geometry/selection_routines.pyx":1585
 *         cm[0] = 1
 *         ia.t = t
 *         ia.dt = dt             # <<<<<<<<<<<<<<
 *         ia.child_mask = cm
 *         ia.hits = 0
 */
  __pyx_v_ia->dt = __pyx_v_dt;

  /* "yt/geometry/selection_routines.pyx":1586
 *         ia.t = t
 *         ia.dt = dt
 *         ia.child_mask = cm             # <<<<<<<<<<<<<<
 *         ia.hits = 0
 *         walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 */
  __pyx_v_ia->child_mask = __pyx_v_cm;

  /* "yt/geometry/selection_routines.pyx":1587
 *         ia.dt = dt
 *         ia.child_mask = cm
 *         ia.hits = 0             # <<<<<<<<<<<<<<
 *         walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 *         rv = 0
 */
  __pyx_v_ia->hits = 0;

  /* "yt/geometry/selection_routines.pyx":1588
 *         ia.child_mask = cm
 *         ia.hits = 0
 *         walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)             # <<<<<<<<<<<<<<
 *         rv = 0
 *         if ia.hits > 0:
 */
  (void)(__pyx_f_2yt_9utilities_3lib_14grid_traversal_walk_volume((&__pyx_v_vc), __pyx_v_self->p1, __pyx_v_self->vec, __pyx_f_2yt_8geometry_18selection_routines_dt_sampler, ((void *)__pyx_v_ia), NULL));

  /* "yt/geometry/selection_routines.pyx":1589
 *         ia.hits = 0
 *         walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 *         rv = 0             # <<<<<<<<<<<<<<
 *         if ia.hits > 0:
 *             rv = 1
 */
  __pyx_v_rv = 0;

  /* "yt/geometry/selection_routines.pyx":1590
 *         walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 *         rv = 0
 *         if ia.hits > 0:             # <<<<<<<<<<<<<<
 *             rv = 1
 *         free(ia)
 */
  __pyx_t_2 = ((__pyx_v_ia->hits > 0) != 0);
  if (__pyx_t_2) {

    /* "yt/geometry/selection_routines.pyx":1591
 *         rv = 0
 *         if ia.hits > 0:
 *             rv = 1             # <<<<<<<<<<<<<<
 *         free(ia)
 *         return rv
 */
    __pyx_v_rv = 1;

    /* "yt/geometry/selection_routines.pyx":1590
 *         walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
 *         rv = 0
 *         if ia.hits > 0:             # <<<<<<<<<<<<<<
 *             rv = 1
 *         free(ia)
 */
  }

  /* "yt/geometry/selection_routines.pyx":1592
 *         if ia.hits > 0:
 *             rv = 1
 *         free(ia)             # <<<<<<<<<<<<<<
 *         return rv
 * 
 */
  free(__pyx_v_ia);

  /* "yt/geometry/selection_routines.pyx":1593
 *             rv = 1
 *         free(ia)
 *         return rv             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = __pyx_v_rv;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1567
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef int i, rv
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1598
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t dds[3]) nogil:
 *         # This is terribly inefficient for Octrees.  For grids, it will never
 */

static int __pyx_f_2yt_8geometry_18selection_routines_11RaySelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_left_edge[3];
  __pyx_t_5numpy_float64_t __pyx_v_right_edge[3];
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":1605
 *         cdef np.float64_t left_edge[3]
 *         cdef np.float64_t right_edge[3]
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             left_edge[i] = pos[i] - dds[i]/2.0
 *             right_edge[i] = pos[i] + dds[i]/2.0
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":1606
 *         cdef np.float64_t right_edge[3]
 *         for i in range(3):
 *             left_edge[i] = pos[i] - dds[i]/2.0             # <<<<<<<<<<<<<<
 *             right_edge[i] = pos[i] + dds[i]/2.0
 *         return self.select_bbox(left_edge, right_edge)
 */
    (__pyx_v_left_edge[__pyx_v_i]) = ((__pyx_v_pos[__pyx_v_i]) - ((__pyx_v_dds[__pyx_v_i]) / 2.0));

    /* "yt/geometry/selection_routines.pyx":1607
 *         for i in range(3):
 *             left_edge[i] = pos[i] - dds[i]/2.0
 *             right_edge[i] = pos[i] + dds[i]/2.0             # <<<<<<<<<<<<<<
 *         return self.select_bbox(left_edge, right_edge)
 * 
 */
    (__pyx_v_right_edge[__pyx_v_i]) = ((__pyx_v_pos[__pyx_v_i]) + ((__pyx_v_dds[__pyx_v_i]) / 2.0));
  }

  /* "yt/geometry/selection_routines.pyx":1608
 *             left_edge[i] = pos[i] - dds[i]/2.0
 *             right_edge[i] = pos[i] + dds[i]/2.0
 *         return self.select_bbox(left_edge, right_edge)             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_r = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_RaySelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.select_bbox(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_left_edge, __pyx_v_right_edge);
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1598
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t dds[3]) nogil:
 *         # This is terribly inefficient for Octrees.  For grids, it will never
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1610
 *         return self.select_bbox(left_edge, right_edge)
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("p1[0]", self.p1[0]),
 *                 ("p1[1]", self.p1[1]),
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_9_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_9_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_8_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_8_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":1611
 * 
 *     def _hash_vals(self):
 *         return (("p1[0]", self.p1[0]),             # <<<<<<<<<<<<<<
 *                 ("p1[1]", self.p1[1]),
 *                 ("p1[2]", self.p1[2]),
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->p1[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_p1_0);
  __Pyx_GIVEREF(__pyx_kp_s_p1_0);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_s_p1_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1612
 *     def _hash_vals(self):
 *         return (("p1[0]", self.p1[0]),
 *                 ("p1[1]", self.p1[1]),             # <<<<<<<<<<<<<<
 *                 ("p1[2]", self.p1[2]),
 *                 ("p2[0]", self.p2[0]),
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->p1[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_kp_s_p1_1);
  __Pyx_GIVEREF(__pyx_kp_s_p1_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_s_p1_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1613
 *         return (("p1[0]", self.p1[0]),
 *                 ("p1[1]", self.p1[1]),
 *                 ("p1[2]", self.p1[2]),             # <<<<<<<<<<<<<<
 *                 ("p2[0]", self.p2[0]),
 *                 ("p2[1]", self.p2[1]),
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->p1[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_kp_s_p1_2);
  __Pyx_GIVEREF(__pyx_kp_s_p1_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_s_p1_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1614
 *                 ("p1[1]", self.p1[1]),
 *                 ("p1[2]", self.p1[2]),
 *                 ("p2[0]", self.p2[0]),             # <<<<<<<<<<<<<<
 *                 ("p2[1]", self.p2[1]),
 *                 ("p2[2]", self.p2[2]),
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->p2[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_kp_s_p2_0);
  __Pyx_GIVEREF(__pyx_kp_s_p2_0);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_s_p2_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1615
 *                 ("p1[2]", self.p1[2]),
 *                 ("p2[0]", self.p2[0]),
 *                 ("p2[1]", self.p2[1]),             # <<<<<<<<<<<<<<
 *                 ("p2[2]", self.p2[2]),
 *                 ("vec[0]", self.vec[0]),
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->p2[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_kp_s_p2_1);
  __Pyx_GIVEREF(__pyx_kp_s_p2_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_p2_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1616
 *                 ("p2[0]", self.p2[0]),
 *                 ("p2[1]", self.p2[1]),
 *                 ("p2[2]", self.p2[2]),             # <<<<<<<<<<<<<<
 *                 ("vec[0]", self.vec[0]),
 *                 ("vec[1]", self.vec[1]),
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->p2[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_kp_s_p2_2);
  __Pyx_GIVEREF(__pyx_kp_s_p2_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_kp_s_p2_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1617
 *                 ("p2[1]", self.p2[1]),
 *                 ("p2[2]", self.p2[2]),
 *                 ("vec[0]", self.vec[0]),             # <<<<<<<<<<<<<<
 *                 ("vec[1]", self.vec[1]),
 *                 ("vec[2]", self.vec[2]))
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->vec[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_kp_s_vec_0);
  __Pyx_GIVEREF(__pyx_kp_s_vec_0);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_kp_s_vec_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1618
 *                 ("p2[2]", self.p2[2]),
 *                 ("vec[0]", self.vec[0]),
 *                 ("vec[1]", self.vec[1]),             # <<<<<<<<<<<<<<
 *                 ("vec[2]", self.vec[2]))
 * 
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->vec[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_kp_s_vec_1);
  __Pyx_GIVEREF(__pyx_kp_s_vec_1);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_kp_s_vec_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1619
 *                 ("vec[0]", self.vec[0]),
 *                 ("vec[1]", self.vec[1]),
 *                 ("vec[2]", self.vec[2]))             # <<<<<<<<<<<<<<
 * 
 * ray_selector = RaySelector
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->vec[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_kp_s_vec_2);
  __Pyx_GIVEREF(__pyx_kp_s_vec_2);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_kp_s_vec_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1611
 * 
 *     def _hash_vals(self):
 *         return (("p1[0]", self.p1[0]),             # <<<<<<<<<<<<<<
 *                 ("p1[1]", self.p1[1]),
 *                 ("p1[2]", self.p1[2]),
 */
  __pyx_t_1 = PyTuple_New(9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_t_10);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1610
 *         return self.select_bbox(left_edge, right_edge)
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("p1[0]", self.p1[0]),
 *                 ("p1[1]", self.p1[1]),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("yt.geometry.selection_routines.RaySelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_10__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.RaySelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_12__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_11RaySelector_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.RaySelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1627
 *     cdef np.int64_t nids
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         self.obj_ids = dobj._obj_ids
 *         self.nids = self.obj_ids.shape[0]
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1627, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1627, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.DataCollectionSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":1628
 * 
 *     def __init__(self, dobj):
 *         self.obj_ids = dobj._obj_ids             # <<<<<<<<<<<<<<
 *         self.nids = self.obj_ids.shape[0]
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_obj_ids); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->obj_ids);
  __Pyx_DECREF(__pyx_v_self->obj_ids);
  __pyx_v_self->obj_ids = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1629
 *     def __init__(self, dobj):
 *         self.obj_ids = dobj._obj_ids
 *         self.nids = self.obj_ids.shape[0]             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->obj_ids, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_npy_int64(__pyx_t_2); if (unlikely((__pyx_t_3 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1629, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self->nids = __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":1627
 *     cdef np.int64_t nids
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         self.obj_ids = dobj._obj_ids
 *         self.nids = self.obj_ids.shape[0]
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("yt.geometry.selection_routines.DataCollectionSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1634
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_grids(self,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=2] left_edges,
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_3select_grids(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_3select_grids(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left_edges = 0;
  CYTHON_UNUSED PyArrayObject *__pyx_v_right_edges = 0;
  CYTHON_UNUSED PyArrayObject *__pyx_v_levels = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("select_grids (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left_edges,&__pyx_n_s_right_edges,&__pyx_n_s_levels,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left_edges)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right_edges)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, 1); __PYX_ERR(0, 1634, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_levels)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, 2); __PYX_ERR(0, 1634, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "select_grids") < 0)) __PYX_ERR(0, 1634, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_left_edges = ((PyArrayObject *)values[0]);
    __pyx_v_right_edges = ((PyArrayObject *)values[1]);
    __pyx_v_levels = ((PyArrayObject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1634, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.DataCollectionSelector.select_grids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left_edges), __pyx_ptype_5numpy_ndarray, 1, "left_edges", 0))) __PYX_ERR(0, 1635, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right_edges), __pyx_ptype_5numpy_ndarray, 1, "right_edges", 0))) __PYX_ERR(0, 1636, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_levels), __pyx_ptype_5numpy_ndarray, 1, "levels", 0))) __PYX_ERR(0, 1637, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector_2select_grids(((struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *)__pyx_v_self), __pyx_v_left_edges, __pyx_v_right_edges, __pyx_v_levels);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector_2select_grids(struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *__pyx_v_self, PyArrayObject *__pyx_v_left_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_right_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_levels) {
  int __pyx_v_n;
  int __pyx_v_ng;
  PyArrayObject *__pyx_v_gridi = 0;
  PyArrayObject *__pyx_v_oids = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_gridi;
  __Pyx_Buffer __pyx_pybuffer_gridi;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left_edges;
  __Pyx_Buffer __pyx_pybuffer_left_edges;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_levels;
  __Pyx_Buffer __pyx_pybuffer_levels;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_oids;
  __Pyx_Buffer __pyx_pybuffer_oids;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right_edges;
  __Pyx_Buffer __pyx_pybuffer_right_edges;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyArrayObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_6;
  __pyx_t_5numpy_int64_t __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("select_grids", 0);
  __pyx_pybuffer_gridi.pybuffer.buf = NULL;
  __pyx_pybuffer_gridi.refcount = 0;
  __pyx_pybuffernd_gridi.data = NULL;
  __pyx_pybuffernd_gridi.rcbuffer = &__pyx_pybuffer_gridi;
  __pyx_pybuffer_oids.pybuffer.buf = NULL;
  __pyx_pybuffer_oids.refcount = 0;
  __pyx_pybuffernd_oids.data = NULL;
  __pyx_pybuffernd_oids.rcbuffer = &__pyx_pybuffer_oids;
  __pyx_pybuffer_left_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_left_edges.refcount = 0;
  __pyx_pybuffernd_left_edges.data = NULL;
  __pyx_pybuffernd_left_edges.rcbuffer = &__pyx_pybuffer_left_edges;
  __pyx_pybuffer_right_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_right_edges.refcount = 0;
  __pyx_pybuffernd_right_edges.data = NULL;
  __pyx_pybuffernd_right_edges.rcbuffer = &__pyx_pybuffer_right_edges;
  __pyx_pybuffer_levels.pybuffer.buf = NULL;
  __pyx_pybuffer_levels.refcount = 0;
  __pyx_pybuffernd_levels.data = NULL;
  __pyx_pybuffernd_levels.rcbuffer = &__pyx_pybuffer_levels;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_left_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1634, __pyx_L1_error)
  }
  __pyx_pybuffernd_left_edges.diminfo[0].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left_edges.diminfo[0].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_left_edges.diminfo[1].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_left_edges.diminfo[1].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_right_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1634, __pyx_L1_error)
  }
  __pyx_pybuffernd_right_edges.diminfo[0].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right_edges.diminfo[0].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_right_edges.diminfo[1].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_right_edges.diminfo[1].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_levels.rcbuffer->pybuffer, (PyObject*)__pyx_v_levels, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1634, __pyx_L1_error)
  }
  __pyx_pybuffernd_levels.diminfo[0].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_levels.diminfo[0].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_levels.diminfo[1].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_levels.diminfo[1].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[1];

  /* "yt/geometry/selection_routines.pyx":1639
 *                      np.ndarray[np.int32_t, ndim=2] levels):
 *         cdef int n
 *         cdef int ng = left_edges.shape[0]             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')
 *         cdef np.ndarray[np.int64_t, ndim=1] oids = self.obj_ids
 */
  __pyx_v_ng = (__pyx_v_left_edges->dimensions[0]);

  /* "yt/geometry/selection_routines.pyx":1640
 *         cdef int n
 *         cdef int ng = left_edges.shape[0]
 *         cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.int64_t, ndim=1] oids = self.obj_ids
 *         with nogil:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_ng); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 1640, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1640, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_gridi.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_gridi = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_gridi.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1640, __pyx_L1_error)
    } else {__pyx_pybuffernd_gridi.diminfo[0].strides = __pyx_pybuffernd_gridi.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_gridi.diminfo[0].shape = __pyx_pybuffernd_gridi.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_5 = 0;
  __pyx_v_gridi = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":1641
 *         cdef int ng = left_edges.shape[0]
 *         cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')
 *         cdef np.ndarray[np.int64_t, ndim=1] oids = self.obj_ids             # <<<<<<<<<<<<<<
 *         with nogil:
 *             for n in range(self.nids):
 */
  if (!(likely(((__pyx_v_self->obj_ids) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_self->obj_ids, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1641, __pyx_L1_error)
  __pyx_t_4 = __pyx_v_self->obj_ids;
  __Pyx_INCREF(__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_oids.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_t_4), &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_oids = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_oids.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1641, __pyx_L1_error)
    } else {__pyx_pybuffernd_oids.diminfo[0].strides = __pyx_pybuffernd_oids.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_oids.diminfo[0].shape = __pyx_pybuffernd_oids.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_v_oids = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":1642
 *         cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')
 *         cdef np.ndarray[np.int64_t, ndim=1] oids = self.obj_ids
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for n in range(self.nids):
 *                 gridi[oids[n]] = 1
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "yt/geometry/selection_routines.pyx":1643
 *         cdef np.ndarray[np.int64_t, ndim=1] oids = self.obj_ids
 *         with nogil:
 *             for n in range(self.nids):             # <<<<<<<<<<<<<<
 *                 gridi[oids[n]] = 1
 *         return gridi.astype("bool")
 */
        __pyx_t_6 = __pyx_v_self->nids;
        __pyx_t_7 = __pyx_t_6;
        for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
          __pyx_v_n = __pyx_t_8;

          /* "yt/geometry/selection_routines.pyx":1644
 *         with nogil:
 *             for n in range(self.nids):
 *                 gridi[oids[n]] = 1             # <<<<<<<<<<<<<<
 *         return gridi.astype("bool")
 * 
 */
          __pyx_t_9 = __pyx_v_n;
          __pyx_t_10 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int64_t *, __pyx_pybuffernd_oids.rcbuffer->pybuffer.buf, __pyx_t_9, __pyx_pybuffernd_oids.diminfo[0].strides));
          *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_gridi.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_gridi.diminfo[0].strides) = 1;
        }
      }

      /* "yt/geometry/selection_routines.pyx":1642
 *         cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')
 *         cdef np.ndarray[np.int64_t, ndim=1] oids = self.obj_ids
 *         with nogil:             # <<<<<<<<<<<<<<
 *             for n in range(self.nids):
 *                 gridi[oids[n]] = 1
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "yt/geometry/selection_routines.pyx":1645
 *             for n in range(self.nids):
 *                 gridi[oids[n]] = 1
 *         return gridi.astype("bool")             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_gridi), __pyx_n_s_astype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_n_s_bool) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_n_s_bool);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1634
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_grids(self,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=2] left_edges,
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_gridi.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_oids.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.DataCollectionSelector.select_grids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_gridi.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_oids.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_gridi);
  __Pyx_XDECREF((PyObject *)__pyx_v_oids);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1650
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def fill_mask(self, gobj):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=3] mask
 *         mask = np.ones(gobj.ActiveDimensions, dtype='uint8')
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_5fill_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_gobj); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_5fill_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_gobj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fill_mask (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector_4fill_mask(((struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *)__pyx_v_self), ((PyObject *)__pyx_v_gobj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector_4fill_mask(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *__pyx_v_self, PyObject *__pyx_v_gobj) {
  PyArrayObject *__pyx_v_mask = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyArrayObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fill_mask", 0);
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;

  /* "yt/geometry/selection_routines.pyx":1652
 *     def fill_mask(self, gobj):
 *         cdef np.ndarray[np.uint8_t, ndim=3] mask
 *         mask = np.ones(gobj.ActiveDimensions, dtype='uint8')             # <<<<<<<<<<<<<<
 *         return mask.astype("bool")
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_ActiveDimensions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 1652, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1652, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
      __pyx_t_7 = __pyx_t_8 = __pyx_t_9 = 0;
    }
    __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mask.diminfo[1].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mask.diminfo[1].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_mask.diminfo[2].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_mask.diminfo[2].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1652, __pyx_L1_error)
  }
  __pyx_t_5 = 0;
  __pyx_v_mask = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":1653
 *         cdef np.ndarray[np.uint8_t, ndim=3] mask
 *         mask = np.ones(gobj.ActiveDimensions, dtype='uint8')
 *         return mask.astype("bool")             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_mask), __pyx_n_s_astype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_n_s_bool) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_n_s_bool);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1650
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def fill_mask(self, gobj):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=3] mask
 *         mask = np.ones(gobj.ActiveDimensions, dtype='uint8')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.DataCollectionSelector.fill_mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1655
 *         return mask.astype("bool")
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (hash(self.obj_ids.tostring()), self.nids)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_7_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_7_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector_6_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector_6_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_hash_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":1656
 * 
 *     def _hash_vals(self):
 *         return (hash(self.obj_ids.tostring()), self.nids)             # <<<<<<<<<<<<<<
 * 
 * data_collection_selector = DataCollectionSelector
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->obj_ids, __pyx_n_s_tostring); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_hash_t)-1))) __PYX_ERR(0, 1656, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->nids); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1655
 *         return mask.astype("bool")
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (hash(self.obj_ids.tostring()), self.nids)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("yt.geometry.selection_routines.DataCollectionSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector_8__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.DataCollectionSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector_10__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_22DataCollectionSelector_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.DataCollectionSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1665
 *     cdef np.float64_t center[3]
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         _ensure_code(dobj.center)
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_17EllipsoidSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_17EllipsoidSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1665, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1665, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.EllipsoidSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_17EllipsoidSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_17EllipsoidSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __pyx_t_5numpy_float64_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":1667
 *     def __init__(self, dobj):
 *         cdef int i
 *         _ensure_code(dobj.center)             # <<<<<<<<<<<<<<
 *         _ensure_code(dobj._e0)
 *         _ensure_code(dobj._e1)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":1668
 *         cdef int i
 *         _ensure_code(dobj.center)
 *         _ensure_code(dobj._e0)             # <<<<<<<<<<<<<<
 *         _ensure_code(dobj._e1)
 *         _ensure_code(dobj._e2)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_e0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1669
 *         _ensure_code(dobj.center)
 *         _ensure_code(dobj._e0)
 *         _ensure_code(dobj._e1)             # <<<<<<<<<<<<<<
 *         _ensure_code(dobj._e2)
 *         _ensure_code(dobj._A)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_e1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":1670
 *         _ensure_code(dobj._e0)
 *         _ensure_code(dobj._e1)
 *         _ensure_code(dobj._e2)             # <<<<<<<<<<<<<<
 *         _ensure_code(dobj._A)
 *         _ensure_code(dobj._B)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_e2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1671
 *         _ensure_code(dobj._e1)
 *         _ensure_code(dobj._e2)
 *         _ensure_code(dobj._A)             # <<<<<<<<<<<<<<
 *         _ensure_code(dobj._B)
 *         _ensure_code(dobj._C)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_A); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":1672
 *         _ensure_code(dobj._e2)
 *         _ensure_code(dobj._A)
 *         _ensure_code(dobj._B)             # <<<<<<<<<<<<<<
 *         _ensure_code(dobj._C)
 *         for i in range(3):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_B); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1673
 *         _ensure_code(dobj._A)
 *         _ensure_code(dobj._B)
 *         _ensure_code(dobj._C)             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             self.center[i] = dobj.center[i]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_C); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_2yt_8geometry_18selection_routines__ensure_code(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":1674
 *         _ensure_code(dobj._B)
 *         _ensure_code(dobj._C)
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             self.center[i] = dobj.center[i]
 *             self.vec[0][i] = dobj._e0[i]
 */
  for (__pyx_t_3 = 0; __pyx_t_3 < 3; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "yt/geometry/selection_routines.pyx":1675
 *         _ensure_code(dobj._C)
 *         for i in range(3):
 *             self.center[i] = dobj.center[i]             # <<<<<<<<<<<<<<
 *             self.vec[0][i] = dobj._e0[i]
 *             self.vec[1][i] = dobj._e1[i]
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_center); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1675, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_self->center[__pyx_v_i]) = __pyx_t_4;

    /* "yt/geometry/selection_routines.pyx":1676
 *         for i in range(3):
 *             self.center[i] = dobj.center[i]
 *             self.vec[0][i] = dobj._e0[i]             # <<<<<<<<<<<<<<
 *             self.vec[1][i] = dobj._e1[i]
 *             self.vec[2][i] = dobj._e2[i]
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_e0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1676, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    ((__pyx_v_self->vec[0])[__pyx_v_i]) = __pyx_t_4;

    /* "yt/geometry/selection_routines.pyx":1677
 *             self.center[i] = dobj.center[i]
 *             self.vec[0][i] = dobj._e0[i]
 *             self.vec[1][i] = dobj._e1[i]             # <<<<<<<<<<<<<<
 *             self.vec[2][i] = dobj._e2[i]
 *         self.mag[0] = dobj._A
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_e1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1677, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1677, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1677, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    ((__pyx_v_self->vec[1])[__pyx_v_i]) = __pyx_t_4;

    /* "yt/geometry/selection_routines.pyx":1678
 *             self.vec[0][i] = dobj._e0[i]
 *             self.vec[1][i] = dobj._e1[i]
 *             self.vec[2][i] = dobj._e2[i]             # <<<<<<<<<<<<<<
 *         self.mag[0] = dobj._A
 *         self.mag[1] = dobj._B
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_e2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1678, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1678, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1678, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    ((__pyx_v_self->vec[2])[__pyx_v_i]) = __pyx_t_4;
  }

  /* "yt/geometry/selection_routines.pyx":1679
 *             self.vec[1][i] = dobj._e1[i]
 *             self.vec[2][i] = dobj._e2[i]
 *         self.mag[0] = dobj._A             # <<<<<<<<<<<<<<
 *         self.mag[1] = dobj._B
 *         self.mag[2] = dobj._C
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_A); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1679, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  (__pyx_v_self->mag[0]) = __pyx_t_4;

  /* "yt/geometry/selection_routines.pyx":1680
 *             self.vec[2][i] = dobj._e2[i]
 *         self.mag[0] = dobj._A
 *         self.mag[1] = dobj._B             # <<<<<<<<<<<<<<
 *         self.mag[2] = dobj._C
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_B); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1680, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  (__pyx_v_self->mag[1]) = __pyx_t_4;

  /* "yt/geometry/selection_routines.pyx":1681
 *         self.mag[0] = dobj._A
 *         self.mag[1] = dobj._B
 *         self.mag[2] = dobj._C             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_C); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1681, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  (__pyx_v_self->mag[2]) = __pyx_t_4;

  /* "yt/geometry/selection_routines.pyx":1665
 *     cdef np.float64_t center[3]
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         _ensure_code(dobj.center)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("yt.geometry.selection_routines.EllipsoidSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1686
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         return self.select_point(pos)
 * 
 */

static int __pyx_f_2yt_8geometry_18selection_routines_17EllipsoidSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1687
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         return self.select_point(pos)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_EllipsoidSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.select_point(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_pos);
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1686
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         return self.select_point(pos)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1692
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef np.float64_t dot_evec[3]
 *         cdef np.float64_t dist
 */

static int __pyx_f_2yt_8geometry_18selection_routines_17EllipsoidSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  __pyx_t_5numpy_float64_t __pyx_v_dot_evec[3];
  __pyx_t_5numpy_float64_t __pyx_v_dist;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "yt/geometry/selection_routines.pyx":1696
 *         cdef np.float64_t dist
 *         cdef int i, j
 *         dot_evec[0] = dot_evec[1] = dot_evec[2] = 0             # <<<<<<<<<<<<<<
 *         # Calculate the rotated dot product
 *         for i in range(3): # axis
 */
  (__pyx_v_dot_evec[0]) = 0.0;
  (__pyx_v_dot_evec[1]) = 0.0;
  (__pyx_v_dot_evec[2]) = 0.0;

  /* "yt/geometry/selection_routines.pyx":1698
 *         dot_evec[0] = dot_evec[1] = dot_evec[2] = 0
 *         # Calculate the rotated dot product
 *         for i in range(3): # axis             # <<<<<<<<<<<<<<
 *             dist = self.difference(pos[i], self.center[i], i)
 *             for j in range(3):
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":1699
 *         # Calculate the rotated dot product
 *         for i in range(3): # axis
 *             dist = self.difference(pos[i], self.center[i], i)             # <<<<<<<<<<<<<<
 *             for j in range(3):
 *                 dot_evec[j] += dist * self.vec[j][i]
 */
    __pyx_v_dist = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_EllipsoidSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.difference(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), (__pyx_v_pos[__pyx_v_i]), (__pyx_v_self->center[__pyx_v_i]), __pyx_v_i);

    /* "yt/geometry/selection_routines.pyx":1700
 *         for i in range(3): # axis
 *             dist = self.difference(pos[i], self.center[i], i)
 *             for j in range(3):             # <<<<<<<<<<<<<<
 *                 dot_evec[j] += dist * self.vec[j][i]
 *         dist = 0.0
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 3; __pyx_t_2+=1) {
      __pyx_v_j = __pyx_t_2;

      /* "yt/geometry/selection_routines.pyx":1701
 *             dist = self.difference(pos[i], self.center[i], i)
 *             for j in range(3):
 *                 dot_evec[j] += dist * self.vec[j][i]             # <<<<<<<<<<<<<<
 *         dist = 0.0
 *         for i in range(3):
 */
      __pyx_t_3 = __pyx_v_j;
      (__pyx_v_dot_evec[__pyx_t_3]) = ((__pyx_v_dot_evec[__pyx_t_3]) + (__pyx_v_dist * ((__pyx_v_self->vec[__pyx_v_j])[__pyx_v_i])));
    }
  }

  /* "yt/geometry/selection_routines.pyx":1702
 *             for j in range(3):
 *                 dot_evec[j] += dist * self.vec[j][i]
 *         dist = 0.0             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             dist += (dot_evec[i] * dot_evec[i])/(self.mag[i] * self.mag[i])
 */
  __pyx_v_dist = 0.0;

  /* "yt/geometry/selection_routines.pyx":1703
 *                 dot_evec[j] += dist * self.vec[j][i]
 *         dist = 0.0
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             dist += (dot_evec[i] * dot_evec[i])/(self.mag[i] * self.mag[i])
 *         if dist <= 1.0: return 1
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":1704
 *         dist = 0.0
 *         for i in range(3):
 *             dist += (dot_evec[i] * dot_evec[i])/(self.mag[i] * self.mag[i])             # <<<<<<<<<<<<<<
 *         if dist <= 1.0: return 1
 *         return 0
 */
    __pyx_v_dist = (__pyx_v_dist + (((__pyx_v_dot_evec[__pyx_v_i]) * (__pyx_v_dot_evec[__pyx_v_i])) / ((__pyx_v_self->mag[__pyx_v_i]) * (__pyx_v_self->mag[__pyx_v_i]))));
  }

  /* "yt/geometry/selection_routines.pyx":1705
 *         for i in range(3):
 *             dist += (dot_evec[i] * dot_evec[i])/(self.mag[i] * self.mag[i])
 *         if dist <= 1.0: return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_4 = ((__pyx_v_dist <= 1.0) != 0);
  if (__pyx_t_4) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":1706
 *             dist += (dot_evec[i] * dot_evec[i])/(self.mag[i] * self.mag[i])
 *         if dist <= 1.0: return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1692
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef np.float64_t dot_evec[3]
 *         cdef np.float64_t dist
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1711
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         # this is the sphere selection
 *         cdef int i
 */

static int __pyx_f_2yt_8geometry_18selection_routines_17EllipsoidSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_dist;
  __pyx_t_5numpy_float64_t __pyx_v_dist2_max;
  __pyx_t_5numpy_float64_t __pyx_v_dist2;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":1714
 *         # this is the sphere selection
 *         cdef int i
 *         cdef np.float64_t dist, dist2_max, dist2 = 0             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             dist = self.difference(pos[i], self.center[i], i)
 */
  __pyx_v_dist2 = 0.0;

  /* "yt/geometry/selection_routines.pyx":1715
 *         cdef int i
 *         cdef np.float64_t dist, dist2_max, dist2 = 0
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             dist = self.difference(pos[i], self.center[i], i)
 *             dist2 += dist * dist
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/selection_routines.pyx":1716
 *         cdef np.float64_t dist, dist2_max, dist2 = 0
 *         for i in range(3):
 *             dist = self.difference(pos[i], self.center[i], i)             # <<<<<<<<<<<<<<
 *             dist2 += dist * dist
 *         dist2_max = (self.mag[0] + radius) * (self.mag[0] + radius)
 */
    __pyx_v_dist = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_EllipsoidSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.difference(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), (__pyx_v_pos[__pyx_v_i]), (__pyx_v_self->center[__pyx_v_i]), __pyx_v_i);

    /* "yt/geometry/selection_routines.pyx":1717
 *         for i in range(3):
 *             dist = self.difference(pos[i], self.center[i], i)
 *             dist2 += dist * dist             # <<<<<<<<<<<<<<
 *         dist2_max = (self.mag[0] + radius) * (self.mag[0] + radius)
 *         if dist2 <= dist2_max:
 */
    __pyx_v_dist2 = (__pyx_v_dist2 + (__pyx_v_dist * __pyx_v_dist));
  }

  /* "yt/geometry/selection_routines.pyx":1718
 *             dist = self.difference(pos[i], self.center[i], i)
 *             dist2 += dist * dist
 *         dist2_max = (self.mag[0] + radius) * (self.mag[0] + radius)             # <<<<<<<<<<<<<<
 *         if dist2 <= dist2_max:
 *             return 1
 */
  __pyx_v_dist2_max = (((__pyx_v_self->mag[0]) + __pyx_v_radius) * ((__pyx_v_self->mag[0]) + __pyx_v_radius));

  /* "yt/geometry/selection_routines.pyx":1719
 *             dist2 += dist * dist
 *         dist2_max = (self.mag[0] + radius) * (self.mag[0] + radius)
 *         if dist2 <= dist2_max:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  __pyx_t_2 = ((__pyx_v_dist2 <= __pyx_v_dist2_max) != 0);
  if (__pyx_t_2) {

    /* "yt/geometry/selection_routines.pyx":1720
 *         dist2_max = (self.mag[0] + radius) * (self.mag[0] + radius)
 *         if dist2 <= dist2_max:
 *             return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":1719
 *             dist2 += dist * dist
 *         dist2_max = (self.mag[0] + radius) * (self.mag[0] + radius)
 *         if dist2 <= dist2_max:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  }

  /* "yt/geometry/selection_routines.pyx":1721
 *         if dist2 <= dist2_max:
 *             return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1711
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         # this is the sphere selection
 *         cdef int i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1726
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         # This is the sphere selection
 */

static int __pyx_f_2yt_8geometry_18selection_routines_17EllipsoidSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_box_center;
  __pyx_t_5numpy_float64_t __pyx_v_relcenter;
  __pyx_t_5numpy_float64_t __pyx_v_closest;
  __pyx_t_5numpy_float64_t __pyx_v_dist;
  __pyx_t_5numpy_float64_t __pyx_v_edge;
  __pyx_t_5numpy_float64_t __pyx_v_dist_max;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "yt/geometry/selection_routines.pyx":1731
 *         cdef int i
 *         cdef np.float64_t box_center, relcenter, closest, dist, edge, dist_max
 *         if left_edge[0] <= self.center[0] <= right_edge[0] and \             # <<<<<<<<<<<<<<
 *            left_edge[1] <= self.center[1] <= right_edge[1] and \
 *            left_edge[2] <= self.center[2] <= right_edge[2]:
 */
  __pyx_t_2 = ((__pyx_v_left_edge[0]) <= (__pyx_v_self->center[0]));
  if (__pyx_t_2) {
    __pyx_t_2 = ((__pyx_v_self->center[0]) <= (__pyx_v_right_edge[0]));
  }
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":1732
 *         cdef np.float64_t box_center, relcenter, closest, dist, edge, dist_max
 *         if left_edge[0] <= self.center[0] <= right_edge[0] and \
 *            left_edge[1] <= self.center[1] <= right_edge[1] and \             # <<<<<<<<<<<<<<
 *            left_edge[2] <= self.center[2] <= right_edge[2]:
 *             return 1
 */
  __pyx_t_3 = ((__pyx_v_left_edge[1]) <= (__pyx_v_self->center[1]));
  if (__pyx_t_3) {
    __pyx_t_3 = ((__pyx_v_self->center[1]) <= (__pyx_v_right_edge[1]));
  }
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":1733
 *         if left_edge[0] <= self.center[0] <= right_edge[0] and \
 *            left_edge[1] <= self.center[1] <= right_edge[1] and \
 *            left_edge[2] <= self.center[2] <= right_edge[2]:             # <<<<<<<<<<<<<<
 *             return 1
 *         # http://www.gamedev.net/topic/335465-is-this-the-simplest-sphere-aabb-collision-test/
 */
  __pyx_t_2 = ((__pyx_v_left_edge[2]) <= (__pyx_v_self->center[2]));
  if (__pyx_t_2) {
    __pyx_t_2 = ((__pyx_v_self->center[2]) <= (__pyx_v_right_edge[2]));
  }
  __pyx_t_3 = (__pyx_t_2 != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;

  /* "yt/geometry/selection_routines.pyx":1731
 *         cdef int i
 *         cdef np.float64_t box_center, relcenter, closest, dist, edge, dist_max
 *         if left_edge[0] <= self.center[0] <= right_edge[0] and \             # <<<<<<<<<<<<<<
 *            left_edge[1] <= self.center[1] <= right_edge[1] and \
 *            left_edge[2] <= self.center[2] <= right_edge[2]:
 */
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":1734
 *            left_edge[1] <= self.center[1] <= right_edge[1] and \
 *            left_edge[2] <= self.center[2] <= right_edge[2]:
 *             return 1             # <<<<<<<<<<<<<<
 *         # http://www.gamedev.net/topic/335465-is-this-the-simplest-sphere-aabb-collision-test/
 *         dist = 0
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":1731
 *         cdef int i
 *         cdef np.float64_t box_center, relcenter, closest, dist, edge, dist_max
 *         if left_edge[0] <= self.center[0] <= right_edge[0] and \             # <<<<<<<<<<<<<<
 *            left_edge[1] <= self.center[1] <= right_edge[1] and \
 *            left_edge[2] <= self.center[2] <= right_edge[2]:
 */
  }

  /* "yt/geometry/selection_routines.pyx":1736
 *             return 1
 *         # http://www.gamedev.net/topic/335465-is-this-the-simplest-sphere-aabb-collision-test/
 *         dist = 0             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             box_center = (right_edge[i] + left_edge[i])/2.0
 */
  __pyx_v_dist = 0.0;

  /* "yt/geometry/selection_routines.pyx":1737
 *         # http://www.gamedev.net/topic/335465-is-this-the-simplest-sphere-aabb-collision-test/
 *         dist = 0
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             box_center = (right_edge[i] + left_edge[i])/2.0
 *             relcenter = self.difference(box_center, self.center[i], i)
 */
  for (__pyx_t_4 = 0; __pyx_t_4 < 3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "yt/geometry/selection_routines.pyx":1738
 *         dist = 0
 *         for i in range(3):
 *             box_center = (right_edge[i] + left_edge[i])/2.0             # <<<<<<<<<<<<<<
 *             relcenter = self.difference(box_center, self.center[i], i)
 *             edge = right_edge[i] - left_edge[i]
 */
    __pyx_v_box_center = (((__pyx_v_right_edge[__pyx_v_i]) + (__pyx_v_left_edge[__pyx_v_i])) / 2.0);

    /* "yt/geometry/selection_routines.pyx":1739
 *         for i in range(3):
 *             box_center = (right_edge[i] + left_edge[i])/2.0
 *             relcenter = self.difference(box_center, self.center[i], i)             # <<<<<<<<<<<<<<
 *             edge = right_edge[i] - left_edge[i]
 *             closest = relcenter - fclip(relcenter, -edge/2.0, edge/2.0)
 */
    __pyx_v_relcenter = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_EllipsoidSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.difference(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self), __pyx_v_box_center, (__pyx_v_self->center[__pyx_v_i]), __pyx_v_i);

    /* "yt/geometry/selection_routines.pyx":1740
 *             box_center = (right_edge[i] + left_edge[i])/2.0
 *             relcenter = self.difference(box_center, self.center[i], i)
 *             edge = right_edge[i] - left_edge[i]             # <<<<<<<<<<<<<<
 *             closest = relcenter - fclip(relcenter, -edge/2.0, edge/2.0)
 *             dist += closest * closest
 */
    __pyx_v_edge = ((__pyx_v_right_edge[__pyx_v_i]) - (__pyx_v_left_edge[__pyx_v_i]));

    /* "yt/geometry/selection_routines.pyx":1741
 *             relcenter = self.difference(box_center, self.center[i], i)
 *             edge = right_edge[i] - left_edge[i]
 *             closest = relcenter - fclip(relcenter, -edge/2.0, edge/2.0)             # <<<<<<<<<<<<<<
 *             dist += closest * closest
 *         dist_max = self.mag[0] * self.mag[0]
 */
    __pyx_v_closest = (__pyx_v_relcenter - __pyx_f_2yt_9utilities_3lib_8fp_utils_fclip(__pyx_v_relcenter, ((-__pyx_v_edge) / 2.0), (__pyx_v_edge / 2.0)));

    /* "yt/geometry/selection_routines.pyx":1742
 *             edge = right_edge[i] - left_edge[i]
 *             closest = relcenter - fclip(relcenter, -edge/2.0, edge/2.0)
 *             dist += closest * closest             # <<<<<<<<<<<<<<
 *         dist_max = self.mag[0] * self.mag[0]
 *         if dist <= dist_max:
 */
    __pyx_v_dist = (__pyx_v_dist + (__pyx_v_closest * __pyx_v_closest));
  }

  /* "yt/geometry/selection_routines.pyx":1743
 *             closest = relcenter - fclip(relcenter, -edge/2.0, edge/2.0)
 *             dist += closest * closest
 *         dist_max = self.mag[0] * self.mag[0]             # <<<<<<<<<<<<<<
 *         if dist <= dist_max:
 *             return 1
 */
  __pyx_v_dist_max = ((__pyx_v_self->mag[0]) * (__pyx_v_self->mag[0]));

  /* "yt/geometry/selection_routines.pyx":1744
 *             dist += closest * closest
 *         dist_max = self.mag[0] * self.mag[0]
 *         if dist <= dist_max:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  __pyx_t_1 = ((__pyx_v_dist <= __pyx_v_dist_max) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":1745
 *         dist_max = self.mag[0] * self.mag[0]
 *         if dist <= dist_max:
 *             return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":1744
 *             dist += closest * closest
 *         dist_max = self.mag[0] * self.mag[0]
 *         if dist <= dist_max:             # <<<<<<<<<<<<<<
 *             return 1
 *         return 0
 */
  }

  /* "yt/geometry/selection_routines.pyx":1746
 *         if dist <= dist_max:
 *             return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1726
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         # This is the sphere selection
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1748
 *         return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("vec[0][0]", self.vec[0][0]),
 *                 ("vec[0][1]", self.vec[0][1]),
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17EllipsoidSelector_3_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17EllipsoidSelector_3_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_17EllipsoidSelector_2_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17EllipsoidSelector_2_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":1749
 * 
 *     def _hash_vals(self):
 *         return (("vec[0][0]", self.vec[0][0]),             # <<<<<<<<<<<<<<
 *                 ("vec[0][1]", self.vec[0][1]),
 *                 ("vec[0][2]", self.vec[0][2]),
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_self->vec[0])[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_vec_0_0);
  __Pyx_GIVEREF(__pyx_kp_s_vec_0_0);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_s_vec_0_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1750
 *     def _hash_vals(self):
 *         return (("vec[0][0]", self.vec[0][0]),
 *                 ("vec[0][1]", self.vec[0][1]),             # <<<<<<<<<<<<<<
 *                 ("vec[0][2]", self.vec[0][2]),
 *                 ("vec[1][0]", self.vec[1][0]),
 */
  __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_self->vec[0])[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_kp_s_vec_0_1);
  __Pyx_GIVEREF(__pyx_kp_s_vec_0_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_s_vec_0_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1751
 *         return (("vec[0][0]", self.vec[0][0]),
 *                 ("vec[0][1]", self.vec[0][1]),
 *                 ("vec[0][2]", self.vec[0][2]),             # <<<<<<<<<<<<<<
 *                 ("vec[1][0]", self.vec[1][0]),
 *                 ("vec[1][1]", self.vec[1][1]),
 */
  __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_self->vec[0])[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_kp_s_vec_0_2);
  __Pyx_GIVEREF(__pyx_kp_s_vec_0_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_s_vec_0_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1752
 *                 ("vec[0][1]", self.vec[0][1]),
 *                 ("vec[0][2]", self.vec[0][2]),
 *                 ("vec[1][0]", self.vec[1][0]),             # <<<<<<<<<<<<<<
 *                 ("vec[1][1]", self.vec[1][1]),
 *                 ("vec[1][2]", self.vec[1][2]),
 */
  __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_self->vec[1])[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_kp_s_vec_1_0);
  __Pyx_GIVEREF(__pyx_kp_s_vec_1_0);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_s_vec_1_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1753
 *                 ("vec[0][2]", self.vec[0][2]),
 *                 ("vec[1][0]", self.vec[1][0]),
 *                 ("vec[1][1]", self.vec[1][1]),             # <<<<<<<<<<<<<<
 *                 ("vec[1][2]", self.vec[1][2]),
 *                 ("vec[2][0]", self.vec[2][0]),
 */
  __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_self->vec[1])[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_kp_s_vec_1_1);
  __Pyx_GIVEREF(__pyx_kp_s_vec_1_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_vec_1_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1754
 *                 ("vec[1][0]", self.vec[1][0]),
 *                 ("vec[1][1]", self.vec[1][1]),
 *                 ("vec[1][2]", self.vec[1][2]),             # <<<<<<<<<<<<<<
 *                 ("vec[2][0]", self.vec[2][0]),
 *                 ("vec[2][1]", self.vec[2][1]),
 */
  __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_self->vec[1])[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_kp_s_vec_1_2);
  __Pyx_GIVEREF(__pyx_kp_s_vec_1_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_kp_s_vec_1_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1755
 *                 ("vec[1][1]", self.vec[1][1]),
 *                 ("vec[1][2]", self.vec[1][2]),
 *                 ("vec[2][0]", self.vec[2][0]),             # <<<<<<<<<<<<<<
 *                 ("vec[2][1]", self.vec[2][1]),
 *                 ("vec[2][2]", self.vec[2][2]),
 */
  __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_self->vec[2])[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_kp_s_vec_2_0);
  __Pyx_GIVEREF(__pyx_kp_s_vec_2_0);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_kp_s_vec_2_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1756
 *                 ("vec[1][2]", self.vec[1][2]),
 *                 ("vec[2][0]", self.vec[2][0]),
 *                 ("vec[2][1]", self.vec[2][1]),             # <<<<<<<<<<<<<<
 *                 ("vec[2][2]", self.vec[2][2]),
 *                 ("mag[0]", self.mag[0]),
 */
  __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_self->vec[2])[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_kp_s_vec_2_1);
  __Pyx_GIVEREF(__pyx_kp_s_vec_2_1);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_kp_s_vec_2_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1757
 *                 ("vec[2][0]", self.vec[2][0]),
 *                 ("vec[2][1]", self.vec[2][1]),
 *                 ("vec[2][2]", self.vec[2][2]),             # <<<<<<<<<<<<<<
 *                 ("mag[0]", self.mag[0]),
 *                 ("mag[1]", self.mag[1]),
 */
  __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_self->vec[2])[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_kp_s_vec_2_2);
  __Pyx_GIVEREF(__pyx_kp_s_vec_2_2);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_kp_s_vec_2_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1758
 *                 ("vec[2][1]", self.vec[2][1]),
 *                 ("vec[2][2]", self.vec[2][2]),
 *                 ("mag[0]", self.mag[0]),             # <<<<<<<<<<<<<<
 *                 ("mag[1]", self.mag[1]),
 *                 ("mag[2]", self.mag[2]),
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->mag[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_INCREF(__pyx_kp_s_mag_0);
  __Pyx_GIVEREF(__pyx_kp_s_mag_0);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_kp_s_mag_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1759
 *                 ("vec[2][2]", self.vec[2][2]),
 *                 ("mag[0]", self.mag[0]),
 *                 ("mag[1]", self.mag[1]),             # <<<<<<<<<<<<<<
 *                 ("mag[2]", self.mag[2]),
 *                 ("center[0]", self.center[0]),
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->mag[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_INCREF(__pyx_kp_s_mag_1);
  __Pyx_GIVEREF(__pyx_kp_s_mag_1);
  PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_kp_s_mag_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1760
 *                 ("mag[0]", self.mag[0]),
 *                 ("mag[1]", self.mag[1]),
 *                 ("mag[2]", self.mag[2]),             # <<<<<<<<<<<<<<
 *                 ("center[0]", self.center[0]),
 *                 ("center[1]", self.center[1]),
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->mag[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_INCREF(__pyx_kp_s_mag_2);
  __Pyx_GIVEREF(__pyx_kp_s_mag_2);
  PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_kp_s_mag_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1761
 *                 ("mag[1]", self.mag[1]),
 *                 ("mag[2]", self.mag[2]),
 *                 ("center[0]", self.center[0]),             # <<<<<<<<<<<<<<
 *                 ("center[1]", self.center[1]),
 *                 ("center[2]", self.center[2]))
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->center[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_INCREF(__pyx_kp_s_center_0);
  __Pyx_GIVEREF(__pyx_kp_s_center_0);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_kp_s_center_0);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1762
 *                 ("mag[2]", self.mag[2]),
 *                 ("center[0]", self.center[0]),
 *                 ("center[1]", self.center[1]),             # <<<<<<<<<<<<<<
 *                 ("center[2]", self.center[2]))
 * 
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->center[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_INCREF(__pyx_kp_s_center_1);
  __Pyx_GIVEREF(__pyx_kp_s_center_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_kp_s_center_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1763
 *                 ("center[0]", self.center[0]),
 *                 ("center[1]", self.center[1]),
 *                 ("center[2]", self.center[2]))             # <<<<<<<<<<<<<<
 * 
 * ellipsoid_selector = EllipsoidSelector
 */
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->center[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_INCREF(__pyx_kp_s_center_2);
  __Pyx_GIVEREF(__pyx_kp_s_center_2);
  PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_kp_s_center_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1749
 * 
 *     def _hash_vals(self):
 *         return (("vec[0][0]", self.vec[0][0]),             # <<<<<<<<<<<<<<
 *                 ("vec[0][1]", self.vec[0][1]),
 *                 ("vec[0][2]", self.vec[0][2]),
 */
  __pyx_t_1 = PyTuple_New(15); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_1, 11, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_1, 12, __pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_15);
  PyTuple_SET_ITEM(__pyx_t_1, 13, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_16);
  PyTuple_SET_ITEM(__pyx_t_1, 14, __pyx_t_16);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_t_15 = 0;
  __pyx_t_16 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1748
 *         return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (("vec[0][0]", self.vec[0][0]),
 *                 ("vec[0][1]", self.vec[0][1]),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("yt.geometry.selection_routines.EllipsoidSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17EllipsoidSelector_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17EllipsoidSelector_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_17EllipsoidSelector_4__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17EllipsoidSelector_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.EllipsoidSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17EllipsoidSelector_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17EllipsoidSelector_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_17EllipsoidSelector_6__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17EllipsoidSelector_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.EllipsoidSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1770
 *     cdef object ind
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         self.ind = dobj.id - dobj._id_offset
 * 
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1770, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1770, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.GridSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_12GridSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_12GridSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":1771
 * 
 *     def __init__(self, dobj):
 *         self.ind = dobj.id - dobj._id_offset             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_id_offset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Subtract(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->ind);
  __Pyx_DECREF(__pyx_v_self->ind);
  __pyx_v_self->ind = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":1770
 *     cdef object ind
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         self.ind = dobj.id - dobj._id_offset
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("yt.geometry.selection_routines.GridSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1776
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_grids(self,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=2] left_edges,
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_3select_grids(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_3select_grids(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left_edges = 0;
  CYTHON_UNUSED PyArrayObject *__pyx_v_right_edges = 0;
  CYTHON_UNUSED PyArrayObject *__pyx_v_levels = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("select_grids (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left_edges,&__pyx_n_s_right_edges,&__pyx_n_s_levels,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left_edges)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right_edges)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, 1); __PYX_ERR(0, 1776, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_levels)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, 2); __PYX_ERR(0, 1776, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "select_grids") < 0)) __PYX_ERR(0, 1776, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_left_edges = ((PyArrayObject *)values[0]);
    __pyx_v_right_edges = ((PyArrayObject *)values[1]);
    __pyx_v_levels = ((PyArrayObject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1776, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.GridSelector.select_grids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left_edges), __pyx_ptype_5numpy_ndarray, 1, "left_edges", 0))) __PYX_ERR(0, 1777, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right_edges), __pyx_ptype_5numpy_ndarray, 1, "right_edges", 0))) __PYX_ERR(0, 1778, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_levels), __pyx_ptype_5numpy_ndarray, 1, "levels", 0))) __PYX_ERR(0, 1779, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_12GridSelector_2select_grids(((struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *)__pyx_v_self), __pyx_v_left_edges, __pyx_v_right_edges, __pyx_v_levels);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_12GridSelector_2select_grids(struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *__pyx_v_self, PyArrayObject *__pyx_v_left_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_right_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_levels) {
  int __pyx_v_ng;
  PyArrayObject *__pyx_v_gridi = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_gridi;
  __Pyx_Buffer __pyx_pybuffer_gridi;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left_edges;
  __Pyx_Buffer __pyx_pybuffer_left_edges;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_levels;
  __Pyx_Buffer __pyx_pybuffer_levels;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right_edges;
  __Pyx_Buffer __pyx_pybuffer_right_edges;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyArrayObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("select_grids", 0);
  __pyx_pybuffer_gridi.pybuffer.buf = NULL;
  __pyx_pybuffer_gridi.refcount = 0;
  __pyx_pybuffernd_gridi.data = NULL;
  __pyx_pybuffernd_gridi.rcbuffer = &__pyx_pybuffer_gridi;
  __pyx_pybuffer_left_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_left_edges.refcount = 0;
  __pyx_pybuffernd_left_edges.data = NULL;
  __pyx_pybuffernd_left_edges.rcbuffer = &__pyx_pybuffer_left_edges;
  __pyx_pybuffer_right_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_right_edges.refcount = 0;
  __pyx_pybuffernd_right_edges.data = NULL;
  __pyx_pybuffernd_right_edges.rcbuffer = &__pyx_pybuffer_right_edges;
  __pyx_pybuffer_levels.pybuffer.buf = NULL;
  __pyx_pybuffer_levels.refcount = 0;
  __pyx_pybuffernd_levels.data = NULL;
  __pyx_pybuffernd_levels.rcbuffer = &__pyx_pybuffer_levels;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_left_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1776, __pyx_L1_error)
  }
  __pyx_pybuffernd_left_edges.diminfo[0].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left_edges.diminfo[0].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_left_edges.diminfo[1].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_left_edges.diminfo[1].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_right_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1776, __pyx_L1_error)
  }
  __pyx_pybuffernd_right_edges.diminfo[0].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right_edges.diminfo[0].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_right_edges.diminfo[1].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_right_edges.diminfo[1].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_levels.rcbuffer->pybuffer, (PyObject*)__pyx_v_levels, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1776, __pyx_L1_error)
  }
  __pyx_pybuffernd_levels.diminfo[0].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_levels.diminfo[0].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_levels.diminfo[1].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_levels.diminfo[1].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[1];

  /* "yt/geometry/selection_routines.pyx":1780
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 *                      np.ndarray[np.int32_t, ndim=2] levels):
 *         cdef int ng = left_edges.shape[0]             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')
 *         gridi[self.ind] = 1
 */
  __pyx_v_ng = (__pyx_v_left_edges->dimensions[0]);

  /* "yt/geometry/selection_routines.pyx":1781
 *                      np.ndarray[np.int32_t, ndim=2] levels):
 *         cdef int ng = left_edges.shape[0]
 *         cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')             # <<<<<<<<<<<<<<
 *         gridi[self.ind] = 1
 *         return gridi.astype("bool")
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_ng); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 1781, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1781, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_gridi.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_gridi = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_gridi.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1781, __pyx_L1_error)
    } else {__pyx_pybuffernd_gridi.diminfo[0].strides = __pyx_pybuffernd_gridi.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_gridi.diminfo[0].shape = __pyx_pybuffernd_gridi.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_5 = 0;
  __pyx_v_gridi = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":1782
 *         cdef int ng = left_edges.shape[0]
 *         cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')
 *         gridi[self.ind] = 1             # <<<<<<<<<<<<<<
 *         return gridi.astype("bool")
 * 
 */
  if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_gridi), __pyx_v_self->ind, __pyx_int_1) < 0)) __PYX_ERR(0, 1782, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":1783
 *         cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')
 *         gridi[self.ind] = 1
 *         return gridi.astype("bool")             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_gridi), __pyx_n_s_astype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_n_s_bool) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_n_s_bool);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1776
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_grids(self,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=2] left_edges,
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_gridi.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.GridSelector.select_grids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_gridi.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_gridi);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1788
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def fill_mask(self, gobj):             # <<<<<<<<<<<<<<
 *         return np.ones(gobj.ActiveDimensions, dtype='bool')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_5fill_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_gobj); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_5fill_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_gobj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fill_mask (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_12GridSelector_4fill_mask(((struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *)__pyx_v_self), ((PyObject *)__pyx_v_gobj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_12GridSelector_4fill_mask(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *__pyx_v_self, PyObject *__pyx_v_gobj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fill_mask", 0);

  /* "yt/geometry/selection_routines.pyx":1789
 *     @cython.cdivision(True)
 *     def fill_mask(self, gobj):
 *         return np.ones(gobj.ActiveDimensions, dtype='bool')             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_gobj, __pyx_n_s_ActiveDimensions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_bool) < 0) __PYX_ERR(0, 1789, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1788
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def fill_mask(self, gobj):             # <<<<<<<<<<<<<<
 *         return np.ones(gobj.ActiveDimensions, dtype='bool')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("yt.geometry.selection_routines.GridSelector.fill_mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1794
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

static int __pyx_f_2yt_8geometry_18selection_routines_12GridSelector_select_cell(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1795
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1794
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1797
 *         return 1
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         # we apparently don't check if the point actually lies in the grid..
 *         return 1
 */

static int __pyx_f_2yt_8geometry_18selection_routines_12GridSelector_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1799
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         # we apparently don't check if the point actually lies in the grid..
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1797
 *         return 1
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         # we apparently don't check if the point actually lies in the grid..
 *         return 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1801
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (self.ind,)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_7_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_7_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_12GridSelector_6_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_12GridSelector_6_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":1802
 * 
 *     def _hash_vals(self):
 *         return (self.ind,)             # <<<<<<<<<<<<<<
 * 
 * grid_selector = GridSelector
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->ind);
  __Pyx_GIVEREF(__pyx_v_self->ind);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->ind);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1801
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (self.ind,)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.GridSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_12GridSelector_8__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_12GridSelector_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.GridSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_12GridSelector_10__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_12GridSelector_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.GridSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1808
 * cdef class OctreeSubsetSelector(SelectorObject):
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         self.base_selector = dobj.base_selector
 *         self.min_level = self.base_selector.min_level
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1808, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1808, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.OctreeSubsetSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int32_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":1809
 * 
 *     def __init__(self, dobj):
 *         self.base_selector = dobj.base_selector             # <<<<<<<<<<<<<<
 *         self.min_level = self.base_selector.min_level
 *         self.max_level = self.base_selector.max_level
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_base_selector); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject))))) __PYX_ERR(0, 1809, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->base_selector);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->base_selector));
  __pyx_v_self->base_selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1810
 *     def __init__(self, dobj):
 *         self.base_selector = dobj.base_selector
 *         self.min_level = self.base_selector.min_level             # <<<<<<<<<<<<<<
 *         self.max_level = self.base_selector.max_level
 *         self.domain_id = dobj.domain_id
 */
  __pyx_t_2 = __pyx_v_self->base_selector->min_level;
  __pyx_v_self->__pyx_base.min_level = __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":1811
 *         self.base_selector = dobj.base_selector
 *         self.min_level = self.base_selector.min_level
 *         self.max_level = self.base_selector.max_level             # <<<<<<<<<<<<<<
 *         self.domain_id = dobj.domain_id
 *         self.overlap_cells = 1
 */
  __pyx_t_2 = __pyx_v_self->base_selector->max_level;
  __pyx_v_self->__pyx_base.max_level = __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":1812
 *         self.min_level = self.base_selector.min_level
 *         self.max_level = self.base_selector.max_level
 *         self.domain_id = dobj.domain_id             # <<<<<<<<<<<<<<
 *         self.overlap_cells = 1
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_domain_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_3 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1812, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->domain_id = __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":1813
 *         self.max_level = self.base_selector.max_level
 *         self.domain_id = dobj.domain_id
 *         self.overlap_cells = 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_v_self->__pyx_base.overlap_cells = 1;

  /* "yt/geometry/selection_routines.pyx":1808
 * cdef class OctreeSubsetSelector(SelectorObject):
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         self.base_selector = dobj.base_selector
 *         self.min_level = self.base_selector.min_level
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.OctreeSubsetSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1818
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_grids(self,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=2] left_edges,
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_3select_grids(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_3select_grids(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyArrayObject *__pyx_v_left_edges = 0;
  CYTHON_UNUSED PyArrayObject *__pyx_v_right_edges = 0;
  CYTHON_UNUSED PyArrayObject *__pyx_v_levels = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("select_grids (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left_edges,&__pyx_n_s_right_edges,&__pyx_n_s_levels,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left_edges)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right_edges)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, 1); __PYX_ERR(0, 1818, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_levels)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, 2); __PYX_ERR(0, 1818, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "select_grids") < 0)) __PYX_ERR(0, 1818, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_left_edges = ((PyArrayObject *)values[0]);
    __pyx_v_right_edges = ((PyArrayObject *)values[1]);
    __pyx_v_levels = ((PyArrayObject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1818, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.OctreeSubsetSelector.select_grids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left_edges), __pyx_ptype_5numpy_ndarray, 1, "left_edges", 0))) __PYX_ERR(0, 1819, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right_edges), __pyx_ptype_5numpy_ndarray, 1, "right_edges", 0))) __PYX_ERR(0, 1820, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_levels), __pyx_ptype_5numpy_ndarray, 1, "levels", 0))) __PYX_ERR(0, 1821, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_2select_grids(((struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *)__pyx_v_self), __pyx_v_left_edges, __pyx_v_right_edges, __pyx_v_levels);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_2select_grids(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED PyArrayObject *__pyx_v_left_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_right_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_levels) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left_edges;
  __Pyx_Buffer __pyx_pybuffer_left_edges;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_levels;
  __Pyx_Buffer __pyx_pybuffer_levels;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right_edges;
  __Pyx_Buffer __pyx_pybuffer_right_edges;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("select_grids", 0);
  __pyx_pybuffer_left_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_left_edges.refcount = 0;
  __pyx_pybuffernd_left_edges.data = NULL;
  __pyx_pybuffernd_left_edges.rcbuffer = &__pyx_pybuffer_left_edges;
  __pyx_pybuffer_right_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_right_edges.refcount = 0;
  __pyx_pybuffernd_right_edges.data = NULL;
  __pyx_pybuffernd_right_edges.rcbuffer = &__pyx_pybuffer_right_edges;
  __pyx_pybuffer_levels.pybuffer.buf = NULL;
  __pyx_pybuffer_levels.refcount = 0;
  __pyx_pybuffernd_levels.data = NULL;
  __pyx_pybuffernd_levels.rcbuffer = &__pyx_pybuffer_levels;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_left_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1818, __pyx_L1_error)
  }
  __pyx_pybuffernd_left_edges.diminfo[0].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left_edges.diminfo[0].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_left_edges.diminfo[1].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_left_edges.diminfo[1].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_right_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1818, __pyx_L1_error)
  }
  __pyx_pybuffernd_right_edges.diminfo[0].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right_edges.diminfo[0].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_right_edges.diminfo[1].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_right_edges.diminfo[1].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_levels.rcbuffer->pybuffer, (PyObject*)__pyx_v_levels, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1818, __pyx_L1_error)
  }
  __pyx_pybuffernd_levels.diminfo[0].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_levels.diminfo[0].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_levels.diminfo[1].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_levels.diminfo[1].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[1];

  /* "yt/geometry/selection_routines.pyx":1822
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 *                      np.ndarray[np.int32_t, ndim=2] levels):
 *         raise RuntimeError             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_Raise(__pyx_builtin_RuntimeError, 0, 0, 0);
  __PYX_ERR(0, 1822, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":1818
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_grids(self,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=2] left_edges,
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 */

  /* function exit code */
  __pyx_L1_error:;
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.OctreeSubsetSelector.select_grids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1827
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

static int __pyx_f_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_sphere(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1828
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1827
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1833
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

static int __pyx_f_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_cell(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1834
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1833
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1839
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

static int __pyx_f_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1840
 *     @cython.cdivision(True)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1839
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1845
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         return self.base_selector.select_bbox(left_edge, right_edge)
 */

static int __pyx_f_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1847
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 *         return self.base_selector.select_bbox(left_edge, right_edge)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->base_selector->__pyx_vtab)->select_bbox(__pyx_v_self->base_selector, __pyx_v_left_edge, __pyx_v_right_edge);
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1845
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         return self.base_selector.select_bbox(left_edge, right_edge)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1852
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_grid *__pyx_optional_args) {

  /* "yt/geometry/selection_routines.pyx":1854
 *     cdef int select_grid(self, np.float64_t left_edge[3],
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:             # <<<<<<<<<<<<<<
 *         # Because visitors now use select_grid, we should be explicitly
 *         # checking this.
 */
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o = ((struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *)NULL);
  int __pyx_v_res;
  int __pyx_r;
  int __pyx_t_1;
  struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_o = __pyx_optional_args->o;
    }
  }

  /* "yt/geometry/selection_routines.pyx":1858
 *         # checking this.
 *         cdef int res
 *         res = self.base_selector.select_grid(left_edge, right_edge, level, o)             # <<<<<<<<<<<<<<
 *         if self.domain_id == -1:
 *             return res
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.o = __pyx_v_o;
  __pyx_t_1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->base_selector->__pyx_vtab)->select_grid(__pyx_v_self->base_selector, __pyx_v_left_edge, __pyx_v_right_edge, __pyx_v_level, &__pyx_t_2); 
  __pyx_v_res = __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":1859
 *         cdef int res
 *         res = self.base_selector.select_grid(left_edge, right_edge, level, o)
 *         if self.domain_id == -1:             # <<<<<<<<<<<<<<
 *             return res
 *         elif res == 1 and o != NULL and o.domain != self.domain_id:
 */
  __pyx_t_3 = ((__pyx_v_self->domain_id == -1LL) != 0);
  if (__pyx_t_3) {

    /* "yt/geometry/selection_routines.pyx":1860
 *         res = self.base_selector.select_grid(left_edge, right_edge, level, o)
 *         if self.domain_id == -1:
 *             return res             # <<<<<<<<<<<<<<
 *         elif res == 1 and o != NULL and o.domain != self.domain_id:
 *             return -1
 */
    __pyx_r = __pyx_v_res;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":1859
 *         cdef int res
 *         res = self.base_selector.select_grid(left_edge, right_edge, level, o)
 *         if self.domain_id == -1:             # <<<<<<<<<<<<<<
 *             return res
 *         elif res == 1 and o != NULL and o.domain != self.domain_id:
 */
  }

  /* "yt/geometry/selection_routines.pyx":1861
 *         if self.domain_id == -1:
 *             return res
 *         elif res == 1 and o != NULL and o.domain != self.domain_id:             # <<<<<<<<<<<<<<
 *             return -1
 *         return res
 */
  __pyx_t_4 = ((__pyx_v_res == 1) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_v_o != NULL) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_v_o->domain != __pyx_v_self->domain_id) != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "yt/geometry/selection_routines.pyx":1862
 *             return res
 *         elif res == 1 and o != NULL and o.domain != self.domain_id:
 *             return -1             # <<<<<<<<<<<<<<
 *         return res
 * 
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":1861
 *         if self.domain_id == -1:
 *             return res
 *         elif res == 1 and o != NULL and o.domain != self.domain_id:             # <<<<<<<<<<<<<<
 *             return -1
 *         return res
 */
  }

  /* "yt/geometry/selection_routines.pyx":1863
 *         elif res == 1 and o != NULL and o.domain != self.domain_id:
 *             return -1
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1852
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1865
 *         return res
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (hash(self.base_selector), self.domain_id)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_5_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_5_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_4_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_4_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":1866
 * 
 *     def _hash_vals(self):
 *         return (hash(self.base_selector), self.domain_id)             # <<<<<<<<<<<<<<
 * 
 * octree_subset_selector = OctreeSubsetSelector
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->base_selector);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(0, 1866, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->domain_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1865
 *         return res
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (hash(self.base_selector), self.domain_id)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("yt.geometry.selection_routines.OctreeSubsetSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pxd":72
 * cdef class OctreeSubsetSelector(SelectorObject):
 *     cdef SelectorObject base_selector
 *     cdef public np.int64_t domain_id             # <<<<<<<<<<<<<<
 * 
 * cdef class BooleanSelector(SelectorObject):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_9domain_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_9domain_id_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_9domain_id___get__(((struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_9domain_id___get__(struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->domain_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.OctreeSubsetSelector.domain_id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_9domain_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_9domain_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_9domain_id_2__set__(((struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_9domain_id_2__set__(struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_int64_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int64(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 72, __pyx_L1_error)
  __pyx_v_self->domain_id = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.OctreeSubsetSelector.domain_id.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_6__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.OctreeSubsetSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_8__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.OctreeSubsetSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1879
 *     cdef np.float64_t DRE[3]
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         self.min_ind = dobj.min_ind
 *         self.max_ind = dobj.max_ind
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1879, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1879, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.IndexedOctreeSubsetSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  long __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_uint64_t __pyx_t_2;
  __pyx_t_5numpy_int32_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  long __pyx_t_6;
  __pyx_t_5numpy_float64_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":1880
 * 
 *     def __init__(self, dobj):
 *         self.min_ind = dobj.min_ind             # <<<<<<<<<<<<<<
 *         self.max_ind = dobj.max_ind
 *         self.base_selector = dobj.base_selector
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_min_ind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_uint64(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->min_ind = __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":1881
 *     def __init__(self, dobj):
 *         self.min_ind = dobj.min_ind
 *         self.max_ind = dobj.max_ind             # <<<<<<<<<<<<<<
 *         self.base_selector = dobj.base_selector
 *         self.min_level = self.base_selector.min_level
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_max_ind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_npy_uint64(__pyx_t_1); if (unlikely((__pyx_t_2 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1881, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->max_ind = __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":1882
 *         self.min_ind = dobj.min_ind
 *         self.max_ind = dobj.max_ind
 *         self.base_selector = dobj.base_selector             # <<<<<<<<<<<<<<
 *         self.min_level = self.base_selector.min_level
 *         self.max_level = self.base_selector.max_level
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_base_selector); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject))))) __PYX_ERR(0, 1882, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->base_selector);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->base_selector));
  __pyx_v_self->base_selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1883
 *         self.max_ind = dobj.max_ind
 *         self.base_selector = dobj.base_selector
 *         self.min_level = self.base_selector.min_level             # <<<<<<<<<<<<<<
 *         self.max_level = self.base_selector.max_level
 *         self.filter_bbox = 0
 */
  __pyx_t_3 = __pyx_v_self->base_selector->min_level;
  __pyx_v_self->__pyx_base.min_level = __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":1884
 *         self.base_selector = dobj.base_selector
 *         self.min_level = self.base_selector.min_level
 *         self.max_level = self.base_selector.max_level             # <<<<<<<<<<<<<<
 *         self.filter_bbox = 0
 *         if getattr(dobj.ds, "filter_bbox", False):
 */
  __pyx_t_3 = __pyx_v_self->base_selector->max_level;
  __pyx_v_self->__pyx_base.max_level = __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":1885
 *         self.min_level = self.base_selector.min_level
 *         self.max_level = self.base_selector.max_level
 *         self.filter_bbox = 0             # <<<<<<<<<<<<<<
 *         if getattr(dobj.ds, "filter_bbox", False):
 *             self.filter_bbox = 1
 */
  __pyx_v_self->filter_bbox = 0;

  /* "yt/geometry/selection_routines.pyx":1886
 *         self.max_level = self.base_selector.max_level
 *         self.filter_bbox = 0
 *         if getattr(dobj.ds, "filter_bbox", False):             # <<<<<<<<<<<<<<
 *             self.filter_bbox = 1
 *         for i in range(3):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_ds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_GetAttr3(__pyx_t_1, __pyx_n_s_filter_bbox, Py_False); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1886, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {

    /* "yt/geometry/selection_routines.pyx":1887
 *         self.filter_bbox = 0
 *         if getattr(dobj.ds, "filter_bbox", False):
 *             self.filter_bbox = 1             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             self.DLE[i] = dobj.ds.domain_left_edge[i]
 */
    __pyx_v_self->filter_bbox = 1;

    /* "yt/geometry/selection_routines.pyx":1886
 *         self.max_level = self.base_selector.max_level
 *         self.filter_bbox = 0
 *         if getattr(dobj.ds, "filter_bbox", False):             # <<<<<<<<<<<<<<
 *             self.filter_bbox = 1
 *         for i in range(3):
 */
  }

  /* "yt/geometry/selection_routines.pyx":1888
 *         if getattr(dobj.ds, "filter_bbox", False):
 *             self.filter_bbox = 1
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             self.DLE[i] = dobj.ds.domain_left_edge[i]
 *             self.DRE[i] = dobj.ds.domain_right_edge[i]
 */
  for (__pyx_t_6 = 0; __pyx_t_6 < 3; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "yt/geometry/selection_routines.pyx":1889
 *             self.filter_bbox = 1
 *         for i in range(3):
 *             self.DLE[i] = dobj.ds.domain_left_edge[i]             # <<<<<<<<<<<<<<
 *             self.DRE[i] = dobj.ds.domain_right_edge[i]
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_ds); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_domain_left_edge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_i, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_7 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1889, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    (__pyx_v_self->DLE[__pyx_v_i]) = __pyx_t_7;

    /* "yt/geometry/selection_routines.pyx":1890
 *         for i in range(3):
 *             self.DLE[i] = dobj.ds.domain_left_edge[i]
 *             self.DRE[i] = dobj.ds.domain_right_edge[i]             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_ds); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_domain_right_edge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_i, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_7 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1890, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    (__pyx_v_self->DRE[__pyx_v_i]) = __pyx_t_7;
  }

  /* "yt/geometry/selection_routines.pyx":1879
 *     cdef np.float64_t DRE[3]
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         self.min_ind = dobj.min_ind
 *         self.max_ind = dobj.max_ind
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("yt.geometry.selection_routines.IndexedOctreeSubsetSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1895
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_grids(self,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=2] left_edges,
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_3select_grids(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_3select_grids(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyArrayObject *__pyx_v_left_edges = 0;
  CYTHON_UNUSED PyArrayObject *__pyx_v_right_edges = 0;
  CYTHON_UNUSED PyArrayObject *__pyx_v_levels = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("select_grids (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left_edges,&__pyx_n_s_right_edges,&__pyx_n_s_levels,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left_edges)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right_edges)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, 1); __PYX_ERR(0, 1895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_levels)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, 2); __PYX_ERR(0, 1895, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "select_grids") < 0)) __PYX_ERR(0, 1895, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_left_edges = ((PyArrayObject *)values[0]);
    __pyx_v_right_edges = ((PyArrayObject *)values[1]);
    __pyx_v_levels = ((PyArrayObject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1895, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.IndexedOctreeSubsetSelector.select_grids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left_edges), __pyx_ptype_5numpy_ndarray, 1, "left_edges", 0))) __PYX_ERR(0, 1896, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right_edges), __pyx_ptype_5numpy_ndarray, 1, "right_edges", 0))) __PYX_ERR(0, 1897, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_levels), __pyx_ptype_5numpy_ndarray, 1, "levels", 0))) __PYX_ERR(0, 1898, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_2select_grids(((struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *)__pyx_v_self), __pyx_v_left_edges, __pyx_v_right_edges, __pyx_v_levels);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_2select_grids(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED PyArrayObject *__pyx_v_left_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_right_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_levels) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left_edges;
  __Pyx_Buffer __pyx_pybuffer_left_edges;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_levels;
  __Pyx_Buffer __pyx_pybuffer_levels;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right_edges;
  __Pyx_Buffer __pyx_pybuffer_right_edges;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("select_grids", 0);
  __pyx_pybuffer_left_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_left_edges.refcount = 0;
  __pyx_pybuffernd_left_edges.data = NULL;
  __pyx_pybuffernd_left_edges.rcbuffer = &__pyx_pybuffer_left_edges;
  __pyx_pybuffer_right_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_right_edges.refcount = 0;
  __pyx_pybuffernd_right_edges.data = NULL;
  __pyx_pybuffernd_right_edges.rcbuffer = &__pyx_pybuffer_right_edges;
  __pyx_pybuffer_levels.pybuffer.buf = NULL;
  __pyx_pybuffer_levels.refcount = 0;
  __pyx_pybuffernd_levels.data = NULL;
  __pyx_pybuffernd_levels.rcbuffer = &__pyx_pybuffer_levels;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_left_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1895, __pyx_L1_error)
  }
  __pyx_pybuffernd_left_edges.diminfo[0].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left_edges.diminfo[0].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_left_edges.diminfo[1].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_left_edges.diminfo[1].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_right_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1895, __pyx_L1_error)
  }
  __pyx_pybuffernd_right_edges.diminfo[0].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right_edges.diminfo[0].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_right_edges.diminfo[1].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_right_edges.diminfo[1].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_levels.rcbuffer->pybuffer, (PyObject*)__pyx_v_levels, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1895, __pyx_L1_error)
  }
  __pyx_pybuffernd_levels.diminfo[0].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_levels.diminfo[0].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_levels.diminfo[1].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_levels.diminfo[1].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[1];

  /* "yt/geometry/selection_routines.pyx":1899
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 *                      np.ndarray[np.int32_t, ndim=2] levels):
 *         raise RuntimeError             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_Raise(__pyx_builtin_RuntimeError, 0, 0, 0);
  __PYX_ERR(0, 1899, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":1895
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_grids(self,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=2] left_edges,
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 */

  /* function exit code */
  __pyx_L1_error:;
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.IndexedOctreeSubsetSelector.select_grids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1904
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

static int __pyx_f_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_sphere(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1905
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1904
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1910
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

static int __pyx_f_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_cell(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1911
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1910
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1916
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         if self.filter_bbox == 0:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_v_i;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":1918
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         cdef int i
 *         if self.filter_bbox == 0:             # <<<<<<<<<<<<<<
 *             return 1
 *         for i in range(3):
 */
  __pyx_t_1 = ((__pyx_v_self->filter_bbox == 0) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":1919
 *         cdef int i
 *         if self.filter_bbox == 0:
 *             return 1             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             if pos[i] < self.DLE[i] or pos[i] > self.DRE[i]:
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":1918
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         cdef int i
 *         if self.filter_bbox == 0:             # <<<<<<<<<<<<<<
 *             return 1
 *         for i in range(3):
 */
  }

  /* "yt/geometry/selection_routines.pyx":1920
 *         if self.filter_bbox == 0:
 *             return 1
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             if pos[i] < self.DLE[i] or pos[i] > self.DRE[i]:
 *                 return 0
 */
  for (__pyx_t_2 = 0; __pyx_t_2 < 3; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "yt/geometry/selection_routines.pyx":1921
 *             return 1
 *         for i in range(3):
 *             if pos[i] < self.DLE[i] or pos[i] > self.DRE[i]:             # <<<<<<<<<<<<<<
 *                 return 0
 *         return 1
 */
    __pyx_t_3 = (((__pyx_v_pos[__pyx_v_i]) < (__pyx_v_self->DLE[__pyx_v_i])) != 0);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_3 = (((__pyx_v_pos[__pyx_v_i]) > (__pyx_v_self->DRE[__pyx_v_i])) != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_1) {

      /* "yt/geometry/selection_routines.pyx":1922
 *         for i in range(3):
 *             if pos[i] < self.DLE[i] or pos[i] > self.DRE[i]:
 *                 return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "yt/geometry/selection_routines.pyx":1921
 *             return 1
 *         for i in range(3):
 *             if pos[i] < self.DLE[i] or pos[i] > self.DRE[i]:             # <<<<<<<<<<<<<<
 *                 return 0
 *         return 1
 */
    }
  }

  /* "yt/geometry/selection_routines.pyx":1923
 *             if pos[i] < self.DLE[i] or pos[i] > self.DRE[i]:
 *                 return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1916
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         if self.filter_bbox == 0:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1928
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         return self.base_selector.select_bbox(left_edge, right_edge)
 */

static int __pyx_f_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1930
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 *         return self.base_selector.select_bbox(left_edge, right_edge)             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],
 */
  __pyx_r = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->base_selector->__pyx_vtab)->select_bbox(__pyx_v_self->base_selector, __pyx_v_left_edge, __pyx_v_right_edge);
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1928
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         return self.base_selector.select_bbox(left_edge, right_edge)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1932
 *         return self.base_selector.select_bbox(left_edge, right_edge)
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_grid *__pyx_optional_args) {

  /* "yt/geometry/selection_routines.pyx":1934
 *     cdef int select_grid(self, np.float64_t left_edge[3],
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:             # <<<<<<<<<<<<<<
 *         # Because visitors now use select_grid, we should be explicitly
 *         # checking this.
 */
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o = ((struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *)NULL);
  int __pyx_r;
  int __pyx_t_1;
  struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid __pyx_t_2;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_o = __pyx_optional_args->o;
    }
  }

  /* "yt/geometry/selection_routines.pyx":1937
 *         # Because visitors now use select_grid, we should be explicitly
 *         # checking this.
 *         return self.base_selector.select_grid(left_edge, right_edge, level, o)             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.o = __pyx_v_o;
  __pyx_t_1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->base_selector->__pyx_vtab)->select_grid(__pyx_v_self->base_selector, __pyx_v_left_edge, __pyx_v_right_edge, __pyx_v_level, &__pyx_t_2); 
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1932
 *         return self.base_selector.select_bbox(left_edge, right_edge)
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1939
 *         return self.base_selector.select_grid(left_edge, right_edge, level, o)
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (hash(self.base_selector), self.min_ind, self.max_ind)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_5_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_5_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_4_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_4_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":1940
 * 
 *     def _hash_vals(self):
 *         return (hash(self.base_selector), self.min_ind, self.max_ind)             # <<<<<<<<<<<<<<
 * 
 * indexed_octree_subset_selector = IndexedOctreeSubsetSelector
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->base_selector);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(0, 1940, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->min_ind); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->max_ind); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1939
 *         return self.base_selector.select_grid(left_edge, right_edge, level, o)
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (hash(self.base_selector), self.min_ind, self.max_ind)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("yt.geometry.selection_routines.IndexedOctreeSubsetSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_6__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.IndexedOctreeSubsetSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_8__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.IndexedOctreeSubsetSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1946
 * cdef class AlwaysSelector(SelectorObject):
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         self.overlap_cells = 1
 * 
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_14AlwaysSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_14AlwaysSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1946, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1946, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.AlwaysSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14AlwaysSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_14AlwaysSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_dobj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":1947
 * 
 *     def __init__(self, dobj):
 *         self.overlap_cells = 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_v_self->__pyx_base.overlap_cells = 1;

  /* "yt/geometry/selection_routines.pyx":1946
 * cdef class AlwaysSelector(SelectorObject):
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         self.overlap_cells = 1
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1952
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_grids(self,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=2] left_edges,
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14AlwaysSelector_3select_grids(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14AlwaysSelector_3select_grids(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left_edges = 0;
  CYTHON_UNUSED PyArrayObject *__pyx_v_right_edges = 0;
  CYTHON_UNUSED PyArrayObject *__pyx_v_levels = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("select_grids (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left_edges,&__pyx_n_s_right_edges,&__pyx_n_s_levels,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left_edges)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right_edges)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, 1); __PYX_ERR(0, 1952, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_levels)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, 2); __PYX_ERR(0, 1952, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "select_grids") < 0)) __PYX_ERR(0, 1952, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_left_edges = ((PyArrayObject *)values[0]);
    __pyx_v_right_edges = ((PyArrayObject *)values[1]);
    __pyx_v_levels = ((PyArrayObject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1952, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.AlwaysSelector.select_grids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left_edges), __pyx_ptype_5numpy_ndarray, 1, "left_edges", 0))) __PYX_ERR(0, 1953, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right_edges), __pyx_ptype_5numpy_ndarray, 1, "right_edges", 0))) __PYX_ERR(0, 1954, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_levels), __pyx_ptype_5numpy_ndarray, 1, "levels", 0))) __PYX_ERR(0, 1955, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14AlwaysSelector_2select_grids(((struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *)__pyx_v_self), __pyx_v_left_edges, __pyx_v_right_edges, __pyx_v_levels);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14AlwaysSelector_2select_grids(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self, PyArrayObject *__pyx_v_left_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_right_edges, CYTHON_UNUSED PyArrayObject *__pyx_v_levels) {
  int __pyx_v_ng;
  PyArrayObject *__pyx_v_gridi = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_gridi;
  __Pyx_Buffer __pyx_pybuffer_gridi;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left_edges;
  __Pyx_Buffer __pyx_pybuffer_left_edges;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_levels;
  __Pyx_Buffer __pyx_pybuffer_levels;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right_edges;
  __Pyx_Buffer __pyx_pybuffer_right_edges;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyArrayObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("select_grids", 0);
  __pyx_pybuffer_gridi.pybuffer.buf = NULL;
  __pyx_pybuffer_gridi.refcount = 0;
  __pyx_pybuffernd_gridi.data = NULL;
  __pyx_pybuffernd_gridi.rcbuffer = &__pyx_pybuffer_gridi;
  __pyx_pybuffer_left_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_left_edges.refcount = 0;
  __pyx_pybuffernd_left_edges.data = NULL;
  __pyx_pybuffernd_left_edges.rcbuffer = &__pyx_pybuffer_left_edges;
  __pyx_pybuffer_right_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_right_edges.refcount = 0;
  __pyx_pybuffernd_right_edges.data = NULL;
  __pyx_pybuffernd_right_edges.rcbuffer = &__pyx_pybuffer_right_edges;
  __pyx_pybuffer_levels.pybuffer.buf = NULL;
  __pyx_pybuffer_levels.refcount = 0;
  __pyx_pybuffernd_levels.data = NULL;
  __pyx_pybuffernd_levels.rcbuffer = &__pyx_pybuffer_levels;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_left_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1952, __pyx_L1_error)
  }
  __pyx_pybuffernd_left_edges.diminfo[0].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left_edges.diminfo[0].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_left_edges.diminfo[1].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_left_edges.diminfo[1].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_right_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1952, __pyx_L1_error)
  }
  __pyx_pybuffernd_right_edges.diminfo[0].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right_edges.diminfo[0].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_right_edges.diminfo[1].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_right_edges.diminfo[1].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_levels.rcbuffer->pybuffer, (PyObject*)__pyx_v_levels, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1952, __pyx_L1_error)
  }
  __pyx_pybuffernd_levels.diminfo[0].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_levels.diminfo[0].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_levels.diminfo[1].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_levels.diminfo[1].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[1];

  /* "yt/geometry/selection_routines.pyx":1956
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 *                      np.ndarray[np.int32_t, ndim=2] levels):
 *         cdef int ng = left_edges.shape[0]             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.ones(ng, dtype='uint8')
 *         return gridi.astype("bool")
 */
  __pyx_v_ng = (__pyx_v_left_edges->dimensions[0]);

  /* "yt/geometry/selection_routines.pyx":1957
 *                      np.ndarray[np.int32_t, ndim=2] levels):
 *         cdef int ng = left_edges.shape[0]
 *         cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.ones(ng, dtype='uint8')             # <<<<<<<<<<<<<<
 *         return gridi.astype("bool")
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_ng); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 1957, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1957, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_gridi.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_gridi = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_gridi.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1957, __pyx_L1_error)
    } else {__pyx_pybuffernd_gridi.diminfo[0].strides = __pyx_pybuffernd_gridi.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_gridi.diminfo[0].shape = __pyx_pybuffernd_gridi.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_5 = 0;
  __pyx_v_gridi = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":1958
 *         cdef int ng = left_edges.shape[0]
 *         cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.ones(ng, dtype='uint8')
 *         return gridi.astype("bool")             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_gridi), __pyx_n_s_astype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_n_s_bool) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_n_s_bool);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1952
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_grids(self,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=2] left_edges,
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_gridi.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.AlwaysSelector.select_grids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_gridi.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_gridi);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1963
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14AlwaysSelector_select_cell(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1964
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1963
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1966
 *         return 1
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14AlwaysSelector_select_grid(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge, CYTHON_UNUSED __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14AlwaysSelector_select_grid *__pyx_optional_args) {
  int __pyx_r;
  if (__pyx_optional_args) {
  }

  /* "yt/geometry/selection_routines.pyx":1969
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1966
 *         return 1
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1971
 *         return 1
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14AlwaysSelector_select_point(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1972
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1971
 *         return 1
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1974
 *         return 1
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14AlwaysSelector_select_sphere(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_pos, CYTHON_UNUSED __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1975
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1974
 *         return 1
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1977
 *         return 1
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         return 1
 */

static int __pyx_f_2yt_8geometry_18selection_routines_14AlwaysSelector_select_bbox(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":1979
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1977
 *         return 1
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         return 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1981
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return ("always", 1,)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14AlwaysSelector_5_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14AlwaysSelector_5_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14AlwaysSelector_4_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14AlwaysSelector_4_hash_vals(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":1982
 * 
 *     def _hash_vals(self):
 *         return ("always", 1,)             # <<<<<<<<<<<<<<
 * 
 * always_selector = AlwaysSelector
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_tuple__37);
  __pyx_r = __pyx_tuple__37;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1981
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return ("always", 1,)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14AlwaysSelector_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14AlwaysSelector_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14AlwaysSelector_6__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14AlwaysSelector_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__38, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.AlwaysSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14AlwaysSelector_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_14AlwaysSelector_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_14AlwaysSelector_8__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_14AlwaysSelector_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__39, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.AlwaysSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1990
 *     cdef SelectorObject selector2
 * 
 *     def __init__(self, dobj, selector1, selector2):             # <<<<<<<<<<<<<<
 *         self.selector1 = selector1
 *         self.selector2 = selector2
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_15ComposeSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_15ComposeSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_dobj = 0;
  PyObject *__pyx_v_selector1 = 0;
  PyObject *__pyx_v_selector2 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,&__pyx_n_s_selector1,&__pyx_n_s_selector2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_selector1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 1990, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_selector2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 1990, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1990, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_dobj = values[0];
    __pyx_v_selector1 = values[1];
    __pyx_v_selector2 = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1990, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.ComposeSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_15ComposeSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *)__pyx_v_self), __pyx_v_dobj, __pyx_v_selector1, __pyx_v_selector2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_15ComposeSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_dobj, PyObject *__pyx_v_selector1, PyObject *__pyx_v_selector2) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":1991
 * 
 *     def __init__(self, dobj, selector1, selector2):
 *         self.selector1 = selector1             # <<<<<<<<<<<<<<
 *         self.selector2 = selector2
 * 
 */
  if (!(likely(((__pyx_v_selector1) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_selector1, __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject))))) __PYX_ERR(0, 1991, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_selector1;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->selector1);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->selector1));
  __pyx_v_self->selector1 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1992
 *     def __init__(self, dobj, selector1, selector2):
 *         self.selector1 = selector1
 *         self.selector2 = selector2             # <<<<<<<<<<<<<<
 * 
 *     def select_grids(self,
 */
  if (!(likely(((__pyx_v_selector2) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_selector2, __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject))))) __PYX_ERR(0, 1992, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_selector2;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->selector2);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->selector2));
  __pyx_v_self->selector2 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":1990
 *     cdef SelectorObject selector2
 * 
 *     def __init__(self, dobj, selector1, selector2):             # <<<<<<<<<<<<<<
 *         self.selector1 = selector1
 *         self.selector2 = selector2
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.ComposeSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":1994
 *         self.selector2 = selector2
 * 
 *     def select_grids(self,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=2] left_edges,
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_15ComposeSelector_3select_grids(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_15ComposeSelector_3select_grids(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left_edges = 0;
  PyArrayObject *__pyx_v_right_edges = 0;
  PyArrayObject *__pyx_v_levels = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("select_grids (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left_edges,&__pyx_n_s_right_edges,&__pyx_n_s_levels,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left_edges)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right_edges)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, 1); __PYX_ERR(0, 1994, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_levels)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, 2); __PYX_ERR(0, 1994, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "select_grids") < 0)) __PYX_ERR(0, 1994, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_left_edges = ((PyArrayObject *)values[0]);
    __pyx_v_right_edges = ((PyArrayObject *)values[1]);
    __pyx_v_levels = ((PyArrayObject *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("select_grids", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1994, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.ComposeSelector.select_grids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left_edges), __pyx_ptype_5numpy_ndarray, 1, "left_edges", 0))) __PYX_ERR(0, 1995, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right_edges), __pyx_ptype_5numpy_ndarray, 1, "right_edges", 0))) __PYX_ERR(0, 1996, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_levels), __pyx_ptype_5numpy_ndarray, 1, "levels", 0))) __PYX_ERR(0, 1997, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_15ComposeSelector_2select_grids(((struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *)__pyx_v_self), __pyx_v_left_edges, __pyx_v_right_edges, __pyx_v_levels);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_15ComposeSelector_2select_grids(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self, PyArrayObject *__pyx_v_left_edges, PyArrayObject *__pyx_v_right_edges, PyArrayObject *__pyx_v_levels) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left_edges;
  __Pyx_Buffer __pyx_pybuffer_left_edges;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_levels;
  __Pyx_Buffer __pyx_pybuffer_levels;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right_edges;
  __Pyx_Buffer __pyx_pybuffer_right_edges;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("select_grids", 0);
  __pyx_pybuffer_left_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_left_edges.refcount = 0;
  __pyx_pybuffernd_left_edges.data = NULL;
  __pyx_pybuffernd_left_edges.rcbuffer = &__pyx_pybuffer_left_edges;
  __pyx_pybuffer_right_edges.pybuffer.buf = NULL;
  __pyx_pybuffer_right_edges.refcount = 0;
  __pyx_pybuffernd_right_edges.data = NULL;
  __pyx_pybuffernd_right_edges.rcbuffer = &__pyx_pybuffer_right_edges;
  __pyx_pybuffer_levels.pybuffer.buf = NULL;
  __pyx_pybuffer_levels.refcount = 0;
  __pyx_pybuffernd_levels.data = NULL;
  __pyx_pybuffernd_levels.rcbuffer = &__pyx_pybuffer_levels;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_left_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1994, __pyx_L1_error)
  }
  __pyx_pybuffernd_left_edges.diminfo[0].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left_edges.diminfo[0].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_left_edges.diminfo[1].strides = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_left_edges.diminfo[1].shape = __pyx_pybuffernd_left_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer, (PyObject*)__pyx_v_right_edges, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1994, __pyx_L1_error)
  }
  __pyx_pybuffernd_right_edges.diminfo[0].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right_edges.diminfo[0].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_right_edges.diminfo[1].strides = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_right_edges.diminfo[1].shape = __pyx_pybuffernd_right_edges.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_levels.rcbuffer->pybuffer, (PyObject*)__pyx_v_levels, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1994, __pyx_L1_error)
  }
  __pyx_pybuffernd_levels.diminfo[0].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_levels.diminfo[0].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_levels.diminfo[1].strides = __pyx_pybuffernd_levels.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_levels.diminfo[1].shape = __pyx_pybuffernd_levels.rcbuffer->pybuffer.shape[1];

  /* "yt/geometry/selection_routines.pyx":1998
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 *                      np.ndarray[np.int32_t, ndim=2] levels):
 *         return np.logical_or(             # <<<<<<<<<<<<<<
 *                     self.selector1.select_grids(left_edges, right_edges, levels),
 *                     self.selector2.select_grids(left_edges, right_edges, levels))
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_logical_or); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":1999
 *                      np.ndarray[np.int32_t, ndim=2] levels):
 *         return np.logical_or(
 *                     self.selector1.select_grids(left_edges, right_edges, levels),             # <<<<<<<<<<<<<<
 *                     self.selector2.select_grids(left_edges, right_edges, levels))
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->selector1), __pyx_n_s_select_grids); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, ((PyObject *)__pyx_v_left_edges), ((PyObject *)__pyx_v_right_edges), ((PyObject *)__pyx_v_levels)};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1999, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, ((PyObject *)__pyx_v_left_edges), ((PyObject *)__pyx_v_right_edges), ((PyObject *)__pyx_v_levels)};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1999, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_left_edges));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_left_edges));
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_left_edges));
    __Pyx_INCREF(((PyObject *)__pyx_v_right_edges));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_right_edges));
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, ((PyObject *)__pyx_v_right_edges));
    __Pyx_INCREF(((PyObject *)__pyx_v_levels));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_levels));
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, ((PyObject *)__pyx_v_levels));
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "yt/geometry/selection_routines.pyx":2000
 *         return np.logical_or(
 *                     self.selector1.select_grids(left_edges, right_edges, levels),
 *                     self.selector2.select_grids(left_edges, right_edges, levels))             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->selector2), __pyx_n_s_select_grids); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_7)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, ((PyObject *)__pyx_v_left_edges), ((PyObject *)__pyx_v_right_edges), ((PyObject *)__pyx_v_levels)};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2000, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, ((PyObject *)__pyx_v_left_edges), ((PyObject *)__pyx_v_right_edges), ((PyObject *)__pyx_v_levels)};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2000, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2000, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_left_edges));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_left_edges));
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_6, ((PyObject *)__pyx_v_left_edges));
    __Pyx_INCREF(((PyObject *)__pyx_v_right_edges));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_right_edges));
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_6, ((PyObject *)__pyx_v_right_edges));
    __Pyx_INCREF(((PyObject *)__pyx_v_levels));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_levels));
    PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_6, ((PyObject *)__pyx_v_levels));
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2000, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1998, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1998, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_6, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_6, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":1994
 *         self.selector2 = selector2
 * 
 *     def select_grids(self,             # <<<<<<<<<<<<<<
 *                      np.ndarray[np.float64_t, ndim=2] left_edges,
 *                      np.ndarray[np.float64_t, ndim=2] right_edges,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.ComposeSelector.select_grids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left_edges.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_levels.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right_edges.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2002
 *                     self.selector2.select_grids(left_edges, right_edges, levels))
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         if self.selector1.select_cell(pos, dds) and \
 *                 self.selector2.select_cell(pos, dds):
 */

static int __pyx_f_2yt_8geometry_18selection_routines_15ComposeSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":2003
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         if self.selector1.select_cell(pos, dds) and \             # <<<<<<<<<<<<<<
 *                 self.selector2.select_cell(pos, dds):
 *             return 1
 */
  __pyx_t_2 = (((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->selector1->__pyx_vtab)->select_cell(__pyx_v_self->selector1, __pyx_v_pos, __pyx_v_dds) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":2004
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         if self.selector1.select_cell(pos, dds) and \
 *                 self.selector2.select_cell(pos, dds):             # <<<<<<<<<<<<<<
 *             return 1
 *         else:
 */
  __pyx_t_2 = (((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->selector2->__pyx_vtab)->select_cell(__pyx_v_self->selector2, __pyx_v_pos, __pyx_v_dds) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "yt/geometry/selection_routines.pyx":2003
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         if self.selector1.select_cell(pos, dds) and \             # <<<<<<<<<<<<<<
 *                 self.selector2.select_cell(pos, dds):
 *             return 1
 */
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":2005
 *         if self.selector1.select_cell(pos, dds) and \
 *                 self.selector2.select_cell(pos, dds):
 *             return 1             # <<<<<<<<<<<<<<
 *         else:
 *             return 0
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":2003
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         if self.selector1.select_cell(pos, dds) and \             # <<<<<<<<<<<<<<
 *                 self.selector2.select_cell(pos, dds):
 *             return 1
 */
  }

  /* "yt/geometry/selection_routines.pyx":2007
 *             return 1
 *         else:
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],
 */
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2002
 *                     self.selector2.select_grids(left_edges, right_edges, levels))
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         if self.selector1.select_cell(pos, dds) and \
 *                 self.selector2.select_cell(pos, dds):
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2009
 *             return 0
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_15ComposeSelector_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_15ComposeSelector_select_grid *__pyx_optional_args) {

  /* "yt/geometry/selection_routines.pyx":2011
 *     cdef int select_grid(self, np.float64_t left_edge[3],
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:             # <<<<<<<<<<<<<<
 *         if self.selector1.select_grid(left_edge, right_edge, level, o) or \
 *                 self.selector2.select_grid(left_edge, right_edge, level, o):
 */
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o = ((struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *)NULL);
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid __pyx_t_3;
  int __pyx_t_4;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_o = __pyx_optional_args->o;
    }
  }

  /* "yt/geometry/selection_routines.pyx":2012
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 *         if self.selector1.select_grid(left_edge, right_edge, level, o) or \             # <<<<<<<<<<<<<<
 *                 self.selector2.select_grid(left_edge, right_edge, level, o):
 *             return 1
 */
  __pyx_t_3.__pyx_n = 1;
  __pyx_t_3.o = __pyx_v_o;
  __pyx_t_2 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->selector1->__pyx_vtab)->select_grid(__pyx_v_self->selector1, __pyx_v_left_edge, __pyx_v_right_edge, __pyx_v_level, &__pyx_t_3); 
  __pyx_t_4 = (__pyx_t_2 != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":2013
 *                          Oct *o = NULL) nogil:
 *         if self.selector1.select_grid(left_edge, right_edge, level, o) or \
 *                 self.selector2.select_grid(left_edge, right_edge, level, o):             # <<<<<<<<<<<<<<
 *             return 1
 *         else:
 */
  __pyx_t_3.__pyx_n = 1;
  __pyx_t_3.o = __pyx_v_o;
  __pyx_t_2 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->selector2->__pyx_vtab)->select_grid(__pyx_v_self->selector2, __pyx_v_left_edge, __pyx_v_right_edge, __pyx_v_level, &__pyx_t_3); 
  __pyx_t_4 = (__pyx_t_2 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;

  /* "yt/geometry/selection_routines.pyx":2012
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 *         if self.selector1.select_grid(left_edge, right_edge, level, o) or \             # <<<<<<<<<<<<<<
 *                 self.selector2.select_grid(left_edge, right_edge, level, o):
 *             return 1
 */
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":2014
 *         if self.selector1.select_grid(left_edge, right_edge, level, o) or \
 *                 self.selector2.select_grid(left_edge, right_edge, level, o):
 *             return 1             # <<<<<<<<<<<<<<
 *         else:
 *             return 0
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":2012
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 *         if self.selector1.select_grid(left_edge, right_edge, level, o) or \             # <<<<<<<<<<<<<<
 *                 self.selector2.select_grid(left_edge, right_edge, level, o):
 *             return 1
 */
  }

  /* "yt/geometry/selection_routines.pyx":2016
 *             return 1
 *         else:
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 */
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2009
 *             return 0
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2018
 *             return 0
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         if self.selector1.select_point(pos) and \
 *                 self.selector2.select_point(pos):
 */

static int __pyx_f_2yt_8geometry_18selection_routines_15ComposeSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":2019
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         if self.selector1.select_point(pos) and \             # <<<<<<<<<<<<<<
 *                 self.selector2.select_point(pos):
 *             return 1
 */
  __pyx_t_2 = (((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->selector1->__pyx_vtab)->select_point(__pyx_v_self->selector1, __pyx_v_pos) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":2020
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         if self.selector1.select_point(pos) and \
 *                 self.selector2.select_point(pos):             # <<<<<<<<<<<<<<
 *             return 1
 *         else:
 */
  __pyx_t_2 = (((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->selector2->__pyx_vtab)->select_point(__pyx_v_self->selector2, __pyx_v_pos) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "yt/geometry/selection_routines.pyx":2019
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         if self.selector1.select_point(pos) and \             # <<<<<<<<<<<<<<
 *                 self.selector2.select_point(pos):
 *             return 1
 */
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":2021
 *         if self.selector1.select_point(pos) and \
 *                 self.selector2.select_point(pos):
 *             return 1             # <<<<<<<<<<<<<<
 *         else:
 *             return 0
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":2019
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         if self.selector1.select_point(pos) and \             # <<<<<<<<<<<<<<
 *                 self.selector2.select_point(pos):
 *             return 1
 */
  }

  /* "yt/geometry/selection_routines.pyx":2023
 *             return 1
 *         else:
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 */
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2018
 *             return 0
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         if self.selector1.select_point(pos) and \
 *                 self.selector2.select_point(pos):
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2025
 *             return 0
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         if self.selector1.select_sphere(pos, radius) and \
 *                 self.selector2.select_sphere(pos, radius):
 */

static int __pyx_f_2yt_8geometry_18selection_routines_15ComposeSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":2026
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         if self.selector1.select_sphere(pos, radius) and \             # <<<<<<<<<<<<<<
 *                 self.selector2.select_sphere(pos, radius):
 *             return 1
 */
  __pyx_t_2 = (((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->selector1->__pyx_vtab)->select_sphere(__pyx_v_self->selector1, __pyx_v_pos, __pyx_v_radius) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":2027
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         if self.selector1.select_sphere(pos, radius) and \
 *                 self.selector2.select_sphere(pos, radius):             # <<<<<<<<<<<<<<
 *             return 1
 *         else:
 */
  __pyx_t_2 = (((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->selector2->__pyx_vtab)->select_sphere(__pyx_v_self->selector2, __pyx_v_pos, __pyx_v_radius) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "yt/geometry/selection_routines.pyx":2026
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         if self.selector1.select_sphere(pos, radius) and \             # <<<<<<<<<<<<<<
 *                 self.selector2.select_sphere(pos, radius):
 *             return 1
 */
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":2028
 *         if self.selector1.select_sphere(pos, radius) and \
 *                 self.selector2.select_sphere(pos, radius):
 *             return 1             # <<<<<<<<<<<<<<
 *         else:
 *             return 0
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":2026
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         if self.selector1.select_sphere(pos, radius) and \             # <<<<<<<<<<<<<<
 *                 self.selector2.select_sphere(pos, radius):
 *             return 1
 */
  }

  /* "yt/geometry/selection_routines.pyx":2030
 *             return 1
 *         else:
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 */
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2025
 *             return 0
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         if self.selector1.select_sphere(pos, radius) and \
 *                 self.selector2.select_sphere(pos, radius):
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2032
 *             return 0
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         if self.selector1.select_bbox(left_edge, right_edge) and \
 */

static int __pyx_f_2yt_8geometry_18selection_routines_15ComposeSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":2034
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 *         if self.selector1.select_bbox(left_edge, right_edge) and \             # <<<<<<<<<<<<<<
 *                 self.selector2.select_bbox(left_edge, right_edge):
 *             return 1
 */
  __pyx_t_2 = (((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->selector1->__pyx_vtab)->select_bbox(__pyx_v_self->selector1, __pyx_v_left_edge, __pyx_v_right_edge) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "yt/geometry/selection_routines.pyx":2035
 *                                np.float64_t right_edge[3]) nogil:
 *         if self.selector1.select_bbox(left_edge, right_edge) and \
 *                 self.selector2.select_bbox(left_edge, right_edge):             # <<<<<<<<<<<<<<
 *             return 1
 *         else:
 */
  __pyx_t_2 = (((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->selector2->__pyx_vtab)->select_bbox(__pyx_v_self->selector2, __pyx_v_left_edge, __pyx_v_right_edge) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "yt/geometry/selection_routines.pyx":2034
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 *         if self.selector1.select_bbox(left_edge, right_edge) and \             # <<<<<<<<<<<<<<
 *                 self.selector2.select_bbox(left_edge, right_edge):
 *             return 1
 */
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pyx":2036
 *         if self.selector1.select_bbox(left_edge, right_edge) and \
 *                 self.selector2.select_bbox(left_edge, right_edge):
 *             return 1             # <<<<<<<<<<<<<<
 *         else:
 *             return 0
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "yt/geometry/selection_routines.pyx":2034
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 *         if self.selector1.select_bbox(left_edge, right_edge) and \             # <<<<<<<<<<<<<<
 *                 self.selector2.select_bbox(left_edge, right_edge):
 *             return 1
 */
  }

  /* "yt/geometry/selection_routines.pyx":2038
 *             return 1
 *         else:
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2032
 *             return 0
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         if self.selector1.select_bbox(left_edge, right_edge) and \
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2040
 *             return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (hash(self.selector1), hash(self.selector2))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_15ComposeSelector_5_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_15ComposeSelector_5_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_15ComposeSelector_4_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_15ComposeSelector_4_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":2041
 * 
 *     def _hash_vals(self):
 *         return (hash(self.selector1), hash(self.selector2))             # <<<<<<<<<<<<<<
 * 
 * compose_selector = ComposeSelector
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->selector1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(0, 2041, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((PyObject *)__pyx_v_self->selector2);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_2 = PyObject_Hash(__pyx_t_3); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(0, 2041, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_FromHash_t(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2040
 *             return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (hash(self.selector1), hash(self.selector2))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("yt.geometry.selection_routines.ComposeSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_15ComposeSelector_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_15ComposeSelector_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_15ComposeSelector_6__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_15ComposeSelector_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__40, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.ComposeSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_15ComposeSelector_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_15ComposeSelector_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_15ComposeSelector_8__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_15ComposeSelector_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__41, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.ComposeSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2050
 *     cdef object pind
 *     cdef public np.int64_t halo_id
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         self.base_source = dobj.base_source
 *         self.base_selector = self.base_source.selector
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2050, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2050, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.HaloParticlesSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":2051
 *     cdef public np.int64_t halo_id
 *     def __init__(self, dobj):
 *         self.base_source = dobj.base_source             # <<<<<<<<<<<<<<
 *         self.base_selector = self.base_source.selector
 *         self.pind = dobj.particle_indices
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_base_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->base_source);
  __Pyx_DECREF(__pyx_v_self->base_source);
  __pyx_v_self->base_source = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":2052
 *     def __init__(self, dobj):
 *         self.base_source = dobj.base_source
 *         self.base_selector = self.base_source.selector             # <<<<<<<<<<<<<<
 *         self.pind = dobj.particle_indices
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->base_source, __pyx_n_s_selector); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject))))) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->base_selector);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->base_selector));
  __pyx_v_self->base_selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":2053
 *         self.base_source = dobj.base_source
 *         self.base_selector = self.base_source.selector
 *         self.pind = dobj.particle_indices             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_particle_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->pind);
  __Pyx_DECREF(__pyx_v_self->pind);
  __pyx_v_self->pind = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":2050
 *     cdef object pind
 *     cdef public np.int64_t halo_id
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         self.base_source = dobj.base_source
 *         self.base_selector = self.base_source.selector
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.HaloParticlesSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2055
 *         self.pind = dobj.particle_indices
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return ("halo_particles", self.halo_id)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_3_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_3_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_2_hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_2_hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":2056
 * 
 *     def _hash_vals(self):
 *         return ("halo_particles", self.halo_id)             # <<<<<<<<<<<<<<
 * 
 * halo_particles_selector = HaloParticlesSelector
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->halo_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_halo_particles);
  __Pyx_GIVEREF(__pyx_n_s_halo_particles);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_halo_particles);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2055
 *         self.pind = dobj.particle_indices
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return ("halo_particles", self.halo_id)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("yt.geometry.selection_routines.HaloParticlesSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2046
 * 
 * cdef class HaloParticlesSelector(SelectorObject):
 *     cdef public object base_source             # <<<<<<<<<<<<<<
 *     cdef SelectorObject base_selector
 *     cdef object pind
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source___get__(((struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source___get__(struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->base_source);
  __pyx_r = __pyx_v_self->base_source;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source_2__set__(((struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source_2__set__(struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->base_source);
  __Pyx_DECREF(__pyx_v_self->base_source);
  __pyx_v_self->base_source = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source_4__del__(((struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source_4__del__(struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->base_source);
  __Pyx_DECREF(__pyx_v_self->base_source);
  __pyx_v_self->base_source = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2049
 *     cdef SelectorObject base_selector
 *     cdef object pind
 *     cdef public np.int64_t halo_id             # <<<<<<<<<<<<<<
 *     def __init__(self, dobj):
 *         self.base_source = dobj.base_source
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_7halo_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_7halo_id_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_7halo_id___get__(((struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_7halo_id___get__(struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->halo_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.HaloParticlesSelector.halo_id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_7halo_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_7halo_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_7halo_id_2__set__(((struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_7halo_id_2__set__(struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_int64_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int64(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2049, __pyx_L1_error)
  __pyx_v_self->halo_id = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.HaloParticlesSelector.halo_id.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_4__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__42, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.HaloParticlesSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_6__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_21HaloParticlesSelector_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__43, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.HaloParticlesSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2063
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * def points_in_cells(             # <<<<<<<<<<<<<<
 *         np.float64_t[:] cx,
 *         np.float64_t[:] cy,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_5points_in_cells(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_8geometry_18selection_routines_5points_in_cells = {"points_in_cells", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_18selection_routines_5points_in_cells, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_5points_in_cells(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_cx = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_cy = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_cz = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_dx = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_dy = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_dz = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_px = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_py = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_pz = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("points_in_cells (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cx,&__pyx_n_s_cy,&__pyx_n_s_cz,&__pyx_n_s_dx,&__pyx_n_s_dy,&__pyx_n_s_dz,&__pyx_n_s_px,&__pyx_n_s_py,&__pyx_n_s_pz,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cx)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("points_in_cells", 1, 9, 9, 1); __PYX_ERR(0, 2063, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("points_in_cells", 1, 9, 9, 2); __PYX_ERR(0, 2063, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("points_in_cells", 1, 9, 9, 3); __PYX_ERR(0, 2063, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("points_in_cells", 1, 9, 9, 4); __PYX_ERR(0, 2063, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("points_in_cells", 1, 9, 9, 5); __PYX_ERR(0, 2063, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_px)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("points_in_cells", 1, 9, 9, 6); __PYX_ERR(0, 2063, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_py)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("points_in_cells", 1, 9, 9, 7); __PYX_ERR(0, 2063, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("points_in_cells", 1, 9, 9, 8); __PYX_ERR(0, 2063, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "points_in_cells") < 0)) __PYX_ERR(0, 2063, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_cx = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_cx.memview)) __PYX_ERR(0, 2064, __pyx_L3_error)
    __pyx_v_cy = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_cy.memview)) __PYX_ERR(0, 2065, __pyx_L3_error)
    __pyx_v_cz = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_cz.memview)) __PYX_ERR(0, 2066, __pyx_L3_error)
    __pyx_v_dx = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_dx.memview)) __PYX_ERR(0, 2067, __pyx_L3_error)
    __pyx_v_dy = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_dy.memview)) __PYX_ERR(0, 2068, __pyx_L3_error)
    __pyx_v_dz = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_dz.memview)) __PYX_ERR(0, 2069, __pyx_L3_error)
    __pyx_v_px = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_px.memview)) __PYX_ERR(0, 2070, __pyx_L3_error)
    __pyx_v_py = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_py.memview)) __PYX_ERR(0, 2071, __pyx_L3_error)
    __pyx_v_pz = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_pz.memview)) __PYX_ERR(0, 2072, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("points_in_cells", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2063, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.points_in_cells", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_4points_in_cells(__pyx_self, __pyx_v_cx, __pyx_v_cy, __pyx_v_cz, __pyx_v_dx, __pyx_v_dy, __pyx_v_dz, __pyx_v_px, __pyx_v_py, __pyx_v_pz);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_4points_in_cells(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_cx, __Pyx_memviewslice __pyx_v_cy, __Pyx_memviewslice __pyx_v_cz, __Pyx_memviewslice __pyx_v_dx, __Pyx_memviewslice __pyx_v_dy, __Pyx_memviewslice __pyx_v_dz, __Pyx_memviewslice __pyx_v_px, __Pyx_memviewslice __pyx_v_py, __Pyx_memviewslice __pyx_v_pz) {
  int __pyx_v_p;
  int __pyx_v_c;
  int __pyx_v_n_p;
  int __pyx_v_n_c;
  PyArrayObject *__pyx_v_mask = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyArrayObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("points_in_cells", 0);
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;

  /* "yt/geometry/selection_routines.pyx":2081
 *     cdef np.ndarray[np.uint8_t, ndim=1, cast=True] mask
 * 
 *     n_p = px.size             # <<<<<<<<<<<<<<
 *     n_c = cx.size
 *     mask = np.ones(n_p, dtype="bool")
 */
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_px, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2081, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_n_p = __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":2082
 * 
 *     n_p = px.size
 *     n_c = cx.size             # <<<<<<<<<<<<<<
 *     mask = np.ones(n_p, dtype="bool")
 * 
 */
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_cx, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2082, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2082, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2082, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_n_c = __pyx_t_3;

  /* "yt/geometry/selection_routines.pyx":2083
 *     n_p = px.size
 *     n_c = cx.size
 *     mask = np.ones(n_p, dtype="bool")             # <<<<<<<<<<<<<<
 * 
 *     for p in range(n_p):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_n_p); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_bool) < 0) __PYX_ERR(0, 2083, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 2083, __pyx_L1_error)
  __pyx_t_6 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 1, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 1, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
      __pyx_t_7 = __pyx_t_8 = __pyx_t_9 = 0;
    }
    __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 2083, __pyx_L1_error)
  }
  __pyx_t_6 = 0;
  __pyx_v_mask = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "yt/geometry/selection_routines.pyx":2085
 *     mask = np.ones(n_p, dtype="bool")
 * 
 *     for p in range(n_p):             # <<<<<<<<<<<<<<
 *         for c in range(n_c):
 *             if fabs(px[p] - cx[c]) > 0.5 * dx[c]:
 */
  __pyx_t_3 = __pyx_v_n_p;
  __pyx_t_10 = __pyx_t_3;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_p = __pyx_t_11;

    /* "yt/geometry/selection_routines.pyx":2086
 * 
 *     for p in range(n_p):
 *         for c in range(n_c):             # <<<<<<<<<<<<<<
 *             if fabs(px[p] - cx[c]) > 0.5 * dx[c]:
 *                 mask[p] = False
 */
    __pyx_t_12 = __pyx_v_n_c;
    __pyx_t_13 = __pyx_t_12;
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
      __pyx_v_c = __pyx_t_14;

      /* "yt/geometry/selection_routines.pyx":2087
 *     for p in range(n_p):
 *         for c in range(n_c):
 *             if fabs(px[p] - cx[c]) > 0.5 * dx[c]:             # <<<<<<<<<<<<<<
 *                 mask[p] = False
 *                 continue
 */
      __pyx_t_15 = __pyx_v_p;
      __pyx_t_16 = __pyx_v_c;
      __pyx_t_17 = __pyx_v_c;
      __pyx_t_18 = ((fabs(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_px.data + __pyx_t_15 * __pyx_v_px.strides[0]) ))) - (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_cx.data + __pyx_t_16 * __pyx_v_cx.strides[0]) ))))) > (0.5 * (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_dx.data + __pyx_t_17 * __pyx_v_dx.strides[0]) ))))) != 0);
      if (__pyx_t_18) {

        /* "yt/geometry/selection_routines.pyx":2088
 *         for c in range(n_c):
 *             if fabs(px[p] - cx[c]) > 0.5 * dx[c]:
 *                 mask[p] = False             # <<<<<<<<<<<<<<
 *                 continue
 *             if fabs(py[p] - cy[c]) > 0.5 * dy[c]:
 */
        __pyx_t_17 = __pyx_v_p;
        *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_mask.diminfo[0].strides) = 0;

        /* "yt/geometry/selection_routines.pyx":2089
 *             if fabs(px[p] - cx[c]) > 0.5 * dx[c]:
 *                 mask[p] = False
 *                 continue             # <<<<<<<<<<<<<<
 *             if fabs(py[p] - cy[c]) > 0.5 * dy[c]:
 *                 mask[p] = False
 */
        goto __pyx_L5_continue;

        /* "yt/geometry/selection_routines.pyx":2087
 *     for p in range(n_p):
 *         for c in range(n_c):
 *             if fabs(px[p] - cx[c]) > 0.5 * dx[c]:             # <<<<<<<<<<<<<<
 *                 mask[p] = False
 *                 continue
 */
      }

      /* "yt/geometry/selection_routines.pyx":2090
 *                 mask[p] = False
 *                 continue
 *             if fabs(py[p] - cy[c]) > 0.5 * dy[c]:             # <<<<<<<<<<<<<<
 *                 mask[p] = False
 *                 continue
 */
      __pyx_t_16 = __pyx_v_p;
      __pyx_t_15 = __pyx_v_c;
      __pyx_t_19 = __pyx_v_c;
      __pyx_t_18 = ((fabs(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_py.data + __pyx_t_16 * __pyx_v_py.strides[0]) ))) - (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_cy.data + __pyx_t_15 * __pyx_v_cy.strides[0]) ))))) > (0.5 * (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_dy.data + __pyx_t_19 * __pyx_v_dy.strides[0]) ))))) != 0);
      if (__pyx_t_18) {

        /* "yt/geometry/selection_routines.pyx":2091
 *                 continue
 *             if fabs(py[p] - cy[c]) > 0.5 * dy[c]:
 *                 mask[p] = False             # <<<<<<<<<<<<<<
 *                 continue
 *             if fabs(pz[p] - cz[c]) > 0.5 * dz[c]:
 */
        __pyx_t_19 = __pyx_v_p;
        *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_19, __pyx_pybuffernd_mask.diminfo[0].strides) = 0;

        /* "yt/geometry/selection_routines.pyx":2092
 *             if fabs(py[p] - cy[c]) > 0.5 * dy[c]:
 *                 mask[p] = False
 *                 continue             # <<<<<<<<<<<<<<
 *             if fabs(pz[p] - cz[c]) > 0.5 * dz[c]:
 *                 mask[p] = False
 */
        goto __pyx_L5_continue;

        /* "yt/geometry/selection_routines.pyx":2090
 *                 mask[p] = False
 *                 continue
 *             if fabs(py[p] - cy[c]) > 0.5 * dy[c]:             # <<<<<<<<<<<<<<
 *                 mask[p] = False
 *                 continue
 */
      }

      /* "yt/geometry/selection_routines.pyx":2093
 *                 mask[p] = False
 *                 continue
 *             if fabs(pz[p] - cz[c]) > 0.5 * dz[c]:             # <<<<<<<<<<<<<<
 *                 mask[p] = False
 *                 continue
 */
      __pyx_t_15 = __pyx_v_p;
      __pyx_t_16 = __pyx_v_c;
      __pyx_t_20 = __pyx_v_c;
      __pyx_t_18 = ((fabs(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_pz.data + __pyx_t_15 * __pyx_v_pz.strides[0]) ))) - (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_cz.data + __pyx_t_16 * __pyx_v_cz.strides[0]) ))))) > (0.5 * (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_dz.data + __pyx_t_20 * __pyx_v_dz.strides[0]) ))))) != 0);
      if (__pyx_t_18) {

        /* "yt/geometry/selection_routines.pyx":2094
 *                 continue
 *             if fabs(pz[p] - cz[c]) > 0.5 * dz[c]:
 *                 mask[p] = False             # <<<<<<<<<<<<<<
 *                 continue
 *             if mask[p]: break
 */
        __pyx_t_20 = __pyx_v_p;
        *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_20, __pyx_pybuffernd_mask.diminfo[0].strides) = 0;

        /* "yt/geometry/selection_routines.pyx":2095
 *             if fabs(pz[p] - cz[c]) > 0.5 * dz[c]:
 *                 mask[p] = False
 *                 continue             # <<<<<<<<<<<<<<
 *             if mask[p]: break
 * 
 */
        goto __pyx_L5_continue;

        /* "yt/geometry/selection_routines.pyx":2093
 *                 mask[p] = False
 *                 continue
 *             if fabs(pz[p] - cz[c]) > 0.5 * dz[c]:             # <<<<<<<<<<<<<<
 *                 mask[p] = False
 *                 continue
 */
      }

      /* "yt/geometry/selection_routines.pyx":2096
 *                 mask[p] = False
 *                 continue
 *             if mask[p]: break             # <<<<<<<<<<<<<<
 * 
 *     return mask
 */
      __pyx_t_16 = __pyx_v_p;
      __pyx_t_18 = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_mask.diminfo[0].strides)) != 0);
      if (__pyx_t_18) {
        goto __pyx_L6_break;
      }
      __pyx_L5_continue:;
    }
    __pyx_L6_break:;
  }

  /* "yt/geometry/selection_routines.pyx":2098
 *             if mask[p]: break
 * 
 *     return mask             # <<<<<<<<<<<<<<
 * 
 * cdef class BooleanSelector(SelectorObject):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_mask));
  __pyx_r = ((PyObject *)__pyx_v_mask);
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2063
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * def points_in_cells(             # <<<<<<<<<<<<<<
 *         np.float64_t[:] cx,
 *         np.float64_t[:] cy,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.selection_routines.points_in_cells", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mask);
  __PYX_XDEC_MEMVIEW(&__pyx_v_cx, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_cy, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_cz, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_dx, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_dy, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_dz, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_px, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_py, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_pz, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2102
 * cdef class BooleanSelector(SelectorObject):
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         # Note that this has a different API than the other selector objects,
 *         # so will not work as a traditional data selector.
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2102, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2102, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":2105
 *         # Note that this has a different API than the other selector objects,
 *         # so will not work as a traditional data selector.
 *         if not hasattr(dobj.dobj1, "selector"):             # <<<<<<<<<<<<<<
 *             self.sel1 = dobj.dobj1
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_dobj1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_HasAttr(__pyx_t_1, __pyx_n_s_selector); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_3) {

    /* "yt/geometry/selection_routines.pyx":2106
 *         # so will not work as a traditional data selector.
 *         if not hasattr(dobj.dobj1, "selector"):
 *             self.sel1 = dobj.dobj1             # <<<<<<<<<<<<<<
 *         else:
 *             self.sel1 = dobj.dobj1.selector
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_dobj1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject))))) __PYX_ERR(0, 2106, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->sel1);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->sel1));
    __pyx_v_self->sel1 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "yt/geometry/selection_routines.pyx":2105
 *         # Note that this has a different API than the other selector objects,
 *         # so will not work as a traditional data selector.
 *         if not hasattr(dobj.dobj1, "selector"):             # <<<<<<<<<<<<<<
 *             self.sel1 = dobj.dobj1
 *         else:
 */
    goto __pyx_L3;
  }

  /* "yt/geometry/selection_routines.pyx":2108
 *             self.sel1 = dobj.dobj1
 *         else:
 *             self.sel1 = dobj.dobj1.selector             # <<<<<<<<<<<<<<
 *         if not hasattr(dobj.dobj2, "selector"):
 *             self.sel2 = dobj.dobj2
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_dobj1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_selector); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject))))) __PYX_ERR(0, 2108, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_4);
    __Pyx_GOTREF(__pyx_v_self->sel1);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->sel1));
    __pyx_v_self->sel1 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "yt/geometry/selection_routines.pyx":2109
 *         else:
 *             self.sel1 = dobj.dobj1.selector
 *         if not hasattr(dobj.dobj2, "selector"):             # <<<<<<<<<<<<<<
 *             self.sel2 = dobj.dobj2
 *         else:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_dobj2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_HasAttr(__pyx_t_4, __pyx_n_s_selector); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 2109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = ((!(__pyx_t_3 != 0)) != 0);
  if (__pyx_t_2) {

    /* "yt/geometry/selection_routines.pyx":2110
 *             self.sel1 = dobj.dobj1.selector
 *         if not hasattr(dobj.dobj2, "selector"):
 *             self.sel2 = dobj.dobj2             # <<<<<<<<<<<<<<
 *         else:
 *             self.sel2 = dobj.dobj2.selector
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_dobj2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject))))) __PYX_ERR(0, 2110, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_4);
    __Pyx_GOTREF(__pyx_v_self->sel2);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->sel2));
    __pyx_v_self->sel2 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "yt/geometry/selection_routines.pyx":2109
 *         else:
 *             self.sel1 = dobj.dobj1.selector
 *         if not hasattr(dobj.dobj2, "selector"):             # <<<<<<<<<<<<<<
 *             self.sel2 = dobj.dobj2
 *         else:
 */
    goto __pyx_L4;
  }

  /* "yt/geometry/selection_routines.pyx":2112
 *             self.sel2 = dobj.dobj2
 *         else:
 *             self.sel2 = dobj.dobj2.selector             # <<<<<<<<<<<<<<
 * 
 * cdef class BooleanANDSelector(BooleanSelector):
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_dobj2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_selector); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject))))) __PYX_ERR(0, 2112, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->sel2);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->sel2));
    __pyx_v_self->sel2 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L4:;

  /* "yt/geometry/selection_routines.pyx":2102
 * cdef class BooleanSelector(SelectorObject):
 * 
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         # Note that this has a different API than the other selector objects,
 *         # so will not work as a traditional data selector.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pxd":75
 * 
 * cdef class BooleanSelector(SelectorObject):
 *     cdef public SelectorObject sel1             # <<<<<<<<<<<<<<
 *     cdef public SelectorObject sel2
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1___get__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1___get__(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->sel1));
  __pyx_r = ((PyObject *)__pyx_v_self->sel1);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1_2__set__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1_2__set__(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject))))) __PYX_ERR(1, 75, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->sel1);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->sel1));
  __pyx_v_self->sel1 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanSelector.sel1.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1_4__del__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1_4__del__(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->sel1);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->sel1));
  __pyx_v_self->sel1 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pxd":76
 * cdef class BooleanSelector(SelectorObject):
 *     cdef public SelectorObject sel1
 *     cdef public SelectorObject sel2             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.float64_t _periodic_dist(np.float64_t x1, np.float64_t x2,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2___get__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2___get__(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->sel2));
  __pyx_r = ((PyObject *)__pyx_v_self->sel2);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2_2__set__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2_2__set__(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject))))) __PYX_ERR(1, 76, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->sel2);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->sel2));
  __pyx_v_self->sel2 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanSelector.sel2.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2_4__del__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2_4__del__(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->sel2);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->sel2));
  __pyx_v_self->sel2 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_2__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__44, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_15BooleanSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__45, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2115
 * 
 * cdef class BooleanANDSelector(BooleanSelector):
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef int rv1 = self.sel1.select_bbox(left_edge, right_edge)
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_v_rv1;
  int __pyx_v_rv2;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2117
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef int rv1 = self.sel1.select_bbox(left_edge, right_edge)             # <<<<<<<<<<<<<<
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_bbox(left_edge, right_edge)
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_bbox(__pyx_v_self->__pyx_base.sel1, __pyx_v_left_edge, __pyx_v_right_edge);

  /* "yt/geometry/selection_routines.pyx":2118
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef int rv1 = self.sel1.select_bbox(left_edge, right_edge)
 *         if rv1 == 0: return 0             # <<<<<<<<<<<<<<
 *         cdef int rv2 = self.sel2.select_bbox(left_edge, right_edge)
 *         if rv2 == 0: return 0
 */
  __pyx_t_1 = ((__pyx_v_rv1 == 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2119
 *         cdef int rv1 = self.sel1.select_bbox(left_edge, right_edge)
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_bbox(left_edge, right_edge)             # <<<<<<<<<<<<<<
 *         if rv2 == 0: return 0
 *         return 1
 */
  __pyx_v_rv2 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel2->__pyx_vtab)->select_bbox(__pyx_v_self->__pyx_base.sel2, __pyx_v_left_edge, __pyx_v_right_edge);

  /* "yt/geometry/selection_routines.pyx":2120
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_bbox(left_edge, right_edge)
 *         if rv2 == 0: return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv2 == 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2121
 *         cdef int rv2 = self.sel2.select_bbox(left_edge, right_edge)
 *         if rv2 == 0: return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2115
 * 
 * cdef class BooleanANDSelector(BooleanSelector):
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef int rv1 = self.sel1.select_bbox(left_edge, right_edge)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2123
 *         return 1
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_grid *__pyx_optional_args) {

  /* "yt/geometry/selection_routines.pyx":2125
 *     cdef int select_grid(self, np.float64_t left_edge[3],
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_grid(left_edge, right_edge, level, o)
 *         if rv1 == 0: return 0
 */
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o = ((struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *)NULL);
  int __pyx_v_rv1;
  int __pyx_v_rv2;
  int __pyx_r;
  int __pyx_t_1;
  struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid __pyx_t_2;
  int __pyx_t_3;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_o = __pyx_optional_args->o;
    }
  }

  /* "yt/geometry/selection_routines.pyx":2126
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 *         cdef int rv1 = self.sel1.select_grid(left_edge, right_edge, level, o)             # <<<<<<<<<<<<<<
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_grid(left_edge, right_edge, level, o)
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.o = __pyx_v_o;
  __pyx_t_1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_grid(__pyx_v_self->__pyx_base.sel1, __pyx_v_left_edge, __pyx_v_right_edge, __pyx_v_level, &__pyx_t_2); 
  __pyx_v_rv1 = __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2127
 *                          Oct *o = NULL) nogil:
 *         cdef int rv1 = self.sel1.select_grid(left_edge, right_edge, level, o)
 *         if rv1 == 0: return 0             # <<<<<<<<<<<<<<
 *         cdef int rv2 = self.sel2.select_grid(left_edge, right_edge, level, o)
 *         if rv2 == 0: return 0
 */
  __pyx_t_3 = ((__pyx_v_rv1 == 0) != 0);
  if (__pyx_t_3) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2128
 *         cdef int rv1 = self.sel1.select_grid(left_edge, right_edge, level, o)
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_grid(left_edge, right_edge, level, o)             # <<<<<<<<<<<<<<
 *         if rv2 == 0: return 0
 *         return 1
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.o = __pyx_v_o;
  __pyx_t_1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel2->__pyx_vtab)->select_grid(__pyx_v_self->__pyx_base.sel2, __pyx_v_left_edge, __pyx_v_right_edge, __pyx_v_level, &__pyx_t_2); 
  __pyx_v_rv2 = __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2129
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_grid(left_edge, right_edge, level, o)
 *         if rv2 == 0: return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  __pyx_t_3 = ((__pyx_v_rv2 == 0) != 0);
  if (__pyx_t_3) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2130
 *         cdef int rv2 = self.sel2.select_grid(left_edge, right_edge, level, o)
 *         if rv2 == 0: return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2123
 *         return 1
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2132
 *         return 1
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         if rv1 == 0: return 0
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_v_rv1;
  int __pyx_v_rv2;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2133
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)             # <<<<<<<<<<<<<<
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_cell(__pyx_v_self->__pyx_base.sel1, __pyx_v_pos, __pyx_v_dds);

  /* "yt/geometry/selection_routines.pyx":2134
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         if rv1 == 0: return 0             # <<<<<<<<<<<<<<
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 *         if rv2 == 0: return 0
 */
  __pyx_t_1 = ((__pyx_v_rv1 == 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2135
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)             # <<<<<<<<<<<<<<
 *         if rv2 == 0: return 0
 *         return 1
 */
  __pyx_v_rv2 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel2->__pyx_vtab)->select_cell(__pyx_v_self->__pyx_base.sel2, __pyx_v_pos, __pyx_v_dds);

  /* "yt/geometry/selection_routines.pyx":2136
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 *         if rv2 == 0: return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv2 == 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2137
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 *         if rv2 == 0: return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2132
 *         return 1
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         if rv1 == 0: return 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2139
 *         return 1
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         if rv1 == 0: return 0
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_v_rv1;
  int __pyx_v_rv2;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2140
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         cdef int rv1 = self.sel1.select_point(pos)             # <<<<<<<<<<<<<<
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_point(pos)
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_point(__pyx_v_self->__pyx_base.sel1, __pyx_v_pos);

  /* "yt/geometry/selection_routines.pyx":2141
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         if rv1 == 0: return 0             # <<<<<<<<<<<<<<
 *         cdef int rv2 = self.sel2.select_point(pos)
 *         if rv2 == 0: return 0
 */
  __pyx_t_1 = ((__pyx_v_rv1 == 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2142
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_point(pos)             # <<<<<<<<<<<<<<
 *         if rv2 == 0: return 0
 *         return 1
 */
  __pyx_v_rv2 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel2->__pyx_vtab)->select_point(__pyx_v_self->__pyx_base.sel2, __pyx_v_pos);

  /* "yt/geometry/selection_routines.pyx":2143
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_point(pos)
 *         if rv2 == 0: return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv2 == 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2144
 *         cdef int rv2 = self.sel2.select_point(pos)
 *         if rv2 == 0: return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2139
 *         return 1
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         if rv1 == 0: return 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2146
 *         return 1
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         if rv1 == 0: return 0
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_v_rv1;
  int __pyx_v_rv2;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2147
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)             # <<<<<<<<<<<<<<
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_sphere(__pyx_v_self->__pyx_base.sel1, __pyx_v_pos, __pyx_v_radius);

  /* "yt/geometry/selection_routines.pyx":2148
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         if rv1 == 0: return 0             # <<<<<<<<<<<<<<
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 *         if rv2 == 0: return 0
 */
  __pyx_t_1 = ((__pyx_v_rv1 == 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2149
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)             # <<<<<<<<<<<<<<
 *         if rv2 == 0: return 0
 *         return 1
 */
  __pyx_v_rv2 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel2->__pyx_vtab)->select_sphere(__pyx_v_self->__pyx_base.sel2, __pyx_v_pos, __pyx_v_radius);

  /* "yt/geometry/selection_routines.pyx":2150
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 *         if rv2 == 0: return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv2 == 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2151
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 *         if rv2 == 0: return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2146
 *         return 1
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         if rv1 == 0: return 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2153
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (self.sel1._hash_vals() +
 *                 ("and",) +
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanANDSelector_1_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanANDSelector_1_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_18BooleanANDSelector__hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanANDSelector__hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":2154
 * 
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +             # <<<<<<<<<<<<<<
 *                 ("and",) +
 *                 self.sel2._hash_vals())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.sel1), __pyx_n_s_hash_vals); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":2155
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +
 *                 ("and",) +             # <<<<<<<<<<<<<<
 *                 self.sel2._hash_vals())
 * 
 */
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_tuple__46); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":2156
 *         return (self.sel1._hash_vals() +
 *                 ("and",) +
 *                 self.sel2._hash_vals())             # <<<<<<<<<<<<<<
 * 
 * cdef class BooleanORSelector(BooleanSelector):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.sel2), __pyx_n_s_hash_vals); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":2155
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +
 *                 ("and",) +             # <<<<<<<<<<<<<<
 *                 self.sel2._hash_vals())
 * 
 */
  __pyx_t_3 = PyNumber_Add(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2153
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (self.sel1._hash_vals() +
 *                 ("and",) +
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanANDSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanANDSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanANDSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_18BooleanANDSelector_2__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanANDSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanANDSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanANDSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanANDSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_18BooleanANDSelector_4__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanANDSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanANDSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2159
 * 
 * cdef class BooleanORSelector(BooleanSelector):
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef int rv1 = self.sel1.select_bbox(left_edge, right_edge)
 */

static int __pyx_f_2yt_8geometry_18selection_routines_17BooleanORSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_v_rv1;
  int __pyx_v_rv2;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2161
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef int rv1 = self.sel1.select_bbox(left_edge, right_edge)             # <<<<<<<<<<<<<<
 *         if rv1 == 1: return 1
 *         cdef int rv2 = self.sel2.select_bbox(left_edge, right_edge)
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_bbox(__pyx_v_self->__pyx_base.sel1, __pyx_v_left_edge, __pyx_v_right_edge);

  /* "yt/geometry/selection_routines.pyx":2162
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef int rv1 = self.sel1.select_bbox(left_edge, right_edge)
 *         if rv1 == 1: return 1             # <<<<<<<<<<<<<<
 *         cdef int rv2 = self.sel2.select_bbox(left_edge, right_edge)
 *         if rv2 == 1: return 1
 */
  __pyx_t_1 = ((__pyx_v_rv1 == 1) != 0);
  if (__pyx_t_1) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2163
 *         cdef int rv1 = self.sel1.select_bbox(left_edge, right_edge)
 *         if rv1 == 1: return 1
 *         cdef int rv2 = self.sel2.select_bbox(left_edge, right_edge)             # <<<<<<<<<<<<<<
 *         if rv2 == 1: return 1
 *         return 0
 */
  __pyx_v_rv2 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel2->__pyx_vtab)->select_bbox(__pyx_v_self->__pyx_base.sel2, __pyx_v_left_edge, __pyx_v_right_edge);

  /* "yt/geometry/selection_routines.pyx":2164
 *         if rv1 == 1: return 1
 *         cdef int rv2 = self.sel2.select_bbox(left_edge, right_edge)
 *         if rv2 == 1: return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv2 == 1) != 0);
  if (__pyx_t_1) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2165
 *         cdef int rv2 = self.sel2.select_bbox(left_edge, right_edge)
 *         if rv2 == 1: return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2159
 * 
 * cdef class BooleanORSelector(BooleanSelector):
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         cdef int rv1 = self.sel1.select_bbox(left_edge, right_edge)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2167
 *         return 0
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_17BooleanORSelector_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_17BooleanORSelector_select_grid *__pyx_optional_args) {

  /* "yt/geometry/selection_routines.pyx":2169
 *     cdef int select_grid(self, np.float64_t left_edge[3],
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_grid(left_edge, right_edge, level, o)
 *         if rv1 == 1: return 1
 */
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o = ((struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *)NULL);
  int __pyx_v_rv1;
  int __pyx_v_rv2;
  int __pyx_r;
  int __pyx_t_1;
  struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid __pyx_t_2;
  int __pyx_t_3;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_o = __pyx_optional_args->o;
    }
  }

  /* "yt/geometry/selection_routines.pyx":2170
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 *         cdef int rv1 = self.sel1.select_grid(left_edge, right_edge, level, o)             # <<<<<<<<<<<<<<
 *         if rv1 == 1: return 1
 *         cdef int rv2 = self.sel2.select_grid(left_edge, right_edge, level, o)
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.o = __pyx_v_o;
  __pyx_t_1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_grid(__pyx_v_self->__pyx_base.sel1, __pyx_v_left_edge, __pyx_v_right_edge, __pyx_v_level, &__pyx_t_2); 
  __pyx_v_rv1 = __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2171
 *                          Oct *o = NULL) nogil:
 *         cdef int rv1 = self.sel1.select_grid(left_edge, right_edge, level, o)
 *         if rv1 == 1: return 1             # <<<<<<<<<<<<<<
 *         cdef int rv2 = self.sel2.select_grid(left_edge, right_edge, level, o)
 *         if rv2 == 1: return 1
 */
  __pyx_t_3 = ((__pyx_v_rv1 == 1) != 0);
  if (__pyx_t_3) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2172
 *         cdef int rv1 = self.sel1.select_grid(left_edge, right_edge, level, o)
 *         if rv1 == 1: return 1
 *         cdef int rv2 = self.sel2.select_grid(left_edge, right_edge, level, o)             # <<<<<<<<<<<<<<
 *         if rv2 == 1: return 1
 *         return 0
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.o = __pyx_v_o;
  __pyx_t_1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel2->__pyx_vtab)->select_grid(__pyx_v_self->__pyx_base.sel2, __pyx_v_left_edge, __pyx_v_right_edge, __pyx_v_level, &__pyx_t_2); 
  __pyx_v_rv2 = __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2173
 *         if rv1 == 1: return 1
 *         cdef int rv2 = self.sel2.select_grid(left_edge, right_edge, level, o)
 *         if rv2 == 1: return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_3 = ((__pyx_v_rv2 == 1) != 0);
  if (__pyx_t_3) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2174
 *         cdef int rv2 = self.sel2.select_grid(left_edge, right_edge, level, o)
 *         if rv2 == 1: return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2167
 *         return 0
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2176
 *         return 0
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         if rv1 == 1: return 1
 */

static int __pyx_f_2yt_8geometry_18selection_routines_17BooleanORSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_v_rv1;
  int __pyx_v_rv2;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2177
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)             # <<<<<<<<<<<<<<
 *         if rv1 == 1: return 1
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_cell(__pyx_v_self->__pyx_base.sel1, __pyx_v_pos, __pyx_v_dds);

  /* "yt/geometry/selection_routines.pyx":2178
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         if rv1 == 1: return 1             # <<<<<<<<<<<<<<
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 *         if rv2 == 1: return 1
 */
  __pyx_t_1 = ((__pyx_v_rv1 == 1) != 0);
  if (__pyx_t_1) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2179
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         if rv1 == 1: return 1
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)             # <<<<<<<<<<<<<<
 *         if rv2 == 1: return 1
 *         return 0
 */
  __pyx_v_rv2 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel2->__pyx_vtab)->select_cell(__pyx_v_self->__pyx_base.sel2, __pyx_v_pos, __pyx_v_dds);

  /* "yt/geometry/selection_routines.pyx":2180
 *         if rv1 == 1: return 1
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 *         if rv2 == 1: return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv2 == 1) != 0);
  if (__pyx_t_1) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2181
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 *         if rv2 == 1: return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2176
 *         return 0
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         if rv1 == 1: return 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2183
 *         return 0
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         if rv1 == 1: return 1
 */

static int __pyx_f_2yt_8geometry_18selection_routines_17BooleanORSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_v_rv1;
  int __pyx_v_rv2;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2184
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         cdef int rv1 = self.sel1.select_point(pos)             # <<<<<<<<<<<<<<
 *         if rv1 == 1: return 1
 *         cdef int rv2 = self.sel2.select_point(pos)
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_point(__pyx_v_self->__pyx_base.sel1, __pyx_v_pos);

  /* "yt/geometry/selection_routines.pyx":2185
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         if rv1 == 1: return 1             # <<<<<<<<<<<<<<
 *         cdef int rv2 = self.sel2.select_point(pos)
 *         if rv2 == 1: return 1
 */
  __pyx_t_1 = ((__pyx_v_rv1 == 1) != 0);
  if (__pyx_t_1) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2186
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         if rv1 == 1: return 1
 *         cdef int rv2 = self.sel2.select_point(pos)             # <<<<<<<<<<<<<<
 *         if rv2 == 1: return 1
 *         return 0
 */
  __pyx_v_rv2 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel2->__pyx_vtab)->select_point(__pyx_v_self->__pyx_base.sel2, __pyx_v_pos);

  /* "yt/geometry/selection_routines.pyx":2187
 *         if rv1 == 1: return 1
 *         cdef int rv2 = self.sel2.select_point(pos)
 *         if rv2 == 1: return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv2 == 1) != 0);
  if (__pyx_t_1) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2188
 *         cdef int rv2 = self.sel2.select_point(pos)
 *         if rv2 == 1: return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2183
 *         return 0
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         if rv1 == 1: return 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2190
 *         return 0
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         if rv1 == 1: return 1
 */

static int __pyx_f_2yt_8geometry_18selection_routines_17BooleanORSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_v_rv1;
  int __pyx_v_rv2;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2191
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)             # <<<<<<<<<<<<<<
 *         if rv1 == 1: return 1
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_sphere(__pyx_v_self->__pyx_base.sel1, __pyx_v_pos, __pyx_v_radius);

  /* "yt/geometry/selection_routines.pyx":2192
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         if rv1 == 1: return 1             # <<<<<<<<<<<<<<
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 *         if rv2 == 1: return 1
 */
  __pyx_t_1 = ((__pyx_v_rv1 == 1) != 0);
  if (__pyx_t_1) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2193
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         if rv1 == 1: return 1
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)             # <<<<<<<<<<<<<<
 *         if rv2 == 1: return 1
 *         return 0
 */
  __pyx_v_rv2 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel2->__pyx_vtab)->select_sphere(__pyx_v_self->__pyx_base.sel2, __pyx_v_pos, __pyx_v_radius);

  /* "yt/geometry/selection_routines.pyx":2194
 *         if rv1 == 1: return 1
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 *         if rv2 == 1: return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv2 == 1) != 0);
  if (__pyx_t_1) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2195
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 *         if rv2 == 1: return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2190
 *         return 0
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         if rv1 == 1: return 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2197
 *         return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (self.sel1._hash_vals() +
 *                 ("or",) +
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17BooleanORSelector_1_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17BooleanORSelector_1_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_17BooleanORSelector__hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17BooleanORSelector__hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":2198
 * 
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +             # <<<<<<<<<<<<<<
 *                 ("or",) +
 *                 self.sel2._hash_vals())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.sel1), __pyx_n_s_hash_vals); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":2199
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +
 *                 ("or",) +             # <<<<<<<<<<<<<<
 *                 self.sel2._hash_vals())
 * 
 */
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_tuple__49); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":2200
 *         return (self.sel1._hash_vals() +
 *                 ("or",) +
 *                 self.sel2._hash_vals())             # <<<<<<<<<<<<<<
 * 
 * cdef class BooleanNOTSelector(BooleanSelector):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.sel2), __pyx_n_s_hash_vals); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":2199
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +
 *                 ("or",) +             # <<<<<<<<<<<<<<
 *                 self.sel2._hash_vals())
 * 
 */
  __pyx_t_3 = PyNumber_Add(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2197
 *         return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (self.sel1._hash_vals() +
 *                 ("or",) +
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanORSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17BooleanORSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17BooleanORSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_17BooleanORSelector_2__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17BooleanORSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__50, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanORSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17BooleanORSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_17BooleanORSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_17BooleanORSelector_4__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_17BooleanORSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__51, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanORSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2203
 * 
 * cdef class BooleanNOTSelector(BooleanSelector):
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         # We always return True here, because we don't have a "fully included"
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_bbox(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":2207
 *         # We always return True here, because we don't have a "fully included"
 *         # check anywhere else.
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2203
 * 
 * cdef class BooleanNOTSelector(BooleanSelector):
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         # We always return True here, because we don't have a "fully included"
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2209
 *         return 1
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_grid(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge, CYTHON_UNUSED __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_grid *__pyx_optional_args) {
  int __pyx_r;
  if (__pyx_optional_args) {
  }

  /* "yt/geometry/selection_routines.pyx":2212
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2209
 *         return 1
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2214
 *         return 1
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         if rv1 == 0: return 1
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_v_rv1;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2215
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)             # <<<<<<<<<<<<<<
 *         if rv1 == 0: return 1
 *         return 0
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_cell(__pyx_v_self->__pyx_base.sel1, __pyx_v_pos, __pyx_v_dds);

  /* "yt/geometry/selection_routines.pyx":2216
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         if rv1 == 0: return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv1 == 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2217
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         if rv1 == 0: return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2214
 *         return 1
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         if rv1 == 0: return 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2219
 *         return 0
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         if rv1 == 0: return 1
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_v_rv1;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2220
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         cdef int rv1 = self.sel1.select_point(pos)             # <<<<<<<<<<<<<<
 *         if rv1 == 0: return 1
 *         return 0
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_point(__pyx_v_self->__pyx_base.sel1, __pyx_v_pos);

  /* "yt/geometry/selection_routines.pyx":2221
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         if rv1 == 0: return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv1 == 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2222
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         if rv1 == 0: return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2219
 *         return 0
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         if rv1 == 0: return 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2224
 *         return 0
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         if rv1 == 0: return 1
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_v_rv1;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2225
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)             # <<<<<<<<<<<<<<
 *         if rv1 == 0: return 1
 *         return 0
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_sphere(__pyx_v_self->__pyx_base.sel1, __pyx_v_pos, __pyx_v_radius);

  /* "yt/geometry/selection_routines.pyx":2226
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         if rv1 == 0: return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv1 == 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2227
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         if rv1 == 0: return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2224
 *         return 0
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         if rv1 == 0: return 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2229
 *         return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (self.sel1._hash_vals() +
 *                 ("not",))
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNOTSelector_1_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNOTSelector_1_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_18BooleanNOTSelector__hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanNOTSelector__hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":2230
 * 
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +             # <<<<<<<<<<<<<<
 *                 ("not",))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.sel1), __pyx_n_s_hash_vals); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":2231
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +
 *                 ("not",))             # <<<<<<<<<<<<<<
 * 
 * cdef class BooleanXORSelector(BooleanSelector):
 */
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_tuple__52); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2229
 *         return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (self.sel1._hash_vals() +
 *                 ("not",))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanNOTSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNOTSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNOTSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_18BooleanNOTSelector_2__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanNOTSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__53, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanNOTSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNOTSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNOTSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_18BooleanNOTSelector_4__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanNOTSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__54, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanNOTSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2235
 * cdef class BooleanXORSelector(BooleanSelector):
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         # We always return True here, because we don't have a "fully included"
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_bbox(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":2239
 *         # We always return True here, because we don't have a "fully included"
 *         # check anywhere else.
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2235
 * cdef class BooleanXORSelector(BooleanSelector):
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         # We always return True here, because we don't have a "fully included"
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2241
 *         return 1
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_grid(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_v_self, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_left_edge, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_right_edge, CYTHON_UNUSED __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_grid *__pyx_optional_args) {
  int __pyx_r;
  if (__pyx_optional_args) {
  }

  /* "yt/geometry/selection_routines.pyx":2244
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2241
 *         return 1
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2246
 *         return 1
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_v_rv1;
  int __pyx_v_rv2;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2247
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)             # <<<<<<<<<<<<<<
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 *         if rv1 == rv2: return 0
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_cell(__pyx_v_self->__pyx_base.sel1, __pyx_v_pos, __pyx_v_dds);

  /* "yt/geometry/selection_routines.pyx":2248
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)             # <<<<<<<<<<<<<<
 *         if rv1 == rv2: return 0
 *         return 1
 */
  __pyx_v_rv2 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel2->__pyx_vtab)->select_cell(__pyx_v_self->__pyx_base.sel2, __pyx_v_pos, __pyx_v_dds);

  /* "yt/geometry/selection_routines.pyx":2249
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 *         if rv1 == rv2: return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv1 == __pyx_v_rv2) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2250
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 *         if rv1 == rv2: return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2246
 *         return 1
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2252
 *         return 1
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         cdef int rv2 = self.sel2.select_point(pos)
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_v_rv1;
  int __pyx_v_rv2;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2253
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         cdef int rv1 = self.sel1.select_point(pos)             # <<<<<<<<<<<<<<
 *         cdef int rv2 = self.sel2.select_point(pos)
 *         if rv1 == rv2: return 0
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_point(__pyx_v_self->__pyx_base.sel1, __pyx_v_pos);

  /* "yt/geometry/selection_routines.pyx":2254
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         cdef int rv2 = self.sel2.select_point(pos)             # <<<<<<<<<<<<<<
 *         if rv1 == rv2: return 0
 *         return 1
 */
  __pyx_v_rv2 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel2->__pyx_vtab)->select_point(__pyx_v_self->__pyx_base.sel2, __pyx_v_pos);

  /* "yt/geometry/selection_routines.pyx":2255
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         cdef int rv2 = self.sel2.select_point(pos)
 *         if rv1 == rv2: return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv1 == __pyx_v_rv2) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2256
 *         cdef int rv2 = self.sel2.select_point(pos)
 *         if rv1 == rv2: return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2252
 *         return 1
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         cdef int rv2 = self.sel2.select_point(pos)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2258
 *         return 1
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_v_rv1;
  int __pyx_v_rv2;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2259
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)             # <<<<<<<<<<<<<<
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 *         if rv1 == rv2: return 0
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_sphere(__pyx_v_self->__pyx_base.sel1, __pyx_v_pos, __pyx_v_radius);

  /* "yt/geometry/selection_routines.pyx":2260
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)             # <<<<<<<<<<<<<<
 *         if rv1 == rv2: return 0
 *         return 1
 */
  __pyx_v_rv2 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel2->__pyx_vtab)->select_sphere(__pyx_v_self->__pyx_base.sel2, __pyx_v_pos, __pyx_v_radius);

  /* "yt/geometry/selection_routines.pyx":2261
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 *         if rv1 == rv2: return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv1 == __pyx_v_rv2) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2262
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 *         if rv1 == rv2: return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2258
 *         return 1
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2264
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (self.sel1._hash_vals() +
 *                 ("xor",) +
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanXORSelector_1_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanXORSelector_1_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_18BooleanXORSelector__hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanXORSelector__hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":2265
 * 
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +             # <<<<<<<<<<<<<<
 *                 ("xor",) +
 *                 self.sel2._hash_vals())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.sel1), __pyx_n_s_hash_vals); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":2266
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +
 *                 ("xor",) +             # <<<<<<<<<<<<<<
 *                 self.sel2._hash_vals())
 * 
 */
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_tuple__55); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":2267
 *         return (self.sel1._hash_vals() +
 *                 ("xor",) +
 *                 self.sel2._hash_vals())             # <<<<<<<<<<<<<<
 * 
 * cdef class BooleanNEGSelector(BooleanSelector):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.sel2), __pyx_n_s_hash_vals); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":2266
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +
 *                 ("xor",) +             # <<<<<<<<<<<<<<
 *                 self.sel2._hash_vals())
 * 
 */
  __pyx_t_3 = PyNumber_Add(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2264
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (self.sel1._hash_vals() +
 *                 ("xor",) +
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanXORSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanXORSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanXORSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_18BooleanXORSelector_2__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanXORSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__56, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanXORSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanXORSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanXORSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_18BooleanXORSelector_4__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanXORSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__57, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanXORSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2271
 * cdef class BooleanNEGSelector(BooleanSelector):
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         # We always return True here, because we don't have a "fully included"
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_r;

  /* "yt/geometry/selection_routines.pyx":2275
 *         # We always return True here, because we don't have a "fully included"
 *         # check anywhere else.
 *         return self.sel1.select_bbox(left_edge, right_edge)             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],
 */
  __pyx_r = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_bbox(__pyx_v_self->__pyx_base.sel1, __pyx_v_left_edge, __pyx_v_right_edge);
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2271
 * cdef class BooleanNEGSelector(BooleanSelector):
 * 
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3]) nogil:
 *         # We always return True here, because we don't have a "fully included"
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2277
 *         return self.sel1.select_bbox(left_edge, right_edge)
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_grid *__pyx_optional_args) {

  /* "yt/geometry/selection_routines.pyx":2279
 *     cdef int select_grid(self, np.float64_t left_edge[3],
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:             # <<<<<<<<<<<<<<
 *         return self.sel1.select_grid(left_edge, right_edge, level, o)
 * 
 */
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o = ((struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *)NULL);
  int __pyx_r;
  int __pyx_t_1;
  struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid __pyx_t_2;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_o = __pyx_optional_args->o;
    }
  }

  /* "yt/geometry/selection_routines.pyx":2280
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 *         return self.sel1.select_grid(left_edge, right_edge, level, o)             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.o = __pyx_v_o;
  __pyx_t_1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_grid(__pyx_v_self->__pyx_base.sel1, __pyx_v_left_edge, __pyx_v_right_edge, __pyx_v_level, &__pyx_t_2); 
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2277
 *         return self.sel1.select_bbox(left_edge, right_edge)
 * 
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2282
 *         return self.sel1.select_grid(left_edge, right_edge, level, o)
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         if rv1 == 0: return 0
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_v_rv1;
  int __pyx_v_rv2;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2283
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)             # <<<<<<<<<<<<<<
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_cell(__pyx_v_self->__pyx_base.sel1, __pyx_v_pos, __pyx_v_dds);

  /* "yt/geometry/selection_routines.pyx":2284
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         if rv1 == 0: return 0             # <<<<<<<<<<<<<<
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 *         if rv2 == 1: return 0
 */
  __pyx_t_1 = ((__pyx_v_rv1 == 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2285
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)             # <<<<<<<<<<<<<<
 *         if rv2 == 1: return 0
 *         return 1
 */
  __pyx_v_rv2 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel2->__pyx_vtab)->select_cell(__pyx_v_self->__pyx_base.sel2, __pyx_v_pos, __pyx_v_dds);

  /* "yt/geometry/selection_routines.pyx":2286
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 *         if rv2 == 1: return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv2 == 1) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2287
 *         cdef int rv2 = self.sel2.select_cell(pos, dds)
 *         if rv2 == 1: return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2282
 *         return self.sel1.select_grid(left_edge, right_edge, level, o)
 * 
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_cell(pos, dds)
 *         if rv1 == 0: return 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2289
 *         return 1
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         if rv1 == 0: return 0
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_v_rv1;
  int __pyx_v_rv2;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2290
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         cdef int rv1 = self.sel1.select_point(pos)             # <<<<<<<<<<<<<<
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_point(pos)
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_point(__pyx_v_self->__pyx_base.sel1, __pyx_v_pos);

  /* "yt/geometry/selection_routines.pyx":2291
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         if rv1 == 0: return 0             # <<<<<<<<<<<<<<
 *         cdef int rv2 = self.sel2.select_point(pos)
 *         if rv2 == 1: return 0
 */
  __pyx_t_1 = ((__pyx_v_rv1 == 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2292
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_point(pos)             # <<<<<<<<<<<<<<
 *         if rv2 == 1: return 0
 *         return 1
 */
  __pyx_v_rv2 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel2->__pyx_vtab)->select_point(__pyx_v_self->__pyx_base.sel2, __pyx_v_pos);

  /* "yt/geometry/selection_routines.pyx":2293
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_point(pos)
 *         if rv2 == 1: return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv2 == 1) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2294
 *         cdef int rv2 = self.sel2.select_point(pos)
 *         if rv2 == 1: return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2289
 *         return 1
 * 
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_point(pos)
 *         if rv1 == 0: return 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2296
 *         return 1
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         if rv1 == 0: return 0
 */

static int __pyx_f_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_v_rv1;
  int __pyx_v_rv2;
  int __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pyx":2297
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)             # <<<<<<<<<<<<<<
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 */
  __pyx_v_rv1 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel1->__pyx_vtab)->select_sphere(__pyx_v_self->__pyx_base.sel1, __pyx_v_pos, __pyx_v_radius);

  /* "yt/geometry/selection_routines.pyx":2298
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         if rv1 == 0: return 0             # <<<<<<<<<<<<<<
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 *         if rv2 == 1: return 0
 */
  __pyx_t_1 = ((__pyx_v_rv1 == 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2299
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)             # <<<<<<<<<<<<<<
 *         if rv2 == 1: return 0
 *         return 1
 */
  __pyx_v_rv2 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->__pyx_base.sel2->__pyx_vtab)->select_sphere(__pyx_v_self->__pyx_base.sel2, __pyx_v_pos, __pyx_v_radius);

  /* "yt/geometry/selection_routines.pyx":2300
 *         if rv1 == 0: return 0
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 *         if rv2 == 1: return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv2 == 1) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pyx":2301
 *         cdef int rv2 = self.sel2.select_sphere(pos, radius)
 *         if rv2 == 1: return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2296
 *         return 1
 * 
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         cdef int rv1 = self.sel1.select_sphere(pos, radius)
 *         if rv1 == 0: return 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2303
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (self.sel1._hash_vals() +
 *                 ("neg",) +
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNEGSelector_1_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNEGSelector_1_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_18BooleanNEGSelector__hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanNEGSelector__hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":2304
 * 
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +             # <<<<<<<<<<<<<<
 *                 ("neg",) +
 *                 self.sel2._hash_vals())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.sel1), __pyx_n_s_hash_vals); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/selection_routines.pyx":2305
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +
 *                 ("neg",) +             # <<<<<<<<<<<<<<
 *                 self.sel2._hash_vals())
 * 
 */
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_tuple__58); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":2306
 *         return (self.sel1._hash_vals() +
 *                 ("neg",) +
 *                 self.sel2._hash_vals())             # <<<<<<<<<<<<<<
 * 
 * cdef class ChainedBooleanSelector(SelectorObject):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.sel2), __pyx_n_s_hash_vals); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":2305
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +
 *                 ("neg",) +             # <<<<<<<<<<<<<<
 *                 self.sel2._hash_vals())
 * 
 */
  __pyx_t_3 = PyNumber_Add(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2303
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         return (self.sel1._hash_vals() +
 *                 ("neg",) +
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanNEGSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNEGSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNEGSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_18BooleanNEGSelector_2__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanNEGSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__59, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanNEGSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNEGSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNEGSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_18BooleanNEGSelector_4__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_18BooleanNEGSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__60, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.BooleanNEGSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2311
 *     cdef int n_obj
 *     cdef np.ndarray selectors
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         # These are data objects, not selectors
 *         self.n_obj = len(dobj.data_objects)
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dobj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dobj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dobj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2311, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_dobj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2311, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.selection_routines.ChainedBooleanSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_22ChainedBooleanSelector___init__(((struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *)__pyx_v_self), __pyx_v_dobj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_18selection_routines_22ChainedBooleanSelector___init__(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *__pyx_v_self, PyObject *__pyx_v_dobj) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/selection_routines.pyx":2313
 *     def __init__(self, dobj):
 *         # These are data objects, not selectors
 *         self.n_obj = len(dobj.data_objects)             # <<<<<<<<<<<<<<
 *         self.selectors = np.empty(self.n_obj, dtype="object")
 *         for i in range(self.n_obj):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_data_objects); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2313, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->n_obj = __pyx_t_2;

  /* "yt/geometry/selection_routines.pyx":2314
 *         # These are data objects, not selectors
 *         self.n_obj = len(dobj.data_objects)
 *         self.selectors = np.empty(self.n_obj, dtype="object")             # <<<<<<<<<<<<<<
 *         for i in range(self.n_obj):
 *             self.selectors[i] = dobj.data_objects[i].selector
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->n_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_object) < 0) __PYX_ERR(0, 2314, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 2314, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v_self->selectors);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->selectors));
  __pyx_v_self->selectors = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "yt/geometry/selection_routines.pyx":2315
 *         self.n_obj = len(dobj.data_objects)
 *         self.selectors = np.empty(self.n_obj, dtype="object")
 *         for i in range(self.n_obj):             # <<<<<<<<<<<<<<
 *             self.selectors[i] = dobj.data_objects[i].selector
 * 
 */
  __pyx_t_6 = __pyx_v_self->n_obj;
  __pyx_t_7 = __pyx_t_6;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_i = __pyx_t_8;

    /* "yt/geometry/selection_routines.pyx":2316
 *         self.selectors = np.empty(self.n_obj, dtype="object")
 *         for i in range(self.n_obj):
 *             self.selectors[i] = dobj.data_objects[i].selector             # <<<<<<<<<<<<<<
 * 
 * cdef class ChainedBooleanANDSelector(ChainedBooleanSelector):
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_dobj, __pyx_n_s_data_objects); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_5, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_selector); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(__Pyx_SetItemInt(((PyObject *)__pyx_v_self->selectors), __pyx_v_i, __pyx_t_5, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 2316, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }

  /* "yt/geometry/selection_routines.pyx":2311
 *     cdef int n_obj
 *     cdef np.ndarray selectors
 *     def __init__(self, dobj):             # <<<<<<<<<<<<<<
 *         # These are data objects, not selectors
 *         self.n_obj = len(dobj.data_objects)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("yt.geometry.selection_routines.ChainedBooleanSelector.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_2__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__61, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.ChainedBooleanSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_4__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__62, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.ChainedBooleanSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2322
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3]) nogil:
 *         with gil:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("select_bbox", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "yt/geometry/selection_routines.pyx":2324
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                          np.float64_t right_edge[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(
 */
  /*try:*/ {
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "yt/geometry/selection_routines.pyx":2325
 *                          np.float64_t right_edge[3]) nogil:
 *         with gil:
 *             for i in range(self.n_obj):             # <<<<<<<<<<<<<<
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(
 *                         left_edge, right_edge) == 0:
 */
          __pyx_t_1 = __pyx_v_self->__pyx_base.n_obj;
          __pyx_t_2 = __pyx_t_1;
          for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
            __pyx_v_i = __pyx_t_3;

            /* "yt/geometry/selection_routines.pyx":2326
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(             # <<<<<<<<<<<<<<
 *                         left_edge, right_edge) == 0:
 *                     return 0
 */
            __pyx_t_4 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self->__pyx_base.selectors), __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2326, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);

            /* "yt/geometry/selection_routines.pyx":2327
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(
 *                         left_edge, right_edge) == 0:             # <<<<<<<<<<<<<<
 *                     return 0
 *         return 1
 */
            __pyx_t_5 = ((((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4)->__pyx_vtab)->select_bbox(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4), __pyx_v_left_edge, __pyx_v_right_edge) == 0) != 0);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "yt/geometry/selection_routines.pyx":2326
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(             # <<<<<<<<<<<<<<
 *                         left_edge, right_edge) == 0:
 *                     return 0
 */
            if (__pyx_t_5) {

              /* "yt/geometry/selection_routines.pyx":2328
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(
 *                         left_edge, right_edge) == 0:
 *                     return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
              __pyx_r = 0;
              goto __pyx_L6_return;

              /* "yt/geometry/selection_routines.pyx":2326
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(             # <<<<<<<<<<<<<<
 *                         left_edge, right_edge) == 0:
 *                     return 0
 */
            }
          }
        }

        /* "yt/geometry/selection_routines.pyx":2324
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                          np.float64_t right_edge[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L6_return: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L3_return;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }

    /* "yt/geometry/selection_routines.pyx":2329
 *                         left_edge, right_edge) == 0:
 *                     return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
    __pyx_r = 1;
    goto __pyx_L3_return;
  }

  /* "yt/geometry/selection_routines.pyx":2324
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                          np.float64_t right_edge[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "yt/geometry/selection_routines.pyx":2322
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3]) nogil:
 *         with gil:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("yt.geometry.selection_routines.ChainedBooleanANDSelector.select_bbox", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2334
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_grid *__pyx_optional_args) {

  /* "yt/geometry/selection_routines.pyx":2336
 *     cdef int select_grid(self, np.float64_t left_edge[3],
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:             # <<<<<<<<<<<<<<
 *         with gil:
 *             for i in range(self.n_obj):
 */
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o = ((struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *)NULL);
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("select_grid", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_o = __pyx_optional_args->o;
    }
  }
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "yt/geometry/selection_routines.pyx":2337
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_grid(
 */
  /*try:*/ {
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "yt/geometry/selection_routines.pyx":2338
 *                          Oct *o = NULL) nogil:
 *         with gil:
 *             for i in range(self.n_obj):             # <<<<<<<<<<<<<<
 *                 if (<SelectorObject>self.selectors[i]).select_grid(
 *                         left_edge, right_edge, level, o) == 0:
 */
          __pyx_t_1 = __pyx_v_self->__pyx_base.n_obj;
          __pyx_t_2 = __pyx_t_1;
          for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
            __pyx_v_i = __pyx_t_3;

            /* "yt/geometry/selection_routines.pyx":2339
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_grid(             # <<<<<<<<<<<<<<
 *                         left_edge, right_edge, level, o) == 0:
 *                     return 0
 */
            __pyx_t_4 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self->__pyx_base.selectors), __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2339, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);

            /* "yt/geometry/selection_routines.pyx":2340
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_grid(
 *                         left_edge, right_edge, level, o) == 0:             # <<<<<<<<<<<<<<
 *                     return 0
 *         return 1
 */
            __pyx_t_6.__pyx_n = 1;
            __pyx_t_6.o = __pyx_v_o;
            __pyx_t_5 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4)->__pyx_vtab)->select_grid(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4), __pyx_v_left_edge, __pyx_v_right_edge, __pyx_v_level, &__pyx_t_6); 
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_7 = ((__pyx_t_5 == 0) != 0);

            /* "yt/geometry/selection_routines.pyx":2339
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_grid(             # <<<<<<<<<<<<<<
 *                         left_edge, right_edge, level, o) == 0:
 *                     return 0
 */
            if (__pyx_t_7) {

              /* "yt/geometry/selection_routines.pyx":2341
 *                 if (<SelectorObject>self.selectors[i]).select_grid(
 *                         left_edge, right_edge, level, o) == 0:
 *                     return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
              __pyx_r = 0;
              goto __pyx_L6_return;

              /* "yt/geometry/selection_routines.pyx":2339
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_grid(             # <<<<<<<<<<<<<<
 *                         left_edge, right_edge, level, o) == 0:
 *                     return 0
 */
            }
          }
        }

        /* "yt/geometry/selection_routines.pyx":2337
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_grid(
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L6_return: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L3_return;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }

    /* "yt/geometry/selection_routines.pyx":2342
 *                         left_edge, right_edge, level, o) == 0:
 *                     return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
    __pyx_r = 1;
    goto __pyx_L3_return;
  }

  /* "yt/geometry/selection_routines.pyx":2337
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_grid(
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "yt/geometry/selection_routines.pyx":2334
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("yt.geometry.selection_routines.ChainedBooleanANDSelector.select_grid", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2347
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         with gil:
 *             for i in range(self.n_obj):
 */

static int __pyx_f_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("select_cell", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "yt/geometry/selection_routines.pyx":2348
 *     @cython.wraparound(False)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_cell(
 */
  /*try:*/ {
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "yt/geometry/selection_routines.pyx":2349
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         with gil:
 *             for i in range(self.n_obj):             # <<<<<<<<<<<<<<
 *                 if (<SelectorObject>self.selectors[i]).select_cell(
 *                         pos, dds) == 0:
 */
          __pyx_t_1 = __pyx_v_self->__pyx_base.n_obj;
          __pyx_t_2 = __pyx_t_1;
          for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
            __pyx_v_i = __pyx_t_3;

            /* "yt/geometry/selection_routines.pyx":2350
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_cell(             # <<<<<<<<<<<<<<
 *                         pos, dds) == 0:
 *                     return 0
 */
            __pyx_t_4 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self->__pyx_base.selectors), __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2350, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);

            /* "yt/geometry/selection_routines.pyx":2351
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_cell(
 *                         pos, dds) == 0:             # <<<<<<<<<<<<<<
 *                     return 0
 *         return 1
 */
            __pyx_t_5 = ((((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4)->__pyx_vtab)->select_cell(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4), __pyx_v_pos, __pyx_v_dds) == 0) != 0);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "yt/geometry/selection_routines.pyx":2350
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_cell(             # <<<<<<<<<<<<<<
 *                         pos, dds) == 0:
 *                     return 0
 */
            if (__pyx_t_5) {

              /* "yt/geometry/selection_routines.pyx":2352
 *                 if (<SelectorObject>self.selectors[i]).select_cell(
 *                         pos, dds) == 0:
 *                     return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
              __pyx_r = 0;
              goto __pyx_L6_return;

              /* "yt/geometry/selection_routines.pyx":2350
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_cell(             # <<<<<<<<<<<<<<
 *                         pos, dds) == 0:
 *                     return 0
 */
            }
          }
        }

        /* "yt/geometry/selection_routines.pyx":2348
 *     @cython.wraparound(False)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_cell(
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L6_return: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L3_return;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }

    /* "yt/geometry/selection_routines.pyx":2353
 *                         pos, dds) == 0:
 *                     return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
    __pyx_r = 1;
    goto __pyx_L3_return;
  }

  /* "yt/geometry/selection_routines.pyx":2348
 *     @cython.wraparound(False)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_cell(
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "yt/geometry/selection_routines.pyx":2347
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         with gil:
 *             for i in range(self.n_obj):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("yt.geometry.selection_routines.ChainedBooleanANDSelector.select_cell", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2358
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         with gil:
 *             for i in range(self.n_obj):
 */

static int __pyx_f_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("select_point", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "yt/geometry/selection_routines.pyx":2359
 *     @cython.wraparound(False)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_point(pos) == 0:
 */
  /*try:*/ {
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "yt/geometry/selection_routines.pyx":2360
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         with gil:
 *             for i in range(self.n_obj):             # <<<<<<<<<<<<<<
 *                 if (<SelectorObject>self.selectors[i]).select_point(pos) == 0:
 *                     return 0
 */
          __pyx_t_1 = __pyx_v_self->__pyx_base.n_obj;
          __pyx_t_2 = __pyx_t_1;
          for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
            __pyx_v_i = __pyx_t_3;

            /* "yt/geometry/selection_routines.pyx":2361
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_point(pos) == 0:             # <<<<<<<<<<<<<<
 *                     return 0
 *         return 1
 */
            __pyx_t_4 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self->__pyx_base.selectors), __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2361, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_5 = ((((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4)->__pyx_vtab)->select_point(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4), __pyx_v_pos) == 0) != 0);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (__pyx_t_5) {

              /* "yt/geometry/selection_routines.pyx":2362
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_point(pos) == 0:
 *                     return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
              __pyx_r = 0;
              goto __pyx_L6_return;

              /* "yt/geometry/selection_routines.pyx":2361
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_point(pos) == 0:             # <<<<<<<<<<<<<<
 *                     return 0
 *         return 1
 */
            }
          }
        }

        /* "yt/geometry/selection_routines.pyx":2359
 *     @cython.wraparound(False)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_point(pos) == 0:
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L6_return: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L3_return;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }

    /* "yt/geometry/selection_routines.pyx":2363
 *                 if (<SelectorObject>self.selectors[i]).select_point(pos) == 0:
 *                     return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
    __pyx_r = 1;
    goto __pyx_L3_return;
  }

  /* "yt/geometry/selection_routines.pyx":2359
 *     @cython.wraparound(False)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_point(pos) == 0:
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "yt/geometry/selection_routines.pyx":2358
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         with gil:
 *             for i in range(self.n_obj):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("yt.geometry.selection_routines.ChainedBooleanANDSelector.select_point", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2368
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         with gil:
 *             for i in range(self.n_obj):
 */

static int __pyx_f_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("select_sphere", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "yt/geometry/selection_routines.pyx":2369
 *     @cython.wraparound(False)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(
 */
  /*try:*/ {
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "yt/geometry/selection_routines.pyx":2370
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         with gil:
 *             for i in range(self.n_obj):             # <<<<<<<<<<<<<<
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(
 *                         pos, radius) == 0:
 */
          __pyx_t_1 = __pyx_v_self->__pyx_base.n_obj;
          __pyx_t_2 = __pyx_t_1;
          for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
            __pyx_v_i = __pyx_t_3;

            /* "yt/geometry/selection_routines.pyx":2371
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(             # <<<<<<<<<<<<<<
 *                         pos, radius) == 0:
 *                     return 0
 */
            __pyx_t_4 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self->__pyx_base.selectors), __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2371, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);

            /* "yt/geometry/selection_routines.pyx":2372
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(
 *                         pos, radius) == 0:             # <<<<<<<<<<<<<<
 *                     return 0
 *         return 1
 */
            __pyx_t_5 = ((((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4)->__pyx_vtab)->select_sphere(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4), __pyx_v_pos, __pyx_v_radius) == 0) != 0);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "yt/geometry/selection_routines.pyx":2371
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(             # <<<<<<<<<<<<<<
 *                         pos, radius) == 0:
 *                     return 0
 */
            if (__pyx_t_5) {

              /* "yt/geometry/selection_routines.pyx":2373
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(
 *                         pos, radius) == 0:
 *                     return 0             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
              __pyx_r = 0;
              goto __pyx_L6_return;

              /* "yt/geometry/selection_routines.pyx":2371
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(             # <<<<<<<<<<<<<<
 *                         pos, radius) == 0:
 *                     return 0
 */
            }
          }
        }

        /* "yt/geometry/selection_routines.pyx":2369
 *     @cython.wraparound(False)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L6_return: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L3_return;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }

    /* "yt/geometry/selection_routines.pyx":2374
 *                         pos, radius) == 0:
 *                     return 0
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
    __pyx_r = 1;
    goto __pyx_L3_return;
  }

  /* "yt/geometry/selection_routines.pyx":2369
 *     @cython.wraparound(False)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "yt/geometry/selection_routines.pyx":2368
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         with gil:
 *             for i in range(self.n_obj):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("yt.geometry.selection_routines.ChainedBooleanANDSelector.select_sphere", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2376
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         v = ("chained_and",)
 *         for s in self.selectors:
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_1_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_1_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector__hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector__hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_v_self) {
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":2377
 * 
 *     def _hash_vals(self):
 *         v = ("chained_and",)             # <<<<<<<<<<<<<<
 *         for s in self.selectors:
 *             v += s._hash_vals()
 */
  __Pyx_INCREF(__pyx_tuple__63);
  __pyx_v_v = __pyx_tuple__63;

  /* "yt/geometry/selection_routines.pyx":2378
 *     def _hash_vals(self):
 *         v = ("chained_and",)
 *         for s in self.selectors:             # <<<<<<<<<<<<<<
 *             v += s._hash_vals()
 *         return v
 */
  if (likely(PyList_CheckExact(((PyObject *)__pyx_v_self->__pyx_base.selectors))) || PyTuple_CheckExact(((PyObject *)__pyx_v_self->__pyx_base.selectors))) {
    __pyx_t_1 = ((PyObject *)__pyx_v_self->__pyx_base.selectors); __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(((PyObject *)__pyx_v_self->__pyx_base.selectors)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2378, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2378, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 2378, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2378, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 2378, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2378, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2378, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "yt/geometry/selection_routines.pyx":2379
 *         v = ("chained_and",)
 *         for s in self.selectors:
 *             v += s._hash_vals()             # <<<<<<<<<<<<<<
 *         return v
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_s, __pyx_n_s_hash_vals); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_v_v, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_v, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "yt/geometry/selection_routines.pyx":2378
 *     def _hash_vals(self):
 *         v = ("chained_and",)
 *         for s in self.selectors:             # <<<<<<<<<<<<<<
 *             v += s._hash_vals()
 *         return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":2380
 *         for s in self.selectors:
 *             v += s._hash_vals()
 *         return v             # <<<<<<<<<<<<<<
 * 
 * intersection_selector = ChainedBooleanANDSelector
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_v);
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2376
 *         return 1
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         v = ("chained_and",)
 *         for s in self.selectors:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("yt.geometry.selection_routines.ChainedBooleanANDSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_2__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__64, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.ChainedBooleanANDSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_4__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__65, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.ChainedBooleanANDSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2388
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3]) nogil:
 *         with gil:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_bbox(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("select_bbox", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "yt/geometry/selection_routines.pyx":2390
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                          np.float64_t right_edge[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(
 */
  /*try:*/ {
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "yt/geometry/selection_routines.pyx":2391
 *                          np.float64_t right_edge[3]) nogil:
 *         with gil:
 *             for i in range(self.n_obj):             # <<<<<<<<<<<<<<
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(
 *                         left_edge, right_edge) == 1:
 */
          __pyx_t_1 = __pyx_v_self->__pyx_base.n_obj;
          __pyx_t_2 = __pyx_t_1;
          for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
            __pyx_v_i = __pyx_t_3;

            /* "yt/geometry/selection_routines.pyx":2392
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(             # <<<<<<<<<<<<<<
 *                         left_edge, right_edge) == 1:
 *                     return 1
 */
            __pyx_t_4 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self->__pyx_base.selectors), __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2392, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);

            /* "yt/geometry/selection_routines.pyx":2393
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(
 *                         left_edge, right_edge) == 1:             # <<<<<<<<<<<<<<
 *                     return 1
 *         return 0
 */
            __pyx_t_5 = ((((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4)->__pyx_vtab)->select_bbox(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4), __pyx_v_left_edge, __pyx_v_right_edge) == 1) != 0);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "yt/geometry/selection_routines.pyx":2392
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(             # <<<<<<<<<<<<<<
 *                         left_edge, right_edge) == 1:
 *                     return 1
 */
            if (__pyx_t_5) {

              /* "yt/geometry/selection_routines.pyx":2394
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(
 *                         left_edge, right_edge) == 1:
 *                     return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
              __pyx_r = 1;
              goto __pyx_L6_return;

              /* "yt/geometry/selection_routines.pyx":2392
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(             # <<<<<<<<<<<<<<
 *                         left_edge, right_edge) == 1:
 *                     return 1
 */
            }
          }
        }

        /* "yt/geometry/selection_routines.pyx":2390
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                          np.float64_t right_edge[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L6_return: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L3_return;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }

    /* "yt/geometry/selection_routines.pyx":2395
 *                         left_edge, right_edge) == 1:
 *                     return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
    __pyx_r = 0;
    goto __pyx_L3_return;
  }

  /* "yt/geometry/selection_routines.pyx":2390
 *     cdef int select_bbox(self, np.float64_t left_edge[3],
 *                          np.float64_t right_edge[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_bbox(
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "yt/geometry/selection_routines.pyx":2388
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_bbox(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3]) nogil:
 *         with gil:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("yt.geometry.selection_routines.ChainedBooleanORSelector.select_bbox", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2400
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

static int __pyx_f_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_grid(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_left_edge, __pyx_t_5numpy_float64_t *__pyx_v_right_edge, __pyx_t_5numpy_int32_t __pyx_v_level, struct __pyx_opt_args_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_grid *__pyx_optional_args) {

  /* "yt/geometry/selection_routines.pyx":2402
 *     cdef int select_grid(self, np.float64_t left_edge[3],
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:             # <<<<<<<<<<<<<<
 *         with gil:
 *             for i in range(self.n_obj):
 */
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o = ((struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *)NULL);
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("select_grid", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_o = __pyx_optional_args->o;
    }
  }
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "yt/geometry/selection_routines.pyx":2403
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_grid(
 */
  /*try:*/ {
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "yt/geometry/selection_routines.pyx":2404
 *                          Oct *o = NULL) nogil:
 *         with gil:
 *             for i in range(self.n_obj):             # <<<<<<<<<<<<<<
 *                 if (<SelectorObject>self.selectors[i]).select_grid(
 *                         left_edge, right_edge, level, o) == 1:
 */
          __pyx_t_1 = __pyx_v_self->__pyx_base.n_obj;
          __pyx_t_2 = __pyx_t_1;
          for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
            __pyx_v_i = __pyx_t_3;

            /* "yt/geometry/selection_routines.pyx":2405
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_grid(             # <<<<<<<<<<<<<<
 *                         left_edge, right_edge, level, o) == 1:
 *                     return 1
 */
            __pyx_t_4 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self->__pyx_base.selectors), __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2405, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);

            /* "yt/geometry/selection_routines.pyx":2406
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_grid(
 *                         left_edge, right_edge, level, o) == 1:             # <<<<<<<<<<<<<<
 *                     return 1
 *         return 0
 */
            __pyx_t_6.__pyx_n = 1;
            __pyx_t_6.o = __pyx_v_o;
            __pyx_t_5 = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4)->__pyx_vtab)->select_grid(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4), __pyx_v_left_edge, __pyx_v_right_edge, __pyx_v_level, &__pyx_t_6); 
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_7 = ((__pyx_t_5 == 1) != 0);

            /* "yt/geometry/selection_routines.pyx":2405
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_grid(             # <<<<<<<<<<<<<<
 *                         left_edge, right_edge, level, o) == 1:
 *                     return 1
 */
            if (__pyx_t_7) {

              /* "yt/geometry/selection_routines.pyx":2407
 *                 if (<SelectorObject>self.selectors[i]).select_grid(
 *                         left_edge, right_edge, level, o) == 1:
 *                     return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
              __pyx_r = 1;
              goto __pyx_L6_return;

              /* "yt/geometry/selection_routines.pyx":2405
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_grid(             # <<<<<<<<<<<<<<
 *                         left_edge, right_edge, level, o) == 1:
 *                     return 1
 */
            }
          }
        }

        /* "yt/geometry/selection_routines.pyx":2403
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_grid(
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L6_return: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L3_return;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }

    /* "yt/geometry/selection_routines.pyx":2408
 *                         left_edge, right_edge, level, o) == 1:
 *                     return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
    __pyx_r = 0;
    goto __pyx_L3_return;
  }

  /* "yt/geometry/selection_routines.pyx":2403
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_grid(
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "yt/geometry/selection_routines.pyx":2400
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                          np.float64_t right_edge[3], np.int32_t level,
 *                          Oct *o = NULL) nogil:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("yt.geometry.selection_routines.ChainedBooleanORSelector.select_grid", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2413
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         with gil:
 *             for i in range(self.n_obj):
 */

static int __pyx_f_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_cell(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("select_cell", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "yt/geometry/selection_routines.pyx":2414
 *     @cython.wraparound(False)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_cell(
 */
  /*try:*/ {
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "yt/geometry/selection_routines.pyx":2415
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         with gil:
 *             for i in range(self.n_obj):             # <<<<<<<<<<<<<<
 *                 if (<SelectorObject>self.selectors[i]).select_cell(
 *                         pos, dds) == 1:
 */
          __pyx_t_1 = __pyx_v_self->__pyx_base.n_obj;
          __pyx_t_2 = __pyx_t_1;
          for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
            __pyx_v_i = __pyx_t_3;

            /* "yt/geometry/selection_routines.pyx":2416
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_cell(             # <<<<<<<<<<<<<<
 *                         pos, dds) == 1:
 *                     return 1
 */
            __pyx_t_4 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self->__pyx_base.selectors), __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2416, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);

            /* "yt/geometry/selection_routines.pyx":2417
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_cell(
 *                         pos, dds) == 1:             # <<<<<<<<<<<<<<
 *                     return 1
 *         return 0
 */
            __pyx_t_5 = ((((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4)->__pyx_vtab)->select_cell(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4), __pyx_v_pos, __pyx_v_dds) == 1) != 0);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "yt/geometry/selection_routines.pyx":2416
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_cell(             # <<<<<<<<<<<<<<
 *                         pos, dds) == 1:
 *                     return 1
 */
            if (__pyx_t_5) {

              /* "yt/geometry/selection_routines.pyx":2418
 *                 if (<SelectorObject>self.selectors[i]).select_cell(
 *                         pos, dds) == 1:
 *                     return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
              __pyx_r = 1;
              goto __pyx_L6_return;

              /* "yt/geometry/selection_routines.pyx":2416
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_cell(             # <<<<<<<<<<<<<<
 *                         pos, dds) == 1:
 *                     return 1
 */
            }
          }
        }

        /* "yt/geometry/selection_routines.pyx":2414
 *     @cython.wraparound(False)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_cell(
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L6_return: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L3_return;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }

    /* "yt/geometry/selection_routines.pyx":2419
 *                         pos, dds) == 1:
 *                     return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
    __pyx_r = 0;
    goto __pyx_L3_return;
  }

  /* "yt/geometry/selection_routines.pyx":2414
 *     @cython.wraparound(False)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_cell(
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "yt/geometry/selection_routines.pyx":2413
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:             # <<<<<<<<<<<<<<
 *         with gil:
 *             for i in range(self.n_obj):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("yt.geometry.selection_routines.ChainedBooleanORSelector.select_cell", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2424
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         with gil:
 *             for i in range(self.n_obj):
 */

static int __pyx_f_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_point(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("select_point", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "yt/geometry/selection_routines.pyx":2425
 *     @cython.wraparound(False)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_point(pos) == 1:
 */
  /*try:*/ {
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "yt/geometry/selection_routines.pyx":2426
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         with gil:
 *             for i in range(self.n_obj):             # <<<<<<<<<<<<<<
 *                 if (<SelectorObject>self.selectors[i]).select_point(pos) == 1:
 *                     return 1
 */
          __pyx_t_1 = __pyx_v_self->__pyx_base.n_obj;
          __pyx_t_2 = __pyx_t_1;
          for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
            __pyx_v_i = __pyx_t_3;

            /* "yt/geometry/selection_routines.pyx":2427
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_point(pos) == 1:             # <<<<<<<<<<<<<<
 *                     return 1
 *         return 0
 */
            __pyx_t_4 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self->__pyx_base.selectors), __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2427, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_5 = ((((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4)->__pyx_vtab)->select_point(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4), __pyx_v_pos) == 1) != 0);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (__pyx_t_5) {

              /* "yt/geometry/selection_routines.pyx":2428
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_point(pos) == 1:
 *                     return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
              __pyx_r = 1;
              goto __pyx_L6_return;

              /* "yt/geometry/selection_routines.pyx":2427
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_point(pos) == 1:             # <<<<<<<<<<<<<<
 *                     return 1
 *         return 0
 */
            }
          }
        }

        /* "yt/geometry/selection_routines.pyx":2425
 *     @cython.wraparound(False)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_point(pos) == 1:
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L6_return: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L3_return;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }

    /* "yt/geometry/selection_routines.pyx":2429
 *                 if (<SelectorObject>self.selectors[i]).select_point(pos) == 1:
 *                     return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
    __pyx_r = 0;
    goto __pyx_L3_return;
  }

  /* "yt/geometry/selection_routines.pyx":2425
 *     @cython.wraparound(False)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_point(pos) == 1:
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "yt/geometry/selection_routines.pyx":2424
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_point(self, np.float64_t pos[3]) nogil:             # <<<<<<<<<<<<<<
 *         with gil:
 *             for i in range(self.n_obj):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("yt.geometry.selection_routines.ChainedBooleanORSelector.select_point", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2434
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         with gil:
 *             for i in range(self.n_obj):
 */

static int __pyx_f_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_sphere(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t __pyx_v_radius) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("select_sphere", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "yt/geometry/selection_routines.pyx":2435
 *     @cython.wraparound(False)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(
 */
  /*try:*/ {
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "yt/geometry/selection_routines.pyx":2436
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         with gil:
 *             for i in range(self.n_obj):             # <<<<<<<<<<<<<<
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(
 *                         pos, radius) == 1:
 */
          __pyx_t_1 = __pyx_v_self->__pyx_base.n_obj;
          __pyx_t_2 = __pyx_t_1;
          for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
            __pyx_v_i = __pyx_t_3;

            /* "yt/geometry/selection_routines.pyx":2437
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(             # <<<<<<<<<<<<<<
 *                         pos, radius) == 1:
 *                     return 1
 */
            __pyx_t_4 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self->__pyx_base.selectors), __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2437, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);

            /* "yt/geometry/selection_routines.pyx":2438
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(
 *                         pos, radius) == 1:             # <<<<<<<<<<<<<<
 *                     return 1
 *         return 0
 */
            __pyx_t_5 = ((((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4)->__pyx_vtab)->select_sphere(((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_4), __pyx_v_pos, __pyx_v_radius) == 1) != 0);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "yt/geometry/selection_routines.pyx":2437
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(             # <<<<<<<<<<<<<<
 *                         pos, radius) == 1:
 *                     return 1
 */
            if (__pyx_t_5) {

              /* "yt/geometry/selection_routines.pyx":2439
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(
 *                         pos, radius) == 1:
 *                     return 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
              __pyx_r = 1;
              goto __pyx_L6_return;

              /* "yt/geometry/selection_routines.pyx":2437
 *         with gil:
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(             # <<<<<<<<<<<<<<
 *                         pos, radius) == 1:
 *                     return 1
 */
            }
          }
        }

        /* "yt/geometry/selection_routines.pyx":2435
 *     @cython.wraparound(False)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L6_return: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L3_return;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }

    /* "yt/geometry/selection_routines.pyx":2440
 *                         pos, radius) == 1:
 *                     return 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _hash_vals(self):
 */
    __pyx_r = 0;
    goto __pyx_L3_return;
  }

  /* "yt/geometry/selection_routines.pyx":2435
 *     @cython.wraparound(False)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
 *         with gil:             # <<<<<<<<<<<<<<
 *             for i in range(self.n_obj):
 *                 if (<SelectorObject>self.selectors[i]).select_sphere(
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "yt/geometry/selection_routines.pyx":2434
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:             # <<<<<<<<<<<<<<
 *         with gil:
 *             for i in range(self.n_obj):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("yt.geometry.selection_routines.ChainedBooleanORSelector.select_sphere", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pyx":2442
 *         return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         v = ("chained_or",)
 *         for s in self.selectors:
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_1_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_1_hash_vals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hash_vals (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector__hash_vals(((struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector__hash_vals(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_v_self) {
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_hash_vals", 0);

  /* "yt/geometry/selection_routines.pyx":2443
 * 
 *     def _hash_vals(self):
 *         v = ("chained_or",)             # <<<<<<<<<<<<<<
 *         for s in self.selectors:
 *             v += s._hash_vals()
 */
  __Pyx_INCREF(__pyx_tuple__66);
  __pyx_v_v = __pyx_tuple__66;

  /* "yt/geometry/selection_routines.pyx":2444
 *     def _hash_vals(self):
 *         v = ("chained_or",)
 *         for s in self.selectors:             # <<<<<<<<<<<<<<
 *             v += s._hash_vals()
 *         return v
 */
  if (likely(PyList_CheckExact(((PyObject *)__pyx_v_self->__pyx_base.selectors))) || PyTuple_CheckExact(((PyObject *)__pyx_v_self->__pyx_base.selectors))) {
    __pyx_t_1 = ((PyObject *)__pyx_v_self->__pyx_base.selectors); __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(((PyObject *)__pyx_v_self->__pyx_base.selectors)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2444, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 2444, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2444, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 2444, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2444, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2444, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "yt/geometry/selection_routines.pyx":2445
 *         v = ("chained_or",)
 *         for s in self.selectors:
 *             v += s._hash_vals()             # <<<<<<<<<<<<<<
 *         return v
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_s, __pyx_n_s_hash_vals); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_v_v, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_v, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "yt/geometry/selection_routines.pyx":2444
 *     def _hash_vals(self):
 *         v = ("chained_or",)
 *         for s in self.selectors:             # <<<<<<<<<<<<<<
 *             v += s._hash_vals()
 *         return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":2446
 *         for s in self.selectors:
 *             v += s._hash_vals()
 *         return v             # <<<<<<<<<<<<<<
 * 
 * union_selector = ChainedBooleanORSelector
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_v);
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pyx":2442
 *         return 0
 * 
 *     def _hash_vals(self):             # <<<<<<<<<<<<<<
 *         v = ("chained_or",)
 *         for s in self.selectors:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("yt.geometry.selection_routines.ChainedBooleanORSelector._hash_vals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_2__reduce_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__67, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.ChainedBooleanORSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_4__setstate_cython__(((struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__68, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.selection_routines.ChainedBooleanORSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyArray_Descr *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":265
 * 
 *             cdef int i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":266
 *             cdef int i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":268
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":271
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_C_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__69, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(3, 272, __pyx_L1_error)

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":275
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_F_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__70, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(3, 276, __pyx_L1_error)

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":278
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":279
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":283
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * 2) * ((size_t)__pyx_v_ndim))));

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":284
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":285
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_4 = __pyx_v_ndim;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":286
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":287
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
    goto __pyx_L9;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":289
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
  /*else*/ {
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L9:;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":291
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":292
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":293
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!(PyArray_ISWRITEABLE(__pyx_v_self) != 0));

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":296
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)
 *             cdef int offset
 */
  __pyx_v_f = NULL;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":297
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 *             cdef int offset
 * 
 */
  __pyx_t_7 = PyArray_DESCR(__pyx_v_self);
  __pyx_t_3 = ((PyObject *)__pyx_t_7);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":300
 *             cdef int offset
 * 
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not PyDataType_HASFIELDS(descr):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = ((!(PyDataType_HASFIELDS(__pyx_v_descr) != 0)) != 0);
  if (__pyx_t_1) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":303
 * 
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_4 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_4;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '>') != 0);
    if (!__pyx_t_2) {
      goto __pyx_L15_next_or;
    } else {
    }
    __pyx_t_2 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_L15_next_or:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":305
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '<') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_2 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L14_bool_binop_done:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_1)) {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__71, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(3, 306, __pyx_L1_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":307
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    switch (__pyx_v_t) {
      case NPY_BYTE:
      __pyx_v_f = ((char *)"b");
      break;
      case NPY_UBYTE:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":308
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
      __pyx_v_f = ((char *)"B");
      break;
      case NPY_SHORT:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":309
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
      __pyx_v_f = ((char *)"h");
      break;
      case NPY_USHORT:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":310
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
      __pyx_v_f = ((char *)"H");
      break;
      case NPY_INT:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":311
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
      __pyx_v_f = ((char *)"i");
      break;
      case NPY_UINT:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":312
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
      __pyx_v_f = ((char *)"I");
      break;
      case NPY_LONG:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":313
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
      __pyx_v_f = ((char *)"l");
      break;
      case NPY_ULONG:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":314
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
      __pyx_v_f = ((char *)"L");
      break;
      case NPY_LONGLONG:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":315
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
      __pyx_v_f = ((char *)"q");
      break;
      case NPY_ULONGLONG:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":316
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
      __pyx_v_f = ((char *)"Q");
      break;
      case NPY_FLOAT:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":317
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
      __pyx_v_f = ((char *)"f");
      break;
      case NPY_DOUBLE:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":318
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
      __pyx_v_f = ((char *)"d");
      break;
      case NPY_LONGDOUBLE:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":319
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
      __pyx_v_f = ((char *)"g");
      break;
      case NPY_CFLOAT:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":320
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
      __pyx_v_f = ((char *)"Zf");
      break;
      case NPY_CDOUBLE:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":321
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
      __pyx_v_f = ((char *)"Zd");
      break;
      case NPY_CLONGDOUBLE:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":322
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
      __pyx_v_f = ((char *)"Zg");
      break;
      case NPY_OBJECT:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":323
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_v_f = ((char *)"O");
      break;
      default:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":325
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(3, 325, __pyx_L1_error)
      break;
    }

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":326
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":327
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":329
 *                 return
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
  /*else*/ {
    __pyx_v_info->format = ((char *)PyObject_Malloc(0xFF));

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":330
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":331
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":332
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,             # <<<<<<<<<<<<<<
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 */
    __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 0xFF), (&__pyx_v_offset)); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(3, 332, __pyx_L1_error)
    __pyx_v_f = __pyx_t_9;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":335
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = (PyArray_HASFIELDS(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":339
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)
 */
    PyObject_Free(__pyx_v_info->format);

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":341
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    PyObject_Free(__pyx_v_info->strides);

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":820
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":820
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":823
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":824
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":823
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":826
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":827
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":826
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":829
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":830
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":829
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":832
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":833
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":832
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":835
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":836
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":836
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":839
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":835
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":841
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  char *__pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":846
 * 
 *     cdef dtype child
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":847
 *     cdef dtype child
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":850
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(__pyx_v_descr->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(3, 850, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_descr->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(3, 850, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 850, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_childname, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":851
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    if (unlikely(__pyx_v_descr->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(3, 851, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(3, 851, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fields, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":852
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 */
    if (likely(__pyx_v_fields != Py_None)) {
      PyObject* sequence = __pyx_v_fields;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(3, 852, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(3, 852, __pyx_L1_error)
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) __PYX_ERR(3, 852, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_child, ((PyArray_Descr *)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_offset, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":854
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 854, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 854, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(3, 854, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = ((((__pyx_v_end - __pyx_v_f) - ((int)__pyx_t_5)) < 15) != 0);
    if (unlikely(__pyx_t_6)) {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":855
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__72, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 855, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(3, 855, __pyx_L1_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":854
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    }

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":857
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '>') != 0);
    if (!__pyx_t_7) {
      goto __pyx_L8_next_or;
    } else {
    }
    __pyx_t_7 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_L8_next_or:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":858
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '<') != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":857
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_6)) {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":859
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__71, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 859, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(3, 859, __pyx_L1_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":857
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":869
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 869, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 869, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 869, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_6) break;

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":870
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 0x78;

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":871
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":872
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_8 = 0;
      (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + 1);
    }

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":874
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_8 = 0;
    (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + __pyx_v_child->elsize);

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":876
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_6 = ((!(PyDataType_HASFIELDS(__pyx_v_child) != 0)) != 0);
    if (__pyx_t_6) {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":877
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_child->type_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":878
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_6 = (((__pyx_v_end - __pyx_v_f) < 5) != 0);
      if (unlikely(__pyx_t_6)) {

        /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__73, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 879, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(3, 879, __pyx_L1_error)

        /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":878
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":882
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_BYTE); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 882, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 882, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 882, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":883
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UBYTE); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":884
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_SHORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x68;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":885
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_USHORT); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 885, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":886
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_INT); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x69;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":887
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UINT); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 887, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":888
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x6C;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":889
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":890
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGLONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x71;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":891
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONGLONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":892
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_FLOAT); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 892, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x66;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":893
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_DOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x64;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":894
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x67;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":895
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CFLOAT); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x66;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":896
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 896, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x64;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":897
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x67;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":898
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_OBJECT); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (likely(__pyx_t_6)) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":900
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
      /*else*/ {
        __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 900, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 900, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(3, 900, __pyx_L1_error)
      }
      __pyx_L15:;

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":901
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":876
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
      goto __pyx_L13;
    }

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":905
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
    /*else*/ {
      __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(3, 905, __pyx_L1_error)
      __pyx_v_f = __pyx_t_9;
    }
    __pyx_L13:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":850
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":906
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":841
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1021
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1023
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1021
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1025
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1027
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1027
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1029
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1025
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1033
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * cdef inline int import_array() except -1:
 *     try:
 *         _import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 1035, __pyx_L3_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1036
 *     try:
 *         _import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 1036, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1037
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__74, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 1037, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(3, 1037, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1033
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1039
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 1041, __pyx_L3_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1042
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 1042, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1043
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__75, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 1043, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(3, 1043, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1039
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1045
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 1047, __pyx_L3_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1048
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 1048, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1049
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__75, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 1049, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(3, 1049, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1045
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "oct_visitors.pxd":51
 *     cdef void visit(self, Oct*, np.uint8_t selected)
 * 
 *     cdef inline int oind(self):             # <<<<<<<<<<<<<<
 *         cdef int d = (1 << self.oref)
 *         return (((self.ind[0]*d)+self.ind[1])*d+self.ind[2])
 */

static CYTHON_INLINE int __pyx_f_2yt_8geometry_12oct_visitors_10OctVisitor_oind(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *__pyx_v_self) {
  int __pyx_v_d;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("oind", 0);

  /* "oct_visitors.pxd":52
 * 
 *     cdef inline int oind(self):
 *         cdef int d = (1 << self.oref)             # <<<<<<<<<<<<<<
 *         return (((self.ind[0]*d)+self.ind[1])*d+self.ind[2])
 * 
 */
  __pyx_v_d = (1 << __pyx_v_self->oref);

  /* "oct_visitors.pxd":53
 *     cdef inline int oind(self):
 *         cdef int d = (1 << self.oref)
 *         return (((self.ind[0]*d)+self.ind[1])*d+self.ind[2])             # <<<<<<<<<<<<<<
 * 
 *     cdef inline int rind(self):
 */
  __pyx_r = (((((__pyx_v_self->ind[0]) * __pyx_v_d) + (__pyx_v_self->ind[1])) * __pyx_v_d) + (__pyx_v_self->ind[2]));
  goto __pyx_L0;

  /* "oct_visitors.pxd":51
 *     cdef void visit(self, Oct*, np.uint8_t selected)
 * 
 *     cdef inline int oind(self):             # <<<<<<<<<<<<<<
 *         cdef int d = (1 << self.oref)
 *         return (((self.ind[0]*d)+self.ind[1])*d+self.ind[2])
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "oct_visitors.pxd":55
 *         return (((self.ind[0]*d)+self.ind[1])*d+self.ind[2])
 * 
 *     cdef inline int rind(self):             # <<<<<<<<<<<<<<
 *         cdef int d = (1 << self.oref)
 *         return (((self.ind[2]*d)+self.ind[1])*d+self.ind[0])
 */

static CYTHON_INLINE int __pyx_f_2yt_8geometry_12oct_visitors_10OctVisitor_rind(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *__pyx_v_self) {
  int __pyx_v_d;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("rind", 0);

  /* "oct_visitors.pxd":56
 * 
 *     cdef inline int rind(self):
 *         cdef int d = (1 << self.oref)             # <<<<<<<<<<<<<<
 *         return (((self.ind[2]*d)+self.ind[1])*d+self.ind[0])
 * 
 */
  __pyx_v_d = (1 << __pyx_v_self->oref);

  /* "oct_visitors.pxd":57
 *     cdef inline int rind(self):
 *         cdef int d = (1 << self.oref)
 *         return (((self.ind[2]*d)+self.ind[1])*d+self.ind[0])             # <<<<<<<<<<<<<<
 * 
 * cdef class CountTotalOcts(OctVisitor):
 */
  __pyx_r = (((((__pyx_v_self->ind[2]) * __pyx_v_d) + (__pyx_v_self->ind[1])) * __pyx_v_d) + (__pyx_v_self->ind[0]));
  goto __pyx_L0;

  /* "oct_visitors.pxd":55
 *         return (((self.ind[0]*d)+self.ind[1])*d+self.ind[2])
 * 
 *     cdef inline int rind(self):             # <<<<<<<<<<<<<<
 *         cdef int d = (1 << self.oref)
 *         return (((self.ind[2]*d)+self.ind[1])*d+self.ind[0])
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "oct_visitors.pxd":123
 *     cdef np.uint64_t *nfinest
 * 
 * cdef inline int cind(int i, int j, int k):             # <<<<<<<<<<<<<<
 *     # THIS ONLY WORKS FOR CHILDREN.  It is not general for zones.
 *     return (((i*2)+j)*2+k)
 */

static CYTHON_INLINE int __pyx_f_2yt_8geometry_12oct_visitors_cind(int __pyx_v_i, int __pyx_v_j, int __pyx_v_k) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cind", 0);

  /* "oct_visitors.pxd":125
 * cdef inline int cind(int i, int j, int k):
 *     # THIS ONLY WORKS FOR CHILDREN.  It is not general for zones.
 *     return (((i*2)+j)*2+k)             # <<<<<<<<<<<<<<
 * 
 */
  __pyx_r = ((((__pyx_v_i * 2) + __pyx_v_j) * 2) + __pyx_v_k);
  goto __pyx_L0;

  /* "oct_visitors.pxd":123
 *     cdef np.uint64_t *nfinest
 * 
 * cdef inline int cind(int i, int j, int k):             # <<<<<<<<<<<<<<
 *     # THIS ONLY WORKS FOR CHILDREN.  It is not general for zones.
 *     return (((i*2)+j)*2+k)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pxd":22
 *     grid_visitor_function, check_child_masked
 * 
 * cdef inline _ensure_code(arr):             # <<<<<<<<<<<<<<
 *     if hasattr(arr, "units"):
 *         if "code_length" == str(arr.units):
 */

static CYTHON_INLINE PyObject *__pyx_f_2yt_8geometry_18selection_routines__ensure_code(PyObject *__pyx_v_arr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ensure_code", 0);

  /* "yt/geometry/selection_routines.pxd":23
 * 
 * cdef inline _ensure_code(arr):
 *     if hasattr(arr, "units"):             # <<<<<<<<<<<<<<
 *         if "code_length" == str(arr.units):
 *             return arr
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_arr, __pyx_n_s_units); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 23, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "yt/geometry/selection_routines.pxd":24
 * cdef inline _ensure_code(arr):
 *     if hasattr(arr, "units"):
 *         if "code_length" == str(arr.units):             # <<<<<<<<<<<<<<
 *             return arr
 *         arr.convert_to_units("code_length")
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_arr, __pyx_n_s_units); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_n_s_code_length, __pyx_t_4, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_2) {

      /* "yt/geometry/selection_routines.pxd":25
 *     if hasattr(arr, "units"):
 *         if "code_length" == str(arr.units):
 *             return arr             # <<<<<<<<<<<<<<
 *         arr.convert_to_units("code_length")
 *     return arr
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_arr);
      __pyx_r = __pyx_v_arr;
      goto __pyx_L0;

      /* "yt/geometry/selection_routines.pxd":24
 * cdef inline _ensure_code(arr):
 *     if hasattr(arr, "units"):
 *         if "code_length" == str(arr.units):             # <<<<<<<<<<<<<<
 *             return arr
 *         arr.convert_to_units("code_length")
 */
    }

    /* "yt/geometry/selection_routines.pxd":26
 *         if "code_length" == str(arr.units):
 *             return arr
 *         arr.convert_to_units("code_length")             # <<<<<<<<<<<<<<
 *     return arr
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_arr, __pyx_n_s_convert_to_units); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 26, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_n_s_code_length) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_s_code_length);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "yt/geometry/selection_routines.pxd":23
 * 
 * cdef inline _ensure_code(arr):
 *     if hasattr(arr, "units"):             # <<<<<<<<<<<<<<
 *         if "code_length" == str(arr.units):
 *             return arr
 */
  }

  /* "yt/geometry/selection_routines.pxd":27
 *             return arr
 *         arr.convert_to_units("code_length")
 *     return arr             # <<<<<<<<<<<<<<
 * 
 * cdef class SelectorObject:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_arr);
  __pyx_r = __pyx_v_arr;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pxd":22
 *     grid_visitor_function, check_child_masked
 * 
 * cdef inline _ensure_code(arr):             # <<<<<<<<<<<<<<
 *     if hasattr(arr, "units"):
 *         if "code_length" == str(arr.units):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("yt.geometry.selection_routines._ensure_code", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/selection_routines.pxd":78
 *     cdef public SelectorObject sel2
 * 
 * cdef inline np.float64_t _periodic_dist(np.float64_t x1, np.float64_t x2,             # <<<<<<<<<<<<<<
 *                                         np.float64_t dw, bint periodic) nogil:
 *     cdef np.float64_t rel = x1 - x2
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_18selection_routines__periodic_dist(__pyx_t_5numpy_float64_t __pyx_v_x1, __pyx_t_5numpy_float64_t __pyx_v_x2, __pyx_t_5numpy_float64_t __pyx_v_dw, int __pyx_v_periodic) {
  __pyx_t_5numpy_float64_t __pyx_v_rel;
  __pyx_t_5numpy_float64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/geometry/selection_routines.pxd":80
 * cdef inline np.float64_t _periodic_dist(np.float64_t x1, np.float64_t x2,
 *                                         np.float64_t dw, bint periodic) nogil:
 *     cdef np.float64_t rel = x1 - x2             # <<<<<<<<<<<<<<
 *     if not periodic: return rel
 *     if rel > dw * 0.5:
 */
  __pyx_v_rel = (__pyx_v_x1 - __pyx_v_x2);

  /* "yt/geometry/selection_routines.pxd":81
 *                                         np.float64_t dw, bint periodic) nogil:
 *     cdef np.float64_t rel = x1 - x2
 *     if not periodic: return rel             # <<<<<<<<<<<<<<
 *     if rel > dw * 0.5:
 *         rel -= dw
 */
  __pyx_t_1 = ((!(__pyx_v_periodic != 0)) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_rel;
    goto __pyx_L0;
  }

  /* "yt/geometry/selection_routines.pxd":82
 *     cdef np.float64_t rel = x1 - x2
 *     if not periodic: return rel
 *     if rel > dw * 0.5:             # <<<<<<<<<<<<<<
 *         rel -= dw
 *     elif rel < -dw * 0.5:
 */
  __pyx_t_1 = ((__pyx_v_rel > (__pyx_v_dw * 0.5)) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pxd":83
 *     if not periodic: return rel
 *     if rel > dw * 0.5:
 *         rel -= dw             # <<<<<<<<<<<<<<
 *     elif rel < -dw * 0.5:
 *         rel += dw
 */
    __pyx_v_rel = (__pyx_v_rel - __pyx_v_dw);

    /* "yt/geometry/selection_routines.pxd":82
 *     cdef np.float64_t rel = x1 - x2
 *     if not periodic: return rel
 *     if rel > dw * 0.5:             # <<<<<<<<<<<<<<
 *         rel -= dw
 *     elif rel < -dw * 0.5:
 */
    goto __pyx_L4;
  }

  /* "yt/geometry/selection_routines.pxd":84
 *     if rel > dw * 0.5:
 *         rel -= dw
 *     elif rel < -dw * 0.5:             # <<<<<<<<<<<<<<
 *         rel += dw
 *     return rel
 */
  __pyx_t_1 = ((__pyx_v_rel < ((-__pyx_v_dw) * 0.5)) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/selection_routines.pxd":85
 *         rel -= dw
 *     elif rel < -dw * 0.5:
 *         rel += dw             # <<<<<<<<<<<<<<
 *     return rel
 */
    __pyx_v_rel = (__pyx_v_rel + __pyx_v_dw);

    /* "yt/geometry/selection_routines.pxd":84
 *     if rel > dw * 0.5:
 *         rel -= dw
 *     elif rel < -dw * 0.5:             # <<<<<<<<<<<<<<
 *         rel += dw
 *     return rel
 */
  }
  __pyx_L4:;

  /* "yt/geometry/selection_routines.pxd":86
 *     elif rel < -dw * 0.5:
 *         rel += dw
 *     return rel             # <<<<<<<<<<<<<<
 */
  __pyx_r = __pyx_v_rel;
  goto __pyx_L0;

  /* "yt/geometry/selection_routines.pxd":78
 *     cdef public SelectorObject sel2
 * 
 * cdef inline np.float64_t _periodic_dist(np.float64_t x1, np.float64_t x2,             # <<<<<<<<<<<<<<
 *                                         np.float64_t dw, bint periodic) nogil:
 *     cdef np.float64_t rel = x1 - x2
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":19
 * cimport cython
 * 
 * cdef inline np.int64_t imax(np.int64_t i0, np.int64_t i1) nogil:             # <<<<<<<<<<<<<<
 *     if i0 > i1: return i0
 *     return i1
 */

static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_imax(__pyx_t_5numpy_int64_t __pyx_v_i0, __pyx_t_5numpy_int64_t __pyx_v_i1) {
  __pyx_t_5numpy_int64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":20
 * 
 * cdef inline np.int64_t imax(np.int64_t i0, np.int64_t i1) nogil:
 *     if i0 > i1: return i0             # <<<<<<<<<<<<<<
 *     return i1
 * 
 */
  __pyx_t_1 = ((__pyx_v_i0 > __pyx_v_i1) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_i0;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":21
 * cdef inline np.int64_t imax(np.int64_t i0, np.int64_t i1) nogil:
 *     if i0 > i1: return i0
 *     return i1             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.float64_t fmax(np.float64_t f0, np.float64_t f1) nogil:
 */
  __pyx_r = __pyx_v_i1;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":19
 * cimport cython
 * 
 * cdef inline np.int64_t imax(np.int64_t i0, np.int64_t i1) nogil:             # <<<<<<<<<<<<<<
 *     if i0 > i1: return i0
 *     return i1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":23
 *     return i1
 * 
 * cdef inline np.float64_t fmax(np.float64_t f0, np.float64_t f1) nogil:             # <<<<<<<<<<<<<<
 *     if f0 > f1: return f0
 *     return f1
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_fmax(__pyx_t_5numpy_float64_t __pyx_v_f0, __pyx_t_5numpy_float64_t __pyx_v_f1) {
  __pyx_t_5numpy_float64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":24
 * 
 * cdef inline np.float64_t fmax(np.float64_t f0, np.float64_t f1) nogil:
 *     if f0 > f1: return f0             # <<<<<<<<<<<<<<
 *     return f1
 * 
 */
  __pyx_t_1 = ((__pyx_v_f0 > __pyx_v_f1) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_f0;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":25
 * cdef inline np.float64_t fmax(np.float64_t f0, np.float64_t f1) nogil:
 *     if f0 > f1: return f0
 *     return f1             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.int64_t imin(np.int64_t i0, np.int64_t i1) nogil:
 */
  __pyx_r = __pyx_v_f1;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":23
 *     return i1
 * 
 * cdef inline np.float64_t fmax(np.float64_t f0, np.float64_t f1) nogil:             # <<<<<<<<<<<<<<
 *     if f0 > f1: return f0
 *     return f1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":27
 *     return f1
 * 
 * cdef inline np.int64_t imin(np.int64_t i0, np.int64_t i1) nogil:             # <<<<<<<<<<<<<<
 *     if i0 < i1: return i0
 *     return i1
 */

static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_imin(__pyx_t_5numpy_int64_t __pyx_v_i0, __pyx_t_5numpy_int64_t __pyx_v_i1) {
  __pyx_t_5numpy_int64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":28
 * 
 * cdef inline np.int64_t imin(np.int64_t i0, np.int64_t i1) nogil:
 *     if i0 < i1: return i0             # <<<<<<<<<<<<<<
 *     return i1
 * 
 */
  __pyx_t_1 = ((__pyx_v_i0 < __pyx_v_i1) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_i0;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":29
 * cdef inline np.int64_t imin(np.int64_t i0, np.int64_t i1) nogil:
 *     if i0 < i1: return i0
 *     return i1             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.float64_t fmin(np.float64_t f0, np.float64_t f1) nogil:
 */
  __pyx_r = __pyx_v_i1;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":27
 *     return f1
 * 
 * cdef inline np.int64_t imin(np.int64_t i0, np.int64_t i1) nogil:             # <<<<<<<<<<<<<<
 *     if i0 < i1: return i0
 *     return i1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":31
 *     return i1
 * 
 * cdef inline np.float64_t fmin(np.float64_t f0, np.float64_t f1) nogil:             # <<<<<<<<<<<<<<
 *     if f0 < f1: return f0
 *     return f1
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_fmin(__pyx_t_5numpy_float64_t __pyx_v_f0, __pyx_t_5numpy_float64_t __pyx_v_f1) {
  __pyx_t_5numpy_float64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":32
 * 
 * cdef inline np.float64_t fmin(np.float64_t f0, np.float64_t f1) nogil:
 *     if f0 < f1: return f0             # <<<<<<<<<<<<<<
 *     return f1
 * 
 */
  __pyx_t_1 = ((__pyx_v_f0 < __pyx_v_f1) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_f0;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":33
 * cdef inline np.float64_t fmin(np.float64_t f0, np.float64_t f1) nogil:
 *     if f0 < f1: return f0
 *     return f1             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.float64_t fabs(np.float64_t f0) nogil:
 */
  __pyx_r = __pyx_v_f1;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":31
 *     return i1
 * 
 * cdef inline np.float64_t fmin(np.float64_t f0, np.float64_t f1) nogil:             # <<<<<<<<<<<<<<
 *     if f0 < f1: return f0
 *     return f1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":35
 *     return f1
 * 
 * cdef inline np.float64_t fabs(np.float64_t f0) nogil:             # <<<<<<<<<<<<<<
 *     if f0 < 0.0: return -f0
 *     return f0
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_fabs(__pyx_t_5numpy_float64_t __pyx_v_f0) {
  __pyx_t_5numpy_float64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":36
 * 
 * cdef inline np.float64_t fabs(np.float64_t f0) nogil:
 *     if f0 < 0.0: return -f0             # <<<<<<<<<<<<<<
 *     return f0
 * 
 */
  __pyx_t_1 = ((__pyx_v_f0 < 0.0) != 0);
  if (__pyx_t_1) {
    __pyx_r = (-__pyx_v_f0);
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":37
 * cdef inline np.float64_t fabs(np.float64_t f0) nogil:
 *     if f0 < 0.0: return -f0
 *     return f0             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.int64_t iclip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:
 */
  __pyx_r = __pyx_v_f0;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":35
 *     return f1
 * 
 * cdef inline np.float64_t fabs(np.float64_t f0) nogil:             # <<<<<<<<<<<<<<
 *     if f0 < 0.0: return -f0
 *     return f0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":39
 *     return f0
 * 
 * cdef inline np.int64_t iclip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:             # <<<<<<<<<<<<<<
 *     if i < a: return a
 *     if i > b: return b
 */

static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_iclip(__pyx_t_5numpy_int64_t __pyx_v_i, __pyx_t_5numpy_int64_t __pyx_v_a, __pyx_t_5numpy_int64_t __pyx_v_b) {
  __pyx_t_5numpy_int64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":40
 * 
 * cdef inline np.int64_t iclip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:
 *     if i < a: return a             # <<<<<<<<<<<<<<
 *     if i > b: return b
 *     return i
 */
  __pyx_t_1 = ((__pyx_v_i < __pyx_v_a) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_a;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":41
 * cdef inline np.int64_t iclip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:
 *     if i < a: return a
 *     if i > b: return b             # <<<<<<<<<<<<<<
 *     return i
 * 
 */
  __pyx_t_1 = ((__pyx_v_i > __pyx_v_b) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_b;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":42
 *     if i < a: return a
 *     if i > b: return b
 *     return i             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.int64_t i64clip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:
 */
  __pyx_r = __pyx_v_i;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":39
 *     return f0
 * 
 * cdef inline np.int64_t iclip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:             # <<<<<<<<<<<<<<
 *     if i < a: return a
 *     if i > b: return b
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":44
 *     return i
 * 
 * cdef inline np.int64_t i64clip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:             # <<<<<<<<<<<<<<
 *     if i < a: return a
 *     if i > b: return b
 */

static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_i64clip(__pyx_t_5numpy_int64_t __pyx_v_i, __pyx_t_5numpy_int64_t __pyx_v_a, __pyx_t_5numpy_int64_t __pyx_v_b) {
  __pyx_t_5numpy_int64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":45
 * 
 * cdef inline np.int64_t i64clip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:
 *     if i < a: return a             # <<<<<<<<<<<<<<
 *     if i > b: return b
 *     return i
 */
  __pyx_t_1 = ((__pyx_v_i < __pyx_v_a) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_a;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":46
 * cdef inline np.int64_t i64clip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:
 *     if i < a: return a
 *     if i > b: return b             # <<<<<<<<<<<<<<
 *     return i
 * 
 */
  __pyx_t_1 = ((__pyx_v_i > __pyx_v_b) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_b;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":47
 *     if i < a: return a
 *     if i > b: return b
 *     return i             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.float64_t fclip(np.float64_t f,
 */
  __pyx_r = __pyx_v_i;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":44
 *     return i
 * 
 * cdef inline np.int64_t i64clip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:             # <<<<<<<<<<<<<<
 *     if i < a: return a
 *     if i > b: return b
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":49
 *     return i
 * 
 * cdef inline np.float64_t fclip(np.float64_t f,             # <<<<<<<<<<<<<<
 *                       np.float64_t a, np.float64_t b) nogil:
 *     return fmin(fmax(f, a), b)
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_fclip(__pyx_t_5numpy_float64_t __pyx_v_f, __pyx_t_5numpy_float64_t __pyx_v_a, __pyx_t_5numpy_float64_t __pyx_v_b) {
  __pyx_t_5numpy_float64_t __pyx_r;

  /* "yt/utilities/lib/fp_utils.pxd":51
 * cdef inline np.float64_t fclip(np.float64_t f,
 *                       np.float64_t a, np.float64_t b) nogil:
 *     return fmin(fmax(f, a), b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.int64_t i64max(np.int64_t i0, np.int64_t i1) nogil:
 */
  __pyx_r = __pyx_f_2yt_9utilities_3lib_8fp_utils_fmin(__pyx_f_2yt_9utilities_3lib_8fp_utils_fmax(__pyx_v_f, __pyx_v_a), __pyx_v_b);
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":49
 *     return i
 * 
 * cdef inline np.float64_t fclip(np.float64_t f,             # <<<<<<<<<<<<<<
 *                       np.float64_t a, np.float64_t b) nogil:
 *     return fmin(fmax(f, a), b)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":53
 *     return fmin(fmax(f, a), b)
 * 
 * cdef inline np.int64_t i64max(np.int64_t i0, np.int64_t i1) nogil:             # <<<<<<<<<<<<<<
 *     if i0 > i1: return i0
 *     return i1
 */

static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_i64max(__pyx_t_5numpy_int64_t __pyx_v_i0, __pyx_t_5numpy_int64_t __pyx_v_i1) {
  __pyx_t_5numpy_int64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":54
 * 
 * cdef inline np.int64_t i64max(np.int64_t i0, np.int64_t i1) nogil:
 *     if i0 > i1: return i0             # <<<<<<<<<<<<<<
 *     return i1
 * 
 */
  __pyx_t_1 = ((__pyx_v_i0 > __pyx_v_i1) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_i0;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":55
 * cdef inline np.int64_t i64max(np.int64_t i0, np.int64_t i1) nogil:
 *     if i0 > i1: return i0
 *     return i1             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.int64_t i64min(np.int64_t i0, np.int64_t i1) nogil:
 */
  __pyx_r = __pyx_v_i1;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":53
 *     return fmin(fmax(f, a), b)
 * 
 * cdef inline np.int64_t i64max(np.int64_t i0, np.int64_t i1) nogil:             # <<<<<<<<<<<<<<
 *     if i0 > i1: return i0
 *     return i1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":57
 *     return i1
 * 
 * cdef inline np.int64_t i64min(np.int64_t i0, np.int64_t i1) nogil:             # <<<<<<<<<<<<<<
 *     if i0 < i1: return i0
 *     return i1
 */

static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_i64min(__pyx_t_5numpy_int64_t __pyx_v_i0, __pyx_t_5numpy_int64_t __pyx_v_i1) {
  __pyx_t_5numpy_int64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":58
 * 
 * cdef inline np.int64_t i64min(np.int64_t i0, np.int64_t i1) nogil:
 *     if i0 < i1: return i0             # <<<<<<<<<<<<<<
 *     return i1
 * 
 */
  __pyx_t_1 = ((__pyx_v_i0 < __pyx_v_i1) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_i0;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":59
 * cdef inline np.int64_t i64min(np.int64_t i0, np.int64_t i1) nogil:
 *     if i0 < i1: return i0
 *     return i1             # <<<<<<<<<<<<<<
 * 
 */
  __pyx_r = __pyx_v_i1;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":57
 *     return i1
 * 
 * cdef inline np.int64_t i64min(np.int64_t i0, np.int64_t i1) nogil:             # <<<<<<<<<<<<<<
 *     if i0 < i1: return i0
 *     return i1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "oct_container.pxd":56
 * 
 * cdef class OctObjectPool(ObjectPool):
 *     cdef inline OctAllocationContainer *get_cont(self, int i):             # <<<<<<<<<<<<<<
 *         return <OctAllocationContainer*> (&self.containers[i])
 * 
 */

static CYTHON_INLINE struct __pyx_t_2yt_8geometry_13oct_container_OctAllocationContainer *__pyx_f_2yt_8geometry_13oct_container_13OctObjectPool_get_cont(struct __pyx_obj_2yt_8geometry_13oct_container_OctObjectPool *__pyx_v_self, int __pyx_v_i) {
  struct __pyx_t_2yt_8geometry_13oct_container_OctAllocationContainer *__pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_cont", 0);

  /* "oct_container.pxd":57
 * cdef class OctObjectPool(ObjectPool):
 *     cdef inline OctAllocationContainer *get_cont(self, int i):
 *         return <OctAllocationContainer*> (&self.containers[i])             # <<<<<<<<<<<<<<
 * 
 * cdef OctList *OctList_append(OctList *list, Oct *o)
 */
  __pyx_r = ((struct __pyx_t_2yt_8geometry_13oct_container_OctAllocationContainer *)(&(__pyx_v_self->__pyx_base.containers[__pyx_v_i])));
  goto __pyx_L0;

  /* "oct_container.pxd":56
 * 
 * cdef class OctObjectPool(ObjectPool):
 *     cdef inline OctAllocationContainer *get_cont(self, int i):             # <<<<<<<<<<<<<<
 *         return <OctAllocationContainer*> (&self.containers[i])
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/volume_container.pxd":30
 *     int dims[3]
 * 
 * cdef inline int vc_index(VolumeContainer *vc, int i, int j, int k):             # <<<<<<<<<<<<<<
 *     return (i*vc.dims[1]+j)*vc.dims[2]+k
 * 
 */

static CYTHON_INLINE int __pyx_f_2yt_9utilities_3lib_16volume_container_vc_index(struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer *__pyx_v_vc, int __pyx_v_i, int __pyx_v_j, int __pyx_v_k) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("vc_index", 0);

  /* "yt/utilities/lib/volume_container.pxd":31
 * 
 * cdef inline int vc_index(VolumeContainer *vc, int i, int j, int k):
 *     return (i*vc.dims[1]+j)*vc.dims[2]+k             # <<<<<<<<<<<<<<
 * 
 * cdef inline int vc_pos_index(VolumeContainer *vc, np.float64_t *spos):
 */
  __pyx_r = ((((__pyx_v_i * (__pyx_v_vc->dims[1])) + __pyx_v_j) * (__pyx_v_vc->dims[2])) + __pyx_v_k);
  goto __pyx_L0;

  /* "yt/utilities/lib/volume_container.pxd":30
 *     int dims[3]
 * 
 * cdef inline int vc_index(VolumeContainer *vc, int i, int j, int k):             # <<<<<<<<<<<<<<
 *     return (i*vc.dims[1]+j)*vc.dims[2]+k
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/volume_container.pxd":33
 *     return (i*vc.dims[1]+j)*vc.dims[2]+k
 * 
 * cdef inline int vc_pos_index(VolumeContainer *vc, np.float64_t *spos):             # <<<<<<<<<<<<<<
 *     cdef int index[3]
 *     cdef int i
 */

static CYTHON_INLINE int __pyx_f_2yt_9utilities_3lib_16volume_container_vc_pos_index(struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer *__pyx_v_vc, __pyx_t_5numpy_float64_t *__pyx_v_spos) {
  int __pyx_v_index[3];
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("vc_pos_index", 0);

  /* "yt/utilities/lib/volume_container.pxd":36
 *     cdef int index[3]
 *     cdef int i
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         index[i] = <int> ((spos[i] - vc.left_edge[i]) * vc.idds[i])
 *     return vc_index(vc, index[0], index[1], index[2])
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/volume_container.pxd":37
 *     cdef int i
 *     for i in range(3):
 *         index[i] = <int> ((spos[i] - vc.left_edge[i]) * vc.idds[i])             # <<<<<<<<<<<<<<
 *     return vc_index(vc, index[0], index[1], index[2])
 * 
 */
    (__pyx_v_index[__pyx_v_i]) = ((int)(((__pyx_v_spos[__pyx_v_i]) - (__pyx_v_vc->left_edge[__pyx_v_i])) * (__pyx_v_vc->idds[__pyx_v_i])));
  }

  /* "yt/utilities/lib/volume_container.pxd":38
 *     for i in range(3):
 *         index[i] = <int> ((spos[i] - vc.left_edge[i]) * vc.idds[i])
 *     return vc_index(vc, index[0], index[1], index[2])             # <<<<<<<<<<<<<<
 * 
 */
  __pyx_r = __pyx_f_2yt_9utilities_3lib_16volume_container_vc_index(__pyx_v_vc, (__pyx_v_index[0]), (__pyx_v_index[1]), (__pyx_v_index[2]));
  goto __pyx_L0;

  /* "yt/utilities/lib/volume_container.pxd":33
 *     return (i*vc.dims[1]+j)*vc.dims[2]+k
 * 
 * cdef inline int vc_pos_index(VolumeContainer *vc, np.float64_t *spos):             # <<<<<<<<<<<<<<
 *     cdef int index[3]
 *     cdef int i
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/bitarray.pxd":20
 * cimport cython
 * 
 * cdef inline void ba_set_value(np.uint8_t *buf, np.uint64_t ind,             # <<<<<<<<<<<<<<
 *                               np.uint8_t val) nogil:
 *     if val > 0: val = 1
 */

static CYTHON_INLINE void __pyx_f_2yt_9utilities_3lib_8bitarray_ba_set_value(__pyx_t_5numpy_uint8_t *__pyx_v_buf, __pyx_t_5numpy_uint64_t __pyx_v_ind, __pyx_t_5numpy_uint8_t __pyx_v_val) {
  int __pyx_t_1;
  __pyx_t_5numpy_uint64_t __pyx_t_2;

  /* "yt/utilities/lib/bitarray.pxd":22
 * cdef inline void ba_set_value(np.uint8_t *buf, np.uint64_t ind,
 *                               np.uint8_t val) nogil:
 *     if val > 0: val = 1             # <<<<<<<<<<<<<<
 *     buf[ind >> 3] |= (val << (ind & 7))
 * 
 */
  __pyx_t_1 = ((__pyx_v_val > 0) != 0);
  if (__pyx_t_1) {
    __pyx_v_val = 1;
  }

  /* "yt/utilities/lib/bitarray.pxd":23
 *                               np.uint8_t val) nogil:
 *     if val > 0: val = 1
 *     buf[ind >> 3] |= (val << (ind & 7))             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.uint8_t ba_get_value(np.uint8_t *buf, np.uint64_t ind) nogil:
 */
  __pyx_t_2 = (__pyx_v_ind >> 3);
  (__pyx_v_buf[__pyx_t_2]) = ((__pyx_v_buf[__pyx_t_2]) | (__pyx_v_val << (__pyx_v_ind & 7)));

  /* "yt/utilities/lib/bitarray.pxd":20
 * cimport cython
 * 
 * cdef inline void ba_set_value(np.uint8_t *buf, np.uint64_t ind,             # <<<<<<<<<<<<<<
 *                               np.uint8_t val) nogil:
 *     if val > 0: val = 1
 */

  /* function exit code */
}

/* "yt/utilities/lib/bitarray.pxd":25
 *     buf[ind >> 3] |= (val << (ind & 7))
 * 
 * cdef inline np.uint8_t ba_get_value(np.uint8_t *buf, np.uint64_t ind) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t rv = (buf[ind >> 3] & (1 << (ind & 7)))
 *     if rv == 0: return 0
 */

static CYTHON_INLINE __pyx_t_5numpy_uint8_t __pyx_f_2yt_9utilities_3lib_8bitarray_ba_get_value(__pyx_t_5numpy_uint8_t *__pyx_v_buf, __pyx_t_5numpy_uint64_t __pyx_v_ind) {
  __pyx_t_5numpy_uint8_t __pyx_v_rv;
  __pyx_t_5numpy_uint8_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/bitarray.pxd":26
 * 
 * cdef inline np.uint8_t ba_get_value(np.uint8_t *buf, np.uint64_t ind) nogil:
 *     cdef np.uint8_t rv = (buf[ind >> 3] & (1 << (ind & 7)))             # <<<<<<<<<<<<<<
 *     if rv == 0: return 0
 *     return 1
 */
  __pyx_v_rv = ((__pyx_v_buf[(__pyx_v_ind >> 3)]) & (1 << (__pyx_v_ind & 7)));

  /* "yt/utilities/lib/bitarray.pxd":27
 * cdef inline np.uint8_t ba_get_value(np.uint8_t *buf, np.uint64_t ind) nogil:
 *     cdef np.uint8_t rv = (buf[ind >> 3] & (1 << (ind & 7)))
 *     if rv == 0: return 0             # <<<<<<<<<<<<<<
 *     return 1
 * 
 */
  __pyx_t_1 = ((__pyx_v_rv == 0) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/bitarray.pxd":28
 *     cdef np.uint8_t rv = (buf[ind >> 3] & (1 << (ind & 7)))
 *     if rv == 0: return 0
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * cdef class bitarray:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "yt/utilities/lib/bitarray.pxd":25
 *     buf[ind >> 3] |= (val << (ind & 7))
 * 
 * cdef inline np.uint8_t ba_get_value(np.uint8_t *buf, np.uint64_t ind) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t rv = (buf[ind >> 3] & (1 << (ind & 7)))
 *     if rv == 0: return 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

/* Python wrapper */
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shape = 0;
  Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_format = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_v_allocate_buffer;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_itemsize,&__pyx_n_s_format,&__pyx_n_s_mode,&__pyx_n_s_allocate_buffer,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject *)__pyx_n_s_c);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_itemsize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 1); __PYX_ERR(2, 122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 2); __PYX_ERR(2, 122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_allocate_buffer);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(2, 122, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_shape = ((PyObject*)values[0]);
    __pyx_v_itemsize = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_itemsize == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 122, __pyx_L3_error)
    __pyx_v_format = values[2];
    __pyx_v_mode = values[3];
    if (values[4]) {
      __pyx_v_allocate_buffer = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_allocate_buffer == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 123, __pyx_L3_error)
    } else {

      /* "View.MemoryView":123
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,
 *                   mode="c", bint allocate_buffer=True):             # <<<<<<<<<<<<<<
 * 
 *         cdef int idx
 */
      __pyx_v_allocate_buffer = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 122, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shape), (&PyTuple_Type), 1, "shape", 1))) __PYX_ERR(2, 122, __pyx_L1_error)
  if (unlikely(((PyObject *)__pyx_v_format) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "format"); __PYX_ERR(2, 122, __pyx_L1_error)
  }
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v_shape, __pyx_v_itemsize, __pyx_v_format, __pyx_v_mode, __pyx_v_allocate_buffer);

  /* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_dim;
  PyObject **__pyx_v_p;
  char __pyx_v_order;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_format);

  /* "View.MemoryView":129
 *         cdef PyObject **p
 * 
 *         self.ndim = <int> len(shape)             # <<<<<<<<<<<<<<
 *         self.itemsize = itemsize
 * 
 */
  if (unlikely(__pyx_v_shape == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 129, __pyx_L1_error)
  }
  __pyx_t_1 = PyTuple_GET_SIZE(__pyx_v_shape); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(2, 129, __pyx_L1_error)
  __pyx_v_self->ndim = ((int)__pyx_t_1);

  /* "View.MemoryView":130
 * 
 *         self.ndim = <int> len(shape)
 *         self.itemsize = itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not self.ndim:
 */
  __pyx_v_self->itemsize = __pyx_v_itemsize;

  /* "View.MemoryView":132
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_self->ndim != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":133
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__76, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 133, __pyx_L1_error)

    /* "View.MemoryView":132
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  }

  /* "View.MemoryView":135
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  __pyx_t_2 = ((__pyx_v_itemsize <= 0) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":136
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__77, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 136, __pyx_L1_error)

    /* "View.MemoryView":135
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  }

  /* "View.MemoryView":138
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_format); 
  __pyx_t_4 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":139
 * 
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')             # <<<<<<<<<<<<<<
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_n_s_ASCII) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_s_ASCII);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":138
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  }

  /* "View.MemoryView":140
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string             # <<<<<<<<<<<<<<
 *         self.format = self._format
 * 
 */
  if (!(likely(PyBytes_CheckExact(__pyx_v_format))||((__pyx_v_format) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_format)->tp_name), 0))) __PYX_ERR(2, 140, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_format;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_format);
  __Pyx_DECREF(__pyx_v_self->_format);
  __pyx_v_self->_format = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":141
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_self->_format == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(2, 141, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->_format); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(2, 141, __pyx_L1_error)
  __pyx_v_self->format = __pyx_t_7;

  /* "View.MemoryView":144
 * 
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)             # <<<<<<<<<<<<<<
 *         self._strides = self._shape + self.ndim
 * 
 */
  __pyx_v_self->_shape = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * __pyx_v_self->ndim) * 2)));

  /* "View.MemoryView":145
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)
 *         self._strides = self._shape + self.ndim             # <<<<<<<<<<<<<<
 * 
 *         if not self._shape:
 */
  __pyx_v_self->_strides = (__pyx_v_self->_shape + __pyx_v_self->ndim);

  /* "View.MemoryView":147
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  __pyx_t_4 = ((!(__pyx_v_self->_shape != 0)) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "View.MemoryView":148
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__78, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 148, __pyx_L1_error)

    /* "View.MemoryView":147
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  }

  /* "View.MemoryView":151
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  __pyx_t_8 = 0;
  __pyx_t_3 = __pyx_v_shape; __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;
  for (;;) {
    if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(2, 151, __pyx_L1_error)
    #else
    __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 151, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_9;
    __pyx_v_idx = __pyx_t_8;
    __pyx_t_8 = (__pyx_t_8 + 1);

    /* "View.MemoryView":152
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    __pyx_t_4 = ((__pyx_v_dim <= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":153
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))             # <<<<<<<<<<<<<<
 *             self._shape[idx] = dim
 * 
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_6);
      __pyx_t_5 = 0;
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(2, 153, __pyx_L1_error)

      /* "View.MemoryView":152
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    }

    /* "View.MemoryView":154
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim             # <<<<<<<<<<<<<<
 * 
 *         cdef char order
 */
    (__pyx_v_self->_shape[__pyx_v_idx]) = __pyx_v_dim;

    /* "View.MemoryView":151
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":157
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_fortran, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(2, 157, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "View.MemoryView":158
 *         cdef char order
 *         if mode == 'fortran':
 *             order = b'F'             # <<<<<<<<<<<<<<
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 */
    __pyx_v_order = 'F';

    /* "View.MemoryView":159
 *         if mode == 'fortran':
 *             order = b'F'
 *             self.mode = u'fortran'             # <<<<<<<<<<<<<<
 *         elif mode == 'c':
 *             order = b'C'
 */
    __Pyx_INCREF(__pyx_n_u_fortran);
    __Pyx_GIVEREF(__pyx_n_u_fortran);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_fortran;

    /* "View.MemoryView":157
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":160
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_c, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(2, 160, __pyx_L1_error)
  if (likely(__pyx_t_4)) {

    /* "View.MemoryView":161
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 *             order = b'C'             # <<<<<<<<<<<<<<
 *             self.mode = u'c'
 *         else:
 */
    __pyx_v_order = 'C';

    /* "View.MemoryView":162
 *         elif mode == 'c':
 *             order = b'C'
 *             self.mode = u'c'             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 */
    __Pyx_INCREF(__pyx_n_u_c);
    __Pyx_GIVEREF(__pyx_n_u_c);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_c;

    /* "View.MemoryView":160
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":164
 *             self.mode = u'c'
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)             # <<<<<<<<<<<<<<
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_v_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_10, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_ERR(2, 164, __pyx_L1_error)
  }
  __pyx_L10:;

  /* "View.MemoryView":166
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,             # <<<<<<<<<<<<<<
 *                                              itemsize, self.ndim, order)
 * 
 */
  __pyx_v_self->len = __pyx_fill_contig_strides_array(__pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_itemsize, __pyx_v_self->ndim, __pyx_v_order);

  /* "View.MemoryView":169
 *                                              itemsize, self.ndim, order)
 * 
 *         self.free_data = allocate_buffer             # <<<<<<<<<<<<<<
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:
 */
  __pyx_v_self->free_data = __pyx_v_allocate_buffer;

  /* "View.MemoryView":170
 * 
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'             # <<<<<<<<<<<<<<
 *         if allocate_buffer:
 * 
 */
  __pyx_t_10 = PyObject_RichCompare(__pyx_v_format, __pyx_n_b_O, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 170, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_self->dtype_is_object = __pyx_t_4;

  /* "View.MemoryView":171
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = (__pyx_v_allocate_buffer != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":174
 * 
 * 
 *             self.data = <char *>malloc(self.len)             # <<<<<<<<<<<<<<
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")
 */
    __pyx_v_self->data = ((char *)malloc(__pyx_v_self->len));

    /* "View.MemoryView":175
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    __pyx_t_4 = ((!(__pyx_v_self->data != 0)) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":176
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__79, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(2, 176, __pyx_L1_error)

      /* "View.MemoryView":175
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    }

    /* "View.MemoryView":178
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    __pyx_t_4 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":179
 * 
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data             # <<<<<<<<<<<<<<
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 */
      __pyx_v_p = ((PyObject **)__pyx_v_self->data);

      /* "View.MemoryView":180
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):             # <<<<<<<<<<<<<<
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)
 */
      if (unlikely(__pyx_v_itemsize == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
        __PYX_ERR(2, 180, __pyx_L1_error)
      }
      else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_self->len))) {
        PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
        __PYX_ERR(2, 180, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_div_Py_ssize_t(__pyx_v_self->len, __pyx_v_itemsize);
      __pyx_t_9 = __pyx_t_1;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_9; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "View.MemoryView":181
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None             # <<<<<<<<<<<<<<
 *                     Py_INCREF(Py_None)
 * 
 */
        (__pyx_v_p[__pyx_v_i]) = Py_None;

        /* "View.MemoryView":182
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
        Py_INCREF(Py_None);
      }

      /* "View.MemoryView":178
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    }

    /* "View.MemoryView":171
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_format);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":185
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(((struct __pyx_array_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_bufmode;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":186
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1             # <<<<<<<<<<<<<<
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
  __pyx_v_bufmode = -1;

  /* "View.MemoryView":187
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_c, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 187, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":188
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
    __pyx_v_bufmode = (PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":187
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":189
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_fortran, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 189, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":190
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 */
    __pyx_v_bufmode = (PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":189
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":191
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  __pyx_t_1 = ((!((__pyx_v_flags & __pyx_v_bufmode) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":192
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__80, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 192, __pyx_L1_error)

    /* "View.MemoryView":191
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  }

  /* "View.MemoryView":193
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data             # <<<<<<<<<<<<<<
 *         info.len = self.len
 *         info.ndim = self.ndim
 */
  __pyx_t_4 = __pyx_v_self->data;
  __pyx_v_info->buf = __pyx_t_4;

  /* "View.MemoryView":194
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 *         info.len = self.len             # <<<<<<<<<<<<<<
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 */
  __pyx_t_5 = __pyx_v_self->len;
  __pyx_v_info->len = __pyx_t_5;

  /* "View.MemoryView":195
 *         info.buf = self.data
 *         info.len = self.len
 *         info.ndim = self.ndim             # <<<<<<<<<<<<<<
 *         info.shape = self._shape
 *         info.strides = self._strides
 */
  __pyx_t_6 = __pyx_v_self->ndim;
  __pyx_v_info->ndim = __pyx_t_6;

  /* "View.MemoryView":196
 *         info.len = self.len
 *         info.ndim = self.ndim
 *         info.shape = self._shape             # <<<<<<<<<<<<<<
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 */
  __pyx_t_7 = __pyx_v_self->_shape;
  __pyx_v_info->shape = __pyx_t_7;

  /* "View.MemoryView":197
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 *         info.strides = self._strides             # <<<<<<<<<<<<<<
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 */
  __pyx_t_7 = __pyx_v_self->_strides;
  __pyx_v_info->strides = __pyx_t_7;

  /* "View.MemoryView":198
 *         info.shape = self._shape
 *         info.strides = self._strides
 *         info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 */
  __pyx_v_info->suboffsets = NULL;

  /* "View.MemoryView":199
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize             # <<<<<<<<<<<<<<
 *         info.readonly = 0
 * 
 */
  __pyx_t_5 = __pyx_v_self->itemsize;
  __pyx_v_info->itemsize = __pyx_t_5;

  /* "View.MemoryView":200
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 *         info.readonly = 0             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  __pyx_v_info->readonly = 0;

  /* "View.MemoryView":202
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":203
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_4 = __pyx_v_self->format;
    __pyx_v_info->format = __pyx_t_4;

    /* "View.MemoryView":202
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
    goto __pyx_L5;
  }

  /* "View.MemoryView":205
 *             info.format = self.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.obj = self
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L5:;

  /* "View.MemoryView":207
 *             info.format = NULL
 * 
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":185
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":211
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

/* Python wrapper */
static void __pyx_array___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_array___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":212
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
  __pyx_t_1 = ((__pyx_v_self->callback_free_data != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":213
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)             # <<<<<<<<<<<<<<
 *         elif self.free_data:
 *             if self.dtype_is_object:
 */
    __pyx_v_self->callback_free_data(__pyx_v_self->data);

    /* "View.MemoryView":212
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":214
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  __pyx_t_1 = (__pyx_v_self->free_data != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":215
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    __pyx_t_1 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":216
 *         elif self.free_data:
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,             # <<<<<<<<<<<<<<
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_self->data, __pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_self->ndim, 0);

      /* "View.MemoryView":215
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    }

    /* "View.MemoryView":218
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 *             free(self.data)             # <<<<<<<<<<<<<<
 *         PyObject_Free(self._shape)
 * 
 */
    free(__pyx_v_self->data);

    /* "View.MemoryView":214
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":219
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 *         PyObject_Free(self._shape)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  PyObject_Free(__pyx_v_self->_shape);

  /* "View.MemoryView":211
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":222
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_5array_7memview___get__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":223
 *     @property
 *     def memview(self):
 *         return self.get_memview()             # <<<<<<<<<<<<<<
 * 
 *     @cname('get_memview')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_array *)__pyx_v_self->__pyx_vtab)->get_memview(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":222
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.memview.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":226
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_memview", 0);

  /* "View.MemoryView":227
 *     @cname('get_memview')
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE             # <<<<<<<<<<<<<<
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 */
  __pyx_v_flags = ((PyBUF_ANY_CONTIGUOUS | PyBUF_FORMAT) | PyBUF_WRITABLE);

  /* "View.MemoryView":228
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":226
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.get_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":230
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":231
 * 
 *     def __len__(self):
 *         return self._shape[0]             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, attr):
 */
  __pyx_r = (__pyx_v_self->_shape[0]);
  goto __pyx_L0;

  /* "View.MemoryView":230
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":233
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr); /*proto*/
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_attr));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattr__", 0);

  /* "View.MemoryView":234
 * 
 *     def __getattr__(self, attr):
 *         return getattr(self.memview, attr)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, item):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":233
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":236
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":237
 * 
 *     def __getitem__(self, item):
 *         return self.memview[item]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, item, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_item); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":236
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":239
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

/* Python wrapper */
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /*proto*/
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "View.MemoryView":240
 * 
 *     def __setitem__(self, item, value):
 *         self.memview[item] = value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_item, __pyx_v_value) < 0)) __PYX_ERR(2, 240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":239
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array___reduce_cython__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__81, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array_2__setstate_cython__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__82, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":244
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

static struct __pyx_array_obj *__pyx_array_new(PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, char *__pyx_v_format, char *__pyx_v_mode, char *__pyx_v_buf) {
  struct __pyx_array_obj *__pyx_v_result = 0;
  struct __pyx_array_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("array_cwrapper", 0);

  /* "View.MemoryView":248
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_buf == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":249
 * 
 *     if buf == NULL:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))             # <<<<<<<<<<<<<<
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":248
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":251
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
  /*else*/ {
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_3 = 0;

    /* "View.MemoryView":252
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)             # <<<<<<<<<<<<<<
 *         result.data = buf
 * 
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_allocate_buffer, Py_False) < 0) __PYX_ERR(2, 252, __pyx_L1_error)

    /* "View.MemoryView":251
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "View.MemoryView":253
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)
 *         result.data = buf             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
    __pyx_v_result->data = __pyx_v_buf;
  }
  __pyx_L3:;

  /* "View.MemoryView":255
 *         result.data = buf
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":244
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.array_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":281
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

/* Python wrapper */
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(2, 281, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_name = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 281, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.Enum.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "View.MemoryView":282
 *     cdef object name
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return self.name
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "View.MemoryView":281
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":283
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":284
 *         self.name = name
 *     def __repr__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* "View.MemoryView":283
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum___reduce_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->name != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.Enum.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum_2__setstate_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_unpickle_Enum__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":298
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

static void *__pyx_align_pointer(void *__pyx_v_memory, size_t __pyx_v_alignment) {
  Py_intptr_t __pyx_v_aligned_p;
  size_t __pyx_v_offset;
  void *__pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":300
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory             # <<<<<<<<<<<<<<
 *     cdef size_t offset
 * 
 */
  __pyx_v_aligned_p = ((Py_intptr_t)__pyx_v_memory);

  /* "View.MemoryView":304
 * 
 *     with cython.cdivision(True):
 *         offset = aligned_p % alignment             # <<<<<<<<<<<<<<
 * 
 *     if offset > 0:
 */
  __pyx_v_offset = (__pyx_v_aligned_p % __pyx_v_alignment);

  /* "View.MemoryView":306
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  __pyx_t_1 = ((__pyx_v_offset > 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":307
 * 
 *     if offset > 0:
 *         aligned_p += alignment - offset             # <<<<<<<<<<<<<<
 * 
 *     return <void *> aligned_p
 */
    __pyx_v_aligned_p = (__pyx_v_aligned_p + (__pyx_v_alignment - __pyx_v_offset));

    /* "View.MemoryView":306
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  }

  /* "View.MemoryView":309
 *         aligned_p += alignment - offset
 * 
 *     return <void *> aligned_p             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((void *)__pyx_v_aligned_p);
  goto __pyx_L0;

  /* "View.MemoryView":298
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":345
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

/* Python wrapper */
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  int __pyx_v_flags;
  int __pyx_v_dtype_is_object;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_flags,&__pyx_n_s_dtype_is_object,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flags)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); __PYX_ERR(2, 345, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dtype_is_object);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(2, 345, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_obj = values[0];
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 345, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_dtype_is_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_dtype_is_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 345, __pyx_L3_error)
    } else {
      __pyx_v_dtype_is_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 345, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_obj, __pyx_v_flags, __pyx_v_dtype_is_object);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "View.MemoryView":346
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 */
  __Pyx_INCREF(__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(__pyx_v_self->obj);
  __pyx_v_self->obj = __pyx_v_obj;

  /* "View.MemoryView":347
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj
 *         self.flags = flags             # <<<<<<<<<<<<<<
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 */
  __pyx_v_self->flags = __pyx_v_flags;

  /* "View.MemoryView":348
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  __pyx_t_2 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))) == ((PyObject *)__pyx_memoryview_type));
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_obj != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "View.MemoryView":349
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)             # <<<<<<<<<<<<<<
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 */
    __pyx_t_4 = __Pyx_GetBuffer(__pyx_v_obj, (&__pyx_v_self->view), __pyx_v_flags); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 349, __pyx_L1_error)

    /* "View.MemoryView":350
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    __pyx_t_1 = ((((PyObject *)__pyx_v_self->view.obj) == NULL) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":351
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None             # <<<<<<<<<<<<<<
 *                 Py_INCREF(Py_None)
 * 
 */
      ((Py_buffer *)(&__pyx_v_self->view))->obj = Py_None;

      /* "View.MemoryView":352
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         global __pyx_memoryview_thread_locks_used
 */
      Py_INCREF(Py_None);

      /* "View.MemoryView":350
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    }

    /* "View.MemoryView":348
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  }

  /* "View.MemoryView":355
 * 
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 */
  __pyx_t_1 = ((__pyx_memoryview_thread_locks_used < 8) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":356
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]             # <<<<<<<<<<<<<<
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:
 */
    __pyx_v_self->lock = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

    /* "View.MemoryView":357
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1             # <<<<<<<<<<<<<<
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 */
    __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used + 1);

    /* "View.MemoryView":355
 * 
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 */
  }

  /* "View.MemoryView":358
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:             # <<<<<<<<<<<<<<
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 */
  __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":359
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()             # <<<<<<<<<<<<<<
 *             if self.lock is NULL:
 *                 raise MemoryError
 */
    __pyx_v_self->lock = PyThread_allocate_lock();

    /* "View.MemoryView":360
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "View.MemoryView":361
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
      PyErr_NoMemory(); __PYX_ERR(2, 361, __pyx_L1_error)

      /* "View.MemoryView":360
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    }

    /* "View.MemoryView":358
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:             # <<<<<<<<<<<<<<
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 */
  }

  /* "View.MemoryView":363
 *                 raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":364
 * 
 *         if flags & PyBUF_FORMAT:
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')             # <<<<<<<<<<<<<<
 *         else:
 *             self.dtype_is_object = dtype_is_object
 */
    __pyx_t_2 = (((__pyx_v_self->view.format[0]) == 'O') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_2 = (((__pyx_v_self->view.format[1]) == '\x00') != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L11_bool_binop_done:;
    __pyx_v_self->dtype_is_object = __pyx_t_1;

    /* "View.MemoryView":363
 *                 raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":366
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 *             self.dtype_is_object = dtype_is_object             # <<<<<<<<<<<<<<
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 */
  /*else*/ {
    __pyx_v_self->dtype_is_object = __pyx_v_dtype_is_object;
  }
  __pyx_L10:;

  /* "View.MemoryView":368
 *             self.dtype_is_object = dtype_is_object
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(             # <<<<<<<<<<<<<<
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL
 */
  __pyx_v_self->acquisition_count_aligned_p = ((__pyx_atomic_int *)__pyx_align_pointer(((void *)(&(__pyx_v_self->acquisition_count[0]))), (sizeof(__pyx_atomic_int))));

  /* "View.MemoryView":370
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(memoryview self):
 */
  __pyx_v_self->typeinfo = NULL;

  /* "View.MemoryView":345
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":372
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

/* Python wrapper */
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self) {
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyThread_type_lock __pyx_t_6;
  PyThread_type_lock __pyx_t_7;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":373
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
  __pyx_t_1 = (__pyx_v_self->obj != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":374
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)             # <<<<<<<<<<<<<<
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 */
    __Pyx_ReleaseBuffer((&__pyx_v_self->view));

    /* "View.MemoryView":373
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":375
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  __pyx_t_2 = ((((Py_buffer *)(&__pyx_v_self->view))->obj == Py_None) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":377
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL             # <<<<<<<<<<<<<<
 *             Py_DECREF(Py_None)
 * 
 */
    ((Py_buffer *)(&__pyx_v_self->view))->obj = NULL;

    /* "View.MemoryView":378
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 *             Py_DECREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         cdef int i
 */
    Py_DECREF(Py_None);

    /* "View.MemoryView":375
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":382
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  __pyx_t_2 = ((__pyx_v_self->lock != NULL) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":383
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):             # <<<<<<<<<<<<<<
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 */
    __pyx_t_3 = __pyx_memoryview_thread_locks_used;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "View.MemoryView":384
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      __pyx_t_2 = (((__pyx_memoryview_thread_locks[__pyx_v_i]) == __pyx_v_self->lock) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":385
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1             # <<<<<<<<<<<<<<
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 */
        __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used - 1);

        /* "View.MemoryView":386
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        __pyx_t_2 = ((__pyx_v_i != __pyx_memoryview_thread_locks_used) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":388
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */
          __pyx_t_6 = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);
          __pyx_t_7 = (__pyx_memoryview_thread_locks[__pyx_v_i]);

          /* "View.MemoryView":387
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (             # <<<<<<<<<<<<<<
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break
 */
          (__pyx_memoryview_thread_locks[__pyx_v_i]) = __pyx_t_6;
          (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]) = __pyx_t_7;

          /* "View.MemoryView":386
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        }

        /* "View.MemoryView":389
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 PyThread_free_lock(self.lock)
 */
        goto __pyx_L6_break;

        /* "View.MemoryView":384
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      }
    }
    /*else*/ {

      /* "View.MemoryView":391
 *                     break
 *             else:
 *                 PyThread_free_lock(self.lock)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 */
      PyThread_free_lock(__pyx_v_self->lock);
    }
    __pyx_L6_break:;

    /* "View.MemoryView":382
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  }

  /* "View.MemoryView":372
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":393
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  Py_ssize_t __pyx_v_dim;
  char *__pyx_v_itemp;
  PyObject *__pyx_v_idx = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  char *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_item_pointer", 0);

  /* "View.MemoryView":395
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf             # <<<<<<<<<<<<<<
 * 
 *         for dim, idx in enumerate(index):
 */
  __pyx_v_itemp = ((char *)__pyx_v_self->view.buf);

  /* "View.MemoryView":397
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_v_index)) || PyTuple_CheckExact(__pyx_v_index)) {
    __pyx_t_2 = __pyx_v_index; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 397, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(2, 397, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(2, 397, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 397, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_1;
    __pyx_t_1 = (__pyx_t_1 + 1);

    /* "View.MemoryView":398
 * 
 *         for dim, idx in enumerate(index):
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)             # <<<<<<<<<<<<<<
 * 
 *         return itemp
 */
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 398, __pyx_L1_error)
    __pyx_t_7 = __pyx_pybuffer_index((&__pyx_v_self->view), __pyx_v_itemp, __pyx_t_6, __pyx_v_dim); if (unlikely(__pyx_t_7 == ((char *)NULL))) __PYX_ERR(2, 398, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_7;

    /* "View.MemoryView":397
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":400
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 *         return itemp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_itemp;
  goto __pyx_L0;

  /* "View.MemoryView":393
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.get_item_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":403
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index); /*proto*/
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_indices = NULL;
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":404
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  __pyx_t_1 = (__pyx_v_index == __pyx_builtin_Ellipsis);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":405
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __pyx_r = ((PyObject *)__pyx_v_self);
    goto __pyx_L0;

    /* "View.MemoryView":404
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  }

  /* "View.MemoryView":407
 *             return self
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         cdef char *itemp
 */
  __pyx_t_3 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(__pyx_t_3 != Py_None)) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(2, 407, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(2, 407, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_indices = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "View.MemoryView":410
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 410, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "View.MemoryView":411
 *         cdef char *itemp
 *         if have_slices:
 *             return memview_slice(self, indices)             # <<<<<<<<<<<<<<
 *         else:
 *             itemp = self.get_item_pointer(indices)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((PyObject *)__pyx_memview_slice(__pyx_v_self, __pyx_v_indices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":410
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  }

  /* "View.MemoryView":413
 *             return memview_slice(self, indices)
 *         else:
 *             itemp = self.get_item_pointer(indices)             # <<<<<<<<<<<<<<
 *             return self.convert_item_to_object(itemp)
 * 
 */
  /*else*/ {
    __pyx_t_6 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_indices); if (unlikely(__pyx_t_6 == ((char *)NULL))) __PYX_ERR(2, 413, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_6;

    /* "View.MemoryView":414
 *         else:
 *             itemp = self.get_item_pointer(indices)
 *             return self.convert_item_to_object(itemp)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(memoryview self, object index, object value):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->convert_item_to_object(__pyx_v_self, __pyx_v_itemp); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":403
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":416
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

/* Python wrapper */
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /*proto*/
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_obj = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);
  __Pyx_INCREF(__pyx_v_index);

  /* "View.MemoryView":417
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  __pyx_t_1 = (__pyx_v_self->view.readonly != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":418
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__83, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 418, __pyx_L1_error)

    /* "View.MemoryView":417
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":420
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         if have_slices:
 */
  __pyx_t_2 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(__pyx_t_2 != Py_None)) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(2, 420, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(2, 420, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_3;
  __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "View.MemoryView":422
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 422, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "View.MemoryView":423
 * 
 *         if have_slices:
 *             obj = self.is_slice(value)             # <<<<<<<<<<<<<<
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->is_slice(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_obj = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "View.MemoryView":424
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 424, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "View.MemoryView":425
 *             obj = self.is_slice(value)
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)
 */
      __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assignment(__pyx_v_self, __pyx_t_2, __pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "View.MemoryView":424
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":427
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)             # <<<<<<<<<<<<<<
 *         else:
 *             self.setitem_indexed(index, value)
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_memoryview_type))))) __PYX_ERR(2, 427, __pyx_L1_error)
      __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assign_scalar(__pyx_v_self, ((struct __pyx_memoryview_obj *)__pyx_t_4), __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L5:;

    /* "View.MemoryView":422
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":429
 *                 self.setitem_slice_assign_scalar(self[index], value)
 *         else:
 *             self.setitem_indexed(index, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef is_slice(self, obj):
 */
  /*else*/ {
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_indexed(__pyx_v_self, __pyx_v_index, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "View.MemoryView":416
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":431
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_slice", 0);
  __Pyx_INCREF(__pyx_v_obj);

  /* "View.MemoryView":432
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_memoryview_type); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "View.MemoryView":434
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_self->flags & (~PyBUF_WRITABLE)) | PyBUF_ANY_CONTIGUOUS)); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":435
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 return None
 */
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 435, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "View.MemoryView":434
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_obj);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7);
        __pyx_t_6 = 0;
        __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "View.MemoryView":436
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(2, 436, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":437
 *                                  self.dtype_is_object)
 *             except TypeError:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         return obj
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L7_except_return;
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L7_except_return:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L0;
      __pyx_L9_try_end:;
    }

    /* "View.MemoryView":432
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  }

  /* "View.MemoryView":439
 *                 return None
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assignment(self, dst, src):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_obj);
  __pyx_r = __pyx_v_obj;
  goto __pyx_L0;

  /* "View.MemoryView":431
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":441
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src) {
  __Pyx_memviewslice __pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_src_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  __Pyx_memviewslice *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assignment", 0);

  /* "View.MemoryView":445
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_memoryview_type))))) __PYX_ERR(2, 445, __pyx_L1_error)
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_src), (&__pyx_v_src_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 445, __pyx_L1_error)

  /* "View.MemoryView":446
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],             # <<<<<<<<<<<<<<
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 */
  if (!(likely(((__pyx_v_dst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_dst, __pyx_memoryview_type))))) __PYX_ERR(2, 446, __pyx_L1_error)
  __pyx_t_2 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_dst), (&__pyx_v_dst_slice)); if (unlikely(__pyx_t_2 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 446, __pyx_L1_error)

  /* "View.MemoryView":447
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":445
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  __pyx_t_6 = __pyx_memoryview_copy_contents((__pyx_t_1[0]), (__pyx_t_2[0]), __pyx_t_4, __pyx_t_5, __pyx_v_self->dtype_is_object); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 445, __pyx_L1_error)

  /* "View.MemoryView":441
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":449
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value) {
  int __pyx_v_array[0x80];
  void *__pyx_v_tmp;
  void *__pyx_v_item;
  __Pyx_memviewslice *__pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_tmp_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assign_scalar", 0);

  /* "View.MemoryView":451
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 *         cdef int array[128]
 *         cdef void *tmp = NULL             # <<<<<<<<<<<<<<
 *         cdef void *item
 * 
 */
  __pyx_v_tmp = NULL;

  /* "View.MemoryView":456
 *         cdef __Pyx_memviewslice *dst_slice
 *         cdef __Pyx_memviewslice tmp_slice
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)             # <<<<<<<<<<<<<<
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_dst, (&__pyx_v_tmp_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 456, __pyx_L1_error)
  __pyx_v_dst_slice = __pyx_t_1;

  /* "View.MemoryView":458
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
  __pyx_t_2 = ((((size_t)__pyx_v_self->view.itemsize) > (sizeof(__pyx_v_array))) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":459
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)             # <<<<<<<<<<<<<<
 *             if tmp == NULL:
 *                 raise MemoryError
 */
    __pyx_v_tmp = PyMem_Malloc(__pyx_v_self->view.itemsize);

    /* "View.MemoryView":460
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    __pyx_t_2 = ((__pyx_v_tmp == NULL) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":461
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             item = tmp
 *         else:
 */
      PyErr_NoMemory(); __PYX_ERR(2, 461, __pyx_L1_error)

      /* "View.MemoryView":460
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    }

    /* "View.MemoryView":462
 *             if tmp == NULL:
 *                 raise MemoryError
 *             item = tmp             # <<<<<<<<<<<<<<
 *         else:
 *             item = <void *> array
 */
    __pyx_v_item = __pyx_v_tmp;

    /* "View.MemoryView":458
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":464
 *             item = tmp
 *         else:
 *             item = <void *> array             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  /*else*/ {
    __pyx_v_item = ((void *)__pyx_v_array);
  }
  __pyx_L3:;

  /* "View.MemoryView":466
 *             item = <void *> array
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value
 */
  /*try:*/ {

    /* "View.MemoryView":467
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
    __pyx_t_2 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":468
 *         try:
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value             # <<<<<<<<<<<<<<
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)
 */
      (((PyObject **)__pyx_v_item)[0]) = ((PyObject *)__pyx_v_value);

      /* "View.MemoryView":467
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
      goto __pyx_L8;
    }

    /* "View.MemoryView":470
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else*/ {
      __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, ((char *)__pyx_v_item), __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 470, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_L8:;

    /* "View.MemoryView":474
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    __pyx_t_2 = ((__pyx_v_self->view.suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":475
 * 
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)             # <<<<<<<<<<<<<<
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 *                                 item, self.dtype_is_object)
 */
      __pyx_t_3 = assert_direct_dimensions(__pyx_v_self->view.suboffsets, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 475, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "View.MemoryView":474
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    }

    /* "View.MemoryView":476
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,             # <<<<<<<<<<<<<<
 *                                 item, self.dtype_is_object)
 *         finally:
 */
    __pyx_memoryview_slice_assign_scalar(__pyx_v_dst_slice, __pyx_v_dst->view.ndim, __pyx_v_self->view.itemsize, __pyx_v_item, __pyx_v_self->dtype_is_object);
  }

  /* "View.MemoryView":479
 *                                 item, self.dtype_is_object)
 *         finally:
 *             PyMem_Free(tmp)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_indexed(self, index, value):
 */
  /*finally:*/ {
    /*normal exit:*/{
      PyMem_Free(__pyx_v_tmp);
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;
      {
        PyMem_Free(__pyx_v_tmp);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "View.MemoryView":449
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":481
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_indexed", 0);

  /* "View.MemoryView":482
 * 
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)             # <<<<<<<<<<<<<<
 *         self.assign_item_from_object(itemp, value)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_index); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(2, 482, __pyx_L1_error)
  __pyx_v_itemp = __pyx_t_1;

  /* "View.MemoryView":483
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":481
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_indexed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":485
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_bytesitem = 0;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":488
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef bytes bytesitem
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":491
 *         cdef bytes bytesitem
 * 
 *         bytesitem = itemp[:self.view.itemsize]             # <<<<<<<<<<<<<<
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_itemp + 0, __pyx_v_self->view.itemsize - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bytesitem = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "View.MemoryView":493
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)             # <<<<<<<<<<<<<<
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 493, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 493, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 493, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_6);
        __Pyx_INCREF(__pyx_v_bytesitem);
        __Pyx_GIVEREF(__pyx_v_bytesitem);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_bytesitem);
        __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 493, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_result = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    }

    /* "View.MemoryView":497
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
    /*else:*/ {
      __pyx_t_10 = strlen(__pyx_v_self->view.format); 
      __pyx_t_11 = ((__pyx_t_10 == 1) != 0);
      if (__pyx_t_11) {

        /* "View.MemoryView":498
 *         else:
 *             if len(self.view.format) == 1:
 *                 return result[0]             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_result, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 498, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L6_except_return;

        /* "View.MemoryView":497
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
      }

      /* "View.MemoryView":499
 *             if len(self.view.format) == 1:
 *                 return result[0]
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L6_except_return;
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "View.MemoryView":494
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unable to convert item to object")
 *         else:
 */
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_5, &__pyx_t_9);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 494, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_5, __pyx_t_9);
    __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_9 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(2, 494, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_1);

      /* "View.MemoryView":495
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__84, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 495, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(2, 495, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
  }

  /* "View.MemoryView":485
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesitem);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":501
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_struct = NULL;
  char __pyx_v_c;
  PyObject *__pyx_v_bytesvalue = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  char *__pyx_t_11;
  char *__pyx_t_12;
  char *__pyx_t_13;
  char *__pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":504
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef char c
 *         cdef bytes bytesvalue
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":509
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "View.MemoryView":510
 * 
 *         if isinstance(value, tuple):
 *             bytesvalue = struct.pack(self.view.format, *value)             # <<<<<<<<<<<<<<
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(2, 510, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":509
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":512
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)             # <<<<<<<<<<<<<<
 * 
 *         for i, c in enumerate(bytesvalue):
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 512, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 512, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 512, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_value);
      __Pyx_GIVEREF(__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_value);
      __pyx_t_1 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 512, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(2, 512, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
  __pyx_t_9 = 0;
  if (unlikely(__pyx_v_bytesvalue == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
    __PYX_ERR(2, 514, __pyx_L1_error)
  }
  __Pyx_INCREF(__pyx_v_bytesvalue);
  __pyx_t_10 = __pyx_v_bytesvalue;
  __pyx_t_12 = PyBytes_AS_STRING(__pyx_t_10);
  __pyx_t_13 = (__pyx_t_12 + PyBytes_GET_SIZE(__pyx_t_10));
  for (__pyx_t_14 = __pyx_t_12; __pyx_t_14 < __pyx_t_13; __pyx_t_14++) {
    __pyx_t_11 = __pyx_t_14;
    __pyx_v_c = (__pyx_t_11[0]);

    /* "View.MemoryView":515
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    __pyx_v_i = __pyx_t_9;

    /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
    __pyx_t_9 = (__pyx_t_9 + 1);

    /* "View.MemoryView":515
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    (__pyx_v_itemp[__pyx_v_i]) = __pyx_v_c;
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "View.MemoryView":501
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.memoryview.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":518
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  char *__pyx_t_5;
  void *__pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":519
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  __pyx_t_2 = ((__pyx_v_flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->view.readonly != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":520
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__85, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 520, __pyx_L1_error)

    /* "View.MemoryView":519
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":522
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":523
 * 
 *         if flags & PyBUF_ND:
 *             info.shape = self.view.shape             # <<<<<<<<<<<<<<
 *         else:
 *             info.shape = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.shape;
    __pyx_v_info->shape = __pyx_t_4;

    /* "View.MemoryView":522
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
    goto __pyx_L6;
  }

  /* "View.MemoryView":525
 *             info.shape = self.view.shape
 *         else:
 *             info.shape = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */
  /*else*/ {
    __pyx_v_info->shape = NULL;
  }
  __pyx_L6:;

  /* "View.MemoryView":527
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":528
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.strides = self.view.strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.strides = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.strides;
    __pyx_v_info->strides = __pyx_t_4;

    /* "View.MemoryView":527
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
    goto __pyx_L7;
  }

  /* "View.MemoryView":530
 *             info.strides = self.view.strides
 *         else:
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_INDIRECT:
 */
  /*else*/ {
    __pyx_v_info->strides = NULL;
  }
  __pyx_L7:;

  /* "View.MemoryView":532
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_INDIRECT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":533
 * 
 *         if flags & PyBUF_INDIRECT:
 *             info.suboffsets = self.view.suboffsets             # <<<<<<<<<<<<<<
 *         else:
 *             info.suboffsets = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.suboffsets;
    __pyx_v_info->suboffsets = __pyx_t_4;

    /* "View.MemoryView":532
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
    goto __pyx_L8;
  }

  /* "View.MemoryView":535
 *             info.suboffsets = self.view.suboffsets
 *         else:
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  /*else*/ {
    __pyx_v_info->suboffsets = NULL;
  }
  __pyx_L8:;

  /* "View.MemoryView":537
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":538
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.view.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_5 = __pyx_v_self->view.format;
    __pyx_v_info->format = __pyx_t_5;

    /* "View.MemoryView":537
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":540
 *             info.format = self.view.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.buf = self.view.buf
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L9:;

  /* "View.MemoryView":542
 *             info.format = NULL
 * 
 *         info.buf = self.view.buf             # <<<<<<<<<<<<<<
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 */
  __pyx_t_6 = __pyx_v_self->view.buf;
  __pyx_v_info->buf = __pyx_t_6;

  /* "View.MemoryView":543
 * 
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim             # <<<<<<<<<<<<<<
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 */
  __pyx_t_7 = __pyx_v_self->view.ndim;
  __pyx_v_info->ndim = __pyx_t_7;

  /* "View.MemoryView":544
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize             # <<<<<<<<<<<<<<
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 */
  __pyx_t_8 = __pyx_v_self->view.itemsize;
  __pyx_v_info->itemsize = __pyx_t_8;

  /* "View.MemoryView":545
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len             # <<<<<<<<<<<<<<
 *         info.readonly = self.view.readonly
 *         info.obj = self
 */
  __pyx_t_8 = __pyx_v_self->view.len;
  __pyx_v_info->len = __pyx_t_8;

  /* "View.MemoryView":546
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */
  __pyx_t_1 = __pyx_v_self->view.readonly;
  __pyx_v_info->readonly = __pyx_t_1;

  /* "View.MemoryView":547
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":518
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":553
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":554
 *     @property
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&result.from_slice)
 *         return result
 */
  __pyx_t_1 = __pyx_memoryview_copy_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_memoryviewslice_type))))) __PYX_ERR(2, 554, __pyx_L1_error)
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":555
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __pyx_t_2 = __pyx_memslice_transpose((&__pyx_v_result->from_slice)); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(2, 555, __pyx_L1_error)

  /* "View.MemoryView":556
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":553
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.T.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":559
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":560
 *     @property
 *     def base(self):
 *         return self.obj             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->obj);
  __pyx_r = __pyx_v_self->obj;
  goto __pyx_L0;

  /* "View.MemoryView":559
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":563
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":564
 *     @property
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
  for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_length = (__pyx_t_2[0]);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(2, 564, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = PyList_AsTuple(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":563
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":567
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_stride;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":568
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  __pyx_t_1 = ((__pyx_v_self->view.strides == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":570
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__86, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 570, __pyx_L1_error)

    /* "View.MemoryView":568
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  }

  /* "View.MemoryView":572
 *             raise ValueError("Buffer view does not expose strides")
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = (__pyx_v_self->view.strides + __pyx_v_self->view.ndim);
  for (__pyx_t_5 = __pyx_v_self->view.strides; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
    __pyx_t_3 = __pyx_t_5;
    __pyx_v_stride = (__pyx_t_3[0]);
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_stride); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(2, 572, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":567
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.strides.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":575
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":576
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.suboffsets == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":577
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Multiply(__pyx_tuple__87, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":576
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  }

  /* "View.MemoryView":579
 *             return (-1,) * self.view.ndim
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = (__pyx_v_self->view.suboffsets + __pyx_v_self->view.ndim);
  for (__pyx_t_6 = __pyx_v_self->view.suboffsets; __pyx_t_6 < __pyx_t_5; __pyx_t_6++) {
    __pyx_t_4 = __pyx_t_6;
    __pyx_v_suboffset = (__pyx_t_4[0]);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_suboffset); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(2, 579, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PyList_AsTuple(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":575
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.suboffsets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":582
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":583
 *     @property
 *     def ndim(self):
 *         return self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":582
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.ndim.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":586
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":587
 *     @property
 *     def itemsize(self):
 *         return self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":586
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.itemsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":590
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":591
 *     @property
 *     def nbytes(self):
 *         return self.size * self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":590
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.nbytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":594
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":595
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_size == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":596
 *     def size(self):
 *         if self._size is None:
 *             result = 1             # <<<<<<<<<<<<<<
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 */
    __Pyx_INCREF(__pyx_int_1);
    __pyx_v_result = __pyx_int_1;

    /* "View.MemoryView":598
 *             result = 1
 * 
 *             for length in self.view.shape[:self.view.ndim]:             # <<<<<<<<<<<<<<
 *                 result *= length
 * 
 */
    __pyx_t_4 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
    for (__pyx_t_5 = __pyx_v_self->view.shape; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
      __pyx_t_3 = __pyx_t_5;
      __pyx_t_6 = PyInt_FromSsize_t((__pyx_t_3[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "View.MemoryView":599
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 *                 result *= length             # <<<<<<<<<<<<<<
 * 
 *             self._size = result
 */
      __pyx_t_6 = PyNumber_InPlaceMultiply(__pyx_v_result, __pyx_v_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 599, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_6);
      __pyx_t_6 = 0;
    }

    /* "View.MemoryView":601
 *                 result *= length
 * 
 *             self._size = result             # <<<<<<<<<<<<<<
 * 
 *         return self._size
 */
    __Pyx_INCREF(__pyx_v_result);
    __Pyx_GIVEREF(__pyx_v_result);
    __Pyx_GOTREF(__pyx_v_self->_size);
    __Pyx_DECREF(__pyx_v_self->_size);
    __pyx_v_self->_size = __pyx_v_result;

    /* "View.MemoryView":595
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  }

  /* "View.MemoryView":603
 *             self._size = result
 * 
 *         return self._size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_size);
  __pyx_r = __pyx_v_self->_size;
  goto __pyx_L0;

  /* "View.MemoryView":594
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":605
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

/* Python wrapper */
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":606
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.ndim >= 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":607
 *     def __len__(self):
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */
    __pyx_r = (__pyx_v_self->view.shape[0]);
    goto __pyx_L0;

    /* "View.MemoryView":606
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  }

  /* "View.MemoryView":609
 *             return self.view.shape[0]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":605
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":611
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":612
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":613
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "View.MemoryView":612
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":611
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":615
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "View.MemoryView":616
 * 
 *     def __str__(self):
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_object, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":615
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":619
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_c_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_c_contig", 0);

  /* "View.MemoryView":622
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 622, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":623
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def is_f_contig(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'C', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":619
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_c_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":625
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_f_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_f_contig", 0);

  /* "View.MemoryView":628
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 628, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":629
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'F', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":625
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_f_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":631
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_mslice;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "View.MemoryView":633
 *     def copy(self):
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &mslice)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_F_CONTIGUOUS));

  /* "View.MemoryView":635
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 * 
 *         slice_copy(self, &mslice)             # <<<<<<<<<<<<<<
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,
 *                                    self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_mslice));

  /* "View.MemoryView":636
 * 
 *         slice_copy(self, &mslice)
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                    self.view.itemsize,
 *                                    flags|PyBUF_C_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_mslice), ((char *)"c"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_C_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 636, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":641
 *                                    self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &mslice)             # <<<<<<<<<<<<<<
 * 
 *     def copy_fortran(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_mslice)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":631
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":643
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy_fortran (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy_fortran", 0);

  /* "View.MemoryView":645
 *     def copy_fortran(self):
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &src)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_C_CONTIGUOUS));

  /* "View.MemoryView":647
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 * 
 *         slice_copy(self, &src)             # <<<<<<<<<<<<<<
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,
 *                                 self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_src));

  /* "View.MemoryView":648
 * 
 *         slice_copy(self, &src)
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                 self.view.itemsize,
 *                                 flags|PyBUF_F_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_src), ((char *)"fortran"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_F_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 648, __pyx_L1_error)
  __pyx_v_dst = __pyx_t_1;

  /* "View.MemoryView":653
 *                                 self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_dst)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":643
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy_fortran", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview___reduce_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__88, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview_2__setstate_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__89, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":657
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

static PyObject *__pyx_memoryview_new(PyObject *__pyx_v_o, int __pyx_v_flags, int __pyx_v_dtype_is_object, __Pyx_TypeInfo *__pyx_v_typeinfo) {
  struct __pyx_memoryview_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_cwrapper", 0);

  /* "View.MemoryView":658
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)             # <<<<<<<<<<<<<<
 *     result.typeinfo = typeinfo
 *     return result
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_o);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryview_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":659
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_v_result->typeinfo = __pyx_v_typeinfo;

  /* "View.MemoryView":660
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_check')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":657
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":663
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

static CYTHON_INLINE int __pyx_memoryview_check(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("memoryview_check", 0);

  /* "View.MemoryView":664
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):
 *     return isinstance(o, memoryview)             # <<<<<<<<<<<<<<
 * 
 * cdef tuple _unellipsify(object index, int ndim):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_memoryview_type); 
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "View.MemoryView":663
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":666
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

static PyObject *_unellipsify(PyObject *__pyx_v_index, int __pyx_v_ndim) {
  PyObject *__pyx_v_tup = NULL;
  PyObject *__pyx_v_result = NULL;
  int __pyx_v_have_slices;
  int __pyx_v_seen_ellipsis;
  CYTHON_UNUSED PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_item = NULL;
  Py_ssize_t __pyx_v_nslices;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unellipsify", 0);

  /* "View.MemoryView":671
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
  __pyx_t_1 = PyTuple_Check(__pyx_v_index); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":672
 *     """
 *     if not isinstance(index, tuple):
 *         tup = (index,)             # <<<<<<<<<<<<<<
 *     else:
 *         tup = index
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_index);
    __Pyx_GIVEREF(__pyx_v_index);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_index);
    __pyx_v_tup = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "View.MemoryView":671
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":674
 *         tup = (index,)
 *     else:
 *         tup = index             # <<<<<<<<<<<<<<
 * 
 *     result = []
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_index);
    __pyx_v_tup = __pyx_v_index;
  }
  __pyx_L3:;

  /* "View.MemoryView":676
 *         tup = index
 * 
 *     result = []             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_result = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":677
 * 
 *     result = []
 *     have_slices = False             # <<<<<<<<<<<<<<
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 */
  __pyx_v_have_slices = 0;

  /* "View.MemoryView":678
 *     result = []
 *     have_slices = False
 *     seen_ellipsis = False             # <<<<<<<<<<<<<<
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 */
  __pyx_v_seen_ellipsis = 0;

  /* "View.MemoryView":679
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_3 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_tup)) || PyTuple_CheckExact(__pyx_v_tup)) {
    __pyx_t_4 = __pyx_v_tup; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_tup); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 679, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(2, 679, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(2, 679, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_6(__pyx_t_4);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 679, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_3);
    __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "View.MemoryView":680
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
    __pyx_t_2 = (__pyx_v_item == __pyx_builtin_Ellipsis);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":681
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
      __pyx_t_1 = ((!(__pyx_v_seen_ellipsis != 0)) != 0);
      if (__pyx_t_1) {

        /* "View.MemoryView":682
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */
        __pyx_t_8 = PyObject_Length(__pyx_v_tup); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(2, 682, __pyx_L1_error)
        __pyx_t_7 = PyList_New(1 * ((((__pyx_v_ndim - __pyx_t_8) + 1)<0) ? 0:((__pyx_v_ndim - __pyx_t_8) + 1))); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 682, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        { Py_ssize_t __pyx_temp;
          for (__pyx_temp=0; __pyx_temp < ((__pyx_v_ndim - __pyx_t_8) + 1); __pyx_temp++) {
            __Pyx_INCREF(__pyx_slice__90);
            __Pyx_GIVEREF(__pyx_slice__90);
            PyList_SET_ITEM(__pyx_t_7, __pyx_temp, __pyx_slice__90);
          }
        }
        __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_7); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 682, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "View.MemoryView":683
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True             # <<<<<<<<<<<<<<
 *             else:
 *                 result.append(slice(None))
 */
        __pyx_v_seen_ellipsis = 1;

        /* "View.MemoryView":681
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":685
 *                 seen_ellipsis = True
 *             else:
 *                 result.append(slice(None))             # <<<<<<<<<<<<<<
 *             have_slices = True
 *         else:
 */
      /*else*/ {
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_slice__90); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 685, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":686
 *             else:
 *                 result.append(slice(None))
 *             have_slices = True             # <<<<<<<<<<<<<<
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 */
      __pyx_v_have_slices = 1;

      /* "View.MemoryView":680
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":688
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
    /*else*/ {
      __pyx_t_2 = PySlice_Check(__pyx_v_item); 
      __pyx_t_10 = ((!(__pyx_t_2 != 0)) != 0);
      if (__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = ((!(PyIndex_Check(__pyx_v_item) != 0)) != 0);
      __pyx_t_1 = __pyx_t_10;
      __pyx_L9_bool_binop_done:;
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":689
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 *                 raise TypeError("Cannot index with type '%s'" % type(item))             # <<<<<<<<<<<<<<
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 */
        __pyx_t_7 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_index_with_type_s, ((PyObject *)Py_TYPE(__pyx_v_item))); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 689, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 689, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(2, 689, __pyx_L1_error)

        /* "View.MemoryView":688
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
      }

      /* "View.MemoryView":691
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 *             have_slices = have_slices or isinstance(item, slice)             # <<<<<<<<<<<<<<
 *             result.append(item)
 * 
 */
      __pyx_t_10 = (__pyx_v_have_slices != 0);
      if (!__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = PySlice_Check(__pyx_v_item); 
      __pyx_t_2 = (__pyx_t_10 != 0);
      __pyx_t_1 = __pyx_t_2;
      __pyx_L11_bool_binop_done:;
      __pyx_v_have_slices = __pyx_t_1;

      /* "View.MemoryView":692
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 *             result.append(item)             # <<<<<<<<<<<<<<
 * 
 *     nslices = ndim - len(result)
 */
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_item); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 692, __pyx_L1_error)
    }
    __pyx_L6:;

    /* "View.MemoryView":679
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":694
 *             result.append(item)
 * 
 *     nslices = ndim - len(result)             # <<<<<<<<<<<<<<
 *     if nslices:
 *         result.extend([slice(None)] * nslices)
 */
  __pyx_t_5 = PyList_GET_SIZE(__pyx_v_result); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(2, 694, __pyx_L1_error)
  __pyx_v_nslices = (__pyx_v_ndim - __pyx_t_5);

  /* "View.MemoryView":695
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  __pyx_t_1 = (__pyx_v_nslices != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":696
 *     nslices = ndim - len(result)
 *     if nslices:
 *         result.extend([slice(None)] * nslices)             # <<<<<<<<<<<<<<
 * 
 *     return have_slices or nslices, tuple(result)
 */
    __pyx_t_3 = PyList_New(1 * ((__pyx_v_nslices<0) ? 0:__pyx_v_nslices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_v_nslices; __pyx_temp++) {
        __Pyx_INCREF(__pyx_slice__90);
        __Pyx_GIVEREF(__pyx_slice__90);
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, __pyx_slice__90);
      }
    }
    __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 696, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "View.MemoryView":695
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  }

  /* "View.MemoryView":698
 *         result.extend([slice(None)] * nslices)
 * 
 *     return have_slices or nslices, tuple(result)             # <<<<<<<<<<<<<<
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!__pyx_v_have_slices) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_have_slices); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L14_bool_binop_done;
  }
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_nslices); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_L14_bool_binop_done:;
  __pyx_t_4 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = ((PyObject*)__pyx_t_11);
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":666
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("View.MemoryView._unellipsify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tup);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

static PyObject *assert_direct_dimensions(Py_ssize_t *__pyx_v_suboffsets, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assert_direct_dimensions", 0);

  /* "View.MemoryView":701
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")
 */
  __pyx_t_2 = (__pyx_v_suboffsets + __pyx_v_ndim);
  for (__pyx_t_3 = __pyx_v_suboffsets; __pyx_t_3 < __pyx_t_2; __pyx_t_3++) {
    __pyx_t_1 = __pyx_t_3;
    __pyx_v_suboffset = (__pyx_t_1[0]);

    /* "View.MemoryView":702
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    __pyx_t_4 = ((__pyx_v_suboffset >= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":703
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__91, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(2, 703, __pyx_L1_error)

      /* "View.MemoryView":702
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    }
  }

  /* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.assert_direct_dimensions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":710
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *__pyx_v_memview, PyObject *__pyx_v_indices) {
  int __pyx_v_new_ndim;
  int __pyx_v_suboffset_dim;
  int __pyx_v_dim;
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  __Pyx_memviewslice *__pyx_v_p_src;
  struct __pyx_memoryviewslice_obj *__pyx_v_memviewsliceobj = 0;
  __Pyx_memviewslice *__pyx_v_p_dst;
  int *__pyx_v_p_suboffset_dim;
  Py_ssize_t __pyx_v_start;
  Py_ssize_t __pyx_v_stop;
  Py_ssize_t __pyx_v_step;
  int __pyx_v_have_start;
  int __pyx_v_have_stop;
  int __pyx_v_have_step;
  PyObject *__pyx_v_index = NULL;
  struct __pyx_memoryview_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_memoryview_obj *__pyx_t_4;
  char *__pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memview_slice", 0);

  /* "View.MemoryView":711
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim             # <<<<<<<<<<<<<<
 *     cdef bint negative_step
 *     cdef __Pyx_memviewslice src, dst
 */
  __pyx_v_new_ndim = 0;
  __pyx_v_suboffset_dim = -1;

  /* "View.MemoryView":718
 * 
 * 
 *     memset(&dst, 0, sizeof(dst))             # <<<<<<<<<<<<<<
 * 
 *     cdef _memoryviewslice memviewsliceobj
 */
  (void)(memset((&__pyx_v_dst), 0, (sizeof(__pyx_v_dst))));

  /* "View.MemoryView":722
 *     cdef _memoryviewslice memviewsliceobj
 * 
 *     assert memview.view.ndim > 0             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_memview->view.ndim > 0) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(2, 722, __pyx_L1_error)
    }
  }
  #endif

  /* "View.MemoryView":724
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":725
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview             # <<<<<<<<<<<<<<
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(2, 725, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_memviewsliceobj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":726
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, &src)
 */
    __pyx_v_p_src = (&__pyx_v_memviewsliceobj->from_slice);

    /* "View.MemoryView":724
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":728
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 *         slice_copy(memview, &src)             # <<<<<<<<<<<<<<
 *         p_src = &src
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_src));

    /* "View.MemoryView":729
 *     else:
 *         slice_copy(memview, &src)
 *         p_src = &src             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_p_src = (&__pyx_v_src);
  }
  __pyx_L3:;

  /* "View.MemoryView":735
 * 
 * 
 *     dst.memview = p_src.memview             # <<<<<<<<<<<<<<
 *     dst.data = p_src.data
 * 
 */
  __pyx_t_4 = __pyx_v_p_src->memview;
  __pyx_v_dst.memview = __pyx_t_4;

  /* "View.MemoryView":736
 * 
 *     dst.memview = p_src.memview
 *     dst.data = p_src.data             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_v_p_src->data;
  __pyx_v_dst.data = __pyx_t_5;

  /* "View.MemoryView":741
 * 
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst             # <<<<<<<<<<<<<<
 *     cdef int *p_suboffset_dim = &suboffset_dim
 *     cdef Py_ssize_t start, stop, step
 */
  __pyx_v_p_dst = (&__pyx_v_dst);

  /* "View.MemoryView":742
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst
 *     cdef int *p_suboffset_dim = &suboffset_dim             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t start, stop, step
 *     cdef bint have_start, have_stop, have_step
 */
  __pyx_v_p_suboffset_dim = (&__pyx_v_suboffset_dim);

  /* "View.MemoryView":746
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  __pyx_t_6 = 0;
  if (likely(PyList_CheckExact(__pyx_v_indices)) || PyTuple_CheckExact(__pyx_v_indices)) {
    __pyx_t_3 = __pyx_v_indices; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 746, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 746, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(2, 746, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(2, 746, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      }
    } else {
      __pyx_t_9 = __pyx_t_8(__pyx_t_3);
      if (unlikely(!__pyx_t_9)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 746, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_v_dim = __pyx_t_6;
    __pyx_t_6 = (__pyx_t_6 + 1);

    /* "View.MemoryView":747
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
    __pyx_t_2 = (PyIndex_Check(__pyx_v_index) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":751
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 *                 index, 0, 0, # start, stop, step             # <<<<<<<<<<<<<<
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 */
      __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 751, __pyx_L1_error)

      /* "View.MemoryView":748
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_t_10, 0, 0, 0, 0, 0, 0); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(2, 748, __pyx_L1_error)

      /* "View.MemoryView":747
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":754
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
    __pyx_t_2 = (__pyx_v_index == Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":755
 *                 False)
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1             # <<<<<<<<<<<<<<
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 */
      (__pyx_v_p_dst->shape[__pyx_v_new_ndim]) = 1;

      /* "View.MemoryView":756
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0             # <<<<<<<<<<<<<<
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1
 */
      (__pyx_v_p_dst->strides[__pyx_v_new_ndim]) = 0;

      /* "View.MemoryView":757
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1             # <<<<<<<<<<<<<<
 *             new_ndim += 1
 *         else:
 */
      (__pyx_v_p_dst->suboffsets[__pyx_v_new_ndim]) = -1L;

      /* "View.MemoryView":758
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 *         else:
 *             start = index.start or 0
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

      /* "View.MemoryView":754
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":760
 *             new_ndim += 1
 *         else:
 *             start = index.start or 0             # <<<<<<<<<<<<<<
 *             stop = index.stop or 0
 *             step = index.step or 0
 */
    /*else*/ {
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 760, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 760, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 760, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L7_bool_binop_done:;
      __pyx_v_start = __pyx_t_10;

      /* "View.MemoryView":761
 *         else:
 *             start = index.start or 0
 *             stop = index.stop or 0             # <<<<<<<<<<<<<<
 *             step = index.step or 0
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 761, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 761, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 761, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L9_bool_binop_done:;
      __pyx_v_stop = __pyx_t_10;

      /* "View.MemoryView":762
 *             start = index.start or 0
 *             stop = index.stop or 0
 *             step = index.step or 0             # <<<<<<<<<<<<<<
 * 
 *             have_start = index.start is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 762, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 762, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L11_bool_binop_done:;
      __pyx_v_step = __pyx_t_10;

      /* "View.MemoryView":764
 *             step = index.step or 0
 * 
 *             have_start = index.start is not None             # <<<<<<<<<<<<<<
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 764, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_start = __pyx_t_1;

      /* "View.MemoryView":765
 * 
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None             # <<<<<<<<<<<<<<
 *             have_step = index.step is not None
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 765, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_stop = __pyx_t_1;

      /* "View.MemoryView":766
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None             # <<<<<<<<<<<<<<
 * 
 *             slice_memviewslice(
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_step = __pyx_t_1;

      /* "View.MemoryView":768
 *             have_step = index.step is not None
 * 
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_have_start, __pyx_v_have_stop, __pyx_v_have_step, 1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(2, 768, __pyx_L1_error)

      /* "View.MemoryView":774
 *                 have_start, have_stop, have_step,
 *                 True)
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);
    }
    __pyx_L6:;

    /* "View.MemoryView":746
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":776
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":777
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":778
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_dtype_func,
 *                                     memview.dtype_is_object)
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(2, 778, __pyx_L1_error) }

    /* "View.MemoryView":779
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 *     else:
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(2, 779, __pyx_L1_error) }

    /* "View.MemoryView":777
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, __pyx_v_memviewsliceobj->to_object_func, __pyx_v_memviewsliceobj->to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(2, 777, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":776
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  }

  /* "View.MemoryView":782
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":783
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,
 *                                     memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, NULL, NULL, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "View.MemoryView":782
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(2, 782, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":710
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memview_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_memviewsliceobj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":807
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *__pyx_v_dst, Py_ssize_t __pyx_v_shape, Py_ssize_t __pyx_v_stride, Py_ssize_t __pyx_v_suboffset, int __pyx_v_dim, int __pyx_v_new_ndim, int *__pyx_v_suboffset_dim, Py_ssize_t __pyx_v_start, Py_ssize_t __pyx_v_stop, Py_ssize_t __pyx_v_step, int __pyx_v_have_start, int __pyx_v_have_stop, int __pyx_v_have_step, int __pyx_v_is_slice) {
  Py_ssize_t __pyx_v_new_shape;
  int __pyx_v_negative_step;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":827
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
  __pyx_t_1 = ((!(__pyx_v_is_slice != 0)) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":829
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    __pyx_t_1 = ((__pyx_v_start < 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":830
 * 
 *         if start < 0:
 *             start += shape             # <<<<<<<<<<<<<<
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 */
      __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

      /* "View.MemoryView":829
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    }

    /* "View.MemoryView":831
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    __pyx_t_1 = (0 <= __pyx_v_start);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_start < __pyx_v_shape);
    }
    __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":832
 *             start += shape
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)             # <<<<<<<<<<<<<<
 *     else:
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"Index out of bounds (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 832, __pyx_L1_error)

      /* "View.MemoryView":831
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    }

    /* "View.MemoryView":827
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":835
 *     else:
 * 
 *         negative_step = have_step != 0 and step < 0             # <<<<<<<<<<<<<<
 * 
 *         if have_step and step == 0:
 */
  /*else*/ {
    __pyx_t_1 = ((__pyx_v_have_step != 0) != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step < 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L6_bool_binop_done:;
    __pyx_v_negative_step = __pyx_t_2;

    /* "View.MemoryView":837
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    __pyx_t_1 = (__pyx_v_have_step != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step == 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_2) {

      /* "View.MemoryView":838
 * 
 *         if have_step and step == 0:
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Step may not be zero (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 838, __pyx_L1_error)

      /* "View.MemoryView":837
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    }

    /* "View.MemoryView":841
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
    __pyx_t_2 = (__pyx_v_have_start != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":842
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
      __pyx_t_2 = ((__pyx_v_start < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":843
 *         if have_start:
 *             if start < 0:
 *                 start += shape             # <<<<<<<<<<<<<<
 *                 if start < 0:
 *                     start = 0
 */
        __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

        /* "View.MemoryView":844
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        __pyx_t_2 = ((__pyx_v_start < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":845
 *                 start += shape
 *                 if start < 0:
 *                     start = 0             # <<<<<<<<<<<<<<
 *             elif start >= shape:
 *                 if negative_step:
 */
          __pyx_v_start = 0;

          /* "View.MemoryView":844
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        }

        /* "View.MemoryView":842
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
        goto __pyx_L12;
      }

      /* "View.MemoryView":846
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      __pyx_t_2 = ((__pyx_v_start >= __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":847
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
        __pyx_t_2 = (__pyx_v_negative_step != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":848
 *             elif start >= shape:
 *                 if negative_step:
 *                     start = shape - 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     start = shape
 */
          __pyx_v_start = (__pyx_v_shape - 1);

          /* "View.MemoryView":847
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
          goto __pyx_L14;
        }

        /* "View.MemoryView":850
 *                     start = shape - 1
 *                 else:
 *                     start = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        /*else*/ {
          __pyx_v_start = __pyx_v_shape;
        }
        __pyx_L14:;

        /* "View.MemoryView":846
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      }
      __pyx_L12:;

      /* "View.MemoryView":841
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
      goto __pyx_L11;
    }

    /* "View.MemoryView":852
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":853
 *         else:
 *             if negative_step:
 *                 start = shape - 1             # <<<<<<<<<<<<<<
 *             else:
 *                 start = 0
 */
        __pyx_v_start = (__pyx_v_shape - 1);

        /* "View.MemoryView":852
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
        goto __pyx_L15;
      }

      /* "View.MemoryView":855
 *                 start = shape - 1
 *             else:
 *                 start = 0             # <<<<<<<<<<<<<<
 * 
 *         if have_stop:
 */
      /*else*/ {
        __pyx_v_start = 0;
      }
      __pyx_L15:;
    }
    __pyx_L11:;

    /* "View.MemoryView":857
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
    __pyx_t_2 = (__pyx_v_have_stop != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":858
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
      __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":859
 *         if have_stop:
 *             if stop < 0:
 *                 stop += shape             # <<<<<<<<<<<<<<
 *                 if stop < 0:
 *                     stop = 0
 */
        __pyx_v_stop = (__pyx_v_stop + __pyx_v_shape);

        /* "View.MemoryView":860
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":861
 *                 stop += shape
 *                 if stop < 0:
 *                     stop = 0             # <<<<<<<<<<<<<<
 *             elif stop > shape:
 *                 stop = shape
 */
          __pyx_v_stop = 0;

          /* "View.MemoryView":860
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        }

        /* "View.MemoryView":858
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
        goto __pyx_L17;
      }

      /* "View.MemoryView":862
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      __pyx_t_2 = ((__pyx_v_stop > __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":863
 *                     stop = 0
 *             elif stop > shape:
 *                 stop = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        __pyx_v_stop = __pyx_v_shape;

        /* "View.MemoryView":862
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      }
      __pyx_L17:;

      /* "View.MemoryView":857
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
      goto __pyx_L16;
    }

    /* "View.MemoryView":865
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":866
 *         else:
 *             if negative_step:
 *                 stop = -1             # <<<<<<<<<<<<<<
 *             else:
 *                 stop = shape
 */
        __pyx_v_stop = -1L;

        /* "View.MemoryView":865
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
        goto __pyx_L19;
      }

      /* "View.MemoryView":868
 *                 stop = -1
 *             else:
 *                 stop = shape             # <<<<<<<<<<<<<<
 * 
 *         if not have_step:
 */
      /*else*/ {
        __pyx_v_stop = __pyx_v_shape;
      }
      __pyx_L19:;
    }
    __pyx_L16:;

    /* "View.MemoryView":870
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    __pyx_t_2 = ((!(__pyx_v_have_step != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":871
 * 
 *         if not have_step:
 *             step = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_step = 1;

      /* "View.MemoryView":870
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    }

    /* "View.MemoryView":875
 * 
 *         with cython.cdivision(True):
 *             new_shape = (stop - start) // step             # <<<<<<<<<<<<<<
 * 
 *             if (stop - start) - step * new_shape:
 */
    __pyx_v_new_shape = ((__pyx_v_stop - __pyx_v_start) / __pyx_v_step);

    /* "View.MemoryView":877
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    __pyx_t_2 = (((__pyx_v_stop - __pyx_v_start) - (__pyx_v_step * __pyx_v_new_shape)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":878
 * 
 *             if (stop - start) - step * new_shape:
 *                 new_shape += 1             # <<<<<<<<<<<<<<
 * 
 *         if new_shape < 0:
 */
      __pyx_v_new_shape = (__pyx_v_new_shape + 1);

      /* "View.MemoryView":877
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    }

    /* "View.MemoryView":880
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    __pyx_t_2 = ((__pyx_v_new_shape < 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":881
 * 
 *         if new_shape < 0:
 *             new_shape = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_new_shape = 0;

      /* "View.MemoryView":880
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    }

    /* "View.MemoryView":884
 * 
 * 
 *         dst.strides[new_ndim] = stride * step             # <<<<<<<<<<<<<<
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset
 */
    (__pyx_v_dst->strides[__pyx_v_new_ndim]) = (__pyx_v_stride * __pyx_v_step);

    /* "View.MemoryView":885
 * 
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape             # <<<<<<<<<<<<<<
 *         dst.suboffsets[new_ndim] = suboffset
 * 
 */
    (__pyx_v_dst->shape[__pyx_v_new_ndim]) = __pyx_v_new_shape;

    /* "View.MemoryView":886
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_dst->suboffsets[__pyx_v_new_ndim]) = __pyx_v_suboffset;
  }
  __pyx_L3:;

  /* "View.MemoryView":889
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
  __pyx_t_2 = (((__pyx_v_suboffset_dim[0]) < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":890
 * 
 *     if suboffset_dim[0] < 0:
 *         dst.data += start * stride             # <<<<<<<<<<<<<<
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 */
    __pyx_v_dst->data = (__pyx_v_dst->data + (__pyx_v_start * __pyx_v_stride));

    /* "View.MemoryView":889
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
    goto __pyx_L23;
  }

  /* "View.MemoryView":892
 *         dst.data += start * stride
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride             # <<<<<<<<<<<<<<
 * 
 *     if suboffset >= 0:
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_suboffset_dim[0]);
    (__pyx_v_dst->suboffsets[__pyx_t_3]) = ((__pyx_v_dst->suboffsets[__pyx_t_3]) + (__pyx_v_start * __pyx_v_stride));
  }
  __pyx_L23:;

  /* "View.MemoryView":894
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":895
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
    __pyx_t_2 = ((!(__pyx_v_is_slice != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":896
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
      __pyx_t_2 = ((__pyx_v_new_ndim == 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":897
 *         if not is_slice:
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 */
        __pyx_v_dst->data = ((((char **)__pyx_v_dst->data)[0]) + __pyx_v_suboffset);

        /* "View.MemoryView":896
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
        goto __pyx_L26;
      }

      /* "View.MemoryView":899
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "             # <<<<<<<<<<<<<<
 *                                      "must be indexed and not sliced", dim)
 *         else:
 */
      /*else*/ {

        /* "View.MemoryView":900
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 *                                      "must be indexed and not sliced", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             suboffset_dim[0] = new_ndim
 */
        __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"All dimensions preceding dimension %d must be indexed and not sliced"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 899, __pyx_L1_error)
      }
      __pyx_L26:;

      /* "View.MemoryView":895
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
      goto __pyx_L25;
    }

    /* "View.MemoryView":902
 *                                      "must be indexed and not sliced", dim)
 *         else:
 *             suboffset_dim[0] = new_ndim             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    /*else*/ {
      (__pyx_v_suboffset_dim[0]) = __pyx_v_new_ndim;
    }
    __pyx_L25:;

    /* "View.MemoryView":894
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  }

  /* "View.MemoryView":904
 *             suboffset_dim[0] = new_ndim
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":807
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.slice_memviewslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":910
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

static char *__pyx_pybuffer_index(Py_buffer *__pyx_v_view, char *__pyx_v_bufp, Py_ssize_t __pyx_v_index, Py_ssize_t __pyx_v_dim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_suboffset;
  Py_ssize_t __pyx_v_itemsize;
  char *__pyx_v_resultp;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pybuffer_index", 0);

  /* "View.MemoryView":912
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t itemsize = view.itemsize
 *     cdef char *resultp
 */
  __pyx_v_suboffset = -1L;

  /* "View.MemoryView":913
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 *     cdef Py_ssize_t itemsize = view.itemsize             # <<<<<<<<<<<<<<
 *     cdef char *resultp
 * 
 */
  __pyx_t_1 = __pyx_v_view->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":916
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
  __pyx_t_2 = ((__pyx_v_view->ndim == 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":917
 * 
 *     if view.ndim == 0:
 *         shape = view.len / itemsize             # <<<<<<<<<<<<<<
 *         stride = itemsize
 *     else:
 */
    if (unlikely(__pyx_v_itemsize == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(2, 917, __pyx_L1_error)
    }
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_view->len))) {
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
      __PYX_ERR(2, 917, __pyx_L1_error)
    }
    __pyx_v_shape = __Pyx_div_Py_ssize_t(__pyx_v_view->len, __pyx_v_itemsize);

    /* "View.MemoryView":918
 *     if view.ndim == 0:
 *         shape = view.len / itemsize
 *         stride = itemsize             # <<<<<<<<<<<<<<
 *     else:
 *         shape = view.shape[dim]
 */
    __pyx_v_stride = __pyx_v_itemsize;

    /* "View.MemoryView":916
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":920
 *         stride = itemsize
 *     else:
 *         shape = view.shape[dim]             # <<<<<<<<<<<<<<
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 */
  /*else*/ {
    __pyx_v_shape = (__pyx_v_view->shape[__pyx_v_dim]);

    /* "View.MemoryView":921
 *     else:
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]             # <<<<<<<<<<<<<<
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]
 */
    __pyx_v_stride = (__pyx_v_view->strides[__pyx_v_dim]);

    /* "View.MemoryView":922
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    __pyx_t_2 = ((__pyx_v_view->suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":923
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]             # <<<<<<<<<<<<<<
 * 
 *     if index < 0:
 */
      __pyx_v_suboffset = (__pyx_v_view->suboffsets[__pyx_v_dim]);

      /* "View.MemoryView":922
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":925
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  __pyx_t_2 = ((__pyx_v_index < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":926
 * 
 *     if index < 0:
 *         index += view.shape[dim]             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 */
    __pyx_v_index = (__pyx_v_index + (__pyx_v_view->shape[__pyx_v_dim]));

    /* "View.MemoryView":927
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    __pyx_t_2 = ((__pyx_v_index < 0) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":928
 *         index += view.shape[dim]
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     if index >= shape:
 */
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 928, __pyx_L1_error)

      /* "View.MemoryView":927
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    }

    /* "View.MemoryView":925
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  }

  /* "View.MemoryView":930
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_index >= __pyx_v_shape) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":931
 * 
 *     if index >= shape:
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     resultp = bufp + index * stride
 */
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 931, __pyx_L1_error)

    /* "View.MemoryView":930
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  }

  /* "View.MemoryView":933
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     resultp = bufp + index * stride             # <<<<<<<<<<<<<<
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset
 */
  __pyx_v_resultp = (__pyx_v_bufp + (__pyx_v_index * __pyx_v_stride));

  /* "View.MemoryView":934
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":935
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset             # <<<<<<<<<<<<<<
 * 
 *     return resultp
 */
    __pyx_v_resultp = ((((char **)__pyx_v_resultp)[0]) + __pyx_v_suboffset);

    /* "View.MemoryView":934
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  }

  /* "View.MemoryView":937
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 *     return resultp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_resultp;
  goto __pyx_L0;

  /* "View.MemoryView":910
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.pybuffer_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":943
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

static int __pyx_memslice_transpose(__Pyx_memviewslice *__pyx_v_memslice) {
  int __pyx_v_ndim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  int __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":944
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:
 *     cdef int ndim = memslice.memview.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 */
  __pyx_t_1 = __pyx_v_memslice->memview->view.ndim;
  __pyx_v_ndim = __pyx_t_1;

  /* "View.MemoryView":946
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 *     cdef Py_ssize_t *shape = memslice.shape             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t *strides = memslice.strides
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->shape;
  __pyx_v_shape = __pyx_t_2;

  /* "View.MemoryView":947
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 *     cdef Py_ssize_t *strides = memslice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->strides;
  __pyx_v_strides = __pyx_t_2;

  /* "View.MemoryView":951
 * 
 *     cdef int i, j
 *     for i in range(ndim / 2):             # <<<<<<<<<<<<<<
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 */
  __pyx_t_3 = __Pyx_div_long(__pyx_v_ndim, 2);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_4; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":952
 *     cdef int i, j
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i             # <<<<<<<<<<<<<<
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]
 */
    __pyx_v_j = ((__pyx_v_ndim - 1) - __pyx_v_i);

    /* "View.MemoryView":953
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]             # <<<<<<<<<<<<<<
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 */
    __pyx_t_5 = (__pyx_v_strides[__pyx_v_j]);
    __pyx_t_6 = (__pyx_v_strides[__pyx_v_i]);
    (__pyx_v_strides[__pyx_v_i]) = __pyx_t_5;
    (__pyx_v_strides[__pyx_v_j]) = __pyx_t_6;

    /* "View.MemoryView":954
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]             # <<<<<<<<<<<<<<
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 */
    __pyx_t_6 = (__pyx_v_shape[__pyx_v_j]);
    __pyx_t_5 = (__pyx_v_shape[__pyx_v_i]);
    (__pyx_v_shape[__pyx_v_i]) = __pyx_t_6;
    (__pyx_v_shape[__pyx_v_j]) = __pyx_t_5;

    /* "View.MemoryView":956
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_i]) >= 0) != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_j]) >= 0) != 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_7) {

      /* "View.MemoryView":957
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")             # <<<<<<<<<<<<<<
 * 
 *     return 1
 */
      __pyx_t_9 = __pyx_memoryview_err(__pyx_builtin_ValueError, ((char *)"Cannot transpose memoryview with indirect dimensions")); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 957, __pyx_L1_error)

      /* "View.MemoryView":956
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    }
  }

  /* "View.MemoryView":959
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "View.MemoryView":943
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.transpose_memslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":976
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

/* Python wrapper */
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":977
 * 
 *     def __dealloc__(self):
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __PYX_XDEC_MEMVIEW((&__pyx_v_self->from_slice), 1);

  /* "View.MemoryView":976
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":979
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":980
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_object_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":981
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)             # <<<<<<<<<<<<<<
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_self->to_object_func(__pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 981, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":980
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  }

  /* "View.MemoryView":983
 *             return self.to_object_func(itemp)
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_memoryview_convert_item_to_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":979
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":985
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":986
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_dtype_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":987
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)             # <<<<<<<<<<<<<<
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)
 */
    __pyx_t_2 = __pyx_v_self->to_dtype_func(__pyx_v_itemp, __pyx_v_value); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(2, 987, __pyx_L1_error)

    /* "View.MemoryView":986
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":989
 *             self.to_dtype_func(itemp, value)
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  /*else*/ {
    __pyx_t_3 = __pyx_memoryview_assign_item_from_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 989, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":985
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":992
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":993
 *     @property
 *     def base(self):
 *         return self.from_object             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->from_object);
  __pyx_r = __pyx_v_self->from_object;
  goto __pyx_L0;

  /* "View.MemoryView":992
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice___reduce_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__92, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice_2__setstate_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__93, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice __pyx_v_memviewslice, int __pyx_v_ndim, PyObject *(*__pyx_v_to_object_func)(char *), int (*__pyx_v_to_dtype_func)(char *, PyObject *), int __pyx_v_dtype_is_object) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_TypeInfo *__pyx_t_4;
  Py_buffer __pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  Py_ssize_t *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_fromslice", 0);

  /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_t_1 = ((((PyObject *)__pyx_v_memviewslice.memview) == Py_None) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1008
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  }

  /* "View.MemoryView":1013
 * 
 * 
 *     result = _memoryviewslice(None, 0, dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     result.from_slice = memviewslice
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryviewslice_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":1015
 *     result = _memoryviewslice(None, 0, dtype_is_object)
 * 
 *     result.from_slice = memviewslice             # <<<<<<<<<<<<<<
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 */
  __pyx_v_result->from_slice = __pyx_v_memviewslice;

  /* "View.MemoryView":1016
 * 
 *     result.from_slice = memviewslice
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)             # <<<<<<<<<<<<<<
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 */
  __PYX_INC_MEMVIEW((&__pyx_v_memviewslice), 1);

  /* "View.MemoryView":1018
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base             # <<<<<<<<<<<<<<
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_memviewslice.memview), __pyx_n_s_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_result->from_object);
  __Pyx_DECREF(__pyx_v_result->from_object);
  __pyx_v_result->from_object = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "View.MemoryView":1019
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 *     result.typeinfo = memviewslice.memview.typeinfo             # <<<<<<<<<<<<<<
 * 
 *     result.view = memviewslice.memview.view
 */
  __pyx_t_4 = __pyx_v_memviewslice.memview->typeinfo;
  __pyx_v_result->__pyx_base.typeinfo = __pyx_t_4;

  /* "View.MemoryView":1021
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 *     result.view = memviewslice.memview.view             # <<<<<<<<<<<<<<
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 */
  __pyx_t_5 = __pyx_v_memviewslice.memview->view;
  __pyx_v_result->__pyx_base.view = __pyx_t_5;

  /* "View.MemoryView":1022
 * 
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data             # <<<<<<<<<<<<<<
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 */
  __pyx_v_result->__pyx_base.view.buf = ((void *)__pyx_v_memviewslice.data);

  /* "View.MemoryView":1023
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim             # <<<<<<<<<<<<<<
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)
 */
  __pyx_v_result->__pyx_base.view.ndim = __pyx_v_ndim;

  /* "View.MemoryView":1024
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None             # <<<<<<<<<<<<<<
 *     Py_INCREF(Py_None)
 * 
 */
  ((Py_buffer *)(&__pyx_v_result->__pyx_base.view))->obj = Py_None;

  /* "View.MemoryView":1025
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 */
  Py_INCREF(Py_None);

  /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
  __pyx_t_1 = ((((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1028
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 *         result.flags = PyBUF_RECORDS             # <<<<<<<<<<<<<<
 *     else:
 *         result.flags = PyBUF_RECORDS_RO
 */
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS;

    /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":1030
 *         result.flags = PyBUF_RECORDS
 *     else:
 *         result.flags = PyBUF_RECORDS_RO             # <<<<<<<<<<<<<<
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 */
  /*else*/ {
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS_RO;
  }
  __pyx_L4:;

  /* "View.MemoryView":1032
 *         result.flags = PyBUF_RECORDS_RO
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape             # <<<<<<<<<<<<<<
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides
 * 
 */
  __pyx_v_result->__pyx_base.view.shape = ((Py_ssize_t *)__pyx_v_result->from_slice.shape);

  /* "View.MemoryView":1033
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_result->__pyx_base.view.strides = ((Py_ssize_t *)__pyx_v_result->from_slice.strides);

  /* "View.MemoryView":1036
 * 
 * 
 *     result.view.suboffsets = NULL             # <<<<<<<<<<<<<<
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 */
  __pyx_v_result->__pyx_base.view.suboffsets = NULL;

  /* "View.MemoryView":1037
 * 
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 */
  __pyx_t_7 = (__pyx_v_result->from_slice.suboffsets + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->from_slice.suboffsets; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_v_suboffset = (__pyx_t_6[0]);

    /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    __pyx_t_1 = ((__pyx_v_suboffset >= 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":1039
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_result->__pyx_base.view.suboffsets = ((Py_ssize_t *)__pyx_v_result->from_slice.suboffsets);

      /* "View.MemoryView":1040
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break             # <<<<<<<<<<<<<<
 * 
 *     result.view.len = result.view.itemsize
 */
      goto __pyx_L6_break;

      /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    }
  }
  __pyx_L6_break:;

  /* "View.MemoryView":1042
 *             break
 * 
 *     result.view.len = result.view.itemsize             # <<<<<<<<<<<<<<
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length
 */
  __pyx_t_9 = __pyx_v_result->__pyx_base.view.itemsize;
  __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

  /* "View.MemoryView":1043
 * 
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         result.view.len *= length
 * 
 */
  __pyx_t_7 = (__pyx_v_result->__pyx_base.view.shape + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->__pyx_base.view.shape; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_6[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "View.MemoryView":1044
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length             # <<<<<<<<<<<<<<
 * 
 *     result.to_object_func = to_object_func
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_result->__pyx_base.view.len); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_InPlaceMultiply(__pyx_t_2, __pyx_v_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 1044, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result->__pyx_base.view.len = __pyx_t_9;
  }

  /* "View.MemoryView":1046
 *         result.view.len *= length
 * 
 *     result.to_object_func = to_object_func             # <<<<<<<<<<<<<<
 *     result.to_dtype_func = to_dtype_func
 * 
 */
  __pyx_v_result->to_object_func = __pyx_v_to_object_func;

  /* "View.MemoryView":1047
 * 
 *     result.to_object_func = to_object_func
 *     result.to_dtype_func = to_dtype_func             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  __pyx_v_result->to_dtype_func = __pyx_v_to_dtype_func;

  /* "View.MemoryView":1049
 *     result.to_dtype_func = to_dtype_func
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_fromslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_mslice) {
  struct __pyx_memoryviewslice_obj *__pyx_v_obj = 0;
  __Pyx_memviewslice *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_slice_from_memview", 0);

  /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1056
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview             # <<<<<<<<<<<<<<
 *         return &obj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(2, 1056, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_obj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":1057
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview
 *         return &obj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, mslice)
 */
    __pyx_r = (&__pyx_v_obj->from_slice);
    goto __pyx_L0;

    /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  }

  /* "View.MemoryView":1059
 *         return &obj.from_slice
 *     else:
 *         slice_copy(memview, mslice)             # <<<<<<<<<<<<<<
 *         return mslice
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, __pyx_v_mslice);

    /* "View.MemoryView":1060
 *     else:
 *         slice_copy(memview, mslice)
 *         return mslice             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_slice_copy')
 */
    __pyx_r = __pyx_v_mslice;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.get_slice_from_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_dst) {
  int __pyx_v_dim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  Py_ssize_t *__pyx_v_suboffsets;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  __Pyx_RefNannySetupContext("slice_copy", 0);

  /* "View.MemoryView":1067
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 * 
 *     shape = memview.view.shape             # <<<<<<<<<<<<<<
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets
 */
  __pyx_t_1 = __pyx_v_memview->view.shape;
  __pyx_v_shape = __pyx_t_1;

  /* "View.MemoryView":1068
 * 
 *     shape = memview.view.shape
 *     strides = memview.view.strides             # <<<<<<<<<<<<<<
 *     suboffsets = memview.view.suboffsets
 * 
 */
  __pyx_t_1 = __pyx_v_memview->view.strides;
  __pyx_v_strides = __pyx_t_1;

  /* "View.MemoryView":1069
 *     shape = memview.view.shape
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets             # <<<<<<<<<<<<<<
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 */
  __pyx_t_1 = __pyx_v_memview->view.suboffsets;
  __pyx_v_suboffsets = __pyx_t_1;

  /* "View.MemoryView":1071
 *     suboffsets = memview.view.suboffsets
 * 
 *     dst.memview = <__pyx_memoryview *> memview             # <<<<<<<<<<<<<<
 *     dst.data = <char *> memview.view.buf
 * 
 */
  __pyx_v_dst->memview = ((struct __pyx_memoryview_obj *)__pyx_v_memview);

  /* "View.MemoryView":1072
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 *     dst.data = <char *> memview.view.buf             # <<<<<<<<<<<<<<
 * 
 *     for dim in range(memview.view.ndim):
 */
  __pyx_v_dst->data = ((char *)__pyx_v_memview->view.buf);

  /* "View.MemoryView":1074
 *     dst.data = <char *> memview.view.buf
 * 
 *     for dim in range(memview.view.ndim):             # <<<<<<<<<<<<<<
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 */
  __pyx_t_2 = __pyx_v_memview->view.ndim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_dim = __pyx_t_4;

    /* "View.MemoryView":1075
 * 
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]             # <<<<<<<<<<<<<<
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 */
    (__pyx_v_dst->shape[__pyx_v_dim]) = (__pyx_v_shape[__pyx_v_dim]);

    /* "View.MemoryView":1076
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]             # <<<<<<<<<<<<<<
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 * 
 */
    (__pyx_v_dst->strides[__pyx_v_dim]) = (__pyx_v_strides[__pyx_v_dim]);

    /* "View.MemoryView":1077
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object')
 */
    if ((__pyx_v_suboffsets != 0)) {
      __pyx_t_5 = (__pyx_v_suboffsets[__pyx_v_dim]);
    } else {
      __pyx_t_5 = -1L;
    }
    (__pyx_v_dst->suboffsets[__pyx_v_dim]) = __pyx_t_5;
  }

  /* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *__pyx_v_memview) {
  __Pyx_memviewslice __pyx_v_memviewslice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy", 0);

  /* "View.MemoryView":1083
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)             # <<<<<<<<<<<<<<
 *     return memoryview_copy_from_slice(memview, &memviewslice)
 * 
 */
  __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_memviewslice));

  /* "View.MemoryView":1084
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)
 *     return memoryview_copy_from_slice(memview, &memviewslice)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_memoryview_copy_object_from_slice(__pyx_v_memview, (&__pyx_v_memviewslice)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_memviewslice) {
  PyObject *(*__pyx_v_to_object_func)(char *);
  int (*__pyx_v_to_dtype_func)(char *, PyObject *);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *(*__pyx_t_3)(char *);
  int (*__pyx_t_4)(char *, PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy_from_slice", 0);

  /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1095
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func             # <<<<<<<<<<<<<<
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 */
    __pyx_t_3 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_object_func;
    __pyx_v_to_object_func = __pyx_t_3;

    /* "View.MemoryView":1096
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func             # <<<<<<<<<<<<<<
 *     else:
 *         to_object_func = NULL
 */
    __pyx_t_4 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_dtype_func;
    __pyx_v_to_dtype_func = __pyx_t_4;

    /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1098
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 *         to_object_func = NULL             # <<<<<<<<<<<<<<
 *         to_dtype_func = NULL
 * 
 */
  /*else*/ {
    __pyx_v_to_object_func = NULL;

    /* "View.MemoryView":1099
 *     else:
 *         to_object_func = NULL
 *         to_dtype_func = NULL             # <<<<<<<<<<<<<<
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 */
    __pyx_v_to_dtype_func = NULL;
  }
  __pyx_L3:;

  /* "View.MemoryView":1101
 *         to_dtype_func = NULL
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,             # <<<<<<<<<<<<<<
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "View.MemoryView":1103
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_memoryview_fromslice((__pyx_v_memviewslice[0]), __pyx_v_memview->view.ndim, __pyx_v_to_object_func, __pyx_v_to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_from_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

static Py_ssize_t abs_py_ssize_t(Py_ssize_t __pyx_v_arg) {
  Py_ssize_t __pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":1110
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_arg < 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1111
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:
 *         return -arg             # <<<<<<<<<<<<<<
 *     else:
 *         return arg
 */
    __pyx_r = (-__pyx_v_arg);
    goto __pyx_L0;

    /* "View.MemoryView":1110
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  }

  /* "View.MemoryView":1113
 *         return -arg
 *     else:
 *         return arg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_get_best_slice_order')
 */
  /*else*/ {
    __pyx_r = __pyx_v_arg;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1116
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

static char __pyx_get_best_slice_order(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim) {
  int __pyx_v_i;
  Py_ssize_t __pyx_v_c_stride;
  Py_ssize_t __pyx_v_f_stride;
  char __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1121
 *     """
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t f_stride = 0
 * 
 */
  __pyx_v_c_stride = 0;

  /* "View.MemoryView":1122
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0
 *     cdef Py_ssize_t f_stride = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_f_stride = 0;

  /* "View.MemoryView":1124
 *     cdef Py_ssize_t f_stride = 0
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1125
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1126
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_c_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1127
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
      goto __pyx_L4_break;

      /* "View.MemoryView":1125
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L4_break:;

  /* "View.MemoryView":1129
 *             break
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 */
  __pyx_t_1 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1130
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1131
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_f_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1132
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 */
      goto __pyx_L7_break;

      /* "View.MemoryView":1130
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L7_break:;

  /* "View.MemoryView":1134
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  __pyx_t_2 = ((abs_py_ssize_t(__pyx_v_c_stride) <= abs_py_ssize_t(__pyx_v_f_stride)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1135
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 *         return 'C'             # <<<<<<<<<<<<<<
 *     else:
 *         return 'F'
 */
    __pyx_r = 'C';
    goto __pyx_L0;

    /* "View.MemoryView":1134
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  }

  /* "View.MemoryView":1137
 *         return 'C'
 *     else:
 *         return 'F'             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  /*else*/ {
    __pyx_r = 'F';
    goto __pyx_L0;
  }

  /* "View.MemoryView":1116
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1140
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

static void _copy_strided_to_strided(char *__pyx_v_src_data, Py_ssize_t *__pyx_v_src_strides, char *__pyx_v_dst_data, Py_ssize_t *__pyx_v_dst_strides, Py_ssize_t *__pyx_v_src_shape, Py_ssize_t *__pyx_v_dst_shape, int __pyx_v_ndim, size_t __pyx_v_itemsize) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED Py_ssize_t __pyx_v_src_extent;
  Py_ssize_t __pyx_v_dst_extent;
  Py_ssize_t __pyx_v_src_stride;
  Py_ssize_t __pyx_v_dst_stride;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;

  /* "View.MemoryView":1147
 * 
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 */
  __pyx_v_src_extent = (__pyx_v_src_shape[0]);

  /* "View.MemoryView":1148
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 */
  __pyx_v_dst_extent = (__pyx_v_dst_shape[0]);

  /* "View.MemoryView":1149
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 */
  __pyx_v_src_stride = (__pyx_v_src_strides[0]);

  /* "View.MemoryView":1150
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_dst_stride = (__pyx_v_dst_strides[0]);

  /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    __pyx_t_2 = ((__pyx_v_src_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_dst_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }

    /* "View.MemoryView":1154
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):             # <<<<<<<<<<<<<<
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 */
    __pyx_t_2 = (((size_t)__pyx_v_src_stride) == __pyx_v_itemsize);
    if (__pyx_t_2) {
      __pyx_t_2 = (__pyx_v_itemsize == ((size_t)__pyx_v_dst_stride));
    }
    __pyx_t_3 = (__pyx_t_2 != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L5_bool_binop_done:;

    /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    if (__pyx_t_1) {

      /* "View.MemoryView":1155
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)             # <<<<<<<<<<<<<<
 *        else:
 *            for i in range(dst_extent):
 */
      (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, (__pyx_v_itemsize * __pyx_v_dst_extent)));

      /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
      goto __pyx_L4;
    }

    /* "View.MemoryView":1157
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 *            for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 */
    /*else*/ {
      __pyx_t_4 = __pyx_v_dst_extent;
      __pyx_t_5 = __pyx_t_4;
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
        __pyx_v_i = __pyx_t_6;

        /* "View.MemoryView":1158
 *        else:
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)             # <<<<<<<<<<<<<<
 *                src_data += src_stride
 *                dst_data += dst_stride
 */
        (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, __pyx_v_itemsize));

        /* "View.MemoryView":1159
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride             # <<<<<<<<<<<<<<
 *                dst_data += dst_stride
 *     else:
 */
        __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

        /* "View.MemoryView":1160
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 *                dst_data += dst_stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(dst_extent):
 */
        __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
      }
    }
    __pyx_L4:;

    /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1162
 *                dst_data += dst_stride
 *     else:
 *         for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *             _copy_strided_to_strided(src_data, src_strides + 1,
 *                                      dst_data, dst_strides + 1,
 */
  /*else*/ {
    __pyx_t_4 = __pyx_v_dst_extent;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "View.MemoryView":1163
 *     else:
 *         for i in range(dst_extent):
 *             _copy_strided_to_strided(src_data, src_strides + 1,             # <<<<<<<<<<<<<<
 *                                      dst_data, dst_strides + 1,
 *                                      src_shape + 1, dst_shape + 1,
 */
      _copy_strided_to_strided(__pyx_v_src_data, (__pyx_v_src_strides + 1), __pyx_v_dst_data, (__pyx_v_dst_strides + 1), (__pyx_v_src_shape + 1), (__pyx_v_dst_shape + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize);

      /* "View.MemoryView":1167
 *                                      src_shape + 1, dst_shape + 1,
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride             # <<<<<<<<<<<<<<
 *             dst_data += dst_stride
 * 
 */
      __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

      /* "View.MemoryView":1168
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride
 *             dst_data += dst_stride             # <<<<<<<<<<<<<<
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,
 */
      __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1140
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

  /* function exit code */
}

/* "View.MemoryView":1170
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

static void copy_strided_to_strided(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize) {

  /* "View.MemoryView":1173
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 *     _copy_strided_to_strided(src.data, src.strides, dst.data, dst.strides,             # <<<<<<<<<<<<<<
 *                              src.shape, dst.shape, ndim, itemsize)
 * 
 */
  _copy_strided_to_strided(__pyx_v_src->data, __pyx_v_src->strides, __pyx_v_dst->data, __pyx_v_dst->strides, __pyx_v_src->shape, __pyx_v_dst->shape, __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1170
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1177
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *__pyx_v_src, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_r;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;

  /* "View.MemoryView":1179
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     for shape in src.shape[:ndim]:
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_size = __pyx_t_1;

  /* "View.MemoryView":1181
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 * 
 *     for shape in src.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         size *= shape
 * 
 */
  __pyx_t_3 = (__pyx_v_src->shape + __pyx_v_ndim);
  for (__pyx_t_4 = __pyx_v_src->shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_shape = (__pyx_t_2[0]);

    /* "View.MemoryView":1182
 * 
 *     for shape in src.shape[:ndim]:
 *         size *= shape             # <<<<<<<<<<<<<<
 * 
 *     return size
 */
    __pyx_v_size = (__pyx_v_size * __pyx_v_shape);
  }

  /* "View.MemoryView":1184
 *         size *= shape
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_fill_contig_strides_array')
 */
  __pyx_r = __pyx_v_size;
  goto __pyx_L0;

  /* "View.MemoryView":1177
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1187
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, Py_ssize_t __pyx_v_stride, int __pyx_v_ndim, char __pyx_v_order) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1196
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
  __pyx_t_1 = ((__pyx_v_order == 'F') != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1197
 * 
 *     if order == 'F':
 *         for idx in range(ndim):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
    __pyx_t_2 = __pyx_v_ndim;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_idx = __pyx_t_4;

      /* "View.MemoryView":1198
 *     if order == 'F':
 *         for idx in range(ndim):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 *     else:
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1199
 *         for idx in range(ndim):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }

    /* "View.MemoryView":1196
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1201
 *             stride *= shape[idx]
 *     else:
 *         for idx in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
  /*else*/ {
    for (__pyx_t_2 = (__pyx_v_ndim - 1); __pyx_t_2 > -1; __pyx_t_2-=1) {
      __pyx_v_idx = __pyx_t_2;

      /* "View.MemoryView":1202
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 * 
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1203
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 * 
 *     return stride
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1205
 *             stride *= shape[idx]
 * 
 *     return stride             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 */
  __pyx_r = __pyx_v_stride;
  goto __pyx_L0;

  /* "View.MemoryView":1187
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1208
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_tmpslice, char __pyx_v_order, int __pyx_v_ndim) {
  int __pyx_v_i;
  void *__pyx_v_result;
  size_t __pyx_v_itemsize;
  size_t __pyx_v_size;
  void *__pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  struct __pyx_memoryview_obj *__pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":1219
 *     cdef void *result
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1220
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef size_t size = slice_get_size(src, ndim)             # <<<<<<<<<<<<<<
 * 
 *     result = malloc(size)
 */
  __pyx_v_size = __pyx_memoryview_slice_get_size(__pyx_v_src, __pyx_v_ndim);

  /* "View.MemoryView":1222
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 *     result = malloc(size)             # <<<<<<<<<<<<<<
 *     if not result:
 *         _err(MemoryError, NULL)
 */
  __pyx_v_result = malloc(__pyx_v_size);

  /* "View.MemoryView":1223
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_result != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1224
 *     result = malloc(size)
 *     if not result:
 *         _err(MemoryError, NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_err(__pyx_builtin_MemoryError, NULL); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 1224, __pyx_L1_error)

    /* "View.MemoryView":1223
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  }

  /* "View.MemoryView":1227
 * 
 * 
 *     tmpslice.data = <char *> result             # <<<<<<<<<<<<<<
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 */
  __pyx_v_tmpslice->data = ((char *)__pyx_v_result);

  /* "View.MemoryView":1228
 * 
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview             # <<<<<<<<<<<<<<
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 */
  __pyx_t_4 = __pyx_v_src->memview;
  __pyx_v_tmpslice->memview = __pyx_t_4;

  /* "View.MemoryView":1229
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1230
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]             # <<<<<<<<<<<<<<
 *         tmpslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_tmpslice->shape[__pyx_v_i]) = (__pyx_v_src->shape[__pyx_v_i]);

    /* "View.MemoryView":1231
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,
 */
    (__pyx_v_tmpslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1233
 *         tmpslice.suboffsets[i] = -1
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,             # <<<<<<<<<<<<<<
 *                               ndim, order)
 * 
 */
  (void)(__pyx_fill_contig_strides_array((&(__pyx_v_tmpslice->shape[0])), (&(__pyx_v_tmpslice->strides[0])), __pyx_v_itemsize, __pyx_v_ndim, __pyx_v_order));

  /* "View.MemoryView":1237
 * 
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1238
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    __pyx_t_2 = (((__pyx_v_tmpslice->shape[__pyx_v_i]) == 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1239
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     if slice_is_contig(src[0], order, ndim):
 */
      (__pyx_v_tmpslice->strides[__pyx_v_i]) = 0;

      /* "View.MemoryView":1238
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    }
  }

  /* "View.MemoryView":1241
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
  __pyx_t_2 = (__pyx_memviewslice_is_contig((__pyx_v_src[0]), __pyx_v_order, __pyx_v_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1242
 * 
 *     if slice_is_contig(src[0], order, ndim):
 *         memcpy(result, src.data, size)             # <<<<<<<<<<<<<<
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 */
    (void)(memcpy(__pyx_v_result, __pyx_v_src->data, __pyx_v_size));

    /* "View.MemoryView":1241
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":1244
 *         memcpy(result, src.data, size)
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  /*else*/ {
    copy_strided_to_strided(__pyx_v_src, __pyx_v_tmpslice, __pyx_v_ndim, __pyx_v_itemsize);
  }
  __pyx_L9:;

  /* "View.MemoryView":1246
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":1208
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.copy_data_to_temp", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1251
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

static int __pyx_memoryview_err_extents(int __pyx_v_i, Py_ssize_t __pyx_v_extent1, Py_ssize_t __pyx_v_extent2) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_extents", 0);

  /* "View.MemoryView":1254
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 *                                                         (i, extent1, extent2))             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_dim')
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_extent1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_extent2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;

  /* "View.MemoryView":1253
 * cdef int _err_extents(int i, Py_ssize_t extent1,
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %             # <<<<<<<<<<<<<<
 *                                                         (i, extent1, extent2))
 * 
 */
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(2, 1253, __pyx_L1_error)

  /* "View.MemoryView":1251
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1257
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

static int __pyx_memoryview_err_dim(PyObject *__pyx_v_error, char *__pyx_v_msg, int __pyx_v_dim) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_dim", 0);
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1258
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:
 *     raise error(msg.decode('ascii') % dim)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err')
 */
  __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyUnicode_Format(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_v_error);
  __pyx_t_3 = __pyx_v_error; __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 1258, __pyx_L1_error)

  /* "View.MemoryView":1257
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_dim", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1261
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

static int __pyx_memoryview_err(PyObject *__pyx_v_error, char *__pyx_v_msg) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err", 0);
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1262
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_msg != NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":1263
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))             # <<<<<<<<<<<<<<
 *     else:
 *         raise error
 */
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_error);
    __pyx_t_4 = __pyx_v_error; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 1263, __pyx_L1_error)

    /* "View.MemoryView":1262
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  }

  /* "View.MemoryView":1265
 *         raise error(msg.decode('ascii'))
 *     else:
 *         raise error             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_contents')
 */
  /*else*/ {
    __Pyx_Raise(__pyx_v_error, 0, 0, 0);
    __PYX_ERR(2, 1265, __pyx_L1_error)
  }

  /* "View.MemoryView":1261
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView._err", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1268
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

static int __pyx_memoryview_copy_contents(__Pyx_memviewslice __pyx_v_src, __Pyx_memviewslice __pyx_v_dst, int __pyx_v_src_ndim, int __pyx_v_dst_ndim, int __pyx_v_dtype_is_object) {
  void *__pyx_v_tmpdata;
  size_t __pyx_v_itemsize;
  int __pyx_v_i;
  char __pyx_v_order;
  int __pyx_v_broadcasting;
  int __pyx_v_direct_copy;
  __Pyx_memviewslice __pyx_v_tmp;
  int __pyx_v_ndim;
  int __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":1276
 *     Check for overlapping memory and verify the shapes.
 *     """
 *     cdef void *tmpdata = NULL             # <<<<<<<<<<<<<<
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 */
  __pyx_v_tmpdata = NULL;

  /* "View.MemoryView":1277
 *     """
 *     cdef void *tmpdata = NULL
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 */
  __pyx_t_1 = __pyx_v_src.memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1279
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)             # <<<<<<<<<<<<<<
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False
 */
  __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_src), __pyx_v_src_ndim);

  /* "View.MemoryView":1280
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False             # <<<<<<<<<<<<<<
 *     cdef bint direct_copy = False
 *     cdef __Pyx_memviewslice tmp
 */
  __pyx_v_broadcasting = 0;

  /* "View.MemoryView":1281
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False             # <<<<<<<<<<<<<<
 *     cdef __Pyx_memviewslice tmp
 * 
 */
  __pyx_v_direct_copy = 0;

  /* "View.MemoryView":1284
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
  __pyx_t_2 = ((__pyx_v_src_ndim < __pyx_v_dst_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1285
 * 
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_src), __pyx_v_src_ndim, __pyx_v_dst_ndim);

    /* "View.MemoryView":1284
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1286
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_dst_ndim < __pyx_v_src_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1287
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)             # <<<<<<<<<<<<<<
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_dst), __pyx_v_dst_ndim, __pyx_v_src_ndim);

    /* "View.MemoryView":1286
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":1289
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
  __pyx_t_3 = __pyx_v_dst_ndim;
  __pyx_t_4 = __pyx_v_src_ndim;
  if (((__pyx_t_3 > __pyx_t_4) != 0)) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_v_ndim = __pyx_t_5;

  /* "View.MemoryView":1291
 *     cdef int ndim = max(src_ndim, dst_ndim)
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 */
  __pyx_t_5 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_5;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1292
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) != (__pyx_v_dst.shape[__pyx_v_i])) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1293
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
      __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) == 1) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":1294
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 *                 broadcasting = True             # <<<<<<<<<<<<<<
 *                 src.strides[i] = 0
 *             else:
 */
        __pyx_v_broadcasting = 1;

        /* "View.MemoryView":1295
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 *                 src.strides[i] = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 */
        (__pyx_v_src.strides[__pyx_v_i]) = 0;

        /* "View.MemoryView":1293
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":1297
 *                 src.strides[i] = 0
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])             # <<<<<<<<<<<<<<
 * 
 *         if src.suboffsets[i] >= 0:
 */
      /*else*/ {
        __pyx_t_6 = __pyx_memoryview_err_extents(__pyx_v_i, (__pyx_v_dst.shape[__pyx_v_i]), (__pyx_v_src.shape[__pyx_v_i])); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 1297, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":1292
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    }

    /* "View.MemoryView":1299
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    __pyx_t_2 = (((__pyx_v_src.suboffsets[__pyx_v_i]) >= 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1300
 * 
 *         if src.suboffsets[i] >= 0:
 *             _err_dim(ValueError, "Dimension %d is not direct", i)             # <<<<<<<<<<<<<<
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 */
      __pyx_t_6 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Dimension %d is not direct"), __pyx_v_i); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 1300, __pyx_L1_error)

      /* "View.MemoryView":1299
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    }
  }

  /* "View.MemoryView":1302
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  __pyx_t_2 = (__pyx_slices_overlap((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1304
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    __pyx_t_2 = ((!(__pyx_memviewslice_is_contig(__pyx_v_src, __pyx_v_order, __pyx_v_ndim) != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1305
 * 
 *         if not slice_is_contig(src, order, ndim):
 *             order = get_best_order(&dst, ndim)             # <<<<<<<<<<<<<<
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 */
      __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim);

      /* "View.MemoryView":1304
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    }

    /* "View.MemoryView":1307
 *             order = get_best_order(&dst, ndim)
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)             # <<<<<<<<<<<<<<
 *         src = tmp
 * 
 */
    __pyx_t_7 = __pyx_memoryview_copy_data_to_temp((&__pyx_v_src), (&__pyx_v_tmp), __pyx_v_order, __pyx_v_ndim); if (unlikely(__pyx_t_7 == ((void *)NULL))) __PYX_ERR(2, 1307, __pyx_L1_error)
    __pyx_v_tmpdata = __pyx_t_7;

    /* "View.MemoryView":1308
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 *         src = tmp             # <<<<<<<<<<<<<<
 * 
 *     if not broadcasting:
 */
    __pyx_v_src = __pyx_v_tmp;

    /* "View.MemoryView":1302
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  }

  /* "View.MemoryView":1310
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_broadcasting != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1313
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'C', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1314
 * 
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)             # <<<<<<<<<<<<<<
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'C', __pyx_v_ndim);

      /* "View.MemoryView":1313
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
      goto __pyx_L12;
    }

    /* "View.MemoryView":1315
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'F', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1316
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)             # <<<<<<<<<<<<<<
 * 
 *         if direct_copy:
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'F', __pyx_v_ndim);

      /* "View.MemoryView":1315
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    }
    __pyx_L12:;

    /* "View.MemoryView":1318
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __pyx_t_2 = (__pyx_v_direct_copy != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1320
 *         if direct_copy:
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

      /* "View.MemoryView":1321
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))             # <<<<<<<<<<<<<<
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 */
      (void)(memcpy(__pyx_v_dst.data, __pyx_v_src.data, __pyx_memoryview_slice_get_size((&__pyx_v_src), __pyx_v_ndim)));

      /* "View.MemoryView":1322
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 *             free(tmpdata)
 *             return 0
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

      /* "View.MemoryView":1323
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
      free(__pyx_v_tmpdata);

      /* "View.MemoryView":1324
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "View.MemoryView":1318
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    }

    /* "View.MemoryView":1310
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1326
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (__pyx_v_order == 'F');
  if (__pyx_t_2) {
    __pyx_t_2 = ('F' == __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim));
  }
  __pyx_t_8 = (__pyx_t_2 != 0);
  if (__pyx_t_8) {

    /* "View.MemoryView":1329
 * 
 * 
 *         transpose_memslice(&src)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&dst)
 * 
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_src)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(2, 1329, __pyx_L1_error)

    /* "View.MemoryView":1330
 * 
 *         transpose_memslice(&src)
 *         transpose_memslice(&dst)             # <<<<<<<<<<<<<<
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_dst)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(2, 1330, __pyx_L1_error)

    /* "View.MemoryView":1326
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1332
 *         transpose_memslice(&dst)
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1333
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)             # <<<<<<<<<<<<<<
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 */
  copy_strided_to_strided((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1334
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 *     free(tmpdata)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1336
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 *     free(tmpdata)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  free(__pyx_v_tmpdata);

  /* "View.MemoryView":1337
 * 
 *     free(tmpdata)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1268
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.memoryview_copy_contents", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1340
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim, int __pyx_v_ndim_other) {
  int __pyx_v_i;
  int __pyx_v_offset;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "View.MemoryView":1344
 *                             int ndim_other) nogil:
 *     cdef int i
 *     cdef int offset = ndim_other - ndim             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_offset = (__pyx_v_ndim_other - __pyx_v_ndim);

  /* "View.MemoryView":1346
 *     cdef int offset = ndim_other - ndim
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1347
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]             # <<<<<<<<<<<<<<
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 */
    (__pyx_v_mslice->shape[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->shape[__pyx_v_i]);

    /* "View.MemoryView":1348
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 */
    (__pyx_v_mslice->strides[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->strides[__pyx_v_i]);

    /* "View.MemoryView":1349
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(offset):
 */
    (__pyx_v_mslice->suboffsets[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->suboffsets[__pyx_v_i]);
  }

  /* "View.MemoryView":1351
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 *     for i in range(offset):             # <<<<<<<<<<<<<<
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 */
  __pyx_t_1 = __pyx_v_offset;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1352
 * 
 *     for i in range(offset):
 *         mslice.shape[i] = 1             # <<<<<<<<<<<<<<
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1
 */
    (__pyx_v_mslice->shape[__pyx_v_i]) = 1;

    /* "View.MemoryView":1353
 *     for i in range(offset):
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_mslice->strides[__pyx_v_i]) = (__pyx_v_mslice->strides[0]);

    /* "View.MemoryView":1354
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_mslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1340
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1362
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_dtype_is_object, int __pyx_v_ndim, int __pyx_v_inc) {
  int __pyx_t_1;

  /* "View.MemoryView":1366
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  __pyx_t_1 = (__pyx_v_dtype_is_object != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1367
 * 
 *     if dtype_is_object:
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,             # <<<<<<<<<<<<<<
 *                                            dst.strides, ndim, inc)
 * 
 */
    __pyx_memoryview_refcount_objects_in_slice_with_gil(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_inc);

    /* "View.MemoryView":1366
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  }

  /* "View.MemoryView":1362
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

  /* function exit code */
}

/* "View.MemoryView":1371
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  __Pyx_RefNannyDeclarations
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("refcount_objects_in_slice_with_gil", 0);

  /* "View.MemoryView":1374
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 *     refcount_objects_in_slice(data, shape, strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 */
  __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, __pyx_v_shape, __pyx_v_strides, __pyx_v_ndim, __pyx_v_inc);

  /* "View.MemoryView":1371
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "View.MemoryView":1377
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

static void __pyx_memoryview_refcount_objects_in_slice(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("refcount_objects_in_slice", 0);

  /* "View.MemoryView":1381
 *     cdef Py_ssize_t i
 * 
 *     for i in range(shape[0]):             # <<<<<<<<<<<<<<
 *         if ndim == 1:
 *             if inc:
 */
  __pyx_t_1 = (__pyx_v_shape[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1382
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
    __pyx_t_4 = ((__pyx_v_ndim == 1) != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":1383
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
      __pyx_t_4 = (__pyx_v_inc != 0);
      if (__pyx_t_4) {

        /* "View.MemoryView":1384
 *         if ndim == 1:
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])
 */
        Py_INCREF((((PyObject **)__pyx_v_data)[0]));

        /* "View.MemoryView":1383
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
        goto __pyx_L6;
      }

      /* "View.MemoryView":1386
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 */
      /*else*/ {
        Py_DECREF((((PyObject **)__pyx_v_data)[0]));
      }
      __pyx_L6:;

      /* "View.MemoryView":1382
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":1388
 *                 Py_DECREF((<PyObject **> data)[0])
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                       ndim - 1, inc)
 * 
 */
    /*else*/ {

      /* "View.MemoryView":1389
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 *                                       ndim - 1, inc)             # <<<<<<<<<<<<<<
 * 
 *         data += strides[0]
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_inc);
    }
    __pyx_L5:;

    /* "View.MemoryView":1391
 *                                       ndim - 1, inc)
 * 
 *         data += strides[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_data = (__pyx_v_data + (__pyx_v_strides[0]));
  }

  /* "View.MemoryView":1377
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1397
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item, int __pyx_v_dtype_is_object) {

  /* "View.MemoryView":1400
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1401
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,             # <<<<<<<<<<<<<<
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)
 */
  __pyx_memoryview__slice_assign_scalar(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_itemsize, __pyx_v_item);

  /* "View.MemoryView":1403
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1397
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1407
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

static void __pyx_memoryview__slice_assign_scalar(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_extent;
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;

  /* "View.MemoryView":1411
 *                               size_t itemsize, void *item) nogil:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t extent = shape[0]
 * 
 */
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1412
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]
 *     cdef Py_ssize_t extent = shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_extent = (__pyx_v_shape[0]);

  /* "View.MemoryView":1414
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1415
 * 
 *     if ndim == 1:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             memcpy(data, item, itemsize)
 *             data += stride
 */
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1416
 *     if ndim == 1:
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)             # <<<<<<<<<<<<<<
 *             data += stride
 *     else:
 */
      (void)(memcpy(__pyx_v_data, __pyx_v_item, __pyx_v_itemsize));

      /* "View.MemoryView":1417
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 *             data += stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(extent):
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }

    /* "View.MemoryView":1414
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1419
 *             data += stride
 *     else:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 */
  /*else*/ {
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1420
 *     else:
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                 ndim - 1, itemsize, item)
 *             data += stride
 */
      __pyx_memoryview__slice_assign_scalar(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize, __pyx_v_item);

      /* "View.MemoryView":1422
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 *             data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1407
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum = {"__pyx_unpickle_Enum", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Enum") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xb068931) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xb068931:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xb0, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_MemviewEnum_type), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_unpickle_Enum__set_state(((struct __pyx_MemviewEnum_obj *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 1) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_vtable_2yt_8geometry_18selection_routines_SelectorObject;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)o);
  p->__pyx_vtab = __pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  if (unlikely(__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_2yt_8geometry_18selection_routines_14SelectorObject_min_level(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_9min_level_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_18selection_routines_14SelectorObject_min_level(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_9min_level_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_18selection_routines_14SelectorObject_max_level(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_9max_level_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_18selection_routines_14SelectorObject_max_level(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_9max_level_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_SelectorObject[] = {
  {"select_grids", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_3select_grids, METH_VARARGS|METH_KEYWORDS, 0},
  {"count_octs", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_5count_octs, METH_VARARGS|METH_KEYWORDS, 0},
  {"count_oct_cells", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_7count_oct_cells, METH_VARARGS|METH_KEYWORDS, 0},
  {"fill_mesh_mask", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_9fill_mesh_mask, METH_O, 0},
  {"fill_mesh_cell_mask", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_11fill_mesh_cell_mask, METH_O, 0},
  {"fill_mask", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_13fill_mask, METH_O, 0},
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_21_hash_vals, METH_NOARGS, 0},
  {"_base_hash", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_23_base_hash, METH_NOARGS, 0},
  {"__pyx_fuse_0count_points", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_0__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_25count_points, METH_VARARGS|METH_KEYWORDS, 0},
  {"__pyx_fuse_1count_points", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_1__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_27count_points, METH_VARARGS|METH_KEYWORDS, 0},
  {"__pyx_fuse_0select_points", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_0__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_31select_points, METH_VARARGS|METH_KEYWORDS, 0},
  {"__pyx_fuse_1select_points", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_1__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_33select_points, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_37__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_39__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_2yt_8geometry_18selection_routines_SelectorObject[] = {
  {(char *)"min_level", __pyx_getprop_2yt_8geometry_18selection_routines_14SelectorObject_min_level, __pyx_setprop_2yt_8geometry_18selection_routines_14SelectorObject_min_level, (char *)0, 0},
  {(char *)"max_level", __pyx_getprop_2yt_8geometry_18selection_routines_14SelectorObject_max_level, __pyx_setprop_2yt_8geometry_18selection_routines_14SelectorObject_max_level, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_SelectorObject = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.SelectorObject", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_2yt_8geometry_18selection_routines_14SelectorObject_19__hash__, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_SelectorObject, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_2yt_8geometry_18selection_routines_SelectorObject, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_AlwaysSelector __pyx_vtable_2yt_8geometry_18selection_routines_AlwaysSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_AlwaysSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_AlwaysSelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_AlwaysSelector[] = {
  {"select_grids", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_18selection_routines_14AlwaysSelector_3select_grids, METH_VARARGS|METH_KEYWORDS, 0},
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_14AlwaysSelector_5_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_14AlwaysSelector_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_14AlwaysSelector_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_AlwaysSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.AlwaysSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_AlwaysSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_14AlwaysSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_AlwaysSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_OctreeSubsetSelector __pyx_vtable_2yt_8geometry_18selection_routines_OctreeSubsetSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_OctreeSubsetSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_OctreeSubsetSelector;
  p->base_selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_2yt_8geometry_18selection_routines_OctreeSubsetSelector(PyObject *o) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->base_selector);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject(o);
}

static int __pyx_tp_traverse_2yt_8geometry_18selection_routines_OctreeSubsetSelector(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *)o;
  e = ((likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) ? ((__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse) ? __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_2yt_8geometry_18selection_routines_OctreeSubsetSelector)); if (e) return e;
  if (p->base_selector) {
    e = (*v)(((PyObject *)p->base_selector), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_2yt_8geometry_18selection_routines_OctreeSubsetSelector(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector *)o;
  if (likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) { if (__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear) __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_2yt_8geometry_18selection_routines_OctreeSubsetSelector);
  tmp = ((PyObject*)p->base_selector);
  p->base_selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_domain_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_9domain_id_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_domain_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_9domain_id_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_OctreeSubsetSelector[] = {
  {"select_grids", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_3select_grids, METH_VARARGS|METH_KEYWORDS, 0},
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_5_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_2yt_8geometry_18selection_routines_OctreeSubsetSelector[] = {
  {(char *)"domain_id", __pyx_getprop_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_domain_id, __pyx_setprop_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_domain_id, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_OctreeSubsetSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.OctreeSubsetSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_OctreeSubsetSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines_OctreeSubsetSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines_OctreeSubsetSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_OctreeSubsetSelector, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_2yt_8geometry_18selection_routines_OctreeSubsetSelector, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_OctreeSubsetSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanSelector __pyx_vtable_2yt_8geometry_18selection_routines_BooleanSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_BooleanSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanSelector;
  p->sel1 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None); Py_INCREF(Py_None);
  p->sel2 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_2yt_8geometry_18selection_routines_BooleanSelector(PyObject *o) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->sel1);
  Py_CLEAR(p->sel2);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject(o);
}

static int __pyx_tp_traverse_2yt_8geometry_18selection_routines_BooleanSelector(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *)o;
  e = ((likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) ? ((__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse) ? __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_2yt_8geometry_18selection_routines_BooleanSelector)); if (e) return e;
  if (p->sel1) {
    e = (*v)(((PyObject *)p->sel1), a); if (e) return e;
  }
  if (p->sel2) {
    e = (*v)(((PyObject *)p->sel2), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_2yt_8geometry_18selection_routines_BooleanSelector(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector *)o;
  if (likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) { if (__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear) __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_2yt_8geometry_18selection_routines_BooleanSelector);
  tmp = ((PyObject*)p->sel1);
  p->sel1 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->sel2);
  p->sel2 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_2yt_8geometry_18selection_routines_15BooleanSelector_sel1(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_18selection_routines_15BooleanSelector_sel1(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1_3__set__(o, v);
  }
  else {
    return __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel1_5__del__(o);
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_18selection_routines_15BooleanSelector_sel2(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_18selection_routines_15BooleanSelector_sel2(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2_3__set__(o, v);
  }
  else {
    return __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_4sel2_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_BooleanSelector[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_2yt_8geometry_18selection_routines_BooleanSelector[] = {
  {(char *)"sel1", __pyx_getprop_2yt_8geometry_18selection_routines_15BooleanSelector_sel1, __pyx_setprop_2yt_8geometry_18selection_routines_15BooleanSelector_sel1, (char *)0, 0},
  {(char *)"sel2", __pyx_getprop_2yt_8geometry_18selection_routines_15BooleanSelector_sel2, __pyx_setprop_2yt_8geometry_18selection_routines_15BooleanSelector_sel2, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_BooleanSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.BooleanSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_PointSelector __pyx_vtable_2yt_8geometry_18selection_routines_PointSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_PointSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_PointSelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_PointSelector[] = {
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_13PointSelector_3_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_13PointSelector_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_13PointSelector_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_PointSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.PointSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_PointSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_PointSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_13PointSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_PointSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SphereSelector __pyx_vtable_2yt_8geometry_18selection_routines_SphereSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_SphereSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_SphereSelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_SphereSelector[] = {
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_14SphereSelector_3_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_14SphereSelector_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_14SphereSelector_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_SphereSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.SphereSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_SphereSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_SphereSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_14SphereSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_SphereSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_RegionSelector __pyx_vtable_2yt_8geometry_18selection_routines_RegionSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_RegionSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_RegionSelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_RegionSelector[] = {
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_14RegionSelector_3_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_14RegionSelector_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_14RegionSelector_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_RegionSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.RegionSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_RegionSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_RegionSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_14RegionSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_RegionSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_CutRegionSelector __pyx_vtable_2yt_8geometry_18selection_routines_CutRegionSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_CutRegionSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_CutRegionSelector;
  p->_positions = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_conditionals = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_2yt_8geometry_18selection_routines_CutRegionSelector(PyObject *o) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_positions);
  Py_CLEAR(p->_conditionals);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject(o);
}

static int __pyx_tp_traverse_2yt_8geometry_18selection_routines_CutRegionSelector(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *)o;
  e = ((likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) ? ((__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse) ? __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_2yt_8geometry_18selection_routines_CutRegionSelector)); if (e) return e;
  if (p->_positions) {
    e = (*v)(p->_positions, a); if (e) return e;
  }
  if (p->_conditionals) {
    e = (*v)(p->_conditionals, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_2yt_8geometry_18selection_routines_CutRegionSelector(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector *)o;
  if (likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) { if (__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear) __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_2yt_8geometry_18selection_routines_CutRegionSelector);
  tmp = ((PyObject*)p->_positions);
  p->_positions = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_conditionals);
  p->_conditionals = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_CutRegionSelector[] = {
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_17CutRegionSelector_3_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_17CutRegionSelector_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_17CutRegionSelector_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_CutRegionSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.CutRegionSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_CutRegionSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_CutRegionSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines_CutRegionSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines_CutRegionSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_CutRegionSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_17CutRegionSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_CutRegionSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_DiskSelector __pyx_vtable_2yt_8geometry_18selection_routines_DiskSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_DiskSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_DiskSelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_DiskSelector[] = {
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_12DiskSelector_3_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_12DiskSelector_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_12DiskSelector_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_DiskSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.DiskSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_DiskSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_DiskSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_12DiskSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_DiskSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_CuttingPlaneSelector __pyx_vtable_2yt_8geometry_18selection_routines_CuttingPlaneSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_CuttingPlaneSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_CuttingPlaneSelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_CuttingPlaneSelector[] = {
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_3_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_CuttingPlaneSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.CuttingPlaneSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_CuttingPlaneSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_CuttingPlaneSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_CuttingPlaneSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SliceSelector __pyx_vtable_2yt_8geometry_18selection_routines_SliceSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_SliceSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_SliceSelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_SliceSelector[] = {
  {"fill_mask", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_13SliceSelector_3fill_mask, METH_O, 0},
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_13SliceSelector_5_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_13SliceSelector_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_13SliceSelector_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_SliceSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.SliceSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_SliceSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_SliceSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_13SliceSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_SliceSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_OrthoRaySelector __pyx_vtable_2yt_8geometry_18selection_routines_OrthoRaySelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_OrthoRaySelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_OrthoRaySelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_OrthoRaySelector[] = {
  {"fill_mask", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_16OrthoRaySelector_3fill_mask, METH_O, 0},
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_16OrthoRaySelector_5_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_16OrthoRaySelector_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_16OrthoRaySelector_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_OrthoRaySelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.OrthoRaySelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_OrthoRaySelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_OrthoRaySelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_16OrthoRaySelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_OrthoRaySelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_RaySelector __pyx_vtable_2yt_8geometry_18selection_routines_RaySelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_RaySelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_RaySelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_RaySelector[] = {
  {"fill_mask", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_3fill_mask, METH_O, 0},
  {"get_dt", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_5get_dt, METH_O, 0},
  {"get_dt_mesh", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_7get_dt_mesh, METH_VARARGS|METH_KEYWORDS, 0},
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_9_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_11__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_13__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_RaySelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.RaySelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_RaySelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_RaySelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_11RaySelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_RaySelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_DataCollectionSelector __pyx_vtable_2yt_8geometry_18selection_routines_DataCollectionSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_DataCollectionSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_DataCollectionSelector;
  p->obj_ids = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_2yt_8geometry_18selection_routines_DataCollectionSelector(PyObject *o) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->obj_ids);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject(o);
}

static int __pyx_tp_traverse_2yt_8geometry_18selection_routines_DataCollectionSelector(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *)o;
  e = ((likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) ? ((__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse) ? __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_2yt_8geometry_18selection_routines_DataCollectionSelector)); if (e) return e;
  if (p->obj_ids) {
    e = (*v)(p->obj_ids, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_2yt_8geometry_18selection_routines_DataCollectionSelector(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector *)o;
  if (likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) { if (__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear) __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_2yt_8geometry_18selection_routines_DataCollectionSelector);
  tmp = ((PyObject*)p->obj_ids);
  p->obj_ids = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_DataCollectionSelector[] = {
  {"select_grids", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_3select_grids, METH_VARARGS|METH_KEYWORDS, 0},
  {"fill_mask", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_5fill_mask, METH_O, 0},
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_7_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_9__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_11__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_DataCollectionSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.DataCollectionSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_DataCollectionSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_DataCollectionSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines_DataCollectionSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines_DataCollectionSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_DataCollectionSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_22DataCollectionSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_DataCollectionSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_EllipsoidSelector __pyx_vtable_2yt_8geometry_18selection_routines_EllipsoidSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_EllipsoidSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_EllipsoidSelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_EllipsoidSelector[] = {
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_17EllipsoidSelector_3_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_17EllipsoidSelector_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_17EllipsoidSelector_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_EllipsoidSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.EllipsoidSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_EllipsoidSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_EllipsoidSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_17EllipsoidSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_EllipsoidSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_GridSelector __pyx_vtable_2yt_8geometry_18selection_routines_GridSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_GridSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_GridSelector;
  p->ind = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_2yt_8geometry_18selection_routines_GridSelector(PyObject *o) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->ind);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject(o);
}

static int __pyx_tp_traverse_2yt_8geometry_18selection_routines_GridSelector(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *)o;
  e = ((likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) ? ((__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse) ? __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_2yt_8geometry_18selection_routines_GridSelector)); if (e) return e;
  if (p->ind) {
    e = (*v)(p->ind, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_2yt_8geometry_18selection_routines_GridSelector(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector *)o;
  if (likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) { if (__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear) __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_2yt_8geometry_18selection_routines_GridSelector);
  tmp = ((PyObject*)p->ind);
  p->ind = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_GridSelector[] = {
  {"select_grids", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_3select_grids, METH_VARARGS|METH_KEYWORDS, 0},
  {"fill_mask", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_5fill_mask, METH_O, 0},
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_7_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_9__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_11__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_GridSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.GridSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_GridSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_GridSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines_GridSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines_GridSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_GridSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_12GridSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_GridSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector __pyx_vtable_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector;
  p->base_selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector(PyObject *o) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->base_selector);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject(o);
}

static int __pyx_tp_traverse_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *)o;
  e = ((likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) ? ((__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse) ? __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector)); if (e) return e;
  if (p->base_selector) {
    e = (*v)(((PyObject *)p->base_selector), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector *)o;
  if (likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) { if (__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear) __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector);
  tmp = ((PyObject*)p->base_selector);
  p->base_selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector[] = {
  {"select_grids", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_3select_grids, METH_VARARGS|METH_KEYWORDS, 0},
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_5_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.IndexedOctreeSubsetSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_ComposeSelector __pyx_vtable_2yt_8geometry_18selection_routines_ComposeSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_ComposeSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_ComposeSelector;
  p->selector1 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None); Py_INCREF(Py_None);
  p->selector2 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_2yt_8geometry_18selection_routines_ComposeSelector(PyObject *o) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->selector1);
  Py_CLEAR(p->selector2);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject(o);
}

static int __pyx_tp_traverse_2yt_8geometry_18selection_routines_ComposeSelector(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *)o;
  e = ((likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) ? ((__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse) ? __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_2yt_8geometry_18selection_routines_ComposeSelector)); if (e) return e;
  if (p->selector1) {
    e = (*v)(((PyObject *)p->selector1), a); if (e) return e;
  }
  if (p->selector2) {
    e = (*v)(((PyObject *)p->selector2), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_2yt_8geometry_18selection_routines_ComposeSelector(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector *)o;
  if (likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) { if (__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear) __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_2yt_8geometry_18selection_routines_ComposeSelector);
  tmp = ((PyObject*)p->selector1);
  p->selector1 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->selector2);
  p->selector2 = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_ComposeSelector[] = {
  {"select_grids", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_18selection_routines_15ComposeSelector_3select_grids, METH_VARARGS|METH_KEYWORDS, 0},
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_15ComposeSelector_5_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_15ComposeSelector_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_15ComposeSelector_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_ComposeSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.ComposeSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_ComposeSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_ComposeSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines_ComposeSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines_ComposeSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_ComposeSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_15ComposeSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_ComposeSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_HaloParticlesSelector __pyx_vtable_2yt_8geometry_18selection_routines_HaloParticlesSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_HaloParticlesSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_HaloParticlesSelector;
  p->base_source = Py_None; Py_INCREF(Py_None);
  p->base_selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None); Py_INCREF(Py_None);
  p->pind = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_2yt_8geometry_18selection_routines_HaloParticlesSelector(PyObject *o) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->base_source);
  Py_CLEAR(p->base_selector);
  Py_CLEAR(p->pind);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject(o);
}

static int __pyx_tp_traverse_2yt_8geometry_18selection_routines_HaloParticlesSelector(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *)o;
  e = ((likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) ? ((__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse) ? __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_2yt_8geometry_18selection_routines_HaloParticlesSelector)); if (e) return e;
  if (p->base_source) {
    e = (*v)(p->base_source, a); if (e) return e;
  }
  if (p->base_selector) {
    e = (*v)(((PyObject *)p->base_selector), a); if (e) return e;
  }
  if (p->pind) {
    e = (*v)(p->pind, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_2yt_8geometry_18selection_routines_HaloParticlesSelector(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector *)o;
  if (likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) { if (__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear) __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_2yt_8geometry_18selection_routines_HaloParticlesSelector);
  tmp = ((PyObject*)p->base_source);
  p->base_source = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->base_selector);
  p->base_selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->pind);
  p->pind = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_2yt_8geometry_18selection_routines_21HaloParticlesSelector_base_source(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_18selection_routines_21HaloParticlesSelector_base_source(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source_3__set__(o, v);
  }
  else {
    return __pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_11base_source_5__del__(o);
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_18selection_routines_21HaloParticlesSelector_halo_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_7halo_id_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_18selection_routines_21HaloParticlesSelector_halo_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_7halo_id_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_HaloParticlesSelector[] = {
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_3_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_2yt_8geometry_18selection_routines_HaloParticlesSelector[] = {
  {(char *)"base_source", __pyx_getprop_2yt_8geometry_18selection_routines_21HaloParticlesSelector_base_source, __pyx_setprop_2yt_8geometry_18selection_routines_21HaloParticlesSelector_base_source, (char *)0, 0},
  {(char *)"halo_id", __pyx_getprop_2yt_8geometry_18selection_routines_21HaloParticlesSelector_halo_id, __pyx_setprop_2yt_8geometry_18selection_routines_21HaloParticlesSelector_halo_id, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_HaloParticlesSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.HaloParticlesSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_HaloParticlesSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_HaloParticlesSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines_HaloParticlesSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines_HaloParticlesSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_HaloParticlesSelector, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_2yt_8geometry_18selection_routines_HaloParticlesSelector, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_21HaloParticlesSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_HaloParticlesSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanANDSelector __pyx_vtable_2yt_8geometry_18selection_routines_BooleanANDSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_BooleanANDSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_BooleanSelector(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanANDSelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_BooleanANDSelector[] = {
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_18BooleanANDSelector_1_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_18BooleanANDSelector_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_18BooleanANDSelector_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_BooleanANDSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.BooleanANDSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanANDSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_BooleanANDSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_BooleanANDSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanORSelector __pyx_vtable_2yt_8geometry_18selection_routines_BooleanORSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_BooleanORSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_BooleanSelector(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanORSelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_BooleanORSelector[] = {
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_17BooleanORSelector_1_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_17BooleanORSelector_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_17BooleanORSelector_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_BooleanORSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.BooleanORSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanORSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_BooleanORSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_BooleanORSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanNOTSelector __pyx_vtable_2yt_8geometry_18selection_routines_BooleanNOTSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_BooleanNOTSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_BooleanSelector(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanNOTSelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_BooleanNOTSelector[] = {
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNOTSelector_1_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNOTSelector_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNOTSelector_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_BooleanNOTSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.BooleanNOTSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNOTSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_BooleanNOTSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_BooleanNOTSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanXORSelector __pyx_vtable_2yt_8geometry_18selection_routines_BooleanXORSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_BooleanXORSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_BooleanSelector(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanXORSelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_BooleanXORSelector[] = {
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_18BooleanXORSelector_1_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_18BooleanXORSelector_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_18BooleanXORSelector_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_BooleanXORSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.BooleanXORSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanXORSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_BooleanXORSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_BooleanXORSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanNEGSelector __pyx_vtable_2yt_8geometry_18selection_routines_BooleanNEGSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_BooleanNEGSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_BooleanSelector(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanNEGSelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_BooleanNEGSelector[] = {
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNEGSelector_1_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNEGSelector_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_18BooleanNEGSelector_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_BooleanNEGSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.BooleanNEGSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanNEGSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines_BooleanSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_BooleanNEGSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_2yt_8geometry_18selection_routines_15BooleanSelector_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_BooleanNEGSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_ChainedBooleanSelector __pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_ChainedBooleanSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_SelectorObject(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_ChainedBooleanSelector;
  p->selectors = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_2yt_8geometry_18selection_routines_ChainedBooleanSelector(PyObject *o) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->selectors);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_SelectorObject(o);
}

static int __pyx_tp_traverse_2yt_8geometry_18selection_routines_ChainedBooleanSelector(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *)o;
  e = ((likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) ? ((__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse) ? __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_2yt_8geometry_18selection_routines_ChainedBooleanSelector)); if (e) return e;
  if (p->selectors) {
    e = (*v)(((PyObject *)p->selectors), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_2yt_8geometry_18selection_routines_ChainedBooleanSelector(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *p = (struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector *)o;
  if (likely(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject)) { if (__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear) __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_2yt_8geometry_18selection_routines_ChainedBooleanSelector);
  tmp = ((PyObject*)p->selectors);
  p->selectors = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_ChainedBooleanSelector[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.ChainedBooleanSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_ChainedBooleanSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines_ChainedBooleanSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines_ChainedBooleanSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_ChainedBooleanSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_ChainedBooleanSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector __pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_ChainedBooleanSelector(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector[] = {
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_1_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.ChainedBooleanANDSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_ChainedBooleanSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines_ChainedBooleanSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines_ChainedBooleanSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_ChainedBooleanORSelector __pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanORSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines_ChainedBooleanORSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *p;
  PyObject *o = __pyx_tp_new_2yt_8geometry_18selection_routines_ChainedBooleanSelector(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__pyx_vtabptr_2yt_8geometry_18selection_routines_ChainedBooleanORSelector;
  return o;
}

static PyMethodDef __pyx_methods_2yt_8geometry_18selection_routines_ChainedBooleanORSelector[] = {
  {"_hash_vals", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_1_hash_vals, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanORSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.ChainedBooleanORSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_ChainedBooleanORSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines_ChainedBooleanSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines_ChainedBooleanSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines_ChainedBooleanSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_18selection_routines_ChainedBooleanORSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_2yt_8geometry_18selection_routines_22ChainedBooleanSelector_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines_ChainedBooleanORSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ *__pyx_freelist_2yt_8geometry_18selection_routines___pyx_scope_struct____init__[8];
static int __pyx_freecount_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ = 0;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines___pyx_scope_struct____init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__)))) {
    o = (PyObject*)__pyx_freelist_2yt_8geometry_18selection_routines___pyx_scope_struct____init__[--__pyx_freecount_2yt_8geometry_18selection_routines___pyx_scope_struct____init__];
    memset(o, 0, sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_2yt_8geometry_18selection_routines___pyx_scope_struct____init__(PyObject *o) {
  struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ *p = (struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_positions);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__)))) {
    __pyx_freelist_2yt_8geometry_18selection_routines___pyx_scope_struct____init__[__pyx_freecount_2yt_8geometry_18selection_routines___pyx_scope_struct____init__++] = ((struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_2yt_8geometry_18selection_routines___pyx_scope_struct____init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ *p = (struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ *)o;
  if (p->__pyx_v_positions) {
    e = (*v)(p->__pyx_v_positions, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_2yt_8geometry_18selection_routines___pyx_scope_struct____init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ *p = (struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_positions);
  p->__pyx_v_positions = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.__pyx_scope_struct____init__", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct____init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines___pyx_scope_struct____init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines___pyx_scope_struct____init__, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_18selection_routines___pyx_scope_struct____init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines___pyx_scope_struct____init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr *__pyx_freelist_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr[--__pyx_freecount_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_position);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr[__pyx_freecount_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_position) {
    e = (*v)(p->__pyx_v_position, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_array __pyx_vtable_array;

static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_array_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_array_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_array;
  p->mode = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_format = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_array___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_array(PyObject *o) {
  struct __pyx_array_obj *p = (struct __pyx_array_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_array___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->mode);
  Py_CLEAR(p->_format);
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_array(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_array(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_array___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_tp_getattro_array(PyObject *o, PyObject *n) {
  PyObject *v = __Pyx_PyObject_GenericGetAttr(o, n);
  if (!v && PyErr_ExceptionMatches(PyExc_AttributeError)) {
    PyErr_Clear();
    v = __pyx_array___getattr__(o, n);
  }
  return v;
}

static PyObject *__pyx_getprop___pyx_array_memview(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(o);
}

static PyMethodDef __pyx_methods_array[] = {
  {"__getattr__", (PyCFunction)__pyx_array___getattr__, METH_O|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_array_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_array_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_array[] = {
  {(char *)"memview", __pyx_getprop___pyx_array_memview, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_array = {
  __pyx_array___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_array, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_array = {
  __pyx_array___len__, /*mp_length*/
  __pyx_array___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_array, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_array = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_array_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_array = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.array", /*tp_name*/
  sizeof(struct __pyx_array_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_array, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_array, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_array, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  __pyx_tp_getattro_array, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_array, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_array, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_array, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_array, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_MemviewEnum_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_MemviewEnum_obj *)o);
  p->name = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_Enum(PyObject *o) {
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_Enum(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  if (p->name) {
    e = (*v)(p->name, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_Enum(PyObject *o) {
  PyObject* tmp;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  tmp = ((PyObject*)p->name);
  p->name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_Enum[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_MemviewEnum = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.Enum", /*tp_name*/
  sizeof(struct __pyx_MemviewEnum_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_Enum, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_MemviewEnum___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_Enum, /*tp_traverse*/
  __pyx_tp_clear_Enum, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_Enum, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_MemviewEnum___init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_Enum, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_memoryview __pyx_vtable_memoryview;

static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryview_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryview_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_memoryview;
  p->obj = Py_None; Py_INCREF(Py_None);
  p->_size = Py_None; Py_INCREF(Py_None);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  p->view.obj = NULL;
  if (unlikely(__pyx_memoryview___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_memoryview(PyObject *o) {
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_memoryview___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->obj);
  Py_CLEAR(p->_size);
  Py_CLEAR(p->_array_interface);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_memoryview(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  if (p->obj) {
    e = (*v)(p->obj, a); if (e) return e;
  }
  if (p->_size) {
    e = (*v)(p->_size, a); if (e) return e;
  }
  if (p->_array_interface) {
    e = (*v)(p->_array_interface, a); if (e) return e;
  }
  if (p->view.obj) {
    e = (*v)(p->view.obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_memoryview(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  tmp = ((PyObject*)p->obj);
  p->obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_size);
  p->_size = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_array_interface);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  Py_CLEAR(p->view.obj);
  return 0;
}
static PyObject *__pyx_sq_item_memoryview(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_memoryview(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_memoryview___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_getprop___pyx_memoryview_T(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_shape(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_strides(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_suboffsets(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_ndim(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_itemsize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_nbytes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(o);
}

static PyMethodDef __pyx_methods_memoryview[] = {
  {"is_c_contig", (PyCFunction)__pyx_memoryview_is_c_contig, METH_NOARGS, 0},
  {"is_f_contig", (PyCFunction)__pyx_memoryview_is_f_contig, METH_NOARGS, 0},
  {"copy", (PyCFunction)__pyx_memoryview_copy, METH_NOARGS, 0},
  {"copy_fortran", (PyCFunction)__pyx_memoryview_copy_fortran, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_memoryview[] = {
  {(char *)"T", __pyx_getprop___pyx_memoryview_T, 0, (char *)0, 0},
  {(char *)"base", __pyx_getprop___pyx_memoryview_base, 0, (char *)0, 0},
  {(char *)"shape", __pyx_getprop___pyx_memoryview_shape, 0, (char *)0, 0},
  {(char *)"strides", __pyx_getprop___pyx_memoryview_strides, 0, (char *)0, 0},
  {(char *)"suboffsets", __pyx_getprop___pyx_memoryview_suboffsets, 0, (char *)0, 0},
  {(char *)"ndim", __pyx_getprop___pyx_memoryview_ndim, 0, (char *)0, 0},
  {(char *)"itemsize", __pyx_getprop___pyx_memoryview_itemsize, 0, (char *)0, 0},
  {(char *)"nbytes", __pyx_getprop___pyx_memoryview_nbytes, 0, (char *)0, 0},
  {(char *)"size", __pyx_getprop___pyx_memoryview_size, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_memoryview = {
  __pyx_memoryview___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_memoryview, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_memoryview = {
  __pyx_memoryview___len__, /*mp_length*/
  __pyx_memoryview___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_memoryview, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_memoryview = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_memoryview = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines.memoryview", /*tp_name*/
  sizeof(struct __pyx_memoryview_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_memoryview, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_memoryview___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_memoryview, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_memoryview, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_memoryview___str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_memoryview, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_memoryview, /*tp_traverse*/
  __pyx_tp_clear_memoryview, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_memoryview, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_memoryview, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_memoryview, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct__memoryviewslice __pyx_vtable__memoryviewslice;

static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryviewslice_obj *p;
  PyObject *o = __pyx_tp_new_memoryview(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryviewslice_obj *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_memoryview*)__pyx_vtabptr__memoryviewslice;
  p->from_object = Py_None; Py_INCREF(Py_None);
  p->from_slice.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc__memoryviewslice(PyObject *o) {
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_memoryviewslice___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->from_object);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_memoryview(o);
}

static int __pyx_tp_traverse__memoryviewslice(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  e = __pyx_tp_traverse_memoryview(o, v, a); if (e) return e;
  if (p->from_object) {
    e = (*v)(p->from_object, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear__memoryviewslice(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  __pyx_tp_clear_memoryview(o);
  tmp = ((PyObject*)p->from_object);
  p->from_object = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  __PYX_XDEC_MEMVIEW(&p->from_slice, 1);
  return 0;
}

static PyObject *__pyx_getprop___pyx_memoryviewslice_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(o);
}

static PyMethodDef __pyx_methods__memoryviewslice[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets__memoryviewslice[] = {
  {(char *)"base", __pyx_getprop___pyx_memoryviewslice_base, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_memoryviewslice = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.selection_routines._memoryviewslice", /*tp_name*/
  sizeof(struct __pyx_memoryviewslice_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc__memoryviewslice, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Internal class for passing memoryview slices to Python", /*tp_doc*/
  __pyx_tp_traverse__memoryviewslice, /*tp_traverse*/
  __pyx_tp_clear__memoryviewslice, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods__memoryviewslice, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets__memoryviewslice, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new__memoryviewslice, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_selection_routines(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_selection_routines},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "selection_routines",
    __pyx_k_Geometry_selection_routines, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 1, 0},
  {&__pyx_n_s_A, __pyx_k_A, sizeof(__pyx_k_A), 0, 0, 1, 1},
  {&__pyx_n_s_ASCII, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 0, 1, 1},
  {&__pyx_n_s_ActiveDimensions, __pyx_k_ActiveDimensions, sizeof(__pyx_k_ActiveDimensions), 0, 0, 1, 1},
  {&__pyx_n_s_AlwaysSelector, __pyx_k_AlwaysSelector, sizeof(__pyx_k_AlwaysSelector), 0, 0, 1, 1},
  {&__pyx_n_s_B, __pyx_k_B, sizeof(__pyx_k_B), 0, 0, 1, 1},
  {&__pyx_n_s_BooleanANDSelector, __pyx_k_BooleanANDSelector, sizeof(__pyx_k_BooleanANDSelector), 0, 0, 1, 1},
  {&__pyx_n_s_BooleanNEGSelector, __pyx_k_BooleanNEGSelector, sizeof(__pyx_k_BooleanNEGSelector), 0, 0, 1, 1},
  {&__pyx_n_s_BooleanNOTSelector, __pyx_k_BooleanNOTSelector, sizeof(__pyx_k_BooleanNOTSelector), 0, 0, 1, 1},
  {&__pyx_n_s_BooleanORSelector, __pyx_k_BooleanORSelector, sizeof(__pyx_k_BooleanORSelector), 0, 0, 1, 1},
  {&__pyx_n_s_BooleanSelector, __pyx_k_BooleanSelector, sizeof(__pyx_k_BooleanSelector), 0, 0, 1, 1},
  {&__pyx_n_s_BooleanXORSelector, __pyx_k_BooleanXORSelector, sizeof(__pyx_k_BooleanXORSelector), 0, 0, 1, 1},
  {&__pyx_kp_s_Buffer_view_does_not_expose_stri, __pyx_k_Buffer_view_does_not_expose_stri, sizeof(__pyx_k_Buffer_view_does_not_expose_stri), 0, 0, 1, 0},
  {&__pyx_n_s_C, __pyx_k_C, sizeof(__pyx_k_C), 0, 0, 1, 1},
  {&__pyx_kp_s_Can_only_create_a_buffer_that_is, __pyx_k_Can_only_create_a_buffer_that_is, sizeof(__pyx_k_Can_only_create_a_buffer_that_is), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_assign_to_read_only_memor, __pyx_k_Cannot_assign_to_read_only_memor, sizeof(__pyx_k_Cannot_assign_to_read_only_memor), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_create_writable_memory_vi, __pyx_k_Cannot_create_writable_memory_vi, sizeof(__pyx_k_Cannot_create_writable_memory_vi), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_index_with_type_s, __pyx_k_Cannot_index_with_type_s, sizeof(__pyx_k_Cannot_index_with_type_s), 0, 0, 1, 0},
  {&__pyx_n_s_ChainedBooleanANDSelector, __pyx_k_ChainedBooleanANDSelector, sizeof(__pyx_k_ChainedBooleanANDSelector), 0, 0, 1, 1},
  {&__pyx_n_s_ChainedBooleanORSelector, __pyx_k_ChainedBooleanORSelector, sizeof(__pyx_k_ChainedBooleanORSelector), 0, 0, 1, 1},
  {&__pyx_n_s_ChainedBooleanSelector, __pyx_k_ChainedBooleanSelector, sizeof(__pyx_k_ChainedBooleanSelector), 0, 0, 1, 1},
  {&__pyx_n_s_ComposeSelector, __pyx_k_ComposeSelector, sizeof(__pyx_k_ComposeSelector), 0, 0, 1, 1},
  {&__pyx_n_s_CutRegionSelector, __pyx_k_CutRegionSelector, sizeof(__pyx_k_CutRegionSelector), 0, 0, 1, 1},
  {&__pyx_n_s_CuttingPlaneSelector, __pyx_k_CuttingPlaneSelector, sizeof(__pyx_k_CuttingPlaneSelector), 0, 0, 1, 1},
  {&__pyx_n_s_DataCollectionSelector, __pyx_k_DataCollectionSelector, sizeof(__pyx_k_DataCollectionSelector), 0, 0, 1, 1},
  {&__pyx_n_s_DiskSelector, __pyx_k_DiskSelector, sizeof(__pyx_k_DiskSelector), 0, 0, 1, 1},
  {&__pyx_n_s_Ellipsis, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
  {&__pyx_n_s_EllipsoidSelector, __pyx_k_EllipsoidSelector, sizeof(__pyx_k_EllipsoidSelector), 0, 0, 1, 1},
  {&__pyx_kp_s_Empty_shape_tuple_for_cython_arr, __pyx_k_Empty_shape_tuple_for_cython_arr, sizeof(__pyx_k_Empty_shape_tuple_for_cython_arr), 0, 0, 1, 0},
  {&__pyx_kp_s_Error_yt_attempted_to_read_outsi, __pyx_k_Error_yt_attempted_to_read_outsi, sizeof(__pyx_k_Error_yt_attempted_to_read_outsi), 0, 0, 1, 0},
  {&__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_k_Expected_at_least_d_argument_s_g, sizeof(__pyx_k_Expected_at_least_d_argument_s_g), 0, 0, 1, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor, __pyx_k_Format_string_allocated_too_shor, sizeof(__pyx_k_Format_string_allocated_too_shor), 0, 1, 0, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor_2, __pyx_k_Format_string_allocated_too_shor_2, sizeof(__pyx_k_Format_string_allocated_too_shor_2), 0, 1, 0, 0},
  {&__pyx_kp_s_Function_call_with_ambiguous_arg, __pyx_k_Function_call_with_ambiguous_arg, sizeof(__pyx_k_Function_call_with_ambiguous_arg), 0, 0, 1, 0},
  {&__pyx_n_s_GridSelector, __pyx_k_GridSelector, sizeof(__pyx_k_GridSelector), 0, 0, 1, 1},
  {&__pyx_n_s_HaloParticlesSelector, __pyx_k_HaloParticlesSelector, sizeof(__pyx_k_HaloParticlesSelector), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xb0, __pyx_k_Incompatible_checksums_s_vs_0xb0, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xb0), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_n_s_IndexedOctreeSubsetSelector, __pyx_k_IndexedOctreeSubsetSelector, sizeof(__pyx_k_IndexedOctreeSubsetSelector), 0, 0, 1, 1},
  {&__pyx_kp_s_Indirect_dimensions_not_supporte, __pyx_k_Indirect_dimensions_not_supporte, sizeof(__pyx_k_Indirect_dimensions_not_supporte), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_k_Invalid_mode_expected_c_or_fortr, sizeof(__pyx_k_Invalid_mode_expected_c_or_fortr), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_k_Invalid_shape_in_axis_d_d, sizeof(__pyx_k_Invalid_shape_in_axis_d_d), 0, 0, 1, 0},
  {&__pyx_n_s_LeftEdge, __pyx_k_LeftEdge, sizeof(__pyx_k_LeftEdge), 0, 0, 1, 1},
  {&__pyx_n_s_Level, __pyx_k_Level, sizeof(__pyx_k_Level), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_k_MemoryView_of_r_at_0x_x, sizeof(__pyx_k_MemoryView_of_r_at_0x_x), 0, 0, 1, 0},
  {&__pyx_kp_s_MemoryView_of_r_object, __pyx_k_MemoryView_of_r_object, sizeof(__pyx_k_MemoryView_of_r_object), 0, 0, 1, 0},
  {&__pyx_kp_s_No_matching_signature_found, __pyx_k_No_matching_signature_found, sizeof(__pyx_k_No_matching_signature_found), 0, 0, 1, 0},
  {&__pyx_kp_u_Non_native_byte_order_not_suppor, __pyx_k_Non_native_byte_order_not_suppor, sizeof(__pyx_k_Non_native_byte_order_not_suppor), 0, 1, 0, 0},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_b_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 0, 1},
  {&__pyx_n_s_OctreeSubsetSelector, __pyx_k_OctreeSubsetSelector, sizeof(__pyx_k_OctreeSubsetSelector), 0, 0, 1, 1},
  {&__pyx_n_s_OrthoRaySelector, __pyx_k_OrthoRaySelector, sizeof(__pyx_k_OrthoRaySelector), 0, 0, 1, 1},
  {&__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_k_Out_of_bounds_on_buffer_access_a, sizeof(__pyx_k_Out_of_bounds_on_buffer_access_a), 0, 0, 1, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_PointSelector, __pyx_k_PointSelector, sizeof(__pyx_k_PointSelector), 0, 0, 1, 1},
  {&__pyx_n_s_RaySelector, __pyx_k_RaySelector, sizeof(__pyx_k_RaySelector), 0, 0, 1, 1},
  {&__pyx_n_s_RegionSelector, __pyx_k_RegionSelector, sizeof(__pyx_k_RegionSelector), 0, 0, 1, 1},
  {&__pyx_kp_s_Region_right_edge_s_left_edge_wi, __pyx_k_Region_right_edge_s_left_edge_wi, sizeof(__pyx_k_Region_right_edge_s_left_edge_wi), 0, 0, 1, 0},
  {&__pyx_n_s_RightEdge, __pyx_k_RightEdge, sizeof(__pyx_k_RightEdge), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_SelectorObject, __pyx_k_SelectorObject, sizeof(__pyx_k_SelectorObject), 0, 0, 1, 1},
  {&__pyx_n_s_SelectorObject_count_points, __pyx_k_SelectorObject_count_points, sizeof(__pyx_k_SelectorObject_count_points), 0, 0, 1, 1},
  {&__pyx_n_s_SelectorObject_select_points, __pyx_k_SelectorObject_select_points, sizeof(__pyx_k_SelectorObject_select_points), 0, 0, 1, 1},
  {&__pyx_n_s_SliceSelector, __pyx_k_SliceSelector, sizeof(__pyx_k_SliceSelector), 0, 0, 1, 1},
  {&__pyx_n_s_SphereSelector, __pyx_k_SphereSelector, sizeof(__pyx_k_SphereSelector), 0, 0, 1, 1},
  {&__pyx_n_s_T, __pyx_k_T, sizeof(__pyx_k_T), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unable_to_convert_item_to_object, __pyx_k_Unable_to_convert_item_to_object, sizeof(__pyx_k_Unable_to_convert_item_to_object), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_View_MemoryView, __pyx_k_View_MemoryView, sizeof(__pyx_k_View_MemoryView), 0, 0, 1, 1},
  {&__pyx_kp_s__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 0, 1, 0},
  {&__pyx_n_s_allocate_buffer, __pyx_k_allocate_buffer, sizeof(__pyx_k_allocate_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_always, __pyx_k_always, sizeof(__pyx_k_always), 0, 0, 1, 1},
  {&__pyx_n_s_always_selector, __pyx_k_always_selector, sizeof(__pyx_k_always_selector), 0, 0, 1, 1},
  {&__pyx_n_s_and, __pyx_k_and, sizeof(__pyx_k_and), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_ascii, __pyx_k_ascii, sizeof(__pyx_k_ascii), 0, 0, 1, 1},
  {&__pyx_n_s_astype, __pyx_k_astype, sizeof(__pyx_k_astype), 0, 0, 1, 1},
  {&__pyx_n_s_axis, __pyx_k_axis, sizeof(__pyx_k_axis), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_base_hash, __pyx_k_base_hash, sizeof(__pyx_k_base_hash), 0, 0, 1, 1},
  {&__pyx_n_s_base_selector, __pyx_k_base_selector, sizeof(__pyx_k_base_selector), 0, 0, 1, 1},
  {&__pyx_n_s_base_source, __pyx_k_base_source, sizeof(__pyx_k_base_source), 0, 0, 1, 1},
  {&__pyx_n_s_bool, __pyx_k_bool, sizeof(__pyx_k_bool), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
  {&__pyx_n_s_center, __pyx_k_center, sizeof(__pyx_k_center), 0, 0, 1, 1},
  {&__pyx_kp_s_center_0, __pyx_k_center_0, sizeof(__pyx_k_center_0), 0, 0, 1, 0},
  {&__pyx_kp_s_center_1, __pyx_k_center_1, sizeof(__pyx_k_center_1), 0, 0, 1, 0},
  {&__pyx_kp_s_center_2, __pyx_k_center_2, sizeof(__pyx_k_center_2), 0, 0, 1, 0},
  {&__pyx_n_s_chained_and, __pyx_k_chained_and, sizeof(__pyx_k_chained_and), 0, 0, 1, 1},
  {&__pyx_n_s_chained_or, __pyx_k_chained_or, sizeof(__pyx_k_chained_or), 0, 0, 1, 1},
  {&__pyx_n_s_child_mask, __pyx_k_child_mask, sizeof(__pyx_k_child_mask), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_code_length, __pyx_k_code_length, sizeof(__pyx_k_code_length), 0, 0, 1, 1},
  {&__pyx_n_s_compose_selector, __pyx_k_compose_selector, sizeof(__pyx_k_compose_selector), 0, 0, 1, 1},
  {&__pyx_kp_s_conditional_s, __pyx_k_conditional_s, sizeof(__pyx_k_conditional_s), 0, 0, 1, 0},
  {&__pyx_n_s_conditionals, __pyx_k_conditionals, sizeof(__pyx_k_conditionals), 0, 0, 1, 1},
  {&__pyx_n_s_connectivity_coords, __pyx_k_connectivity_coords, sizeof(__pyx_k_connectivity_coords), 0, 0, 1, 1},
  {&__pyx_n_s_connectivity_indices, __pyx_k_connectivity_indices, sizeof(__pyx_k_connectivity_indices), 0, 0, 1, 1},
  {&__pyx_n_s_connectivity_length, __pyx_k_connectivity_length, sizeof(__pyx_k_connectivity_length), 0, 0, 1, 1},
  {&__pyx_kp_s_contiguous_and_direct, __pyx_k_contiguous_and_direct, sizeof(__pyx_k_contiguous_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_contiguous_and_indirect, __pyx_k_contiguous_and_indirect, sizeof(__pyx_k_contiguous_and_indirect), 0, 0, 1, 0},
  {&__pyx_n_s_convert_mask_to_indices, __pyx_k_convert_mask_to_indices, sizeof(__pyx_k_convert_mask_to_indices), 0, 0, 1, 1},
  {&__pyx_n_s_convert_to_units, __pyx_k_convert_to_units, sizeof(__pyx_k_convert_to_units), 0, 0, 1, 1},
  {&__pyx_n_s_coord, __pyx_k_coord, sizeof(__pyx_k_coord), 0, 0, 1, 1},
  {&__pyx_n_s_count, __pyx_k_count, sizeof(__pyx_k_count), 0, 0, 1, 1},
  {&__pyx_n_s_count_points, __pyx_k_count_points, sizeof(__pyx_k_count_points), 0, 0, 1, 1},
  {&__pyx_n_s_cpos, __pyx_k_cpos, sizeof(__pyx_k_cpos), 0, 0, 1, 1},
  {&__pyx_n_s_cut_region_selector, __pyx_k_cut_region_selector, sizeof(__pyx_k_cut_region_selector), 0, 0, 1, 1},
  {&__pyx_n_s_cutting_selector, __pyx_k_cutting_selector, sizeof(__pyx_k_cutting_selector), 0, 0, 1, 1},
  {&__pyx_n_s_cx, __pyx_k_cx, sizeof(__pyx_k_cx), 0, 0, 1, 1},
  {&__pyx_n_s_cy, __pyx_k_cy, sizeof(__pyx_k_cy), 0, 0, 1, 1},
  {&__pyx_n_s_cz, __pyx_k_cz, sizeof(__pyx_k_cz), 0, 0, 1, 1},
  {&__pyx_n_s_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 1},
  {&__pyx_n_s_d_2, __pyx_k_d_2, sizeof(__pyx_k_d_2), 0, 0, 1, 1},
  {&__pyx_n_s_data_collection_selector, __pyx_k_data_collection_selector, sizeof(__pyx_k_data_collection_selector), 0, 0, 1, 1},
  {&__pyx_n_s_data_objects, __pyx_k_data_objects, sizeof(__pyx_k_data_objects), 0, 0, 1, 1},
  {&__pyx_n_s_dds, __pyx_k_dds, sizeof(__pyx_k_dds), 0, 0, 1, 1},
  {&__pyx_n_s_defaults, __pyx_k_defaults, sizeof(__pyx_k_defaults), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_disk_selector, __pyx_k_disk_selector, sizeof(__pyx_k_disk_selector), 0, 0, 1, 1},
  {&__pyx_n_s_dobj, __pyx_k_dobj, sizeof(__pyx_k_dobj), 0, 0, 1, 1},
  {&__pyx_n_s_dobj1, __pyx_k_dobj1, sizeof(__pyx_k_dobj1), 0, 0, 1, 1},
  {&__pyx_n_s_dobj2, __pyx_k_dobj2, sizeof(__pyx_k_dobj2), 0, 0, 1, 1},
  {&__pyx_n_s_domain_id, __pyx_k_domain_id, sizeof(__pyx_k_domain_id), 0, 0, 1, 1},
  {&__pyx_n_s_domain_left_edge, __pyx_k_domain_left_edge, sizeof(__pyx_k_domain_left_edge), 0, 0, 1, 1},
  {&__pyx_n_s_domain_right_edge, __pyx_k_domain_right_edge, sizeof(__pyx_k_domain_right_edge), 0, 0, 1, 1},
  {&__pyx_n_s_domain_width, __pyx_k_domain_width, sizeof(__pyx_k_domain_width), 0, 0, 1, 1},
  {&__pyx_kp_s_domain_width_0, __pyx_k_domain_width_0, sizeof(__pyx_k_domain_width_0), 0, 0, 1, 0},
  {&__pyx_kp_s_domain_width_1, __pyx_k_domain_width_1, sizeof(__pyx_k_domain_width_1), 0, 0, 1, 0},
  {&__pyx_kp_s_domain_width_2, __pyx_k_domain_width_2, sizeof(__pyx_k_domain_width_2), 0, 0, 1, 0},
  {&__pyx_n_s_double, __pyx_k_double, sizeof(__pyx_k_double), 0, 0, 1, 1},
  {&__pyx_n_s_ds, __pyx_k_ds, sizeof(__pyx_k_ds), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dtype_is_object, __pyx_k_dtype_is_object, sizeof(__pyx_k_dtype_is_object), 0, 0, 1, 1},
  {&__pyx_n_s_dx, __pyx_k_dx, sizeof(__pyx_k_dx), 0, 0, 1, 1},
  {&__pyx_n_s_dy, __pyx_k_dy, sizeof(__pyx_k_dy), 0, 0, 1, 1},
  {&__pyx_n_s_dz, __pyx_k_dz, sizeof(__pyx_k_dz), 0, 0, 1, 1},
  {&__pyx_n_s_e0, __pyx_k_e0, sizeof(__pyx_k_e0), 0, 0, 1, 1},
  {&__pyx_n_s_e1, __pyx_k_e1, sizeof(__pyx_k_e1), 0, 0, 1, 1},
  {&__pyx_n_s_e2, __pyx_k_e2, sizeof(__pyx_k_e2), 0, 0, 1, 1},
  {&__pyx_n_s_ellipsoid_selector, __pyx_k_ellipsoid_selector, sizeof(__pyx_k_ellipsoid_selector), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_end_point, __pyx_k_end_point, sizeof(__pyx_k_end_point), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_eps, __pyx_k_eps, sizeof(__pyx_k_eps), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_extend, __pyx_k_extend, sizeof(__pyx_k_extend), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_filter_bbox, __pyx_k_filter_bbox, sizeof(__pyx_k_filter_bbox), 0, 0, 1, 1},
  {&__pyx_n_s_finfo, __pyx_k_finfo, sizeof(__pyx_k_finfo), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_float, __pyx_k_float, sizeof(__pyx_k_float), 0, 0, 1, 1},
  {&__pyx_n_s_float32, __pyx_k_float32, sizeof(__pyx_k_float32), 0, 0, 1, 1},
  {&__pyx_n_s_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
  {&__pyx_n_s_fmod, __pyx_k_fmod, sizeof(__pyx_k_fmod), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 0, 1, 1},
  {&__pyx_n_u_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 1, 0, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_k_got_differing_extents_in_dimensi, sizeof(__pyx_k_got_differing_extents_in_dimensi), 0, 0, 1, 0},
  {&__pyx_n_s_grid_selector, __pyx_k_grid_selector, sizeof(__pyx_k_grid_selector), 0, 0, 1, 1},
  {&__pyx_n_s_halo_particles, __pyx_k_halo_particles, sizeof(__pyx_k_halo_particles), 0, 0, 1, 1},
  {&__pyx_n_s_halo_particles_selector, __pyx_k_halo_particles_selector, sizeof(__pyx_k_halo_particles_selector), 0, 0, 1, 1},
  {&__pyx_n_s_hash_vals, __pyx_k_hash_vals, sizeof(__pyx_k_hash_vals), 0, 0, 1, 1},
  {&__pyx_n_s_height, __pyx_k_height, sizeof(__pyx_k_height), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_id_offset, __pyx_k_id_offset, sizeof(__pyx_k_id_offset), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_index_offset, __pyx_k_index_offset, sizeof(__pyx_k_index_offset), 0, 0, 1, 1},
  {&__pyx_n_s_indexed_octree_subset_selector, __pyx_k_indexed_octree_subset_selector, sizeof(__pyx_k_indexed_octree_subset_selector), 0, 0, 1, 1},
  {&__pyx_n_s_indices, __pyx_k_indices, sizeof(__pyx_k_indices), 0, 0, 1, 1},
  {&__pyx_n_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 1},
  {&__pyx_n_s_init___locals_genexpr, __pyx_k_init___locals_genexpr, sizeof(__pyx_k_init___locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_int64, __pyx_k_int64, sizeof(__pyx_k_int64), 0, 0, 1, 1},
  {&__pyx_n_s_intersection_selector, __pyx_k_intersection_selector, sizeof(__pyx_k_intersection_selector), 0, 0, 1, 1},
  {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {&__pyx_kp_s_itemsize_0_for_cython_array, __pyx_k_itemsize_0_for_cython_array, sizeof(__pyx_k_itemsize_0_for_cython_array), 0, 0, 1, 0},
  {&__pyx_n_s_j, __pyx_k_j, sizeof(__pyx_k_j), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_kind, __pyx_k_kind, sizeof(__pyx_k_kind), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_left_edge, __pyx_k_left_edge, sizeof(__pyx_k_left_edge), 0, 0, 1, 1},
  {&__pyx_kp_s_left_edge_0, __pyx_k_left_edge_0, sizeof(__pyx_k_left_edge_0), 0, 0, 1, 0},
  {&__pyx_kp_s_left_edge_1, __pyx_k_left_edge_1, sizeof(__pyx_k_left_edge_1), 0, 0, 1, 0},
  {&__pyx_kp_s_left_edge_2, __pyx_k_left_edge_2, sizeof(__pyx_k_left_edge_2), 0, 0, 1, 0},
  {&__pyx_n_s_left_edges, __pyx_k_left_edges, sizeof(__pyx_k_left_edges), 0, 0, 1, 1},
  {&__pyx_n_s_levels, __pyx_k_levels, sizeof(__pyx_k_levels), 0, 0, 1, 1},
  {&__pyx_n_s_logical_or, __pyx_k_logical_or, sizeof(__pyx_k_logical_or), 0, 0, 1, 1},
  {&__pyx_n_s_loose_selection, __pyx_k_loose_selection, sizeof(__pyx_k_loose_selection), 0, 0, 1, 1},
  {&__pyx_kp_s_mag_0, __pyx_k_mag_0, sizeof(__pyx_k_mag_0), 0, 0, 1, 0},
  {&__pyx_kp_s_mag_1, __pyx_k_mag_1, sizeof(__pyx_k_mag_1), 0, 0, 1, 0},
  {&__pyx_kp_s_mag_2, __pyx_k_mag_2, sizeof(__pyx_k_mag_2), 0, 0, 1, 0},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mask, __pyx_k_mask, sizeof(__pyx_k_mask), 0, 0, 1, 1},
  {&__pyx_n_s_mask_fill, __pyx_k_mask_fill, sizeof(__pyx_k_mask_fill), 0, 0, 1, 1},
  {&__pyx_n_s_max_ind, __pyx_k_max_ind, sizeof(__pyx_k_max_ind), 0, 0, 1, 1},
  {&__pyx_n_s_max_level, __pyx_k_max_level, sizeof(__pyx_k_max_level), 0, 0, 1, 1},
  {&__pyx_n_s_memview, __pyx_k_memview, sizeof(__pyx_k_memview), 0, 0, 1, 1},
  {&__pyx_n_s_mesh, __pyx_k_mesh, sizeof(__pyx_k_mesh), 0, 0, 1, 1},
  {&__pyx_n_s_min_ind, __pyx_k_min_ind, sizeof(__pyx_k_min_ind), 0, 0, 1, 1},
  {&__pyx_n_s_min_level, __pyx_k_min_level, sizeof(__pyx_k_min_level), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_n_c, __pyx_k_n_c, sizeof(__pyx_k_n_c), 0, 0, 1, 1},
  {&__pyx_n_s_n_p, __pyx_k_n_p, sizeof(__pyx_k_n_p), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_kp_u_ndarray_is_not_C_contiguous, __pyx_k_ndarray_is_not_C_contiguous, sizeof(__pyx_k_ndarray_is_not_C_contiguous), 0, 1, 0, 0},
  {&__pyx_kp_u_ndarray_is_not_Fortran_contiguou, __pyx_k_ndarray_is_not_Fortran_contiguou, sizeof(__pyx_k_ndarray_is_not_Fortran_contiguou), 0, 1, 0, 0},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_s_neg, __pyx_k_neg, sizeof(__pyx_k_neg), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_norm_vec, __pyx_k_norm_vec, sizeof(__pyx_k_norm_vec), 0, 0, 1, 1},
  {&__pyx_kp_s_norm_vec_0, __pyx_k_norm_vec_0, sizeof(__pyx_k_norm_vec_0), 0, 0, 1, 0},
  {&__pyx_kp_s_norm_vec_1, __pyx_k_norm_vec_1, sizeof(__pyx_k_norm_vec_1), 0, 0, 1, 0},
  {&__pyx_kp_s_norm_vec_2, __pyx_k_norm_vec_2, sizeof(__pyx_k_norm_vec_2), 0, 0, 1, 0},
  {&__pyx_n_s_not, __pyx_k_not, sizeof(__pyx_k_not), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_nz, __pyx_k_nz, sizeof(__pyx_k_nz), 0, 0, 1, 1},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_obj_ids, __pyx_k_obj_ids, sizeof(__pyx_k_obj_ids), 0, 0, 1, 1},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_s_octree, __pyx_k_octree, sizeof(__pyx_k_octree), 0, 0, 1, 1},
  {&__pyx_n_s_octree_subset_selector, __pyx_k_octree_subset_selector, sizeof(__pyx_k_octree_subset_selector), 0, 0, 1, 1},
  {&__pyx_n_s_offset, __pyx_k_offset, sizeof(__pyx_k_offset), 0, 0, 1, 1},
  {&__pyx_n_s_ones, __pyx_k_ones, sizeof(__pyx_k_ones), 0, 0, 1, 1},
  {&__pyx_n_s_or, __pyx_k_or, sizeof(__pyx_k_or), 0, 0, 1, 1},
  {&__pyx_n_s_ortho_ray_selector, __pyx_k_ortho_ray_selector, sizeof(__pyx_k_ortho_ray_selector), 0, 0, 1, 1},
  {&__pyx_n_s_out, __pyx_k_out, sizeof(__pyx_k_out), 0, 0, 1, 1},
  {&__pyx_n_s_overlap_cells, __pyx_k_overlap_cells, sizeof(__pyx_k_overlap_cells), 0, 0, 1, 1},
  {&__pyx_n_s_p, __pyx_k_p, sizeof(__pyx_k_p), 0, 0, 1, 1},
  {&__pyx_kp_s_p1_0, __pyx_k_p1_0, sizeof(__pyx_k_p1_0), 0, 0, 1, 0},
  {&__pyx_kp_s_p1_1, __pyx_k_p1_1, sizeof(__pyx_k_p1_1), 0, 0, 1, 0},
  {&__pyx_kp_s_p1_2, __pyx_k_p1_2, sizeof(__pyx_k_p1_2), 0, 0, 1, 0},
  {&__pyx_kp_s_p2_0, __pyx_k_p2_0, sizeof(__pyx_k_p2_0), 0, 0, 1, 0},
  {&__pyx_kp_s_p2_1, __pyx_k_p2_1, sizeof(__pyx_k_p2_1), 0, 0, 1, 0},
  {&__pyx_kp_s_p2_2, __pyx_k_p2_2, sizeof(__pyx_k_p2_2), 0, 0, 1, 0},
  {&__pyx_kp_s_p_0, __pyx_k_p_0, sizeof(__pyx_k_p_0), 0, 0, 1, 0},
  {&__pyx_kp_s_p_1, __pyx_k_p_1, sizeof(__pyx_k_p_1), 0, 0, 1, 0},
  {&__pyx_kp_s_p_2, __pyx_k_p_2, sizeof(__pyx_k_p_2), 0, 0, 1, 0},
  {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {&__pyx_n_s_particle_indices, __pyx_k_particle_indices, sizeof(__pyx_k_particle_indices), 0, 0, 1, 1},
  {&__pyx_n_s_periodicity, __pyx_k_periodicity, sizeof(__pyx_k_periodicity), 0, 0, 1, 1},
  {&__pyx_kp_s_periodicity_0, __pyx_k_periodicity_0, sizeof(__pyx_k_periodicity_0), 0, 0, 1, 0},
  {&__pyx_kp_s_periodicity_1, __pyx_k_periodicity_1, sizeof(__pyx_k_periodicity_1), 0, 0, 1, 0},
  {&__pyx_kp_s_periodicity_2, __pyx_k_periodicity_2, sizeof(__pyx_k_periodicity_2), 0, 0, 1, 0},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_point_selector, __pyx_k_point_selector, sizeof(__pyx_k_point_selector), 0, 0, 1, 1},
  {&__pyx_n_s_points_in_cells, __pyx_k_points_in_cells, sizeof(__pyx_k_points_in_cells), 0, 0, 1, 1},
  {&__pyx_n_s_pos, __pyx_k_pos, sizeof(__pyx_k_pos), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_px, __pyx_k_px, sizeof(__pyx_k_px), 0, 0, 1, 1},
  {&__pyx_n_s_px_ax, __pyx_k_px_ax, sizeof(__pyx_k_px_ax), 0, 0, 1, 1},
  {&__pyx_n_s_py, __pyx_k_py, sizeof(__pyx_k_py), 0, 0, 1, 1},
  {&__pyx_n_s_py_ax, __pyx_k_py_ax, sizeof(__pyx_k_py_ax), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_getbuffer, __pyx_k_pyx_getbuffer, sizeof(__pyx_k_pyx_getbuffer), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Enum, __pyx_k_pyx_unpickle_Enum, sizeof(__pyx_k_pyx_unpickle_Enum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_pz, __pyx_k_pz, sizeof(__pyx_k_pz), 0, 0, 1, 1},
  {&__pyx_n_s_radius, __pyx_k_radius, sizeof(__pyx_k_radius), 0, 0, 1, 1},
  {&__pyx_n_s_radius2, __pyx_k_radius2, sizeof(__pyx_k_radius2), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_ray_selector, __pyx_k_ray_selector, sizeof(__pyx_k_ray_selector), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_region_selector, __pyx_k_region_selector, sizeof(__pyx_k_region_selector), 0, 0, 1, 1},
  {&__pyx_n_s_right_edge, __pyx_k_right_edge, sizeof(__pyx_k_right_edge), 0, 0, 1, 1},
  {&__pyx_kp_s_right_edge_0, __pyx_k_right_edge_0, sizeof(__pyx_k_right_edge_0), 0, 0, 1, 0},
  {&__pyx_kp_s_right_edge_1, __pyx_k_right_edge_1, sizeof(__pyx_k_right_edge_1), 0, 0, 1, 0},
  {&__pyx_kp_s_right_edge_2, __pyx_k_right_edge_2, sizeof(__pyx_k_right_edge_2), 0, 0, 1, 0},
  {&__pyx_n_s_right_edges, __pyx_k_right_edges, sizeof(__pyx_k_right_edges), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_n_s_select_grids, __pyx_k_select_grids, sizeof(__pyx_k_select_grids), 0, 0, 1, 1},
  {&__pyx_n_s_select_points, __pyx_k_select_points, sizeof(__pyx_k_select_points), 0, 0, 1, 1},
  {&__pyx_n_s_selector, __pyx_k_selector, sizeof(__pyx_k_selector), 0, 0, 1, 1},
  {&__pyx_n_s_selector1, __pyx_k_selector1, sizeof(__pyx_k_selector1), 0, 0, 1, 1},
  {&__pyx_n_s_selector2, __pyx_k_selector2, sizeof(__pyx_k_selector2), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_slice_selector, __pyx_k_slice_selector, sizeof(__pyx_k_slice_selector), 0, 0, 1, 1},
  {&__pyx_n_s_sphere_selector, __pyx_k_sphere_selector, sizeof(__pyx_k_sphere_selector), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_start_point, __pyx_k_start_point, sizeof(__pyx_k_start_point), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_kp_s_strided_and_direct, __pyx_k_strided_and_direct, sizeof(__pyx_k_strided_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_direct_or_indirect, __pyx_k_strided_and_direct_or_indirect, sizeof(__pyx_k_strided_and_direct_or_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_indirect, __pyx_k_strided_and_indirect, sizeof(__pyx_k_strided_and_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_to_ndarray, __pyx_k_to_ndarray, sizeof(__pyx_k_to_ndarray), 0, 0, 1, 1},
  {&__pyx_n_s_tostring, __pyx_k_tostring, sizeof(__pyx_k_tostring), 0, 0, 1, 1},
  {&__pyx_n_s_transpose, __pyx_k_transpose, sizeof(__pyx_k_transpose), 0, 0, 1, 1},
  {&__pyx_n_s_uint8, __pyx_k_uint8, sizeof(__pyx_k_uint8), 0, 0, 1, 1},
  {&__pyx_kp_s_unable_to_allocate_array_data, __pyx_k_unable_to_allocate_array_data, sizeof(__pyx_k_unable_to_allocate_array_data), 0, 0, 1, 0},
  {&__pyx_kp_s_unable_to_allocate_shape_and_str, __pyx_k_unable_to_allocate_shape_and_str, sizeof(__pyx_k_unable_to_allocate_shape_and_str), 0, 0, 1, 0},
  {&__pyx_n_s_union_selector, __pyx_k_union_selector, sizeof(__pyx_k_union_selector), 0, 0, 1, 1},
  {&__pyx_n_s_units, __pyx_k_units, sizeof(__pyx_k_units), 0, 0, 1, 1},
  {&__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_k_unknown_dtype_code_in_numpy_pxd, sizeof(__pyx_k_unknown_dtype_code_in_numpy_pxd), 0, 1, 0, 0},
  {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_vals, __pyx_k_vals, sizeof(__pyx_k_vals), 0, 0, 1, 1},
  {&__pyx_n_s_vec, __pyx_k_vec, sizeof(__pyx_k_vec), 0, 0, 1, 1},
  {&__pyx_kp_s_vec_0, __pyx_k_vec_0, sizeof(__pyx_k_vec_0), 0, 0, 1, 0},
  {&__pyx_kp_s_vec_0_0, __pyx_k_vec_0_0, sizeof(__pyx_k_vec_0_0), 0, 0, 1, 0},
  {&__pyx_kp_s_vec_0_1, __pyx_k_vec_0_1, sizeof(__pyx_k_vec_0_1), 0, 0, 1, 0},
  {&__pyx_kp_s_vec_0_2, __pyx_k_vec_0_2, sizeof(__pyx_k_vec_0_2), 0, 0, 1, 0},
  {&__pyx_kp_s_vec_1, __pyx_k_vec_1, sizeof(__pyx_k_vec_1), 0, 0, 1, 0},
  {&__pyx_kp_s_vec_1_0, __pyx_k_vec_1_0, sizeof(__pyx_k_vec_1_0), 0, 0, 1, 0},
  {&__pyx_kp_s_vec_1_1, __pyx_k_vec_1_1, sizeof(__pyx_k_vec_1_1), 0, 0, 1, 0},
  {&__pyx_kp_s_vec_1_2, __pyx_k_vec_1_2, sizeof(__pyx_k_vec_1_2), 0, 0, 1, 0},
  {&__pyx_kp_s_vec_2, __pyx_k_vec_2, sizeof(__pyx_k_vec_2), 0, 0, 1, 0},
  {&__pyx_kp_s_vec_2_0, __pyx_k_vec_2_0, sizeof(__pyx_k_vec_2_0), 0, 0, 1, 0},
  {&__pyx_kp_s_vec_2_1, __pyx_k_vec_2_1, sizeof(__pyx_k_vec_2_1), 0, 0, 1, 0},
  {&__pyx_kp_s_vec_2_2, __pyx_k_vec_2_2, sizeof(__pyx_k_vec_2_2), 0, 0, 1, 0},
  {&__pyx_n_s_view, __pyx_k_view, sizeof(__pyx_k_view), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_xor, __pyx_k_xor, sizeof(__pyx_k_xor), 0, 0, 1, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_n_s_yt_geometry_selection_routines, __pyx_k_yt_geometry_selection_routines, sizeof(__pyx_k_yt_geometry_selection_routines), 0, 0, 1, 1},
  {&__pyx_kp_s_yt_geometry_selection_routines_p, __pyx_k_yt_geometry_selection_routines_p, sizeof(__pyx_k_yt_geometry_selection_routines_p), 0, 0, 1, 0},
  {&__pyx_n_s_z, __pyx_k_z, sizeof(__pyx_k_z), 0, 0, 1, 1},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 69, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 111, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 545, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 627, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 989, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(3, 272, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(3, 1037, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(2, 148, __pyx_L1_error)
  __pyx_builtin_Ellipsis = __Pyx_GetBuiltinName(__pyx_n_s_Ellipsis); if (!__pyx_builtin_Ellipsis) __PYX_ERR(2, 404, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(2, 613, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(2, 832, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "yt/geometry/selection_routines.pyx":545
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def count_points(self, np.ndarray[floating, ndim=1] x,             # <<<<<<<<<<<<<<
 *                            np.ndarray[floating, ndim=1] y,
 *                            np.ndarray[floating, ndim=1] z,
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_No_matching_signature_found); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_Function_call_with_ambiguous_arg); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "yt/geometry/selection_routines.pyx":1522
 *             raise NotImplementedError
 *         cdef VolumeContainer vc
 *         child_mask = np.ones((1,1,1), dtype="uint8")             # <<<<<<<<<<<<<<
 *         t = np.zeros((1,1,1), dtype="float64")
 *         dt = np.zeros((1,1,1), dtype="float64") - 1
 */
  __pyx_tuple__23 = PyTuple_Pack(3, __pyx_int_1, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 1522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_tuple__23); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 1522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "yt/geometry/selection_routines.pyx":1982
 * 
 *     def _hash_vals(self):
 *         return ("always", 1,)             # <<<<<<<<<<<<<<
 * 
 * always_selector = AlwaysSelector
 */
  __pyx_tuple__37 = PyTuple_Pack(2, __pyx_n_s_always, __pyx_int_1); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 1982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__39 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__40 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__41 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__42 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__43 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__44 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__45 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "yt/geometry/selection_routines.pyx":2155
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +
 *                 ("and",) +             # <<<<<<<<<<<<<<
 *                 self.sel2._hash_vals())
 * 
 */
  __pyx_tuple__46 = PyTuple_Pack(1, __pyx_n_s_and); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 2155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__47 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__48 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);

  /* "yt/geometry/selection_routines.pyx":2199
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +
 *                 ("or",) +             # <<<<<<<<<<<<<<
 *                 self.sel2._hash_vals())
 * 
 */
  __pyx_tuple__49 = PyTuple_Pack(1, __pyx_n_s_or); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 2199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__50 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__51 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);

  /* "yt/geometry/selection_routines.pyx":2231
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +
 *                 ("not",))             # <<<<<<<<<<<<<<
 * 
 * cdef class BooleanXORSelector(BooleanSelector):
 */
  __pyx_tuple__52 = PyTuple_Pack(1, __pyx_n_s_not); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__53 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__54 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);

  /* "yt/geometry/selection_routines.pyx":2266
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +
 *                 ("xor",) +             # <<<<<<<<<<<<<<
 *                 self.sel2._hash_vals())
 * 
 */
  __pyx_tuple__55 = PyTuple_Pack(1, __pyx_n_s_xor); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(0, 2266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__56 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__57 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);

  /* "yt/geometry/selection_routines.pyx":2305
 *     def _hash_vals(self):
 *         return (self.sel1._hash_vals() +
 *                 ("neg",) +             # <<<<<<<<<<<<<<
 *                 self.sel2._hash_vals())
 * 
 */
  __pyx_tuple__58 = PyTuple_Pack(1, __pyx_n_s_neg); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(0, 2305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__59 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__60 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__61 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__62 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);

  /* "yt/geometry/selection_routines.pyx":2377
 * 
 *     def _hash_vals(self):
 *         v = ("chained_and",)             # <<<<<<<<<<<<<<
 *         for s in self.selectors:
 *             v += s._hash_vals()
 */
  __pyx_tuple__63 = PyTuple_Pack(1, __pyx_n_s_chained_and); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(0, 2377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__64 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__65 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);

  /* "yt/geometry/selection_routines.pyx":2443
 * 
 *     def _hash_vals(self):
 *         v = ("chained_or",)             # <<<<<<<<<<<<<<
 *         for s in self.selectors:
 *             v += s._hash_vals()
 */
  __pyx_tuple__66 = PyTuple_Pack(1, __pyx_n_s_chained_or); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(0, 2443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__67 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__68 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_tuple__69 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_C_contiguous); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(3, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_tuple__70 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_Fortran_contiguou); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(3, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_tuple__71 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(3, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":855
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_tuple__72 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor); if (unlikely(!__pyx_tuple__72)) __PYX_ERR(3, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__72);
  __Pyx_GIVEREF(__pyx_tuple__72);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_tuple__73 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor_2); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(3, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1037
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__74 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(3, 1037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pl/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1043
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__75 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__75)) __PYX_ERR(3, 1043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);

  /* "View.MemoryView":133
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
  __pyx_tuple__76 = PyTuple_Pack(1, __pyx_kp_s_Empty_shape_tuple_for_cython_arr); if (unlikely(!__pyx_tuple__76)) __PYX_ERR(2, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__76);
  __Pyx_GIVEREF(__pyx_tuple__76);

  /* "View.MemoryView":136
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
  __pyx_tuple__77 = PyTuple_Pack(1, __pyx_kp_s_itemsize_0_for_cython_array); if (unlikely(!__pyx_tuple__77)) __PYX_ERR(2, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__77);
  __Pyx_GIVEREF(__pyx_tuple__77);

  /* "View.MemoryView":148
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__78 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_shape_and_str); if (unlikely(!__pyx_tuple__78)) __PYX_ERR(2, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__78);
  __Pyx_GIVEREF(__pyx_tuple__78);

  /* "View.MemoryView":176
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
  __pyx_tuple__79 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_array_data); if (unlikely(!__pyx_tuple__79)) __PYX_ERR(2, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__79);
  __Pyx_GIVEREF(__pyx_tuple__79);

  /* "View.MemoryView":192
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
  __pyx_tuple__80 = PyTuple_Pack(1, __pyx_kp_s_Can_only_create_a_buffer_that_is); if (unlikely(!__pyx_tuple__80)) __PYX_ERR(2, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__80);
  __Pyx_GIVEREF(__pyx_tuple__80);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__81 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__81)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__81);
  __Pyx_GIVEREF(__pyx_tuple__81);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__82 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__82)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__82);
  __Pyx_GIVEREF(__pyx_tuple__82);

  /* "View.MemoryView":418
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
  __pyx_tuple__83 = PyTuple_Pack(1, __pyx_kp_s_Cannot_assign_to_read_only_memor); if (unlikely(!__pyx_tuple__83)) __PYX_ERR(2, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__83);
  __Pyx_GIVEREF(__pyx_tuple__83);

  /* "View.MemoryView":495
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
  __pyx_tuple__84 = PyTuple_Pack(1, __pyx_kp_s_Unable_to_convert_item_to_object); if (unlikely(!__pyx_tuple__84)) __PYX_ERR(2, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__84);
  __Pyx_GIVEREF(__pyx_tuple__84);

  /* "View.MemoryView":520
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
  __pyx_tuple__85 = PyTuple_Pack(1, __pyx_kp_s_Cannot_create_writable_memory_vi); if (unlikely(!__pyx_tuple__85)) __PYX_ERR(2, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__85);
  __Pyx_GIVEREF(__pyx_tuple__85);

  /* "View.MemoryView":570
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
  __pyx_tuple__86 = PyTuple_Pack(1, __pyx_kp_s_Buffer_view_does_not_expose_stri); if (unlikely(!__pyx_tuple__86)) __PYX_ERR(2, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);

  /* "View.MemoryView":577
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
  __pyx_tuple__87 = PyTuple_New(1); if (unlikely(!__pyx_tuple__87)) __PYX_ERR(2, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__87);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_tuple__87, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_tuple__87);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__88 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__88)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__88);
  __Pyx_GIVEREF(__pyx_tuple__88);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__89 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__89)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__89);
  __Pyx_GIVEREF(__pyx_tuple__89);

  /* "View.MemoryView":682
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */
  __pyx_slice__90 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__90)) __PYX_ERR(2, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__90);
  __Pyx_GIVEREF(__pyx_slice__90);

  /* "View.MemoryView":703
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__91 = PyTuple_Pack(1, __pyx_kp_s_Indirect_dimensions_not_supporte); if (unlikely(!__pyx_tuple__91)) __PYX_ERR(2, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__91);
  __Pyx_GIVEREF(__pyx_tuple__91);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__92 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__92)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__92);
  __Pyx_GIVEREF(__pyx_tuple__92);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__93 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__93)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__93);
  __Pyx_GIVEREF(__pyx_tuple__93);

  /* "yt/geometry/selection_routines.pyx":63
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def convert_mask_to_indices(np.ndarray[np.uint8_t, ndim=3, cast=True] mask,             # <<<<<<<<<<<<<<
 *             int count, int transpose = 0):
 *     cdef int i, j, k, cpos
 */
  __pyx_tuple__94 = PyTuple_Pack(8, __pyx_n_s_mask, __pyx_n_s_count, __pyx_n_s_transpose, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_k, __pyx_n_s_cpos, __pyx_n_s_indices); if (unlikely(!__pyx_tuple__94)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__94);
  __Pyx_GIVEREF(__pyx_tuple__94);
  __pyx_codeobj__95 = (PyObject*)__Pyx_PyCode_New(3, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__94, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_yt_geometry_selection_routines_p, __pyx_n_s_convert_mask_to_indices, 63, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__95)) __PYX_ERR(0, 63, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":102
 *     return count
 * 
 * def mask_fill(np.ndarray[np.float64_t, ndim=1] out,             # <<<<<<<<<<<<<<
 *               np.int64_t offset,
 *               np.ndarray[np.uint8_t, ndim=3, cast=True] mask,
 */
  __pyx_tuple__96 = PyTuple_Pack(4, __pyx_n_s_out, __pyx_n_s_offset, __pyx_n_s_mask, __pyx_n_s_vals); if (unlikely(!__pyx_tuple__96)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__96);
  __Pyx_GIVEREF(__pyx_tuple__96);
  __pyx_codeobj__97 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__96, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_yt_geometry_selection_routines_p, __pyx_n_s_mask_fill, 102, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__97)) __PYX_ERR(0, 102, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":545
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def count_points(self, np.ndarray[floating, ndim=1] x,             # <<<<<<<<<<<<<<
 *                            np.ndarray[floating, ndim=1] y,
 *                            np.ndarray[floating, ndim=1] z,
 */
  __pyx_tuple__98 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_x, __pyx_n_s_y, __pyx_n_s_z, __pyx_n_s_radius, __pyx_n_s_count, __pyx_n_s_i, __pyx_n_s_pos); if (unlikely(!__pyx_tuple__98)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__98);
  __Pyx_GIVEREF(__pyx_tuple__98);
  __pyx_codeobj__99 = (PyObject*)__Pyx_PyCode_New(5, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__98, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_yt_geometry_selection_routines_p, __pyx_n_s_count_points, 545, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__99)) __PYX_ERR(0, 545, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":573
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_points(self, np.ndarray[floating, ndim=1] x,             # <<<<<<<<<<<<<<
 *                             np.ndarray[floating, ndim=1] y,
 *                             np.ndarray[floating, ndim=1] z,
 */
  __pyx_tuple__100 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_x, __pyx_n_s_y, __pyx_n_s_z, __pyx_n_s_radius, __pyx_n_s_count, __pyx_n_s_i, __pyx_n_s_pos, __pyx_n_s_mask); if (unlikely(!__pyx_tuple__100)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__100);
  __Pyx_GIVEREF(__pyx_tuple__100);
  __pyx_codeobj__101 = (PyObject*)__Pyx_PyCode_New(5, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__100, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_yt_geometry_selection_routines_p, __pyx_n_s_select_points, 573, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__101)) __PYX_ERR(0, 573, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":2063
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * def points_in_cells(             # <<<<<<<<<<<<<<
 *         np.float64_t[:] cx,
 *         np.float64_t[:] cy,
 */
  __pyx_tuple__102 = PyTuple_Pack(14, __pyx_n_s_cx, __pyx_n_s_cy, __pyx_n_s_cz, __pyx_n_s_dx, __pyx_n_s_dy, __pyx_n_s_dz, __pyx_n_s_px, __pyx_n_s_py, __pyx_n_s_pz, __pyx_n_s_p, __pyx_n_s_c, __pyx_n_s_n_p, __pyx_n_s_n_c, __pyx_n_s_mask); if (unlikely(!__pyx_tuple__102)) __PYX_ERR(0, 2063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__102);
  __Pyx_GIVEREF(__pyx_tuple__102);
  __pyx_codeobj__103 = (PyObject*)__Pyx_PyCode_New(9, 0, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__102, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_yt_geometry_selection_routines_p, __pyx_n_s_points_in_cells, 2063, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__103)) __PYX_ERR(0, 2063, __pyx_L1_error)

  /* "View.MemoryView":286
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_tuple__104 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct_or_indirect); if (unlikely(!__pyx_tuple__104)) __PYX_ERR(2, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__104);
  __Pyx_GIVEREF(__pyx_tuple__104);

  /* "View.MemoryView":287
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_tuple__105 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct); if (unlikely(!__pyx_tuple__105)) __PYX_ERR(2, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__105);
  __Pyx_GIVEREF(__pyx_tuple__105);

  /* "View.MemoryView":288
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__106 = PyTuple_Pack(1, __pyx_kp_s_strided_and_indirect); if (unlikely(!__pyx_tuple__106)) __PYX_ERR(2, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__106);
  __Pyx_GIVEREF(__pyx_tuple__106);

  /* "View.MemoryView":291
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_tuple__107 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_direct); if (unlikely(!__pyx_tuple__107)) __PYX_ERR(2, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__107);
  __Pyx_GIVEREF(__pyx_tuple__107);

  /* "View.MemoryView":292
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__108 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_indirect); if (unlikely(!__pyx_tuple__108)) __PYX_ERR(2, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__108);
  __Pyx_GIVEREF(__pyx_tuple__108);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__109 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__109)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__109);
  __Pyx_GIVEREF(__pyx_tuple__109);
  __pyx_codeobj__110 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__109, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Enum, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__110)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyByteArray_Type_extend.type = (PyObject*)&PyByteArray_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_1_0 = PyFloat_FromDouble(1.0); if (unlikely(!__pyx_float_1_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_99 = PyInt_FromLong(99); if (unlikely(!__pyx_int_99)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_184977713 = PyInt_FromLong(184977713L); if (unlikely(!__pyx_int_184977713)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  generic = Py_None; Py_INCREF(Py_None);
  strided = Py_None; Py_INCREF(Py_None);
  indirect = Py_None; Py_INCREF(Py_None);
  contiguous = Py_None; Py_INCREF(Py_None);
  indirect_contiguous = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject = &__pyx_vtable_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtable_2yt_8geometry_18selection_routines_SelectorObject.recursively_visit_octs = (void (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, int, struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_recursively_visit_octs *__pyx_optional_args))__pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_recursively_visit_octs;
  __pyx_vtable_2yt_8geometry_18selection_routines_SelectorObject.visit_oct_cells = (void (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *, int, int, int))__pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_visit_oct_cells;
  __pyx_vtable_2yt_8geometry_18selection_routines_SelectorObject.select_grid = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int32_t, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid *__pyx_optional_args))__pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_select_grid;
  __pyx_vtable_2yt_8geometry_18selection_routines_SelectorObject.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_SelectorObject.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_SelectorObject.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_SelectorObject.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_select_bbox;
  __pyx_vtable_2yt_8geometry_18selection_routines_SelectorObject.fill_mask_selector = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, int *, PyArrayObject *, PyArrayObject *, int))__pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_fill_mask_selector;
  __pyx_vtable_2yt_8geometry_18selection_routines_SelectorObject.visit_grid_cells = (void (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, struct __pyx_t_2yt_8geometry_13grid_visitors_GridVisitorData *, __pyx_t_2yt_8geometry_13grid_visitors_grid_visitor_function *, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_visit_grid_cells *__pyx_optional_args))__pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_visit_grid_cells;
  __pyx_vtable_2yt_8geometry_18selection_routines_SelectorObject.difference = (__pyx_t_5numpy_float64_t (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, int))__pyx_f_2yt_8geometry_18selection_routines_14SelectorObject_difference;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_SelectorObject) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_SelectorObject.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_SelectorObject.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_SelectorObject.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_SelectorObject.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_SelectorObject.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SelectorObject, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_SelectorObject) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_SelectorObject) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject = &__pyx_type_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_AlwaysSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_AlwaysSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_AlwaysSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtable_2yt_8geometry_18selection_routines_AlwaysSelector.__pyx_base.select_grid = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int32_t, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid *__pyx_optional_args))__pyx_f_2yt_8geometry_18selection_routines_14AlwaysSelector_select_grid;
  __pyx_vtable_2yt_8geometry_18selection_routines_AlwaysSelector.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_14AlwaysSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_AlwaysSelector.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_14AlwaysSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_AlwaysSelector.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_14AlwaysSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_AlwaysSelector.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_14AlwaysSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_AlwaysSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_AlwaysSelector) < 0) __PYX_ERR(0, 1944, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_AlwaysSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_AlwaysSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_AlwaysSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_AlwaysSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_AlwaysSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_AlwaysSelector) < 0) __PYX_ERR(0, 1944, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_AlwaysSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_AlwaysSelector) < 0) __PYX_ERR(0, 1944, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_AlwaysSelector) < 0) __PYX_ERR(0, 1944, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_AlwaysSelector = &__pyx_type_2yt_8geometry_18selection_routines_AlwaysSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_OctreeSubsetSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_OctreeSubsetSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_OctreeSubsetSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtable_2yt_8geometry_18selection_routines_OctreeSubsetSelector.__pyx_base.select_grid = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int32_t, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid *__pyx_optional_args))__pyx_f_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_grid;
  __pyx_vtable_2yt_8geometry_18selection_routines_OctreeSubsetSelector.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_OctreeSubsetSelector.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_OctreeSubsetSelector.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_OctreeSubsetSelector.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_20OctreeSubsetSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_OctreeSubsetSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_OctreeSubsetSelector) < 0) __PYX_ERR(0, 1806, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_OctreeSubsetSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_OctreeSubsetSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_OctreeSubsetSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_OctreeSubsetSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_OctreeSubsetSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_OctreeSubsetSelector) < 0) __PYX_ERR(0, 1806, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_OctreeSubsetSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_OctreeSubsetSelector) < 0) __PYX_ERR(0, 1806, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_OctreeSubsetSelector) < 0) __PYX_ERR(0, 1806, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_OctreeSubsetSelector = &__pyx_type_2yt_8geometry_18selection_routines_OctreeSubsetSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_BooleanSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_type_2yt_8geometry_18selection_routines_BooleanSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_BooleanSelector) < 0) __PYX_ERR(0, 2100, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_BooleanSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_BooleanSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_BooleanSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_BooleanSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_BooleanSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanSelector) < 0) __PYX_ERR(0, 2100, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BooleanSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_BooleanSelector) < 0) __PYX_ERR(0, 2100, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_BooleanSelector) < 0) __PYX_ERR(0, 2100, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_BooleanSelector = &__pyx_type_2yt_8geometry_18selection_routines_BooleanSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_PointSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_PointSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_PointSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtable_2yt_8geometry_18selection_routines_PointSelector.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_13PointSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_PointSelector.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_13PointSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_PointSelector.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_13PointSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_PointSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_PointSelector) < 0) __PYX_ERR(0, 641, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_PointSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_PointSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_PointSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_PointSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_PointSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_PointSelector) < 0) __PYX_ERR(0, 641, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PointSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_PointSelector) < 0) __PYX_ERR(0, 641, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_PointSelector) < 0) __PYX_ERR(0, 641, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_PointSelector = &__pyx_type_2yt_8geometry_18selection_routines_PointSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_SphereSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_SphereSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_SphereSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtable_2yt_8geometry_18selection_routines_SphereSelector.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_14SphereSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_SphereSelector.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_14SphereSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_SphereSelector.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_14SphereSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_SphereSelector.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_14SphereSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_SphereSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_SphereSelector) < 0) __PYX_ERR(0, 698, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_SphereSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_SphereSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_SphereSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_SphereSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_SphereSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_SphereSelector) < 0) __PYX_ERR(0, 698, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SphereSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_SphereSelector) < 0) __PYX_ERR(0, 698, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_SphereSelector) < 0) __PYX_ERR(0, 698, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_SphereSelector = &__pyx_type_2yt_8geometry_18selection_routines_SphereSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_RegionSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_RegionSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_RegionSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtable_2yt_8geometry_18selection_routines_RegionSelector.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_14RegionSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_RegionSelector.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_14RegionSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_RegionSelector.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_14RegionSelector_select_bbox;
  __pyx_vtable_2yt_8geometry_18selection_routines_RegionSelector.__pyx_base.fill_mask_selector = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, int *, PyArrayObject *, PyArrayObject *, int))__pyx_f_2yt_8geometry_18selection_routines_14RegionSelector_fill_mask_selector;
  __pyx_type_2yt_8geometry_18selection_routines_RegionSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_RegionSelector) < 0) __PYX_ERR(0, 800, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_RegionSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_RegionSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_RegionSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_RegionSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_RegionSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_RegionSelector) < 0) __PYX_ERR(0, 800, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_RegionSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_RegionSelector) < 0) __PYX_ERR(0, 800, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_RegionSelector) < 0) __PYX_ERR(0, 800, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_RegionSelector = &__pyx_type_2yt_8geometry_18selection_routines_RegionSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_CutRegionSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_CutRegionSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_CutRegionSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtable_2yt_8geometry_18selection_routines_CutRegionSelector.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_17CutRegionSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_CutRegionSelector.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_17CutRegionSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_CutRegionSelector.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_17CutRegionSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_CutRegionSelector.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_17CutRegionSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_CutRegionSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_CutRegionSelector) < 0) __PYX_ERR(0, 961, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_CutRegionSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_CutRegionSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_CutRegionSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_CutRegionSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_CutRegionSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_CutRegionSelector) < 0) __PYX_ERR(0, 961, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CutRegionSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_CutRegionSelector) < 0) __PYX_ERR(0, 961, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_CutRegionSelector) < 0) __PYX_ERR(0, 961, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_CutRegionSelector = &__pyx_type_2yt_8geometry_18selection_routines_CutRegionSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_DiskSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_DiskSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_DiskSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtable_2yt_8geometry_18selection_routines_DiskSelector.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_12DiskSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_DiskSelector.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_12DiskSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_DiskSelector.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_12DiskSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_DiskSelector.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_12DiskSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_DiskSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_DiskSelector) < 0) __PYX_ERR(0, 995, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_DiskSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_DiskSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_DiskSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_DiskSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_DiskSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_DiskSelector) < 0) __PYX_ERR(0, 995, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_DiskSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_DiskSelector) < 0) __PYX_ERR(0, 995, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_DiskSelector) < 0) __PYX_ERR(0, 995, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_DiskSelector = &__pyx_type_2yt_8geometry_18selection_routines_DiskSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_CuttingPlaneSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_CuttingPlaneSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_CuttingPlaneSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtable_2yt_8geometry_18selection_routines_CuttingPlaneSelector.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_CuttingPlaneSelector.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_CuttingPlaneSelector.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_CuttingPlaneSelector.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_20CuttingPlaneSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_CuttingPlaneSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_CuttingPlaneSelector) < 0) __PYX_ERR(0, 1110, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_CuttingPlaneSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_CuttingPlaneSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_CuttingPlaneSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_CuttingPlaneSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_CuttingPlaneSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_CuttingPlaneSelector) < 0) __PYX_ERR(0, 1110, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CuttingPlaneSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_CuttingPlaneSelector) < 0) __PYX_ERR(0, 1110, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_CuttingPlaneSelector) < 0) __PYX_ERR(0, 1110, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_CuttingPlaneSelector = &__pyx_type_2yt_8geometry_18selection_routines_CuttingPlaneSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_SliceSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_SliceSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_SliceSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtable_2yt_8geometry_18selection_routines_SliceSelector.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_13SliceSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_SliceSelector.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_13SliceSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_SliceSelector.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_13SliceSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_SliceSelector.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_13SliceSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_SliceSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_SliceSelector) < 0) __PYX_ERR(0, 1190, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_SliceSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_SliceSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_SliceSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_SliceSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_SliceSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_SliceSelector) < 0) __PYX_ERR(0, 1190, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SliceSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_SliceSelector) < 0) __PYX_ERR(0, 1190, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_SliceSelector) < 0) __PYX_ERR(0, 1190, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_SliceSelector = &__pyx_type_2yt_8geometry_18selection_routines_SliceSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_OrthoRaySelector = &__pyx_vtable_2yt_8geometry_18selection_routines_OrthoRaySelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_OrthoRaySelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtable_2yt_8geometry_18selection_routines_OrthoRaySelector.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_16OrthoRaySelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_OrthoRaySelector.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_16OrthoRaySelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_OrthoRaySelector.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_16OrthoRaySelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_OrthoRaySelector.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_16OrthoRaySelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_OrthoRaySelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_OrthoRaySelector) < 0) __PYX_ERR(0, 1279, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_OrthoRaySelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_OrthoRaySelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_OrthoRaySelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_OrthoRaySelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_OrthoRaySelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_OrthoRaySelector) < 0) __PYX_ERR(0, 1279, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_OrthoRaySelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_OrthoRaySelector) < 0) __PYX_ERR(0, 1279, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_OrthoRaySelector) < 0) __PYX_ERR(0, 1279, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_OrthoRaySelector = &__pyx_type_2yt_8geometry_18selection_routines_OrthoRaySelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_RaySelector = &__pyx_vtable_2yt_8geometry_18selection_routines_RaySelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_RaySelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtable_2yt_8geometry_18selection_routines_RaySelector.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_11RaySelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_RaySelector.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_11RaySelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_RaySelector.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_11RaySelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_RaySelector.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_11RaySelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_RaySelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_RaySelector) < 0) __PYX_ERR(0, 1403, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_RaySelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_RaySelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_RaySelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_RaySelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_RaySelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_RaySelector) < 0) __PYX_ERR(0, 1403, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_RaySelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_RaySelector) < 0) __PYX_ERR(0, 1403, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_RaySelector) < 0) __PYX_ERR(0, 1403, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_RaySelector = &__pyx_type_2yt_8geometry_18selection_routines_RaySelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_DataCollectionSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_DataCollectionSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_DataCollectionSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_type_2yt_8geometry_18selection_routines_DataCollectionSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_DataCollectionSelector) < 0) __PYX_ERR(0, 1623, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_DataCollectionSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_DataCollectionSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_DataCollectionSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_DataCollectionSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_DataCollectionSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_DataCollectionSelector) < 0) __PYX_ERR(0, 1623, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_DataCollectionSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_DataCollectionSelector) < 0) __PYX_ERR(0, 1623, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_DataCollectionSelector) < 0) __PYX_ERR(0, 1623, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_DataCollectionSelector = &__pyx_type_2yt_8geometry_18selection_routines_DataCollectionSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_EllipsoidSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_EllipsoidSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_EllipsoidSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtable_2yt_8geometry_18selection_routines_EllipsoidSelector.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_17EllipsoidSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_EllipsoidSelector.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_17EllipsoidSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_EllipsoidSelector.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_17EllipsoidSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_EllipsoidSelector.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_17EllipsoidSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_EllipsoidSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_EllipsoidSelector) < 0) __PYX_ERR(0, 1660, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_EllipsoidSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_EllipsoidSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_EllipsoidSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_EllipsoidSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_EllipsoidSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_EllipsoidSelector) < 0) __PYX_ERR(0, 1660, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_EllipsoidSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_EllipsoidSelector) < 0) __PYX_ERR(0, 1660, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_EllipsoidSelector) < 0) __PYX_ERR(0, 1660, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_EllipsoidSelector = &__pyx_type_2yt_8geometry_18selection_routines_EllipsoidSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_GridSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_GridSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_GridSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtable_2yt_8geometry_18selection_routines_GridSelector.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_12GridSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_GridSelector.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_12GridSelector_select_point;
  __pyx_type_2yt_8geometry_18selection_routines_GridSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_GridSelector) < 0) __PYX_ERR(0, 1767, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_GridSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_GridSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_GridSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_GridSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_GridSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_GridSelector) < 0) __PYX_ERR(0, 1767, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GridSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_GridSelector) < 0) __PYX_ERR(0, 1767, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_GridSelector) < 0) __PYX_ERR(0, 1767, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_GridSelector = &__pyx_type_2yt_8geometry_18selection_routines_GridSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtable_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector.__pyx_base.select_grid = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int32_t, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid *__pyx_optional_args))__pyx_f_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_grid;
  __pyx_vtable_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_27IndexedOctreeSubsetSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector) < 0) __PYX_ERR(0, 1870, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector) < 0) __PYX_ERR(0, 1870, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IndexedOctreeSubsetSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector) < 0) __PYX_ERR(0, 1870, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector) < 0) __PYX_ERR(0, 1870, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector = &__pyx_type_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_ComposeSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_ComposeSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_ComposeSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_vtable_2yt_8geometry_18selection_routines_ComposeSelector.__pyx_base.select_grid = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int32_t, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid *__pyx_optional_args))__pyx_f_2yt_8geometry_18selection_routines_15ComposeSelector_select_grid;
  __pyx_vtable_2yt_8geometry_18selection_routines_ComposeSelector.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_15ComposeSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_ComposeSelector.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_15ComposeSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_ComposeSelector.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_15ComposeSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_ComposeSelector.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_15ComposeSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_ComposeSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_ComposeSelector) < 0) __PYX_ERR(0, 1986, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_ComposeSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_ComposeSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_ComposeSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_ComposeSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_ComposeSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_ComposeSelector) < 0) __PYX_ERR(0, 1986, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ComposeSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_ComposeSelector) < 0) __PYX_ERR(0, 1986, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_ComposeSelector) < 0) __PYX_ERR(0, 1986, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_ComposeSelector = &__pyx_type_2yt_8geometry_18selection_routines_ComposeSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_HaloParticlesSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_HaloParticlesSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_HaloParticlesSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_type_2yt_8geometry_18selection_routines_HaloParticlesSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_HaloParticlesSelector) < 0) __PYX_ERR(0, 2045, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_HaloParticlesSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_HaloParticlesSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_HaloParticlesSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_HaloParticlesSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_HaloParticlesSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_HaloParticlesSelector) < 0) __PYX_ERR(0, 2045, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_HaloParticlesSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_HaloParticlesSelector) < 0) __PYX_ERR(0, 2045, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_HaloParticlesSelector) < 0) __PYX_ERR(0, 2045, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_HaloParticlesSelector = &__pyx_type_2yt_8geometry_18selection_routines_HaloParticlesSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanANDSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_BooleanANDSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanANDSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanANDSelector.__pyx_base.__pyx_base.select_grid = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int32_t, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid *__pyx_optional_args))__pyx_f_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_grid;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanANDSelector.__pyx_base.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanANDSelector.__pyx_base.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanANDSelector.__pyx_base.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanANDSelector.__pyx_base.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_18BooleanANDSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_BooleanANDSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_BooleanSelector;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_BooleanANDSelector) < 0) __PYX_ERR(0, 2114, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_BooleanANDSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_BooleanANDSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_BooleanANDSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_BooleanANDSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_BooleanANDSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanANDSelector) < 0) __PYX_ERR(0, 2114, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BooleanANDSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_BooleanANDSelector) < 0) __PYX_ERR(0, 2114, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_BooleanANDSelector) < 0) __PYX_ERR(0, 2114, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_BooleanANDSelector = &__pyx_type_2yt_8geometry_18selection_routines_BooleanANDSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanORSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_BooleanORSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanORSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanORSelector.__pyx_base.__pyx_base.select_grid = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int32_t, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid *__pyx_optional_args))__pyx_f_2yt_8geometry_18selection_routines_17BooleanORSelector_select_grid;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanORSelector.__pyx_base.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_17BooleanORSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanORSelector.__pyx_base.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_17BooleanORSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanORSelector.__pyx_base.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_17BooleanORSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanORSelector.__pyx_base.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_17BooleanORSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_BooleanORSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_BooleanSelector;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_BooleanORSelector) < 0) __PYX_ERR(0, 2158, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_BooleanORSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_BooleanORSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_BooleanORSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_BooleanORSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_BooleanORSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanORSelector) < 0) __PYX_ERR(0, 2158, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BooleanORSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_BooleanORSelector) < 0) __PYX_ERR(0, 2158, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_BooleanORSelector) < 0) __PYX_ERR(0, 2158, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_BooleanORSelector = &__pyx_type_2yt_8geometry_18selection_routines_BooleanORSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanNOTSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_BooleanNOTSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanNOTSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanNOTSelector.__pyx_base.__pyx_base.select_grid = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int32_t, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid *__pyx_optional_args))__pyx_f_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_grid;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanNOTSelector.__pyx_base.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanNOTSelector.__pyx_base.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanNOTSelector.__pyx_base.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanNOTSelector.__pyx_base.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_18BooleanNOTSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_BooleanNOTSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_BooleanSelector;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_BooleanNOTSelector) < 0) __PYX_ERR(0, 2202, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_BooleanNOTSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_BooleanNOTSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_BooleanNOTSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_BooleanNOTSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_BooleanNOTSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanNOTSelector) < 0) __PYX_ERR(0, 2202, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BooleanNOTSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_BooleanNOTSelector) < 0) __PYX_ERR(0, 2202, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_BooleanNOTSelector) < 0) __PYX_ERR(0, 2202, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_BooleanNOTSelector = &__pyx_type_2yt_8geometry_18selection_routines_BooleanNOTSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanXORSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_BooleanXORSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanXORSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanXORSelector.__pyx_base.__pyx_base.select_grid = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int32_t, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid *__pyx_optional_args))__pyx_f_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_grid;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanXORSelector.__pyx_base.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanXORSelector.__pyx_base.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanXORSelector.__pyx_base.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanXORSelector.__pyx_base.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_18BooleanXORSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_BooleanXORSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_BooleanSelector;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_BooleanXORSelector) < 0) __PYX_ERR(0, 2233, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_BooleanXORSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_BooleanXORSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_BooleanXORSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_BooleanXORSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_BooleanXORSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanXORSelector) < 0) __PYX_ERR(0, 2233, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BooleanXORSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_BooleanXORSelector) < 0) __PYX_ERR(0, 2233, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_BooleanXORSelector) < 0) __PYX_ERR(0, 2233, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_BooleanXORSelector = &__pyx_type_2yt_8geometry_18selection_routines_BooleanXORSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanNEGSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_BooleanNEGSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanNEGSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanNEGSelector.__pyx_base.__pyx_base.select_grid = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int32_t, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid *__pyx_optional_args))__pyx_f_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_grid;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanNEGSelector.__pyx_base.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanNEGSelector.__pyx_base.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanNEGSelector.__pyx_base.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_BooleanNEGSelector.__pyx_base.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_18BooleanNEGSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_BooleanNEGSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_BooleanSelector;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_BooleanNEGSelector) < 0) __PYX_ERR(0, 2269, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_BooleanNEGSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_BooleanNEGSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_BooleanNEGSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_BooleanNEGSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_BooleanNEGSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanNEGSelector) < 0) __PYX_ERR(0, 2269, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BooleanNEGSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_BooleanNEGSelector) < 0) __PYX_ERR(0, 2269, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_BooleanNEGSelector) < 0) __PYX_ERR(0, 2269, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_BooleanNEGSelector = &__pyx_type_2yt_8geometry_18selection_routines_BooleanNEGSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_ChainedBooleanSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;
  __pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanSelector) < 0) __PYX_ERR(0, 2308, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_ChainedBooleanSelector) < 0) __PYX_ERR(0, 2308, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ChainedBooleanSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanSelector) < 0) __PYX_ERR(0, 2308, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanSelector) < 0) __PYX_ERR(0, 2308, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_ChainedBooleanSelector = &__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_ChainedBooleanSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector.__pyx_base.__pyx_base.select_grid = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int32_t, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid *__pyx_optional_args))__pyx_f_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_grid;
  __pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector.__pyx_base.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector.__pyx_base.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector.__pyx_base.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector.__pyx_base.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_25ChainedBooleanANDSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_ChainedBooleanSelector;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector) < 0) __PYX_ERR(0, 2318, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector) < 0) __PYX_ERR(0, 2318, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ChainedBooleanANDSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector) < 0) __PYX_ERR(0, 2318, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector) < 0) __PYX_ERR(0, 2318, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector = &__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector;
  __pyx_vtabptr_2yt_8geometry_18selection_routines_ChainedBooleanORSelector = &__pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanORSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanORSelector.__pyx_base = *__pyx_vtabptr_2yt_8geometry_18selection_routines_ChainedBooleanSelector;
  __pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanORSelector.__pyx_base.__pyx_base.select_grid = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int32_t, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid *__pyx_optional_args))__pyx_f_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_grid;
  __pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanORSelector.__pyx_base.__pyx_base.select_cell = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_cell;
  __pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanORSelector.__pyx_base.__pyx_base.select_point = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_point;
  __pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanORSelector.__pyx_base.__pyx_base.select_sphere = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t))__pyx_f_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_sphere;
  __pyx_vtable_2yt_8geometry_18selection_routines_ChainedBooleanORSelector.__pyx_base.__pyx_base.select_bbox = (int (*)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *))__pyx_f_2yt_8geometry_18selection_routines_24ChainedBooleanORSelector_select_bbox;
  __pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanORSelector.tp_base = __pyx_ptype_2yt_8geometry_18selection_routines_ChainedBooleanSelector;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanORSelector) < 0) __PYX_ERR(0, 2384, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanORSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanORSelector.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanORSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanORSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanORSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_18selection_routines_ChainedBooleanORSelector) < 0) __PYX_ERR(0, 2384, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ChainedBooleanORSelector, (PyObject *)&__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanORSelector) < 0) __PYX_ERR(0, 2384, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanORSelector) < 0) __PYX_ERR(0, 2384, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_ChainedBooleanORSelector = &__pyx_type_2yt_8geometry_18selection_routines_ChainedBooleanORSelector;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines___pyx_scope_struct____init__) < 0) __PYX_ERR(0, 965, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines___pyx_scope_struct____init__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines___pyx_scope_struct____init__.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines___pyx_scope_struct____init__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines___pyx_scope_struct____init__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_2yt_8geometry_18selection_routines___pyx_scope_struct____init__ = &__pyx_type_2yt_8geometry_18selection_routines___pyx_scope_struct____init__;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 968, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr = &__pyx_type_2yt_8geometry_18selection_routines___pyx_scope_struct_1_genexpr;
  __pyx_vtabptr_array = &__pyx_vtable_array;
  __pyx_vtable_array.get_memview = (PyObject *(*)(struct __pyx_array_obj *))__pyx_array_get_memview;
  if (PyType_Ready(&__pyx_type___pyx_array) < 0) __PYX_ERR(2, 105, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_array.tp_print = 0;
  #endif
  if (__Pyx_SetVtable(__pyx_type___pyx_array.tp_dict, __pyx_vtabptr_array) < 0) __PYX_ERR(2, 105, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_array) < 0) __PYX_ERR(2, 105, __pyx_L1_error)
  __pyx_array_type = &__pyx_type___pyx_array;
  if (PyType_Ready(&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(2, 279, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_MemviewEnum.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_MemviewEnum.tp_dictoffset && __pyx_type___pyx_MemviewEnum.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_MemviewEnum.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(2, 279, __pyx_L1_error)
  __pyx_MemviewEnum_type = &__pyx_type___pyx_MemviewEnum;
  __pyx_vtabptr_memoryview = &__pyx_vtable_memoryview;
  __pyx_vtable_memoryview.get_item_pointer = (char *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_get_item_pointer;
  __pyx_vtable_memoryview.is_slice = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_is_slice;
  __pyx_vtable_memoryview.setitem_slice_assignment = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_slice_assignment;
  __pyx_vtable_memoryview.setitem_slice_assign_scalar = (PyObject *(*)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_setitem_slice_assign_scalar;
  __pyx_vtable_memoryview.setitem_indexed = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_indexed;
  __pyx_vtable_memoryview.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryview_convert_item_to_object;
  __pyx_vtable_memoryview.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryview_assign_item_from_object;
  if (PyType_Ready(&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(2, 330, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryview.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryview.tp_dictoffset && __pyx_type___pyx_memoryview.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryview.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryview.tp_dict, __pyx_vtabptr_memoryview) < 0) __PYX_ERR(2, 330, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(2, 330, __pyx_L1_error)
  __pyx_memoryview_type = &__pyx_type___pyx_memoryview;
  __pyx_vtabptr__memoryviewslice = &__pyx_vtable__memoryviewslice;
  __pyx_vtable__memoryviewslice.__pyx_base = *__pyx_vtabptr_memoryview;
  __pyx_vtable__memoryviewslice.__pyx_base.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryviewslice_convert_item_to_object;
  __pyx_vtable__memoryviewslice.__pyx_base.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryviewslice_assign_item_from_object;
  __pyx_type___pyx_memoryviewslice.tp_base = __pyx_memoryview_type;
  if (PyType_Ready(&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(2, 965, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryviewslice.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryviewslice.tp_dictoffset && __pyx_type___pyx_memoryviewslice.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryviewslice.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryviewslice.tp_dict, __pyx_vtabptr__memoryviewslice) < 0) __PYX_ERR(2, 965, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(2, 965, __pyx_L1_error)
  __pyx_memoryviewslice_type = &__pyx_type___pyx_memoryviewslice;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(3, 206, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(3, 229, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(3, 233, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(3, 242, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(3, 917, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("yt.geometry.oct_visitors"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2yt_8geometry_12oct_visitors_OctVisitor = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "OctVisitor", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_OctVisitor) __PYX_ERR(5, 33, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_OctVisitor = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_OctVisitor->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_OctVisitor)) __PYX_ERR(5, 33, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "CountTotalOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_CountTotalOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalOcts) __PYX_ERR(5, 59, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_CountTotalOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountTotalOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_CountTotalOcts)) __PYX_ERR(5, 59, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalCells = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "CountTotalCells", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_CountTotalCells), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalCells) __PYX_ERR(5, 62, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_CountTotalCells = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountTotalCells*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalCells->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_CountTotalCells)) __PYX_ERR(5, 62, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_MarkOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "MarkOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_MarkOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_MarkOcts) __PYX_ERR(5, 65, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_MarkOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MarkOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_MarkOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_MarkOcts)) __PYX_ERR(5, 65, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_MaskOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "MaskOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_MaskOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_MaskOcts) __PYX_ERR(5, 69, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_MaskOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MaskOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_MaskOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_MaskOcts)) __PYX_ERR(5, 69, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_IndexOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "IndexOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_IndexOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_IndexOcts) __PYX_ERR(5, 72, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_IndexOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IndexOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_IndexOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_IndexOcts)) __PYX_ERR(5, 72, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_ICoordsOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "ICoordsOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_ICoordsOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_ICoordsOcts) __PYX_ERR(5, 75, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_ICoordsOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_ICoordsOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_ICoordsOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_ICoordsOcts)) __PYX_ERR(5, 75, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_IResOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "IResOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_IResOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_IResOcts) __PYX_ERR(5, 78, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_IResOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IResOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_IResOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_IResOcts)) __PYX_ERR(5, 78, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_FCoordsOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "FCoordsOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_FCoordsOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_FCoordsOcts) __PYX_ERR(5, 81, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_FCoordsOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FCoordsOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_FCoordsOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_FCoordsOcts)) __PYX_ERR(5, 81, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_FWidthOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "FWidthOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_FWidthOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_FWidthOcts) __PYX_ERR(5, 84, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_FWidthOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FWidthOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_FWidthOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_FWidthOcts)) __PYX_ERR(5, 84, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_CopyArrayI64 = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "CopyArrayI64", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_CopyArrayI64), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_CopyArrayI64) __PYX_ERR(5, 87, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_CopyArrayI64 = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CopyArrayI64*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_CopyArrayI64->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_CopyArrayI64)) __PYX_ERR(5, 87, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_CopyArrayF64 = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "CopyArrayF64", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_CopyArrayF64), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_CopyArrayF64) __PYX_ERR(5, 91, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_CopyArrayF64 = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CopyArrayF64*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_CopyArrayF64->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_CopyArrayF64)) __PYX_ERR(5, 91, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_IdentifyOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "IdentifyOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_IdentifyOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_IdentifyOcts) __PYX_ERR(5, 95, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_IdentifyOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IdentifyOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_IdentifyOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_IdentifyOcts)) __PYX_ERR(5, 95, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_AssignDomainInd = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "AssignDomainInd", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_AssignDomainInd), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_AssignDomainInd) __PYX_ERR(5, 98, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_AssignDomainInd = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_AssignDomainInd*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_AssignDomainInd->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_AssignDomainInd)) __PYX_ERR(5, 98, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_FillFileIndicesO = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "FillFileIndicesO", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_FillFileIndicesO), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_FillFileIndicesO) __PYX_ERR(5, 101, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_FillFileIndicesO = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FillFileIndicesO*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_FillFileIndicesO->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_FillFileIndicesO)) __PYX_ERR(5, 101, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_FillFileIndicesR = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "FillFileIndicesR", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_FillFileIndicesR), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_FillFileIndicesR) __PYX_ERR(5, 106, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_FillFileIndicesR = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FillFileIndicesR*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_FillFileIndicesR->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_FillFileIndicesR)) __PYX_ERR(5, 106, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_CountByDomain = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "CountByDomain", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_CountByDomain), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_CountByDomain) __PYX_ERR(5, 111, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_CountByDomain = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountByDomain*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_CountByDomain->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_CountByDomain)) __PYX_ERR(5, 111, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_StoreOctree = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "StoreOctree", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_StoreOctree), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_StoreOctree) __PYX_ERR(5, 114, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_StoreOctree = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_StoreOctree*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_StoreOctree->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_StoreOctree)) __PYX_ERR(5, 114, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_LoadOctree = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "LoadOctree", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_LoadOctree), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_LoadOctree) __PYX_ERR(5, 117, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_LoadOctree = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_LoadOctree*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_LoadOctree->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_LoadOctree)) __PYX_ERR(5, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("yt.utilities.lib.allocation_container"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2yt_9utilities_3lib_20allocation_container_ObjectPool = __Pyx_ImportType(__pyx_t_1, "yt.utilities.lib.allocation_container", "ObjectPool", sizeof(struct __pyx_obj_2yt_9utilities_3lib_20allocation_container_ObjectPool), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_9utilities_3lib_20allocation_container_ObjectPool) __PYX_ERR(6, 26, __pyx_L1_error)
  __pyx_vtabptr_2yt_9utilities_3lib_20allocation_container_ObjectPool = (struct __pyx_vtabstruct_2yt_9utilities_3lib_20allocation_container_ObjectPool*)__Pyx_GetVtable(__pyx_ptype_2yt_9utilities_3lib_20allocation_container_ObjectPool->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_9utilities_3lib_20allocation_container_ObjectPool)) __PYX_ERR(6, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("yt.geometry.oct_container"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2yt_8geometry_13oct_container_OctObjectPool = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_container", "OctObjectPool", sizeof(struct __pyx_obj_2yt_8geometry_13oct_container_OctObjectPool), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_13oct_container_OctObjectPool) __PYX_ERR(7, 55, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_13oct_container_OctObjectPool = (struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctObjectPool*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_13oct_container_OctObjectPool->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_13oct_container_OctObjectPool)) __PYX_ERR(7, 55, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_container", "OctreeContainer", sizeof(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer) __PYX_ERR(7, 63, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_13oct_container_OctreeContainer = (struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctreeContainer*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_13oct_container_OctreeContainer)) __PYX_ERR(7, 63, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_container", "SparseOctreeContainer", sizeof(struct __pyx_obj_2yt_8geometry_13oct_container_SparseOctreeContainer), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer) __PYX_ERR(7, 94, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_13oct_container_SparseOctreeContainer = (struct __pyx_vtabstruct_2yt_8geometry_13oct_container_SparseOctreeContainer*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_13oct_container_SparseOctreeContainer)) __PYX_ERR(7, 94, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_13oct_container_RAMSESOctreeContainer = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_container", "RAMSESOctreeContainer", sizeof(struct __pyx_obj_2yt_8geometry_13oct_container_RAMSESOctreeContainer), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_13oct_container_RAMSESOctreeContainer) __PYX_ERR(7, 102, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_13oct_container_RAMSESOctreeContainer = (struct __pyx_vtabstruct_2yt_8geometry_13oct_container_RAMSESOctreeContainer*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_13oct_container_RAMSESOctreeContainer->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_13oct_container_RAMSESOctreeContainer)) __PYX_ERR(7, 102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("yt.utilities.lib.partitioned_grid"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2yt_9utilities_3lib_16partitioned_grid_PartitionedGrid = __Pyx_ImportType(__pyx_t_1, "yt.utilities.lib.partitioned_grid", "PartitionedGrid", sizeof(struct __pyx_obj_2yt_9utilities_3lib_16partitioned_grid_PartitionedGrid), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_9utilities_3lib_16partitioned_grid_PartitionedGrid) __PYX_ERR(8, 22, __pyx_L1_error)
  __pyx_vtabptr_2yt_9utilities_3lib_16partitioned_grid_PartitionedGrid = (struct __pyx_vtabstruct_2yt_9utilities_3lib_16partitioned_grid_PartitionedGrid*)__Pyx_GetVtable(__pyx_ptype_2yt_9utilities_3lib_16partitioned_grid_PartitionedGrid->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_9utilities_3lib_16partitioned_grid_PartitionedGrid)) __PYX_ERR(8, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("yt.utilities.lib.image_samplers"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2yt_9utilities_3lib_14image_samplers_ImageSampler = __Pyx_ImportType(__pyx_t_1, "yt.utilities.lib.image_samplers", "ImageSampler", sizeof(struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_ImageSampler), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_9utilities_3lib_14image_samplers_ImageSampler) __PYX_ERR(9, 42, __pyx_L1_error)
  __pyx_vtabptr_2yt_9utilities_3lib_14image_samplers_ImageSampler = (struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_ImageSampler*)__Pyx_GetVtable(__pyx_ptype_2yt_9utilities_3lib_14image_samplers_ImageSampler->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_9utilities_3lib_14image_samplers_ImageSampler)) __PYX_ERR(9, 42, __pyx_L1_error)
  __pyx_ptype_2yt_9utilities_3lib_14image_samplers_ProjectionSampler = __Pyx_ImportType(__pyx_t_1, "yt.utilities.lib.image_samplers", "ProjectionSampler", sizeof(struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_ProjectionSampler), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_9utilities_3lib_14image_samplers_ProjectionSampler) __PYX_ERR(9, 75, __pyx_L1_error)
  __pyx_vtabptr_2yt_9utilities_3lib_14image_samplers_ProjectionSampler = (struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_ProjectionSampler*)__Pyx_GetVtable(__pyx_ptype_2yt_9utilities_3lib_14image_samplers_ProjectionSampler->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_9utilities_3lib_14image_samplers_ProjectionSampler)) __PYX_ERR(9, 75, __pyx_L1_error)
  __pyx_ptype_2yt_9utilities_3lib_14image_samplers_InterpolatedProjectionSampler = __Pyx_ImportType(__pyx_t_1, "yt.utilities.lib.image_samplers", "InterpolatedProjectionSampler", sizeof(struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_InterpolatedProjectionSampler), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_9utilities_3lib_14image_samplers_InterpolatedProjectionSampler) __PYX_ERR(9, 78, __pyx_L1_error)
  __pyx_vtabptr_2yt_9utilities_3lib_14image_samplers_InterpolatedProjectionSampler = (struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_InterpolatedProjectionSampler*)__Pyx_GetVtable(__pyx_ptype_2yt_9utilities_3lib_14image_samplers_InterpolatedProjectionSampler->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_9utilities_3lib_14image_samplers_InterpolatedProjectionSampler)) __PYX_ERR(9, 78, __pyx_L1_error)
  __pyx_ptype_2yt_9utilities_3lib_14image_samplers_VolumeRenderSampler = __Pyx_ImportType(__pyx_t_1, "yt.utilities.lib.image_samplers", "VolumeRenderSampler", sizeof(struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_VolumeRenderSampler), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_9utilities_3lib_14image_samplers_VolumeRenderSampler) __PYX_ERR(9, 83, __pyx_L1_error)
  __pyx_vtabptr_2yt_9utilities_3lib_14image_samplers_VolumeRenderSampler = (struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_VolumeRenderSampler*)__Pyx_GetVtable(__pyx_ptype_2yt_9utilities_3lib_14image_samplers_VolumeRenderSampler->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_9utilities_3lib_14image_samplers_VolumeRenderSampler)) __PYX_ERR(9, 83, __pyx_L1_error)
  __pyx_ptype_2yt_9utilities_3lib_14image_samplers_LightSourceRenderSampler = __Pyx_ImportType(__pyx_t_1, "yt.utilities.lib.image_samplers", "LightSourceRenderSampler", sizeof(struct __pyx_obj_2yt_9utilities_3lib_14image_samplers_LightSourceRenderSampler), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_9utilities_3lib_14image_samplers_LightSourceRenderSampler) __PYX_ERR(9, 89, __pyx_L1_error)
  __pyx_vtabptr_2yt_9utilities_3lib_14image_samplers_LightSourceRenderSampler = (struct __pyx_vtabstruct_2yt_9utilities_3lib_14image_samplers_LightSourceRenderSampler*)__Pyx_GetVtable(__pyx_ptype_2yt_9utilities_3lib_14image_samplers_LightSourceRenderSampler->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_9utilities_3lib_14image_samplers_LightSourceRenderSampler)) __PYX_ERR(9, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("yt.utilities.lib.bitarray"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2yt_9utilities_3lib_8bitarray_bitarray = __Pyx_ImportType(__pyx_t_1, "yt.utilities.lib.bitarray", "bitarray", sizeof(struct __pyx_obj_2yt_9utilities_3lib_8bitarray_bitarray), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_9utilities_3lib_8bitarray_bitarray) __PYX_ERR(10, 30, __pyx_L1_error)
  __pyx_vtabptr_2yt_9utilities_3lib_8bitarray_bitarray = (struct __pyx_vtabstruct_2yt_9utilities_3lib_8bitarray_bitarray*)__Pyx_GetVtable(__pyx_ptype_2yt_9utilities_3lib_8bitarray_bitarray->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_9utilities_3lib_8bitarray_bitarray)) __PYX_ERR(10, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("yt.geometry.oct_container"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "ORDER_MAX", (void **)&__pyx_vp_2yt_8geometry_13oct_container_ORDER_MAX, "int") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("yt.geometry.grid_visitors"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "check_child_masked", (void (**)(void))&__pyx_f_2yt_8geometry_13grid_visitors_check_child_masked, "__pyx_t_5numpy_uint8_t (struct __pyx_t_2yt_8geometry_13grid_visitors_GridVisitorData *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = PyImport_ImportModule("yt.utilities.lib.fnv_hash"); if (!__pyx_t_2) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "c_fnv_hash", (void (**)(void))&__pyx_f_2yt_9utilities_3lib_8fnv_hash_c_fnv_hash, "__pyx_t_5numpy_int64_t (__Pyx_memviewslice)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = PyImport_ImportModule("yt.utilities.lib.grid_traversal"); if (!__pyx_t_3) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "walk_volume", (void (**)(void))&__pyx_f_2yt_9utilities_3lib_14grid_traversal_walk_volume, "int (struct __pyx_t_2yt_9utilities_3lib_16volume_container_VolumeContainer *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_2yt_9utilities_3lib_14grid_traversal_sampler_function *, void *, struct __pyx_opt_args_2yt_9utilities_3lib_14grid_traversal_walk_volume *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initselection_routines(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initselection_routines(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_selection_routines(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_selection_routines(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_selection_routines(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_5numpy_float64_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  static PyThread_type_lock __pyx_t_7[8];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'selection_routines' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_selection_routines(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("selection_routines", __pyx_methods, __pyx_k_Geometry_selection_routines, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_yt__geometry__selection_routines) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "yt.geometry.selection_routines")) {
      if (unlikely(PyDict_SetItemString(modules, "yt.geometry.selection_routines", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_variable_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "yt/geometry/selection_routines.pyx":17
 * #-----------------------------------------------------------------------------
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * cimport cython
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":46
 * # use this as an epsilon test for grids aligned with selector
 * # define here to avoid the gil later
 * cdef np.float64_t grid_eps = np.finfo(np.float64).eps             # <<<<<<<<<<<<<<
 * grid_eps = 0.0
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_finfo); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float64); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_eps); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_2yt_8geometry_18selection_routines_grid_eps = __pyx_t_4;

  /* "yt/geometry/selection_routines.pyx":47
 * # define here to avoid the gil later
 * cdef np.float64_t grid_eps = np.finfo(np.float64).eps
 * grid_eps = 0.0             # <<<<<<<<<<<<<<
 * 
 * # These routines are separated into a couple different categories:
 */
  __pyx_v_2yt_8geometry_18selection_routines_grid_eps = 0.0;

  /* "yt/geometry/selection_routines.pyx":63
 * @cython.wraparound(False)
 * @cython.cdivision(True)
 * def convert_mask_to_indices(np.ndarray[np.uint8_t, ndim=3, cast=True] mask,             # <<<<<<<<<<<<<<
 *             int count, int transpose = 0):
 *     cdef int i, j, k, cpos
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_2yt_8geometry_18selection_routines_1convert_mask_to_indices, NULL, __pyx_n_s_yt_geometry_selection_routines); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_convert_mask_to_indices, __pyx_t_3) < 0) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":102
 *     return count
 * 
 * def mask_fill(np.ndarray[np.float64_t, ndim=1] out,             # <<<<<<<<<<<<<<
 *               np.int64_t offset,
 *               np.ndarray[np.uint8_t, ndim=3, cast=True] mask,
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_2yt_8geometry_18selection_routines_3mask_fill, NULL, __pyx_n_s_yt_geometry_selection_routines); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_mask_fill, __pyx_t_3) < 0) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "yt/geometry/selection_routines.pyx":545
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def count_points(self, np.ndarray[floating, ndim=1] x,             # <<<<<<<<<<<<<<
 *                            np.ndarray[floating, ndim=1] y,
 *                            np.ndarray[floating, ndim=1] z,
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_fuse_0__pyx_mdef_2yt_8geometry_18selection_routines_14SelectorObject_25count_points, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SelectorObject_count_points, NULL, __pyx_n_s_yt_geometry_selection_routines, __pyx_d, ((PyObject *)__pyx_codeobj__99)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_float, __pyx_t_1) < 0) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_fuse_1__pyx_mdef_2yt_8geometry_18selection_routines_14SelectorObject_27count_points, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SelectorObject_count_points, NULL, __pyx_n_s_yt_geometry_selection_routines, __pyx_d, ((PyObject *)__pyx_codeobj__99)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_double, __pyx_t_1) < 0) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_FusedFunction_New(&__pyx_mdef_2yt_8geometry_18selection_routines_14SelectorObject_15count_points, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SelectorObject_count_points, NULL, __pyx_n_s_yt_geometry_selection_routines, __pyx_d, ((PyObject *)__pyx_codeobj__99)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_1)->__signatures__ = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_dict, __pyx_n_s_count_points, __pyx_t_1) < 0) __PYX_ERR(0, 545, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/selection_routines.pyx":573
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def select_points(self, np.ndarray[floating, ndim=1] x,             # <<<<<<<<<<<<<<
 *                             np.ndarray[floating, ndim=1] y,
 *                             np.ndarray[floating, ndim=1] z,
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __pyx_FusedFunction_New(&__pyx_fuse_0__pyx_mdef_2yt_8geometry_18selection_routines_14SelectorObject_31select_points, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SelectorObject_select_points, NULL, __pyx_n_s_yt_geometry_selection_routines, __pyx_d, ((PyObject *)__pyx_codeobj__101)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_float, __pyx_t_5) < 0) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_FusedFunction_New(&__pyx_fuse_1__pyx_mdef_2yt_8geometry_18selection_routines_14SelectorObject_33select_points, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SelectorObject_select_points, NULL, __pyx_n_s_yt_geometry_selection_routines, __pyx_d, ((PyObject *)__pyx_codeobj__101)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_double, __pyx_t_5) < 0) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_FusedFunction_New(&__pyx_mdef_2yt_8geometry_18selection_routines_14SelectorObject_17select_points, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SelectorObject_select_points, NULL, __pyx_n_s_yt_geometry_selection_routines, __pyx_d, ((PyObject *)__pyx_codeobj__101)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_5)->__signatures__ = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_dict, __pyx_n_s_select_points, __pyx_t_5) < 0) __PYX_ERR(0, 573, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "yt/geometry/selection_routines.pyx":695
 *                 ("p[2]", self.p[2]))
 * 
 * point_selector = PointSelector             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_point_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_PointSelector)) < 0) __PYX_ERR(0, 695, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":798
 *                 ("center[2]", self.center[2]))
 * 
 * sphere_selector = SphereSelector             # <<<<<<<<<<<<<<
 * 
 * cdef class RegionSelector(SelectorObject):
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sphere_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_SphereSelector)) < 0) __PYX_ERR(0, 798, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":959
 *                 ("right_edge[2]", self.right_edge[2]))
 * 
 * region_selector = RegionSelector             # <<<<<<<<<<<<<<
 * 
 * cdef class CutRegionSelector(SelectorObject):
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_region_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_RegionSelector)) < 0) __PYX_ERR(0, 959, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":993
 *         return ("conditionals", t)
 * 
 * cut_region_selector = CutRegionSelector             # <<<<<<<<<<<<<<
 * 
 * cdef class DiskSelector(SelectorObject):
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cut_region_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_CutRegionSelector)) < 0) __PYX_ERR(0, 993, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":1108
 *                 ("height", self.height))
 * 
 * disk_selector = DiskSelector             # <<<<<<<<<<<<<<
 * 
 * cdef class CuttingPlaneSelector(SelectorObject):
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_disk_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_DiskSelector)) < 0) __PYX_ERR(0, 1108, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":1188
 *                 ("d", self.d))
 * 
 * cutting_selector = CuttingPlaneSelector             # <<<<<<<<<<<<<<
 * 
 * cdef class SliceSelector(SelectorObject):
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cutting_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_CuttingPlaneSelector)) < 0) __PYX_ERR(0, 1188, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":1277
 *                 ("coord", self.coord))
 * 
 * slice_selector = SliceSelector             # <<<<<<<<<<<<<<
 * 
 * cdef class OrthoRaySelector(SelectorObject):
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_slice_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_SliceSelector)) < 0) __PYX_ERR(0, 1277, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":1379
 *                 ("axis", self.axis))
 * 
 * ortho_ray_selector = OrthoRaySelector             # <<<<<<<<<<<<<<
 * 
 * cdef struct IntegrationAccumulator:
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ortho_ray_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_OrthoRaySelector)) < 0) __PYX_ERR(0, 1379, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":1621
 *                 ("vec[2]", self.vec[2]))
 * 
 * ray_selector = RaySelector             # <<<<<<<<<<<<<<
 * 
 * cdef class DataCollectionSelector(SelectorObject):
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ray_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_RaySelector)) < 0) __PYX_ERR(0, 1621, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":1658
 *         return (hash(self.obj_ids.tostring()), self.nids)
 * 
 * data_collection_selector = DataCollectionSelector             # <<<<<<<<<<<<<<
 * 
 * cdef class EllipsoidSelector(SelectorObject):
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_data_collection_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_DataCollectionSelector)) < 0) __PYX_ERR(0, 1658, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":1765
 *                 ("center[2]", self.center[2]))
 * 
 * ellipsoid_selector = EllipsoidSelector             # <<<<<<<<<<<<<<
 * 
 * cdef class GridSelector(SelectorObject):
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ellipsoid_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_EllipsoidSelector)) < 0) __PYX_ERR(0, 1765, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":1804
 *         return (self.ind,)
 * 
 * grid_selector = GridSelector             # <<<<<<<<<<<<<<
 * 
 * cdef class OctreeSubsetSelector(SelectorObject):
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_grid_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_GridSelector)) < 0) __PYX_ERR(0, 1804, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":1868
 *         return (hash(self.base_selector), self.domain_id)
 * 
 * octree_subset_selector = OctreeSubsetSelector             # <<<<<<<<<<<<<<
 * 
 * cdef class IndexedOctreeSubsetSelector(SelectorObject):
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_octree_subset_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_OctreeSubsetSelector)) < 0) __PYX_ERR(0, 1868, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":1942
 *         return (hash(self.base_selector), self.min_ind, self.max_ind)
 * 
 * indexed_octree_subset_selector = IndexedOctreeSubsetSelector             # <<<<<<<<<<<<<<
 * 
 * cdef class AlwaysSelector(SelectorObject):
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_indexed_octree_subset_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_IndexedOctreeSubsetSelector)) < 0) __PYX_ERR(0, 1942, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":1984
 *         return ("always", 1,)
 * 
 * always_selector = AlwaysSelector             # <<<<<<<<<<<<<<
 * 
 * cdef class ComposeSelector(SelectorObject):
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_always_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_AlwaysSelector)) < 0) __PYX_ERR(0, 1984, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":2043
 *         return (hash(self.selector1), hash(self.selector2))
 * 
 * compose_selector = ComposeSelector             # <<<<<<<<<<<<<<
 * 
 * cdef class HaloParticlesSelector(SelectorObject):
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_compose_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_ComposeSelector)) < 0) __PYX_ERR(0, 2043, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":2058
 *         return ("halo_particles", self.halo_id)
 * 
 * halo_particles_selector = HaloParticlesSelector             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_halo_particles_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_HaloParticlesSelector)) < 0) __PYX_ERR(0, 2058, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":2063
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * def points_in_cells(             # <<<<<<<<<<<<<<
 *         np.float64_t[:] cx,
 *         np.float64_t[:] cy,
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_2yt_8geometry_18selection_routines_5points_in_cells, NULL, __pyx_n_s_yt_geometry_selection_routines); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_points_in_cells, __pyx_t_6) < 0) __PYX_ERR(0, 2063, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "yt/geometry/selection_routines.pyx":2382
 *         return v
 * 
 * intersection_selector = ChainedBooleanANDSelector             # <<<<<<<<<<<<<<
 * 
 * cdef class ChainedBooleanORSelector(ChainedBooleanSelector):
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_intersection_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_ChainedBooleanANDSelector)) < 0) __PYX_ERR(0, 2382, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":2448
 *         return v
 * 
 * union_selector = ChainedBooleanORSelector             # <<<<<<<<<<<<<<
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_union_selector, ((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_ChainedBooleanORSelector)) < 0) __PYX_ERR(0, 2448, __pyx_L1_error)

  /* "yt/geometry/selection_routines.pyx":1
 * """             # <<<<<<<<<<<<<<
 * Geometry selection routines.
 * 
 */
  __pyx_t_6 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_6) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "View.MemoryView":209
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(array self):
 */
  __pyx_t_6 = __pyx_capsule_create(((void *)(&__pyx_array_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem((PyObject *)__pyx_array_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_6) < 0) __PYX_ERR(2, 209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_array_type);

  /* "View.MemoryView":286
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__104, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_XGOTREF(generic);
  __Pyx_DECREF_SET(generic, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "View.MemoryView":287
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__105, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_XGOTREF(strided);
  __Pyx_DECREF_SET(strided, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "View.MemoryView":288
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__106, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_XGOTREF(indirect);
  __Pyx_DECREF_SET(indirect, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "View.MemoryView":291
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__107, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_XGOTREF(contiguous);
  __Pyx_DECREF_SET(contiguous, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "View.MemoryView":292
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__108, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_XGOTREF(indirect_contiguous);
  __Pyx_DECREF_SET(indirect_contiguous, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "View.MemoryView":316
 * 
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0             # <<<<<<<<<<<<<<
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [
 *     PyThread_allocate_lock(),
 */
  __pyx_memoryview_thread_locks_used = 0;

  /* "View.MemoryView":317
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [             # <<<<<<<<<<<<<<
 *     PyThread_allocate_lock(),
 *     PyThread_allocate_lock(),
 */
  __pyx_t_7[0] = PyThread_allocate_lock();
  __pyx_t_7[1] = PyThread_allocate_lock();
  __pyx_t_7[2] = PyThread_allocate_lock();
  __pyx_t_7[3] = PyThread_allocate_lock();
  __pyx_t_7[4] = PyThread_allocate_lock();
  __pyx_t_7[5] = PyThread_allocate_lock();
  __pyx_t_7[6] = PyThread_allocate_lock();
  __pyx_t_7[7] = PyThread_allocate_lock();
  memcpy(&(__pyx_memoryview_thread_locks[0]), __pyx_t_7, sizeof(__pyx_memoryview_thread_locks[0]) * (8));

  /* "View.MemoryView":549
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_6 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem((PyObject *)__pyx_memoryview_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_6) < 0) __PYX_ERR(2, 549, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_memoryview_type);

  /* "View.MemoryView":995
 *         return self.from_object
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_6 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem((PyObject *)__pyx_memoryviewslice_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_6) < 0) __PYX_ERR(2, 995, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_memoryviewslice_type);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum, NULL, __pyx_n_s_View_MemoryView); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Enum, __pyx_t_6) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init yt.geometry.selection_routines", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init yt.geometry.selection_routines");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* BufferGetAndValidate */
  static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (unlikely(info->buf == NULL)) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}
static void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static int __Pyx__GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  buf->buf = NULL;
  if (unlikely(__Pyx_GetBuffer(obj, buf, flags) == -1)) {
    __Pyx_ZeroBuffer(buf);
    return -1;
  }
  if (unlikely(buf->ndim != nd)) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if (unlikely((size_t)buf->itemsize != dtype->size)) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_SafeReleaseBuffer(buf);
  return -1;
}

/* PyDictVersioning */
  #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
  #if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* ExtTypeTest */
  static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* BufferFallbackError */
  static void __Pyx_RaiseBufferFallbackError(void) {
  PyErr_SetString(PyExc_ValueError,
     "Buffer acquisition failed on assignment; and then reacquiring the old buffer failed too!");
}

/* PyErrFetchRestore */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
  #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyErrExceptionMatches */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
  static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
  static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* BufferIndexError */
  static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

/* GetItemInt */
  static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* MemviewSliceInit */
  static int
__Pyx_init_memviewslice(struct __pyx_memoryview_obj *memview,
                        int ndim,
                        __Pyx_memviewslice *memviewslice,
                        int memview_is_new_reference)
{
    __Pyx_RefNannyDeclarations
    int i, retval=-1;
    Py_buffer *buf = &memview->view;
    __Pyx_RefNannySetupContext("init_memviewslice", 0);
    if (unlikely(memviewslice->memview || memviewslice->data)) {
        PyErr_SetString(PyExc_ValueError,
            "memviewslice is already initialized!");
        goto fail;
    }
    if (buf->strides) {
        for (i = 0; i < ndim; i++) {
            memviewslice->strides[i] = buf->strides[i];
        }
    } else {
        Py_ssize_t stride = buf->itemsize;
        for (i = ndim - 1; i >= 0; i--) {
            memviewslice->strides[i] = stride;
            stride *= buf->shape[i];
        }
    }
    for (i = 0; i < ndim; i++) {
        memviewslice->shape[i]   = buf->shape[i];
        if (buf->suboffsets) {
            memviewslice->suboffsets[i] = buf->suboffsets[i];
        } else {
            memviewslice->suboffsets[i] = -1;
        }
    }
    memviewslice->memview = memview;
    memviewslice->data = (char *)buf->buf;
    if (__pyx_add_acquisition_count(memview) == 0 && !memview_is_new_reference) {
        Py_INCREF(memview);
    }
    retval = 0;
    goto no_fail;
fail:
    memviewslice->memview = 0;
    memviewslice->data = 0;
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}
#ifndef Py_NO_RETURN
#define Py_NO_RETURN
#endif
static void __pyx_fatalerror(const char *fmt, ...) Py_NO_RETURN {
    va_list vargs;
    char msg[200];
#ifdef HAVE_STDARG_PROTOTYPES
    va_start(vargs, fmt);
#else
    va_start(vargs);
#endif
    vsnprintf(msg, 200, fmt, vargs);
    va_end(vargs);
    Py_FatalError(msg);
}
static CYTHON_INLINE int
__pyx_add_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)++;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE int
__pyx_sub_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)--;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE void
__Pyx_INC_MEMVIEW(__Pyx_memviewslice *memslice, int have_gil, int lineno)
{
    int first_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None))
        return;
    if (unlikely(__pyx_get_slice_count(memview) < 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    first_time = __pyx_add_acquisition_count(memview) == 0;
    if (unlikely(first_time)) {
        if (have_gil) {
            Py_INCREF((PyObject *) memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_INCREF((PyObject *) memview);
            PyGILState_Release(_gilstate);
        }
    }
}
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *memslice,
                                             int have_gil, int lineno) {
    int last_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None)) {
        memslice->memview = NULL;
        return;
    }
    if (unlikely(__pyx_get_slice_count(memview) <= 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    last_time = __pyx_sub_acquisition_count(memview) == 1;
    memslice->data = NULL;
    if (unlikely(last_time)) {
        if (have_gil) {
            Py_CLEAR(memslice->memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_CLEAR(memslice->memview);
            PyGILState_Release(_gilstate);
        }
    } else {
        memslice->memview = NULL;
    }
}

/* PyCFunctionFastCall */
  #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
  #if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall2Args */
  static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* WriteUnraisableException */
  static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* DictGetItem */
  #if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* UnicodeAsUCS4 */
  static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject* x) {
   Py_ssize_t length;
   #if CYTHON_PEP393_ENABLED
   length = PyUnicode_GET_LENGTH(x);
   if (likely(length == 1)) {
       return PyUnicode_READ_CHAR(x, 0);
   }
   #else
   length = PyUnicode_GET_SIZE(x);
   if (likely(length == 1)) {
       return PyUnicode_AS_UNICODE(x)[0];
   }
   #if Py_UNICODE_SIZE == 2
   else if (PyUnicode_GET_SIZE(x) == 2) {
       Py_UCS4 high_val = PyUnicode_AS_UNICODE(x)[0];
       if (high_val >= 0xD800 && high_val <= 0xDBFF) {
           Py_UCS4 low_val = PyUnicode_AS_UNICODE(x)[1];
           if (low_val >= 0xDC00 && low_val <= 0xDFFF) {
               return 0x10000 + (((high_val & ((1<<10)-1)) << 10) | (low_val & ((1<<10)-1)));
           }
       }
   }
   #endif
   #endif
   PyErr_Format(PyExc_ValueError,
                "only single character unicode strings can be converted to Py_UCS4, "
                "got length %" CYTHON_FORMAT_SSIZE_T "d", length);
   return (Py_UCS4)-1;
}

/* object_ord */
  static long __Pyx__PyObject_Ord(PyObject* c) {
    Py_ssize_t size;
    if (PyBytes_Check(c)) {
        size = PyBytes_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyBytes_AS_STRING(c)[0];
        }
#if PY_MAJOR_VERSION < 3
    } else if (PyUnicode_Check(c)) {
        return (long)__Pyx_PyUnicode_AsPy_UCS4(c);
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    } else if (PyByteArray_Check(c)) {
        size = PyByteArray_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyByteArray_AS_STRING(c)[0];
        }
#endif
    } else {
        PyErr_Format(PyExc_TypeError,
            "ord() expected string of length 1, but %.200s found", c->ob_type->tp_name);
        return (long)(Py_UCS4)-1;
    }
    PyErr_Format(PyExc_TypeError,
        "ord() expected a character, but string of length %zd found", size);
    return (long)(Py_UCS4)-1;
}

/* SetItemInt */
  static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* IterFinish */
  static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* PyObjectCallNoArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObjectGetMethod */
  static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
  static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
  static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
  static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
  static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* RaiseNoneIterError */
  static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
  static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
  static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
  static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* UnpackUnboundCMethod */
  static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod1 */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            if (PY_VERSION_HEX >= 0x030700A0) {
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            } else {
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            }
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* None */
  static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* pyfrozenset_new */
  static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if (likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* PySetContains */
  static int __Pyx_PySet_ContainsUnhashable(PyObject *set, PyObject *key) {
    int result = -1;
    if (PySet_Check(key) && PyErr_ExceptionMatches(PyExc_TypeError)) {
        PyObject *tmpkey;
        PyErr_Clear();
        tmpkey = __Pyx_PyFrozenSet_New(key);
        if (tmpkey != NULL) {
            result = PySet_Contains(set, tmpkey);
            Py_DECREF(tmpkey);
        }
    }
    return result;
}
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq) {
    int result = PySet_Contains(set, key);
    if (unlikely(result < 0)) {
        result = __Pyx_PySet_ContainsUnhashable(set, key);
    }
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyIntBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* None */
  static CYTHON_INLINE long __Pyx_mod_long(long a, long b) {
    long r = a % b;
    r += ((r != 0) & ((r ^ b) < 0)) * b;
    return r;
}

/* PyIntBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* PyFloatBinop */
  #if !CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyFloat_DivideCObj_ZeroDivisionError(operand) if (unlikely(zerodivision_check && ((operand) == 0))) {\
    PyErr_SetString(PyExc_ZeroDivisionError, "float division by zero");\
    return NULL;\
}
static PyObject* __Pyx_PyFloat_DivideCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double a = floatval;
    double b, result;
    (void)inplace;
    (void)zerodivision_check;
    if (likely(PyFloat_CheckExact(op2))) {
        b = PyFloat_AS_DOUBLE(op2);
        __Pyx_PyFloat_DivideCObj_ZeroDivisionError(b)
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        b = (double) PyInt_AS_LONG(op2);
        __Pyx_PyFloat_DivideCObj_ZeroDivisionError(b)
    } else
    #endif
    if (likely(PyLong_CheckExact(op2))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        const digit* digits = ((PyLongObject*)op2)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op2);
        switch (size) {
            case  0: __Pyx_PyFloat_DivideCObj_ZeroDivisionError(0) break;
            case -1: b = -(double) digits[0]; break;
            case  1: b = (double) digits[0]; break;
            case -2:
            case 2:
                if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                    b = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -2)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -3:
            case 3:
                if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                    b = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -3)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -4:
            case 4:
                if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                    b = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -4)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            default:
        #else
        {
        #endif
            b = PyLong_AsDouble(op2);
            if (unlikely(b == -1.0 && PyErr_Occurred())) return NULL;
            __Pyx_PyFloat_DivideCObj_ZeroDivisionError(b)
        }
    } else {
        return (inplace ? __Pyx_PyNumber_InPlaceDivide(op1, op2) : __Pyx_PyNumber_Divide(op1, op2));
    }
        __Pyx_PyFloat_DivideCObj_ZeroDivisionError(b)
        PyFPE_START_PROTECT("divide", return NULL)
        result = a / b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* HasAttr */
    static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* GetTopmostException */
    #if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
    #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
    #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* BytesEquals */
    static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
    static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* None */
    static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t a, Py_ssize_t b) {
    Py_ssize_t q = a / b;
    Py_ssize_t r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* ObjectGetItem */
    #if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* decode_c_string */
    static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return PyUnicode_FromUnicode(NULL, 0);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* SwapException */
    #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* Import */
    static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* FastTypeChecks */
    #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* None */
    static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* None */
    static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* ImportFrom */
    static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CallNextTpTraverse */
    static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_traverse != current_tp_traverse)
        type = type->tp_base;
    while (type && type->tp_traverse == current_tp_traverse)
        type = type->tp_base;
    if (type && type->tp_traverse)
        return type->tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
    static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_clear != current_tp_clear)
        type = type->tp_base;
    while (type && type->tp_clear == current_tp_clear)
        type = type->tp_base;
    if (type && type->tp_clear)
        type->tp_clear(obj);
}

/* PyObject_GenericGetAttrNoDict */
    #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
    #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
    static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
    static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
    static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
    #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
    static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* FetchCommonType */
    static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunctionShared */
    #include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* FusedFunction */
    static PyObject *
__pyx_FusedFunction_New(PyMethodDef *ml, int flags,
                        PyObject *qualname, PyObject *closure,
                        PyObject *module, PyObject *globals,
                        PyObject *code)
{
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_FusedFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        __pyx_FusedFunctionObject *fusedfunc = (__pyx_FusedFunctionObject *) op;
        fusedfunc->__signatures__ = NULL;
        fusedfunc->type = NULL;
        fusedfunc->self = NULL;
        PyObject_GC_Track(op);
    }
    return op;
}
static void
__pyx_FusedFunction_dealloc(__pyx_FusedFunctionObject *self)
{
    PyObject_GC_UnTrack(self);
    Py_CLEAR(self->self);
    Py_CLEAR(self->type);
    Py_CLEAR(self->__signatures__);
    __Pyx__CyFunction_dealloc((__pyx_CyFunctionObject *) self);
}
static int
__pyx_FusedFunction_traverse(__pyx_FusedFunctionObject *self,
                             visitproc visit,
                             void *arg)
{
    Py_VISIT(self->self);
    Py_VISIT(self->type);
    Py_VISIT(self->__signatures__);
    return __Pyx_CyFunction_traverse((__pyx_CyFunctionObject *) self, visit, arg);
}
static int
__pyx_FusedFunction_clear(__pyx_FusedFunctionObject *self)
{
    Py_CLEAR(self->self);
    Py_CLEAR(self->type);
    Py_CLEAR(self->__signatures__);
    return __Pyx_CyFunction_clear((__pyx_CyFunctionObject *) self);
}
static PyObject *
__pyx_FusedFunction_descr_get(PyObject *self, PyObject *obj, PyObject *type)
{
    __pyx_FusedFunctionObject *func, *meth;
    func = (__pyx_FusedFunctionObject *) self;
    if (func->self || func->func.flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(self);
        return self;
    }
    if (obj == Py_None)
        obj = NULL;
    meth = (__pyx_FusedFunctionObject *) __pyx_FusedFunction_New(
                    ((PyCFunctionObject *) func)->m_ml,
                    ((__pyx_CyFunctionObject *) func)->flags,
                    ((__pyx_CyFunctionObject *) func)->func_qualname,
                    ((__pyx_CyFunctionObject *) func)->func_closure,
                    ((PyCFunctionObject *) func)->m_module,
                    ((__pyx_CyFunctionObject *) func)->func_globals,
                    ((__pyx_CyFunctionObject *) func)->func_code);
    if (!meth)
        return NULL;
    if (func->func.defaults) {
        PyObject **pydefaults;
        int i;
        if (!__Pyx_CyFunction_InitDefaults((PyObject*)meth,
                                      func->func.defaults_size,
                                      func->func.defaults_pyobjects)) {
            Py_XDECREF((PyObject*)meth);
            return NULL;
        }
        memcpy(meth->func.defaults, func->func.defaults, func->func.defaults_size);
        pydefaults = __Pyx_CyFunction_Defaults(PyObject *, meth);
        for (i = 0; i < meth->func.defaults_pyobjects; i++)
            Py_XINCREF(pydefaults[i]);
    }
    Py_XINCREF(func->func.func_classobj);
    meth->func.func_classobj = func->func.func_classobj;
    Py_XINCREF(func->__signatures__);
    meth->__signatures__ = func->__signatures__;
    Py_XINCREF(type);
    meth->type = type;
    Py_XINCREF(func->func.defaults_tuple);
    meth->func.defaults_tuple = func->func.defaults_tuple;
    if (func->func.flags & __Pyx_CYFUNCTION_CLASSMETHOD)
        obj = type;
    Py_XINCREF(obj);
    meth->self = obj;
    return (PyObject *) meth;
}
static PyObject *
_obj_to_str(PyObject *obj)
{
    if (PyType_Check(obj))
        return PyObject_GetAttr(obj, __pyx_n_s_name_2);
    else
        return PyObject_Str(obj);
}
static PyObject *
__pyx_FusedFunction_getitem(__pyx_FusedFunctionObject *self, PyObject *idx)
{
    PyObject *signature = NULL;
    PyObject *unbound_result_func;
    PyObject *result_func = NULL;
    if (self->__signatures__ == NULL) {
        PyErr_SetString(PyExc_TypeError, "Function is not fused");
        return NULL;
    }
    if (PyTuple_Check(idx)) {
        PyObject *list = PyList_New(0);
        Py_ssize_t n = PyTuple_GET_SIZE(idx);
        PyObject *sep = NULL;
        int i;
        if (unlikely(!list))
            return NULL;
        for (i = 0; i < n; i++) {
            int ret;
            PyObject *string;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyObject *item = PyTuple_GET_ITEM(idx, i);
#else
            PyObject *item = PySequence_ITEM(idx, i);  if (unlikely(!item)) goto __pyx_err;
#endif
            string = _obj_to_str(item);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
            Py_DECREF(item);
#endif
            if (unlikely(!string)) goto __pyx_err;
            ret = PyList_Append(list, string);
            Py_DECREF(string);
            if (unlikely(ret < 0)) goto __pyx_err;
        }
        sep = PyUnicode_FromString("|");
        if (likely(sep))
            signature = PyUnicode_Join(sep, list);
__pyx_err:
;
        Py_DECREF(list);
        Py_XDECREF(sep);
    } else {
        signature = _obj_to_str(idx);
    }
    if (!signature)
        return NULL;
    unbound_result_func = PyObject_GetItem(self->__signatures__, signature);
    if (unbound_result_func) {
        if (self->self || self->type) {
            __pyx_FusedFunctionObject *unbound = (__pyx_FusedFunctionObject *) unbound_result_func;
            Py_CLEAR(unbound->func.func_classobj);
            Py_XINCREF(self->func.func_classobj);
            unbound->func.func_classobj = self->func.func_classobj;
            result_func = __pyx_FusedFunction_descr_get(unbound_result_func,
                                                        self->self, self->type);
        } else {
            result_func = unbound_result_func;
            Py_INCREF(result_func);
        }
    }
    Py_DECREF(signature);
    Py_XDECREF(unbound_result_func);
    return result_func;
}
static PyObject *
__pyx_FusedFunction_callfunction(PyObject *func, PyObject *args, PyObject *kw)
{
     __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    int static_specialized = (cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD &&
                              !((__pyx_FusedFunctionObject *) func)->__signatures__);
    if (cyfunc->flags & __Pyx_CYFUNCTION_CCLASS && !static_specialized) {
        return __Pyx_CyFunction_CallAsMethod(func, args, kw);
    } else {
        return __Pyx_CyFunction_Call(func, args, kw);
    }
}
static PyObject *
__pyx_FusedFunction_call(PyObject *func, PyObject *args, PyObject *kw)
{
    __pyx_FusedFunctionObject *binding_func = (__pyx_FusedFunctionObject *) func;
    Py_ssize_t argc = PyTuple_GET_SIZE(args);
    PyObject *new_args = NULL;
    __pyx_FusedFunctionObject *new_func = NULL;
    PyObject *result = NULL;
    PyObject *self = NULL;
    int is_staticmethod = binding_func->func.flags & __Pyx_CYFUNCTION_STATICMETHOD;
    int is_classmethod = binding_func->func.flags & __Pyx_CYFUNCTION_CLASSMETHOD;
    if (binding_func->self) {
        Py_ssize_t i;
        new_args = PyTuple_New(argc + 1);
        if (!new_args)
            return NULL;
        self = binding_func->self;
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_INCREF(self);
#endif
        Py_INCREF(self);
        PyTuple_SET_ITEM(new_args, 0, self);
        for (i = 0; i < argc; i++) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyObject *item = PyTuple_GET_ITEM(args, i);
            Py_INCREF(item);
#else
            PyObject *item = PySequence_ITEM(args, i);  if (unlikely(!item)) goto bad;
#endif
            PyTuple_SET_ITEM(new_args, i + 1, item);
        }
        args = new_args;
    } else if (binding_func->type) {
        if (argc < 1) {
            PyErr_SetString(PyExc_TypeError, "Need at least one argument, 0 given.");
            return NULL;
        }
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        self = PyTuple_GET_ITEM(args, 0);
#else
        self = PySequence_ITEM(args, 0);  if (unlikely(!self)) return NULL;
#endif
    }
    if (self && !is_classmethod && !is_staticmethod) {
        int is_instance = PyObject_IsInstance(self, binding_func->type);
        if (unlikely(!is_instance)) {
            PyErr_Format(PyExc_TypeError,
                         "First argument should be of type %.200s, got %.200s.",
                         ((PyTypeObject *) binding_func->type)->tp_name,
                         self->ob_type->tp_name);
            goto bad;
        } else if (unlikely(is_instance == -1)) {
            goto bad;
        }
    }
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
    Py_XDECREF(self);
    self = NULL;
#endif
    if (binding_func->__signatures__) {
        PyObject *tup;
        if (is_staticmethod && binding_func->func.flags & __Pyx_CYFUNCTION_CCLASS) {
            tup = PyTuple_Pack(3, args,
                               kw == NULL ? Py_None : kw,
                               binding_func->func.defaults_tuple);
            if (unlikely(!tup)) goto bad;
            new_func = (__pyx_FusedFunctionObject *) __Pyx_CyFunction_CallMethod(
                func, binding_func->__signatures__, tup, NULL);
        } else {
            tup = PyTuple_Pack(4, binding_func->__signatures__, args,
                               kw == NULL ? Py_None : kw,
                               binding_func->func.defaults_tuple);
            if (unlikely(!tup)) goto bad;
            new_func = (__pyx_FusedFunctionObject *) __pyx_FusedFunction_callfunction(func, tup, NULL);
        }
        Py_DECREF(tup);
        if (unlikely(!new_func))
            goto bad;
        Py_XINCREF(binding_func->func.func_classobj);
        Py_CLEAR(new_func->func.func_classobj);
        new_func->func.func_classobj = binding_func->func.func_classobj;
        func = (PyObject *) new_func;
    }
    result = __pyx_FusedFunction_callfunction(func, args, kw);
bad:
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
    Py_XDECREF(self);
#endif
    Py_XDECREF(new_args);
    Py_XDECREF((PyObject *) new_func);
    return result;
}
static PyMemberDef __pyx_FusedFunction_members[] = {
    {(char *) "__signatures__",
     T_OBJECT,
     offsetof(__pyx_FusedFunctionObject, __signatures__),
     READONLY,
     0},
    {0, 0, 0, 0, 0},
};
static PyMappingMethods __pyx_FusedFunction_mapping_methods = {
    0,
    (binaryfunc) __pyx_FusedFunction_getitem,
    0,
};
static PyTypeObject __pyx_FusedFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "fused_cython_function",
    sizeof(__pyx_FusedFunctionObject),
    0,
    (destructor) __pyx_FusedFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    0,
    0,
    0,
    &__pyx_FusedFunction_mapping_methods,
    0,
    (ternaryfunc) __pyx_FusedFunction_call,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __pyx_FusedFunction_traverse,
    (inquiry) __pyx_FusedFunction_clear,
    0,
    0,
    0,
    0,
    0,
    __pyx_FusedFunction_members,
    __pyx_CyFunction_getsets,
    &__pyx_CyFunctionType_type,
    0,
    __pyx_FusedFunction_descr_get,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_FusedFunction_init(void) {
    __pyx_FusedFunctionType_type.tp_base = __pyx_CyFunctionType;
    __pyx_FusedFunctionType = __Pyx_FetchCommonType(&__pyx_FusedFunctionType_type);
    if (__pyx_FusedFunctionType == NULL) {
        return -1;
    }
    return 0;
}

/* CLineInTraceback */
    #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
    static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
    #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) return __pyx_pw_5numpy_7ndarray_1__getbuffer__(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_array_type)) return __pyx_array_getbuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_memoryview_type)) return __pyx_memoryview_getbuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
        else if (__Pyx_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) __pyx_pw_5numpy_7ndarray_3__releasebuffer__(obj, view);
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


    /* MemviewSliceIsContig */
    static int
__pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim)
{
    int i, index, step, start;
    Py_ssize_t itemsize = mvs.memview->view.itemsize;
    if (order == 'F') {
        step = 1;
        start = 0;
    } else {
        step = -1;
        start = ndim - 1;
    }
    for (i = 0; i < ndim; i++) {
        index = start + step * i;
        if (mvs.suboffsets[index] >= 0 || mvs.strides[index] != itemsize)
            return 0;
        itemsize *= mvs.shape[index];
    }
    return 1;
}

/* OverlappingSlices */
    static void
__pyx_get_array_memory_extents(__Pyx_memviewslice *slice,
                               void **out_start, void **out_end,
                               int ndim, size_t itemsize)
{
    char *start, *end;
    int i;
    start = end = slice->data;
    for (i = 0; i < ndim; i++) {
        Py_ssize_t stride = slice->strides[i];
        Py_ssize_t extent = slice->shape[i];
        if (extent == 0) {
            *out_start = *out_end = start;
            return;
        } else {
            if (stride > 0)
                end += stride * (extent - 1);
            else
                start += stride * (extent - 1);
        }
    }
    *out_start = start;
    *out_end = end + itemsize;
}
static int
__pyx_slices_overlap(__Pyx_memviewslice *slice1,
                     __Pyx_memviewslice *slice2,
                     int ndim, size_t itemsize)
{
    void *start1, *end1, *start2, *end2;
    __pyx_get_array_memory_extents(slice1, &start1, &end1, ndim, itemsize);
    __pyx_get_array_memory_extents(slice2, &start2, &end2, ndim, itemsize);
    return (start1 < end2) && (start2 < end1);
}

/* Capsule */
    static CYTHON_INLINE PyObject *
__pyx_capsule_create(void *p, CYTHON_UNUSED const char *sig)
{
    PyObject *cobj;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, NULL);
#else
    cobj = PyCObject_FromVoidPtr(p, NULL);
#endif
    return cobj;
}

/* TypeInfoCompare */
    static int
__pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b)
{
    int i;
    if (!a || !b)
        return 0;
    if (a == b)
        return 1;
    if (a->size != b->size || a->typegroup != b->typegroup ||
            a->is_unsigned != b->is_unsigned || a->ndim != b->ndim) {
        if (a->typegroup == 'H' || b->typegroup == 'H') {
            return a->size == b->size;
        } else {
            return 0;
        }
    }
    if (a->ndim) {
        for (i = 0; i < a->ndim; i++)
            if (a->arraysize[i] != b->arraysize[i])
                return 0;
    }
    if (a->typegroup == 'S') {
        if (a->flags != b->flags)
            return 0;
        if (a->fields || b->fields) {
            if (!(a->fields && b->fields))
                return 0;
            for (i = 0; a->fields[i].type && b->fields[i].type; i++) {
                __Pyx_StructField *field_a = a->fields + i;
                __Pyx_StructField *field_b = b->fields + i;
                if (field_a->offset != field_b->offset ||
                    !__pyx_typeinfo_cmp(field_a->type, field_b->type))
                    return 0;
            }
            return !a->fields[i].type && !b->fields[i].type;
        }
    }
    return 1;
}

/* MemviewSliceValidateAndInit */
    static int
__pyx_check_strides(Py_buffer *buf, int dim, int ndim, int spec)
{
    if (buf->shape[dim] <= 1)
        return 1;
    if (buf->strides) {
        if (spec & __Pyx_MEMVIEW_CONTIG) {
            if (spec & (__Pyx_MEMVIEW_PTR|__Pyx_MEMVIEW_FULL)) {
                if (unlikely(buf->strides[dim] != sizeof(void *))) {
                    PyErr_Format(PyExc_ValueError,
                                 "Buffer is not indirectly contiguous "
                                 "in dimension %d.", dim);
                    goto fail;
                }
            } else if (unlikely(buf->strides[dim] != buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
        if (spec & __Pyx_MEMVIEW_FOLLOW) {
            Py_ssize_t stride = buf->strides[dim];
            if (stride < 0)
                stride = -stride;
            if (unlikely(stride < buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
    } else {
        if (unlikely(spec & __Pyx_MEMVIEW_CONTIG && dim != ndim - 1)) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not contiguous in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(spec & (__Pyx_MEMVIEW_PTR))) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not indirect in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(buf->suboffsets)) {
            PyErr_SetString(PyExc_ValueError,
                            "Buffer exposes suboffsets but no strides");
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_check_suboffsets(Py_buffer *buf, int dim, CYTHON_UNUSED int ndim, int spec)
{
    if (spec & __Pyx_MEMVIEW_DIRECT) {
        if (unlikely(buf->suboffsets && buf->suboffsets[dim] >= 0)) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer not compatible with direct access "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    if (spec & __Pyx_MEMVIEW_PTR) {
        if (unlikely(!buf->suboffsets || (buf->suboffsets[dim] < 0))) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer is not indirectly accessible "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_verify_contig(Py_buffer *buf, int ndim, int c_or_f_flag)
{
    int i;
    if (c_or_f_flag & __Pyx_IS_F_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = 0; i < ndim; i++) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not fortran contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    } else if (c_or_f_flag & __Pyx_IS_C_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = ndim - 1; i >- 1; i--) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not C contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    }
    return 1;
fail:
    return 0;
}
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj)
{
    struct __pyx_memoryview_obj *memview, *new_memview;
    __Pyx_RefNannyDeclarations
    Py_buffer *buf;
    int i, spec = 0, retval = -1;
    __Pyx_BufFmt_Context ctx;
    int from_memoryview = __pyx_memoryview_check(original_obj);
    __Pyx_RefNannySetupContext("ValidateAndInit_memviewslice", 0);
    if (from_memoryview && __pyx_typeinfo_cmp(dtype, ((struct __pyx_memoryview_obj *)
                                                            original_obj)->typeinfo)) {
        memview = (struct __pyx_memoryview_obj *) original_obj;
        new_memview = NULL;
    } else {
        memview = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                            original_obj, buf_flags, 0, dtype);
        new_memview = memview;
        if (unlikely(!memview))
            goto fail;
    }
    buf = &memview->view;
    if (unlikely(buf->ndim != ndim)) {
        PyErr_Format(PyExc_ValueError,
                "Buffer has wrong number of dimensions (expected %d, got %d)",
                ndim, buf->ndim);
        goto fail;
    }
    if (new_memview) {
        __Pyx_BufFmt_Init(&ctx, stack, dtype);
        if (unlikely(!__Pyx_BufFmt_CheckString(&ctx, buf->format))) goto fail;
    }
    if (unlikely((unsigned) buf->itemsize != dtype->size)) {
        PyErr_Format(PyExc_ValueError,
                     "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "u byte%s) "
                     "does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "u byte%s)",
                     buf->itemsize,
                     (buf->itemsize > 1) ? "s" : "",
                     dtype->name,
                     dtype->size,
                     (dtype->size > 1) ? "s" : "");
        goto fail;
    }
    for (i = 0; i < ndim; i++) {
        spec = axes_specs[i];
        if (unlikely(!__pyx_check_strides(buf, i, ndim, spec)))
            goto fail;
        if (unlikely(!__pyx_check_suboffsets(buf, i, ndim, spec)))
            goto fail;
    }
    if (unlikely(buf->strides && !__pyx_verify_contig(buf, ndim, c_or_f_flag)))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview, ndim, memviewslice,
                                         new_memview != NULL) == -1)) {
        goto fail;
    }
    retval = 0;
    goto no_fail;
fail:
    Py_XDECREF(new_memview);
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}

/* ObjectToMemviewSlice */
    static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_float(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_float, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
    static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* CIntFromPyVerify */
    #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* ObjectToMemviewSlice */
    static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value) {
    const Py_intptr_t neg_one = (Py_intptr_t) ((Py_intptr_t) 0 - (Py_intptr_t) 1), const_zero = (Py_intptr_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(Py_intptr_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(Py_intptr_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(Py_intptr_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int64(npy_int64 value) {
    const npy_int64 neg_one = (npy_int64) ((npy_int64) 0 - (npy_int64) 1), const_zero = (npy_int64) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_int64) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_int64) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int64) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_int64) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int64) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_int64),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint64(npy_uint64 value) {
    const npy_uint64 neg_one = (npy_uint64) ((npy_uint64) 0 - (npy_uint64) 1), const_zero = (npy_uint64) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_uint64) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_uint64) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint64) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_uint64) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint64) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_uint64),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int32(npy_int32 value) {
    const npy_int32 neg_one = (npy_int32) ((npy_int32) 0 - (npy_int32) 1), const_zero = (npy_int32) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_int32) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_int32) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int32) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_int32) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int32) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_int32),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint8(npy_uint8 value) {
    const npy_uint8 neg_one = (npy_uint8) ((npy_uint8) 0 - (npy_uint8) 1), const_zero = (npy_uint8) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_uint8) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_uint8) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint8) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_uint8) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint8) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_uint8),
                                     little, !is_unsigned);
    }
}

/* Print */
    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static PyObject *__Pyx_GetStdout(void) {
    PyObject *f = PySys_GetObject((char *)"stdout");
    if (!f) {
        PyErr_SetString(PyExc_RuntimeError, "lost sys.stdout");
    }
    return f;
}
static int __Pyx_Print(PyObject* f, PyObject *arg_tuple, int newline) {
    int i;
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    for (i=0; i < PyTuple_GET_SIZE(arg_tuple); i++) {
        PyObject* v;
        if (PyFile_SoftSpace(f, 1)) {
            if (PyFile_WriteString(" ", f) < 0)
                goto error;
        }
        v = PyTuple_GET_ITEM(arg_tuple, i);
        if (PyFile_WriteObject(v, f, Py_PRINT_RAW) < 0)
            goto error;
        if (PyString_Check(v)) {
            char *s = PyString_AsString(v);
            Py_ssize_t len = PyString_Size(v);
            if (len > 0) {
                switch (s[len-1]) {
                    case ' ': break;
                    case '\f': case '\r': case '\n': case '\t': case '\v':
                        PyFile_SoftSpace(f, 0);
                        break;
                    default:  break;
                }
            }
        }
    }
    if (newline) {
        if (PyFile_WriteString("\n", f) < 0)
            goto error;
        PyFile_SoftSpace(f, 0);
    }
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
}
#else
static int __Pyx_Print(PyObject* stream, PyObject *arg_tuple, int newline) {
    PyObject* kwargs = 0;
    PyObject* result = 0;
    PyObject* end_string;
    if (unlikely(!__pyx_print)) {
        __pyx_print = PyObject_GetAttr(__pyx_b, __pyx_n_s_print);
        if (!__pyx_print)
            return -1;
    }
    if (stream) {
        kwargs = PyDict_New();
        if (unlikely(!kwargs))
            return -1;
        if (unlikely(PyDict_SetItem(kwargs, __pyx_n_s_file, stream) < 0))
            goto bad;
        if (!newline) {
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                goto bad;
            if (PyDict_SetItem(kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                goto bad;
            }
            Py_DECREF(end_string);
        }
    } else if (!newline) {
        if (unlikely(!__pyx_print_kwargs)) {
            __pyx_print_kwargs = PyDict_New();
            if (unlikely(!__pyx_print_kwargs))
                return -1;
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                return -1;
            if (PyDict_SetItem(__pyx_print_kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                return -1;
            }
            Py_DECREF(end_string);
        }
        kwargs = __pyx_print_kwargs;
    }
    result = PyObject_Call(__pyx_print, arg_tuple, kwargs);
    if (unlikely(kwargs) && (kwargs != __pyx_print_kwargs))
        Py_DECREF(kwargs);
    if (!result)
        return -1;
    Py_DECREF(result);
    return 0;
bad:
    if (kwargs != __pyx_print_kwargs)
        Py_XDECREF(kwargs);
    return -1;
}
#endif

/* MemviewDtypeToObject */
    static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_float64_t(const char *itemp) {
    return (PyObject *) PyFloat_FromDouble(*(__pyx_t_5numpy_float64_t *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_float64_t(const char *itemp, PyObject *obj) {
    __pyx_t_5numpy_float64_t value = __pyx_PyFloat_AsDouble(obj);
    if ((value == ((npy_float64)-1)) && PyErr_Occurred())
        return 0;
    *(__pyx_t_5numpy_float64_t *) itemp = value;
    return 1;
}

/* Declarations */
    #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
    #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
    #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
    #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value) {
    const enum NPY_TYPES neg_one = (enum NPY_TYPES) ((enum NPY_TYPES) 0 - (enum NPY_TYPES) 1), const_zero = (enum NPY_TYPES) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum NPY_TYPES) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum NPY_TYPES) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum NPY_TYPES),
                                     little, !is_unsigned);
    }
}

/* MemviewSliceCopyTemplate */
    static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object)
{
    __Pyx_RefNannyDeclarations
    int i;
    __Pyx_memviewslice new_mvs = { 0, 0, { 0 }, { 0 }, { 0 } };
    struct __pyx_memoryview_obj *from_memview = from_mvs->memview;
    Py_buffer *buf = &from_memview->view;
    PyObject *shape_tuple = NULL;
    PyObject *temp_int = NULL;
    struct __pyx_array_obj *array_obj = NULL;
    struct __pyx_memoryview_obj *memview_obj = NULL;
    __Pyx_RefNannySetupContext("__pyx_memoryview_copy_new_contig", 0);
    for (i = 0; i < ndim; i++) {
        if (unlikely(from_mvs->suboffsets[i] >= 0)) {
            PyErr_Format(PyExc_ValueError, "Cannot copy memoryview slice with "
                                           "indirect dimensions (axis %d)", i);
            goto fail;
        }
    }
    shape_tuple = PyTuple_New(ndim);
    if (unlikely(!shape_tuple)) {
        goto fail;
    }
    __Pyx_GOTREF(shape_tuple);
    for(i = 0; i < ndim; i++) {
        temp_int = PyInt_FromSsize_t(from_mvs->shape[i]);
        if(unlikely(!temp_int)) {
            goto fail;
        } else {
            PyTuple_SET_ITEM(shape_tuple, i, temp_int);
            temp_int = NULL;
        }
    }
    array_obj = __pyx_array_new(shape_tuple, sizeof_dtype, buf->format, (char *) mode, NULL);
    if (unlikely(!array_obj)) {
        goto fail;
    }
    __Pyx_GOTREF(array_obj);
    memview_obj = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                    (PyObject *) array_obj, contig_flag,
                                    dtype_is_object,
                                    from_mvs->memview->typeinfo);
    if (unlikely(!memview_obj))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview_obj, ndim, &new_mvs, 1) < 0))
        goto fail;
    if (unlikely(__pyx_memoryview_copy_contents(*from_mvs, new_mvs, ndim, ndim,
                                                dtype_is_object) < 0))
        goto fail;
    goto no_fail;
fail:
    __Pyx_XDECREF(new_mvs.memview);
    new_mvs.memview = NULL;
    new_mvs.data = NULL;
no_fail:
    __Pyx_XDECREF(shape_tuple);
    __Pyx_XDECREF(temp_int);
    __Pyx_XDECREF(array_obj);
    __Pyx_RefNannyFinishContext();
    return new_mvs;
}

/* CIntFromPy */
    static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
    static CYTHON_INLINE npy_int64 __Pyx_PyInt_As_npy_int64(PyObject *x) {
    const npy_int64 neg_one = (npy_int64) ((npy_int64) 0 - (npy_int64) 1), const_zero = (npy_int64) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_int64) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_int64, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_int64) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int64) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_int64, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_int64) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) >= 2 * PyLong_SHIFT) {
                            return (npy_int64) (((((npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int64) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) >= 3 * PyLong_SHIFT) {
                            return (npy_int64) (((((((npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int64) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) >= 4 * PyLong_SHIFT) {
                            return (npy_int64) (((((((((npy_int64)digits[3]) << PyLong_SHIFT) | (npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_int64) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_int64) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int64) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int64) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_int64, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_int64,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_int64) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int64) (((npy_int64)-1)*(((((npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_int64) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int64) ((((((npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_int64) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int64) (((npy_int64)-1)*(((((((npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int64) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int64) ((((((((npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_int64) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int64) (((npy_int64)-1)*(((((((((npy_int64)digits[3]) << PyLong_SHIFT) | (npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int64) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int64) ((((((((((npy_int64)digits[3]) << PyLong_SHIFT) | (npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_int64) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int64) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_int64 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_int64) -1;
        }
    } else {
        npy_int64 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_int64) -1;
        val = __Pyx_PyInt_As_npy_int64(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_int64");
    return (npy_int64) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_int64");
    return (npy_int64) -1;
}

/* CIntFromPy */
    static CYTHON_INLINE npy_int32 __Pyx_PyInt_As_npy_int32(PyObject *x) {
    const npy_int32 neg_one = (npy_int32) ((npy_int32) 0 - (npy_int32) 1), const_zero = (npy_int32) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_int32) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_int32, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_int32) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int32) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_int32, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_int32) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) >= 2 * PyLong_SHIFT) {
                            return (npy_int32) (((((npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int32) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) >= 3 * PyLong_SHIFT) {
                            return (npy_int32) (((((((npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int32) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) >= 4 * PyLong_SHIFT) {
                            return (npy_int32) (((((((((npy_int32)digits[3]) << PyLong_SHIFT) | (npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_int32) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_int32) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int32, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int32) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int32, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int32) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_int32, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_int32,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_int32) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int32) (((npy_int32)-1)*(((((npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_int32) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int32) ((((((npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_int32) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int32) (((npy_int32)-1)*(((((((npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int32) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int32) ((((((((npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_int32) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int32) (((npy_int32)-1)*(((((((((npy_int32)digits[3]) << PyLong_SHIFT) | (npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int32) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int32) ((((((((((npy_int32)digits[3]) << PyLong_SHIFT) | (npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_int32) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int32, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int32) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int32, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_int32 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_int32) -1;
        }
    } else {
        npy_int32 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_int32) -1;
        val = __Pyx_PyInt_As_npy_int32(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_int32");
    return (npy_int32) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_int32");
    return (npy_int32) -1;
}

/* CIntFromPy */
    static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
    static CYTHON_INLINE npy_uint8 __Pyx_PyInt_As_npy_uint8(PyObject *x) {
    const npy_uint8 neg_one = (npy_uint8) ((npy_uint8) 0 - (npy_uint8) 1), const_zero = (npy_uint8) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_uint8) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_uint8, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_uint8) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_uint8) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_uint8, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_uint8) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) >= 2 * PyLong_SHIFT) {
                            return (npy_uint8) (((((npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_uint8) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) >= 3 * PyLong_SHIFT) {
                            return (npy_uint8) (((((((npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_uint8) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) >= 4 * PyLong_SHIFT) {
                            return (npy_uint8) (((((((((npy_uint8)digits[3]) << PyLong_SHIFT) | (npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_uint8) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_uint8) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint8, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_uint8) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint8, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_uint8) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_uint8, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_uint8,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_uint8) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_uint8) (((npy_uint8)-1)*(((((npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_uint8) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_uint8) ((((((npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_uint8) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_uint8) (((npy_uint8)-1)*(((((((npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_uint8) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_uint8) ((((((((npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_uint8) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_uint8) (((npy_uint8)-1)*(((((((((npy_uint8)digits[3]) << PyLong_SHIFT) | (npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_uint8) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_uint8) ((((((((((npy_uint8)digits[3]) << PyLong_SHIFT) | (npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_uint8) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint8, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_uint8) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint8, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_uint8 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_uint8) -1;
        }
    } else {
        npy_uint8 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_uint8) -1;
        val = __Pyx_PyInt_As_npy_uint8(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_uint8");
    return (npy_uint8) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_uint8");
    return (npy_uint8) -1;
}

/* CIntFromPy */
    static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_As_Py_intptr_t(PyObject *x) {
    const Py_intptr_t neg_one = (Py_intptr_t) ((Py_intptr_t) 0 - (Py_intptr_t) 1), const_zero = (Py_intptr_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(Py_intptr_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (Py_intptr_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (Py_intptr_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(Py_intptr_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(Py_intptr_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(Py_intptr_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(Py_intptr_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (Py_intptr_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(Py_intptr_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(Py_intptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (Py_intptr_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(Py_intptr_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(Py_intptr_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(Py_intptr_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(Py_intptr_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(Py_intptr_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(Py_intptr_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(Py_intptr_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(Py_intptr_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            Py_intptr_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (Py_intptr_t) -1;
        }
    } else {
        Py_intptr_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (Py_intptr_t) -1;
        val = __Pyx_PyInt_As_Py_intptr_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to Py_intptr_t");
    return (Py_intptr_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to Py_intptr_t");
    return (Py_intptr_t) -1;
}

/* CIntFromPy */
    static CYTHON_INLINE npy_uint64 __Pyx_PyInt_As_npy_uint64(PyObject *x) {
    const npy_uint64 neg_one = (npy_uint64) ((npy_uint64) 0 - (npy_uint64) 1), const_zero = (npy_uint64) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_uint64) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_uint64, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_uint64) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_uint64) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_uint64, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_uint64) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) >= 2 * PyLong_SHIFT) {
                            return (npy_uint64) (((((npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_uint64) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) >= 3 * PyLong_SHIFT) {
                            return (npy_uint64) (((((((npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_uint64) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) >= 4 * PyLong_SHIFT) {
                            return (npy_uint64) (((((((((npy_uint64)digits[3]) << PyLong_SHIFT) | (npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_uint64) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_uint64) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint64, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_uint64) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint64, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_uint64) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_uint64, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_uint64,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_uint64) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_uint64) (((npy_uint64)-1)*(((((npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_uint64) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_uint64) ((((((npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_uint64) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_uint64) (((npy_uint64)-1)*(((((((npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_uint64) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_uint64) ((((((((npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_uint64) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_uint64) (((npy_uint64)-1)*(((((((((npy_uint64)digits[3]) << PyLong_SHIFT) | (npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_uint64) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_uint64) ((((((((((npy_uint64)digits[3]) << PyLong_SHIFT) | (npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_uint64) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint64, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_uint64) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint64, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_uint64 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_uint64) -1;
        }
    } else {
        npy_uint64 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_uint64) -1;
        val = __Pyx_PyInt_As_npy_uint64(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_uint64");
    return (npy_uint64) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_uint64");
    return (npy_uint64) -1;
}

/* BytesContains */
    static CYTHON_INLINE int __Pyx_BytesContains(PyObject* bytes, char character) {
    const Py_ssize_t length = PyBytes_GET_SIZE(bytes);
    char* char_start = PyBytes_AS_STRING(bytes);
    return memchr(char_start, (unsigned char)character, (size_t)length) != NULL;
}

/* ImportNumPyArray */
    static PyObject* __Pyx__ImportNumPyArray(void) {
    PyObject *numpy_module, *ndarray_object = NULL;
    numpy_module = __Pyx_Import(__pyx_n_s_numpy, NULL, 0);
    if (likely(numpy_module)) {
        ndarray_object = PyObject_GetAttrString(numpy_module, "ndarray");
        Py_DECREF(numpy_module);
    }
    if (unlikely(!ndarray_object)) {
        PyErr_Clear();
    }
    if (unlikely(!ndarray_object || !PyObject_TypeCheck(ndarray_object, &PyType_Type))) {
        Py_XDECREF(ndarray_object);
        Py_INCREF(Py_None);
        ndarray_object = Py_None;
    }
    return ndarray_object;
}
static CYTHON_INLINE PyObject* __Pyx_ImportNumPyArrayTypeIfAvailable(void) {
    if (unlikely(!__pyx_numpy_ndarray)) {
        __pyx_numpy_ndarray = __Pyx__ImportNumPyArray();
    }
    Py_INCREF(__pyx_numpy_ndarray);
    return __pyx_numpy_ndarray;
}

/* CIntFromPy */
    static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
    const char neg_one = (char) ((char) 0 - (char) 1), const_zero = (char) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(char, digit, digits[0])
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 2 * PyLong_SHIFT) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 3 * PyLong_SHIFT) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 4 * PyLong_SHIFT) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* ObjectToMemviewSlice */
    static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_unsigned_char(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_unsigned_char, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* PyObjectCallMethod1 */
    static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
    #include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
    static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
    #ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
    static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
    static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* VoidPtrImport */
    #ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* FunctionImport */
    #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
    static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
